/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("cloud", ["dataconv", "mysql", "node-uuid", "emailjs", "express"], function(dataconv, mysql, uuid, emailjs, express) {
    'use strict';
    
    const exports = {};

    const pickRand = lst => lst[(lst.length() * Math.random())|0];
    
    const rejectEmpty = msg => rows => (rows.length ? Promise.resolve(rows) : Promise.reject(new Error(msg || 'Empty')));
    const pickFirst = msg => {
        const screen = rejectEmpty(msg);
        return rows => screen(rows).then(rows => rows[0]);
    };

    let sessions = [];
    const addSession = session => sessions.push(session);
    const removeSession = session => sessions.splice(sessions.indexOf(session), 1);

    let groupProto = {
        addSession: function(session) {
            this.sessions.push(session);
        },
        removeSession: function(session) {
            this.sessions.splice(this.sessions.indexOf(session), 1);
        },
        addServer: function(server) {
            if ( this.servers.indexOf(server) < 0 ) {
                this.servers.push(server);
                this.serversById.set(server.serverId, server);
            }
            this.updateFreeMB();
        },
        removeServer: function(server) {
            this.servers.splice(this.servers.indexOf(server), 1);
            this.serversById.delete(server.serverId);
            this.updateFreeMB();
        },
        updateFreeMB: function() {
            let sum = 0.0;
            this.servers.forEach(server => { sum += (+server.freeMB); });
            this.freeMB = this.servers.length ? (sum / this.servers.length) : 0.0;
            return this.freeMB;
        },
        pickUploadServer: function(bytes) {
            let server = pickRand(this.servers);
            return ( server && (server.freeMB > (bytes>>20)) && Promise.resolve(server) ||
                     Promise.reject(server ? 'Disks are full' : 'No data servers online') );
        }
    };
    let groupsOnline = {
        byId: new Map(),
        get: function(groupId) {
            let g = this.byId.get(groupId);
            if ( ! g ) {
                g = Object.assign({
                    groupId: groupId,
                    sessions: [],
                    servers: [],
                    serversById: new Map()
                }, groupProto); // could do actual prototype inheritance if groupProto gets many more methods
                this.byId.set(groupId, g);
            }
            return g;
        }
    };

    function decorateSeatsOnline(seats) {
        var groupId = seats.length ? seats[0].personGroup : 0;
        var groupOnline = groupsOnline.get(groupId);
        var byKey = new Map();
        groupOnline.sessions.forEach(function(s) {
            s.seatKey && byKey.set(s.seatKey, s);
        });
        seats.forEach(function(s) {
            s.isOnline = byKey.has(s.seatKey);
        });
        return seats;
    }

    
    let ConnectionConfig = {
        hostname: "localhost",
        user: "qubcloud",
        password: "jihqprewgop",
        database: "qubcloud",
        inviteHost: "www.qub.buffalo.edu",
        mainInstallPath: "/online",
        emailSender: "qub@www.qub.buffalo.edu",
        requestEmail: "qub@www.qub.buffalo.edu",
        adminLink: "https://www.qub.buffalo.edu/online/"
    };

    exports.Connect = function Connect(config) {
        config = Object.assign({}, ConnectionConfig, config);
        return mysql.createPool({
            connectionLimit: 100,
            host: config.hostname,
            user: config.user,
            password: config.password,
            database: config.database,
            charset: "utf8mb4_unicode_ci"
            /* IMPORTANT: before creating tables, you want mysql to be using full multibyte UTF8.

This goes in /etc/mysql/mysql.conf.d/mysqld.cnf:
----
init_connect='SET collation_connection = utf8mb4_unicode_ci'
init_connect='SET NAMES utf8mb4'
character-set-server=utf8mb4
collation-server=utf8mb4_unicode_ci
skip-character-set-client-handshake
----
If you already make the tables, you have to fix them, e.g.:
----
ALTER DATABASE Models CHARACTER SET = utf8mb4 COLLATE = utf8mb4_unicode_ci;
USE Models;
ALTER TABLE Person CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
-----
*/
        });
    };


    exports.Session = function(config) {
        this.config = Object.assign({}, ConnectionConfig, config);
        this.pool = this.config.pool || exports.Connect(this.config)
        this.seatKey = '';
        this.person = {};
        this.group = {};
    };
    exports.Session.prototype = {
        query: function(q, inserts) {
            console.log('|  mysql>', q,inserts);
            return new Promise((resolve, reject) => {
                this.pool.getConnection((err,connection) => {
                    if (err) {
                        reject("No database connection: "+err);
                        return;
                    }
                    var result;
                    connection.query(q, inserts||[], (err, rows) => {
                        connection.release();
                        if ( err ) {
                            reject("Database error: "+err);
                        } else {
                            if ( q.startsWith('SELECT') ) {
                                resolve(rows);
                            } else {
                                result = rows;
                                connection.commit(function(commErr) {
                                    if ( err ) {
                                        reject("Database error: "+commErr);
                                    } else {
                                        // console.log('committed ',q);
                                        resolve(result);
                                    }
                                });
                            }
                        }
                    }).on('error', err => {
                        reject("Database error: "+err);
                        return;
                    });
                });
            });
        },

        isSuper: function() {
            return this.person.personIsSuper ? Promise.resolve(true) : Promise.reject(new Error('Insufficient credentials'));
        },
        addGroup: function(g) {
            return this.isSuper()
                .then(() => this.query(`INSERT INTO Groups SET ?`, {
                    groupActive: true,
                    groupName: g.groupName,
                    groupAbout: g.groupAbout,
                    groupMaxPerson: g.groupMaxPerson,
                    groupMaxAdmin: g.groupMaxAdmin,
                    groupMaxSeats: g.groupMaxSeats,
                    groupComputeMinutes: g.groupComputeMinutes,
                    groupFreeMB: 0,
                    groupDataKey: uuid.v4(),
                    groupComputeKey: uuid.v4()
                })).then(result => result.insertId);
        },
        listGroups: function(groupId) {
            if ( (! this.person.personIsSuper) &&
                 (this.person.personHoping ||
                  (this.group.groupId !== groupId)) ) {
                return Promise.reject(new Error('Insufficient credentials'));
            }
            const where = groupId ? ` WHERE groupId=${groupId}` : '';
            return this.query(`SELECT * FROM Groups${where};`);
        },
        canEditGroup: function(groupId) {
            if ( (! this.person.personIsSuper) &&
                 ((! this.person.personIsAdmin) || (this.person.personGroup !== groupId)) ) {
                return Promise.reject(new Error('Insufficient credentials'));
            }
            return Promise.resolve(true);
        },
        setGroupActive: function(groupId, active) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`UPDATE Groups SET groupActive=${active} WHERE groupId=${groupId};`))
                .then(() => groupsOnline.get(groupId).sessions.map(sess => sess.eject('Group account deactivated')));
        },
        setGroupName: function(groupId, name) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`UPDATE Groups SET groupName=${mysql.escape(name)} WHERE groupId=${groupId};`));
        },
        setGroupAbout: function(groupId, about) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`UPDATE Groups SET groupAbout=${mysql.escape(about)} WHERE groupId=${groupId};`));
        },
        setGroupDataKey: function(groupId, key) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`UPDATE Groups SET groupDataKey=${mysql.escape(key)} WHERE groupId=${groupId};`))
                .then(() => {
                    groupsOnline.get(groupId).servers.map(server => server.eject('Data key changed'));
                });
        },
        setGroupComputeKey: function(groupId, key) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`UPDATE Groups SET groupComputeKey=${mysql.escape(key)} WHERE groupId=${groupId};`));
        },
        setGroupMaxPerson: function(groupId, maxPerson) {
            return this.isSuper()
                .then(() => this.query(`UPDATE Groups SET groupMaxPerson=${maxPerson} WHERE groupId=${groupId};`));
        },
        setGroupMaxAdmin: function(groupId, maxAdmin) {
            return this.isSuper()
                .then(() => this.query(`UPDATE Groups SET groupMaxAdmin=${maxAdmin} WHERE groupId=${groupId};`));
        },
        setGroupMaxSeats: function(groupId, maxSeats) {
            return this.isSuper()
                .then(() => this.query(`UPDATE Groups SET groupMaxSeats=${maxSeats} WHERE groupId=${groupId};`));
        },
        setGroupComputeMinutes: function(groupId, minutes) {
            return this.isSuper()  /// also let the compute coord decrement them
                .then(() => this.query(`UPDATE Groups SET groupComputeMinutes=${minutes} WHERE groupId=${groupId};`));
        },
        
        addPerson: function(p) {
            if ( (p.personGroup <= 0) ||
                 ((! this.person.personIsSuper) && (p.personGroup !== this.person.personGroup)) ) {
                return Promise.reject(new Error('Invalid Group'));
            }
            if ( ! (this.person.personIsSuper || this.person.personIsAdmin) ) {
                return Promise.reject(new Error('Insufficient permissions'));
            }
            if ( ! this.person.personIsSuper ) {
                p.personIsSuper = false;
            }
            return this.query(`SELECT * FROM Groups WHERE groupId=${p.personGroup};`)
                .then(pickFirst('Invalid group id'))
                .then(group => this.query(`SELECT personIsAdmin FROM Person WHERE personGroup=${p.personGroup} AND personActive=TRUE;`)
                      .then(people => {
                          if ( people.length >= group.groupMaxPerson ) {
                              throw new Error('Group has reached its person limit');
                          }
                          if ( p.personIsAdmin && (people.filter(p2 => p2.personIsAdmin).length >= group.groupMaxAdmin) ) {
                                  throw new Error('Group has reached its admin limit');
                          }
                      }))
                .then(() => this.query(`INSERT INTO Person SET ?`, {
                    personGroup: p.personGroup,
                    personActive: true,
                    personIsAdmin: p.personIsAdmin,
                    personIsContact: p.personIsContact,
                    personIsSuper: p.personIsSuper,
                    personName: p.personName,
                    personAbout: p.personAbout,
                    personEmail: p.personEmail
                })).then(result => result.insertId);
        },
        listPeople: function(groupId) {
            if ( (! this.person.personIsSuper) &&
                 (! (this.person.personIsAdmin && (this.person.personGroup === groupId)) ) ) {
                return Promise.reject(new Error('Insufficient permissions'));
            }
            const where = groupId ? ` WHERE personGroup=${groupId}` : '';
            return this.query(`SELECT * FROM Person${where};`);
        },
        canEditPerson: function(personId, admin) {
            if ( this.person.personIsSuper ||
                 ((! admin) || (this.person.personId === personId)) ) {
                return Promise.resolve(true);
            }
            if ( ! this.person.personIsAdmin ) {
                return Promise.reject(new Error('Insufficient permissions'));
            }
            return this.query(`SELECT personId, personGroup FROM Person WHERE personId=${personId}`)
                .then(pickFirst('No such person'))
                .then(person => {
                    if ( person.personGroup !== this.person.personGroup ) {
                        throw new Error('Insufficient permissions');
                    }
                    return true;
                });
        },
        setPersonGroup: function(personId, groupId) {
            return this.isSuper()
                .then(() => this.query(`UPDATE Person SET personGroup=${groupId} WHERE personId=${personId};`))
                .then(() => {
                    if ( personId === this.person.personId ) {
                        this.person.personGroup = groupId;
                    }
                });
        },
        setPersonActive: function(personId, active) {
            return this.canEditPerson(personId, true)
                .then(() => this.query(`UPDATE Person SET personActive=${active} WHERE personId=${personId};`))
                .then(() => {
                    if ( ! active ) {
                        for (let sess of sessions) {
                            if ( sess.person.personId === personId ) {
                                sess.eject('Account deactivated');
                            }
                        }
                    }
                    if ( personId === this.person.personId ) {
                        this.person.personActive = active;
                    }
                });
        },
        setPersonIsAdmin: function(personId, isAdmin) {
            return this.canEditPerson(personId, true)
                .then(t => {
                    if ( isAdmin ) {
                        return this.query(`SELECT personId, personGroup, personIsAdmin, groupId, groupMaxAdmin FROM Person INNER JOIN Groups ON personGroup=groupId WHERE personId=${personId}`).then(pickFirst()).then(pg => {
                            return this.query(`SELECT personGroup, personIsAdmin FROM Person WHERE personGroup=${pg.personGroup} and personIsAdmin=TRUE;`)
                                .then(people => {
                                    if ( pg.groupMaxAdmin <= people.length ) {
                                        throw new Error('Group has reached its admin limit');
                                    }
                                    if ( personId === this.person.personId ) {
                                        this.person.personIsAdmin = isAdmin;
                                    }
                                });
                        });
                    }
                    return t;
                }).then(() => this.query(`UPDATE Person SET personIsAdmin=${mysql.escape(isAdmin)} WHERE personId=${personId};`));
        },
        setPersonIsContact: function(personId, isContact) {
            return this.canEditPerson(personId, true)
                .then(() => this.query(`UPDATE Person SET personIsContact=${mysql.escape(isContact)} WHERE personId=${personId};`))
                .then(() => {
                    if ( personId === this.person.personId ) {
                        this.person.personIsContact = isContact;
                    }
                });
        },
        setPersonIsSuper: function(personId, isSuper) {
            this.isSuper()
                .then(() => this.query(`UPDATE Person SET personIsSuper=${mysql.escape(isSuper)} WHERE personId=${personId};`));
        },
        setPersonName: function(personId, name) {
            return this.canEditPerson(personId)
                .then(() => this.query(`UPDATE Person SET personName=${mysql.escape(name)} WHERE personId=${personId};`))
                .then(() => {
                    if ( personId === this.person.personId ) {
                        this.person.personName = name;
                    }
                });
        },
        setPersonAbout: function(personId, about) {
            return this.canEditPerson(personId)
                .then(() => this.query(`UPDATE Person SET personAbout=${mysql.escape(about)} WHERE personId=${personId};`))
                .then(() => {
                    if ( personId === this.person.personId ) {
                        this.person.personAbout = about;
                    }
                });
        },
        setPersonEmail: function(personId, email) {
            return this.canEditPerson(personId)
                .then(() => this.query(`UPDATE Person SET personEmail=${mysql.escape(email)} WHERE personId=${personId};`))
                .then(() => {
                    if ( personId === this.person.personId ) {
                        this.person.personEmail = email;
                    }
                });
        },

        requestAccount: function(h, ip, agent) {
            console.log('reqAccount', h);
            const pat = h.groupInfo ? `${mysql.escape('%'+h.groupInfo.toLowerCase()+'%')}` : '';
            return ((this.person.personId <= 0) ? Promise.resolve(true) : this.signOut())
                .then(() => this.query(`INSERT INTO Hopeful SET ?`, {
                    hopefulGroup: 0,
                    hopefulGroupInfo: h.groupInfo,
                    hopefulName: h.name,
                    hopefulEmail: h.email,
                    hopefulAbout: h.about
                })).then(result => {
                    return (this.seatKey ? Promise.resolve(this.seatKey) : this.newSeat())
                        .then(seatKey => this.query(`UPDATE Seat SET seatPerson=${-result.insertId}, seatWhen=NOW(), seatIP=${mysql.escape(ip)}, seatAgent=${mysql.escape(agent)} WHERE seatKey=${mysql.escape(seatKey)};`))
                        .then(() => {
                            if ( pat ) {
                                return this.query(`SELECT groupId, groupName, groupAbout FROM Groups WHERE (LOWER(CONCAT_WS(groupName, groupAbout)) LIKE ${pat});`)
                            } else {
                                return Promise.resolve([]);
                            }
                        }).then(groups => {
                            if ( groups.length === 1 ) {
                                this.fwdHopeful(result.insertId, groups[0].groupId);
                            } else {
                                this.bwdHopeful(result.insertId);
                            }
                            return this.seatKey;
                        });
                });
        },
        listHopeful: function(groupId) {
            return ((groupId === 0) ? this.isSuper() : Promise.resolve(true))
                .then(() => this.canEditGroup(groupId))
                .then(() => this.query(`SELECT * from Hopeful WHERE hopefulGroup=${groupId};`))
                .then(rows => Promise.all(rows.map(h => this.whoHasEmail(h.hopefulEmail, groupId)
                                                   .then(candidates => candidates.filter(c => ((groupId === 0) || (groupId === c.personGroup))))
                                                   .then(candidates => Object.assign(h, {candidates: candidates})))));
        },
        fwdHopeful: function(hopefulId, groupId) {
            return this.isSuper()
                .then(() => this.query(`UPDATE Hopeful SET hopefulGroup=${groupId} WHERE hopefulId=${hopefulId};`))
                .then(() => this.query(`SELECT * FROM Hopeful WHERE hopefulId=${hopefulId};`))
                .then(pickFirst())
                .then(h => {
                    const groupOnline = groupsOnline.get(groupId);
                    for (var i=0; i<groupOnline.sessions.length; ++i) {
                        let s = groupOnline.sessions[i];
                        if ( s.person.personIsAdmin && s.person.personIsContact ) {
                            this.query(`SELECT * from Hopeful WHERE hopefulId=${hopefulId};`)
                                .then(hopefuls => s.notifyOfHopefuls(hopefuls))
                                .catch(err => {
                                    console.log('notifying of hopefuls for group '+groupId, err);
                                });
                            break; // just notify one person
                        }
                    }
                    if ( i === groupOnline.sessions.length ) { // no contact online
                        this.query(`SELECT personName, personEmail, personGroup, personIsAdmin, personIsContact FROM Person WHERE personGroup=${groupId} and personIsAdmin=TRUE and personIsContact=TRUE;`)
                                    .then(rejectEmpty('No Admin Contacts for group '+groupId))
                                    .then(contacts => Promise.all(contacts.map(ct => {
                                        return emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                                            .send({
                                                from: this.config.emailSender,
                                                to: `${ct.personName} <${ct.personEmail}>`,
                                                subject: 'QUB Online - New account request',
                                                text: `${h.hopefulName} wants to join your QUB Online group (they asked for "${h.hopefulGroupInfo}"). Please <a href="${this.config.adminLink}">Sign in</a> to accept the request, or to forward it to the appropriate place.

`
                                            }, (err, msg) => {
                                                if ( err ) {
                                                    console.log(`Sending notification to ${ct.personName} of ${ct.personGroup}:`, err);
                                                }
                                            });
                                    }))).catch(err => {
                                        console.log('Notifying group '+groupId, err);
                                    });
                    }
                });
        },
        bwdHopeful: function(hopefulId) {
            return this.query(`SELECT * from Hopeful WHERE hopefulId=${hopefulId};`)
                .then(pickFirst())
                .then(h => {
                    return (h.hopefulGroup ? this.canEditGroup(h.hopefulGroup) : Promise.resolve(true))
                        .then(() => this.query(`UPDATE Hopeful SET hopefulGroup=0 WHERE hopefulId=${hopefulId};`))
                        .then(() => {
                            for (var i=0; i<sessions.length; ++i) {
                                let s = sessions[i];
                                if ( s.person.personIsSuper ) {
                                    s.notifyOfHopefuls([h]);
                                    break; // just notify one person
                                }
                            }
                            if ( i === sessions.length ) { // no super online
                                this.query(`SELECT personName, personEmail, personIsSuper FROM Person WHERE personIsSuper=TRUE;`)
                                    .then(rejectEmpty('No Super!'))
                                    .then(supers => Promise.all(supers.map(sup => {
                                        return emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                                            .send({
                                                from: this.config.emailSender,
                                                to: `${sup.personName} <${sup.personEmail}>`,
                                                subject: 'QUB Online - New account/Unknown group',
                                                text: `${h.hopefulName} is looking for the group "${h.hopefulGroupInfo}" -- please <a href="${this.config.adminLink}">Sign in</a> to deal with it.

`
                                            }, (err, msg) => {
                                                if ( err ) {
                                                    console.log(`Sending notification to super of hopeful/nogroup:`, err);
                                                }
                                            });
                                    }))).catch(err => {
                                        console.log('Notifying super of hopeful/nogroup:', err);
                                    });
                            }
                        });
                });
        },
        acceptHopeful: function(hopefulId, isAdmin, isContact) {
            return this.query(`SELECT * from Hopeful WHERE hopefulId=${hopefulId};`)
                .then(pickFirst())
                .then(h => {
                    return this.canEditGroup(h.hopefulGroup)
                        .then(() => this.addPerson({
                            personGroup: h.hopefulGroup,
                            personIsAdmin: isAdmin,
                            personIsContact: isContact,
                            personIsSuper: false,
                            personName: h.hopefulName,
                            personAbout: h.hopefulAbout,
                            personEmail: h.hopefulEmail
                        })).then(personId => this.query(`SELECT * FROM Seat WHERE seatPerson=${-hopefulId};`)
                                 .then(pickFirst('No such seat'))
                                 .then(seat => this.query(`UPDATE Seat SET seatPerson=${personId} WHERE seatId=${seat.seatId};`)
                                       .then(() => this.query(`DELETE FROM Hopeful WHERE hopefulId=${hopefulId};`))
                                       .then(() => {
                                           for (let sess of sessions) {
                                               if ( sess.seatKey === seat.seatKey ) {
                                                   sess.accepted(personId, h.hopefulGroup);
                                                   break;
                                               }
                                           }
                                           return personId;
                                       })));
                });
        },
        acceptCandidate: function(hopefulId, personId) {
            return this.aboutPerson(personId)
                .then(p => this.canEditGroup(p.personGroup)
                      .then(() => this.query(`SELECT * FROM Seat WHERE seatPerson=${-hopefulId};`)
                            .then(pickFirst('No such seat'))
                            .then(seat => this.query(`UPDATE Seat SET seatPerson=${personId} WHERE seatId=${seat.seatId};`)
                                  .then(() => this.query(`DELETE FROM Hopeful WHERE hopefulId=${hopefulId};`))
                                  .then(() => {
                                      for (let sess of sessions) {
                                          if ( sess.seatKey === seat.seatKey ) {
                                              sess.accepted(personId, p.personGroup);
                                              break;
                                          }
                                      }
                                      return personId;
                                  }))));
        },
        rejectHopeful: function(hopefulId) {
            return this.query(`SELECT * FROM Hopeful WHERE hopefulId=${hopefulId};`)
                .then(pickFirst())
                .then(h => this.canEditGroup(h.hopefulGroup))
                .then(() => this.query(`UPDATE Seat SET seatPerson=0 WHERE seatPerson=${-hopefulId};`))
                .then(() => this.query(`DELETE FROM Hopeful WHERE hopefulId=${hopefulId};`));
        },
        whoHasEmail: function(email, groupId) {
            groupId = groupId|0;
            let where = [`personEmail=${mysql.escape(email)}`];
            if ( (! this.person.personIsSuper) || groupId ) {
                where.push(`personGroup=${groupId}`);
            }
            return this.query(`SELECT personId, personName, personEmail, personGroup FROM Person WHERE ${where.join(' AND ')};`);
        },
        notifyOfHopefuls: function(hopefuls, groupId) {
            return Promise.all(hopefuls.map(h => this.whoHasEmail(h.hopefulEmail, groupId)
                                            .then(candidates => Object.assign(h, {candidates: candidates}))))
                .then(hopefuls => {
                    return hopefuls.length && this.config.onhopefuls && this.config.onhopefuls(hopefuls) || true;
                }).catch(err => {
                    console.log('notify of hopefuls for group '+groupId, err);
                });
        },
        notifyAllHopeful: function(groupId) {
            return this.query(`SELECT * FROM Hopeful WHERE hopefulGroup=${groupId};`)
                .then(hopefuls => {
                    return this.notifyOfHopefuls(hopefuls, groupId);
                });
        },
        
        invite: function(personId, expiresHours, usesRemain) {
            return this.query(`SELECT * FROM Person WHERE personId=${personId};`)
                .then(pickFirst('No such person'))
                .then(p => {
                    if ( ((! this.person.personIsSuper) || (! p.personGroup)) &&
                         (this.person.personId !== personId) &&
                         ((! this.person.personIsAdmin) || (this.person.personGroup !== p.personGroup)) ) {
                        throw new Error('Insufficient credentials');
                    }
                    const inviteKey = uuid.v4();
                    const expires = new Date();
                    expires.setHours(expires.getHours() + Math.max(1, expiresHours));
                    return this.query(`INSERT INTO Invite SET ?`, {
                        inviteKey: inviteKey,
                        invitePerson: personId,
                        inviteExpires: expires,
                        inviteUsesRemain: Math.max(0, usesRemain)
                    }).then(iResult => {
                        this.sendInvite(inviteKey, p);
                        return inviteKey;
                    });
                });
        },
        sendInvite: function(inviteKey, p) {
            const url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/signin.html#'+inviteKey;
            emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                .send({from: this.config.emailSender,
                       to: p.personName + ' <' + p.personEmail + '>',
                       subject: "Sign in to QuB Online",
                       text: `Use this link to sign your browser into QuB Online as ${p.personName}:

${url}
`
                      },
                      (err, msg) => {
                          if ( err ) {
                              console.log(`Sending invitation to ${p.personName} <${p.personEmail}>: `, err);
                          }
                      });
        },
        
        newSeat: function(s) {
            const seatKey = uuid.v4();
            this.seatKey = seatKey;
            return this.query(`INSERT INTO Seat SET ?`, Object.assign({}, s, {
                seatKey: seatKey,
                seatWhen: new Date()
            })).then(() => seatKey);
        },
        auth: function(q, ip, agent) {
            if ( agent && (agent.length > 255) ) {
                agent = agent.slice(0, 255);
            }
            return this.query(`SELECT seatId, seatPerson FROM Seat WHERE seatKey=${mysql.escape(""+q.seatKey)};`)
                .then(pickFirst('Signin is invalid or expired'))
                .then(seat => {
                    this.query(`UPDATE Seat SET seatWhen=NOW(), seatIP=${mysql.escape(""+ip)}, seatAgent=${mysql.escape(agent)} WHERE seatId=${seat.seatId};`)
                        .catch(err => {
                            console.log("Updating signin info: ", err);
                        });
                    if ( seat.seatPerson < 0 ) { // can be negative seatPerson: actually -hopefulId
                        return this.query(`SELECT hopefulId, hopefulGroup, hopefulName, hopefulEmail FROM Hopeful WHERE hopefulId=${-seat.seatPerson};`)
                            .then(pickFirst('Expired?'))
                            .then(h => {
                                return {
                                    personHoping: true,
                                    personId: seat.seatPerson,
                                    personGroup: h.hopefulGroup,
                                    personName: h.hopefulName,
                                    personEmail: h.hopefulEmail
                                };
                            });
                    } else {
                        return this.query(`SELECT personId, personGroup, personActive, personIsAdmin, personIsContact, personIsSuper, personName, personAbout, personEmail FROM Person WHERE personId=${seat.seatPerson};`)
                            .then(pickFirst('Missing person'));
                    }
                })
                .then(p => {
                    if ( ! (p.personHoping || p.personActive) ) {
                        throw new Error("Account not active");
                    }
                    if ( (q.isAdmin && ! p.personIsAdmin)
                         || (q.isSuper && ! p.personIsSuper) ) {
                        throw new Error("Insufficent permissions");
                    }
                    this.seatKey = q.seatKey;
                    this.person = p;
                    this.group = {groupId: p.personGroup};
                    return this.listGroups(p.personGroup)
                        .then(pickFirst('Missing group '+p.personGroup+' ?!'))
                        .then(g => {
                            this.group = g;
                            if ( ! g.groupActive ) {
                                throw new Error('Group account not active');
                            }
                            if ( this.person.personIsSuper ) {
                                this.notifyAllHopeful(0);
                            }
                            if ( this.person.personIsAdmin && this.person.personIsContact ) {
                                this.notifyAllHopeful(g.groupId);
                            }
                            if ( ! p.personHoping ) {
                                this.groupOnline = groupsOnline.get(g.groupId);
                                if ( this.groupOnline.sessions.length >= g.groupMaxSeats ) {
                                    this.groupOnline = null;
                                    throw new Error("Group has reached the limit for simultaneous logins");
                                }
                                this.groupOnline.addSession(this);
                            }
                            return this.person;
                        });
                });
        },
        finalize: function() {
            this.groupOnline && this.groupOnline.removeSession(this);
            this.groupOnline = null;
        },
        signIn: function(inviteKey, ip, agent) {
            return this.query(`SELECT inviteId, invitePerson, inviteExpires, inviteUsesRemain FROM Invite WHERE inviteKey=${mysql.escape(inviteKey)};`)
                .then(pickFirst('Invitation expired or unknown'))
                .then(inv => {
                    const delInvite = () => {
                        this.query(`DELETE FROM Invite WHERE inviteId=${inv.inviteId};`)
                            .catch(err => {
                                console.log("Deleting invite: ", err);
                            });
                    };
                    if ( ((new Date()) - inv.inviteExpires) > 0 ) {
                        delInvite();
                        throw new Error("Invitation expired");
                    }
                    if ( inv.inviteUsesRemain > 1 ) {
                        this.query(`UPDATE Invite SET inviteUsesRemain=${inv.inviteUsesRemain-1} WHERE inviteId=${inv.inviteId};`)
                            .catch(err => {
                                console.log("Decrementing inviteUsesRemain: ", err);
                            });
                    } else if ( inv.inviteUsesRemain === 1 ) {
                        delInvite();
                    }
                    return this.newSeat({
                        seatPerson: inv.invitePerson,
                        seatIP: ip,
                        seatAgent: agent
                    });
                }).then(seatKey => this.auth({seatKey: seatKey}, ip, agent)
                                       .then(p => seatKey));
        },
        signOut: function() {
            this.person = {};
            this.group = {};
            return this.query(`UPDATE Seat SET seatPerson=0 WHERE seatKey=${mysql.escape(this.seatKey)};`);
        },
        eject: function(reason) {
            this.person = {};
            this.group = {};
            this.serverId = null;
            this.config.oneject && this.config.oneject(reason);
        },
        accepted: function(personId, groupId) {
            this.person.personId = personId;
            this.person.personGroup = groupId;
            this.person.personIsHopeful = false;
            this.group.groupId = groupId;
            this.query(`SELECT * FROM Person WHERE personId=${personId};`)
                .then(pickFirst('No such person'))
                .then(p => {
                    Object.assign(this.person, p);
                    return this.query(`SELECT * FROM Groups WHERE groupId=${groupId} LIMIT 1;`)
                        .then(pickFirst('No such group'));
                }).then(group => {
                    this.group = group;
                    this.config.onaccepted && this.config.onaccepted(this.person);
                });
        },
        
        aboutPerson: function(personId) {
            if ( (! personId) || ((this.person.personId > 0) && (this.person.personId === personId)) ) {
                return Promise.resolve(this.person);
            }
            return this.query(`SELECT * FROM Person WHERE personId=${personId};`)
                .then(pickFirst('No such person'))
                .then(p => {
                    if ( ! (this.person.personIsSuper ||
                            (this.person.personId === p.personId) || 
                            (this.person.personIsAdmin && (this.person.personGroup === p.personGroup))) ) {
                        delete p.personActive;
                        delete p.personIsAdmin;
                        delete p.personIsSuper;
                        delete p.personEmail;
                    }
                    return p;
                });
        },
        listSeats: function(groupId) { // groupId=undefined: seats of current person
            if ( this.person.personId <= 0 ) {
                return Promise.resolve([]);
            }
            return (groupId ?
                    this.canEditGroup(groupId)
                        .then(() => this.query(`SELECT seatPerson, seatKey, seatWhen, seatIP, seatAgent, personId, personName, personGroup FROM Seat INNER JOIN Person ON seatPerson=personId WHERE personGroup=${groupId};`)) : 
                    this.query(`SELECT seatPerson, seatKey, seatWhen, seatIP, seatAgent, personId, personName, personGroup FROM Seat INNER JOIN Person ON seatPerson=personId WHERE seatPerson=${this.person.personId};`))
                .then(decorateSeatsOnline);
        },
        unSignIn: function(seatKey, personId) { // personId=undefined: use this.person.personId
            let pid = this.person.personId;
            let checkAllowed = () => {
                if ( this.person.personId <= 0 ) {
                    return Promise.reject(new Error('Insufficient permissions'));
                }
                if ( personId && this.person.personIsAdmin && ! this.person.personIsSuper ) {
                    pid = personId;
                    return this.query(`SELECT personId, personGroup FROM Person WHERE personId=${personId} AND personGroup=${this.person.personGroup} LIMIT 1;`)
                        .then(pickFirst('No such person'));
                }
                return Promise.resolve(true);
            };
            return checkAllowed()
                .then(() => this.query(`DELETE FROM Seat WHERE seatPerson=${pid} AND seatKey=${mysql.escape(seatKey)};`))
                .then(() => {
                    for (let sess of sessions) {
                        if ( sess.seatKey === seatKey ) {
                            sess.eject('Remotely logged out');
                            break;
                        }
                    }
                });
        },

        digestModel: (m, personName, personAbout) => [m.modelName, m.modelAbout, m.modelAuthors, personName||this.person.personName, personAbout||this.person.personAbout, ""+Nstate+"state", ""+Nclass+"class"].join("\n").toLowerCase(),
        addModel: function(m) {
            if ( this.person.personId <= 0 ) {
                return Promise.reject(new Error('Insufficent credentials'));
            }
            const qmj = m.modelQmj;
            const Nstate = qmj && qmj.states && qmj.states.length || 0;
            const classes = [];
            let Nclass = 0;
            qmj.states && qmj.states.forEach(s => {
                classes[s.Class] += 1;
            });
            classes.forEach(c => { ++Nclass });
            const digest = this.digestModel(m);
            let insModel = {
                modelPerson: this.person.personId,
                modelGroup: this.person.personGroup,
                modelPublic: m.modelPublic,
                modelName: m.modelName,
                modelQmj: JSON.stringify(qmj),
                modelAbout: m.modelAbout,
                modelWhenAdded: new Date(),
                modelWhenModified: new Date(),
                modelAuthors: m.modelAuthors,
                modelDOI: m.modelDOI,
                modelStates: m.modelStates,
                modelClasses: m.modelClasses,
                modelDigest: digest,
                modelFolder: m.modelFolder
            };
            if ( m.modelWhenPub ) {
                insModel.modelWhenPub = m.modelWhenPub;
            }
            return this.query(`INSERT INTO Model SET ?`, insModel)
                .then(result => result.insertId);
        },
        modelFromDB: m => {
            try {
                m.modelQmj = JSON.parse(m.modelQmj);
            } catch (err) {}
            return m;
        },
        updateModel: function(m) {
            var personId, groupId;
            return this.query(`SELECT modelPerson, modelGroup FROM Model WHERE modelId=${m.modelId};`)
                .then(pickFirst('No such model'))
                .then(dbm => {
                    if ( this.person.personIsSuper ||
                         (this.person.personId === db.modelPerson) ||
                         ((this.person.personGroup === db.modelGroup) && this.person.personIsAdmin) ) {
                        personId = db.modelPerson;
                        groupId = db.modelGroup;
                        return Promise.resolve(true);
                    }
                    return Promise.reject(new Error('Insufficient permissions'));
                }).then(() => {
                    let updates = [];
                    let dateToo = false;
                    let doField = (name, trivial) => {
                        const x = m[name];
                        if ( typeof(x) !== 'undefined' ) {
                            updates.push(`${name}=${mysql.escape(x)}`);
                            dateToo = dateToo || ! trivial;
                        }
                    };
                    doField('modelPublic', true);
                    ['modelAbout', 'modelWhenPub', 'modelAuthors', 'modelDOI', 'modelStates', 'modelClasses', 'modelFolder'].forEach(doField);
                    if ( 'modelQmj' in m ) {
                        updates.push(`modelQmj=${mysql.escape(JSON.stringify(m.modelQmj))}`);
                        dateToo = true;
                    }
                    if ( dateToo ) {
                        updates.push(`modelWhenModified=NOW()`);
                    }
                    
                    return this.query(`SELECT personId, personName, personAbout FROM Person WHERE personId=${personId};`)
                        .then(pickFirst())
                        .then(p => {
                            updates.push(`modelDigest=${mysql.escape(this.digestModel(m, p.personName, p.personAbout))}`);
                            this.query(`UPDATE Model SET ${updates.join(', ')} WHERE modelId=${m.modelId};`);
                        });
                });
        },
        listModels: function(q) {
            let where = [];
            let doWhere = name => {
                const x = q[name];
                if ( typeof(x) !== 'undefined' ) {
                    where.push(`${name}=${mysql.escape(x)}`);
                }
            };
            ['modelPerson', 'modelGroup', 'modelStates', 'modelClasses', 'modelFolder'].forEach(doWhere);
            if ( q.modelDigest ) {
                where.push(`(modelDigest LIKE ${mysql.escape('%'+q.modelDigest.toLowerCase()+'%')})`);
            }
            /// limit chronologically
            if ( ! this.person.personIsSuper ) {
                where.push(`(modelPublic=TRUE OR modelGroup=${this.person.personGroup})`);
            }
            return this.query(`SELECT * FROM Model ${where.length ? 'WHERE ' : ''}${where.join(' AND ')};`)
                .then(models => models.map(this.modelFromDB));
        },
        getModel: function(modelId) {
            return this.query(`SELECT * FROM Model WHERE modelId=${modelId};`)
                .then(pickFirst('No such model'))
                .then(m => {
                    if ( (! m.modelPublic) && (m.modelGroup !== this.person.personGroup) && (! this.person.personIsSuper) ) {
                        throw new Error('Insufficient permissions');
                    }
                    return this.modelFromDB(m);
                });
        },

        authServer: function(dataKey, uuid) {
            return this.query(`SELECT * FROM Groups WHERE groupDataKey=${mysql.escape(dataKey)} AND groupActive=TRUE;`)
                .then(pickFirst('Invalid data key'))
                .then(group => this.query(`SELECT serverId, serverUUID FROM DataServer WHERE serverUUID=${mysql.escape(uuid)}`)
                      .then(servers => {
                          this.group = group;
                          this.groupOnline = groupsOnline.get(group.groupId);
                          return (servers.length ? Promise.resolve(servers[0].serverId) : this.query(`INSERT INTO DataServer SET ?`, {
                              serverUUID: uuid,
                              serverWhenFirst: new Date(),
                              serverCoverage: 0.0,
                              serverFreeMB: 0
                          }).then(result => result.insertId)).then(serverId => {
                              this.serverId = serverId;
                              return this.query(`UPDATE DataServer SET serverGroup=${group.groupId}, serverWhenLast=NOW() WHERE serverId=${serverId};`)
                                  .then(() => serverId);
                          });
                      }));
        },
        updateServer: function(coverage, freeMB, baseUrl) {
            let updates = [`serverCoverage=${mysql.escape(coverage)}`,
                           `serverFreeMB=${mysql.escape(freeMB)}`];
            if ( baseUrl ) {
                updates.push(`serverBaseUrl=${mysql.escape(baseUrl)}`);
                updates.push(`serverGroup=${this.group.groupId}`);
            }
            return this.query(`UPDATE DataServer SET ${updates.join(', ')} WHERE serverId=${this.serverId};`);
        },
        updateGroupFreeMB: function(groupFree) {
            return this.query(`UPDATE Groups SET groupFreeMB=${groupFree} WHERE groupId=${this.group.groupId};`);
        },

        forgetServer: function(serverId) {
            serverId = serverId || this.serverId;
            return this.query(`SELECT serverId, serverGroup FROM Server WHERE serverId=${serverId};`)
                .then(pickFirst('No such server'))
                .then(server => {
                    if ( this.person.personIsSuper ||
                         (this.serverId === serverId) ||
                         (this.person.personIsAdmin && (this.group.groupId === server.serverGroup)) ) {
                        return true;
                    }
                    throw new Error('Insufficient credentials');
                }).then(() => this.query(`DELETE FROM DataInstance WHERE instanceServer=${serverId};`))
                .then(() => this.query(`DELETE FROM DataServer WHERE serverId=${serverId};`));
        },
        listServers: function(groupId) {
            return this.canEditGroup(groupId)
                .then(() => this.query(`SELECT * FROM DataServer WHERE serverGroup=${groupId};`))
                .then(servers => servers.map(s => {
                    s.isOnline = this.groupOnline.serversById.has(s.serverId);
                    return s;
                }));
        },

        digestDataset: ds => [ds.datasetName, ds.datasetAbout, `${ds.datasetPoints} points`, ds.datasetMolecule].join('\n').toLowerCase(),
        addDataset: function(ds) {
            if ( this.person.personId <= 0 ) {
                return Promise.reject(new Error('Insufficent credentials'));
            }
            const digest = this.digestDataset(ds);
            return this.query(`INSERT INTO Dataset SET ?`, {
                datasetPerson: this.person.personId,
                datasetGroup: this.person.personGroup,
                datasetPublic: ds.datasetPublic,
                datasetName: ds.datasetName,
                datasetAbout: ds.datasetAbout,
                datasetWhenRec: ds.datasetWhenRec,
                datasetWhenAdded: new Date(),
                datasetWhenIdl: new Date(),
                datasetSampling: ds.datasetSampling,
                datasetPoints: ds.datasetPoints,
                datasetEvents: ds.datasetEvents,
                datasetSegments: ds.datasetSegments,
                datasetStimuli: JSON.stringify(ds.datasetStimuli),
                datasetConstants: JSON.stringify(ds.datasetConstants),
                datasetMolecule: ds.datasetMolecule,
                datasetEtc: JSON.stringify(ds.datasetEtc),
                datasetDigest: digest,
                datasetFolder: ds.datasetFolder
            }).then(result => result.insertId);
        },
        canEditDataset: function(datasetId) {
            return this.isSuper()
                .catch(() => {
                    return this.query(`SELECT datasetPerson, datasetGroup FROM Dataset WHERE datasetId=${datasetId};`)
                        .then(pickFirst('No such dataset'))
                        .then(ds => {
                            if ( this.person.personId === ds.datasetPerson ) {
                                return Promise.resolve(true);
                            }
                            if ( (this.person.personGroup === ds.datasetGroup) && this.person.personIsAdmin ) {
                                return Promise.resolve(true);
                            }
                            return Promise.reject(new Error('Insufficient permissions'));
                        });
                });
        },
        setDatasetPublic: function(datasetId, datasetPublic) {
            return this.canEditDataset(datasetId)
                .then(() => this.query(`UPDATE Dataset SET datasetPublic=${mysql.escape(datasetPublic)} WHERE datasetId=${datasetId};`))
                .then(() => this.groupOnline.servers.map(server => server.publicChanged(datasetId, datasetPublic)));
        },
        datasetIsPublic: function(datasetId) {
            return this.query(`SELECT datasetId, datasetPublic FROM Dataset WHERE datasetId=${datasetId};`)
                .then(pickFirst('No such dataset'))
                .then(ds => ds.datasetPublic);
        },
        setDatasetFolder: function(datasetId, datasetFolder) {
            return this.canEditDataset(datasetId)
                .then(() => this.query(`UPDATE Dataset SET datasetFolder=${mysql.escape(datasetFolder)} WHERE datasetId=${datasetId};`));
        },
        datasetFromDB: ds => {
            try {
                ds.datasetStimuli = JSON.parse(ds.datasetStimuli);
                ds.datasetConstants = JSON.parse(ds.datasetConstants);
                ds.datasetEtc = JSON.parse(ds.datasetEtc);
            } catch (err) {}
            return ds;
        },
        listDatasets: function(q) {
            let where = [];
            q.datasetId && where.push(`datasetId=${q.datasetId}`);
            q.personId && where.push(`datasetPerson=${q.personId}`);
            q.groupId && where.push(`datasetGroup=${q.groupId}`);
            q.datasetDigest && where.push(`(datasetDigest LIKE ${mysql.escape('%'+q.datasetDigest.toLowerCase()+'%')})`);
            q.datasetFolder && where.push(`datasetFolder=${mysql.escape(q.datasetFolder)}`);
            /// limit chronologically
            if ( ! this.person.personIsSuper ) {
                where.push(`(datasetPublic=TRUE OR datasetGroup=${this.group.groupId})`);
            }
            return this.query(`SELECT * FROM Dataset ${where.length ? 'WHERE ' : ''}${where.join(' AND ')};`)
                .then(rows => rows.map(this.datasetFromDB));
        },
        delDataset: function(datasetId) {
            return this.query(`SELECT * FROM DataInstance WHERE instanceDataset=${datasetId};`)
                .then(instances => {
                    if ( instances.length ) {
                        throw new Error('Data instances exist');
                    }
                }).then(() => this.query(`SELECT datasetId, datasetGroup FROM Dataset WHERE DatasetId=${datasetId};`))
                .then(pickFirst('No such dataset'))
                .then(dataset => {
                    if ( (datasetGroup !== this.group.groupId) && ! this.person.personIsSuper ) {
                        throw new Error('Insufficient privileges');
                    }
                }).then(() => this.query(`DELETE FROM Dataset WHERE datasetId=${datasetId};`));
        },

        canInstanceData: function(datasetId) {
            if ( ! this.serverId ) {
                return Promise.reject(new Error('Not signed in as a data server'));
            }
            return this.query(`SELECT datasetId, datasetGroup from Dataset WHERE datasetId=${datasetId};`)
                .then(pickFirst('No such dataset'))
                .then(dataset => {
                    if ( dataset.datasetGroup !== this.group.groupId ) {
                        throw new Error('Dataset belongs to another group');
                    }
                    return true;
                });
        },
        addDataInstance: function(datasetId, sum, bytes, path) {
            return this.canInstanceData(datasetId)
                .then(() => this.query(`INSERT INTO DataInstance SET ?`, {
                    instanceDataset: datasetId,
                    instanceServer: this.serverId,
                    instanceSum: sum,
                    instanceBytes: bytes,
                    instancePath: path
                })).then(result => result.insertId);
        },
        moveDataInstance: function(datasetId, instanceId, path) {
            return this.canInstanceData(datasetId)
                .then(() => this.query(`UPDATE DataInstance SET instancePath=${mysql.escape(path)} WHERE instanceDataset=${datasetId} AND instanceId=${instanceId};`));
        },
        listInstances: function(datasetId) {
            return this.isSuper()
                .catch(() => this.query(`SELECT datasetId, datasetPublic, datasetGroup FROM Dataset WHERE datasetId=${datasetId};`)
                       .then(pickFirst('No such dataset'))
                       .then(dataset => {
                           if ( (datasetGroup !== this.group.groupId) && ! dataset.datasetPublic ) {
                               throw new Error('Dataset is not public');
                           }
                       }))
                .then(() => this.query(`SELECT * FROM DataInstance WHERE instanceDataset=${datasetId};`));
        },
        pickInstanceServer: function(datasetId) {
            return this.listInstances(datasetId)
                .then(instances => instances.map(instance => ({
                    instancePath: instance.instancePath,
                    instanceSum: instance.instanceSum,
                    instanceBytes: instance.instanceBytes,
                    server: this.groupOnline.serversById.get(instance.instanceServer)
                }))).then(servers => servers.filter(server => !! server.server))
                .then(pickRand);
        },
        idlReplaced: function(datasetId, serverId, whenIdl) {
            return this.query(`SELECT instanceDataset, instanceServer FROM DataInstance WHERE instanceDataset=${datasetId} AND instanceServer=${serverId};`)
                .then(pickFirst("server shouldn't have this file"))
                .then(() => this.query(`UPDATE Dataset SET datasetWhenIdl=${mysql.escape(whenIdl)} WHERE datasetId=${datasetId};`))
                .then(() => this.query(`SELECT * FROM DataInstance WHERE instanceDataset=${datasetId};`))
                .then(instances => {
                    let othersOnline = [];
                    let newUrl = '';
                    instances.forEach(instance => {
                        let instServer = this.groupOnline.serversById.get(instance.instanceServer);
                        if ( instServer ) {
                            if ( instance.instanceServer === serverId ) {
                                newUrl = `${instServer.baseUrl}/${instance.path}/idealized.jsotan`;
                            } else {
                                instServer && othersOnline.push({
                                    server: instServer,
                                    path: instance.instancePath
                                });
                            }
                        }
                    });
                    othersOnline.forEach(other => other.server.idlReplacement(datasetId, other.path, newUrl));
                });
        },
        
        digestProject: p => [p.projectName, p.projectAbout, p.projectAuthors, p.projectType].join('\n').toLowerCase(),
        addProject: function(p) {
            if ( this.person.personId <= 0 ) {
                return Promise.reject(new Error('Insufficent credentials'));
            }
            const digest = this.digestProject(p);
            let ins = {
                projectPerson: this.person.personId,
                projectGroup: this.person.personGroup,
                projectPublic: p.projectPublic,
                projectName: p.projectName,
                projectAbout: p.projectAbout,
                projectWhenCreated: new Date(),
                projectWhenModified: new Date(),
                projectAuthors: p.projectAuthors,
                projectDOI: p.projectDOI,
                projectType: p.projectType,
                projectConfig: JSON.stringify(p.projectConfig),
                projectDigest: digest,
                projectFolder: p.projectFolder
            };
            if ( p.projectWhenPub ) {
                ins.projectWhenPub = p.projectWhenPub;
            }
            return this.query(`INSERT INTO Project SET ?`, p)
                .then(result => result.insertId);
        },
        canEditProject: function(projectId) {
            return this.isSuper()
                .catch(() => {
                    return this.query(`SELECT projectPerson, projectGroup FROM Project WHERE projectId=${p.projectId};`)
                        .then(pickFirst('No such project'))
                        .then(dbp => {
                            if ( this.person.personId === db.modelPerson ) {
                                return dbp;
                            }
                            if ( (this.person.personGroup === db.modelGroup) && this.person.personIsAdmin ) {
                                return dbp;
                            }
                            throw new Error('Insufficient permissions');
                        });
                });
        },
        canSeeProject: function(projectId, p) {
            if ( p ) {
                return (this.person.personIsSuper || p.projectPublic || (this.person.personGroup === p.projectGroup)) ? Promise.resolve(true) :
                    Promise.reject(new Error('Insufficient permissions'));
            }
            return this.isSuper()
                .catch(() => {
                    return this.query(`SELECT projectId, projectGroup, projectPublic FROM Project WHERE projectId=${projectId}`)
                        .then(pickFirst('No such project'))
                        .then(dbp => {
                            if ( dbp.projectPublic || (this.person.personGroup === dbp.projectGroup) ) {
                                return dbp;
                            }
                            throw new Error('Insufficient permissions');
                        });
                });
        },
        updateProject: function(p) {
            return this.canEditProject(p.projectId)
                .then(pFull => {
                    let updates = [];
                    let dateToo = dateOnly;
                    let doField = (name, trivial) => {
                        const x = p[name];
                        if ( typeof(x) !== 'undefined' ) {
                            updates.push(`${name}=${mysql.escape(x)}`);
                            dateToo = dateToo || ! trivial;
                        }
                    };
                    doField('projectPublic', true);
                    ['projectAbout', 'projectWhenPub', 'projectAuthors', 'projectDOI', 'projectFolder'].forEach(doField);
                    if ( 'projectConfig' in p ) {
                        updates.push(`projectConfig=${mysql.escape(JSON.stringify(p.projectConfig))}`);
                        dateToo = true;
                    }
                    if ( dateToo ) {
                        updates.push(`projectWhenModified=NOW()`);
                    }
                    updates.length && updates.push(`projectDigest=${mysql.escape(this.digestProject(Object.assign({}, pFull, p)))}`);
                    return this.query(`UPDATE Project SET ${updates.join(', ')} WHERE projectId=${p.projectId};`);
                });
        },
        updateProjectDate: function(projectId) { // assuming you're already vetted
            return this.query(`UPDATE Project SET projectWhenModified=NOW() WHERE projectId=${projectId};`);
        },
        addProjectModel: function(projectId, modelId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`INSERT INTO ProjectModel SET ?`, {
                    pmProject: projectId,
                    pmModel: modelId,
                    pmActive: true
                })).then(() => this.updateProjectDate(projectId));
        },
        removeProjectModel: function(pmId, projectId, modelId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`DELETE FROM ProjectModel WHERE pmId=${pmId} AND pmProject=${projectId} AND pmModel=${modelId};`))
                .then(() => this.updateProjectDate(projectId));
        },
        getProjectModels: function(projectId, p) {
            return this.canSeeProject(projectId, p)
                .then(() => this.query(`SELECT * FROM ProjectModel INNER JOIN Model ON pmModel=modelId WHERE pmProject=${projectId};`))
                .then(rows => rows.map(this.modelFromDB));
        },
        setProjectModelActive: function(projectId, modelId, active) {
            return this.canEditProject(projectId)
                .then(() => this.query(`UPDATE ProjectModel SET pmActive=${mysql.escape(active)} WHERE pmProject=${projectId} AND pmModel=${modelId};`))
                .then(() => this.updateProjectDate(projectId));
        },
        addProjectDataset: function(projectId, datasetId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`INSERT INTO ProjectDataset SET ?`, {
                    pdProject: projectId,
                    pdDataset: datasetId,
                    pdActive: true
                })).then(() => this.updateProjectDate(projectId));
        },
        removeProjectDataset: function(pdId, projectId, datasetId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`DELETE FROM ProjectDataset WHERE pdId=${pdId} AND pdProject=${projectId} AND pdDataset=${datasetId};`))
                .then(() => this.updateProjectDate(projectId));
        },
        getProjectDatasets: function(projectId, p) {
            return this.canSeeProject(projectId, p)
                .then(() => this.query(`SELECT * FROM ProjectDataset INNER JOIN Dataset ON pdDataset=datasetId WHERE pdProject=${projectId};`))
                .then(rows => rows.map(this.datasetFromDB));
        },
        setProjectDatasetActive: function(projectId, datasetId, active) {
            return this.canEditProject(projectId)
                .then(() => this.query(`UPDATE ProjectDataset SET pdActive=${mysql.escape(active)} WHERE pdProject=${projectId} AND pdDataset=${datasetId};`))
                .then(() => this.updateProjectDate(projectId));
        },
        addProjectResult: function(r) {
            return this.canEditProject(r.resultProject)
                .then(() => this.query(`INSERT INTO Result Set ?`, {
                    resultProject: r.resultProject,
                    resultConfig: JSON.stringify(r.resultConfig),
                    resultLL: r.resultLL,
                    resultWhen: r.resultWhen,
                    resultResults: JSON.stringify(r.resultResults)
                })).then(result => {
                    this.updateProjectDate(r.resultProject);
                    return result.insertId;
                });
        },
        delProjectResult: function(projectId, resultId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`DELETE FROM Result WHERE resultId=${resultId} AND resultProject=${projectId};`));
        },
        listProjectResults: function(projectId, p) {
            return this.canSeeProject(projectId, p)
                .then(() => this.query(`SELECT resultId, resultProject FROM Result WHERE resultProject=${projectId};`))
                .then(rows => rows.map(r => r.resultId));
        },
        resultFromDB: r => {
            try {
                r.resultConfig = JSON.parse(r.resultConfig);
                r.resultResults = JSON.parse(r.resultResults);
            } catch (err) {}
            return r;
        },
        getProjectResult: function(projectId, resultId) {
            if ( this.person.personId <= 0 ) {
                return Promise.reject(new Error('Insufficient credentials'));
            }
            return this.query(`SELECT resultConfig, resultLL, resultResults FROM Result WHERE resultProject=${projectId} AND resultId=${resultId};`)
                .then(rows => rows.map(this.resultFromDB));
        },
        listProjects: function(q) {
            let where = [];
            q.personId && where.push(`projectPerson=${q.personId}`);
            q.groupId && where.push(`projectGroup=${q.groupId}`);
            q.projectType && where.push(`projectType=${mysql.escape(q.projectType)}`);
            q.projectDigest && where.push(`(projectDigest LIKE ${mysql.escape('%'+q.projectDigest.toLowerCase()+'%')})`);
            q.projectFolder && where.push(`projectFolder=${mysql.escape(q.projectFolder)}`);
            /// limit chronologically
            if ( ! this.person.personIsSuper ) {
                where.push(`(projectPublic=TRUE OR projectGroup=${this.person.personGroup})`);
            }
            if ( ! (q.datasetId || q.modelId) ) {
                return this.query(`SELECT * FROM Project ${where.length ? 'WHERE ' : ''}${where.join(',')};`);
            }
            let joins = [];
            if ( q.datasetId ) {
                joins.push('INNER JOIN ProjectDataset ON pdProject=projectId');
                where.push(`pdDataset=${q.datasetId}`);
            }
            if ( q.modelId ) {
                joins.push('INNER JOIN ProjectModel ON pmProject=projectId');
                where.push(`pmModel=${q.modelId}`);
            }
            return this.query(`SELECT * FROM Project ${joins.join(' ')} WHERE ${where.join(' AND ')};`);
        },
        delProject: function(projectId) {
            return this.canEditProject(projectId)
                .then(() => this.query(`DELETE FROM Result WHERE resultProject=${projectId};`))
                .then(() => this.query(`DELETE FROM ProjectDataset WHERE pdProject=${projectId};`))
                .then(() => this.query(`DELETE FROM ProjectModel WHERE pmProject=${projectId};`))
                .then(() => this.query(`DELETE FROM Project WHERE projectId=${projectId};`));
        }

        
    };

    const TicketResponseMethods = {
        init: function(ws, tkt, msg) {
            this.ws = ws;
            this.tkt = tkt;
            this.msg = msg;
            return this;
        },
        send: function(r) {
            console.log('|  s> ', r);
            this.ws.send(JSON.stringify(r));
            return r;
        },
        sendErr: function(err) {
            console.log('|  e> ', err);
            this.send({
                t: 'Error',
                err: {
                    message: err.message || err,
                    stack: err.stack
                }
            });
            return err;
        },
        reply: function(r) {
            console.log('|  '+this.tkt+' => ', r);
            r.tkt = this.tkt;
            this.ws.send(JSON.stringify(Object.assign({
                tkt: this.tkt
            }, r)));
            return r;
        },
        replyEmpty: function(r) {
            console.log('|  '+this.tkt+' => ');
            this.ws.send(JSON.stringify({
                tkt: this.tkt
            }));
            return r;
        },
        replyInField: function(f) {
            return r => {
                console.log('|  '+this.tkt+' => ('+f+')',r);
                const rmsg = {tkt: this.tkt};
                rmsg[f] = r;
                this.ws.send(JSON.stringify(rmsg));
                return rmsg;
            };
        },
        replyErr: function(err, label) {
            if ( label ) {
                console.log(label+':err:',err);
                console.log(err.stack);
            }
            this.tkt ? this.reply({
                tktErr: {
                    message: err.message,
                    stack: err.stack
                }
            }) : this.sendErr(err);
        }
    };
    const TicketRouterMethods = {
        init: function(ws, label) {
            this.ws = ws;
            this._label = label;
            this._using = {};
            return this;
        },
        send: TicketResponseMethods.send,
        sendErr: TicketResponseMethods.sendErr,
        use: function(t, act) {
            if ( this._using[t] ) {
                console.log('warning: '+this._label+' method '+t+' is already defined');
            }
            this._using[t] = act;
        },
        hasRoute: function(t) {
            return !! this._using[t];
        },
        useDefault: function(act) {
            this._default = act;
        },
        route: function(data, flags, msg, response) {
            if ( msg ) {
                console.log('|  '+(msg.tkt||'')+' > ', msg.t, msg);
                const act = this._using[msg.t];
                if ( act ) {
                    return Promise.resolve(act(msg, response) || response.replyEmpty());
                }
            }
            if ( this._default ) {
                return Promise.resolve(this._default(data, flags, msg, response));
            }
            return Promise.reject(new Error('Unknown message type:'+msg.t+': '+JSON.stringify(msg)));
        }            
    };
    const newTicketRouter = (ws, label) => {
        label = label ? ('' + label) : 'TicketRouter';
        const router = (data, flags) => {
            let msg;
            try {
                msg = JSON.parse(data);
            } catch(err) {}
            const response = Object.create(TicketResponseMethods).init(ws, msg && msg.tkt, msg);
            return router.route(data, flags, msg, response)
                .catch(err => response.replyErr(err, label));
        };
        return Object.assign(router, TicketRouterMethods).init(ws, label);
    };
    
    const msgStr = (msg, f) => (msg[f] ? (''+msg[f]) : '');
    const copyIfStr = (msg, out, field) => {
        if ( field in msg ) { out[field] = ''+msg[field]; }
    };
    const copyIfBool = (msg, out, field) => {
        if ( field in msg ) { out[field] = !! msg[field]; }
    };
    const copyIfInt = (msg, out, field) => {
        if ( field in msg ) { out[field] = msg[field]|0; }
    };
    const copyIfRaw = (msg, out, field) => {
        if ( field in msg ) { out[field] = msg[field]; }
    };

    exports.router = express.Router();
    exports.setupWS = () => {
        const pool = exports.Connect();

        exports.router.ws("/ui", (ws, req) => {
            const db = new exports.Session({
                pool: pool,
                onhopefuls: hopefuls => router.send({
                    t: 'screenHopefuls',
                    hopefuls: hopefuls
                }),
                oneject: reason => router.send({
                    t: 'ejected',
                    reason: reason
                }),
                onaccepted: person => router.send({
                    t: 'accepted',
                    person: person
                })
            });
            const router = newTicketRouter(ws, 'online:cloud:ui');
            let conv = {};
            //ws.binaryType = 'arraybuffer';
            ws.on('close', () => {
                console.log('ui session closed',db.person.personId, db.seatKey);
                db.finalize();
                removeSession(db);
            });
            ws.on('error', err => {
                console.log('ui session error',db.person.personId, db.seatKey, err);
                console.log(err.stack);
                db.finalize();
                removeSession(db);
            });
            ws.on('message', router);
            addSession(db);

            router.use('ping', (msg, resp) => resp.reply(msg));
            router.use('log', (msg, resp) => {
                console.log('online:cloud:ui:'+db.seatKey+':',msg);
            });
            router.use('addGroup', (msg, resp) => db.addGroup({
                groupName: msgStr(msg, 'groupName'),
                groupAbout: msgStr(msg, 'groupAbout'),
                groupMaxPerson: msg.groupMaxPerson|0,
                groupMaxAdmin: msg.groupMaxAdmin|0,
                groupMaxSeats: msg.groupMaxSeats|0,
                groupComputeMinutes: msg.groupComputeMinutes|0
            }).then(resp.replyInField('groupId')));
            router.use('listGroups', (msg, resp) => db.listGroups(msg.groupId ? (msg.groupId|0) : undefined)
                       .then(resp.replyInField('groups')));
            router.use('setGroupActive', (msg, resp) => db.setGroupActive(msg.groupId|0, !!msg.groupActive)
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupName', (msg, resp) => db.setGroupName(msg.groupId|0, msgStr(msg, 'groupName'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupAbout', (msg, resp) => db.setGroupAbout(msg.groupId|0, msgStr(msg, 'groupAbout'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupDataKey', (msg, resp) => db.setGroupDataKey(msg.groupId|0, msgStr(msg, 'groupDataKey'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupComputeKey', (msg, resp) => db.setGroupComputeKey(msg.groupId|0, msgStr(msg, 'groupComputeKey'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupMaxPerson', (msg, resp) => db.setGroupMaxPerson(msg.groupId|0, msg.groupMaxPerson|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupMaxAdmin', (msg, resp) => db.setGroupMaxAdmin(msg.groupId|0, msg.groupMaxAdmin|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupMaxSeats', (msg, resp) => db.setGroupMaxSeats(msg.groupId|0, msg.groupMaxSeats|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('setGroupComputeMinutes', (msg, resp) => db.setGroupComputeMinutes(msg.groupId|0, msg.groupComputeMinutes|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('addPerson', (msg, resp) => db.addPerson({
                personGroup: msg.personGroup|0,
                personIsAdmin: !! msg.personIsAdmin,
                personIsContact: !! msg.personIsContact,
                personIsSuper: !! msg.personIsSuper,
                personName: msgStr(msg, 'personName'),
                personAbout: msgStr(msg, 'personAbout'),
                personEmail: msgStr(msg, 'personEmail')
            }).then(resp.replyInField('personId')));
            router.use('listPeople', (msg, resp) => db.listPeople(msg.groupId ? (msg.groupId|0) : undefined)
                       .then(resp.replyInField('people')));
            router.use('setPersonGroup', (msg, resp) => db.setPersonGroup(msg.personId|0, msg.groupId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonActive', (msg, resp) => db.setPersonActive(msg.personId|0, !!msg.personActive)
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonIsAdmin', (msg, resp) => db.setPersonIsAdmin(msg.personId|0, !!msg.personIsAdmin)
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonIsContact', (msg, resp) => db.setPersonIsContact(msg.personId|0, !!msg.personIsContact)
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonIsSuper', (msg, resp) => db.setPersonIsSuper(msg.personId|0, !!msg.personIsSuper)
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonName', (msg, resp) => db.setPersonName(msg.personId|0, msgStr(msg, 'personName'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonAbout', (msg, resp) => db.setPersonAbout(msg.personId|0, msgStr(msg, 'personAbout'))
                       .then(x => resp.replyEmpty(x)));
            router.use('setPersonEmail', (msg, resp) => db.setPersonEmail(msg.personId|0, msgStr(msg, 'personEmail'))
                       .then(x => resp.replyEmpty(x)));
            router.use('requestAccount', (msg, resp) => db.requestAccount({
                groupInfo: msgStr(msg, 'groupInfo'),
                name: msgStr(msg, 'name'),
                email: msgStr(msg, 'email'),
                about: msgStr(msg, 'about')
            }, req.ip, req.headers['user-agent'])
                       .then(resp.replyInField('seatKey')));
            router.use('listHopeful', (msg, resp) => db.listHopeful(msg.groupId|0)
                       .then(resp.replyInField('hopefuls')));
            router.use('fwdHopeful', (msg, resp) => db.fwdHopeful(msg.hopefulId|0, msg.groupId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('bwdHopeful', (msg, resp) => db.bwdHopeful(msg.hopefulId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('acceptHopeful', (msg, resp) => db.acceptHopeful(msg.hopefulId|0, !!msg.personIsAdmin, !!msg.personIsContact)
                       .then(resp.replyInField('personId')));
            router.use('acceptCandidate', (msg, resp) => db.acceptCandidate(msg.hopefulId|0, msg.personId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('rejectHopeful', (msg, resp) => db.rejectHopeful(msg.hopefulId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('invite', (msg, resp) => db.invite(msg.personId|0, msg.expiresHours|0, msg.usesRemain|0)
                       .then(resp.replyInField('inviteKey')));
            router.use('auth', (msg, resp) => db.auth({
                seatKey: msgStr(msg, 'seatKey'),
                isAdmin: !! msg.isAdmin,
                isSuper: !! msg.isSuper
            }, req.ip, req.headers['user-agent'])
                       .then(x => resp.replyEmpty(x)));
            router.use('signIn', (msg, resp) => db.signIn(msgStr(msg, 'inviteKey'), req.ip, req.headers['user-agent'])
                       .then(resp.replyInField('seatKey')));
            router.use('signOut', (msg, resp) => db.signOut()
                       .then(x => resp.replyEmpty(x)));
            router.use('aboutPerson', (msg, resp) => db.aboutPerson(msg.personId|0)
                       .then(x => resp.reply(x)));
            router.use('signedIn', (msg, resp) => ((db.person.personId > 0) ? null : Promise.reject(new Error('Not signed in'))));
            router.use('listSeats', (msg, resp) => db.listSeats(msg.groupId ? (msg.groupId|0) : undefined)
                       .then(resp.replyInField('seats')));
            router.use('unSignIn', (msg, resp) => db.unSignIn(msgStr(msg, 'seatKey'), msg.otherPersonId ? (msg.otherPersonId|0) : undefined)
                       .then(x => resp.replyEmpty(x)));
            router.use('addModel', (msg, resp) => {
                let m = {
                    modelPublic: !! msg.modelPublic,
                    modelName: msgStr(msg, 'modelName'),
                    modelQmj: msg.modelQmj,
                    modelAbout: msgStr(msg, 'modelAbout'),
                    modelAuthors: msgStr(msg, 'modelAuthors'),
                    modelDOI: msgStr(msg, 'modelDOI'),
                    modelStates: msg.modelStates|0,
                    modelClasses: msg.modelClasses|0,
                    modelFolder: msgStr(msg, 'modelFolder')
                };
                copyIfRaw(msg, m, 'modelWhenPub');
                return db.addModel(m)
                    .then(resp.replyInField('modelId'));
            });
            router.use('updateModel', (msg, resp) => {
                let upd = {
                    modelId: msg.modelId|0
                };
                copyIfBool(msg, upd, 'modelPublic');
                copyIfStr(msg, upd, 'modelName');
                copyIfRaw(msg, upd, 'modelQmj');
                copyIfStr(msg, upd, 'modelAbout');
                copyIfRaw(msg, upd, 'modelWhenPub');
                copyIfStr(msg, upd, 'modelAuthors');
                copyIfStr(msg, upd, 'modelDOI');
                copyIfInt(msg, upd, 'modelStates');
                copyIfInt(msg, upd, 'modelClasses');
                copyIfStr(msg, upd, 'modelFolder');
                return db.updateModel(upd)
                    .then(x => resp.replyEmpty(x));
            });
            router.use('listModels', (msg, resp) => {
                let q = {};
                copyIfInt(msg, q, 'modelPerson');
                copyIfInt(msg, q, 'modelGroup');
                copyIfInt(msg, q, 'modelStates');
                copyIfInt(msg, q, 'modelClasses');
                copyIfStr(msg, q, 'modelDigest');
                copyIfStr(msg, q, 'modelFolder');
                return db.listModels(q)
                    .then(resp.replyInField('models'));
            });
            router.use('getModel', (msg, resp) => db.getModel(msg.modelId|0)
                       .then(x => resp.reply(x)));
            router.use('setDatasetPublic', (msg, resp) => db.setDatasetPublic(msg.datasetId|0, !!msg.datasetPublic)
                       .then(x => resp.replyEmpty(x)));
            router.use('setDatasetFolder', (msg, resp) => db.setDatasetFolder(msg.datasetId|0, msgStr(msg, 'datasetFolder'))
                       .then(x => resp.replyEmpty(x)));
            router.use('listServers', (msg, resp) => db.listServers(msg.groupId|0)
                       .then(resp.replyInField('servers')));
            router.use('listDatasets', (msg, resp) => {
                let q = {};
                copyIfInt(msg, q, 'personId');
                copyIfInt(msg, q, 'groupId');
                copyIfStr(msg, q, 'datasetDigest');
                copyIfStr(msg, q, 'datasetFolder');
                return db.listDatasets(q)
                    .then(resp.replyInField('datasets'));
            });
            router.use('pickDataInstance', (msg, resp) => db.pickInstanceServer(msg.datasetId|0)
                       .then(server => resp.reply({
                           serverBaseUrl: server.server.baseUrl,
                           instancePath: server.instancePath
                       })));
            router.use('addProject', (msg, resp) => {
                let p = {
                    projectPublic: !! msg.projectPublic,
                    projectName: msgStr(msg, 'projectName'),
                    projectAbout: msgStr(msg, 'projectAbout'),
                    projectAuthors: msgStr(msg, 'projectAuthors'),
                    projectDOI: msgStr(msg, 'projectDOI'),
                    projectType: msgStr(msg, 'projectType'),
                    projectConfig: msg.projectConfig,
                    projectFolder: msgStr(msg, 'projectFolder')
                };
                copyIfRaw(msg, p, 'projectWhenPub');
                return db.addProject(p)
                    .then(resp.replyInField('projectId'));
            });
            router.use('updateProject', (msg, resp) => {
                let upd = {
                    projectId: msg.projectId|0
                };
                copyIfStr(msg, upd, 'projectAbout');
                copyIfRaw(msg, upd, 'projectWhenPub');
                copyIfStr(msg, upd, 'projectAuthors');
                copyIfStr(msg, upd, 'projectDOI')
                copyIfRaw(msg, upd, 'projectConfig');
                copyIfStr(msg, upd, 'projectFolder');
                return db.updateProject(upd)
                    .then(x => resp.replyEmpty(x));
            });
            router.use('addProjectModel', (msg, resp) => db.addProjectModel(msg.projectId|0, msg.modelId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('removeProjectModel', (msg, resp) => db.removeProjectModel(msg.pmId|0, msg.projectId|0, msg.modelId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('getProjectModels', (msg, resp) => db.getProjectModels(msg.projectId|0)
                       .then(resp.replyInField('models')));
            router.use('setProjectModelActive', (msg, resp) => db.setProjectModelActive(msg.projectId|0, msg.modelId|0, !!msg.active)
                       .then(x => resp.replyEmpty(x)));
            router.use('addProjectDataset', (msg, resp) => db.addProjectDataset(msg.projectId|0, msg.datasetId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('removeProjectDataset', (msg, resp) => db.removeProjectDataset(msg.pdId|0, msg.projectId|0, msg.datasetId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('getProjectDatasets', (msg, resp) => db.getProjectDatasets(msg.projectId|0)
                       .then(resp.replyInField('datasets')));
            router.use('setProjectDatasetActive', (msg, resp) => db.setProjectDatasetActive(msg.projectId|0, msg.datasetId|0, !!msg.active)
                       .then(x => resp.replyEmpty(x)));
            router.use('addProjectResult', (msg, resp) => db.addProjectResult({
                resultProject: msg.resultProject|0,
                resultConfig: msg.resultConfig || {},
                resultLL: +msg.resultLL,
                resultWhen: msg.resultWhen || new Date(),
                resultResults: msg.resultResults || {}
            }).then(resp.replyInField('resultId')));
            router.use('delProjectResult', (msg, resp) => db.delProjectResult(msg.projectId|0, msg.resultId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('listProjectResults', (msg, resp) => db.listProjectResults(msg.projectId|0)
                       .then(resp.replyInField('results')));
            router.use('getProjectResult', (msg, resp) => db.getProjectResult(msg.projectId|0, msg.resultId|0)
                       .then(x => resp.reply(x)));
            router.use('listProjects', (msg, resp) => {
                let q = {};
                copyIfInt(msg, q, 'personId');
                copyIfInt(msg, q, 'groupId');
                copyIfStr(msg, q, 'projectType');
                copyIfStr(msg, q, 'projectDigest');
                copyIfInt(msg, q, 'datasetId');
                copyIfInt(msg, q, 'modelId');
                copyIfInt(msg, q, 'projectFolder');
                return db.listProjects(q)
                    .then(resp.replyInField('projects'));
            });
            router.use('delProject', (msg, resp) => db.delProject(msg.projectId|0)
                       .then(x => resp.replyEmpty(x)));
            router.use('startUpload', (msg, resp) => db.groupOnline.pickUploadServer(msg.bytes|0)
                       .then(server => {
                           const fileKey = uuid.v4();
                           return db.addDataset({
                               datasetPublic: !! msg.datasetPublic,
                               datasetGroup: db.group.groupId,
                               datasetName: ''+msg.datasetName,
                               datasetAbout: ''+msg.datasetAbout,
                               datasetWhenRec: msg.datasetWhenRec,
                               datasetSampling: +msg.datasetSampling,
                               datasetPoints: msg.datasetPoints|0,
                               datasetEvents: msg.datasetEvents|0,
                               datasetSegments: msg.datasetSegments|0,
                               datasetStimuli: msg.datasetStimuli || {},
                               datasetConstants: msg.datasetConstants || {},
                               datasetMolecule: ''+msg.datasetMolecule,
                               datasetEtc: msg.datasetEtc || {},
                               datasetFolder: msgStr(msg, 'datasetFolder')
                           }).then(datasetId => server.offer(datasetId, fileKey, msg.datasetBytes|0, msg.datasetSegments|0))
                               .then(path => resp.reply({
                                   datasetId: datasetId,
                                   fileKey: fileKey,
                                   serverId: server.serverId,
                                   serverBaseUrl: server.baseUrl,
                                   serverPutUrl: server.putUrl,
                                   path: path
                               }));
                       }));
            router.use('uploading', (msg, resp) => db.groupOnline.getServerById(msg.serverId|0)
                       .then(server => server.uploading(msg.datasetId|0, ''+msg.fileKey))
                       .then(x => resp.replyEmpty(x)));
            router.use('uploadFailed', (msg, resp) => db.groupOnline.getServerById(msg.serverId|0)
                       .then(server => server.uploadFailed(msg.datasetId|0, ''+msg.fileKey))
                       .then(x => resp.replyEmpty(x)));
            router.use('replaceIdl', (msg, resp) => db.pickInstanceServer(msg.datasetId)
                       .then(server => {
                           const fileKey = uuid.v4();
                           return server.server.replaceIdl(msg.datasetId, fileKey, server.instancePath)
                               .then(() => resp.reply({
                                   fileKey: fileKey,
                                   serverId: server.server.serverId,
                                   serverBaseUrl: server.server.baseUrl,
                                   serverPutUrl: server.server.putUrl
                               }));
                       }));
            router.use('replacing', (msg, resp) => db.groupOnline.getServerById(msg.serverId|0)
                       .then(server => server.replacing(msg.datasetId|0, ''+msg.fileKey))
                       .then(x => resp.replyEmpty(x)));
            router.use('canPutFile', (msg, resp) => dataconv.canPutFile({
                name: ''+msg.name,
                size: msg.size|0
            }).then(x => resp.reply(x)));
            router.use('releaseFile', (msg, resp) => {
                dataconv.releaseFile(msg.url);
            });
            router.use('convertFile', (msg, resp) => new Promise((resolve, reject) => {
                conv = {
                    resolve: resolve,
                    reject: reject,
                    url: msg.url
                };
                dataconv.convert(conv).then(url => {
                    conv.resolve && conv.resolve(resp.replyInField('url'));
                    conv = {};
                    return url;
                }).then(resp.replyInField('url'))
                    .catch(err => {
                        conv.reject && conv.reject(err);
                        conv = {};
                    });
            }));
            router.use('cancelConvertFile', (msg, resp) => {
                if ( conv.reject ) {
                    conv.reject(new Error('Canceled'));
                    conv.canceled = true;
                    conv = {};
                }
                return resp.reply({});
            });
        });
        
        function PromiseReg() {
            this.map = new Map();
        };
        PromiseReg.prototype = {
            register: function(key, obj) {
                if ( this.map.has(key) ) {
                    return Promise.reject(new Error('Key in use'));
                }
                return new Promise((resolve, reject) => {
                    this.map.set(key, Object.assign({
                        resolve: resolve,
                        reject: reject
                    }, obj));
                });
            },
            pop: function(key) {
                let r = this.map.get(key);
                if ( r ) {
                    this.map.delete(key);
                } else {
                    r = {resolve: x=>0, reject: x=>0};
                }
                return r;
            },
            resolve: function(key, val) {
                let r = this.pop(key);
                r.resolve(val);
                return r;
            },
            reject: function(key, err) {
                let r = this.pop(key);
                r.reject(err);
                return r;
            },
            rejectAll: function(err) {
                for (let r of this.map) {
                    r[1].reject(err);
                }
                this.map.clear();
            }
        };
        
        exports.DataServer = function DataServer(config) {
            this.ws = config.ws;
            this.db = config.db;
            this.authorized = false;
            this.uuid = '';
            this.dataKey = '';
            this.group = {};
            this.groupOnline = null;
            this.serverId = '';
            this.baseUrl = '';
            this.putUrl = '';
            this.pending = new PromiseReg();
        };
        exports.DataServer.prototype = {
            finalize: function() {
                this.pending.rejectAll('DataServer socket closed');
                return true;
            },
            eject: function() {
                this.db.eject();
            },
            // messages from data server:
            auth: function(dataKey, uuid) {
                this.dataKey = dataKey;
                this.uuid = uuid;
                return this.db.authServer(dataKey, uuid)
                    .then(serverId => {
                        this.authorized = true;
                        this.serverId = serverId;
                        this.group = this.db.group;
                        this.groupOnline = groupsOnline.get(this.group.groupId);
                    });
            },
            online: function(coverage, freeMB, baseUrl, putUrl) {
                if ( putUrl ) {
                    this.putUrl = putUrl;
                }
                this.coverage = coverage;
                this.freeMB = freeMB;
                this.baseUrl = baseUrl;
                return this.db.updateServer(coverage, freeMB, baseUrl);
            },
            offline: function() {

            },
            forgetMe: function() {
                this.db.forgetServer(this.serverId);
            },
            accepted: function(datasetId, path) {
                this.pending.resolve(datasetId, path);
            },
            rejected: function(datasetId, err) {
                this.pending.reject(datasetId, err);
            },
            instanced: function(datasetId, sum, bytes, path) {
                this.db.addDataInstance(datasetId, sum, bytes, path)
                    .then(instanceId => {
                        this.pending.resolve(datasetId, instanceId);
                        this.db.countInstances(datasetId).then(count => {
                            if ( count === 1 ) { // tell other servers to replicate
                                this.db.listDatasets({datasetId}).then(pickFirst())
                                    .then(ds => this.groupOnline.servers.map(server => server.newDataset(datasetId, ds.datasetSegments, ds.datasetPublic, ds.datasetWhenIdl, this.baseUrl, path, sum, bytes)));
                            }
                        });
                    });
            },
            idlReplaced: function(datasetId, whenIdl) {
                return this.db.idlReplaced(datasetId, this.serverId, whenIdl)
                    .then(() => this.pending.resolve(datasetId, true));
            },
            instanceMoved: function(datasetId, instanceId, path) {
                this.db.moveDataInstance(datasetId, instanceId, path);
            },
            // messages to data server:
            offer: function(datasetId, fileKey, bytes, nSegment) {
                this.router.send({
                    t: 'offer',
                    datasetId: datasetId,
                    fileKey: fileKey,
                    bytes: bytes,
                    segments: nSegment
                });
                return pending.register(datasetId); // -> accepted/rejected
            },
            uploadFailed: function(datasetId, fileKey) {
                let p = this.pending.reject(datasetId, 'Upload failed');
                p.replaceIdl || this.db.delDataset(datasetId);
            },
            uploading: function(datasetId, fileKey) {
                return this.pending.register(datasetId); // -> instanced/uploadFailed
            },
            replaceIdl: function(datasetId, fileKey, instancePath) {
                this.router.send({
                    t: 'replaceIdl',
                    datasetId: datasetId,
                    fileKey: fileKey,
                    instancePath: instancePath
                });
                return pending.register(datasetId, {replaceIdl: true}); // -> accepted/rejected
            },
            replacing: function(datasetId, fileKey) {
                return this.pending.register(datasetId, {replaceIdl: true}); // -> idlReplaced/uploadFailed
            },
            idlReplacement: function(datasetId, instancePath, newUrl) {
                this.router.send({
                    t: 'idlReplacement',
                    datasetId: datasetId,
                    instancePath: instancePath,
                    newUrl: newUrl
                });
            },
            newDataset: function(datasetId, datasetSegments, datasetPublic, datasetWhenIdl, serverBaseUrl, instancePath, instanceSum, instanceBytes) {
                this.router.send({
                    t: 'newDataset',
                    datasetId,
                    datasetSegments,
                    datasetPublic,
                    datasetWhenIdl,
                    serverBaseUrl,
                    instancePath,
                    instanceSum,
                    instanceBytes
                });
            },
            publicChanged: function(datasetId, datasetPublic) {
                this.router.send({
                    t: 'public',
                    datasetId: datasetId,
                    isPublic: datasetPublic
                });
            }
        };

        exports.router.ws("/data", (ws, req) => {
            const db = new exports.Session({
                pool: pool,
                oneject: reason => router.send({
                    t: 'ejected',
                    reason: reason
                })
            });
            const server = new exports.DataServer({
                db: db,
                ws: ws
            });
            const finalize = () => {
                if ( server.authorized ) {
                    server.groupOnline.removeServer(server);
                }
                server.finalize();
            };
            const router = newTicketRouter(ws, 'online:cloud:data');
            ws.on('close', finalize);
            ws.on('error', finalize);
            ws.on('message', router);
            router.use('ping', (msg, resp) => resp.send(msg));
            router.use('log', (msg, resp) => {
                console.log('online:cloud:data:'+db.serverId+':',msg);
                return true; // true: no reply
            });
            router.use('auth', (msg, resp) => server.auth(''+msg.dataKey, ''+msg.uuid)
                       .then(() => {
                           addAuthRoutes();
                           return resp.send({t: 'auth'});
                       }).catch(err => resp.sendErr(err)));
            function addAuthRoutes() {
                if ( (! server.authorized) || router.hasRoute('online') ) {
                    return;
                }
                router.use( 'online', (msg, resp) => server.online(+msg.coverage, msg.freeMB|0, ''+msg.baseUrl, ''+msg.putUrl)
                            .then(() => {
                                server.groupOnline.addServer(server);
                                db.updateGroupFreeMB(server.groupOnline.updateFreeMB());
                                return true;
                            }).catch(err => resp.sendErr(err)));
                router.use('offline', (msg, resp) => {
                    server.offline()
                    return finalize();
                });
                router.use('forgetMe', (msg, resp) => {
                    server.forgetMe();
                    return finalize();
                });
                router.use('accepted', (msg, resp) => {
                    server.accepted(msg.datasetId|0, ''+msg.path);
                    return true;
                });
                router.use('rejected', (msg, resp) => {
                    server.rejected(msg.datasetId|0, ''+msg.err);
                    return true;
                });
                router.use('instanced', (msg, resp) => {
                    server.instanced(msg.datasetId|0, msg.sum|0, msg.bytes|0, ''+msg.path);
                    return true;
                });
                router.use('idlReplaced', (msg, resp) => {
                    server.idlReplaced(msg.datasetId|0, new Date(msg.datasetWhenIdl));
                    return true;
                });
                router.use('instanceMoved', (msg, resp) => {
                    server.instanceMoved(msg.datasetId|0, msg.instanceId|0, ''+msg.path);
                    return true;
                });
                router.use('listDatasets', (msg, resp) => db.listDatasets({
                    groupId: db.group.groupId
                }).then(datasets => resp.send({
                    t: 'datasets',
                    datasets: datasets
                })));
                router.use('pickInstance', (msg, resp) => db.pickInstanceServer(msg.datasetId|0)
                           .then(server => resp.send({
                               t: 'instance',
                               datasetId: msg.datasetId,
                               serverBaseUrl: server.server.baseUrl,
                               instancePath: server.instancePath,
                               instanceSum: server.instanceSum,
                               instanceBytes: server.instanceBytes
                           })));
            }
        });
    };

    return exports;
});

/// query with chronological limits

/// should data conversion be users-only?
///   make the group run its own dataconv node?

/// periodically clear out expired seats and invites?
