"use strict";

define("dataconv", ["express", "body-parser", "fs", "connect-busboy", "node-uuid", 'path'], function(express, bodyParser, fs, busboy, uuid, path) {
    'use strict';

    const STATIC_BASE = '/var/qub-online/data';
    const URL_BASE = 'https://qub.mandelics.com/online/data/';

    let converters = [];
    let working = {};
    let finished = [];

    const router = express.Router();
    router.use(bodyParser.urlencoded({ extended: true }));

    function setupWS() {
        router.ws('/conv', (ws, req) => {
            let converting = null;
            let me = {
                convert: function(url) {
                    if ( converting ) {
                        return Promise.reject('Busy');
                    }
                    return new Promise(function(resolve, reject) {
                        converting = {
                            url: url,
                            id: uuid.v4(),
                            resolve: resolve,
                            reject: reject
                        };
                        working[converting.id] = me;
                        ws.send(JSON.stringify({
                            t: 'next',
                            url: url,
                            id: converting.id
                        }));
                    });
                },
                cancel: function() {
                    converting && ws.send(JSON.stringify({t: 'cancel'}));
                },
                doneConv: function(url) {
                    if ( converting ) {
                        converting.resolve(url);
                        delete working[converting.id];
                        converters.push(me);
                        converting = null;
                    }
                }
            };
            converters.push(me);
            function onTrouble(err) {
                converting && converting.reject(err);
                if ( converting && working[converting.id] ) {
                    delete working[converting.id];
                }
                converting = null;
                let i = converters.indexOf(me);
                if ( i >= 0 ) {
                    converters.splice(i, 1);
                }
            }
            ws.on('close', () => {
                onTrouble('Closed');
            });
            ws.on('message', (data, flags) => {
                try {
                    data = JSON.parse(data);
                }
                catch (err)
                {}
                if ( data.t === 'err' ) {
                    console.log('converter error', data.err);
                    onTrouble(err);
                    converters.push(me);
                } else if ( data.t === 'ping' ) {
                    ws.send(JSON.stringify(data));
                }
            });
        });
    }

    const fileHandler = busboy();
    let filePerm = {};
    const fileRoutes = [
        {
            route: '/doneConv',
            lookup: working,
            outPath: function(id) { return path.join(STATIC_BASE, id+'.qdf'); },
            onDone: function(id, path, res) {
                let c = working[id];
                c && c.doneConv(URL_BASE+id+'.qdf');
                finished.push({
                    path: path,
                    date: Date.now()
                });
                res.json({t: 'OK'});
            }
        },
        {
            route: '/put',
            lookup: filePerm,
            outPath: function(id) { return path.join(STATIC_BASE, id+filePerm[id].ext); },
            onDone: function(id, path, res) {
                if ( filePerm[id] ) {
                    const ext = filePerm[id].ext;
                    delete filePerm[id];
                    res.json({
                        url: URL_BASE+id+ext
                    });
                }
            }
        }
    ];
    fileRoutes.forEach(function(r) {
        router.use(r.route, fileHandler);
        router.post(r.route, (req, res) => {
            if ( req.busboy ) {
                var fstream, id;
                req.pipe(req.busboy);
                req.busboy.on('field', (fieldname, val) => {
                    if ( fieldname === 'id' ) {
                        id = val;
                        if ( ! r.lookup[id] ) {
                            res.json({t: 'nevermind'});
                            id = 'X';
                        }
                    }
                });
                req.busboy.on('file', (fieldname, file, filename) => {
                    if ( id === 'X' ) {
                        return;
                    }
                    id = id || uuid.v4();
                    let filePath = r.outPath(id);
                    fstream = fs.createWriteStream(filePath);
                    file.pipe(fstream);
                    fstream.on('close', () => {
                        r.onDone(id, filePath, res);
                    });
                });
            }
        });
    });
    
    router.use("/", express.static(STATIC_BASE));
    
    const FILE_EXPIRE_MS = 3*60*60000;
    const GETCONV_WAIT_MS = 15000;

    setInterval(() => {    //// delete hours-old files
        const stale = Date.now() - FILE_EXPIRE_MS;
        let iFresh = 0;
        while ( (iFresh < finished.length) && (finished[iFresh].date <= stale) ) {
            fs.unlink(finished[iFresh].path);
            ++iFresh;
        }
        iFresh && finished.splice(0, iFresh);
    }, 5*60000);
    
    function getConverted(opts) {
        const url = opts.url || opts;
        if ( opts.canceled ) {
            return Promise.reject('Canceled');
        }
        return Promise.resolve(true).then(() => {
            if ( converters.length ) {
                let conv = converters.pop();
                return conv.convert(url);
            }
            return Promise.reject('no converters available');
        }).catch(err => {
            console.log('getConverted:',err);
            return new Promise(function(resolve, reject) {
                setTimeout(() => {
                    resolve(getConverted(opts));
                }, GETCONV_WAIT_MS);
            });
        });
    }

    function canPutFile(rec) { // {name, size in bytes}
        /// could reject on lack of free space
        const fileId = uuid.v4();
        filePerm[fileId] = {
            id: fileId,
            ext: path.extname(rec.name)
        };
        return Promise.resolve(filePerm[fileId]);
    }
    function releaseFile(url) {
        if ( url.indexOf(URL_BASE) === 0 ) {
            fs.unlink(path.join(STATIC_BASE, url.substring(URL_BASE.length)));
        }
    }
    
    return {
        setupWS: setupWS,
        router: router,
        convert: getConverted,
        canPutFile: canPutFile,
        releaseFile: releaseFile
    };
});
