"use strict";

define("modeldb", ["mysql", "node-uuid", "emailjs", "os"], function(mysql, uuid, emailjs, os) {
    let exports = {};

    function extend(target) {
        var sources = [].slice.call(arguments, 1);
        sources.forEach(function (source) {
            source && Object.getOwnPropertyNames(source).forEach(function(propName) {
                Object.defineProperty(target, propName,
                    Object.getOwnPropertyDescriptor(source, propName));
            });
        });
        return target;
    };
    
    let ConnectionConfig = {
        hostname: "localhost",
        user: "Models",
        password: "jihqprewgop",
        database: "Models",
        inviteHost: "www.qub.buffalo.edu",
        mainInstallPath: "/online",
        emailSender: "qub@www.qub.buffalo.edu",
        requestEmail: "qub@www.qub.buffalo.edu"
    };
    exports.Connection = function(config) {
        this.config = extend({}, ConnectionConfig, config);
        this.pool = mysql.createPool({
            connectionLimit: 100,
            host: this.config.hostname,
            user: this.config.user,
            password: this.config.password,
            database: this.config.database,
            charset: "utf8mb4_unicode_ci"
        });
    };
    exports.Connection.prototype.query = function(q, inserts) {
        return new Promise((resolve, reject) => {
            this.pool.getConnection((err,connection) => {
                if (err) {
                    reject("No database connection: "+err);
                    return;
                }
                connection.query(q, inserts||[], (err, rows) => {
                    connection.release();
                    if ( err ) {
                        reject("Database error: "+err);
                    } else {
                        resolve(rows);
                    }
                }).on('error', err => {
                    reject("Database error: "+err);
                    return;
                });
            });
        });
    };
    
    exports.Connection.prototype.signIn = function(inviteSecret, ip, agent) {
        return this.query(`SELECT inviteId, invitePerson, inviteExpires, inviteUsesRemain FROM Invite WHERE inviteSecret=${mysql.escape(inviteSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Invitation expired or unknown");
                }
                let inv = rows[0];
                let delInvite = () => {
                    this.query(`DELETE FROM Invite WHERE inviteId=${inv.inviteId};`)
                        .catch(function(err) {
                            console.log("Deleting invite: ", err);
                        });
                };
                if ( ((new Date()) - inv.inviteExpires) > 0 ) {
                    delInvite();
                    throw new Error("Invitation expired");
                }
                if ( inv.inviteUsesRemain > 1 ) {
                    this.query(`UPDATE Invite SET inviteUsesRemain=${inv.inviteUsesRemain-1} WHERE inviteId=${inv.inviteId};`)
                        .catch(err => {
                            console.log("Decrementing inviteUsesRemain: ", err);
                        });
                } else if ( inv.inviteUsesRemain === 1 ) {
                    delInvite();
                }
                let siSecret = uuid.v4();
                return this.query(`INSERT INTO Signin (signinPerson, signinSecret, signinWhen, signinIP, signinAgent) VALUES (${inv.invitePerson},${mysql.escape(siSecret)},NOW(),${mysql.escape(""+ip)},${mysql.escape(agent)});`)
                    .then(result => siSecret);
            });
    };
    exports.Connection.prototype.signOut = function(signinSecret) {
        return this.query(`DELETE FROM Signin WHERE signinSecret=${mysql.escape(signinSecret)};`);
    };
    exports.Connection.prototype.auth = function(q, ip, agent) {
        return this.query(`SELECT signinId, signinPerson FROM Signin WHERE signinSecret=${mysql.escape(""+q.signinSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Signin is invalid or expired");
                }
                this.query(`UPDATE Signin SET signinWhen=NOW(), signinIP=${mysql.escape(""+ip)}, signinAgent=${mysql.escape(agent)} WHERE signinId=${rows[0].signinId};`)
                    .catch(err => {
                        console.log("Updating signin info: ", err);
                    });
                return this.query(`SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person WHERE personId=${rows[0].signinPerson};`);
            })
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Missing person");
                }
                let p = rows[0];
                if ( ! p.personActive ) {
                    throw new Error("Account deactivated");
                }
                if ( (q.canSave && ! p.personCanSave)
                     || (q.canSelfInvite && ! p.personCanSelfInvite)
                     || (q.canInvite && ! p.personCanInvite)
                     || (q.canAdmin && ! p.personCanAdmin) ) {
                    throw new Error("Insufficent permissions");
                }
                return p;
            });
    };
    exports.Connection.prototype.aboutMe = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => ({
            personName: p.personName,
            personAbout: p.personAbout,
            personEmail: p.personEmail,
            personCanSave: p.personCanSave,
            personCanSelfInvite: p.personCanSelfInvite,
            personCanInvite: p.personCanInvite,
            personCanAdmin: p.personCanAdmin
        }));
    };
    exports.Connection.prototype.setAbout = function(signinSecret, ip, agent, about) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personAbout=${mysql.escape(about)} WHERE personId=${p.personId};`));
    };
    exports.Connection.prototype.setEmail = function(signinSecret, ip, agent, email) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personEmail=${mysql.escape(email)} WHERE personId=${p.personId};`));
    };
    exports.Connection.prototype.selfInvite = function(signinSecret, ip, agent, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canSelfInvite: true}, ip, agent).then(p => {
            let inviteSecret = uuid.v4();
            let expires = new Date();
            expires.setHours(expires.getHours() + expiresHours);
            return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${p.personId},${mysql.escape(expires)},${usesRemain});`).then(result => {
                this.sendInvite(inviteSecret, p);
                return {inviteSecret: inviteSecret};
            });
        });
    };
    exports.Connection.prototype.inviteTrusted = function(invitee, expiresHours, usesRemain) {
        return this.query(`INSERT INTO Person (personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin) VALUES (${mysql.escape(""+invitee.personName)},${mysql.escape(""+invitee.personAbout)},${mysql.escape(""+invitee.personEmail)},1,${invitee.personCanSave?1:0},${invitee.personCanSelfInvite?1:0},${invitee.personCanInvite?1:0},${invitee.personCanAdmin?1:0});`)
            .then(result => {
                let personId = result.insertId;
                let inviteSecret = uuid.v4();
                let expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Connection.prototype.invite = function(signinSecret, ip, agent, invitee, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canInvite: true}, ip, agent).then(p => {
            ["personCanSave", "personCanSelfInvite", "personCanInvite", "personCanAdmin"].forEach(perm => {
                if ( ! p[perm] ) {
                    invitee[perm] = false;
                }
            });
            return this.inviteTrusted(invitee, expiresHours, usesRemain);
        });
    };
    exports.Connection.prototype.sendInvite = function(inviteSecret, p) {
        let url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/signin.html#'+inviteSecret;
        emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
            .send({from: this.config.emailSender,
                   to: p.personName + ' <' + p.personEmail + '>',
                   subject: "Sign in to QuB Online",
                   text: `Use this link to sign your browser into QuB Online as ${p.personName}:

${url}
`
                   },
                  (err, msg) => {
                     if ( err ) {
                         console.log(`Sending invitation to ${p.personName} <${p.personEmail}>: `, err);
                     }
                 });
    };
    exports.Connection.prototype.listPeople = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query("SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person;"))
            .then(rows => ({
                people: rows
            }));
    };
    exports.Connection.prototype.reinvite = function(signinSecret, ip, agent, personId, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query(`SELECT personName, personEmail FROM Person WHERE personId=${personId};`))
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such person");
                }
                let invitee = rows[0];
                let inviteSecret = uuid.v4();
                let expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Connection.prototype.updatePerson = function(signinSecret, ip, agent, id, name, about, email, active, canSave, canSelfInvite, canInvite, canAdmin) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personName=${mysql.escape(name)}, personAbout=${mysql.escape(about)}, personEmail=${mysql.escape(email)}, personActive=${active?1:0}, personCanSave=${canSave?1:0}, personCanSelfInvite=${canSelfInvite?1:0}, personCanInvite=${canInvite?1:0}, personCanAdmin=${canAdmin?1:0} WHERE personId=${id};`));
    };
    exports.Connection.prototype.listSignIns = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`SELECT signinSecret, signinWhen, signinIP, signinAgent FROM Signin WHERE signinPerson=${p.personId};`))
             .then(rows => ({
                signins: rows
            }));
    };
    exports.Connection.prototype.unSignIn = function(signinSecret, ip, agent, otherSecret) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`DELETE FROM Signin WHERE signinSecret=${mysql.escape(otherSecret)};`));
    };
    exports.Connection.prototype.saveModel = function(signinSecret, ip, agent, name, qmj, about, doi, whenPub, authors) {
        return this.auth({signinSecret: signinSecret, canSave: true}, ip, agent)
            .then(p => {
                let Nstate = qmj && qmj.states && qmj.states.length || 0;
                let Nclass = 0;
                let classes = [];
                qmj.states && qmj.states.forEach(s => {
                    classes[s.Class] += 1;
                });
                classes.forEach(c => { ++Nclass });
                let digest = [name, about, authors, p.personName, p.personAbout, ""+Nstate+"state", ""+Nclass+"class"].join("\n");
                return this.query("INSERT INTO Model SET ?", {
                    modelName: name,
                    modelQmj: JSON.stringify(qmj),
                    modelAbout: about,
                    modelDOI: doi,
                    modelPerson: p.personId,
                    modelWhenAdded: new Date(),
                    modelWhenPub: whenPub,
                    modelAuthors: authors,
                    modelStates: Nstate,
                    modelClasses: Nclass,
                    modelDigest: digest
                });
            })
            .then(result => ({
                modelId: result.insertId
            }));
    };
    exports.Connection.prototype.listModels = function(queryString) {
        let where = "";
        let kwInserts = [];
        if ( queryString ) {
            let kw = queryString.split(" ");
            let like = kw.reduce((prev, curr) => {
                prev.push("(modelDigest LIKE ?)");
                kwInserts.push(`%${curr}%`);
                return prev;
            }, []);
            where = "WHERE " + like.join(" AND ");
        }
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId "+where+" ORDER BY modelStates, modelClasses, modelName;", kwInserts)
            .then(rows => ({
                models: rows
            }));
    };
    exports.Connection.prototype.getModelQmj = function(modelId) {
        return this.query("SELECT modelQmj FROM Model WHERE modelId="+modelId+";")
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return JSON.parse(rows[0].modelQmj);
            });
    };
    exports.Connection.prototype.getModelInfo = function(modelId) {
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId WHERE modelId="+modelId+";")
            .then((rows) => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return rows[0];
            });
    };
    exports.Connection.prototype.requestInvite = function(ip, agent, name, about, email) {
        return this.query("SELECT personName, personActive FROM Person WHERE personEmail="+mysql.escape(email)+";")
            .then((rows) => {
                let existing = "";
                if ( rows.length ) {
                    existing = ''+rows[0].personName+' <'+email+'> has an existing account. ';
                    if ( ! rows[0].personActive ) {
                        existing += '(DISABLED) ';
                    }
                    existing += `To reissue an invitation for that account, visit ${this.config.adminLink}

`;
                }
                emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                    .send({from: this.config.emailSender,
                           to: this.config.requestEmail,
                           subject: "QuB Online: Invitation Request from "+name,
                           text: `${existing}${name} <${email}> (${about}) has requested an account on QuB Online.  To create the account, log in to the server, change to the QuB Online model_server directory, and enter one of the following commands:

[single-use invitation, can save models:] node invite.js -n "${name}" -e "${email}" -a "${about}" -s -u 1

[can self-invite to authorize new browser:] node invite.js -n "${name}" -e "${email}" -a "${about}" -sv -u 1

[can invite others:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svi -u 1

[can admin:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svim -u 1

Invitations generally expire after a week; for longer expiration add the -x command line arg with the number of hours into the future; e.g. one year:

node invite.js [...] -x 8766

IP address: ${ip}
User agent: ${agent}
`
                          },
                          (err, msg) => {
                              if ( err ) {
                                  console.log("Sending invitation request: ", err);
                              }
                              if ( msg ) {
                                  console.log("invited/msg: ", msg);
                              }
                          });
            })
            .catch(err => {
                console.log("Sending inv rq: ", err);
            });
    };
    exports.Connection.prototype.newPerson = function(ip, agent, name, about, email) {
        this.query("SELECT personName, personAbout FROM Person WHERE email=?", [email])
            .then(rows => {
                rows.forEach(r => {
                    console.log(`Email ${email} matches existing Person:`,r.personName,r.personAbout);
                });
            });
        return this.query("INSERT INTO Person SET ?", {
            personName: name,
            personAbout: about,
            personEmail: email,
            personActive: 1,
            personCanSave: 0,
            personCanSelfInvite: 1,
            personCanInvite: 0,
            personCanAdmin: 0
        }).then(result => {
            let personId = result.insertId;
            let siSecret = uuid.v4();
            return this.query("INSERT INTO Signin SET ?", {
                signinPerson: personId,
                signinSecret: siSecret,
                signinWhen: new Date(),
                signinIP: ip,
                signinAgent: agent
            }).then(result => ({
                signinSecret: siSecret
            }));
        });
    };
    exports.Connection.prototype.requestPermissions = function(signinSecret, ip, agent, what) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            let url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/users.html#u='+p.personId;
            emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                .send({from: this.config.emailSender,
                       to: this.config.requestEmail,
                       subject: "QuB Online: Permission Request from "+p.personName,
                       text: `${p.personName} <${p.personEmail}> (${p.personAbout$}) has requested more permissions (${what}) on QuB Online.  To edit their permissions, visit the user administration page: ${url}
`
                      },
                      (err, msg) => {
                          if ( err ) {
                              console.log("Sending permission request: ", err);
                          }
                      });
        });
    };
    exports.Connection.prototype.createLink = function(signinSecret, ip, agent, name, qmj, about) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            let linkKey = uuid.v4();
            return this.query("INSERT INTO LinkModel SET ?", {
                linkKey: linkKey,
                modelName: name,
                modelAbout: about,
                modelQmj: JSON.stringify(qmj),
                linkPerson: p.personId,
                linkWhenAdded: new Date()
            }).then(result => ({
                linkId: result.insertId,
                linkKey: linkKey
            }));
        });
    };
    exports.Connection.prototype.loadLink = function(linkId, linkKey) {
        return this.query(`SELECT modelName, modelQmj, modelAbout, linkWhenAdded FROM LinkModel WHERE (linkId=${linkId}) AND (linkKey=${mysql.escape(linkKey)});`)
            .then(rows => {
                if ( rows.length ) {
                    rows[0].modelQmj = JSON.parse(rows[0].modelQmj);
                    rows[0].modelPerson = rows[0].linkPerson;
                    rows[0].modelWhenAdded = rows[0].linkWhenAdded;
                    return rows[0];
                }
                throw new Error("No such link");
            });
    };
    exports.Connection.prototype.listLinks = function(signinSecret, ip, agent, queryString, allUsers) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            let where = "";
            let kwInserts = [];
            let like = [];
            if ( queryString ) {
                let kw = queryString.split(" ");
                like = kw.reduce((prev, curr) => {
                    prev.push("((modelName LIKE ?) OR (modelAbout LIKE ?))");
                    kwInserts.push(`%${curr}%`);
                    kwInserts.push(`%${curr}%`);
                    return prev;
                }, []);
            }
            if ( ! (p.personCanAdmin && allUsers) ) {
                like.push(`(linkPerson=${p.personId})`);
            }
            if ( like.length ) {
                where = "WHERE " + like.join(" AND ");
            }
            return this.query(`SELECT linkId, linkKey, modelName, modelAbout, linkWhenAdded, linkPerson, personName, personAbout FROM LinkModel INNER JOIN Person on personId=linkPerson ${where} ORDER BY linkWhenAdded DESC;`, kwInserts)
                .then(rows => {
                    rows.forEach(r => {
                        r.modelPerson = r.linkPerson;
                        r.modelWhenAdded = r.linkWhenAdded;
                    });
                    return rows;
                });
        });
    };
    exports.Connection.prototype.saveSim = function(signinSecret, ip, agent, name, about, config, modelId) {
        return this.auth({signinSecret: signinSecret, canSave: true}, ip, agent)
            .then(p => this.query("INSERT INTO Simulation SET ?", {
                simName: name,
                simAbout: about,
                simConfig: JSON.stringify(config),
                simModel: modelId
            }))
            .then(result => ({
                simId: result.insertId
            }));
    };
    exports.Connection.prototype.listSim = function(simId) {
        var where = (simId >= 0) ? ` WHERE simId=${mysql.escape(simId)}` : "";
        return this.query(`SELECT simId, simName, simAbout, simConfig, simModel FROM Simulation${where};`)
            .then(rows => ({
                sims: rows.map(row => ({
                    simId: row.simId,
                    name: row.simName,
                    about: row.simAbout,
                    config: JSON.parse(row.simConfig),
                    modelId: row.simModel
                }))
            }));
    };
    
    return exports;
});
