/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("modeldb", ["mysql", "node-uuid", "emailjs"], function(mysql, uuid, emailjs) {
    'use strict';
    
    const exports = {};

    let ConnectionConfig = {
        hostname: "localhost",
        user: "Models",
        password: "jihqprewgop",
        database: "Models",
        inviteHost: "www.qub.buffalo.edu",
        mainInstallPath: "/online",
        emailSender: "qub@www.qub.buffalo.edu",
        requestEmail: "qub@www.qub.buffalo.edu",
        adminLink: "https://www.qub.buffalo.edu/online/users.html"
    };

    exports.Connect = function Connect(config) {
        config = Object.assign({}, ConnectionConfig, config);
        return mysql.createPool({
            connectionLimit: 100,
            host: config.hostname,
            user: config.user,
            password: config.password,
            database: config.database,
            charset: "utf8mb4_unicode_ci"
            /* IMPORTANT: before creating tables, you want mysql to be using full multibyte UTF8.

This goes in /etc/mysql/mysql.conf.d/mysqld.cnf:
----
init_connect='SET collation_connection = utf8mb4_unicode_ci'
init_connect='SET NAMES utf8mb4'
character-set-server=utf8mb4
collation-server=utf8mb4_unicode_ci
skip-character-set-client-handshake
----
If you already make the tables, you have to fix them:
----
ALTER DATABASE Models CHARACTER SET = utf8mb4 COLLATE = utf8mb4_unicode_ci;
USE Models;
ALTER TABLE Person CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE Model CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE LinkModel CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE Simulation CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
-----
*/
        });
    };

    exports.Connection = function(config) {
        this.config = Object.assign({}, ConnectionConfig, config);
        this.pool = exports.Connect(this.config);
    };
    exports.Connection.prototype.query = function(q, inserts) {
        return new Promise((resolve, reject) => {
            this.pool.getConnection((err,connection) => {
                if (err) {
                    reject("No database connection: "+err);
                    return;
                }
                connection.query(q, inserts||[], (err, rows) => {
                    connection.release();
                    if ( err ) {
                        reject("Database error: "+err);
                    } else {
                        resolve(rows);
                    }
                }).on('error', err => {
                    reject("Database error: "+err);
                    return;
                });
            });
        });
    };
    
    exports.Connection.prototype.signIn = function(inviteSecret, ip, agent) {
        return this.query(`SELECT inviteId, invitePerson, inviteExpires, inviteUsesRemain FROM Invite WHERE inviteSecret=${mysql.escape(inviteSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Invitation expired or unknown");
                }
                const inv = rows[0];
                const delInvite = () => {
                    this.query(`DELETE FROM Invite WHERE inviteId=${inv.inviteId};`)
                        .catch(function(err) {
                            console.log("Deleting invite: ", err);
                        });
                };
                if ( ((new Date()) - inv.inviteExpires) > 0 ) {
                    delInvite();
                    throw new Error("Invitation expired");
                }
                if ( inv.inviteUsesRemain > 1 ) {
                    this.query(`UPDATE Invite SET inviteUsesRemain=${inv.inviteUsesRemain-1} WHERE inviteId=${inv.inviteId};`)
                        .catch(err => {
                            console.log("Decrementing inviteUsesRemain: ", err);
                        });
                } else if ( inv.inviteUsesRemain === 1 ) {
                    delInvite();
                }
                const siSecret = uuid.v4();
                return this.query(`INSERT INTO Signin (signinPerson, signinSecret, signinWhen, signinIP, signinAgent) VALUES (${inv.invitePerson},${mysql.escape(siSecret)},NOW(),${mysql.escape(""+ip)},${mysql.escape(agent)});`)
                    .then(result => siSecret);
            });
    };
    exports.Connection.prototype.signOut = function(signinSecret) {
        return this.query(`DELETE FROM Signin WHERE signinSecret=${mysql.escape(signinSecret)};`);
    };
    exports.Connection.prototype.auth = function(q, ip, agent) {
        return this.query(`SELECT signinId, signinPerson FROM Signin WHERE signinSecret=${mysql.escape(""+q.signinSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Signin is invalid or expired");
                }
                this.query(`UPDATE Signin SET signinWhen=NOW(), signinIP=${mysql.escape(""+ip)}, signinAgent=${mysql.escape(agent)} WHERE signinId=${rows[0].signinId};`)
                    .catch(err => {
                        console.log("Updating signin info: ", err);
                    });
                return this.query(`SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person WHERE personId=${rows[0].signinPerson};`);
            })
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Missing person");
                }
                const p = rows[0];
                if ( ! p.personActive ) {
                    throw new Error("Account deactivated");
                }
                if ( (q.canSave && ! p.personCanSave)
                     || (q.canSelfInvite && ! p.personCanSelfInvite)
                     || (q.canInvite && ! p.personCanInvite)
                     || (q.canAdmin && ! p.personCanAdmin) ) {
                    throw new Error("Insufficent permissions");
                }
                return p;
            });
    };
    exports.Connection.prototype.aboutMe = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => ({
            personName: p.personName,
            personAbout: p.personAbout,
            personEmail: p.personEmail,
            personCanSave: p.personCanSave,
            personCanSelfInvite: p.personCanSelfInvite,
            personCanInvite: p.personCanInvite,
            personCanAdmin: p.personCanAdmin
        }));
    };
    exports.Connection.prototype.setAbout = function(signinSecret, ip, agent, about) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personAbout=${mysql.escape(about)} WHERE personId=${p.personId};`));
    };
    exports.Connection.prototype.setEmail = function(signinSecret, ip, agent, email) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personEmail=${mysql.escape(email)} WHERE personId=${p.personId};`));
    };
    exports.Connection.prototype.selfInvite = function(signinSecret, ip, agent, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canSelfInvite: true}, ip, agent).then(p => {
            const inviteSecret = uuid.v4();
            const expires = new Date();
            expires.setHours(expires.getHours() + expiresHours);
            return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${p.personId},${mysql.escape(expires)},${usesRemain});`).then(result => {
                this.sendInvite(inviteSecret, p);
                return {inviteSecret: inviteSecret};
            });
        });
    };
    exports.Connection.prototype.inviteTrusted = function(invitee, expiresHours, usesRemain) {
        return this.query(`INSERT INTO Person (personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin) VALUES (${mysql.escape(""+invitee.personName)},${mysql.escape(""+invitee.personAbout)},${mysql.escape(""+invitee.personEmail)},1,${invitee.personCanSave?1:0},${invitee.personCanSelfInvite?1:0},${invitee.personCanInvite?1:0},${invitee.personCanAdmin?1:0});`)
            .then(result => {
                const personId = result.insertId;
                const inviteSecret = uuid.v4();
                const expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Connection.prototype.invite = function(signinSecret, ip, agent, invitee, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canInvite: true}, ip, agent).then(p => {
            ["personCanSave", "personCanSelfInvite", "personCanInvite", "personCanAdmin"].forEach(perm => {
                if ( ! p[perm] ) {
                    invitee[perm] = false;
                }
            });
            return this.inviteTrusted(invitee, expiresHours, usesRemain);
        });
    };
    exports.Connection.prototype.sendInvite = function(inviteSecret, p) {
        const url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/signin.html#'+inviteSecret;
        emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
            .send({from: this.config.emailSender,
                   to: p.personName + ' <' + p.personEmail + '>',
                   subject: "Sign in to QuB Online",
                   text: `Use this link to sign your browser into QuB Online as ${p.personName}:

${url}
`
                   },
                  (err, msg) => {
                     if ( err ) {
                         console.log(`Sending invitation to ${p.personName} <${p.personEmail}>: `, err);
                     }
                 });
    };
    exports.Connection.prototype.listPeople = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query("SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person;"))
            .then(rows => ({
                people: rows
            }));
    };
    exports.Connection.prototype.reinvite = function(signinSecret, ip, agent, personId, expiresHours, usesRemain) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query(`SELECT personName, personEmail FROM Person WHERE personId=${personId};`))
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such person");
                }
                const invitee = rows[0];
                const inviteSecret = uuid.v4();
                const expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Connection.prototype.updatePerson = function(signinSecret, ip, agent, id, name, about, email, active, canSave, canSelfInvite, canInvite, canAdmin) {
        return this.auth({signinSecret: signinSecret, canAdmin: true}, ip, agent)
            .then(p => this.query(`UPDATE Person SET personName=${mysql.escape(name)}, personAbout=${mysql.escape(about)}, personEmail=${mysql.escape(email)}, personActive=${active?1:0}, personCanSave=${canSave?1:0}, personCanSelfInvite=${canSelfInvite?1:0}, personCanInvite=${canInvite?1:0}, personCanAdmin=${canAdmin?1:0} WHERE personId=${id};`));
    };
    exports.Connection.prototype.listSignIns = function(signinSecret, ip, agent) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`SELECT signinSecret, signinWhen, signinIP, signinAgent FROM Signin WHERE signinPerson=${p.personId};`))
             .then(rows => ({
                signins: rows
            }));
    };
    exports.Connection.prototype.unSignIn = function(signinSecret, ip, agent, otherSecret) {
        return this.auth({signinSecret: signinSecret}, ip, agent)
            .then(p => this.query(`DELETE FROM Signin WHERE signinSecret=${mysql.escape(otherSecret)};`));
    };
    exports.Connection.prototype.saveModel = function(signinSecret, ip, agent, name, qmj, about, doi, whenPub, authors) {
        return this.auth({signinSecret: signinSecret, canSave: true}, ip, agent)
            .then(p => {
                const Nstate = qmj && qmj.states && qmj.states.length || 0;
                const classes = [];
                let Nclass = 0;
                qmj.states && qmj.states.forEach(s => {
                    classes[s.Class] += 1;
                });
                classes.forEach(c => { ++Nclass });
                const digest = [name, about, authors, p.personName, p.personAbout, ""+Nstate+"state", ""+Nclass+"class"].join("\n");
                const record = {
                    modelName: name,
                    modelQmj: JSON.stringify(qmj),
                    modelAbout: about,
                    modelDOI: doi,
                    modelPerson: p.personId,
                    modelWhenAdded: new Date(),
                    modelAuthors: authors,
                    modelStates: Nstate,
                    modelClasses: Nclass,
                    modelDigest: digest
                };
                if ( whenPub ) {
                    record.modelWhenPub = whenPub;
                }
                return this.query("INSERT INTO Model SET ?", record);
            })
            .then(result => ({
                modelId: result.insertId
            }));
    };
    exports.Connection.prototype.listModels = function(queryString) {
        let where = "";
        const kwInserts = [];
        if ( queryString ) {
            const kw = queryString.split(" ");
            const like = kw.reduce((prev, curr) => {
                prev.push("(modelDigest LIKE ?)");
                kwInserts.push(`%${curr}%`);
                return prev;
            }, []);
            where = "WHERE " + like.join(" AND ");
        }
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId "+where+" ORDER BY modelStates, modelClasses, modelName;", kwInserts)
            .then(rows => ({
                models: rows
            }));
    };
    exports.Connection.prototype.getModelQmj = function(modelId) {
        return this.query("SELECT modelQmj FROM Model WHERE modelId="+modelId+";")
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return JSON.parse(rows[0].modelQmj);
            });
    };
    exports.Connection.prototype.getModelInfo = function(modelId) {
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId WHERE modelId="+modelId+";")
            .then((rows) => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return rows[0];
            });
    };
    exports.Connection.prototype.requestInvite = function(ip, agent, name, about, email) {
        return this.query("SELECT personName, personActive FROM Person WHERE personEmail="+mysql.escape(email)+";")
            .then((rows) => {
                let existing = "";
                if ( rows.length ) {
                    existing = ''+rows[0].personName+' <'+email+'> has an existing account. ';
                    if ( ! rows[0].personActive ) {
                        existing += '(DISABLED) ';
                    }
                    existing += `To reissue an invitation for that account, visit ${this.config.adminLink}

`;
                }
                emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                    .send({from: this.config.emailSender,
                           to: this.config.requestEmail,
                           subject: "QuB Online: Invitation Request from "+name,
                           text: `${existing}${name} <${email}> (${about}) has requested an account on QuB Online.  To create the account, log in to the server, change to the QuB Online model_server directory, and enter one of the following commands:

[single-use invitation, can save models:] node invite.js -n "${name}" -e "${email}" -a "${about}" -s -u 1

[can self-invite to authorize new browser:] node invite.js -n "${name}" -e "${email}" -a "${about}" -sv -u 1

[can invite others:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svi -u 1

[can admin:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svim -u 1

Invitations generally expire after a week; for longer expiration add the -x command line arg with the number of hours into the future; e.g. one year:

node invite.js [...] -x 8766

IP address: ${ip}
User agent: ${agent}
`
                          },
                          (err, msg) => {
                              if ( err ) {
                                  console.log("Sending invitation request: ", err);
                              }
                              if ( msg ) {
                                  console.log("invited/msg: ", msg);
                              }
                          });
            })
            .catch(err => {
                console.log("Sending inv rq: ", err);
            });
    };
    exports.Connection.prototype.newPerson = function(ip, agent, name, about, email) {
        this.query("SELECT personName, personAbout FROM Person WHERE email=?", [email])
            .then(rows => {
                rows.forEach(r => {
                    console.log(`Email ${email} matches existing Person:`,r.personName,r.personAbout);
                });
            });
        return this.query("INSERT INTO Person SET ?", {
            personName: name,
            personAbout: about,
            personEmail: email,
            personActive: 1,
            personCanSave: 0,
            personCanSelfInvite: 1,
            personCanInvite: 0,
            personCanAdmin: 0
        }).then(result => {
            const personId = result.insertId;
            const siSecret = uuid.v4();
            return this.query("INSERT INTO Signin SET ?", {
                signinPerson: personId,
                signinSecret: siSecret,
                signinWhen: new Date(),
                signinIP: ip,
                signinAgent: agent
            }).then(result => ({
                signinSecret: siSecret
            }));
        });
    };
    exports.Connection.prototype.requestPermissions = function(signinSecret, ip, agent, what) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            const url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/users.html#u='+p.personId;
            emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                .send({from: this.config.emailSender,
                       to: this.config.requestEmail,
                       subject: "QuB Online: Permission Request from "+p.personName,
                       text: `${p.personName} <${p.personEmail}> (${p.personAbout$}) has requested more permissions (${what}) on QuB Online.  To edit their permissions, visit the user administration page: ${url}
`
                      },
                      (err, msg) => {
                          if ( err ) {
                              console.log("Sending permission request: ", err);
                          }
                      });
        });
    };
    exports.Connection.prototype.createLink = function(signinSecret, ip, agent, name, qmj, about) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            const linkKey = uuid.v4().slice(0,4);
            return this.query("INSERT INTO LinkModel SET ?", {
                linkKey: linkKey,
                modelName: name,
                modelAbout: about,
                modelQmj: JSON.stringify(qmj),
                linkPerson: p.personId,
                linkWhenAdded: new Date()
            }).then(result => ({
                linkId: result.insertId,
                linkKey: linkKey
            }));
        });
    };
    exports.Connection.prototype.loadLink = function(linkId, linkKey) {
        return this.query(`SELECT modelName, modelQmj, modelAbout, linkWhenAdded FROM LinkModel WHERE (linkId=${linkId}) AND (linkKey=${mysql.escape(linkKey)});`)
            .then(rows => {
                if ( rows.length ) {
                    rows[0].modelQmj = JSON.parse(rows[0].modelQmj);
                    rows[0].modelPerson = rows[0].linkPerson;
                    rows[0].modelWhenAdded = rows[0].linkWhenAdded;
                    return rows[0];
                }
                throw new Error("No such link");
            });
    };
    exports.Connection.prototype.listLinks = function(signinSecret, ip, agent, queryString, allUsers) {
        return this.auth({signinSecret: signinSecret}, ip, agent).then(p => {
            let where = "";
            let like = [];
            const kwInserts = [];
            if ( queryString ) {
                let kw = queryString.split(" ");
                like = kw.reduce((prev, curr) => {
                    prev.push("((modelName LIKE ?) OR (modelAbout LIKE ?))");
                    kwInserts.push(`%${curr}%`);
                    kwInserts.push(`%${curr}%`);
                    return prev;
                }, []);
            }
            if ( ! (p.personCanAdmin && allUsers) ) {
                like.push(`(linkPerson=${p.personId})`);
            }
            if ( like.length ) {
                where = "WHERE " + like.join(" AND ");
            }
            return this.query(`SELECT linkId, linkKey, modelName, modelAbout, linkWhenAdded, linkPerson, personName, personAbout FROM LinkModel INNER JOIN Person on personId=linkPerson ${where} ORDER BY linkWhenAdded DESC;`, kwInserts)
                .then(rows => {
                    rows.forEach(r => {
                        r.modelPerson = r.linkPerson;
                        r.modelWhenAdded = r.linkWhenAdded;
                    });
                    return rows;
                });
        });
    };
    exports.Connection.prototype.saveSim = function(signinSecret, ip, agent, name, about, config, modelId) {
        return this.auth({signinSecret: signinSecret, canSave: true}, ip, agent)
            .then(p => this.query("INSERT INTO Simulation SET ?", {
                simName: name,
                simAbout: about,
                simConfig: JSON.stringify(config),
                simModel: modelId
            }))
            .then(result => ({
                simId: result.insertId
            }));
    };
    exports.Connection.prototype.listSim = function(simId) {
        const where = (simId >= 0) ? ` WHERE simId=${mysql.escape(simId)}` : "";
        return this.query(`SELECT simId, simName, simAbout, simConfig, simModel FROM Simulation${where};`)
            .then(rows => ({
                sims: rows.map(row => ({
                    simId: row.simId,
                    name: row.simName,
                    about: row.simAbout,
                    config: JSON.parse(row.simConfig),
                    modelId: row.simModel
                }))
            }));
    };
    


    // should really configure with shared pool=modeldb.Connect()
    exports.Session = function(config) {
        this.config = Object.assign({}, ConnectionConfig, config);
        this.pool = this.config.pool || exports.Connect(this.config)
        this.signin = '';
        this.person = {};
    };
    exports.Session.prototype.query = function(q, inserts) {
        return new Promise((resolve, reject) => {
            this.pool.getConnection((err,connection) => {
                if (err) {
                    reject("No database connection: "+err);
                    return;
                }
                connection.query(q, inserts||[], (err, rows) => {
                    connection.release();
                    if ( err ) {
                        reject("Database error: "+err);
                    } else {
                        resolve(rows);
                    }
                }).on('error', err => {
                    reject("Database error: "+err);
                    return;
                });
            });
        });
    };
    
    exports.Session.prototype.signIn = function(inviteSecret, ip, agent) {
        return this.query(`SELECT inviteId, invitePerson, inviteExpires, inviteUsesRemain FROM Invite WHERE inviteSecret=${mysql.escape(inviteSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Invitation expired or unknown");
                }
                const inv = rows[0];
                const delInvite = () => {
                    this.query(`DELETE FROM Invite WHERE inviteId=${inv.inviteId};`)
                        .catch(err => {
                            console.log("Deleting invite: ", err);
                        });
                };
                if ( ((new Date()) - inv.inviteExpires) > 0 ) {
                    delInvite();
                    throw new Error("Invitation expired");
                }
                if ( inv.inviteUsesRemain > 1 ) {
                    this.query(`UPDATE Invite SET inviteUsesRemain=${inv.inviteUsesRemain-1} WHERE inviteId=${inv.inviteId};`)
                        .catch(err => {
                            console.log("Decrementing inviteUsesRemain: ", err);
                        });
                } else if ( inv.inviteUsesRemain === 1 ) {
                    delInvite();
                }
                const siSecret = uuid.v4();
                this.signin = siSecret;
                return this.query(`INSERT INTO Signin (signinPerson, signinSecret, signinWhen, signinIP, signinAgent) VALUES (${inv.invitePerson},${mysql.escape(siSecret)},NOW(),${mysql.escape(""+ip)},${mysql.escape(agent)});`)
                    .then(result => siSecret);
            }).then(siSecret => this.auth({signinSecret: siSecret}, ip, agent)
                    .then(p => siSecret)
                   );
    };
    exports.Session.prototype.signOut = function() {
        if ( ! this.person.personId ) {
            return Promise.reject('Not signed in');
        }
        const signinSecret = this.signin;
        this.person = {};
        this.signin = '';
        return this.query(`DELETE FROM Signin WHERE signinSecret=${mysql.escape(signinSecret)};`);
    };
    exports.Session.prototype.auth = function(q, ip, agent) {
        if ( agent && (agent.length > 255) ) {
            agent = agent.slice(0, 255);
        }
        return this.query(`SELECT signinId, signinPerson FROM Signin WHERE signinSecret=${mysql.escape(""+q.signinSecret)};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Signin is invalid or expired");
                }
                this.query(`UPDATE Signin SET signinWhen=NOW(), signinIP=${mysql.escape(""+ip)}, signinAgent=${mysql.escape(agent)} WHERE signinId=${rows[0].signinId};`)
                    .catch(err => {
                        console.log("Updating signin info: ", err);
                    });
                return this.query(`SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person WHERE personId=${rows[0].signinPerson};`);
            })
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("Missing person");
                }
                const p = rows[0];
                if ( ! p.personActive ) {
                    throw new Error("Account deactivated");
                }
                if ( (q.canSave && ! p.personCanSave)
                     || (q.canSelfInvite && ! p.personCanSelfInvite)
                     || (q.canInvite && ! p.personCanInvite)
                     || (q.canAdmin && ! p.personCanAdmin) ) {
                    throw new Error("Insufficent permissions");
                }
                this.signin = q.signinSecret;
                this.person = p;
                return p;
            });
    };
    exports.Session.prototype.setAbout = function(about) {
        if ( ! this.person.personId ) {
            return Promise.reject('Insufficent permissions');
        }
        return this.query(`UPDATE Person SET personAbout=${mysql.escape(about)} WHERE personId=${this.person.personId};`);
    };
    exports.Session.prototype.setEmail = function(email) {
        if ( ! this.person.personId ) {
            return Promise.reject('Insufficent permissions');
        }
        return this.query(`UPDATE Person SET personEmail=${mysql.escape(email)} WHERE personId=${this.person.personId};`);
    };
    exports.Session.prototype.selfInvite = function(expiresHours, usesRemain) {
        if ( ! this.person.personCanSelfInvite ) {
            return Promise.reject('Insufficent permissions');
        }
        const inviteSecret = uuid.v4();
        const expires = new Date();
        expires.setHours(expires.getHours() + expiresHours);
        return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${this.person.personId},${mysql.escape(expires)},${usesRemain});`).then(result => {
            this.sendInvite(inviteSecret, this.person);
            return {inviteSecret: inviteSecret};
        });
    };
    exports.Session.prototype.inviteTrusted = function(invitee, expiresHours, usesRemain) {
        return this.query(`INSERT INTO Person (personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin) VALUES (${mysql.escape(""+invitee.personName)},${mysql.escape(""+invitee.personAbout)},${mysql.escape(""+invitee.personEmail)},1,${invitee.personCanSave?1:0},${invitee.personCanSelfInvite?1:0},${invitee.personCanInvite?1:0},${invitee.personCanAdmin?1:0});`)
            .then(result => {
                const personId = result.insertId;
                const inviteSecret = uuid.v4();
                const expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Session.prototype.invite = function(invitee, expiresHours, usesRemain) {
        if ( ! this.person.personCanInvite ) {
            return Promise.reject('Insufficent permissions');
        }
        const p = this.person;
        ["personCanSave", "personCanSelfInvite", "personCanInvite", "personCanAdmin"].forEach(perm => {
            if ( ! p[perm] ) {
                invitee[perm] = false;
            }
        });
        return this.inviteTrusted(invitee, expiresHours, usesRemain).then(inviteSecret => ({
            inviteSecret: inviteSecret
        }));
    };
    exports.Session.prototype.sendInvite = function(inviteSecret, p) {
        const url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/signin.html#'+inviteSecret;
        emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
            .send({from: this.config.emailSender,
                   to: p.personName + ' <' + p.personEmail + '>',
                   subject: "Sign in to QuB Online",
                   text: `Use this link to sign your browser into QuB Online as ${p.personName}:

${url}
`
                   },
                  (err, msg) => {
                     if ( err ) {
                         console.log(`Sending invitation to ${p.personName} <${p.personEmail}>: `, err);
                     }
                 });
    };
    exports.Session.prototype.listPeople = function() {
        if ( ! this.person.personCanAdmin ) {
            return Promise.reject('Insufficient permissions');
        }
        return this.query("SELECT personId, personName, personAbout, personEmail, personActive, personCanSave, personCanSelfInvite, personCanInvite, personCanAdmin FROM Person;")
            .then(rows => ({
                people: rows
            }));
    };
    exports.Session.prototype.reinvite = function(personId, expiresHours, usesRemain) {
        if ( ! this.person.personCanAdmin ) {
            return Promise.reject('Insufficient permissions');
        }
        return this.query(`SELECT personName, personEmail FROM Person WHERE personId=${personId};`)
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such person");
                }
                const invitee = rows[0];
                const inviteSecret = uuid.v4();
                const expires = new Date();
                expires.setHours(expires.getHours() + expiresHours);
                return this.query(`INSERT INTO Invite (inviteSecret, invitePerson, inviteExpires, inviteUsesRemain) VALUES (${mysql.escape(inviteSecret)},${personId},${mysql.escape(expires)},${usesRemain});`)
                    .then(iResult => {
                        this.sendInvite(inviteSecret, invitee);
                        return inviteSecret;
                    });
            });
    };
    exports.Session.prototype.updatePerson = function(id, name, about, email, active, canSave, canSelfInvite, canInvite, canAdmin) {
        if ( ! this.person.personCanAdmin ) {
            return Promise.reject('Insufficient permissions');
        }
        return this.query(`UPDATE Person SET personName=${mysql.escape(name)}, personAbout=${mysql.escape(about)}, personEmail=${mysql.escape(email)}, personActive=${active?1:0}, personCanSave=${canSave?1:0}, personCanSelfInvite=${canSelfInvite?1:0}, personCanInvite=${canInvite?1:0}, personCanAdmin=${canAdmin?1:0} WHERE personId=${id};`);
    };
    exports.Session.prototype.listSignIns = function() {
        if ( ! this.person.personId ) {
            return Promise.resolve({signins: []});
        }
        return this.query(`SELECT signinSecret, signinWhen, signinIP, signinAgent FROM Signin WHERE signinPerson=${this.person.personId};`)
            .then(rows => ({
                signins: rows
            }));
    };
    exports.Session.prototype.unSignIn = function(otherSecret) {
        if ( ! this.person.personId ) {
            return Promise.reject('Insufficient permissions');
        }
        return this.query(`DELETE FROM Signin WHERE signinPerson=${this.person.personId} AND signinSecret=${mysql.escape(otherSecret)};`);
    };
    exports.Session.prototype.saveModel = function(name, qmj, about, doi, whenPub, authors) {
        if ( ! this.person.personCanSave ) {
            return Promise.reject('Insufficient permissions');
        }
        const p = this.person;
        const Nstate = qmj && qmj.states && qmj.states.length || 0;
        const classes = [];
        let Nclass = 0;
        qmj.states && qmj.states.forEach(s => {
            classes[s.Class] += 1;
        });
        classes.forEach(c => { ++Nclass });
        const digest = [name, about, authors, p.personName, p.personAbout, ""+Nstate+"state", ""+Nclass+"class"].join("\n");
        return this.query("INSERT INTO Model SET ?", {
            modelName: name,
            modelQmj: JSON.stringify(qmj),
            modelAbout: about,
            modelDOI: doi,
            modelPerson: p.personId,
            modelWhenAdded: new Date(),
            modelWhenPub: whenPub,
            modelAuthors: authors,
            modelStates: Nstate,
            modelClasses: Nclass,
            modelDigest: digest
        }).then(result => ({
            modelId: result.insertId
        }));
    };
    exports.Session.prototype.listModels = function(queryString) {
        let where = "";
        const kwInserts = [];
        if ( queryString ) {
            const kw = queryString.split(" ");
            const like = kw.reduce((prev, curr) => {
                prev.push("(modelDigest LIKE ?)");
                kwInserts.push(`%${curr}%`);
                return prev;
            }, []);
            where = "WHERE " + like.join(" AND ");
        }
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId "+where+" ORDER BY modelStates, modelClasses, modelName;", kwInserts)
            .then(rows => ({
                models: rows
            }));
    };
    exports.Session.prototype.getModelQmj = function(modelId) {
        return this.query("SELECT modelQmj FROM Model WHERE modelId="+modelId+";")
            .then(rows => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return JSON.parse(rows[0].modelQmj);
            });
    };
    exports.Session.prototype.getModelInfo = function(modelId) {
        return this.query("SELECT modelId, modelName, modelAbout, modelDOI, modelWhenAdded, modelWhenPub, modelAuthors, modelStates, modelClasses, personName, personAbout FROM Model INNER JOIN Person ON modelPerson=personId WHERE modelId="+modelId+";")
            .then((rows) => {
                if ( rows.length === 0 ) {
                    throw new Error("No such model");
                }
                return rows[0];
            });
    };
    exports.Session.prototype.requestInvite = function(ip, agent, name, about, email) {
        return this.query("SELECT personName, personActive FROM Person WHERE personEmail="+mysql.escape(email)+";")
            .then((rows) => {
                let existing = "";
                if ( rows.length ) {
                    existing = ''+rows[0].personName+' <'+email+'> has an existing account. ';
                    if ( ! rows[0].personActive ) {
                        existing += '(DISABLED) ';
                    }
                    existing += `To reissue an invitation for that account, visit ${this.config.adminLink}

`;
                }
                emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                    .send({from: this.config.emailSender,
                           to: this.config.requestEmail,
                           subject: "QuB Online: Invitation Request from "+name,
                           text: `${existing}${name} <${email}> (${about}) has requested an account on QuB Online.  To create the account, log in to the server, change to the QuB Online model_server directory, and enter one of the following commands:

[single-use invitation, can save models:] node invite.js -n "${name}" -e "${email}" -a "${about}" -s -u 1

[can self-invite to authorize new browser:] node invite.js -n "${name}" -e "${email}" -a "${about}" -sv -u 1

[can invite others:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svi -u 1

[can admin:] node invite.js -n "${name}" -e "${email}" -a "${about}" -svim -u 1

Invitations generally expire after a week; for longer expiration add the -x command line arg with the number of hours into the future; e.g. one year:

node invite.js [...] -x 8766

IP address: ${ip}
User agent: ${agent}
`
                          },
                          (err, msg) => {
                              if ( err ) {
                                  console.log("Sending invitation request: ", err);
                              }
                              if ( msg ) {
                                  console.log("invited/msg: ", msg);
                              }
                          });
            })
            .catch(err => {
                console.log("Sending inv rq: ", err);
            });
    };
    exports.Session.prototype.newPerson = function(ip, agent, name, about, email) {
        this.query("SELECT personName, personAbout FROM Person WHERE email=?", [email])
            .then(rows => {
                rows.forEach(r => {
                    console.log(`Email ${email} matches existing Person:`,r.personName,r.personAbout);
                });
            });
        return this.query("INSERT INTO Person SET ?", {
            personName: name,
            personAbout: about,
            personEmail: email,
            personActive: 1,
            personCanSave: 0,
            personCanSelfInvite: 1,
            personCanInvite: 0,
            personCanAdmin: 0
        }).then(result => {
            const personId = result.insertId;
            const siSecret = uuid.v4();
            return this.query("INSERT INTO Signin SET ?", {
                signinPerson: personId,
                signinSecret: siSecret,
                signinWhen: new Date(),
                signinIP: ip,
                signinAgent: agent
            }).then(result => ({
                signinSecret: siSecret
            }));
        });
    };
    exports.Session.prototype.requestPermissions = function(what) {
        if ( ! this.person.personId ) {
            return Promise.reject('Not signed in');
        }
        const p = this.person;
        const url = 'http://'+this.config.inviteHost+this.config.mainInstallPath+'/users.html#u='+p.personId;
        emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
            .send({from: this.config.emailSender,
                   to: this.config.requestEmail,
                   subject: "QuB Online: Permission Request from "+p.personName,
                   text: `${p.personName} <${p.personEmail}> (${p.personAbout$}) has requested more permissions (${what}) on QuB Online.  To edit their permissions, visit the user administration page: ${url}
                   `
                  },
                  (err, msg) => {
                      if ( err ) {
                          console.log("Sending permission request: ", err);
                      }
                  });
        return Promise.resolve(true);
    };
    exports.Session.prototype.createLink = function(name, qmj, about) {
        if ( ! this.person.personId ) {
            return Promise.reject('Not signed in');
        }
        const linkKey = uuid.v4().slice(0,4);
        return this.query("INSERT INTO LinkModel SET ?", {
            linkKey: linkKey,
            modelName: name,
            modelAbout: about,
            modelQmj: JSON.stringify(qmj),
            linkPerson: this.person.personId,
            linkWhenAdded: new Date()
        }).then(result => ({
            linkId: result.insertId,
            linkKey: linkKey
        }));
    };
    exports.Session.prototype.loadLink = function(linkId, linkKey) {
        return this.query(`SELECT modelName, modelQmj, modelAbout, linkWhenAdded FROM LinkModel WHERE (linkId=${linkId}) AND (linkKey=${mysql.escape(linkKey)});`)
            .then(rows => {
                if ( rows.length ) {
                    rows[0].modelQmj = JSON.parse(rows[0].modelQmj);
                    rows[0].modelPerson = rows[0].linkPerson;
                    rows[0].modelWhenAdded = rows[0].linkWhenAdded;
                    return rows[0];
                }
                throw new Error("No such link");
            });
    };
    exports.Session.prototype.listLinks = function(queryString, allUsers) {
        if ( ! this.person.personId ) {
            return Promise.reject('Insufficient permissions');
        }
        let where = "";
        let like = [];
        const kwInserts = [];
        if ( queryString ) {
            const kw = queryString.split(" ");
            like = kw.reduce((prev, curr) => {
                prev.push("((modelName LIKE ?) OR (modelAbout LIKE ?))");
                kwInserts.push(`%${curr}%`);
                kwInserts.push(`%${curr}%`);
                return prev;
            }, []);
        }
        if ( ! (this.person.personCanAdmin && allUsers) ) {
            like.push(`(linkPerson=${this.person.personId})`);
        }
        if ( like.length ) {
            where = "WHERE " + like.join(" AND ");
        }
        return this.query(`SELECT linkId, linkKey, modelName, modelAbout, linkWhenAdded, linkPerson, personName, personAbout FROM LinkModel INNER JOIN Person on personId=linkPerson ${where} ORDER BY linkWhenAdded DESC;`, kwInserts)
            .then(rows => {
                rows.forEach(r => {
                    r.modelPerson = r.linkPerson;
                    r.modelWhenAdded = r.linkWhenAdded;
                });
                return rows;
            });
    };
    exports.Session.prototype.saveSim = function(name, about, config, modelId) {
        if ( ! this.person.personCanSave ) {
            return Promise.reject('Insufficient permissions');
        }
        return this.query("INSERT INTO Simulation SET ?", {
            simName: name,
            simAbout: about,
            simConfig: JSON.stringify(config),
            simModel: modelId
        }).then(result => ({
            simId: result.insertId
        }));
    };
    exports.Session.prototype.listSim = function(simId) {
        const where = (simId >= 0) ? ` WHERE simId=${mysql.escape(simId)}` : "";
        return this.query(`SELECT simId, simName, simAbout, simConfig, simModel FROM Simulation${where};`)
            .then(rows => ({
                sims: rows.map(row => ({
                    simId: row.simId,
                    name: row.simName,
                    about: row.simAbout,
                    config: JSON.parse(row.simConfig),
                    modelId: row.simModel
                }))
            }));
    };
    
    return exports;
});
