"use strict";

define("optdb", ["mysql"], function(mysql) {
    let exports = {};

    function extend(target) {
        var sources = [].slice.call(arguments, 1);
        sources.forEach(function (source) {
            Object.getOwnPropertyNames(source).forEach(function(propName) {
                Object.defineProperty(target, propName,
                    Object.getOwnPropertyDescriptor(source, propName));
            });
        });
        return target;
    };
    
    let ConnectionConfig = {
        hostname: "localhost",
        user: "quboptstat",
        password: "jihqprewgop",
        database: "quboptstat"
    };
    exports.Connection = function(config) {
        this.config = extend({}, ConnectionConfig, config);
        this.pool = mysql.createPool({
            connectionLimit: 100,
            host: this.config.hostname,
            user: this.config.user,
            password: this.config.password,
            database: this.config.database
        });
        this.ensureTables()
            // .then(function(x) {console.log("Tables exist");})
            .catch(function(err) {console.log("ensureTables: "+err);});
    };
    exports.Connection.prototype.query = function(q) {
        return new Promise(function(resolve, reject) {
            this.pool.getConnection(function(err,connection) {
                if (err) {
                    reject("No database connection: "+err);
                    return;
                }
                connection.query(q, function(err, rows) {
                    connection.release();
                    if ( err ) {
                        reject("Database error: "+err);
                    } else {
                        resolve(rows);
                    }
                }).on('error', function(err) {
                    reject("Database error: "+err);
                    return;
                });
            });
        }.bind(this));
    };
    exports.Connection.prototype.ensureTable = function(name, sqlFields) {
        return this.query("SHOW TABLES LIKE "+mysql.escape(name)+";").then(function(rows) {
            if ( rows.length === 0 ) {
                return this.query("CREATE TABLE "+name+sqlFields+";");
            }
            return true;
        }.bind(this));
    };
    exports.Connection.prototype.ensureTables = function() {
        return Promise.resolve(true)
            .then(function(x) { return this.ensureTable("Algorithm", "(algorithmId INT PRIMARY KEY AUTO_INCREMENT, algorithmName VARCHAR(32))"); }.bind(this))
            .then(function(x) { return this.ensureTable("Nauty", "(nautyId INT PRIMARY KEY AUTO_INCREMENT, nautyNumber INT, nautyQmj TEXT, nautyStateCount INT)"); }.bind(this))
            .then(function(x) { return this.ensureTable("Stimulus", "(stimId INT PRIMARY KEY AUTO_INCREMENT, stimLbl VARCHAR(64))"); }.bind(this))
            .then(function(x) { return this.ensureTable("Trial", "(trialId INT PRIMARY KEY AUTO_INCREMENT, algorithmId INT, nautyId INT, stimId INT, trialQmj TEXT, channelCount INT, dataCount INT, tDead DOUBLE, sampling DOUBLE, app VARCHAR(32), os VARCHAR(64), ext1 VARCHAR(16), ext2 VARCHAR(16), ext3 VARCHAR(16))"); }.bind(this))
            .then(function(x) { return this.ensureTable("Rate", "(rateId INT PRIMARY KEY AUTO_INCREMENT, trialId INT, rateIndex INT, rateInit DOUBLE, rateVal DOUBLE, rateFactor DOUBLE, rateType INT)"); }.bind(this));
    };
    exports.Connection.prototype.listAlgo = function() {
        return this.query("SELECT algorithmId, algorithmName FROM Algorithm;")
            .then(function(rows) {
                return {algorithms: rows};
            });
    };
    exports.Connection.prototype.listNauty = function() {
        return this.query("SELECT nautyId, nautyNumber, nautyQmj, nautyStateCount FROM Nauty;")
            .then(function(rows) {
                for (var i=0, n=rows.length; i<n; ++i) {
                    rows[i].nautyQmj = JSON.parse(rows[i].nautyQmj);
                }
                return {nauty: rows};
            });
    };
    exports.Connection.prototype.listStim = function() {
        return this.query("SELECT stimId, stimLbl FROM Stimulus;")
            .then(function(rows) {
                return {stimulus: rows};
            });
    };
    exports.Connection.prototype.getNauty = function(nautyId) {
        return this.query("SELECT nautyId, nautyNumber, nautyQmj, nautyStateCount FROM Nauty WHERE nautyId="+nautyId+";")
            .then(function(rows) {
                if ( rows.length === 0 ) {
                    throw new Error("Unknown model id");
                }
                return rows[0];
            });
    };
    exports.Connection.prototype.getTrial = function(trialId) {
        return this.query("SELECT trialId, algorithmId, nautyId, stimId, trialQmj, channelCount, dataCount, tDead, sampling, app, os FROM Trial WHERE trialId="+trialId+";")
            .then(function(rows) {
                if ( rows.length === 0 ) {
                    throw new Error("Unknown trial id");
                }
                rows[0].trialQmj = JSON.parse(rows[0].trialQmj);
                return rows[0];
            });
    };
    exports.Connection.prototype.getRates = function(q) {
        var qs = "SELECT rateId, Rate.trialId, rateIndex, rateInit, rateVal, rateFactor, rateType, nautyId, stimId, algorithmId, channelCount, dataCount, tDead, sampling FROM Rate INNER JOIN Trial ON Rate.trialId = Trial.trialId";
        var active = [];
        ["nautyId", "algorithmId", "stimId", "channelCount", "dataCount", "sampling"].forEach(function(name) {
            if ( typeof(q[name]) !== 'undefined' ) {
                active.push(name);
            }
        });
        var where = active.map(function(name) {
            if ( name === "sampling" ) { // float
                return name+"="+(+q[name]);
            } else { // int
                return name+"="+(q[name]|0);
            }
        }).join(" AND ");
        if ( where ) {
            qs += " WHERE " + where;
        }
        qs += " ORDER BY rateInit;";
        return this.query(qs).then(function(rows) {
            return {rates: rows};
        });
    };
    exports.Connection.prototype.trial = function(t) {
        return this.withTrialId(t).then(function(tResult) {
            let trialId = tResult.trialId;
            let rrPre = t.modelPre.rates;
            let rrPost = t.modelPost.rates;
            let queries = [];
            for (let i=0, n=rrPre.length; i<n; ++i) {
                let rPre = rrPre[i];
                let rPost = rrPost[i];
                queries.push(this.query("INSERT INTO Rate (trialId, rateIndex, rateInit, rateVal, rateFactor, rateType) VALUES ("+trialId+","+i+","+rPre.k0+","+rPost.k0+","+(rPost.k0/rPre.k0)+",0);"));
                if ( rPre.k1 ) {
                    queries.push(this.query("INSERT INTO Rate (trialId, rateIndex, rateInit, rateVal, rateFactor, rateType) VALUES ("+trialId+","+i+","+rPre.k1+","+rPost.k1+","+(rPost.k1/rPre.k1)+",1);"));
                }
                if ( rPre.k2 ) {
                    queries.push(this.query("INSERT INTO Rate (trialId, rateIndex, rateInit, rateVal, rateFactor, rateType) VALUES ("+trialId+","+i+","+rPre.k2+","+rPost.k2+","+(rPost.k2/rPre.k2)+",2);"));
                }
            }
            return Promise.all(queries);
        }.bind(this));
    };
    exports.Connection.prototype.withTrialId = function(t) {
        return Promise.all([this.withNautyId(t.nautyNumber|0, t.modelPre),
                            this.withStimId(t.stimLbl),
                            this.withAlgorithmId(t.algorithmName)])
            .then(function(values) {
                let nautyId = values[0].nautyId;
                let stimId = values[1].stimId;
                let algorithmId = values[2].algorithmId;
                return this.query("INSERT INTO Trial (algorithmId, nautyId, stimId, trialQmj, channelCount, dataCount, tDead, sampling, app, os) VALUES ("+mysql.escape(algorithmId)+","+mysql.escape(nautyId)+","+mysql.escape(stimId)+","+mysql.escape(JSON.stringify(t.modelPre))+","+(t.modelPre.channelCount|0)+","+t.dataCount+","+t.tDead+","+t.sampling+","+mysql.escape(t.app)+","+mysql.escape(t.os)+");")
                    .then(function(result) {
                        return {trialId: result.insertId};
                    });
            }.bind(this));
    };
    exports.Connection.prototype.withNautyId = function(nautyNumber, model) {
        return this.query("SELECT nautyId FROM Nauty WHERE nautyNumber="+nautyNumber+";")
            .then(function(rows) {
                if ( rows.length ) {
                    return {nautyId: rows[0].nautyId};
                }
                let nautyQmj = JSON.parse(JSON.stringify(model));
                let nautyStateCount = nautyQmj.states ? nautyQmj.states.length : 0;
                // rates numbered by Rates.rateIndex
                for (let i=0, n=nautyQmj.rates.length; i<n; ++i) {
                    let r = nautyQmj.rates[i];
                    r.k0 = i;
                    r.k1 = r.k2 = 0;
                }
                return this.query("INSERT INTO Nauty (nautyNumber, nautyQmj, nautyStateCount) VALUES ("+nautyNumber+","+mysql.escape(JSON.stringify(nautyQmj))+","+nautyStateCount+");")
                    .then(function(result) {
                        return {nautyId: result.insertId};
                    });
            }.bind(this));
    };
    exports.Connection.prototype.withStimId = function(stimLbl) {
        return this.query("SELECT stimId FROM Stimulus WHERE stimLbl="+mysql.escape(""+stimLbl)+";")
            .then(function(rows) {
                if ( rows.length ) {
                    return {stimId: rows[0].stimId};
                }
                return this.query("INSERT INTO Stimulus (stimLbl) VALUES ("+mysql.escape(""+stimLbl)+");")
                    .then(function(result) {
                        return {stimId: result.insertId};
                    });
            }.bind(this));
    };
    exports.Connection.prototype.withAlgorithmId = function(name) {
        return this.query("SELECT algorithmId FROM Algorithm WHERE algorithmName="+mysql.escape(""+name)+";")
            .then(function(rows) {
                if ( rows.length ) {
                    return {algorithmId: rows[0].algorithmId};
                }
                return this.query("INSERT INTO Algorithm (algorithmName) VALUES ("+mysql.escape(""+name)+");")
                    .then(function(result) {
                        return {algorithmId: result.insertId};
                    });
            }.bind(this));
    };

    return exports;
});

/*
Algorithm:
  AlgorithmId
  Name (MIL)

Nauty:
  NautyId
  NautyNumber
  NautyQmj
  NautyStateCount

Stimulus:
  StimId
  StimLbl

Trial:
  TrialId
  AlgorithmId
  NautyId
  StimId
  TrialQmj
  ChannelCount
  DataCount
  Tdead
  Sampling
  App
  OS
  Ext1
  Ext2
  Ext3

Rate:
  RateId
  TrialId
  RateIndex
  RateInit
  RateVal
  RateFactor
  RateType (in [0: k0, 1: k1, 2: k2])

*/
