/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("asmffi", [], function() {
    'use strict';

    var asmffiSpec = {
        "dalloc": ["double*", ['size_t']],
        "dfree": ['void', ["double*"]],
        "dget": ['double', ["double*", 'size_t']],
        "dset": ['void', ["double*", 'size_t', 'double']],
        "falloc": ["float*", ['size_t']],
        "ffree": ['void', ["float*"]],
        "fget": ['float', ["float*", 'size_t']],
        "fset": ['void', ["float*", 'size_t', 'float']],
        "ialloc": ["int32*", ['size_t']],
        "ifree": ['void', ["int32*"]],
        "iget": ['int32', ["int32*", 'size_t']],
        "iset": ['void', ["int32*", 'size_t', 'int32']],
        "i64get": ['int64', ["int64*", 'size_t']],
        "i64set": ['void', ["int64*", 'size_t', 'int64']],
        "i16get": ['int16', ["int16*", 'size_t']],
        "i16set": ['void', ["int16*", 'size_t', 'int16']],
        "i8get": ['int8', ["int8*", 'size_t']],
        "i8set": ['void', ["int8*", 'size_t', 'int8']],
        "palloc": ['void*', ['size_t']],
        "pfree": ['void', ['void*']],
        "pget": ['void*', ['void*', 'size_t']],
        "pset": ['void', ['void*', 'size_t', 'void*']],
        "falloc2": ['float**', ['size_t', 'size_t']],
        "ffree2": ['void', ['float**']],
        "frow2": ["float*", ['float**', 'size_t']],
    };
    
    var ffi, ffiLoad;
    try {
        ffi = ( (typeof(window) === 'undefined') &&
                (typeof(require) !== 'undefined') &&
                require('ffi') );
        if ( ffi ) {
            var ref = require('ref');
            var ffiType = {
                'size_t': 'size_t',
                'void': 'void',
                'void*': 'pointer'
            };
            ['double', 'float', 'int', 'int64', 'int32', 'int16', 'int8'].forEach(function(typeName) {
                ffiType[typeName] = typeName;
                var ft = typeName;
                for (var i=0; i<4; ++i) {
                    ffiType[typeName] = ft;
                    typeName += '*';
                    ft = ref.refType(ft);
                }
            });
            ffiLoad = function(libName, spec, rename) {
                return new Promise(function(resolve, reject) {
                    try {
                        var ffiSpec = {};
                        [asmffiSpec, spec].forEach(function (input) {
                            for (var fname in input) {
                                var fspec = input[fname];
                                ffiSpec[fname] = [ffiType[fspec[0]], fspec[1].map(function(t) { return ffiType[t]; })];
                            }
                        });
                        var lib = new ffi.Library('lib'+libName+'.so', ffiSpec);
                        lib.ffi = true;
                        lib.sizeof_p = ref.sizeof['pointer'];
                        for (var fname in (rename || {})) {
                            lib[rename[fname]] = lib[fname];
                            delete lib[fname];
                        }
                        
                        lib.memcpy_in = function(dest, src, srcOffset, bytes) {
                            lib.qubx_memcpy(dest, ref.address(src)+srcOffset, bytes);
                        };
                        lib.memcpy_out = function(dest, destOffset, src, bytes) {
                            lib.qubx_memcpy(ref.address(dest)+destOffset, src, bytes);
                        };
                        lib.memcpy_in_float = function(dest, src, srcOffset, count) {
                            if ( Buffer.isBuffer(src) ) {
                                lib.memcpy_in(dest, src, 4*srcOffset, 4*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    lib.fset(dest, i, src[srcOffset+i]);
                                }
                            }
                        };
                        lib.memcpy_in_double = function(dest, src, srcOffset, count) {
                            if ( Buffer.isBuffer(src) ) {
                                lib.memcpy_in(dest, src, 8*srcOffset, 8*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    lib.dset(dest, i, src[srcOffset+i]);
                                }
                            }
                        };
                        lib.memcpy_in_int = function(dest, src, srcOffset, count) {
                            if ( Buffer.isBuffer(src) ) {
                                lib.memcpy_in(dest, src, 4*srcOffset, 4*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    lib.iset(dest, i, src[srcOffset+i]);
                                }
                            }
                        };
                        lib.memcpy_out_float = function(dest, destOffset, src, count) {
                            if ( Buffer.isBuffer(dest) ) {
                                lib.memcpy_out(dest.buffer, 4*destOffset, src, 4*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    dest[destOffset+i] = lib.fget(src, i);
                                }
                            }
                        };
                        lib.memcpy_out_double = function(dest, destOffset, src, count) {
                            if ( Buffer.isBuffer(dest) ) {
                                lib.memcpy_out(dest.buffer, 8*destOffset, src, 8*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    dest[destOffset+i] = lib.dget(src, i);
                                }
                            }
                        };
                        lib.memcpy_out_int = function(dest, destOffset, src, count) {
                            if ( Buffer.isBuffer(dest) ) {
                                lib.memcpy_out(dest.buffer, 4*destOffset, src, 4*count);
                            } else {
                                for (var i=0; i<count; ++i) {
                                    dest[destOffset+i] = lib.iget(src, i);
                                }
                            }
                        };
                        lib.bzero = function(ptr, count) {
                            var Ni = Math.floor(count/4) | 0;
                            var Nb = count - (4*Ni);
                            for (var i=0; i<Ni; ++i) {
                                lib.iset(ptr, i, 0);
                            }
                            for (var i=0; i<Nb; ++i) {
                                lib.i8set(ptr+4*Ni, i, 0);
                            }
                        };
                        
                        resolve(lib);
                    }
                    catch (err) {
                        reject(err);
                    }
                });
            };
        }
    } catch (err) {
        console.log('ffi:', err);
    }

    var emLoad = function(libName, spec, rename, Module) {
        return new Promise(function(resolve, reject) {
            var asmType = {
                'size_t': 'number',
                'void': null,
                'void*': 'number'
            };
            ['double', 'float', 'int', 'int64', 'int32', 'int16', 'int8'].forEach(function(typeName) {
                asmType[typeName] = 'number';
                asmType[typeName+'*'] = 'number';
                asmType[typeName+'**'] = 'number';
                asmType[typeName+'***'] = 'number';
            });
            Module = Module || {};
            Module.noInitialRun = (typeof(Module.noInitialRun) === 'undefined') ? true : Module.noInitialRun;
            Module.onRuntimeInitialized = function() {
                var lib = {
                    Module: Module,
                    sizeof_p: 4
                };
                [asmffiSpec, spec].forEach(function (input) {
                    for (var fname in input) {
                        var renamed = rename && rename[fname] || fname;
                        var fspec = input[fname];
                        lib[renamed] = Module.cwrap(fname, asmType[fspec[0]], fspec[1].map(function(t) { return asmType[t]; }));
                    }
                });
                lib.memcpy_in = function(dest, src, srcOffset, bytes) {
                    var heapBytes = new Uint8Array(Module.HEAPU8.buffer, dest, bytes);
                    heapBytes.set(new Uint8Array(src, srcOffset, bytes));
                };
                lib.memcpy_out = function(dest, destOffset, src, bytes) {
                    var outBytes = new Uint8Array(dest, destOffset, bytes);
                    outBytes.set(new Uint8Array(Module.HEAPU8.buffer, src, bytes));
                };
                lib.memcpy_in_float = function(dest, src, srcOffset, count) {
                    if ( src.subarray ) {
                        lib.memcpy_in(dest, src.buffer, 4*srcOffset, 4*count);
                    } else {
                        var dbuf = Module.HEAPF32.subarray(dest>>2, (dest>>2)+count);
                        for (var i=0; i<count; ++i) {
                            dbuf[i] = src[srcOffset+i];
                        }
                    }
                };
                lib.memcpy_in_double = function(dest, src, srcOffset, count) {
                    if ( src.subarray ) {
                        lib.memcpy_in(dest, src.buffer, 8*srcOffset, 8*count);
                    } else {
                        var dbuf = Module.HEAPF64.subarray(dest>>3, (dest>>3)+count);
                        for (var i=0; i<count; ++i) {
                            dbuf[i] = src[srcOffset+i];
                        }
                    }
                };
                lib.memcpy_in_int = function(dest, src, srcOffset, count) {
                    if ( src.subarray ) {
                        lib.memcpy_in(dest, src.buffer, 4*srcOffset, 4*count);
                    } else {
                        var dbuf = Module.HEAP32.subarray(dest>>2, (dest>>2)+count);
                        for (var i=0; i<count; ++i) {
                            dbuf[i] = src[srcOffset+i];
                        }
                    }
                };
                lib.memcpy_out_float = function(dest, destOffset, src, count) {
                    if ( dest.subarray ) {
                        lib.memcpy_out(dest.buffer, 4*destOffset, src, 4*count);
                    } else {
                        var sbuf = Module.HEAPF32.subarray(src>>2, (src>>2)+count);
                        for (var i=0; i<count; ++i) {
                            dest[destOffset+i] = sbuf[i];
                        }
                    }
                };
                lib.memcpy_out_double = function(dest, destOffset, src, count) {
                    if ( dest.subarray ) {
                        lib.memcpy_out(dest.buffer, 8*destOffset, src, 8*count);
                    } else {
                        var sbuf = Module.HEAPF32.subarray(src>>3, (src>>3)+count);
                        for (var i=0; i<count; ++i) {
                            dest[destOffset+i] = sbuf[i];
                        }
                    }
                };
                lib.memcpy_out_int = function(dest, destOffset, src, count) {
                    if ( dest.subarray ) {
                        lib.memcpy_out(dest.buffer, 4*destOffset, src, 4*count);
                    } else {
                        var sbuf = Module.HEAP32.subarray(src>>2, (src+count)>>2);
                        for (var i=0; i<count; ++i) {
                            dest[destOffset+i] = sbuf[i];
                        }
                    }
                };
                lib.bzero = function(ptr, count) {
                    var buf = Module.HEAPU8.subarray(ptr, ptr+count);
                    if ( buf.fill ) {
                        buf.fill(0);
                    } else {
                        for (var i=0; i<count; ++i) {
                            buf[i] = 0;
                        }
                    }
                };
                
                resolve(lib);
            };
                           
            require([libName], function(asmMod) {
                if ( asmMod ) {
                    Module = asmMod(Module);
                } else {
                    reject(new Error('Failed to load asm module:',libName));
                }
            });
        });
    };

    var addHelpers = function(lib) {
        var typeInfo = {
            'i8': {
                size: 1,
                get: lib.i8get,
                set: lib.i8set
            },
            'i16': {
                size: 2,
                get: lib.i16get,
                set: lib.i16set
            },
            'i32': {
                size: 4,
                get: lib.iget,
                set: lib.iset
            },
            'i64': {
                size: 8,
                get: lib.i64get,
                set: lib.i64set
            },
            'float': {
                size: 4,
                get: lib.fget,
                set: lib.fset
            },
            'double': {
                size: 8,
                get: lib.dget,
                set: lib.dset
            },
            '*': {
                size: lib.sizeof_p,
                get: lib.pget,
                set: lib.pset
            }
        };
        
        function access(ptr, ftyp, thisVal, name) {
            var accessor;
            var star = ftyp.lastIndexOf('*');
            if ( star > 0 ) {
                var info = typeInfo[ftyp.slice(0, ftyp.indexOf('*'))];
                var levels = ftyp.lastIndexOf('*') - ftyp.indexOf('*');
                // console.log(name, 'of type', typ, 'in array at', arr, 'size', sz)
                // if ( ! info ) { console.log('missing type',typ); }
                accessor = function(i, v) {
                    var arr = lib.pget(ptr, 0);
                    for (var j=0; j<levels; ++j) {
                        arr = lib.pget(arr, 0);
                    }
                    if ( typeof(v) === 'undefined' ) {
                        return info.get(arr, i);
                    }
                    info.set(arr, i, v);
                    return thisVal;
                };
            } else {
                var info = typeInfo[ftyp];
                // console.log(name, 'of type', ftyp, 'at', ptr);
                accessor = function(v) {
                    if ( typeof(v) === 'undefined' ) {
                        return info.get(ptr, 0);
                    }
                    info.set(ptr, 0, v);
                    return thisVal;
                };
            }
            accessor.ptr = ptr;
            return accessor;
        }
        lib.addStructAccess = function(obj, ptr, fields) {
            var at = 0;
            for (var i=0, n=fields.length; i<n; ++i) {
                var f = fields[i];
                var star = f.typ.lastIndexOf('*');
                obj[f.name] = access(ptr+at, f.typ, obj, f.name);
                at += (star !== -1) ? lib.sizeof_p : typeInfo[f.typ].size;
            }
        };
        
        lib.GrowArray = function(byteSize, alloc, get, set, free, memcpy_in, memcpy_out, ArrType, growthFactor) {
            this._byteSize = byteSize;
            this._alloc = alloc || lib.ialloc;
            this._get = get || lib.iget;
            this._set = set || lib.iset;
            this._free = free || lib.ifree;
            this._memcpy_in = memcpy_in || lib.memcpy_in_int;
            this._memcpy_out = memcpy_out || lib.memcpy_out_int;
            this._ArrType = ArrType || Int32Array;
            this._growthFactor = growthFactor || 2;
            this.p = 0;
            this._n = 0;
            this._cap = 0;
        };
        lib.GrowArray.prototype.destroy = function() {
            if ( this.p ) {
                this._free(this.p);
                this.p = 0;
            }
        };
        lib.GrowArray.prototype.size = function(n) {
            if ( typeof(n) === 'undefined' ) {
                return this._n;
            }
            if ( n > this._cap ) {
                if ( this._cap ) {
                    this._free(this.p);
                } else {
                    this._cap = 1;
                }
                while ( this._cap < n ) {
                    this._cap = Math.round(this._cap * this._growthFactor)|0;
                }
                this.p = this._alloc(this._cap);
            }
            this._n = n;
            return this;
        };
        lib.GrowArray.prototype.get = function(i) {
            if ( typeof(i) === 'undefined' ) {
                var all = [];
                for (var j=0; j<this._n; ++j) {
                    all[j] = this.get(j);
                }
                return all;
            }
            return this._get(this.p, i);
        };
        lib.GrowArray.prototype.set = function(i, v) {
            if ( typeof(i.length) !== 'undefined' ) { // args as (v, n)
                v = v || i.length;
                this.size(v);
                this._memcpy_in(this.p, i, 0, v);
            } else {
                this._set(this.p, i, v);
            }
            return this;
        };
        lib.GrowArray.prototype.contents = function(n, arr, offset) {
            if ( typeof(n) === 'undefined' ) {
                n = this._n;
            }
            arr = arr || new this._ArrType(n);
            offset = offset|0;
            this._memcpy_out(arr, 0, this.p+this._byteSize*offset, n);
            return arr;
        };
        
        lib.ivec = function() { return new lib.GrowArray(4); };
        lib.fvec = function() { return new lib.GrowArray(4, lib.falloc, lib.fget, lib.fset, lib.ffree, lib.memcpy_in_float, lib.memcpy_out_float, Float32Array); };
        lib.dvec = function() { return new lib.GrowArray(8, lib.dalloc, lib.dget, lib.dset, lib.dfree, lib.memcpy_in_double, lib.memcpy_out_double, Float64Array); };
        // don't trust ptr-as-int; so don't use pvec's set(arr) or contents()
        lib.pvec = function() { return new lib.GrowArray(lib.sizeof_p, lib.palloc, lib.pget, lib.pset, lib.pfree, lib.memcpy_in_int, lib.memcpy_out_int, Int32Array); }

        return lib;
    };
    
    return {
        load: function(libName, spec, rename, Module) {
            var stepOne = Promise.resolve(false);
            if ( ffiLoad ) {
                stepOne = ffiLoad(libName, spec, rename);
            }
            return stepOne.catch(function(err) {
                console.log('ffi',libName,err);
                return false;
            }).then(function(lib) {
                return lib || emLoad(libName, spec, rename, Module);
            }).then(addHelpers);
        },
        ffi: !!ffiLoad
    };
});
