/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("cloud", ["qubx"], function(qubx) {
    "use strict";

    var gSeatKey = localStorage.getItem('cloud.seatKey');
    var gInviteKey = '';
    var gPersonId = 0;
    var ss = function(v) { return v ? (''+v) : ''; };
    var dd = function(v) {
        var d = new Date(v);
        return Number.isNaN(d.getTime()) ? v : d;
    };
    var dds = function(dateFields) {
        return function(x) {
            dateFields.forEach(function(f) {
                x[f] = dd(x[f]);
            });
            return x;
        }
    };
    var dd_model = dds(['modelWhenAdded', 'modelWhenModified', 'modelWhenPub']);
    var dd_dataset = dds(['datasetWhenRec', 'datasetWhenAdded']);
    var dd_project = dds(['projectWhenCreated', 'projectWhenModified', 'projectWhenPub']);
    var dd_result = dds(['resultWhen']);
    var dd_seat = dds(['seatWhen']);
    var dd_server = dds(['serverWhenFirst', 'serverWhenLast']);
    
    var server = qubx.wsAuth('wss'+window.location.origin.replace(/^https?/, '')+'/online/cloud/ui', {
        requestAuth: function(ws) {
            if ( this.authRq ) {
                return this.authRq;
            }
            var authRq;
            if ( gInviteKey ) {
                var inv = gInviteKey;
                gInviteKey = '';
                authRq = ws.request({
                    t: 'signIn',
                    inviteKey: inv
                }).then(function(reply) {
                    gSeatKey = reply.seatKey;
                    localStorage.setItem("cloud.seatKey", gSeatKey);
                });
            } else if ( gSeatKey ) {
                authRq = ws.request({
                    t: 'auth',
                    seatKey: gSeatKey
                });
            } else {
                return Promise.reject('No credentials');
            }
            this.authRq = authRq.then(function() {
                return ws.request({t: 'aboutPerson'})
                    .then(function(result) {
                        gPersonId = result.personId;
                        exports.personGroup = result.personGroup;
                        exports.personName = result.personName;
                        exports.personAbout = result.personAbout;
                        exports.personEmail = result.personEmail;
                        exports.personIsSuper = result.personIsSuper;
                        exports.personIsContact = result.personIsContact;
                        exports.personIsAdmin = result.personIsAdmin;
                        exports.listGroups(exports.personGroup).then(function(groups) {
                            exports.group = groups[0];
                        });
                        this.authRq = undefined;
                        exports.ev.signin.trigger(result.personIsAdmin)
                        return true;
                    }.bind(this));
            }.bind(this))
                .catch(function(err) {
                    this.authRq = undefined;
                    return Promise.reject(err);
                }.bind(this));
            return this.authRq;
        },
        ondeauth: function() {
            exports.personName = '';
            exports.personAbout = '';
            exports.personEmail = '';
            exports.personIsSuper = exports.personIsAdmin = exports.personIsContact = false;
            exports.group = {};
            gPersonId = 0;
            setTimeout(tryReauth, 5000);
            exports.ev.signout.trigger();
        },
        onmessage: function(event, msg) {
            if ( ! (msg && msg.t) ) {
                console.log('cloud: unexpected empty message');
                return;
            }
            if ( msg.t === 'ejected' ) {
                server.ondeauth();
                exports.ev.ejected.trigger(reason);
            } else if ( msg.t === 'screenHopefuls' ) {
                exports.ev.hopefuls.trigger(msg.hopefuls);
            } else if ( msg.t === 'accepted' ) {
                gPersonId = msg.person.personId;
                exports.personName = msg.person.personName;
                exports.personGroup = msg.person.personGroup;
                exports.personAbout = msg.person.personAbout;
                exports.personEmail = msg.person.personEmail;
                exports.personIsSuper = msg.person.personIsSuper;
                exports.personIsAdmin = msg.person.personIsAdmin;
                exports.personIsContact = msg.person.personIsContact;
                exports.listGroups(exports.personGroup).then(function(groups) {
                    exports.group = groups[0];
                    exports.ev.accepted.trigger(msg.person);
                    exports.ev.signin.trigger(msg.person.personIsAdmin);
                });
            }
        }
    });

    function tryReauth() {
        exports.checkSignedIn()
            .catch(function(err) {
                try {
                    if ( err.indexOf('Socket') >= 0 ) {
                        setTimeout(tryReauth, 5000);
                    }
                }
                catch(maybeFirstErrWasntString)
                {}
            });
    }

    function updateGroup() {
        exports.group.groupId && exports.listGroups(exports.group.groupId).then(function(groups) {
            exports.group = groups[0];
        });
    }
    var schedUpdateGroup = qubx.batchDelayed(updateGroup, 500, true);

    function updatePerson() {
        gPersonId && exports.aboutPerson(gPersonId).then(function(p) {
            exports.personName = p.personName;
            exports.personAbout = p.personAbout;
            exports.personEmail = p.personEmail;
        });
    }
    var schedUpdatePerson = qubx.batchDelayed(updatePerson, 500, true);

    var exports = {
        ev: qubx.makeEvents(['signin', 'signout', 'ejected', 'accepted', 'hopefuls']),
        personName: "",
        personAbout: "",
        personEmail: "",
        group: {},
        checkSignedIn: function() {
            return server.request({t: 'signedIn'});
        },
        signIn: function(key) {
            var signOutFirst = gSeatKey ? exports.signOut() : Promise.resolve(true);
            return signOutFirst.then(function() {
                gInviteKey = key;
                return exports.aboutPerson();
            });
        },
        signOut: function() {
            return server.request({
                t: 'signOut'
            }).then(function(result) {
                server.deauth();
                gSeatKey = "";
                localStorage.setItem("cloud.seatKey", gSeatKey);
                exports.personName = "";
                exports.personAbout = "";
                exports.personEmail = "";
                exports.personIsSuper = exports.personIsAdmin = exports.personIsContact = false;
                return result;
            });
        },
        aboutPerson: function(personId) {
            return server.request(Object.assign({t: 'aboutPerson'}, personId ? {personId: personId|0} : {}));
        },

        ping: function(x) {
            return server.request(Object.assign({
                t: 'ping',
            }, x), true);
        },
        log: function(x) {
            return server.request({
                t: 'log',
                x: JSON.stringify(x)
            }, true);
        },
        addGroup: function(g) {
            // groupName, groupAbout, groupMaxPerson, groupMaxAdmin, groupMaxSeats, groupComputeMinutes => {}.groupId
            return server.request(Object.assign({t: 'addGroup'}, g))
                .then(function(reply) {
                    return reply.groupId;
                });
        },
        listGroups: function(groupId) {
            return server.request(Object.assign({t: 'listGroups'}, groupId ? {groupId: groupId|0} : {}))
                .then(function(reply) { return reply.groups; });
        },
        setGroupActive: function(groupId, active) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupActive',
                groupId: groupId|0,
                groupActive: !! active
            });
        },
        setGroupName: function(groupId, name) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupName',
                groupId: groupId|0,
                groupName: ss(name)
            });
        },
        setGroupAbout: function(groupId, about) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupAbout',
                groupId: groupId|0,
                groupAbout: ss(about)
            });
        },
        setGroupDataKey: function(groupId, dataKey) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupDataKey',
                groupId: groupId|0,
                groupDataKey: ss(dataKey)
            });
        },
        setGroupComputeKey: function(groupId, computeKey) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupComputeKey',
                groupId: groupId|0,
                groupComputeKey: ss(computeKey)
            });
        },
        setGroupMaxPerson: function(groupId, maxPerson) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupMaxPerson',
                groupId: groupId|0,
                groupMaxPerson: maxPerson|0
            });
        },
        setGroupMaxAdmin: function(groupId, maxAdmin) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupMaxAdmin',
                groupId: groupId|0,
                groupMaxAdmin: maxAdmin|0
            });
        },
        setGroupMaxSeats: function(groupId, maxSeats) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupMaxSeats',
                groupId: groupId|0,
                groupMaxSeats: maxSeats|0
            });
        },
        setGroupComputeMinutes: function(groupId, computeMinutes) {
            (groupId === exports.group.groupId) && schedUpdateGroup();
            return server.request({
                t: 'setGroupComputeMinutes',
                groupId: groupId|0,
                groupComputeMinutes: computeMinutes|0
            });
        },
        addPerson: function(p) {
            // personGroup, personIsAdmin, personIsContact, personIsSuper, personName, personAbout, personEmail => {}.personId
            return server.request(Object.assign({t: 'addPerson'}, p))
                .then(function(reply) {
                    return reply.personId;
                });
        },
        listPeople: function(groupId) {
            return server.request(Object.assign({t: 'listPeople'}, groupId ? {groupId: groupId|0} : {}))
                .then(function(reply) { return reply.people; });
        },
        setPersonGroup: function(personId, groupId) {
            if ( typeof(groupId) === 'undefined' ) {
                groupId = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonGroup',
                personId: personId|0,
                groupId: groupId|0
            });
        },
        setPersonActive: function(personId, active) {
            if ( typeof(active) === 'undefined' ) {
                active = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonActive',
                personId: personId|0,
                personActive: !!active
            });
        },
        setPersonIsAdmin: function(personId, isAdmin) {
            if ( typeof(isAdmin) === 'undefined' ) {
                isAdmin = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonIsAdmin',
                personId: personId|0,
                personIsAdmin: !!isAdmin
            });
        },
        setPersonIsContact: function(personId, isContact) {
            if ( typeof(isContact) === 'undefined' ) {
                isContact = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonIsContact',
                personId: personId|0,
                personIsContact: !!isContact
            });
        },
        setPersonIsSuper: function(personId, isSuper) {
            if ( typeof(isSuper) === 'undefined' ) {
                isSuper = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonIsSuper',
                personId: personId|0,
                personIsSuper: !!isSuper
            }).then(function() {
                if ( gPersonId === personId ) {
                    exports.personIsSuper = isSuper;
                }
            });
        },
        setPersonName: function(personId, name) {
            if ( typeof(name) === 'undefined' ) {
                name = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonName',
                personId: personId|0,
                personName: ss(name)
            }).then(function() {
                if ( gPersonId === personId ) {
                    exports.personName = name;
                }
            });
        },
        setPersonAbout: function(personId, about) {
            if ( typeof(about) === 'undefined' ) {
                about = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonAbout',
                personId: personId|0,
                personAbout: ss(about)
            }).then(function() {
                if ( gPersonId === personId ) {
                    exports.personAbout = about;
                }
            });
        },
        setPersonEmail: function(personId, email) {
            if ( typeof(email) === 'undefined' ) {
                email = personId;
                personId = gPersonId;
            }
            (personId === gPersonId) && schedUpdatePerson();
            return server.request({
                t: 'setPersonEmail',
                personId: personId|0,
                personEmail: ss(email)
            }).then(function() {
                if ( gPersonId === personId ) {
                    exports.personEmail = email;
                }
            });
        },
        requestAccount: function(name, email, about, groupInfo) {
            return server.request({
                t: 'requestAccount',
                name: name,
                email: email,
                about: about,
                groupInfo: groupInfo
            }, true).then(function(reply) {
                gSeatKey = reply.seatKey;
                localStorage.setItem("cloud.seatKey", gSeatKey);
                exports.personName = exports.personEmail = exports.personAbout = '';
                exports.personIsSuper = exports.personIsAdmin = exports.personIsContact = false;
                exports.group = {};
            });
        },
        listHopeful: function(groupId) {
            return server.request(Object.assign({t: 'listHopeful'}, groupId ? {groupId: groupId|0} : {}))
                .then(function(reply) { return reply.hopefuls; });
        },
        fwdHopeful: function(hopefulId, groupId) {
            return server.request({
                t: 'fwdHopeful',
                hopefulId: hopefulId|0,
                groupId: groupId|0
            });
        },
        bwdHopeful: function(hopefulId) {
            return server.request({
                t: 'bwdHopeful',
                hopefulId: hopefulId|0
            });
        },
        acceptHopeful: function(hopefulId, isAdmin, isContact) {
            return server.request({
                t: 'acceptHopeful',
                hopefulId: hopefulId|0,
                personIsAdmin: !!isAdmin,
                personIsContact: !!isContact
            }).then(function(reply) {
                return reply.personId;
            });
        },
        acceptCandidate: function(hopefulId, personId) {
            return server.request({
                t: 'acceptCandidate',
                hopefulId: hopefulId|0,
                personId: personId|0
            });
        },
        rejectHopeful: function(hopefulId) {
            return server.request({
                t: 'rejectHopeful',
                hopefulId: hopefulId|0
            });
        },
        invite: function(personId, expiresHours, usesRemain) {
            return server.request({
                t: 'invite',
                personId: personId,
                expiresHours: expiresHours,
                usesRemain: usesRemain
            }).then(function(reply) {
                return reply.inviteKey;
            });
        },
        listSeats: function(groupId) { // no group id: my seats; groupId===false: include this seat
            return server.request(Object.assign({t: 'listSeats'}, groupId ? {groupId: groupId|0} : {}))
                .then(function(reply) {
                    return (groupId === false) ? reply.seats : reply.seats.filter(function(seat) {
                        return seat.seatKey !== gSeatKey;
                    }).map(dd_seat);
                });
        },
        unSignIn: function(otherSeatKey, otherPersonId) {
            var rq = {
                t: 'unSignIn',
                seatKey: otherSeatKey
            };
            if ( otherPersonId ) {
                rq.otherPersonId = otherPersonId;
            }
            return server.request(rq);
        },
        addModel: function(m) {
            // modelPublic, modelName, modelQmj, modelAbout, modelAuthors, modelDOI, modelStates, modelClasses, modelWhenPub, modelFolder => {}.modelId
            return server.request(Object.assign({t: 'addModel'}, m));
        },
        updateModel: function(m) {
            return server.request(Object.assign({t: 'updateModel'}, m));
        },
        listModels: function(q) { //// also getModel: m.when* = dd(m.when*) convert from sql strings; also reading any *When* column
            // modelPerson, modelGroup, modelStates, modelClasses, modelDigest, modelFolder
            return server.request(Object.assign({t: 'listModels'}, q), true)
                .then(function(reply) {
                    return reply.models.map(dd_model);
                });
        },
        getModel: function(modelId) {
            return server.request({
                t: 'getModel',
                modelId: modelId|0
            }, true).then(function(reply) {
                return dd_model(reply);
            });
        },
        listServers: function(groupId) {
            return server.request({
                t: 'listServers',
                groupId: groupId|0
            }).then(function(reply) {
                return reply.servers.map(dd_server);
            });
        },
        setDatasetPublic: function(datasetId, datasetPublic) {
            return server.request({
                t: 'setDatasetPublic',
                datasetId: datasetId|0,
                datasetPublic: !!datasetPublic
            });
        },
        setDatasetFolder: function(datasetId, datasetFolder) {
            return server.request({
                t: 'setDatasetFolder',
                datasetId: datasetId|0,
                datasetFolder: ss(datasetFolder)
            });
        },
        listDatasets: function(q) {
            // personId, groupId, datasetDigest, datasetFolder
            return server.request(Object.assign({t: 'listDatasets'}, q), true)
                .then(function(reply) {
                    return reply.datasets.map(dd_dataset);
                });
        },
        pickDataInstance: function(datasetId) {
            return server.request({
                t: 'pickDataInstance',
                datasetId: datasetId|0
            }); // -> {serverBaseUrl, instancePath}
        },
        addProject: function(p) {
            // projectPublic, projectName, projectAbout, projectAuthors, projectDOI, projectType, projectConfig, projectFolder => {}.projectId
            return server.request(Object.assign({t: 'addProject'}, p));
        },
        updateProject: function(p) {
            return server.request(Object.assign({t: 'updateProject'}, p));
        },
        addProjectModel: function(projectId, modelId) {
            return server.request({
                t: 'addProjectModel',
                projectId: projectId,
                modelId: modelId
            });
        },
        removeProjectModel: function(pmId, projectId, modelId) {
            return server.request({
                t: 'removeProjectModel',
                pmId: pmId,
                projectId: projectId,
                modelId: modelId
            });
        },
        getProjectModels: function(projectId) {
            return server.request({
                t: 'getProjectModels',
                projectId: projectId|0
            }, true).then(function(reply) { return reply.models; });
        },
        setProjectModelActive: function(projectId, modelId, active) {
            return server.request({
                t: 'setProjectModelActive',
                projectId: projectId|0,
                modelId: modelId|0,
                active: !!active
            });
        },
        addProjectDataset: function(projectId, datasetId) {
            return server.request({
                t: 'addProjectDataset',
                projectId: projectId|0,
                datasetId: datasetId|0
            });
        },
        removeProjectDataset: function(pdId, projectId, datasetId) {
            return server.request({
                t: 'removeProjectDataset',
                pdId: pdId|0,
                projectId: projectId|0,
                datasetId: datasetId|0
            });
        },
        getProjectDatasets: function(projectId) {
            return server.request({
                t: 'getProjectDatasets',
                projectId: projectId|0
            }, true).then(function(reply) { return reply.datasets; });
        },
        setProjectDatasetActive: function(projectId, datasetId, active) {
            return server.request({
                t: 'setProjectDatasetActive',
                projectId: projectId|0,
                datasetId: datasetId|0,
                active: !!active
            });
        },
        addProjectResult: function(projectId, r) {
            // resultConfig, resultLL, resultWhen, resultResults
            /// move all times to GMT before sending?
            return server.request(Object.assign({
                t: 'addProjectResult',
                resultProject: projectId|0
            }, r));
        },
        delProjectResult: function(projectId, resultId) {
            return server.request({
                t: 'delProjectResult',
                projectId: projectId|0,
                resultId: resultId|0
            });
        },
        listProjectResults: function(projectId) {
            return server.request({
                t: 'listProjectResults',
                projectId: projectId|0
            }, true).then(function(reply) { return reply.results; });
        },
        getProjectResult: function(projectId, resultId) {
            return server.request({
                t: 'getProjectResult',
                projectId: projectId,
                resultId: resultId
            }, true).then(function(reply) {
                return dd_result(reply);
            });
        },
        listProjects: function(q) {
            // personId, groupId, projectType, projectDigest, datasetId, modelId, projectFolder
            return server.request(Object.assign({t: 'listProjects'}, q), true)
                .then(function(reply) {
                    return reply.projects.map(dd_project);
                });
        },
        delProject: function(projectId) {
            return server.request({
                t: 'delProject',
                projectId: projectId|0
            });
        },
        startUpload: function(f) {
            // datasetPublic, datasetGroup, datasetIsSampled, datasetIsIdealized, datasetName, datasetAbout, datasetWhenRec,
            // datasetSampling, datasetPoints, datasetEvents, datasetSegments, datasetStimuli, datasetConstants, datasetMolecule, datasetEtc, datasetFolder
            // => upload={datasetId, fileKey, serverId, serverBaseUrl, serverPutUrl, path}
            return server.request(Object.assign({t: 'startUpload'}, f));
        },
        uploading: function(upload) {
            return server.request({
                t: 'uploading',
                serverId: upload.serverId|0,
                datasetId: upload.datasetId|0,
                fileKey: ss(upload.fileKey)
            });
        },
        uploadFailed: function(upload) {
            return server.request({
                t: 'uploadFailed',
                serverId: upload.serverId|0,
                datasetId: upload.datasetId|0,
                fileKey: ss(upload.fileKey) 
           });
        },
        // data conversion:
        canPutFile: function(name, size) {
            return server.request({
                t: 'canPutFile',
                name: name,
                size: size
            }, true);
        },
        releaseFile: function(url) {
            return server.request({
                t: 'releaseFile',
                url: url
            }, true);
        },
        convertFile: function(url) {
            return server.request({
                t: 'convertFile',
                url: url
            }, true);
        },
        cancelConvertFile: function() {
            return server.request({
                t: 'cancelConvertFile'
            }, true);
        }
    };

    if ( gSeatKey ) {
        exports.checkSignedIn()
            .catch(function(err) {
                //gSeatKey = '';
                //localStorage.setItem("cloud.seatKey", gSeatKey);
                console.log(err);
            });
    }
    
    exports.intervalHandle = setInterval(exports.ping, 15*60*1000);
    
    return exports;
});

/// logging?
