/* Copyright 2016-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("fitframe", ["qubx"], qubx => {
    'use strict';

    // Class Remote instantiates a sandboxed iframe with the contents of './fitframe.html'.
    // It mediates cross-window function calls using postMessage;
    // methods return a Promise resolving to the result of the call
    // (or rejecting on error with the full returned msg including msg.err, the thrown exception).
    // all methods take a json-ready object and return a json-ready object. Optional second argument
    // transfer array works just like raw postMessage()
    
    var exports = {};

    class Remote {
        constructor(href) {
            this.sched = new qubx.PromiseTickets(),
            this.iframe = document.createElement('iframe');
            this.iframe.sandbox = 'allow-scripts';
            this.iframe.style.display = 'none';
            this.iframe.src = href;
            document.body.appendChild(this.iframe);
            this.remote = this.iframe.contentWindow;
            
            this._listener = this._onMessage.bind(this);
            window.addEventListener('message', this._listener);
            
            this.readyWait = [];
            this.iframe.addEventListener('load', () => {
                this.readyWait.forEach(resolve => resolve(this));
                this.readyWait = null;
            });
        }
        stop() {
            document.body.removeChild(this.iframe);
            window.removeEventListener('message', this._listener);
        }
        waitReady() {
            if ( this.readyWait ) {
                return new Promise((resolve, reject) => this.readyWait.push(resolve));
            } else {
                return Promise.resolve(this);
            }
        }
        method(name) {
            this[name] = (msg, transfer) => new Promise((resolve, reject) => this.remote.postMessage({
                t: name,
                msg: msg,
                tkt: this.sched.register(resolve, reject)
            }, '*', transfer));
            return this;
        }
        _onMessage(event) {
            if ( event.source === this.remote ) {
                if ( event.data.err ) {
                    this.sched.reject(event.data.tkt, event.data);
                } else {
                    this.sched.resolve(event.data.tkt, event.data.msg);
                }
            }
        }
    };

    return qubx.get('fitframe.html').then(html => {
        const blob = new Blob([html], {type: 'text/html'}),
              dataUri = window.URL.createObjectURL(blob);
        return new Remote(dataUri)
            .method('newContext')
            .method('delContext')
            .method('clear')
            .method('setCol')
            .method('setXY')
            .method('setWeight')
            .method('setFunc')
            .method('rss')
            .method('getFit');
    }).then(remote => {
        class Context {
            constructor() {
                // Proxy for fitframe.html:Context
                this.info = remote.newContext();
                this
                    .method('delContext')
                    .method('clear')
                    .method('setCol')
                    .method('setXY')
                    .method('setWeight')
                    .method('setFunc')
                    .method('rss')
                    .method('getFit');
                this.destroy = this.delContext;
            }
            // instance methods are routed by ctx number
            method(name) {
                this[name] = (msg, transfer) => this.info.then(info => {
                    msg.ctx = info.ctx;
                    return remote[name](msg, transfer);
                });
                return this;
            }
        }
        return remote.waitReady()
            .then(() => ({Context}));
    });
});


    
