/* Copyright 2016-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("fitness_core", ["qubx", "qubx_lib", "qubx_optimize", "lmfit", "fitframe"], (qubx, qubx_lib, qubx_optimize, lmfit, fitframe) => {
    'use strict';

    const exports = {};

    // Curve fitting of column data to user-entered functions,
    // using a sandboxed iframe, './fitframe.html', for evaluation,
    // controlled via a Context instance so multiple sessions can coexist
    
    // Columns should all have the same length, and have names
    // which are valid and unique javascript identifiers, e.g.
    //    setCol({name: 'TIME', arr: new Float32Array([.1, .2, .3, ...])})
    
    // A function is minimally in terms of x and some unknown parameter(s), e.g.
    //    scale * x
    // with JS-style math: +,-,*,/,(,) along with common math functions, e.g.
    //    amp * sin(x)
    // The zero-based row index is also available as 'i', e.g.
    //    scale * i
    // We ask you to identify the primary columns:
    //    ctx.setXY(xColName, yColName)
    // but you can additionally specify columns in a function, e.g.
    //    MEASURED = scale * TIME
    // This is useful for fitting multiple columns at once, e.g. with same slope, different intercept:
    //    M1 = slope * x + icept1 # M2 = slope * x + icept2
    // Column names are not necessarily capitalized, but they are case-sensitive.
    // I'm capitalizing them all here for clarity.

    // A function can define auxiliary variables in a "common" block, e.g.
    //    common: {denom: 'a+b'} # MEASURED = a*x/denom + x/(denom+1)
    // Common blocks and other definition sections are set off by hashes.
    // Anything in a common block is available for functions to use.
    // When a common item is defined in quotes ('a+b'), it has access to
    // data, columns by name, and other parameters, and it gets reinterpreted for new parameter values.
    // Without quotes, it can only access the Data object (see below).

    // You might also define functions in a common block, e.g.
    //    amp*fib(i) # common: {fib: function(n) {return (n < 2) ? 1 : (fib(n-1) + fib(n-2));}}
    // See "advanced rules" below.

    // Aside from common blocks, definitions can initialize and constrain parameters, e.g.
    //    amp * sin(x) # amp: {lo: 1e-8}
    //    amp * sin(x) # amp: {lo: 1e-8, hi: 20, val: 2.2}
    // "val", or a plain number (e.g. "phase" below) is the parameter's initial value.
    //    amp * sin(x + phase) # amp: {val: 2.2, canFit: false}, phase: 0.0

    // You can also initialize parameters from data, e.g.
    //    amp * sin(x) # amp: Data.mean('MEASURED')
    // Data.mean, Data.std, Data.min, Data.argmin, Data.max, Data.argmax can be called with any column name.
    // You can even access individual datapoints with Data.column('NAME')[index],
    // with index between 0 and (Data.n - 1).

    // Definitions can also assign weight to a data column, e.g.
    //    MEASURED: {weight: 0.5}
    // which is used to scale the column residual, particularly when there are multiple functions.
    // Quite distinctly, you can setWeightExpr(str), where str is an expression in terms of column names, e.g.
    //    setWeightExpr('1/MEASURED')

    // Advanced rules for common blocks:
    //    - there can be zero or more
    //    - unquoted values can only use Data and the Math functions
    //    - quoted expressions can also refer to column names, parameters, and variables from prior common blocks
    //    - they can't refer to variables defined in the same common block, or a subsequent one
    //    - unknown names in expressions become new parameters to the system
    
    // Make sure to setFunc (then optionally setWeightExpr) after calling setColumns or setXY,
    // before calling fit(). During and after fitting, getFit() will retrieve the latest fit curve(s).
    // After fitting, getErr() fills out param.std from the sqrt diagonal of the inverse second derivative matrix.

    // context events:
    //   ev.data(): after setColumns or setXY
    //   ev.expr(expr, {params, columns, series}): after setExpr
    //   ev.weightExpr(expr): after setWeightExpr
    //   ev.param({name, lo, hi, canFit, val}): after setParam, or during fitting
    //   ev.startFit(): upon fit()
    //   ev.endFit(): after fit()

    // See fitframe.html for further details

    
    const paramStore = {};
    const protoParam = {
        name: '',
        val: 1.0,
        std: 0.0,
        lo: qubx.UNSET,
        hi: qubx.UNSET,
        canFit: true,
        set: function(p) {
            let modified = false;
            if ( (typeof(p.lo) !== 'undefined') && (p.lo !== this.lo) ) {
                this.lo = p.lo;
                modified = true;
            }
            if ( (typeof(p.hi) !== 'undefined') && (p.hi !== this.hi) ) {
                this.hi = p.hi;
                modified = true;
            }
            if ( (typeof(p.canFit) !== 'undefined') && (p.canFit !== this.canFit) ) {
                this.canFit = p.canFit;
                modified = true;
            }
            let v = p.val;
            if ( typeof(v) === 'undefined' ) {
                if ( p === +p ) {
                    v = p;
                }
            }
            if ( (typeof(v) !== 'undefined') && (v !== this.val) ) {
                if ( (this.lo !== qubx.UNSET) && (v < this.lo) ) {
                    return modified;
                }
                if ( (this.hi !== qubx.UNSET) && (v > this.hi) ) {
                    return modified;
                }
                this.val = v;
                modified = true;
            }
            return modified;
        },
        toJSON: function() {
            const tmp = {};
            for (const key of Object.keys(this)) {
                if( typeof this[key] !== 'function' )
                    tmp[key] = this[key];
            }
            return tmp;
        }
    };
    function getParam(name, overrides) {
        // Returns a param object {name, lo, hi, val, canFit} from the persistent store, creating if needed
        //   overrides: optional object used to initialize param if it doesn't exist
        let p = paramStore[name];
        if ( p ) {
            return p;
        }
        p = Object.create(protoParam);
        Object.assign(p, overrides, {name: name});
        paramStore[name] = p;
        return p;
    }
    function newParam(name, overrides, localOverrides) {
        // Returns a "copy" (prototype inheritor) of the persistent param
        //   overrides: used to initialize the persistent param if it doesn't exist
        //   localOverrides: used to customize the new param "copy" object
        // Each Context instance gets its own copies.
        return Object.assign(Object.create(getParam(name, overrides)), localOverrides);
    }

    const ContextDefs = {
        expr: 'm*x + b',
        weightExpr: '1.0'
    };
    exports.Context = class Context {
        constructor(config) {
            this.config = Object.assign({}, ContextDefs, config);
            this.ev = qubx.makeEvents(['data', 'xy', 'expr', 'weightExpr', 'param', 'startFit', 'endFit']);
            this.helper = fitframe.then(module => new module.Context());
            this.paramStore = this.config.paramStore || {};
            this.config.paramStore = this.paramStore;
            for (const name of Object.keys(this.paramStore)) {
                this.paramStore = Object.assign(Object.create(protoParam), this.paramStore[name]);
            }
            this.data = {};
            this.exprInfo = {
                params: [],
                columns: [],
                series: []
            };
        }
        destroy() {
            return this.helper.then(helper => helper.destroy());
        }
        clear() {
            // releases iframe resources in the iframe (and in this proxy)
            return this.helper
                .then(helper => helper.clear({}))
                .then(() => {
                    this.data = {};
                    this.ev.data.trigger();
                });
        }
        setColumns(data) {
            // sets/replaces named columns of data; e.g. {MEASURED1: new Float32Array(n), ...}
            const names = Object.keys(data);
            return this.helper.then(helper => Promise.all(names.map(name => helper.setCol({
                name: name,
                arr: data[name]
            })))).then(() => {
                if ( names.length ) {
                    Object.assign(this.data, data);
                    const arr = data[names[0]];
                    if ( (! this.data.i) || (this.data.i.length !== arr.length) ) {
                        this.data.i = new Float32Array(qubx.range(arr.length));
                    }
                    this.ev.data.trigger();
                }
            });
        }
        setXY(xName, yName) {
            // picks primary columns:
            // xName: selects meaning of 'x' in expressions, decides Data.n if columns disagree
            // yName: fitted to anonymous functions (no equals sign e.g. 'NAME=m*x+b', thus fit to the default yName column)
            return this.helper.then(helper => {
                this.xName = xName;
                this.yName = yName;
                return helper.setXY({
                    x: xName,
                    y: yName
                });
            }).then(() => this.ev.xy.trigger(xName, yName));
        }
        setExpr(expr) {
            // expr: e.g. 'm*x + b; m: 2.0, "b": {lo: 0.0, hi: 10.0, val: 0.5, canFit: false}'
            // resolves: {params, columns: lists of names found in the expression (unknown, existing column)
            //            defs: contents of definition sections in f, except common blocks
            //            series: array of {
            //                name: target fitting column, or undefined to fit against columns[yName]
            //                expr: equation string
            //            }}
            return this.helper.then(helper => helper.setFunc({f: expr}))
                .then(reply => {
                    this.paramOrder = reply.params;
                    this.series = reply.series;
                    const lastParams = this.params || {};
                    this.params = {};
                    reply.params.forEach((name, i) => {
                        let def = reply.defs[name];
                        if ( def === +def ) {
                            def = {val: def};
                        }
                        // apply defs unless param already existed in prev expr
                        this.params[name] = lastParams[name] || Object.assign(this.paramStore[name] || newParam(name, {}), def || {});
                        this.paramStore[name] = this.params[name];
                    });
                    this.config.expr = expr;
                    this.exprInfo = reply;
                    this.ev.expr.trigger(expr, reply);
                    return reply;
                });
        }
        setWeightExpr(expr) {
            // expr: just one function, no equals sign or hashes
            //       in terms of Data and column names
            //       or undefined/blank/falsy for equal 1.0 weighting
            return this.helper.then(helper => helper.setWeight({
                f: expr ? expr.split(';')[0] : '1.0'
            })).then(reply => {
                this.config.weightExpr = expr;
                this.ev.weightExpr.trigger(reply.f);
            });
        }
        setParam(name, p) {
            // updates params[name] with the defined fields of p
            // returns true if any fields were changed
            // e.g. false because !(param.lo < p.val < param.hi)
            const param = this.params[name],
                  didSet = param.set(p);
            didSet && this.ev.param.trigger(param);
            return didSet;
        }
        paramsFromArr(arr, detectOOB) {
            // returns an object with named parameters initialized from an array in optimizer order
            // or null if any parameters out of bounds
            const paramObj = {};
            let ip = 0,
                oob = false;
            this.paramOrder.forEach(name => {
                const p = this.params[name];
                if ( p.canFit ) {
                    const v = arr[ip];
                    if ( detectOOB &&
                         (((p.lo !== qubx.UNSET) && (v < p.lo)) ||
                          ((p.hi !== qubx.UNSET) && (v > p.hi))) ) {
                        oob = true;
                        return;
                    }
                    paramObj[name] = v;
                    ++ip;
                } else {
                    paramObj[name] = p.val;
                }
            });
            return oob ? null : paramObj;
        }
        arrFromParams(params) {
            // returns array of parameter values in optimizer order
            params = params || {};
            const arr = [];
            this.paramOrder.forEach(name => {
                const p = this.params[name];
                if ( p.canFit ) {
                    if ( typeof(params[name]) !== 'undefined' ) {
                        arr.push(params[name]);
                    } else {
                        arr.push(this.params[name].val);
                    }
                }
            });
            return arr;
        }
        setParamsArr(params) {
            let ip = 0;
            this.paramOrder.forEach(name => {
                const p = this.params[name];
                if ( p.canFit ) {
                    this.setParam(name, params[ip]);
                    ++ip;
                }
            });
        }
        fit(opt) {
            // opt: {
            //     maxIter: max iterations
            //     oniteration: function(fitter, params, score, r2, wwRuns, iteration),
            //     lmfit: true,
            //     simplex: false,
            //     dfp: false
            // }
            // resolves: {
            //     params: object with named final param values
            //     score: residual sum of squares
            //     iteration: number of iterations
            // }
            if ( this.fitting ) {
                throw new Error('already fitting');
            }
            this.fitting = true;
            this.stopped = false;
            this.stoppers = [];
            this.ev.startFit.trigger();
            const cleanup = () => {
                this.fitting = false;
                this.ev.endFit.trigger();
                this.stoppers.forEach(resolve => resolve(true));
                this.stoppers = [];
            };
            return this.helper.then(helper => {
                const fitParams = [];
                let r2 = 0.0,
                    wwRuns = 0.0;
                this.paramOrder.forEach(name => {
                    const p = this.params[name];
                    if ( p.canFit ) {
                        fitParams.push(p.val);
                    }
                });
                let result = Promise.resolve({
                    params: fitParams,
                    score: 0,
                    r2: r2,
                    wwRuns: wwRuns,
                    iteration: 0
                });
                const score = params => {
                    const paramObj = this.paramsFromArr(params, true);
                    return (paramObj ? helper.rss({params: paramObj}) : Promise.resolve(1e20))
                        .then(reply => {
                            r2 = reply.r2;
                            wwRuns = reply.wwRuns;
                            return reply.rss;
                        });
                };
                if ( opt.simplex ) {
                    result = result.then(last => new qubx_optimize.Simplex(last.params, {maxIter: opt.maxIter || 100}).run({
                        score: score,
                        iter: (params, score, iteration) => {
                            this.setParamsArr(params);
                            opt.oniteration && opt.oniteration('Simplex', params, score, r2, wwRuns, iteration);
                            return ! this.stopped;
                        }
                    }));
                }
                if ( opt.dfp ) {
                    result = result.then(last => {
                        if ( opt.simplex ) {
                            // perturb starting params for better error estimates
                            last.params.forEach((p, i) => {
                                last.params[i] *= 1 + .02*qubx.gaussRandom();
                            });
                        }
                        return new qubx_optimize.DFP(last.params, {maxIter: opt.maxIter || 100})
                            .run({
                                score: score,
                                grad: qubx_optimize.numGrad2Side(score, 1e-8),
                                iter: (params, score, iteration, grads) => {
                                    this.setParamsArr(params);
                                    opt.oniteration && opt.oniteration('DFP', params, score, r2, wwRuns, iteration, grads)
                                    return ! this.stopped;
                                }
                            }).then(result => {
                                return result;
                            });
                    });
                }

                if ( opt.lmfit || ! (opt.dfp || opt.simplex) ) {
                    let ndata = 0;
                    const ySeries = this.series.map(ser => {
                        const arr = this.data[ser.name || this.yName];
                        ndata += arr.length;
                        return arr;
                    });
                    
                    let arrReuse = undefined;
                    const lmEval = (params, fvec, iter) => {
                        const paramObj = this.paramsFromArr(params, true);
                        if ( ! paramObj ) {
                            return Promise.resolve(-1);
                        }
                        return this.getFit(arrReuse, paramObj).then(arrs => {
                            arrReuse = arrs;
                            let i = 0;
                            ySeries.forEach((dat, s) => {
                                const curve = arrs[s],
                                      n = dat.length;
                                let j = 0;
                                for (; j<n; ++j, ++i) {
                                    fvec[i] = dat[j] - curve[j];
                                }
                            });
                            /// could slow it down calculating stats each iter, or call oniter less frequently...
                            opt.oniteration && opt.oniteration('lmfit', paramObj, 0, 0, 0, iter);
                            return this.stopped ? -1 : 0;
                        });
                    };
                    const control = {
                        maxcall: opt.maxIter
                    };
                    result = result.then(last => {
                        const lmParams = last.params.slice();
                        return lmfit.minimizeSoon(lmParams, lmEval, ndata, control)
                            .then(info => qubx_lib.then(lib => Object.assign(info, {
                                params: lmParams,
                                iteration: control.iteration
                            })));
                    });
                }
                            
                return result.then(last => {
                    this.setParamsArr(last.params);
                    return score(last.params).then(rss => {
                        last.params = this.paramsFromArr(last.params);
                        last.score = last.rss = rss;
                        last.r2 = r2;
                        last.wwRuns = wwRuns;
                        cleanup();
                        return last;
                    });
                }).catch(err => {
                    cleanup();
                    return Promise.reject(err);
                });
            });
        }
        stop() {
            // stops active fitting, resolves true when no longer fitting
            // (of course, consider it might be fitting again by the time you get the resolution)
            if ( ! this.fitting ) {
                return Promise.resolve(true);
            }
            this.stopped = true;
            return new Promise((resolve, reject) => {
                this.stoppers.push(resolve);
            });
        }
        getRSS() {
            const params = {};
            this.exprInfo.params.forEach(name => {
                params[name] = this.params[name].val;
            });
            return this.helper.then(helper => helper.rss({params: params}));
        }
        getFit(arrs, paramArr) {
            // Retrieves the most recent fit curves from the evaluator (during/after fitting).
            //   arrs.length === setExpr()=>series.length
            //   arrs[i].length === Data.n === columns[xName].length
            // Resolves to arrs, which are constructed automatically if undefined
            arrs = arrs || [];
            const n = (arrs.length && arrs[0].length) || (this.data[this.xName] || this.data[this.yName]).length,
                  nser = this.series && this.series.length || 0;
            let i = 0;
            for (; i<nser; ++i) {
                if ( (! arrs[i]) || (arrs[i].length !== n) ) {
                    arrs[i] = new Float32Array(n);
                }
            }
            const msg = {arrs};
            if ( paramArr ) {
                msg.params = paramArr;
            }
            return this.helper.then(helper => helper.getFit(msg, arrs.map(arr => arr.buffer)))
                .then(reply => reply.arrs);
        }
        getErr(fitResult) {
            // Calculates error limits at the current parameters,
            // fills context.params[name].std with standard error of the estimate
            // resolves {
            //     stdErrEst: in context.paramOrder order
            //     correlation: where correlation[i][j] is the cross-correlation between params i and j in paramOrder
            // }
            fitResult = fitResult || {};
            const stdErrEst = {},
                  invHessian = /*fitResult.curvature ? Promise.resolve(fitResult.curvature) :*/ this.helper.then(helper => {
                      let params = fitResult.params;
                      if ( ! (params && params.length) ) {
                          params = this.arrFromParams();
                      }
                      const n = params.length;
                      const score = qubx.memoize1(params => helper.rss({params: this.paramsFromArr(params)}).then(reply => reply.rss), true),
                            steps = params.map(p => Math.abs(1e-7*p)),
                            step = (i, pars, sgn) => {
                                pars = pars.slice();
                                pars[i] += sgn*steps[i];
                                return pars;
                            },
                            d1 = (i, pars) => Promise.all([score(step(i, pars, 1)), score(step(i, pars, -1))]).then(yy => {
                                return (yy[0] - yy[1]) / (2*steps[i]);
                            }),
                            hess = params.map(() => params.map(() => 0));
                      return Promise.all(params.map((vi,i) => Promise.all(params.map((vj,j) => {
                          if ( i > j ) {
                              return Promise.resolve(true);
                          }
                          return Promise.all([d1(i, step(j, params, 1)), d1(i, step(j, params, -1))])
                              .then(yy => hess[i][j] = (yy[0] - yy[1]) / (2*steps[j]));
                      })))).then(() => {
                          let i = 0;
                          for (; i<n; ++i) {
                              let j = i + 1;
                              for (; j<n; ++j) {
                                  hess[i][j] = hess[j][i];
                              }
                          }
                          score(params); // restore central param vals
                          return qubx_lib.then(lib => lib.pinvm(hess));
                      });
                  });
            return invHessian.then(invH => {
                const n = invH.length;
                let ip = 0;
                this.paramOrder.forEach(name => {
                    const p = this.params[name];
                    if ( p.canFit ) {
                        const std = Math.sqrt(Math.abs(invH[ip][ip]));
                        p.std = std;
                        stdErrEst[name] = std;
                        ++ip;
                    }
                });
                const correl = invH.map((row, i) => row.map((cell, j) => {
                    const denom = Math.sqrt(Math.abs(invH[i][i] * invH[j][j]));
                    return Math.abs(denom ? (cell / denom) : cell);
                }));
                return {
                    stdErrEst: stdErrEst,
                    correlation: correl
                };
            });
        }
        getState() {
            const paramOut = {};
            for (const name of Object.keys(this.params)) {
                const p = this.params[name];
                paramOut[name] = {
                    val: p.val,
                    std: p.std
                };
            }
            return {
                config: {
                    expr: this.config.expr,
                    weightExpr: this.config.weightExpr
                },
                xName: this.xName,
                yName: this.yName,
                params: paramOut
            };
        }
        restoreState(state) {
            this.setXY(state.xName, state.yName).then(() => {
                this.setWeightExpr(state.config.weightExpr);
                this.setExpr(state.config.expr).then(() => {
                    for (const name of Object.keys(state.params)) {
                        this.setParam(name, state.params[name]);
                    }
                });
            });
        }
    };

    exports.test = function test() {
        function oniteration(fitter, params, score, r2, wwRuns, iteration, grads) {
            console.log('iter',fitter, params, score, iteration, grads);
        }
        const ctx = new exports.Context();
        Object.keys(ctx.ev).forEach(nm => ctx.ev[nm].sub((...args) => console.log('ev',nm,args)));
        return ctx.setColumns({
            A: new Float32Array([0, 1, 2, 3]),
            B: new Float32Array([0, 2, 4.1, 5.9]),
            C: new Float32Array([1, 1, 5, 1])
        }).then(() => ctx.setXY('A', 'B'))
            .then(() => ctx.setExpr('m*x + b; b:0; C=m*x+c; m: {val: 1, lo: -5, hi: 5, canFit: true}, C: {weight: .01}'))
            .then(() => {
                console.log('params', ctx.params);
                return ctx.fit({
                    oniteration: oniteration
                });
            }).then(result => {
                console.log('fitted',result);
                return ctx.getErr().then(info => {
                    console.log('getErr', info);
                    return result;
                });
            }).then(result => {
                ctx.getFit()
                    .then(arrs => console.log(arrs));
                ctx.setWeightExpr('C')
                    .catch(data => console.log('weight err',data.err.message));
                return ctx.fit({
                    simplex: true,
                    //dfp: true,
                    oniteration: oniteration
                });
            }).then(result => {
                console.log('fitted',result);
                ctx.getErr(result).then(info => console.log('getErr', info));
                return result;
            }).then(result => ctx.getFit().then(arrs => console.log(arrs)))
            .then(() => ctx.setExpr('m*x + b*D; b:0; m: {val: cos(Data.min("x")), lo: -5, hi: 5, canFit: true}; common: {G: -1, H: "m+b"}; common: {D: "C*b*qq*G*H"}'))
            .then(result => ctx.fit({
                oniteration: oniteration
            })).then(result => {
                console.log('fitted',result);
                ctx.getErr(result).then(info => console.log('getErr', info));
                return result;
            }).then(result => ctx.getFit().then(arrs => console.log(arrs)))
            .catch(err => {
                console.log('err',err);
                console.log(err.stack);
            });
    };
    
    return exports;
});
