/* Copyright 2016-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
    jsotan binary data format

    "JavaScript Object and Typed Array Notation"

Encodes and decodes json-like objects that have TypedArrays, as binary blobs, via two exported functions:

    obj = {any: {json_legal: {structure: {having: {typedArrays: [new Int32Array([1, 2, 3]), new Float64Array([0.0])]}}}}}
    blob = jsotan.blobify(obj);
    jsotan.parse(blob).then(obj => {...});

File layout:

    The file begins with a MAGIC string "jsotan\r\n" written as a pair of 32-bit unsigned ints in native byte order.
    We use this to detect the endianness of the file's creator.
    
    Next (byte offset 8) is the byte offset of the JSON text segment, as a 64-bit unsigned int in network byte order (big-endian).

    Next are array contents, stored raw in native byte order.

    The JSON text segment is generated by recursively replacing all typed arrays with records {
        t: data type, in ['i8', 'i16', 'i32', 'u8', 'u16', 'u32', 'f32', 'f64']
        l: length (number of elements)
        jsotanP: byte offset in file
    }, calling JSON.stringify, then appending to the blob constructor array, which encodes it as UTF-8.

    No provision is made for objects other than TypedArrays which cannot be stringified by JSON.
*/

define("jsotan", ["stringview"], StringView => {
    'use strict';
    
    const exports = {};

    const ARR_T = {
        'i8' : Int32Array,
        'i16' : Int16Array,
        'i32' : Int32Array,
        'u8' : Uint32Array,
        'u16' : Uint16Array,
        'u32' : Uint32Array,
        'f32' : Float32Array,
        'f64' : Float64Array
    };
    const ELEMENTSIZE = {
        'i8' : 1,
        'i16' : 2,
        'i32' : 4,
        'u8' : 1,
        'u16' : 2,
        'u32' : 4,
        'f32' : 4,
        'f64' : 8
    };
    const VIEWMETHOD = {
        'i8' : 'getInt8',
        'i16' : 'getInt16',
        'i32' : 'getInt32',
        'u8' : 'getUint8',
        'u16' : 'getUint16',
        'u32' : 'getUint32',
        'f32' : 'getFloat32',
        'f64' : 'getFloat64'
    };

    const MAGIC = [0x6a736f74, 0x616e0d0a],
          MAGIC_LE = [0x746f736a, 0x0a0d6e61];
    
    const bigEndCPU = (() => {
        const arr = new Uint32Array([1]),
              view = new DataView(arr.buffer);
        return !! view[3];
    })();

    // BufBuilder class keeps track of the total bytes added (end)
    class BufBuilder {
        constructor() {
            this.end = 0;
            this.sources = []
        }
        add(buf) {
            this.end += buf.byteLength|0;
            this.sources.push(buf);
            return this;
        }
    }
    
    function stripArrays(obj, bufs) {
        // returns a deep copy of obj; each TypedArray instance is replaced with {t,l,jsotanP},
        // adding arr.buffer to bufs
        let arrayRec = null;
        if ( typeof(obj) === 'object' ) {
            if ( Array.isArray(obj) ) {
                return obj.map(sub => stripArrays(sub, bufs));
            }
            if ( obj instanceof Int8Array ) {
                arrayRec = {t: 'i8'};
            } else if ( obj instanceof Int16Array ) {
                arrayRec = {t: 'i16'};
            } else if ( obj instanceof Int32Array ) {
                arrayRec = {t: 'i32'};
            } else if ( obj instanceof Uint8Array ) {
                arrayRec = {t: 'u8'};
            } else if ( obj instanceof Uint16Array ) {
                arrayRec = {t: 'u16'};
            } else if ( obj instanceof Uint32Array ) {
                arrayRec = {t: 'u32'};
            } else if ( obj instanceof Float32Array ) {
                arrayRec = {t: 'f32'};
            } else if ( obj instanceof Float64Array ) {
                arrayRec = {t: 'f64'};
            }
            if ( arrayRec ) {
                arrayRec.l = obj.length;
                arrayRec.jsotanP = bufs.end;
                bufs.add(obj);
                return arrayRec;
            } else if ( obj ) {
                const keys = Object.keys(obj),
                      stripped = {};
                keys.forEach(k => {
                    stripped[k] = stripArrays(obj[k], bufs);
                });
                return stripped;
            }
        }
        return obj;
    }

    function reconstituteArrays(obj, arrbuf, bigEndFile) {
        // returns a deep copy of obj; each {t,l,jsotanP} is replaced with a TypedArray read from arrbuf.
        if ( typeof(obj) === 'object' ) {
            if ( Array.isArray(obj) ) {
                return obj.map(sub => reconstituteArrays(sub, arrbuf, bigEndFile));
            } else if ( obj.jsotanP ) {
                return reconstituteArray(obj, arrbuf, bigEndFile);
            }
            const keys = Object.keys(obj),
                  rtn = {};
            keys.forEach(k => {
                rtn[k] = reconstituteArrays(obj[k], arrbuf, bigEndFile);
            });
            return rtn;
        }
        return obj;
    }

    function reconstituteArray(rec, arrbuf, bigEndFile) {
        // returns TypedArray specified by rec={t,l,jsotanP}, read from arrbuf.
        // if file and cpu endianness don't match, uses a DataView to convert each element.
        bigEndFile = !! bigEndFile;
        const ArrayType = ARR_T[rec.t],
              arr = new ArrayType(rec.l);
        if ( bigEndFile === bigEndCPU ) {
            const srcArray = new ArrayType(arrbuf, rec.jsotanP, rec.l);
            arr.set(srcArray);
        } else {
            const view = new DataView(arrbuf),
                  littleEndian = ! bigEndFile,
                  getter = VIEWMETHOD[rec.t],
                  sz = ELEMENTSIZE[rec.t];
            let at = rec.jsotanP,
                i = 0;
            for (; i<rec.l; ++i, at += sz) {
                arr[i] = view[getter](at, littleEndian);
            }
        }
        return arr;
    }
                
    exports.blobify = function blobify(stuff) {
        // converts an object into a blob in jsotan format
        const arrbuf = new ArrayBuffer(16),
              view = new DataView(arrbuf),
              bufs = new BufBuilder().add(arrbuf),
              nativeEndian = new Uint32Array(arrbuf);
        nativeEndian[0] = MAGIC[0];
        nativeEndian[1] = MAGIC[1];
        const stripped = stripArrays(stuff, bufs);
        view.setUint32(8, 0);
        view.setUint32(12, bufs.end);
        bufs.add(JSON.stringify(stripped));
        return new Blob(bufs.sources, {
            type: 'application/octet-stream'
        });
    };

    exports.parse = function parse(blob) {
        // reads a blob in jsotan format, returns a promise which resolves to
        // a regular js object.
        return new Promise((resolve, reject) => {
            const rdr = new FileReader();
            rdr.addEventListener('load', e => resolve(e.target.result));
            rdr.addEventListener('error', err => reject(new Error('Failed to read blob ('+err+')')));
            rdr.readAsArrayBuffer(blob);
        }).then(arrbuf => {
            if ( arrbuf.byteLength < 16 ) {
                return {};
            }
            const view = new DataView(arrbuf);
            let bigEndFile = true;
            if ( (view.getUint32(0) !== MAGIC[0]) || (view.getUint32(4) !== MAGIC[1]) ) {
                if ( (view.getUint32(0) !== MAGIC_LE[0]) || (view.getUint32(4) !== MAGIC_LE[1]) ) {
                    throw new Error('Bad magic');
                } else {
                    bigEndFile = false;
                }
            }
            if ( view.getUint32(8) ) {
                throw new Error('Unsupported 64-bit offset');
            }
            const jsonOffset = view.getUint32(12),
                  info = JSON.parse(new StringView(arrbuf, "UTF-8", jsonOffset, arrbuf.byteLength - jsonOffset).toString()),
                  root = reconstituteArrays(info, arrbuf, bigEndFile);

            console.log(root);
            return root;
        });
    };
    
    exports.test = function test() {
        exports.parse(exports.blobify({
            testing: [1,2,3],
            ints: new Int32Array([4,5,6]),
            floats: new Float32Array([7,8,9]),
            nested: [{
                bytes: new Uint8Array([0,255,0])
            }]
        }));
    };

    return exports;
});
