/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see   */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx", ["random", "stringview"], (Random, StringView) => {
    'use strict';
    const exports = {};

    exports.int32ToFloat = a => {
        const buf = new ArrayBuffer(4),
              v = new DataView(buf);
        v.setInt32(0, a);
        return v.getFloat32(0);
    };
    exports.UNSET = exports.int32ToFloat(0x2aaaaaaa);
    //console.log(exports.UNSET, exports.UNSET.toString(2));
    
    exports.extend = Object.assign;

    exports.range = (from, to, step) => {
        if ( typeof(to) === 'undefined' ) {
            return exports.range(0, from);
        }
        if ( typeof(from) === 'undefined' ) {
            throw new Error('range needs endpoint');
        }
        from = +from;
        to = +to;
        step = (+step) || 1;
        const result = [];
        let i = from;
        for (; i<to; i+=step) {
            result.push(i);
        }
        return result;
    };
    
    exports.stopAndPrevent = event => event.stopPropagation() || event.preventDefault();
    
    exports.acceptStr = x => ''+x;
    exports.acceptStrNonempty = x => {
        if ( x ) {
            return ''+x;
        }
        throw new TypeError("Can't be blank");
    };
    
    exports.acceptInt = x => {
        const i = parseInt(x, 10);
        if ( isNaN(i) || (x !== (''+i)) ) {
            throw new TypeError('Expected an integer, got '+x);
        }
        return i;
    };
    exports.acceptIntAtLeast = lb => x => {
        const i = exports.acceptInt(x);
        if ( i < lb ) {
            throw new TypeError('Expected '+lb+' or greater, got '+x);
        }
        return i;
    };
    exports.acceptIntBetween = (lb, ub) => x => {
        const i = exports.acceptInt(x);
        if ( (i < lb)  || (i > ub)) {
            throw new TypeError('Expected '+lb+' <= x <= '+ub+', got '+x);
        }
        return i;
    };
    exports.acceptFloat = x => {
        const f = parseFloat(x);
        if ( isNaN(f) ) {
            throw new TypeError('Expected a number, got '+x);
        }
        return f;
    };
    exports.acceptFloatAtLeast = lb => x => {
        const f = exports.acceptFloat(x);
        if ( f < lb ) {
            throw new TypeError('Expected a number '+lb+' or greater, got '+f);
        }
        return f;
    };
    exports.acceptFloatGreaterThan = lb => x => {
        const f = exports.acceptFloat(x);
        if ( f <= lb ) {
            throw new TypeError('Expected a number greater than '+lb+', got '+f);
        }
        return f;
    };
    exports.acceptFloatBetween = (lb, ub) => x => {
        const f = exports.acceptFloat(x);
        if ( (f < lb) || (f > ub) ) {
            throw new TypeError('Expected a number between '+lb+' and '+ub+', got '+f);
        }
        return f;
    };
    exports.acceptFloatOrUnset = x => x ? exports.acceptFloat(x) : exports.UNSET;
    exports.acceptList = acceptItem => {
        acceptItem = acceptItem || exports.acceptStr;
        const splitter = /[, ]+/;
        return s => s.split(splitter).map(acceptItem);
    };
    exports.formatList = formatItem => {
        formatItem = formatItem || exports.formatStr;
        return l => l.map(formatItem).join(", ");
    };
    exports.acceptNone = x => {
        throw new TypeError('not editable');
    };
    exports.formatStr = x => (x===0) ? '0' : ("" + (x||''));
    exports.formatOrUnset = format => x => (x === exports.UNSET) ? '' : format(x);
    exports.formatK0 = x => {
        try {
            if ( x === 0 ) {
                return '0';
            }
            const m = x*x;
            if ( m < 1e-3 || m > 1e10 ) {
                return x.toExponential(3);
            }
            const s = x.toFixed(3);
            let i = s.length-1;
            while ( i ) {
                if ( s[i] === '.' ) {
                    return s.slice(0, i);
                } else if ( (i > 4) || (s[i] === '0') ) {
                    --i;
                } else {
                    break;
                }
            }
            return (s.indexOf('.') != -1) ? s.slice(0, i+1) : s;
        } catch (err) {
            exports.reportErr(err);
            return ''+x;
        }
    };
    exports.formatK0u = x => exports.formatK0(x).replace('-', '\u2212');
    exports.formatK1u = (k1, V, vlabel) => V ? (vlabel+exports.formatK0u(k1)) : '';
    exports.formatK1 = (k1, k2, L, V, P, unicode_minus) => {
        const parts = [],
              format = unicode_minus ? exports.formatK0u : exports.formatK0;
        L && parts.push("L");
        V && parts.push("V "+format(k1));
        P && parts.push("P "+format(k2));
        return parts.join(" ");
    };
    exports.formatState = row => row.Label || (""+row.Index);

    function ldZero2Dig(s) { return ("0" + s).slice(-2); }
    exports.formatDate = d => {
        const now = new Date(),
              nY = now.getFullYear(),
              dY = d.getFullYear(),
              nM = now.getMonth(),
              dM = d.getMonth(),
              nT = now.getDate(),
              dT = d.getDate(),
              dD = d.getDay(),
              dH = d.getHours(),
              dN = d.getMinutes(),
              time = ldZero2Dig(dH)+':'+ldZero2Dig(dN);
        if ( (nY === dY) && (nM === dM) && (nT === dT) ) {
            return time;
        }
        if ( (now - d) < (1000*60*60*24*6) ) {
            return ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'][dD] + ' ' + time;
        }
        return ''+dT+' '+['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][dM]+' '+dY+' '+time;
    };
    
    exports.pformat = (x, indent, notTop) => {
        let s = "";
        indent = indent || "<br>\n";
        const nextIndent = indent + "&nbsp;&nbsp;";
        if ( Array.isArray(x) ) {
            s += "[";
            if ( x.length && Array.isArray(x[0]) ) {
                let i = 0;
                for (; i<x.length; ++i) {
                    if ( i || notTop ) {
                        s += nextIndent;
                    }
                    s += exports.pformat(x[i], nextIndent, true);
                }
            } else {
                let i = 0;
                for (; i<x.length; ++i) {
                    if ( i ) {
                        s += ", ";
                    }
                    s += exports.pformat(x[i], nextIndent, true);
                }
            }
            s += "]";
        } else if ( typeof(x) === 'object' ) {
            if ( x === null ) {
                return "null";
            }
            for (const p of Object.keys(x)) {
                if ( x.hasOwnProperty(p) ) {
                    s += indent+p+": "+exports.pformat(x[p], nextIndent, true);
                }
            }
        } else if ( ! isNaN(x) ) {
            return exports.formatK0(x);
        } else {
            return ""+x;
        }
        return s;
    };

    exports.arrayMin = (arr, start, end) => {
        start = start || 0;
        end = end || arr.length;
        let min = Infinity;
        while ( (end--) >= start ) {
            if (arr[end] < min) {
                min = arr[end];
            }
        }
        return min;
    };
    exports.arrayMax = (arr, start, end) => {
        start = start || 0;
        end = end || arr.length;
        let max = -Infinity;
        while ( (end--) >= start ) {
            if (arr[end] > max) {
                max = arr[end];
            }
        }
        return max;
    };

    exports.extrema = (buf, start, count) => {
        start = start || 0;
        count = count || buf.length - start;
        if ( ! count ) {
            return {lo: 0, hi: 0, iLo: -1, iHi: -1};
        }
        let low = buf[start],
            high = low,
            iLo = -1,
            iHi = -1,
            i = 1;
        ++start;
        for (; i<count; ++i, ++start) {
            const y = buf[start];
            if ( y < low ) {
                low = y;
                iLo = start;
            }
            if ( y > high ) {
                high = y;
                iHi = start;
            }
        }
        return {lo: low, hi: high, iLo, iHi};
    };

    exports.extremas = (bufs, start, count) => {
        const out = {
            lo: 1e100,
            hi: -1e100,
            iLo: -1,// i: sample index
            iHi: -1,
            jLo: -1,// j: bufs index
            jHi: -1
        };
        bufs.forEach((buf, j) => {
            const ex = exports.extrema(buf, start, Math.min(buf.length-start, count));
            if ( out.lo > ex.lo ) {
                out.lo = ex.lo;
                out.iLo = ex.iLo;
                out.jLo = j;
            }
            if ( out.hi < ex.hi ) {
                out.hi = ex.hi;
                out.iHi = ex.iHi;
                out.jHi = j;
            }
        });
        return out;
    };

    exports.meanstd = (bufs, start, count) => {
        if ( bufs.length && (typeof(bufs[0].length) === 'undefined') ) {
            bufs = [bufs];
        }
        start = start|0;
        let mean = 0.0,
            rss = 0.0,
            n = 0;
        var last_mean, dm, d, x;
        bufs.forEach(buf => {
            const ct = (typeof(count) !== 'undefined') ? (count|0) : (buf.length - start);
            if ( ct <= 0 ) {
                return;
            }
            let i = 0;
            if ( ! n ) {
                mean = buf[start];
                n = 1;
                ++i;
            }
            for (; i<ct; ++i) {
                x = buf[start+i];
                last_mean = mean;
                ++n;
                mean = (1.0/n)*((n-1)*last_mean + x);
                dm = mean - last_mean;
                d = x - last_mean;
                rss += d*d - n*dm*dm;
            }
        });
        return {
            mean, rss, n,
            std: n ? Math.sqrt(rss/n) : 0,
            sampleStd: (n>=2) ? Math.sqrt(rss/(n-1)) : 0
        };
    };

    exports.memoize1 = (f, stringKeys) => {
        const cache = new Map(),
              lookup = x => {
            const k = stringKeys ? (''+x) : x;
            let y = cache.get(k);
            if ( typeof(y) === 'undefined' ) {
                y = f(x);
                cache.set(k, y);
            }
            return y;
        };
        lookup.clear = cache.clear.bind(cache);
        return lookup;
    };

    exports.i = d => d;
    exports.di = (d,i) => i;
    exports.m = exports.memoize1(s => d => d[s]);

    exports.get = (url, responseType, onProgress) => new Promise((resolve, reject) => {
        const req = new XMLHttpRequest();
        req.open('GET', url);
        if ( typeof(responseType) !== 'undefined' ) {
            req.responseType = responseType;
        }
        req.onload = () => {
            if (req.status == 200) {
                resolve(req.response);
            } else {
                reject(Error(req.statusText));
            }
        };
        if ( onProgress ) {
            req.onprogress = evt => {
                if (evt.lengthComputable) {
                    if ( onProgress(evt.loaded / evt.total) ) {
                        req.onprogress = null;
                        req.abort();
                        reject(new Error("Canceled"));
                    }
                }
            };
        }
        req.onerror = err => reject(new Error("Network Error"));
        req.send();
    });
    exports.postJson = (url, data, rawReply) => new Promise((resolve, reject) => {
        const req = new XMLHttpRequest();
        req.open('POST', url);
        req.setRequestHeader('Content-Type', 'application/json');
        req.onload = () => {
            if ( req.status === 200 ) {
                if ( rawReply ) {
                    resolve(req.responseText);
                } else {
                    resolve(JSON.parse(req.responseText));
                }
            } else {
                reject(Error(req.statusText));
            }
        };
        req.onerror = err => reject(Error('Network error'));
        req.send(JSON.stringify(data));
    });

    exports.ws = (url, config) => {
        config = config || {};
        config.binaryType = config.binaryType || 'arraybuffer';
        // onmessage, onerror, onclose, onopen, onreopen
        // immediate: open/reopen before any messages sent
        const tix = new exports.PromiseTickets();
        let ws = null,
            wsOpen = false,
            pending = [],
            firstTime = true;
        const send = msg => {
            if ( wsOpen ) {
                ws.send(msg);
            } else {
                pending.push(msg);
                ws || connect();
            }
        };
        const request = msg => new Promise((resolve, reject) => {
            msg.tkt = tix.register(resolve, reject);
            try {
                send(JSON.stringify(msg));
            } catch(err) {
                tix.reject(msg.tkt, err);
            }
        });
        const connect = () => {
            ws = new WebSocket(url);
            ws.binaryType = config.binaryType;
            ws.addEventListener('error', event => {
                if ( config.onerror ) {
                    config.onerror(event);
                } else {
                    console.log('ws error',event,url);
                }
                ws = null;
                wsOpen = false;
                tix.rejectAll('Socket closed (error)');
                pending = [];
                config.immediate && setTimeout(connect, 5000);
            });
            ws.addEventListener('message', event => {
                if ( event.data.byteLength || event.data.size ) { // binary
                    config.onmessage && config.onmessage(event);
                    return;
                }
                let msg = null;
                if ( event.data ) {
                    try {
                        msg = JSON.parse(event.data);
                        if ( msg.tkt ) {
                            if ( msg.tktErr ) {
                                tix.reject(msg.tkt, msg.tktErr);
                            } else {
                                tix.resolve(msg.tkt, msg);
                            }
                            return;
                        }
                    } catch (err) {
                        console.log(err);
                    }
                }
                config.onmessage && config.onmessage(event, msg || event.data);
            });
            ws.addEventListener('close', event => {
                ws = null;
                wsOpen = false;
                tix.rejectAll('Socket closed');
                pending = [];
                config.onclose && config.onclose(event);
                config.immediate && connect();
            });
            ws.addEventListener('open', event => {
                const cb = (firstTime ? config.onopen : config.onreopen);
                firstTime = false;
                wsOpen = true;
                cb && cb();
                
                let sent = 0;
                if ( ws.readyState === 1 ) {
                    try {
                        for (const pend of pending) {
                            ws.send(pend);
                            ++sent;
                        }
                    } catch (err) {
                        exports.errOut.send(err, url);
                    }
                }
                if ( sent < pending.length ) {
                    console.log('failing send on open; dumping '+(pending.length-sent)+' messages');
                }
                pending = [];
            });
        };
        const close = () => {
            config.immediate = false;
            try {
                ws && ws.close();
            } catch (err)
            {}
        }
        config.immediate && connect();
        return {wsOpen, send, request, close};
    };

    exports.wsAuth = (url, config) => {
        config = config || {};
        const deauth = () => {
            haveAuth = false;
            config.ondeauth && config.ondeauth();
        }
        const ws = exports.ws(url, Object.assign({}, config, {
            onclose: event => {
                deauth();
                config.onclose && config.onclose();
            }
        }));
        let haveAuth = false;
        return {
            send: (msg, noAuth) => (haveAuth || noAuth) ?
                ws.send(msg) :
                config.requestAuth(ws)
                    .then(result => {
                        haveAuth = true;
                        config.onauth && config.onauth(result);
                        return ws.send(msg);
                    }),
            request: (msg, noAuth) => (haveAuth || noAuth) ?
                ws.request(msg) :
                config.requestAuth(ws)
                    .then(result => {
                        haveAuth = true;
                        config.onauth && config.onauth(result);
                        return ws.request(msg);
                    }),
            close: () => {
                haveAuth = false;
                config.ondeauth && config.ondeauth();
                ws.close();
            },
            deauth // call when you have sent a sign-out message
        };
    };

    exports.CircularStock = class CircularStock {
        constructor(capacity) {
            this._next = 0;
            this.size = 0;
            this.flush({capacity: capacity||100});
        }
        firstKey() {
            return (this._next - this.size);
        }
        hasKey(i) {
            return ( (this.firstKey() <= i) && (i < this._next) );
        }
        get(i) {
            return this._stock[i%this._cap];
        }
        push(x) {
            if ( this.size === this._cap ) {
                this.resize(2*this._cap);
            }
            const i = this._next;
            ++this._next;
            ++this.size;
            this._stock[i%this._cap] = x;
            return i;
        }
        pop() {
            if ( this.size === 0 ) {
                throw new Error('Empty');
            }
            const i = this._next - this.size;
            --this.size;
            return this._stock[i%this._cap];
        }
        resize(cap) {
            const stock = [];
            let i = this._next-Math.min(cap, this.size);
            for (; i<this._next; ++i) {
                stock[i%cap] = this._stock[i%this._cap];
            }
            this._cap = cap;
            this._stock = stock;
        }
        flush(options) {
            if ( options.destroy ) {
                let i = this._next-this.size;
                for (; i<this._next; ++i) {
                    options.destroy(this._stock[i%this._cap]);
                }
            }
            if ( typeof(options.capacity) !== 'undefined' ) {
                this._cap = options.capacity|0;
            }
            this._stock = [];
            this.size = 0;
        }
    };
    
    exports.PromisedStock = class PromisedStock {
        constructor(construct, instances, release) {
            this._construct = construct;
            this._release = release;
            this._n = 0;
            this._stock = [];
            this.items = [];
            this._waiting = new exports.CircularStock();
            this.instances(instances|0);
        }
        get(timeoutMS) {
            if ( this._stock.length ) {
                return Promise.resolve(this._stock.pop());
            } else if ( timeoutMS === 0 ) {
                return Promise.reject('Empty');
            }
            return new Promise((resolve, reject) => {
                const num = this._waiting.push({resolve, reject});
                if ( (typeof(timeoutMS) !== 'undefined') && (timeoutMS > 0) ) {
                    setTimeout(() => {
                        if ( this._waiting.hasKey(num) )  {
                            this._waiting.get(num).timedOut = true;
                            reject('Timed out');
                            this._clearTimeouts();
                        }
                    }, timeoutMS);
                }
            });
        }
        put(x) {
            this._clearTimeouts();
            if ( this._waiting.size ) {
                this._waiting.pop().resolve(x);
            } else if ( this._stock.length < this._n ) {
                this._stock.push(x);
            } else if ( this._release ) {
                this._release(x);
            }
            return this;
        }
        reportLost() {
            this.put(this._construct());
        }
        _clearTimeouts() {
            while ( this._waiting.size && this._waiting.get(this._waiting.firstKey()).timedOut ) {
                this._waiting.pop();
            }
        }
        instances(n) {
            if ( typeof(n) === 'undefined' ) {
                return this._n;
            }
            n = n|0;
            const nOut = this._n - this._stock.length;
            while ( this._stock.length > (n-nOut) ) {
                let extra = this._stock.pop();
                this.items.splice(this.items.indexOf(extra));
                this._release && this._release(extra);
            }
            this._n = n;
            while ( this._stock.length < (n-nOut) ) {
                let item = this._construct();
                this.items.push(item);
                this.put(item);
            }
            return this;
        }
        flush() {
            this._waiting.flush({
                destroy: w => w.reject('Timed out (flush)')
            });
            return this;
        }
    };
    
    exports.PromisedBuffers = class PromisedBuffers {
        constructor(count, len, byteWidth, instances) {
            count = count|0;
            len = len|0;
            byteWidth = (byteWidth|0) || 4;
            instances = (instances|0) || 2;
            this.stock = new exports.PromisedStock(() => {
                const buffers = [];
                let i = 0;
                for (; i<count; ++i) {
                    buffers.push(new ArrayBuffer(byteWidth*len));
                }
                return buffers;
            }, instances);
            this.bufferCount = count;
            this.bufferLength = len;
            this.byteWidth = byteWidth;
            this.get = this.stock.get.bind(this.stock);
        }
        put(buffers) {
            if ( (buffers.length != this.bufferCount) || (buffers.length && (buffers[0].byteLength != (this.bufferLength*this.byteWidth))) ) {
                return this; // wrong dimensions--not mine
            }
            return this.stock.put(buffers);
        }
        flush() {
            this.stock.flush();
            return this;
        }
    };

    exports.PromiseTickets = class PromiseTickets {
        constructor() {
            let nextNum = 1;
            const makeTicket = () => {
                const t = nextNum;
                ++nextNum;
                return t;
            };
            this.reuse = [];
            this.getTicket = () => this.reuse.length ? this.reuse.pop() : makeTicket();
            this.registry = [];
            // ticket is index into registry array; reusing tickets for compact storage
        }
        register(resolve, reject, onprogress) {
            const t = this.getTicket();
            this.registry[t] = {resolve, reject, onprogress};
            return t;
        }
        retrieve(ticket) {
            const promise = this.registry[ticket];
            this.registry[ticket] = null;
            this.reuse.push(ticket);
            return promise;
        }
        resolve(ticket, arg) {
            this.retrieve(ticket).resolve(arg);
        }
        reject(ticket, err) {
            this.retrieve(ticket).reject(err);
        }
        rejectAll(err) {
            this.registry.forEach((promise,ticket) => promise && this.reject(ticket, err));
        }
        progress(ticket, progress) {
            const promise = this.registry[ticket];
            promise.onprogress && promise.onprogress(progress);
        }
    };
            
    /*
    exports.testStock = function() {
        var ib = 0;
        var ps = new exports.PromisedStock(function() { ++ib; return [ib]; }, 2);
        for (var ii=0; ii<10; ++ii) {
            console.log('get');
            ps.get().then(function(box) {
                console.log('got',box);
                ps.put(box);
            });
        }
    };
    */
    const field_spec_default = {
        name: "",
        defval: 0.0,
        accept: exports.acceptFloat,
        format: x => x.toPrecision(4),
        choices: [],
        widthChars: 0
    };

    exports.Event = class Event {
        constructor(label) {
            this.label = label;
            this.subs = [];
            Object.freeze(this);
        }
        sub(f) {
            if ( ! f ) {
                throw new Error('null event callback');
            }
            this.subs.push(f);
            return f;
        }
        unsub(f) {
            const i = this.subs.indexOf(f);
            if ( i >= 0 ) {
                this.subs.splice(i, 1);
            } else {
                console.log('not subscribed',f);
            }
        }
        trigger() {
            for (const s of this.subs) {
                try {
                    s.apply(null, arguments);
                } catch (err) {
                    console.log('in event '+this.label+':', err);
                    exports.errOut.send(err, this.label||'Event');
                }
            }
        }
        subscribed() {
            return !! this.subs.length;
        }
    };
    exports.makeEvents = names => {
        const ev = {};
        names.forEach(nm => {
            ev[nm] = new exports.Event();
        });
        Object.freeze(ev);
        return ev;
    };
    
    exports.Table = class Table {
        constructor(config) {
            const defaults = {label: ""};
            this.config = Object.assign({}, defaults, config);
            this.label = this.config.label;
            this.fields = [{name: 'Index', defval: 0, accept: exports.acceptNone, format: exports.formatStr}];
            this.fieldIndex = {'Index' : 0};
            this.entries = [];
            this.row_default = {};
            this.ev = exports.makeEvents(['addField', 'inserting', 'inserted', 'removing', 'removed', 'set']);
            Object.freeze(this);
        }
        [Symbol.iterator]() {
            return this.entries[Symbol.iterator]();
        }
        get size() {
            return this.entries.length;
        }
        addField(spec) {
            const f = Object.assign({}, field_spec_default, spec);
            this.fieldIndex[f.name] = this.fields.length;
            this.fields.push(f);
            for (const entry of this.entries) {
                if ( typeof(entry[f.name]) === 'undefined' ) {
                    entry[f.name] = f.defval;
                }
            }
            this.row_default[f.name] = f.defval;
            this.ev.addField.trigger(f);
            return this;
        }
        clear() {
            let i = this.size - 1;
            for (; i>=0; --i) {
                this.remove(i);
            }
        }
        append(row) {
            this.insert(this.entries.length, row);
            return this;
        }
        insert(ix, row, undoing) {
            const r = Object.assign({}, this.row_default, row);
            r.Index = ix;
            this.ev.inserting.trigger(r, undoing);
            let i = this.entries.length-1;
            for (; i>=ix; --i) {
                const entry = this.entries[i];
                entry.Index = i+1;
                this.entries[i+1] = entry;
            }
            this.entries[ix] = r;
            this.ev.inserted.trigger(r, undoing);
            return this;
        }
        remove(ix, undoing) {
            const row = this.entries[ix];
            this.ev.removing.trigger(row, undoing);
            let i = ix + 1;
            const n = this.entries.length;
            for (; i<n; ++i) {
                const entry = this.entries[i];
                entry.Index = i-1;
                this.entries[i-1] = entry;
            }
            this.entries.splice(this.size-1, 1);
            this.ev.removed.trigger(row, undoing);
            return this;
        }
        get(r, c) {
            if ( typeof(c) === 'undefined' ) {
                return this.entries[r];
            }
            return this.entries[r][c];
        }
        set(r, c, v, undoing) {
            if ( (c in this.fieldIndex) && (r < this.entries.length) ) {
                const entry = this.entries[r];
                if ( entry[c] !== v ) {
                    const info = {row: r, column: c, value: v, prev: entry[c]};
                    entry[c] = v;
                    this.ev.set.trigger(info, undoing);
                }
            }
            return this;
        }
        getByName(name, valueField, nameField) {
            valueField = valueField || 'Value';
            nameField = nameField || 'Name';
            for (const x of this.entries) {
                if ( x[nameField] === name ) {
                    return x[valueField];
                }
            }
        }
        setByName(name, v, valueField, nameField) {
            valueField = valueField || 'Value';
            nameField = nameField || 'Name';
            const n = this.size;
            let i = 0;
            for (; i<n; ++i) {
                if ( this.get(i, nameField) === name ) {
                    this.set(i, valueField, v);
                    return;
                }
            }
            const newRow = [];
            newRow[nameField] = name;
            newRow[valueField] = v;
            this.append(newRow);
        }
        toJson() {
            const entries = [];
            for (const x of this.entries) {
                entries.push(Object.assign({}, x));
            }
            return entries;
        }
        fromJson(entries) {
            this.clear();
            for (const x of entries) {
                this.append(x);
            }
            return this;
        }
    };
    
    exports.Undo = class Undo {
        constructor(config) {
            this.working = false;
            this.enabled = true;
            this.depth = this.depth0 = (config && config.maxDepth) || null;
            this.uu = [];
            this.rr = [];
            this.ev = exports.makeEvents(['change']);
        }
        clear() {
            this.depth = this.depth0;
            this.uu = [];
            this.rr = [];
            this.ev.change.trigger();
        }
        enable() {
            this.enabled = true;
        }
        disable() {
            this.enabled = false;
        }
        canUndo() {
            return (this.uu.length > 0);
        }
        undoLabel() {
            let i = this.uu.length - 1;
            for (; i>=0; --i) {
                if ( typeof(this.uu[i]) === 'string' ) {
                    return this.uu[i];
                }
            }
            return "";
        }
        canRedo() {
            return (this.rr.length > 0);
        }
        redoLabel() {
            let i = this.rr.length - 1;
            for (; i>=0; --i) {
                if ( typeof(this.rr[i]) === 'string' ) {
                    return this.rr[i];
                }
            }
            return "";
        }
        pushUndo(undo, redo) {
            if ( this.working || ! this.enabled ) {
                return;
            }
            this.uu.push( {undo: undo, redo: redo} );
        }
        sealUndo(label) {
            if ( this.working || (this.uu.length === 0) || (typeof(this.uu[this.uu.length-1]) === 'string') || ! this.enabled ) {
                return;
            }
            this.uu.push(label||"");
            this.rr = [];
            if ( this.depth !== null ) {
                if ( this.depth ) {
                    this.depth -= 1;
                } else { // forget the oldest
                    const n = this.uu.length;
                    let i = 0;
                    for (; i<n; ++i) {
                        if ( typeof(this.uu[i]) === 'string' ) {
                            this.uu.splice(0, i+1);
                            break;
                        }
                    }
                }
            }
            this.ev.change.trigger();
        }
        undo() {
            var action;
            if ( ! this.uu.length ) {
                return;
            }
            this.working = true;
            // leftover/unsealed:
            while ( this.uu.length && (typeof(this.uu[this.uu.length-1]) !== 'string') ) {
                action = this.uu.pop();
                action.undo();
                this.rr.push(action);
            }
            let lbl = "";
            if ( this.uu.length ) {
                lbl = this.uu.pop();
            }
            while ( this.uu.length && (typeof(this.uu[this.uu.length-1]) !== 'string') ) {
                action = this.uu.pop();
                action.undo();
                this.rr.push(action);
            }
            this.rr.push(lbl);
            this.working = false;
            if ( this.depth !== null ) {
                this.depth += 1;
            }
            this.ev.change.trigger();
        }
        redo() {
            var action;
            if ( ! this.rr.length ) {
                return;
            }
            this.working = true;
            const lbl = this.rr.pop();
            while ( this.rr.length && (typeof(this.rr[this.rr.length-1]) !== 'string') ) {
                action = this.rr.pop();
                action.redo();
                this.uu.push(action);
            }
            this.uu.push(lbl);
            this.working = false;
            if (this.depth !== null ) {
                this.depth -= 1;
            }
            this.ev.change.trigger();
        }
    };

    exports.TableUndo = class TableUndo {
        constructor(table, undo) {
            function onInserted(row, undoing) {
                if ( undoing ) {
                    return;
                }
                row = Object.assign({}, row);
                undo.pushUndo(() => {table.remove(row.Index, true);},
                              () => {table.insert(row.Index, row, true);});
            }
            function onRemoving(row, undoing) {
                if ( undoing ) {
                    return;
                }
                row = Object.assign({}, row);
                undo.pushUndo(() => {table.insert(row.Index, row, true);},
                              () => {table.remove(row.Index, true);});
            }
            function onSet(info, undoing) {
                if ( undoing ) {
                    return;
                }
                undo.pushUndo(() => {table.set(info.row, info.column, info.prev, true);},
                              () => {table.set(info.row, info.column, info.value, true);});
            }
            this.table = table;
            this.undo = undo;
            table.ev.inserted.sub(onInserted);
            table.ev.removing.sub(onRemoving);
            table.ev.set.sub(onSet);
            this.dispose = () => {
                table.ev.inserted.unsub(onInserted);
                table.ev.removing.unsub(onRemoving);
                table.ev.set.unsub(onSet);
            };
        }
    };

    const getters = {
        c: 'getUint8', // utf8
        d: 'getUint16', // utf16
        e: 'getUint32', // utf32
        p: 'getUint8', // pascal string
        i8: 'getInt8',
        i16: 'getInt16',
        i32: 'getInt32',
        i64: 'getInt64',
        u8: 'getUint8',
        u16: 'getUint16',
        u32: 'getUint32',
        u64: 'getUint64',
        f32: 'getFloat32',
        f64: 'getFloat64',
        v: 'getVoid'
    };
    const setters = {
        c: 'setUint8',
        d: 'setUint16',
        e: 'setUint32',
        p: 'setUint8',
        i8: 'setInt8',
        i16: 'setInt16',
        i32: 'setInt32',
        i64: 'setInt64',
        u8: 'setUint8',
        u16: 'setUint16',
        u32: 'setUint32',
        u64: 'setUint64',
        f32: 'setFloat32',
        f64: 'setFloat64',
        v: 'setVoid'
    };
    function addPointerTypes(accessors, depth, sizeof_p) {
        sizeof_p = sizeof_p || 32;
        Object.keys(accessors).forEach(field => {
            accessors[field+'*'] = accessors[field][0]+'etUint'+sizeof_p;
        });
        depth && addPointerTypes(accessors, (depth|0)-1, sizeof_p);
    };
    addPointerTypes(getters, 2);
    addPointerTypes(setters, 2);
    const sizeof = {
        c: 1, d: 2, e: 4, p: 1
    };
    Object.keys(getters).forEach(field => {
        if ( sizeof[field] ) {
            /// already known
        } else if ( field.endsWith('*') ) {
            sizeof[field] = 4;
        } else {
            sizeof[field] = (field.slice(1)|0) >> 3;
        }
    });

    const arrayExpr = /([^[]+)\[(.*)\]/;
                                
    exports.typedefStruct = function TypedefStruct(spec, littleEndian) {
        const le = !! littleEndian;
        return class Struct {
            constructor(dataview, offset) {
                let off = offset;
                for (const [typ, nm] of spec) {
                    let o = off,
                        arrayMatch = typ.match(arrayExpr);
                    if ( arrayMatch ) {
                        let [atyp, adim] = [arrayMatch[1], (arrayMatch[2]|0)];
                        adim = adim|0;
                        const getterName = getters[atyp];
                        const setterName = setters[atyp];
                        const sz = sizeof[atyp];
                        off += adim*sz;
                        if ( ['c', 'd', 'e'].indexOf(atyp) >= 0 ) {
                            const encoding = {c: 'UTF-8', d: 'UTF-16', e: 'UTF-32'}[atyp];
                            const sView = new StringView(dataview.buffer, encoding, o, adim);
                            Object.defineProperty(this, nm, {
                                get: () => sView.toString(),
                                set: s => {
                                    sView.rawData.fill(0);
                                    sView.rawData.set(new TextEncoder(encoding).encode(s));
                                }
                            });
                        } else if ( atyp === 'p' ) {
                            Object.defineProperty(this, nm, {
                                get: () => {
                                    const n = dataview.getUint8(o);
                                    let s = '',
                                        i = 1;
                                    for (; i<=n; ++i) {
                                        s += String.fromCharCode(dataview.getUint8(o+i));
                                    }
                                    return s;
                                },
                                set: s => {
                                    const n = Math.min(sz-1, s.length);
                                    dataview.setUint8(o, n);
                                    let i = 1;
                                    for (; i<=n; ++i) {
                                        dataview.setUint8(o+i, s.charCodeAt(i-1));
                                    }
                                }
                            });
                        } else {
                            Object.defineProperty(this, nm, {
                                get: () => exports.range(adim).map(i => dataview[getterName](o+i*sz, le)),
                                set: arr => arr.forEach((v,i) => dataview[setterName](o+i*sz, v, le))
                            });
                        }
                    } else {
                        off += sizeof[typ];
                        Object.defineProperty(this, nm, {
                            get: () => dataview[getters[typ]](o, le),
                            set: v => dataview[setters[typ]](o, v, le)
                        });
                    }
                }
                this.byteLength = off - offset;
            }
            copy() {
                const c = {};
                for (const [typ, nm] of spec) {
                    c[nm] = this[nm];
                }
                return c;
            }
        };
    }

    exports.QubtreeNode = exports.typedefStruct([
        ['u8', 'flags'],
        ['u8', 'flags2'],
        ['u8', 'flags3'],
        ['u8', 'dataType'],
        ['u32', 'dataSize'],
        ['u32', 'dataCount'],
        ['u32', 'dataPos'],
        ['u32', 'childPos'],
        ['u32', 'siblingPos'],
        ['u32', 'rsrv4'],
        ['u16', 'rsrv2'],
        ['u8', 'rsrv1'],
        ['u8', 'nameLen']
    ], true);
    
    const QTR_MAGIC = "QUB_(;-)_QFS",
          PRELOAD = 2,
          DATA_IN_NODE = 4,
          QTR_NODE_BYTES = 32,
          QTR_TYPE_STRING = 3,
          QTR_VIEWER = [null, null, null, null, 'getUint8', 'getInt8', 'getUint16', 'getInt16', 'getUint32', 'getInt32', 'getUint64', 'getInt64', 'getFloat32', 'getFloat64', null],
          QTR_BUFTYPE = [null, null, null, null, Uint8Array, Int8Array, Uint16Array, Int16Array, Uint32Array, Int32Array, null, null, Float32Array, Float64Array, null],
          QTR_UNIT = [0, 0, 4, 1, 1, 1, 2, 2, 4, 4, 8, 8, 4, 8, 10];

    exports.readTree = function readTree(arraybuf, view, offset, parent) {
        if ( ! view ) {
            view = new DataView(arraybuf);
        }
        if ( ! offset ) {
            if ( arraybuf.byteLength < QTR_MAGIC.length ) {
                return null;
            }
            let magic = '',
                i = 0;
            for (; i<QTR_MAGIC.length; ++i) {
                magic += String.fromCharCode(view.getUint8(i));
            }
            if ( magic !== QTR_MAGIC ) {
                return null;
            }
            offset = QTR_MAGIC.length;
        }
        if ( arraybuf.byteLength < (offset + QTR_NODE_BYTES) ) {
            return null;
        }
        const ns = new exports.QubtreeNode(view, offset),
              node = {
                  flags: ns.flags,
                  dataType: ns.dataType,
                  dataSize: ns.dataSize,
                  dataCount: ns.dataCount,
                  name: '',
                  rows: 0,
                  columns: 0,
                  children: [],
                  ch: {}
              };
        readTreeData(node, arraybuf, view, offset);
        if ( ns.childPos ) {
            node.children.push(readTree(arraybuf, view, ns.childPos, node));
            node.children.reverse();
            node.ch[node.children[0].name] = node.children[0];
        }
        if ( ns.siblingPos && parent ) {
            const sib = readTree(arraybuf, view, ns.siblingPos, parent);
            parent.children.push(sib);
            parent.ch[sib.name] = sib;
        }
        const namelen = ns.nameLen;
        if ( namelen ) {
            let name = "",
                i = 0;
            for (; i<namelen; ++i) {
                name += String.fromCharCode(view.getUint8(offset+32+i));
            }
            node.name = name;
        }
        Object.freeze(node);
        Object.freeze(node.ch);
        return node;
    };

    function readTreeData(node, arraybuf, view, offset) {
        if ( ! node.dataType ) { return; }
        const unit = QTR_UNIT[node.dataType];
        node.rows = node.dataCount;
        node.columns = node.dataSize / unit;
        const pos = (node.flags & DATA_IN_NODE) ? (offset+12) : view.getUint32(offset+12, true);
        if ( node.dataType === QTR_TYPE_STRING ) {
            const n = pos + node.dataCount;
            let d = "",
                i = pos;
            for (; i<n; ++i) {
                const c = view.getUint8(i);
                if ( c ) {
                    d += String.fromCharCode(c);
                }
            }
            node.data = d;
        } else {
            const method = QTR_VIEWER[node.dataType];
            if ( method ) {
                if ( ! (node.flags & PRELOAD) ) { // potentially large e.g. sampled data
                    node.data = new QTR_BUFTYPE[node.dataType](node.rows*node.columns);
                    if ( pos % unit ) { // have to read them one-by-one if the array is not aligned within arraybuf
                        const n = node.rows*node.columns;
                        let p = pos,
                            i = 0;
                        for (; i<n; ++i, p += unit) {
                            node.data[i] = view[method](p, true);
                        }
                    } else {
                        node.data.set(new QTR_BUFTYPE[node.dataType](arraybuf, pos, node.columns*node.rows));
                    }
                } else {
                    node.data = [];
                    let p = pos,
                        i = 0;
                    for (; i<node.rows; ++i) {
                        if ( node.columns === 1 ) {
                            node.data[i] = view[method](p, true);
                            p += unit;
                        } else {
                            const row = [];
                            node.data[i] = row;
                            let j = 0;
                            for (; j<node.columns; ++j, p += unit) {
                                row[j] = view[method](p, true);
                            }
                        }
                    }
                }
            }
        }
    }
    
    exports.treeToTable = function treeToTable(root) {
        const rows = [];
        let n = 0;
        function setRowCount(count) {
            if ( n ) { return; }
            n = count;
            let i = 0;
            for (; i<count; ++i) {
                rows.push({});
            }
        }
        root && root.children.forEach(ch => {
            if ( ch.dataCount ) {
                setRowCount(ch.dataCount);
                rows.forEach((row, r) => {
                    row[ch.name] = ch.data[r];
                });
            } else {
                setRowCount(ch.children.length);
                rows.forEach((row, r) => {
                    row[ch.name] = ch.children[r].data;
                });
            }
        });
        return rows;
    };
    
    exports.batchDelayed = function batchDelayed(action, msDelay, nowait) {
        let handle = null;
        const timeout = () => {
            handle = null;
            actSoon.active = false;
            action();
        };
        const actSoon = () => {
            if ( handle ) {
                if ( nowait ) {
                    return;
                }
                clearTimeout(handle);
            }
            handle = setTimeout(timeout, msDelay);
            actSoon.active = true;
        };
        return actSoon;
    };
    
    exports.relMouseCoords = function relMouseCoords(currentElement, event) {
        let totalOffsetX = 0,
            totalOffsetY = 0,
            canvasX = 0,
            canvasY = 0;
        do {
            totalOffsetX += currentElement.offsetLeft - currentElement.scrollLeft;
            totalOffsetY += currentElement.offsetTop - currentElement.scrollTop;
        }
        while (currentElement = currentElement.offsetParent);
        canvasX = event.pageX - totalOffsetX;
        canvasY = event.pageY - totalOffsetY;
        return {
            x: canvasX,
            y: canvasY
        };
    };
    
    exports.enhash = function enhash(obj) {
        const pairs = [];
        for (const k of Object.keys(obj)) {
            if ( obj.hasOwnProperty(k) ) {
                pairs.push(k+'='+obj[k]);
            }
        }
        return pairs.join("&");
    };
    exports.unhash = function unhash(hash) {
        const obj = {};
        const h = (hash[0] === '#') ? hash.slice(1) : hash;
        h.split("&").map(pair => {
            const sides = pair.split("=");
            if ( sides.length == 2 ) {
                obj[sides[0]] = sides[1];
            }
        });
        return obj;
    };

    exports.applyTdead = function applyTdead(classes, durs, tdead, opt) {
        const thresh = tdead * .999; // to overcome float/double rounding errors, distinguishing between < and >=
        // opt: subtract, keepInitial
        const L = durs.length;
        let iWr = 0;
        if ( L ) {
            let firstAlive = 0;
            let tm = 0;
            while ( firstAlive < L ) {
                const d = durs[firstAlive];
                if ( d < thresh ) {
                    tm += d;
                    ++firstAlive;
                } else {
                    if ( opt.keepInitial ) {
                        durs[firstAlive] += tm;
                    }
                    break;
                }
            }
            const subtract = opt.subtract;
            let iRd = firstAlive;
            for (; iRd<L; ++iRd) {
                const cls = classes[iRd];
                tm = durs[iRd];
                if ( iWr && ((classes[iWr-1] === cls) || (tm < thresh)) ) {
                    durs[iWr-1] += tm;
                } else {
                    classes[iWr] = cls;
                    if ( subtract ) {
                        tm = Math.max(0.0, tm - tdead);
                    }
                    durs[iWr] = tm;
                    ++iWr;
                }
            }
        }
        return iWr;
    };

    exports.scrolledInt = function scrolledInt(val, offset, fine, coarse, lo, hi) {
        val = val|0;
        if ( fine ) {
            offset /= 10.0;
        } else if ( coarse ) {
            offset *= 10.0;
        }
        var result;
        if ( val < 100 ) {
            offset = ((offset|0) || ((offset>0) ? 1 : -1));
            result = val + offset
        } else {
            const sgn = (val >= 0) ? 1 : -1
            result = (val * Math.pow(1.01, sgn*offset)) | 0;
        }
        if ( typeof(lo) !== 'undefined' ) {
            result = Math.max(lo, result);
        }
        if ( typeof(hi) !== 'undefined' ) {
            result = Math.min(hi, result);
        }
        return result;
    };

    exports.scrolledFloat = function scrolledFloat(val, offset, fine, coarse, lo, hi) {
        if ( fine ) {
            offset /= 10.0;
        } else if ( coarse ) {
            offset *= 10.0;
        }
        const sgn = (val >= 0) ? 1 : -1;
        let result = val * Math.pow(1.01, sgn*offset);
        if ( typeof(lo) !== 'undefined' ) {
            result = Math.max(lo, result);
        }
        if ( typeof(hi) !== 'undefined' ) {
            result = Math.min(hi, result);
        }
        return result;
    };

    exports.scrolledDigits = function scrolledDigits(val, offset, fine, coarse, lo, hi) {
        let incr = val ? 1e15 : 1.0;
        while ( val && (Math.abs(val) < incr) ) {
            incr *= 1e-1;
        }
        if ( (offset < 0) && (Math.abs(val) < 1.1*incr) ) {
            incr *= 1e-1;
        }
        if ( coarse ) {
            // most significant digit
        } else if ( fine ) {
            incr *= 1e-2;
        } else {
            incr *= 1e-1;
        }
        let result = val + offset * incr;
        if ( typeof(lo) !== 'undefined' ) {
            result = Math.max(lo, result);
        }
        if ( typeof(hi) !== 'undefined' ) {
            result = Math.min(hi, result);
        }
        return result;
    };

    exports.fieldOrDef = function fieldOrDef(obj, field, defval) {
        if ( typeof(obj) === 'undefined' ) {
            return defval;
        }
        const f = obj[field];
        if ( typeof(f) === 'undefined' ) {
            return defval;
        }
        return f;
    };

    exports.Kalman1D = class Kalman1D {
        constructor(config) {
            config = config || {};
            const rdConf = exports.fieldOrDef;
            this.A = rdConf(config, 'A', 1.0); // process/transition
            this.Q = rdConf(config, 'Q', 1e-10); // process variance
            this.H = rdConf(config, 'H', 1.0); // process -> measurement
            this.R = rdConf(config, 'R', 1e-6); // measurement variance
            this.X = rdConf(config, 'X', 0.0); // posterior process -- maybe initialize with first data point (divided by H)
            this.Z = rdConf(config, 'Z', 0.0); // posterior measurement
            this.P = rdConf(config, 'P', 0.0); // posterior variance
            this.K = rdConf(config, 'K', 0.0); // kalman gain
        }
        reset() {
            this.P = this.Q;
        }
        next(obs, dest) {
            dest = dest || this;
            const A = this.A,
                  H = this.H,
                  Xprior = A * this.X,
                  Pprior = A * this.P * A + this.Q,
                  K = dest.K = Pprior * H / (H*Pprior*H + this.R);
            dest.X = Xprior + K * (obs - H*Xprior);
            dest.P = (1 - K*H) * Pprior;
            dest.Z = H * dest.X;
            return dest;
        }
        filter(xx, filtered, variance, init) {
            const n = xx.length,
                  ff = filtered || new Float32Array(n);
            if ( init && n ) {
                this.X = xx[0];
            }
            let i = 0;
            for (; i<n; ++i) {
                ff[i] = this.next(xx[i]).Z;
                if ( variance ) {
                    variance[i] = this.P;
                }
            }
            return ff;
        }
    };

    let gaussEngine = null,
        gaussUniDist = null;
    exports.gaussRandom = function gaussRandom() {
        // polar box-muller form
        gaussEngine = gaussEngine || Random.engines.mt19937().autoSeed();
        gaussUniDist = gaussUniDist || Random.real(-1, 1);
        var u, v, s;
        while ( (! s) || (s >= 1) ) {
            u = gaussUniDist(gaussEngine);
            v = gaussUniDist(gaussEngine);
            s = u*u + v*v;
        }
        return u * Math.sqrt(-2*Math.log(s)/s);
    };

    exports.msd = function msd(x) { // most significant digit
        const expScale = Math.pow(10, Math.floor(Math.log10(x)));
        return expScale * Math.round(x / expScale);
    };

    exports.log = obj => console.log(obj);

    exports.errOut = {
        ws: exports.ws('ws'+window.location.origin.replace(/^http/, '')+'/online/errs', {
            onerror: exports.log
        }),
        send: function(err, other) {
            const msg = {
                href: window.location.href,
                userAgent: navigator.userAgent,
                err: err
            };
            if ( typeof(other) !== 'undefined' ) {
                msg.other = other;
            }
            try {
                msg.stack = err.stack;
            } catch (xe) {}
            this.ws.send(JSON.stringify(msg));
        },
        whitelist: ['Canceled', 'Timed out (flush)', 'Layout expired', 'Not ready', 'Use require']
    };
    exports.reportErr = function reportErr(err) {
        if ( (! err.qubReported) && (exports.errOut.whitelist.indexOf(err.tktErr || (''+err)) < 0) ) {
            console.log(err);
            exports.errOut.send(err);
        }
        try { // avoid multi-reporting if possible, when catch(qubx.reportErr) shows up multiple times in the promise chain
            err.qubReported = true;
        } catch (whatever) {}
        return Promise.reject(err);
    };
    
    return exports;
});
    
