/* Copyright 2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_cloud_ui", ["qubx", "qubx_ui", "cloud"], (qubx, qubx_ui, cloud) => {
    'use strict';

    function formatChecked(x) { return x ? '\u2611' : '\u2610'; }

    function recentEnough(expireMS, load) {
        let last = null,
            lastVal = null,
            waiting = [];
        return {
            invalidate: function() {
                last = null;
            },
            load: function() {
                if ( last ) {
                    if ( (new Date().getTime()) > (last + expireMS) ) {
                        last = null;
                    }
                }
                if ( last ) {
                    return Promise.resolve(lastVal);
                } else {
                    return new Promise((resolve, reject) => {
                        waiting.push({
                            resolve: resolve,
                            reject: reject
                        });
                        if ( waiting.length === 1 ) {
                            load().then(v => {
                                last = new Date().getTime();
                                lastVal = v;
                                waiting.forEach(w => w.resolve(v));
                                waiting = [];
                            }).catch(err => {
                                waiting.forEach(w => w.reject(err));
                                waiting = [];
                            });
                        }
                    });
                }
            }
        };
    }
    
    const adminDlg = new qubx_ui.Modal({
        title: 'Administer',
        template: '.qubx_dlg_group',
        canCancel: false,
        okLabel: 'Close'
    });
    adminDlg.dom.then(dom => {
        adminDlg.spinner = new qubx_ui.Spinner(dom.content);
        const dotdot = qubx_ui.querydotdot(dom.content);
        adminDlg.tileGroup = dotdot("#qubx_group_tile_members");
        adminDlg.mnuGroup = dotdot("#qubx_group_pick", mnu => {
            mnu.addEventListener('change', onPickGroup);
        });
        adminDlg.btnRefreshGroup = dotdot.listen("#qubx_group_refresh", 'click', onClickRefreshGroup);
        adminDlg.btnNewGroup = dotdot.listen("#qubx_group_new", 'click', onClickNewGroup);
        adminDlg.btnEditGroup = dotdot.listen("#qubx_group_toEdit", 'click', onClickEditGroup);
        adminDlg.btnHopefuls = dotdot.listen("#qubx_group_hopefuls", 'click', onClickHopefuls);
        adminDlg.btnHopefulsSuper = dotdot.listen("#qubx_group_hopefuls_super", 'click', onClickHopefulsSuper);
        adminDlg.btnHopefulsSuper.style.display = 'none';
        adminDlg.btnAddMember = dotdot.listen("#qubx_group_addMember", 'click', onClickAddMember);
        adminDlg.btnWhosOnline = dotdot.listen("#qubx_group_who", 'click', onClickWhosOnline);
        adminDlg.tblMembers = dotdot("#qubx_group_members", node => new qubx_ui.ColumnTable({
            node: node,
            columns: [
                qubx_ui.ColumnTable.txtCol('Name', 'personName'),
                qubx_ui.ColumnTable.txtCol('Active', 'personActive', formatChecked),
                qubx_ui.ColumnTable.txtCol('Admin', 'personIsAdmin', formatChecked),
                qubx_ui.ColumnTable.txtCol('Contact', 'personIsContact', formatChecked),
                qubx_ui.ColumnTable.btnCol('Edit', '\u270e', onClickEditMember)
            ]
        }));
        adminDlg.tblServers = dotdot("#qubx_group_servers", node => new qubx_ui.ColumnTable({
            node: node,
            columns: [
                qubx_ui.ColumnTable.txtCol('Online', 'isOnline', formatChecked),
                qubx_ui.ColumnTable.txtCol('URL', 'serverBaseUrl'),
                qubx_ui.ColumnTable.txtCol('Seen', 'serverWhenLast', qubx.formatDate),
                qubx_ui.ColumnTable.txtCol('Since', 'serverWhenFirst', qubx.formatDate),
                qubx_ui.ColumnTable.txtCol('Coverage', 'serverCoverage', qubx.formatK0),
                qubx_ui.ColumnTable.txtCol('Free MB', 'serverFreeMB', qubx.formatStr)
            ]
        }));

        adminDlg.tileEdit = dotdot("#qubx_group_tile_edit");
        adminDlg.txtGroupHead = dotdot("#qubx_group_edit_head");
        adminDlg.txtGroupName = dotdot("#qubx_group_name");
        adminDlg.txtGroupAbout = dotdot("#qubx_group_about");
        adminDlg.chkGroupActive = dotdot("#qubx_group_active");
        adminDlg.txtDataKey = dotdot("#qubx_group_dataKey");
        adminDlg.txtComputeKey = dotdot("#qubx_group_computeKey");
        adminDlg.toHideForNewGroup = dotdot("#qubx_group_editOnly");
        adminDlg.tblLimits = dotdot("#qubx_group_limits", node => new qubx_ui.ColumnTable({
            node: node,
            columns: [
                qubx_ui.ColumnTable.txtCol('Limits', 'name'),
                qubx_ui.ColumnTable.inputCol('', 'value', 'number', qubx.acceptIntAtLeast(0), qubx.formatStr, undefined, 0)
            ]
        }));
        adminDlg.btnNewGroupCancel = dotdot.listen("#qubx_group_new_cancel", 'click', onClickNewGroupCancel);
        adminDlg.btnNewGroupOK = dotdot.listen("#qubx_group_new_ok", 'click', onClickNewGroupOK);

        adminDlg.tileMember = dotdot("#qubx_group_tile_member");
        adminDlg.txtMemberHead = dotdot("#qubx_member_head");
        adminDlg.txtMemberName = dotdot("#qubx_member_name");
        adminDlg.txtMemberEmail = dotdot("#qubx_member_email");
        adminDlg.btnInvite = dotdot.listen("#selfInvite", 'click', onClickInvite);
        adminDlg.mnuMemberGroup = dotdot("#qubx_member_group");
        adminDlg.chkMemberIsActive = dotdot("#qubx_member_active");
        adminDlg.chkMemberIsAdmin = dotdot("#qubx_member_isAdmin");
        adminDlg.chkMemberIsContact = dotdot("#qubx_member_isContact");
        adminDlg.chkMemberIsSuper = dotdot("#qubx_member_isSuper");
        adminDlg.txtMemberAbout = dotdot("#qubx_member_about");
        adminDlg.btnMemberSeats = dotdot.listen("#qubx_member_seats", 'click', onClickMemberSeats);
        adminDlg.btnNewMemberCancel = dotdot.listen("#qubx_member_new_cancel", 'click', onClickNewMemberCancel);
        adminDlg.btnNewMemberOK = dotdot.listen("#qubx_member_new_ok", 'click', onClickNewMemberOK);

        adminDlg.mainBtns = dom.buttons;

        adminDlg.groupLister = recentEnough(120000, () => cloud.personIsSuper ? cloud.listGroups() : Promise.resolve([cloud.group]));
        adminDlg.memberLister = recentEnough(120000, () => cloud.listPeople(adminDlg._groupId));
    });
    function adminTile(tile) {
        tile = tile || adminDlg.tileGroup;
        [adminDlg.tileGroup, adminDlg.tileEdit, adminDlg.tileMember].forEach(t => {
            t.style.display = (t !== tile) ? 'none' : '';
        });
        adminDlg.mainBtns.style.display = (tile !== adminDlg.tileGroup) ? 'none' : '';
        if ( tile === adminDlg.tileGroup ) {
            adminDlg.spinner.show();
            adminDlg._groupId = adminDlg._groupId || (cloud.group.groupId|0);
            Promise.all([
                adminDlg.groupLister.load().then(groups => {
                    adminDlg._groups = groups;
                    let iGroup = -1;
                    qubx_ui.setSelectOptions(adminDlg.mnuGroup, groups.map((g, i) => {
                        if ( g.groupId === adminDlg._groupId ) {
                            iGroup = i;
                        }
                        return g.groupName;
                    }));
                    adminDlg.mnuGroup.selectedIndex = iGroup;
                }),
                adminDlg.memberLister.load().then(people => {
                    adminDlg._people = people;
                    adminDlg.tblMembers.render(people, 'personId');
                }),
                cloud.listServers(adminDlg._groupId).then(servers => {
                    adminDlg._servers = servers;
                    adminDlg.tblServers.render(servers, 'serverId');
                }),
                cloud.listHopeful(adminDlg._groupId).then(hopefuls => {
                    adminDlg._hopefuls = hopefuls;
                    if ( hopefuls.length ) {
                        adminDlg.btnHopefuls.textContent = `${hopefuls.length} account request${(hopefuls.length>1) ? 's' : ''} pending...`;
                        adminDlg.btnHopefuls.style.display = '';
                    } else {
                        adminDlg.btnHopefuls.style.display = 'none';
                    }
                }),
                (cloud.personIsSuper ? cloud.listHopeful().then(hopefuls => {
                    adminDlg._unclaimed = hopefuls;
                    if ( hopefuls.length ) {
                        adminDlg.btnHopefulsSuper.textContent = 'See '+hopefuls.length+' unclaimed...';
                        adminDlg.btnHopefulsSuper.style.display = '';
                    } else {
                        adminDlg.btnHopefulsSuper.style.display = 'none';
                    }
                }) : Promise.resolve(true))
            ]).then(() => adminDlg.spinner.hide())
                .catch(qubx.reportErr);
        }
        adminDlg.tile = tile;
    }
    function onPickGroup(event) {
        adminDlg._groupId = adminDlg._groups[adminDlg.mnuGroup.selectedIndex].groupId;
        adminDlg.groupLister.invalidate();
        adminDlg.memberLister.invalidate();
        adminTile();
    }
    function fillGroup(g, isNew) {
        adminDlg.txtGroupHead.textContent = isNew ? 'New Group' : 'Edit Group';
        adminDlg.toHideForNewGroup.style.display = isNew ? 'none' : '';
        adminDlg.txtGroupName.value = g.groupName || 'Untitled Group';
        adminDlg.txtGroupAbout.value = g.groupAbout || '';
        adminDlg.chkGroupActive.parentNode.style.display = cloud.personIsSuper ? '' : 'none';
        adminDlg.chkGroupActive.checked = !! g.groupActive;
        adminDlg.txtDataKey.style.display = isNew ? 'none' : '';
        adminDlg.txtDataKey.value = g.groupDataKey || '';
        adminDlg.txtComputeKey.style.display = isNew ? 'none' : '';
        adminDlg.txtComputeKey.value = g.groupComputeKey || '';
        adminDlg._editGroupLimits = [
            {name: 'Members', value: g.groupMaxPerson|0},
            {name: 'Admins', value: g.groupMaxAdmin|0},
            {name: 'Seats', value: g.groupMaxSeats|0},
            {name: 'Compute minutes', value: g.groupComputeMinutes}
        ];
        adminDlg.tblLimits.render(adminDlg._editGroupLimits, 'name');
    }
    function onClickRefreshGroup() {
        adminDlg.memberLister.invalidate();
        adminTile();
    }
    function onClickNewGroup() {
        fillGroup({
            groupActive: true,
            groupMaxPerson: 1,
            groupMaxAdmin: 1,
            groupMaxSeats: 2,
            groupComputeMinutes: 0
        }, true);
        adminDlg._editGroupId = 0;
        adminTile(adminDlg.tileEdit);
    }
    function onClickEditGroup() {
        const g = adminDlg._groups[adminDlg.mnuGroup.selectedIndex]
        fillGroup(g);
        adminDlg._editGroup = g;
        adminDlg._editGroupId = g.groupId;
        adminTile(adminDlg.tileEdit);
    }
    function onClickNewGroupCancel() {
        // confirm?

        adminTile();
    }
    function onClickNewGroupOK() {
        let step1 = Promise.resolve(true);
        const lim = adminDlg._editGroupLimits.map(l => l.value|0);
        adminDlg.spinner.show();
        if ( adminDlg._editGroupId ) {
            const g = adminDlg._editGroup,
                  toSet = [];
            if ( g.groupName !== adminDlg.txtGroupName.value ) {
                toSet.push(cloud.setGroupName(g.groupId, adminDlg.txtGroupName.value));
            }
            if ( g.groupAbout !== adminDlg.txtGroupAbout.value ) {
                toSet.push(cloud.setGroupAbout(g.groupId, adminDlg.txtGroupAbout.value))
            }
            if ( g.groupDataKey !== adminDlg.txtDataKey.value ) {
                toSet.push(cloud.setGroupDataKey(g.groupId, adminDlg.txtDataKey.value));
            }
            if ( g.groupComputeKey !== adminDlg.txtComputeKey.value ) {
                toSet.push(cloud.setGroupComputeKey(g.groupId, adminDlg.txtComputeKey.value));
            }
            if ( cloud.personIsSuper ) {
                if ( adminDlg.chkGroupActive.checked !== g.groupActive ) {
                    toSet.push(cloud.setGroupActive(g.groupId, adminDlg.chkGroupActive.checked));
                }
                if ( g.groupMaxPerson !== lim[0] ) {
                    toSet.push(cloud.setGroupMaxPerson(g.groupId, lim[0]));
                }
                if ( g.groupMaxAdmin !== lim[1] ) {
                    toSet.push(cloud.setGroupMaxAdmin(g.groupId, lim[1]));
                }
                if ( g.groupMaxSeats !== lim[2] ) {
                    toSet.push(cloud.setGroupMaxSeats(g.groupId, lim[2]));
                }
                if ( g.groupComputeMinutes !== lim[3] ) {
                    toSet.push(cloud.setGroupComputeMinutes(g.groupId, lim[3]));
                }
            }
            step1 = Promise.all(toSet);
        } else {
            step1 = cloud.addGroup({
                groupName: adminDlg.txtGroupName.value,
                groupAbout: adminDlg.txtGroupAbout.value,
                groupMaxPerson: lim[0],
                groupMaxAdmin: lim[1],
                groupMaxSeats: lim[2],
                groupComputeMinutes: lim[3]
            }).then(groupId => {
                adminDlg._groupId = groupId;
            });
        }
        step1.then(() => {
            adminDlg.groupLister.invalidate();
            adminDlg.memberLister.invalidate();
            adminTile();
        })
            .catch(qubx.reportErr)
            .catch(err => {
                alert('When saving the group: '+err);
                adminDlg.spinner.hide();
            });
    }
    function onClickHopefuls() {
        showHopefuls(adminDlg._hopefuls)
            .then(onClickRefreshGroup);
    }
    function onClickHopefulsSuper() {
        showUnclaimed(adminDlg._unclaimed)
            .then(onClickRefreshGroup);
    }
    function onClickWhosOnline() {
        showWhosOnline({groupId: adminDlg._groupId});
    }
    function fillMember(p, isNew) {
        adminDlg.txtMemberHead.textContent = isNew ? 'New Member' : 'Edit Member Info';
        adminDlg.btnMemberSeats.style.display = isNew ? 'none' : '';
        adminDlg.txtMemberName.value = p.personName || 'Unnamed Member';
        adminDlg.txtMemberEmail.value = p.personEmail || '';
        adminDlg.mnuMemberGroup.style.display = cloud.personIsSuper ? '' : 'none';
        let iGroup = -1;
        qubx_ui.setSelectOptions(adminDlg.mnuMemberGroup, adminDlg._groups.map((group, i) => {
            if ( group.groupId === p.personGroup ) {
                iGroup = i;
            }
            return group.groupName;
        }));
        adminDlg.mnuMemberGroup.selectedIndex = iGroup;
        adminDlg.btnMemberSeats.style.display = (p.personEmail === cloud.personEmail) ? '' : 'none';
        adminDlg.chkMemberIsActive.parentNode.style.display = cloud.personIsAdmin ? '' : 'none';
        adminDlg.chkMemberIsActive.checked = !! p.personActive;
        adminDlg.chkMemberIsAdmin.parentNode.style.display = cloud.personIsAdmin ? '' : 'none';
        adminDlg.chkMemberIsAdmin.checked = !! p.personIsAdmin;
        adminDlg.chkMemberIsContact.parentNode.style.display = cloud.personIsAdmin ? '' : 'none';
        adminDlg.chkMemberIsContact.checked = !! p.personIsContact;
        adminDlg.chkMemberIsSuper.parentNode.style.display = cloud.personIsSuper ? '' : 'none';
        adminDlg.chkMemberIsSuper.checked = !! p.personIsSuper;
        adminDlg.txtMemberAbout.value = p.personAbout || '';
    }
    function onClickAddMember() {
        fillMember({
            personActive: true
        }, true);
        adminDlg._editPersonId = 0;
        adminTile(adminDlg.tileMember);
    }
    function onClickEditMember(person, iPerson) {
        adminDlg._editPersonDirect = false;
        fillMember(person);
        adminDlg._editPerson = person;
        adminDlg._editPersonId = person.personId;
        adminTile(adminDlg.tileMember);
    }
    function directEditMember() {
        return cloud.aboutPerson().then(p => {
            adminDlg._groups = adminDlg._groups || [cloud.group];
            adminDlg._groupId = p.personGroup;
            adminDlg._editPerson = p;
            adminDlg._editPersonId = p.personId;
            adminDlg._editPersonDirect = true;
            fillMember(p);
            adminTile(adminDlg.tileMember);
            return adminDlg.run();
        });
    }
    function onClickInvite() {
        promptInvite(adminDlg._editPersonId)
            .catch(qubx.reportErr);
    }
    function onClickMemberSeats() {
        showWhosOnline(); // should be {personId: ...} but server allows groupId or own seats only
    }
        
    function onClickNewMemberCancel() {
        // confirm?

        if ( adminDlg._editPersonDirect ) {
            adminDlg.close();
        } else {
            adminTile();
        }
    }
    function onClickNewMemberOK() {
        let step1 = Promise.resolve(true);
        adminDlg.spinner.show();
        if ( adminDlg._editPersonId ) {
            const p = adminDlg._editPerson,
                  toSet = [];
            if ( p.personEmail !== adminDlg.txtMemberEmail.value ) {
                toSet.push(cloud.setPersonEmail(p.personId, adminDlg.txtMemberEmail.value));
            }
            if ( p.personAbout !== adminDlg.txtMemberAbout.value ) {
                toSet.push(cloud.setPersonAbout(p.personId, adminDlg.txtMemberAbout.value))
            }
            if ( cloud.personIsSuper || cloud.personIsAdmin ) {
                if ( (!!p.personActive) !== adminDlg.chkMemberIsActive.checked ) {
                    toSet.push(cloud.setPersonActive(p.personId, adminDlg.chkMemberIsActive.checked));
                }
                if ( (!!p.personIsAdmin) !== adminDlg.chkMemberIsAdmin.checked ) {
                    toSet.push(cloud.setPersonIsAdmin(p.personId, adminDlg.chkMemberIsAdmin.checked));
                }
                if ( (!!p.personIsContact) !== adminDlg.chkMemberIsContact.checked ) {
                    toSet.push(cloud.setPersonIsContact(p.personId, adminDlg.chkMemberIsContact.checked));
                }
            }
            if ( cloud.personIsSuper ) {
                if ( p.personGroup !== adminDlg._groupId ) {
                    toSet.push(cloud.setPersonGroup(p.personId, adminDlg._groupId));
                }
                if ( (!!p.personIsSuper) !== adminDlg.chkMemberIsSuper.checked ) {
                    toSet.push(cloud.setPersonIsSuper(p.personId, adminDlg.chkMemberIsSuper.checked));
                }
            }
            step1 = Promise.all(toSet);
        } else {
            step1 = cloud.addPerson({
                personGroup: adminDlg._groupId,
                personName: adminDlg.txtMemberName.value,
                personEmail: adminDlg.txtMemberEmail.value,
                personAbout: adminDlg.txtMemberAbout.value,
                personIsAdmin: adminDlg.chkMemberIsAdmin.checked,
                personIsContact: adminDlg.chkMemberIsContact.checked,
                personIsSuper: cloud.personIsSuper && adminDlg.chkMemberIsSuper.checked
            });
        }
        step1.then(() => {
            adminDlg.spinner.hide();
            adminDlg.memberLister.invalidate();
            if ( adminDlg._editPersonDirect ) {
                adminDlg.close();
            } else {
                adminTile();
            }
        })
            .catch(qubx.reportErr)
            .catch(err => {
                alert('When saving the member: '+err);
                adminDlg.spinner.hide();
            });
    }

    function hopefulLabel(c) {
        if ( hopefulDlg._unclaimed ) {
            for (const g of adminDlg._groups) {
                if ( g.groupId === c.personGroup ) {
                    c.personLabel = `${c.personName} (${g.groupName})`;
                    return;
                }
            }
        }
        c.personLabel = c.personName;
    }
    
    const hopefulDlg = new qubx_ui.Modal({
        title: 'Approve new members',
        template: '#qubx_hopefuls',
        canCancel: false,
        okLabel: 'Close'
    });
    hopefulDlg.dom.then(dom => {
        const dotdot = qubx_ui.querydotdot(dom.content);
        hopefulDlg.txtCount = dotdot("#qubx_hopeful_count");
        hopefulDlg.txtName = dotdot("#qubx_hopeful_name");
        hopefulDlg.txtEmail = dotdot("#qubx_hopeful_email");
        hopefulDlg.tblCandidates = dotdot("#qubx_hopeful_candidates", node => new qubx_ui.ColumnTable({
            node: node,
            columns: [
                qubx_ui.ColumnTable.txtCol('Link to an existing account with that email:', 'personLabel'),
                qubx_ui.ColumnTable.btnCol('', '\u2714', onClickCandidate, 'Sign them in under this account')
            ],
            autohide: true
        }));
        hopefulDlg.txtAbout = dotdot("#qubx_hopeful_about");
        hopefulDlg.txtGroupInfo = dotdot("#qubx_hopeful_groupInfo");
        hopefulDlg.btnDeny = dotdot.listen("#qubx_hopeful_deny", 'click', onClickHopefulDeny);
        hopefulDlg.toGroup = dotdot("#qubx_hopeful_toGroup");
        hopefulDlg.mnuGroup = dotdot("#qubx_hopeful_group");
        hopefulDlg.btnFwd = dotdot.listen("#qubx_hopeful_fwd", 'click', onClickHopefulFwd);
        hopefulDlg.btnBwd = dotdot.listen("#qubx_hopeful_bwd", 'click', onClickHopefulBwd);
        hopefulDlg.btnAccept = dotdot.listen("#qubx_hopeful_accept", 'click', onClickHopefulAccept);
        hopefulDlg.spinner = new qubx_ui.Spinner(dom.content);
    });
    function showHopefuls(hopefuls) {
        hopefulDlg._unclaimed = false;
        hopefulDlg._hopefuls = hopefuls;
        hopefulDlg._iHopeful = -1;
        hopefulDlg.toGroup.style.display = 'none';
        [hopefulDlg.btnBwd, hopefulDlg.btnAccept].forEach(btn => {
            btn.style.display = '';
        });
        nextHopeful();
        return hopefulDlg.run();
    }
    function showUnclaimed(hopefuls) {
        hopefulDlg._unclaimed = true;
        hopefulDlg._hopefuls = hopefuls;
        hopefulDlg._iHopeful = -1;
        hopefulDlg.toGroup.style.display = '';
        [hopefulDlg.btnBwd, hopefulDlg.btnAccept].forEach(btn => {
            btn.style.display = 'none';
        });
        qubx_ui.setSelectOptions(hopefulDlg.mnuGroup, adminDlg._groups.map(g => g.groupName));
        nextHopeful();
        return hopefulDlg.run();
    }
    function nextHopeful() {
        hopefulDlg.spinner.hide();
        hopefulDlg._iHopeful += 1;
        if ( hopefulDlg._iHopeful >= hopefulDlg._hopefuls.length ) {
            hopefulDlg.close();
            return;
        }
        const h = hopefulDlg._hopefuls[hopefulDlg._iHopeful];
        hopefulDlg._hopeful = h;
        hopefulDlg.txtCount.textContent = ''+(hopefulDlg._iHopeful+1)+'/'+hopefulDlg._hopefuls.length;
        hopefulDlg.txtName.textContent = h.hopefulName;
        hopefulDlg.txtEmail.textContent = h.hopefulEmail;
        hopefulDlg.txtAbout.textContent = h.hopefulAbout;
        hopefulDlg.txtGroupInfo.textContent = h.hopefulGroupInfo;
        h.candidates.forEach(hopefulLabel);
        hopefulDlg.tblCandidates.render(h.candidates, 'personId');
    }
    function onClickCandidate(candidate) {
        hopefulDlg.spinner.show();
        cloud.acceptCandidate(hopefulDlg._hopeful.hopefulId, candidate.personId)
            .then(() => promptInvite(candidate.personId)
                  .catch(err => {}))
            .then(nextHopeful)
            .catch(qubx.reportErr)
            .catch(err => {
                hopefulDlg.spinner.hide();
                setTimeout(() => alert(err), 10);
            });
    }
    function onClickHopefulDeny() {
        // confirm?
        cloud.rejectHopeful(hopefulDlg._hopeful.hopefulId)
            .catch(qubx.reportErr);
        nextHopeful();
    }
    function onClickHopefulFwd() {
        cloud.fwdHopeful(hopefulDlg._hopeful.hopefulId, adminDlg._groups[hopefulDlg.mnuGroup.selectedIndex].groupId)
            .catch(qubx.reportErr);
        nextHopeful();
    }
    function onClickHopefulBwd() {
        cloud.bwdHopeful(hopefulDlg._hopeful.hopefulId)
            .catch(qubx.reportErr);
        nextHopeful();
    }
    function onClickHopefulAccept() {
        hopefulDlg.spinner.show();
        cloud.acceptHopeful(hopefulDlg._hopeful.hopefulId) // , isAdmin, isContact)
            .then(personId => promptInvite(personId)
                  .catch(err => {}))
            .then(nextHopeful)
            .catch(qubx.reportErr)
            .catch(err => {
                hopefulDlg.spinner.hide();
                setTimeout(() => alert(err.message), 10);
            });
    }

    const inviteDlg = new qubx_ui.Modal({
        title: 'Generate a new signin link',
        template: '.qubx_dlg_sendInvite'
    });
    inviteDlg.dom.then(dom => {
        const dotdot = qubx_ui.querydotdot(dom.content);
        inviteDlg.txtMsg = dotdot(".qubx_selfInvite_msg");
        inviteDlg.txtHours = dotdot("#selfInviteExpiresHours");
        inviteDlg.txtUses = dotdot("#selfInviteUsesRemain");
    });
    function promptInvite(personId) {
        var email, personLabel;
        return cloud.aboutPerson(personId).then(p => {
            personId = personId || p.personId;
            email = p.personEmail;
            personLabel = p.personName+' &lt;'+p.personEmail+'&gt;';
            inviteDlg.txtMsg.innerHTML = 'Generate a new signin link for '+personLabel+'.  The link will be sent by email, and also shown onscreen to be copied and pasted.';
            inviteDlg.txtHours.value = '720';
            inviteDlg.txtUses.value = '1';
            return inviteDlg.run();
        }).then(() => cloud.invite(personId, Math.max(1, inviteDlg.txtHours.value|0), Math.max(0, inviteDlg.txtUses.value)))
            .then(inviteKey => {
                const url = window.location.origin+'/online/signin.html#'+inviteKey;
                return qubx_ui.ShowCopyable('New signin link for '+personLabel+':', url);
            });
    }

    const whosOnlineDlg = new qubx_ui.Modal({
        title: "Who's online",
        template: "#qubx_whosonline",
        canCancel: false,
        okLabel: 'Close'
    });
    whosOnlineDlg.dom.then(dom => {
        whosOnlineDlg.tbl = new qubx_ui.ColumnTable({
            node: dom.content.querySelector("table"),
            columns: [
                qubx_ui.ColumnTable.txtCol('Name', 'personName'),
                qubx_ui.ColumnTable.txtCol('IP', 'seatIP'),
                qubx_ui.ColumnTable.txtCol('Online', 'isOnline', formatChecked),
                qubx_ui.ColumnTable.txtCol('When', 'seatWhen', qubx.formatDate),
                qubx_ui.ColumnTable.btnCol('Sign out', '<span style="color: red">\u2716</span>', onClickSignOut)
            ]
        });
        whosOnlineDlg.spinner = new qubx_ui.Spinner(dom.content);
    });
    function refreshWhosOnline() {
        const groupId = whosOnlineDlg._who.groupId;
        whosOnlineDlg.spinner.show();
        return cloud.listSeats(groupId)
            .then(seats => groupId ? seats.filter(seat => seat.isOnline) : seats)
            .then(seats => {
                whosOnlineDlg.tbl.render(seats, 'seatKey');
                whosOnlineDlg.spinner.hide();
            }).catch(qubx.reportErr)
              .catch(err => wohsOnlineDlg.spinner.hide());
    }        
    function showWhosOnline(who) {
        whosOnlineDlg._who = who || {};
        refreshWhosOnline();
        whosOnlineDlg.run();
    }
    function onClickSignOut(seat, iSeat) {
        qubx_ui.confirm({
            hdr: 'Sign out',
            q: "Remotely end the member's session?"
        }).then(() => cloud.unSignIn(seat.seatKey, seat.personId))
            .then(refreshWhosOnline)
            .catch(qubx.reportErr);
    }

    const signupDlg = new qubx_ui.Modal({
        title: 'Request an account',
        template: ".qubx_dlg_newPerson"
    });
    signupDlg.dom.then(dom => {
        const dotdot = qubx_ui.querydotdot(dom.content);
        signupDlg.txtName = dotdot("#newPersonName");
        signupDlg.txtEmail = dotdot("#newPersonEmail");
        signupDlg.txtAbout = dotdot("#newPersonAbout");
        signupDlg.txtGroupInfo = dotdot("#newPersonGroupInfo");
    });
    function requestAccount() {
        [signupDlg.txtName, signupDlg.txtEmail, signupDlg.txtAbout, signupDlg.txtGroupInfo].forEach(txt => {
            txt.value = '';
        });
        return signupDlg.run()
            .then(() => cloud.requestAccount(signupDlg.txtName.value, signupDlg.txtEmail.value, signupDlg.txtAbout.value, signupDlg.txtGroupInfo.value))
            .then(() => setTimeout(() => alert("Thanks.  We'll pass your request to the next available admin."), 10))
            .catch(qubx.reportErr);
    }

    const FileListingDefs = {
        canSelectFolder: false,
        canSelectFile: true,
        canSelectRoot: false,
        autoExpand: true
        // node, onSelect(item)
    };
    class FileListing {
        constructor(config) {
            this.config = Object.assign({}, FileListingDefs, config);
            this.outline = new qubx_ui.Outline({
                selectHead: this.config.canSelectFolder,
                selectLeaf: this.config.canSelectFile,
                onSelect: this.config.onSelect,
                node: this.config.node
            });
        }
        data(rootLabel, files_, getLabel, getFolder, getGroupName) {
            function sortByKey(items, getKey) {
                const byKey = {};
                items.forEach(item => {
                    const key = getKey(item);
                    let values = byKey[key];
                    if ( ! values ) {
                        values = [];
                        values.key = key;
                        byKey[key] = values;
                    }
                    values.push(item);
                });
                const groups = Object.values(byKey);
                groups.sort((a,b) => a.key.localeCompare(b.key));
                return groups;
            }
            const listFolders = files => {
                return sortByKey(files, getFolder).map(folder => {
                    folder.sort((a,b) => getLabel(a).localeCompare(getLabel(b)));
                    return {
                        expand: this.config.autoExpand,
                        label: folder.key,
                        children: folder.map(item => { label: getLabel(item) })
                    };
                });
            };

            if ( getGroupName ) {
                this.outline.data({
                    expand: true,
                    label: rootLabel,
                    canSelect: this.config.canSelectRoot,
                    children: sortByKey(files_, getGroupName).map(group => ({
                        expand: this.config.autoExpand,
                        label: group.key,
                        children: listFolders(group)
                    }))
                });
            } else {
                this.outline.data({
                    expand: true,
                    label: rootLabel,
                    canSelect: this.config.canSelectRoot,
                    children: listFolders(files_)
                });
            }
        }
    };

    cloud.ev.accepted.sub(p => setTimeout(() => {
            alert('Your account has been approved.  If this is not '+p.personName+', please sign out now.');
    }, 3000));
    cloud.ev.ejected.sub(reason => setTimeout(() => {
        alert('Session terminated: '+reason);
    }, 10));
    cloud.ev.hopefuls.sub(hopefuls => setTimeout(() => {
        const plural = (hopefuls.length>1);
        alert(''+hopefuls.length+' new account'+(plural?'s':'')+' awaiting your approval. To see '+(plural?'them':'it')+', go to QUB:Administer.');
    }, 3000));
    
    const exports = {
        administer: function() {
            if ( cloud.personIsAdmin || cloud.personIsSuper ) {
                adminTile();
                return adminDlg.run();
            } else {
                return directEditMember();
            }
        },
        editMember: directEditMember,
        requestAccount: requestAccount,
        FileListing: FileListing /// 
    };

    return exports;
});

/// project configuration has local modifications [Save]
//// in project view
//// on exiting
//// on switching project (cancelable)

// new model: prompt name and replace (if only one currently in project)


/// new mode: welcome with instructions instead of model/data
/// Project menu could be a button toggling project view in place of model/data/instructions

/// multi data view and analyze (dataset picker and active chk by "chosen" when more than one dataset in proj?)

/// db active flag on project model, dataset
/// collect active flags in result config
/// notice when result references nonexistant model/dataset

/// remind/help them about projects if they try analyzing freeform
/// switching modes changes project type (confirm/explain)?
/// open model adds to project
/// open data confirms that you want to switch to data-entry

/// dataset, model, project each with own "folder" 1 level deep actually e.g. modelFolder column
/// combo popup existing folder names
//// wait and sort big lists by folder on client for file browsing
////  fold in new items listing by date


/// legacy modeldb: fail smoothly if db not available


/// ramp up and down for simu


/// some way to delete old data; delete/hide old models




// reflect own-group edits locally

