/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_coldata", ["qubx", "qubx_ui", "qubx_data", "d3"], (qubx, qubx_ui, qubx_data, d3) => {
    'use strict';

    const exports = {};

    function readRows(rdr, signalNames, index, maxN) {
        const rows = [];
        let i = 0;
        rdr.readSegments((rdr, seg, iSeg) => {
            const segn = seg.n;
            let j = 0;
            for (; j<segn; ++i, ++j) {
                if ( maxN && (i === maxN) ) {
                    break;
                }
                const row = {};
                if ( index ) {
                    index.forEach(iSrc => {
                        row[signalNames[iSrc]] = seg.buffers[iSrc][j];
                    });
                } else {
                    rdr.signalNames.forEach((nm, iSig) =>{
                        row[nm] = seg.buffers[iSig][j];
                    });
                }
                rows[i] = row;
            }
        });
        return rows;
    }
    
    function readColumns(rows, colDefs) {
        const byName = {},
              columns = colDefs.map(col => {
                  const arr = new Float32Array(rows.length);
                  byName[col.rename || col.name] = arr;
                  return arr;
              }),
              n = rows.length,
              m = colDefs.length;
        let i = 0;
        for (; i<n; ++i) {
            const row = rows[i];
            let j = 0;
            for (; j<m; ++j) {
                columns[j][i] = row[colDefs[j].name];
            }
        }
        return byName;
    }
    exports.readColumns = readColumns;
    
    function uncheckDisables(col) {
        if ( col.checked ) {
            col.cells.forEach(cell => cell.classList.remove('qubx_dis'));
        } else {
            col.cells.forEach(cell => cell.classList.add('qubx_dis'));
        }
    }
    
    exports.TableReader = class TableReader {
        constructor(config) {
            config = config || {};
            const appName = config.appName || 'QUB';
            this.columnStore = {};
            this.intakeDlg = new qubx_ui.Modal({
                title: appName+' - Open data file',
                template: "#qubx_coldata_intake"
            });
            this.intakeDlg.dom.then(dom => {
                const dotdot = qubx_ui.querydotdot(dom.content);
                this.previewTable = dotdot("#qubx_coldata_preview > table");
                this.previewNdiv = dotdot("#qubx_coldata_preview_ndiv");
                this.previewN = this.previewNdiv.querySelector("#qubx_coldata_preview_n");
                this.previewNtot = this.previewNdiv.querySelector("#qubx_coldata_preview_ntot");
                if ( typeof(config.msg) !== 'undefined' ) {
                    dotdot("#qubx_coldata_intake_msg").textContent = config.msg;
                }
            }).catch(qubx.reportErr);
            
            this.editDlg = new qubx_ui.Modal({
                title: appName+' - Edit data',
                template: "#qubx_coldata_edit"
            });
            this.editDlg.dom.then(dom => {
                const dotdot = qubx_ui.querydotdot(dom.content);
                this.editDlg.chkSelectAll = dotdot.listen("#qubx_cde_selectall", 'change', this._onSelectAll.bind(this));
                this.editDlg.btnDelRows = dotdot.listen("#qubx_cde_delrows", 'click', this._onClickDelRows.bind(this));
                dotdot.listen("#qubx_cde_addrows", 'click', this._onClickAddRows.bind(this));
                dotdot.listen("#qubx_cde_addcol", 'click', this._onClickAddCol.bind(this));
                const btn = dotdot.listen("#qubx_cde_paste", 'click', this._onClickPaste.bind(this));
                btn.title = 'Paste data from the clipboard...';
                this.editTable = dotdot("#qubx_cde_data > table");
            }).catch(qubx.reportErr);
            
            this.pasteDlg = new qubx_ui.Modal({
                title: appName+' - Paste data',
                template: "#qubx_coldata_paste"
            });
            this.pasteDlg.dom.then(dom => {
                const dotdot = qubx_ui.querydotdot(dom.content);
                this.pasteDlg.chkClear = dotdot("#qubx_cdp_clear");
                this.pasteDlg.txtText = dotdot("#qubx_cdp_text");
            });
        }
        getColumn(name) {
            if ( ! this.columnStore[name] ) {
                this.columnStore[name] = {name: name};
            }
            return this.columnStore[name];
        }
        open(files, silent) {
            const f = files[0];
            return new Promise((resolve, reject) => {
                const reader = new FileReader();
                reader.onload = e => {
                    const rdr = new qubx_data.TextReader(e.target.result, true);
                    let step1 = Promise.resolve(qubx.range(rdr.signalNames.length));
                    if ( ! silent ) {
                        this._showPreview(rdr);
                        step1 = this.intakeDlg.run()
                            .then(() => this._donePreview(rdr));
                    }
                    step1.then(index => {
                        if ( index.length === 0 ) {
                            reject(new Error('No columns selected'));
                        } else {
                            resolve({
                                rows: readRows(rdr, rdr.signalNames, index),
                                columns: index.map((ix, j) =>Object.create(this.getColumn(rdr.signalNames[ix])))
                            });
                        }
                    })
                        .catch(qubx.reportErr)
                        .catch(reject);
                };
                reader.readAsText(f);
            });
        }
        _showPreview(rdr) {
            this.previewTable.innerHTML = "";
            this.preview = new qubx_ui.ColumnTable({
                node: this.previewTable,
                columns: rdr.signalNames.map((nm, i) => qubx_ui.ColumnTable.checkable(
                    qubx_ui.ColumnTable.txtCol(nm, nm, qubx.formatK0), rdr.colOccupied[i], null,
                    col => {
                        if ( col.checked ) {
                            col.cells.forEach(cell => cell.classList.remove('qubx_dis'));
                        } else {
                            col.cells.forEach(cell => cell.classList.add('qubx_dis'));
                        }
                    }))
            });
            const preData = readRows(rdr, rdr.signalNames, null, 100);
            this.preview.render(preData);
            this.previewN.textContent = ''+preData.length;
            let Ntot = 0;
            rdr.counts.forEach(count => {
                Ntot += count;
            });
            this.previewNtot.textContent = ''+Ntot;
            this.previewNdiv.style.display = (Ntot === preData.length) ? 'none' : '';
        }
        _donePreview(rdr) {
            const signalIndex = [];
            this.preview.config.columns.forEach((col, i) => col.checked && signalIndex.push(i));
            this.previewTable.innerHTML = "";
            return signalIndex;
        }
        _setData(rows, columns) {
            this.editRows = rows.map(row => Object.assign({XmeansChecked: false}, row));
            this.editColumns = columns.slice();
            this.editSelCount = 0;
            
            if ( ! this.editColumns.length ) {
                this.editColumns.push(this.getColumn('A'));
            }
            if ( ! this.editRows.length ) {
                const row = {'XmeansChecked': false};
                this.editColumns.forEach(col => {
                    row[col.rename || col.name] = 0.0;
                });
                this.editRows.push(row);
            }
            
            const colDefs = this.editColumns.map((col, i) => {
                const name = col.rename || col.name;
                return qubx_ui.ColumnTable.checkable(qubx_ui.ColumnTable.numCol(name, col.name, qubx.acceptFloat, qubx.formatK0), true, null, uncheckDisables);
            });
            colDefs.splice(0, 0, qubx_ui.ColumnTable.chkCol('', 'XmeansChecked', (row, col, isSel) => {
                this.editSelCount += isSel ? 1 : -1;
                if ( this.editSelCount === 0 ) {
                    this.editDlg.btnDelRows.disabled = true;
                } else if ( this.editSelCount === 1 ) {
                    this.editDlg.btnDelRows.disabled = false;
                }
            }));
            this.editTable.innerHTML = '';
            this.editTableHandler = new qubx_ui.ColumnTable({
                node: this.editTable,
                columns: colDefs
            });
            this.editTableHandler.render(this.editRows);
            
            this.editDlg.btnDelRows.disabled = true;
        }
        edit(rows, columns) {
            this._setData(rows, columns);
            const cleanup = err => {
                if ( err ) {
                    qubx.errOut.send(err);
                    console.log(err);
                }
                this.editTable.innerHTML = '';
                this.editTableHandler = this.editRows = this.editColumns = null;
                return err && Promise.reject(err);
            };
            return this.editDlg.run().then(() => {
                const columns = [];
                this.editColumns.forEach((col, i) => {
                    if ( this.editTableHandler.config.columns[i+1].checked ) {
                        columns.push(col);
                    }
                });
                rows = this.editRows;
                cleanup();
                return {
                    rows: rows,
                    columns: columns
                };
            }).catch(cleanup);
        }
        _onSelectAll() {
            const isSel = this.editDlg.chkSelectAll.checked,
                  colChk = this.editTableHandler.config.columns[0];
            colChk.cells.forEach((cell, i) => {
                const row = this.editRows[i];
                row.XmeansChecked = isSel;
                colChk.update(d3.select(cell), {row: row}, i);
            });
            this.editSelCount = isSel ? this.editRows.length : 0;
            this.editDlg.btnDelRows.disabled = ! isSel;
        }
        _onClickAddRows() {
            const n = prompt("How many rows would you like to add?", "1")|0;
            let i = 0;
            for (; i<n; ++i) {
                const row = {'XmeansChecked': false};
                this.editColumns.forEach(col => {
                    row[col.rename || col.name] = 0.0;
                });
                this.editRows.push(row);
            }
            this.editTableHandler.render(this.editRows);
        }
        _onClickDelRows() {
            const newRows = [];
            this.editRows.forEach(row => row.XmeansChecked || newRows.push(row));
            this.editRows = newRows;
            this.editTable.innerHTML = '';
            this.editTableHandler.render(this.editRows);
            this.editDlg.btnDelRows.disabled = true;
        }
        _onClickAddCol() {
            const iCol = this.editColumns.length,
                  name = prompt("New column name:", qubx_data.ColumnLabel(iCol));
            if ( ! name ) {
                return;
            }
            this.editColumns.push(this.getColumn(name));
            this.editRows.forEach(row => {
                row[name] = 0.0;
            });
            this.editTableHandler.config.columns.splice(1+iCol, 0, qubx_ui.ColumnTable.checkable(qubx_ui.ColumnTable.numCol(name, name, qubx.acceptFloat, qubx.formatK0), true, null, uncheckDisables));
            this.editTable.innerHTML = '';
            this.editTableHandler.render(this.editRows);
        }
        _onClickPaste() {
            var txtData;
            const clearing = this.pasteDlg.chkClear.checked;
            this.pasteDlg.run().then(() => {
                txtData = this.pasteDlg.txtText.value;
                this.pasteDlg.txtText.value = "";
                return this.open([new Blob([txtData], {type: 'text/plain'})], true);
            }).then(data => {
                if ( ! clearing ) {
                    const haveOld = {},
                          haveNew = {};
                    this.editColumns.forEach(col => {
                        haveOld[col.name] = 1;
                    });
                    data.columns.forEach(col => {
                        haveNew[col.name] = 1;
                    });
                    const oldOnly = [],
                          newOnly = [];
                    this.editColumns.forEach(col => haveNew[col.name] || oldOnly.push(col.name));
                    data.columns.forEach(col => {
                        if ( ! haveOld[col.name] ) {
                            newOnly.push(col.name);
                            this.editColumns.push(col);
                        }
                    });
                    this.editRows.forEach(row => newOnly.forEach(name => {
                        row[name] = 0;
                    }));
                    data.rows.forEach(row => oldOnly.forEach(name => {
                        row[name] = 0;
                    }));
                    data.rows = this.editRows.concat(data.rows);
                }
                this._setData(data.rows, clearing ? data.columns : this.editColumns)
            })
                .catch(qubx.reportErr);
        }
    };
    
    return exports;
});
