/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_commitor", ["qubx", "qubx_app", "qubx_lib", "qubx_ui", "qubx_ui_model", "qubx_fit"],
    function(qubx, qubx_app, qubx_lib, qubx_ui, qubx_ui_model, qubx_fit) {
        "use strict";
        var exports = {};

        exports.findCommitor = function findCommitor(mdl, tmeas_ms) {
            return qubx_lib.then(function(lib) {
                var output = "";
                var states = exports.findLinearStates(mdl);
                // warn about out-of-order states. we can't just translate to the true state numbers because the result could be midway between two states.
                for (var i=0, n=states.length; i<n; ++i) {
                    if ( i !== states[i] ) {
                        output += "<div>States are out of order. The answer will be a zero-based index into this array:<br>";
                        output += JSON.stringify(states);
                        output += "</div>";
                        break;
                    }
                }
                
                // get model matrices: Q (rate) and A (transition probability after tmeas)
                var Q = mdl.makeQ();
                var A = lib.expm(Q, tmeas_ms*1e-3);

                // we use a 1d numerical search to find the optimal commitor, so functions are defined in terms of par = [fractionalCommitorIndex]
                // the optimal commitor has Pt[first state] == Pt[last state]
                
                function calcPt(par) { // Pt = P0(commitor) * A
                    // separate commitor param into integer and fraction parts
                    var i = Math.floor(par[0]);
                    var frac = par[0] - i;
                    // use i,frac to initialize state probability vector
                    var P0 = Q.map(function(row) { return 0.0; });
                    P0[states[i]] = 1 - frac;
                    if ( frac ) {
                        P0[states[i+1]] = frac;
                    }
                    // advance probabilities by tmeas
                    return vxm(P0, A);
                }

                function testCommitor(par) {
                    // quantity to minimize: squared difference between start- and end-state probability
                    var Pt = calcPt(par);
                    var d = Pt[states[0]] - Pt[states[states.length-1]];
                    return d*d;
                }
                // search for best commitor using simplex
                var startParam = [0.5*(states.length-1)];
                var fit = qubx_fit.simplex(startParam, {}, bounded(testCommitor, 0, states.length-1));
                var Pt = calcPt(fit.param);
                output += '<div>Commitor: '+fit.param[0].toPrecision(2)+'</div>';
                output += '<div>P('+states[0]+')='+Pt[states[0]].toPrecision(5)+'; ';
                output += 'P('+states[states.length-1]+')='+Pt[states[states.length-1]].toPrecision(5)+'</div>';
                output += '<div>Residual: '+fit.min.toPrecision(3)+' in '+fit.iterations+' iterations</div>';
                return {
                    stateOrder: states,
                    commitor: fit.param[0],
                    Pt: Pt,
                    residual: fit.min,
                    iterations: fit.iterations,
                    message: output
                };
            }).catch(qubx.reportErr);
        };

        function vxm(vec, mat) {
            if ( vec.length === 0 ) {
                return vec;
            }
            var out = [];
            var m = mat.length;
            for (var i=0, n=mat[0].length; i<n; ++i) {
                var s = 0.0;
                for (var j=0; j<m; ++j) {
                    s += vec[j] * mat[j][i];
                }
                out.push(s);
            }
            return out;
        }

        function bounded(func, lo, hi) {
            return function(par) {
                if ( (par[0] < lo) || ( (typeof(hi) !== 'undefined') && (par[0] > hi) ) ) {
                    return 1e10;
                } else {
                    return func(par);
                }
            };
        }
        
        function adjacencyList(model) {
            var connTo = model.states.entries.map(function() { return []; });
            for (var i=0, n=model.rates.size; i<n; ++i) {
                connTo[model.rates.get(i, 'From')].push(model.rates.get(i, 'To'));
            }
            return connTo;
        }
        exports.findLinearStates = function findLinearStates(model) {
            // returns a list of state indices which are linearly connected, in order
            var Ns = model.states.size;
            var connTo = adjacencyList(model);
            var s = 0;
            // find first terminal state: only one neighbor
            while ( (s < Ns) && (connTo[s].length != 1) ) {
                s += 1;
            }
            var linear = [s];
            while ( connTo[s].length ) {
                // visit the neighbor, if any, and remove the reverse connection
                var s2 = connTo[s][0];
                linear.push(s2);
                var iS = connTo[s2].indexOf(s);
                connTo[s2].splice(iS, 1);
                s = s2;
            }
            return linear;
        };
        exports.buildLinearModel = function buildLinearModel(model, n, ffwd, fbwd, sfwd, sbwd) {
            model.clear();
            var dx = 100.0 / (n+1);
            for (var i=0; i<n; ++i) {
                model.states.append({
                    Class: (i == (n-1)) ? 1 : 0,
                    x: (i+1)*dx,
                    y: 50
                });
                if ( i ) {
                    model.rates.append({
                        From: i-1,
                        To: i,
                        k0: ffwd
                    });
                    model.rates.append({
                        From: i,
                        To: i-1,
                        k0: fbwd
                    });
                }
            }
            model.rates.set(0, 'k0', sfwd);
            model.rates.set(model.rates.size-1, 'k0', sbwd);
            model.undo.sealUndo("Build model");
        };

        exports.start = function start(config) {
            var main = new qubx_app.Main(config);
            main.panels.then(function(p) {
                main.runCommitor = function() {
                    exports.findCommitor(p.modelView.model, p.tmeas.value)
                        .then(function(output) {
                            p.output.innerHTML = output.message;
                        });
                };
                p.modelView = new qubx_ui_model.ModelView({node: document.createElement('div')});
                p.modelView.models.insert();
                p.modelView.models.ev.insert.sub(function() {
                    setTimeout(function() {
                        p.modelView.models.remove(0);
                    }, 10);
                });
                var topbot = document.getElementById('qubx_topbot');
                topbot.style.display = 'flex';
                topbot.style.flexFlow = 'column nowrap';
                topbot.style.alignItems = 'stretch';
                p.modelView.node.style.flex = '1 1 auto';
                topbot.appendChild(p.modelView.node);
                qubx_ui.loadTemplate("#qubx_commitor_controlbox").then(function(controls) {
                    p.controls.parentNode.removeChild(p.controls); // don't need central control strip
                    p.controls = controls;
                    p.main.insertBefore(p.controls, p.main.childNodes[0]);

                    main.about = new qubx_ui.Modal({
                        title: 'About',
                        template: '#qubx_commitor_about',
                        canCancel: false,
                        okLabel: 'Close'
                    });
                    
                    main.opener = new qubx_ui.Opener();
                    main.saver = new qubx_ui.Saver();
                    p.mnuFile = new qubx_ui.Menu({
                        node: controls.querySelector("#qubx_mnu_file"),
                        caption: 'File...',
                        groups: [[
                            {
                                caption: 'New model',
                                action: function() {
                                    try {
                                        p.modelView.info.clear();
                                    } catch (err) {
                                        qubx.reportErr(err);
                                    }
                                }
                            },
                            {
                                caption: 'Open model from disk...',
                                action: function() {
                                    main.opener.run()
                                        .then(p.modelView.openFiles.bind(p.modelView))
                                        .then(main.opener.reset.bind(main.opener))
                                        .catch(qubx.reportErr);
;
                                }.bind(this)
                            },
                            {
                                caption: 'Save model as QMJ...',
                                action: function() {
                                    try {
                                        main.saver.prompt({name: "QUB Commitor Model.qmj"}).then(function(fi) {
                                            main.saver.saveJSON(fi.fullname, p.modelView.model.toJson());
                                        });
                                    } catch(err) {
                                        qubx.reportErr(err);
                                    }
                                }.bind(this)
                            },
                            {
                                caption: 'Save model image as SVG...',
                                action: function() {
                                    p.modelView.getSVG()
                                        .then(function(svg) {
                                            main.saver.prompt({name: "Model.svg"}).then(function(fi) {
                                                main.saver.saveText(fi.fullname, svg.outerHTML, "image/svg+xml");
                                            });
                                        }.bind(this))
                                        .catch(qubx.reportErr);
                                }.bind(this)
                            }
                        ]]
                    });

                    p.nstate = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_nstate"),
                                                             5,
                                                             {
                                                                 accept: qubx.acceptIntAtLeast(3),
                                                                 format: qubx.formatStr,
                                                                 localKey: "commitor.nstate",
                                                                 widthChars: 10
                                                             });
                    p.ffwd = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_ffwd"),
                                                           10000000,
                                                           {
                                                               accept: qubx.acceptFloatGreaterThan(0),
                                                               format: qubx.formatK0,
                                                               localKey: "commitor.ffwd",
                                                               widthChars: 10
                                                           });
                    p.fbwd = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_fbwd"),
                                                           10000000,
                                                           {
                                                               accept: qubx.acceptFloatGreaterThan(0),
                                                               format: qubx.formatK0,
                                                               localKey: "commitor.fbwd",
                                                               widthChars: 10
                                                           });
                    p.sfwd = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_sfwd"),
                                                           10,
                                                           {
                                                               accept: qubx.acceptFloatGreaterThan(0),
                                                               format: qubx.formatK0,
                                                               localKey: "commitor.sfwd",
                                                               widthChars: 10
                                                           });
                    p.sbwd = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_sbwd"),
                                                           10,
                                                           {
                                                               accept: qubx.acceptFloatGreaterThan(0),
                                                               format: qubx.formatK0,
                                                               localKey: "commitor.sbwd",
                                                               widthChars: 10
                                                           });
                    p.tmeas = new qubx_ui.AcceptedInput(controls.querySelector("#qubx_commitor_tmeas"),
                                                            100,
                                                            {
                                                                accept: qubx.acceptFloatGreaterThan(0),
                                                                format: qubx.formatK0,
                                                                localKey: "commitor.tmeas",
                                                                widthChars: 10
                                                            });
                    p.tmeas.ev.change.sub(main.runCommitor);
                    var build = controls.querySelector("#qubx_commitor_build");
                    build.addEventListener('click', function() {
                        try {
                            exports.buildLinearModel(p.modelView.model, p.nstate.value, p.ffwd.value, p.fbwd.value, p.sfwd.value, p.sbwd.value);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    })
                    build.click(); // set up an initial model
                    p.output = p.controls.querySelector("#qubx_commitor_output");
                    controls.querySelector("#qubx_commitor_run").addEventListener('click', main.runCommitor);
                    p.controls.querySelector("#qubx_commitor_showabout").addEventListener('click', function() {
                        main.about.run();
                    });
                }.bind(this));
            }.bind(this))
                .catch(qubx.reportErr);
            
            return main;
        };

        return exports;
    });

