/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_data", ["qubx", "qubx_idl", "d3"], (qubx, qubx_idl, d3) => {
    'use strict';
    
    const exports = {};

    function newBuffers(cols, rows) {
        const bufs = [];
        let c = 0;
        for (; c<cols; ++c) {
            bufs[c] = new Float32Array(rows);
        }
        return bufs;
    }
    
    function isTimeColumn(buf, n) {
        if ( buf.length < 3 ) {
            return false;
        }
        const dt = buf[1] - buf[0],
              pred = buf[0] + (n-1)*dt;
        return (Math.abs(buf[n-1] - pred) < dt);
    }

    exports.ColumnLabel = function ColumnLabel(i) {
        if ( i < 26 ) {
            return String.fromCharCode(65+i);
        } else {
            return String.fromCharCode((i / 26)|0) + exports.ColumnLabel(i % 26);
        }
    };

    // signalNames, counts, sampling, readSegments(function({buffers, n}))
    exports.TextReader = class TextReader {
        constructor(t, preserveTimeSeries) {
            this.preserveTimeSeries = !! preserveTimeSeries;
            const parser = (t.indexOf("\t") < 0) ? d3.csv : d3.tsv,
                  rows = parser.parseRows(t);
            this._rows = rows;
            if ( rows && rows[0] && isNaN(rows[0][0]) ) {
                this.signalNames = rows[0];
            }
            this._i0 = this.signalNames ? 1 : 0;
            this.signalNames = (this.signalNames || (rows && rows[0].map(r => null)) || [])
                .map((nm,i) => nm || exports.ColumnLabel(i))
                .map(nm => {
                    if ( nm.startsWith('"') && nm.endsWith('"') ) {
                        return nm.substring(1, nm.length-1);
                    }
                    return nm;
                });
            this.isSampled = this.signalNames.map(x => true);
            this.counts = [];
            this.starts = [this._i0];
            let maxN = 0,
                thisN = 0;
            const pushN = (i, last) => {
                if ( thisN) {
                    this.counts.push(thisN);
                    if ( ! last ) {
                        this.starts.push(i+1);
                    }
                }
                if ( thisN > maxN ) { maxN = thisN; }
                thisN = 0;
            };
            const n = rows.length;
            let i = this._i0;
            for (; i<n; ++i) {
                if ( rows[i].length && rows[i][0] ) {
                    ++thisN;
                } else if ( thisN ) {
                    pushN(i);
                }
            }
            pushN(i, true);
            this.maxN = maxN;
            this._bufs = newBuffers(this.signalNames.length, maxN);
            this._bufsOut = this._bufs;
            this.constants = {};
            // pre-read first segment to detect time column
            this.counts.length && this._readSegment(this.counts[0], 0);
        }
        readSegments(onseg) {
            return this.counts.map((thisN, s) => {
                (s!==this._lastS) && this._readSegment(thisN, s);
                return onseg(this, {
                    buffers: this._bufsOut,
                    n: thisN
                }, s);
            });
        }
        readSegment(s, onseg) {
            (s!==this._lastS) && this._readSegment(this.counts[s], s);
            return onseg(this, {
                buffers: this._bufsOut,
                n: this.counts[s]
            }, s);
        }
        _readSegment(thisN, s) {
            let i = this.starts[s],
                r = 0;
            for (; r<thisN; ++i) {
                if ( this._rows[i].length && this._rows[i][0] ) {
                    this._rows[i].forEach((x, c) => {
                        this._bufs[c][r] = +x;
                    })
                    ++r;
                }
            }
            this.colOccupied = this._bufs.map(col => {
                const n = col.length;
                let r = 0;
                for (; r<n; ++r) {
                    if ( col[r] ) {
                        return true;
                    }
                }
                return false;
            });
            if ( (s === 0) && (this.signalNames.length === this._bufs.length) && isTimeColumn(this._bufs[0], thisN) ) {
                this.sampling = this._bufs[0][1] - this._bufs[0][0];
                if ( ! this.preserveTimeSeries ) {
                    this.signalNames.splice(0, 1);
                    this._bufsOut = this._bufs.slice(1, this._bufs.length);
                }
            }
            this._lastS = s;
        }
    };

    const unquote = s => (s.startsWith('"') && s.endsWith('"')) ? s.substring(1, s.length-1) : s;
    const unreturn = s => s.replace(/\r/g, '');
    
    exports.ATFReader = class ATFReader {
        constructor(t) {
            if ( ! t.startsWith('ATF') ) {
                this.asText = new TextReader(t);
                this.signalNames = this.asText.signalNames;
                this.isSampled = this.asText.isSampled;
                this.counts = this.asText.counts;
                this.maxN = this.asText.maxN;
                this.constants = this.asText.constants;
                this.sampling = this.asText.sampling;
                this.signals = this.asText.signals;
                this.headers = {};
                return;
            }
            const iSignals = t.indexOf('\n"Signals="') + 1,
                  headers = {};
            unreturn(t.substring(0, iSignals)).split('\n').map(unquote).forEach(line => {
                const iEq = line.indexOf('=');
                if ( iEq >= 0 ) {
                    headers[line.substring(0, iEq)] = line.substring(iEq+1);
                }
            });
            const startMS = (headers.SweepStartTimesMS || '0').split(',').map(x => +x),
                  iNames = t.indexOf('\n', iSignals) + 1,
                  iData = t.indexOf('\n', iNames) + 1,
                  colNames = unreturn(t.substring(iSignals, iNames-1)).split('\t').map(unquote).slice(1),
                  colHeads = unreturn(t.substring(iNames, iData-1)).split('\t').map(unquote).slice(1),
                  signalNames = colNames.slice(colNames.length/startMS.length),
                  units = [],
                  rows = d3.tsv.parseRows(t.substring(iData));
            headers.SignalsExported && headers.SignalsExported.split(',').forEach((name, i) => {
                signalNames[i] = name;
            });
            signalNames.forEach((name, i) => {
                const match = colHeads[i].match(/([^(]*) \(([^)]*)\)/),
                      headName = match ? match[1] : colHeads[i];
                units[i] = match ? match[2] : '';
                if ( name.startsWith('Sweep') ) {
                    name = headName;
                }
            });
            this.signals = signalNames.map((name, i) => ({
                name: name,
                units: units[i],
                scaling: 1.0,
                offset: 0.0
            }));
            this.headers = headers;
            this.maxN = rows.length;
            this.counts = startMS.map(x => this.maxN),
            this.signalNames = signalNames;
            this.isSampled = signalNames.map(x => true);
            this.sampling = (this.maxN > 1) ? ((+ rows[1][0]) - (+ rows[0][0])) : 1.0;
            this.constants = {};
            let c = 1;
            this.segments = this.counts.map(() => {
                const signals = [];
                let iSig = 0;
                for (; iSig < this.signalNames.length; ++iSig) {
                    signals.push(new Float32Array(rows.map(r => +r[c])));
                    ++c;
                }
                return signals;
            });
        }
        readSegments(onseg) {
            if ( this.asText ) {
                const rtn = this.asText.readSegments(onseg);
                this.sampling = this.asText.sampling;
                return rtn;
            }
            return this.segments.map((buffers, s) => onseg(this, {
                buffers,
                n: this.counts[s]
            }, s));
        }
        readSegment(s, onseg) {
            if ( this.asText ) {
                return this.asText.readSegment(s, onseg);
            }
            return onseg(this, {
                buffers: this.segments[s],
                n: this.counts[s]
            }, s);
        }
    };
    
    exports.DWTReader = class DWTReader {
        constructor(t) {
            this.signalNames = ['I'];
            this.isSampled = [false];
            this.counts = [];
            this.maxN = 0;
            this.constants = {};
            this.idealized = [[]];
            const hdr = /.*Sampling.ms.:\s*([0-9-eE.]+).*ClassCount:\s*(\d+)\s*(.*)/;
            let dtMS = 1.0;
            while ( t ) {
                let nextSeg = t.indexOf("Segment", 1);
                if ( nextSeg < 0 ) {
                    nextSeg = t.length;
                }
                const nextLine = t.indexOf("\n")+1,
                      info = t.slice(0, nextLine).match(hdr);
                if ( info ) {
                    dtMS = +info[1] || dtMS;
                    const Nclass = info[2]|0,
                          clsinfo = info[3].split(/\s+/).map(function(s) { return +s; }),
                          amp = [],
                          std = [];
                    let countSum = 0,
                        c = 0;
                    for (; c<Nclass; ++c) {
                        amp.push(clsinfo[2*c]);
                        std.push(clsinfo[2*c+1]);
                    }
                    const rows = d3.tsv.parseRows(t.slice(nextLine, nextSeg)),
                          Nd = rows.length,
                          classes = new Int32Array(Nd),
                          counts = new Int32Array(Nd);
                    if ( Nd && rows[0][0].length ) {
                        let i = 0;
                        for (; i<Nd; ++i) {
                            classes[i] = +rows[i][0];
                            count = Math.round(rows[i][1]/dtMS)|0;
                            counts[i] = count;
                            countSum += count;
                        }
                    } else { // some dwts have rows starting with tab
                        let i = 0;
                        for (; i<Nd; ++i) {
                            classes[i] = +rows[i][1];
                            count = Math.round(rows[i][2]/dtMS)|0;
                            durations[i] = count;
                            countSum += count;
                        }
                    }
                    this.counts.push(countSum);
                    if ( countSum > this.maxN ) {
                        this.maxN = countSum;
                    }
                    this.idealized[0].push({
                        amps: amp,
                        stds: std,
                        classes: classes,
                        counts: counts
                    });
                    t = t.slice(nextSeg);
                }
            }
            this.sampling = 1e-3*dtMS;
        }
        readSegments(onseg) {
            return this.counts.map((thisN, s) => {
                return onseg(this, {
                    buffers: [],
                    n: thisN
                }, s);
            });
        }
        readSegment(s, onseg) {
            return onseg(this, {
                buffers: [],
                n: this.counts[s]
            }, s);
        }
    };
    
    const dataAtPath = function dataAtPath(root, ...names) {
        if ( ! root ) { return null }
        for (const name of names) {
            root = root.ch[name];
            if ( ! root ) {
                return null;
            }
        }
        return root.data;
    };

    // signalNames, counts, sampling, readSegments(function({buffers, n}))
    exports.QDFReader = class QDFReader {
        constructor(arraybuf) {
            const qdf = qubx.readTree(arraybuf);
            this.qdf = qdf;
            const qsf = qdf.ch.SessionFile;
            this.sampling = (dataAtPath(qsf, 'Sampling') || dataAtPath(qdf, 'Sampling'))[0];
            this.maxN = 0;
            this.scaling = (dataAtPath(qsf, 'Scaling') || dataAtPath(qdf, 'Scaling'))[0];
            let signals = qsf && qsf.ch.Signals && qubx.treeToTable(qsf.ch.Signals).map(sig => ({
                name: sig.Name,
                units: sig.Units,
                scaling: 1/sig.Scale,
                offset: sig.Offset
            }));
            if ( ! signals ) {
                signals = qsf && qsf.ch.DataChannels && qsf.ch.DataChannels.children.map(ch => ({
                    name: ch.ch.Name.data,
                    units: ch.ch.Units.data,
                    scaling: ch.ch.Scaling.data[0],
                    offset: ch.ch.Offset && ch.ch.Offset.data[0] || 0.0
                }));
            }
            if ( ! signals ) {
                const nsig = qdf.ch.ADChannelCount.data[0];
                signals = [];
                let i = 0;
                for (; i<nsig; ++i) {
                    signals[i] = {
                        name: 'Signal'+(i+1),
                        units: "",
                        scaling: 1.0,
                        offset: 0.0
                    };
                }
            }
            this.signalNames = signals.map(qubx.m('name'));
            this.isSampled = this.signalNames.map(() => true);
            this.signals = signals;
            this.counts = qdf.ch.Segments.children.map(ch => {
                const n = ch.ch.Channels.dataCount;
                if ( n > this.maxN ) {
                    this.maxN = n;
                }
                return n;
            });
            this._bufs = newBuffers(signals.length, this.maxN);
            this.constants = {};
            if ( qsf && qsf.ch.Constants ) {
                qubx.treeToTable(qsf.ch.Constants).forEach(c => {
                    this.constants[c.Name] = c.Value;
                });
            }
        }
        readSegments(onseg) {
            const nc = this.qdf.ch.ADChannelCount.data[0];
            return this.qdf.ch.Segments.children.map((ch, iSeg) => this._readSegment(ch, iSeg, nc, onseg));
        }
        readSegment(s, onseg) {
            return this._readSegment(this.qdf.ch.Segments.children[s], s, this.qdf.ch.ADChannelCount.data[0], onseg);
        }
        _readSegment(segNode, iSeg, nc, onseg) {
            const data = segNode.ch.Channels.data,
                  nr = segNode.ch.Channels.rows,
                  scale = this.signals.map(sig => 1.0 / (this.scaling * sig.scaling));
            let p = 0,
                r = 0;
            for (; r<nr; ++r) {
                let c = 0;
                for (; c<nc; ++c, ++p) {
                    this._bufs[c][r] = data[p] * scale[c];
                }
            }
            return onseg(this, {
                n: nr,
                buffers: this._bufs
            }, iSeg);
        }
    };

    const DataDefs = {
        sampling: 1e-3,
        tCrit: 0, // ms
        // tDead: (ms) default 1e3*sampling
    };
    exports.Data = class Data {
        constructor(config) {
            this._sampling = config && config.sampling || DataDefs.sampling;
            this._tDead = config && config.tDead || (1e3 * this._sampling);
            this._tCrit = config && config.tCrit || DataDefs.tCrit;
            this.segments = [];
            this.signals = [];
            this.constants = new qubx.Table({label: 'Constants'});
            this.constants.addField({name: 'Name', defval: '', format: qubx.formatStr});
            this.constants.addField({name: 'Value', defval: 0.0, format: qubx.formatK0});
            this._nSignal = 0;
            this._reboundSoon = qubx.batchDelayed(this._rebound.bind(this), 500, true);
            this.ev = qubx.makeEvents(["changeSampling", "changeSignals", "changeSegCount", "changeSegSize", "doneSegSize", "changeRaw", "changeSamples", "changeBounds", "changeIdeal", "changeFit", "changePointLL", "changeBaseline", "changeTDead", "changeTCrit", "clear"]);
        }
        sampling(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._sampling;
            }
            if ( x !== this._sampling ) {
                const td = (this._tDead * x/this._sampling) || (1e3 * x);
                this._sampling = x;
                this.ev.changeSampling.trigger(x);
                this.tDead(td);
            }
            return this;
        }
        segmentCount(n) {
            const nPrev = this.segments.length;
            if ( typeof(n) === 'undefined' ) {
                return nPrev;
            }
            if ( n < nPrev ) {
                this.segments.splice(n, nPrev-n);
                this.ev.changeSegCount.trigger(n);
                return this;
            }
            while ( n > this.segments.length ) {
                this.segments.push({
                    n: 0,
                    seen: 0,
                    sampled: [],
                    raw: [],
                    ideal: [],
                    fit: [],
                    pointLL: null,
                    baseline: null,
                    indVar: null,
                    resampled: null
                });
            }
            return this;
        }
        tDead(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._tDead;
            }
            if ( x !== this._tDead ) {
                this._tDead = x;
                this.ev.changeTDead.trigger(x);
            }
            return this;
        }
        tCrit(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._tCrit;
            }
            if ( x !== this._tCrit ) {
                this._tCrit = x;
                this.ev.changeTCrit.trigger(x);
            }
            return this;
        }
        signalCount(nSignal, init) {
            if ( typeof(nSignal) === 'undefined' ) {
                return this._nSignal;
            }
            this._nSignal = nSignal|0;
            if ( nSignal < this.signals.length ) {
                this.signals.splice(nSignal, this.signals.length-nSignal);
            }
            while ( this.signals.length < nSignal ) {
                this.signals.push({
                    name: "",
                    units: ""
                });
            }
            init && this.signals.forEach(init);
            this.ev.changeSignals.trigger(this.signals);
        }
        segmentSize(i, nData) {
            const seg = this.segments[i],
                  nPrev = seg.n;
            if ( typeof(nData) === 'undefined' ) {
                return seg.n;
            }
            if ( (nData !== nPrev) || (seg.ideal.length !== this._nSignal) ) {
                seg.sampled = [];
                seg.raw = seg.raw && new Float32Array(nData);
                seg.ideal = [];
                seg.fit = new Float32Array(2*nData); // only for signal 0
                let j = 0;
                for (; j<nData; ++j) {
                    seg.fit[2*j+1] = -1; // negative std::no draw
                }
                seg.pointLL = null;
                let s = 0;
                for (; s<this._nSignal; ++s) {
                    seg.ideal[s] = new qubx_idl.Idealization(this._sampling, nData);
                    seg.ideal[s].amp = [];
                    seg.ideal[s].std = [];
                }
                seg.n = nData;
                seg.seen = 0;
                this.ev.changeSegSize.trigger(i, nData);
            }
            return this;
        }
        doneSegmentSize() {
            this.ev.doneSegSize.trigger();
        }
        clear(iSeg) {
            const oneSeg = (typeof(iSeg)!=='undefined');
            (oneSeg ? [this.segments[iSeg]] : this.segments).forEach(seg => {
                seg.seen = 0;
                seg.resampled = null;
                seg.ideal.forEach(idl => idl.clear());
                seg.pointLL = null;
                seg.baseline = null;
                const nData = seg.n;
                let j = 0;
                for (; j<nData; ++j) {
                    seg.fit[2*j+1] = -1; // negative std::no draw
                }
            });
            this.ev.clear.trigger(iSeg);
        }
        getSelection(iFrom, iTo, iSeg) {
            if ( typeof(iSeg) === 'undefined' ) {
                return this.segments.map((seg, i) => this.getSelection(iFrom, iTo, i))
                    .filter(qubx.i);
            }
            const seg = this.segments[iSeg];
            iFrom = iFrom || 0;
            if ( iFrom >= seg.n ) {
                return null;
            }
            if ( typeof(iTo) === 'undefined' ) {
                iTo = seg.n-1;
            } else {
                iTo = Math.min(iTo, seg.n-1);
            }
            return {
                seg: seg,
                first: iFrom,
                last: iTo
            };
        }
        setSampled(iSeg, iSignal, offset, x, count, xOffset, immediate, processed) {
            xOffset = xOffset|0;
            const seg = this.segments[iSeg],
                  buf = seg.sampled[iSignal] = seg.sampled[iSignal] || new Float32Array(seg.n),
                  raw = seg.raw = seg.raw || new Float32Array(seg.n);
            if ( typeof(x) === 'undefined' ) {
                let i = 0;
                for (; i<count; ++i) {
                    buf[offset+i] = 0;
                }
                if ( (iSignal === 0) && ! processed ) {
                    //seg.sampled[iSignal] = null;
                    let j = 0;
                    for (; j<count; ++j) {
                        raw[offset+j] = 0;
                    }
                }
            } else if ( x.length ) {
                count = count || x.length;
                if ( x.copyWithin ) {
                    buf.set(x.subarray(xOffset, xOffset+count), offset);
                    if ( (iSignal === 0) && ! processed ) {
                        raw.set(x.subarray(xOffset, xOffset+count), offset);
                    }
                } else {
                    let i = 0;
                    for (; i<count; ++i) {
                        buf[offset+i] = x[xOffset+i];
                    }
                    if ( (iSignal === 0) && ! processed ) {
                        let j = 0;
                        for (; j<count; ++j) {
                            raw[offset+j] =  buf[offset+j];
                        }
                    }
                }
            } else {
                count = 1;
                buf[offset] = x;
                if ( (iSignal === 0) && ! processed ) {
                    raw[offset] = x;
                }
            }
            seg.seen = Math.max(seg.seen, offset+count);
            if ( iSignal || processed || ! this.ev.changeRaw.subscribed() ) {
                this.ev.changeSamples.trigger(iSeg, iSignal, offset, count, immediate);
            } else {
                this.ev.changeRaw.trigger(iSeg, offset, count, immediate);
            }
        }
        editSampled(iSeg, iSignal, f) {
            /// todo: allow editing of signal 0 (raw or processed?)
            const seg = this.segments[iSeg],
                  result = f(iSignal ? seg.sampled[iSignal] : seg.raw, seg.seen);
            if ( iSignal || ! this.ev.changeRaw.subscribed() ) {
                this.ev.changeSamples.trigger(iSeg, iSignal, 0, seg.seen);
            } else {
                seg.sampled[0].set(seg.raw);
                this.ev.changeRaw.trigger(iSeg, 0, seg.seen);
            }
            this.reboundSoon(iSignal);
            return result;
        }
        reboundSoon(iSignal) {
            this._staleBounds = this._staleBounds || [];
            this._staleBounds[iSignal] = true;
            this._reboundSoon();
        }
        _rebound() {
            let i = 0;
            for (; i<this.signals.length; ++i) {
                if ( this._staleBounds[i] ) {
                    this._staleBounds[i] = false;
                    this.ev.changeBounds.trigger(i);
                }
            }
        }
        setIdeal(iSeg, iSignal, counts, classes, offset, amp, std) {
            const seg = this.segments[iSeg],
                  idl = seg.ideal[iSignal] = seg.ideal[iSignal] || new qubx_idl.Idealization(this._sampling, seg.n);
            idl.amp = idl.amp || [];
            idl.std = idl.std || [];
            idl.setDwellsCounts(counts, classes, offset);
            let i = 0;
            for (; i<amp.length; ++i) {
                if ( typeof(amp[i]) !== 'undefined' ) {
                    idl.amp[i] = amp[i];
                    idl.std[i] = std[i];
                }
            }
            this.ev.changeIdeal.trigger(iSeg, iSignal);
        }
        setFit(iSeg, offset, x, count, xOffset) {
            // x: just mean
            xOffset = xOffset|0;
            count = count || x.length;
            const seg = this.segments[iSeg],
                  points = seg.fit;
            let i = 0;
            for (; i<count; ++i) {
                points[2*i] = x[i];
                points[2*i+1] = 0.0;
            }
            this.ev.changeFit.trigger(iSeg);
        }
        setFitStd(iSeg, offset, x, count, xOffset, stride, isVar) {
            // xOffset, count of pairs, alternating mean, std; pairs separated by stride indices
            stride = stride || 2;
            xOffset = xOffset|0;
            count = count || x.length/stride;
            const seg = this.segments[iSeg],
                  points = seg.fit;
            if ( x.copyWithin && (stride === 2) && ! isVar ) {
                points.set(x.subarray(stride*xOffset, stride*(xOffset+count)), stride*offset);
            } else {
                let i = 0;
                for (; i<count; ++i) {
                    const j = 2*(offset+i),
                          k = stride*(offset+i);
                    points[j] = x[k];
                    points[j+1] = isVar ? Math.sqrt(x[k+1]) : x[k+1];
                }
            }
            this.ev.changeFit.trigger(iSeg);
        }
        setPointLL(iSeg, arr) { // whole segment only, steals arr
            const seg = this.segments[iSeg];
            seg.pointLL = arr;
            this.ev.changePointLL.trigger(iSeg);
        }
        setBaseline(iSeg, act) {
            const seg = this.segments[iSeg];
            seg.baseline = seg.baseline || new qubx_idl.BaselineNodes(this._sampling, seg.n);
            if ( act ) {
                act(seg.baseline, seg, iSeg);
                this.ev.changeBaseline.trigger(iSeg);
            }
        }
        clearBaseline(iSeg) {
            const seg = this.segments[iSeg];
            if ( seg.baseline ) {
                seg.baseline = null;
                this.ev.changeBaseline.trigger(iSeg);
            }
        }
        getTXT(opts) {
            const sep = opts.sep || '\t',
                  idealTraces = !! opts.idealTraces, // replace raw samples with sampled idl if avail
                  haveFit = !! (this.segments.length && this.segments[0].pointLL),
                  hdrs = ['Time'].concat(this.signals.map(qubx.m('name')));
            if ( haveFit ) {
                hdrs.push('Fit');
                hdrs.push('FitStd');
                hdrs.push('LL');
            }
            const dt = this._sampling;
            let result = hdrs.join(sep),
                t0 = 0;
            for (const seg of this.segments) {
                const sampled = seg.sampled.slice();
                if ( idealTraces ) {
                    sampled.forEach((points, iSig) => {
                        const idl = seg.ideal[iSig];
                        if ( idl && (idl.classAt(0) >= 0) ) {
                            sampled[iSig] = qubx_idl.dwellsToTrace(idl.getDwellsFLC(0, points.length-1), idl.amp);
                        }
                    });
                }
                const points = sampled[0],
                      fit = seg.fit,
                      pointLL = seg.pointLL,
                      nr = points.length,
                      nc = sampled.length;
                let r = 0;
                for (; r<nr; ++r) {
                    result += '\n'+(t0+r*dt)+sep+points[r];
                    let c = 1;
                    for (; c<nc; ++c) {
                        result += sep+sampled[c][r];
                    }
                    if ( haveFit ) {
                        result += sep+fit[2*r]+sep+fit[2*r+1]+sep+pointLL[r];
                    }
                }
                result += '\n';
                t0 += nr*dt;
            }
            return result;
        }
        getDWT(withStats) {
            const dtMS = 1e3 * this.sampling(),
                  digitsT = Math.max(0, 1-Math.floor(Math.log10(dtMS))),
                  digitsI = 8;
            let startMS = 0.0;
            return this.segments.map((seg, iSeg) => {
                const idl = seg.ideal[0].getDwellsCD(0, seg.n-1, null, null, dtMS),
                      amp = seg.ideal[0].amp || [],
                      std = seg.ideal[0].std,
                      Nd = idl.classes.length,
                      start1 = startMS;
                startMS += Nd * dtMS;
                let body = '';
                if ( withStats ) {
                    let sample = 0,
                        i = 0;
                    for (; i<Nd; ++i) {
                        const dur = idl.durations[i],
                              count = Math.round(dur/dtMS)|0,
                              stats = qubx.meanstd(seg.sampled[0].slice(sample, sample+count-1));
                        sample += count;
                        body += idl.classes[i]+'\t'+dur.toFixed(digitsT)+'\t'+stats.mean.toPrecision(digitsI)+'\t'+stats.std.toPrecision(digitsI)+'\n';
                    }
                } else {
                    let i = 0;
                    for (; i<Nd; ++i) {
                        body += idl.classes[i]+'\t'+idl.durations[i].toFixed(digitsT)+'\n';
                    }
                }
                return 'Segment: '+(iSeg+1)+' Dwells: '+Nd+' Sampling(ms): '+dtMS.toFixed(digitsT)+' Start(ms): '+start1.toFixed(digitsT)+' ClassCount: '+amp.length+' ' +
                    amp.map((a,c) => {
                        return a.toPrecision(digitsI)+' '+std[c].toPrecision(digitsI);
                    }).join(' ') + '\n' +
                    body;
            }).join('\n');
        }
        eachChunk(iSeg, iSignal, chunkSize, callback, idealized) {
            iSeg = (typeof(iSeg) === 'undefined') ? -1 : (iSeg|0);
            iSignal = iSignal|0;
            let maxN = 0;
            this.segments.forEach(seg => {
                maxN = Math.max(maxN, seg.n);
            });
            chunkSize = Math.min(chunkSize|0, maxN) || maxN;
            const buffers = qubx.memoize1(n => new Float32Array(n));
            let s = (iSeg < 0) ? 0 : iSeg,
                d = 0;
            const doNext = () => {
                if ( this.segments.length === 0 ) {
                    return false;
                }
                let seg = this.segments[s];
                if ( d === seg.n ) {
                    if ( (iSeg >= 0) || (s === (this.segments.length-1)) ) {
                        return true;
                    } else {
                        ++s;
                        seg = this.segments[s];
                        d = 0;
                    }
                }
                const n = Math.min(chunkSize, seg.n - d);
                if ( idealized ) {
                    const nd = seg.ideal[iSignal].countDwells(d, d+n),
                          idl = {
                              firsts: new Int32Array(nd),
                              lasts: new Int32Array(nd),
                              classes: new Int32Array(nd)
                          };
                    seg.ideal[iSignal].getDwellsFLC(d, d+n-1, idl.firsts, idl.lasts, idl.classes);
                    d += n;
                    return Promise.resolve(callback(idl, s))
                        .then(doNext);
                } else {
                    const buf = buffers(n);
                    let i = 0;
                    for (; i<n; ++i, ++d) {
                        buf[i] = seg.sampled[iSignal][d];
                    }
                    return Promise.resolve(callback(buf, s))
                        .then(doNext);
                }
            };
            return Promise.resolve(true)
                .then(doNext)
                .catch(qubx.reportErr);
        }
    };


    return exports;
});
