

define('qubx_data_axg', ['qubx'], qubx => {
    'use strict';

    const typedefStruct = qubx.typedefStruct;
    const exports = {
        // file format id's
        kAxoGraph_Graph_Format: 1,
        kAxoGraph_Digitized_Format: 2,
        kAxoGraph_X_Format: 6,
        kAxoGraph_X_Digitized_Format: 6,
        
        kAxoGraph4DocType: 'AxGr', // ?
        kAxoGraphXDocType: 'axgx', // .axgx
        
        // column header for AxoGraph graph files
        ColumnHeader: typedefStruct([
            ['i32', 'points'],
            ['p[80]', 'title']
        ]),

        // x-axis column header for AxoGraph digitized files
        DigitizedFirstColumnHeader: typedefStruct([
            ['i32', 'points'],
            ['p[80]', 'title'],
            ['f32', 'firstPoint'],
            ['f32', 'sampleInterval']
        ]),

        // y-axis column header for AxoGraph digitized files
        DigitizedColumnHeader: typedefStruct([
            ['i32', 'points'],
            ['p[80]', 'title'],
            ['f32', 'scalingFactor']
        ]),
        
        // column header for AxoGraph X files
        AxoGraphXColumnHeader: typedefStruct([
            ['i32', 'points'],
            ['i32', 'dataType'],
            ['i32', 'titleLength']
        ]),

        // translator's note: interpreting 'long' as 'i32' and 'bool' as 'u8'
        // also, I can't find example usage for readTraceHeaders -- where in the file are they?
        AxoGraphXTraceHeader: typedefStruct([
            ['i32', 'nHeaderVersion'], // header version number (currently = 2)
            ['i32', 'nColumnX'], // X column number
            ['i32', 'nColumnY'], // Y column number
            ['i32', 'nError'],  // Error bar column number or -1 if no error bars
            ['i32', 'nErrorBarColumn'], // Negative error bar column number or -1 if no negative error bars

            ['i32', 'nGroup'], // Group number that this column belongs to
            ['u8', 'isShown'], // Trace shown? False if trace is hidden

            ['f64', 'dMinX'], //      Minimum X data point in this trace
            ['f64', 'dMaxX'], //    Maximum X data point in this trace (if both are zero, they will be recalculated)
            ['f64', 'dMinPosX'], //     Minimum positive X data point in this trace (used in log-axis format)

            ['u8', 'bXEqSpaced'], //       True if X axis data is regularly spaced
            ['u8', 'bXMonotonic'], //      True if X axis data is monotonic (each point > previous point)
            ['f64', 'dXInterval'], //    Interval between points for regular X axis data

            ['f64','dMinY'], //     Minimum Y data point in this trace
            ['f64', 'dMaxY'], //     Maximum Y data point in this trace (if both are zero, they will be recalculated)
            ['f64', 'dMinPosY'], //     Minimum positive Y data point in this trace (used in log-axis format)

            ['u32', 'nRGB'], //       Trace color with RGB values serialized into a long int

            ['u8','bLinejoin'], //       True if a line plot joining the data points is displayed
            ['f64', 'dLineThickness'], //      Thickness of the line plot (can be less than 1.0 for fine lines)
            ['i32', 'nLineStyle'], //        Pen style (zero for solid line, non zero for dashed lines)

            ['u8', 'bSymbol'], //       True if symbols are displayed
            ['i32', 'nSymbolType'], //       Symbol type
            ['i32', 'nSymbolSize'], //       Symbol size (radius in pixels)

            ['u8', 'bSymbolSkip'], //       True if some symbols are to be skipped
            ['u8', 'bSymbolSkipDist'], //       True if symbols are to be skipped by distance instead of number of points
            ['i32', 'nSympolDist'], //       Minimum separation of symbols in pixes is previous parameter is true
            
            ['u8', 'bHisto'], //       True for a histogram plot
            ['i32', 'nHistoType'], //       Type of histogram (zero for standard solid fill)
            ['i32', 'nHistoDist'], //       Separation between adjacent histogram bars expressed as a percentage of bar width

            ['u8', 'bError'], //       True if error bars are displayed
            ['u8', 'bErrorPos'], //       True if a positive error bar is displayed
            ['u8', 'bErrorNeg'], //       True if a negative error bar is displayed
            ['i32', 'nErrorWidth'] //       Error bar width in pixels
        ]),

        SeriesArray: typedefStruct([
            ['f64', 'firstValue'],
            ['f64', 'increment']
        ]),

        //============= ColumnData structure ======================
        
        // This enum is copied from AxoGraph X source code
        // The only types used for data file columns are...
        //   ShortArrayType = 4     IntArrayType = 5
        //   FloatArrayType = 6     DoubleArrayType = 7
        //   SeriesArrayType = 9    ScaledShortArrayType = 10
        ColumnType: {
            IntType: 0,
            DoubleType: 1,
            BoolType: 2,
            StringType: 3,
            ShortArrayType: 4,
            IntArrayType: 5,
            FloatArrayType: 6,
            DoubleArrayType: 7,
            BoolArrayType: 8,
            SeriesArrayType: 9,
            ScaledShortArrayType: 10,
            StringArrayType: 11,
            ReferenceType: 12
        },
        
        getFileFormat: function getFileFormat(dataview) {
            let prefix = '',
                i = 0;
            for (; i<4; ++i) {
                prefix += String.fromCharCode(dataview.getUint8(i));
            }
            if ( prefix === exports.kAxoGraph4DocType ) {
                // Got an AxoGraph version 4 format file. Read the file type.
                const version = dataview.getInt16(4);
                if ( (version !== exports.kAxoGraph_Graph_Format) &&
                     (version !== exports.kAxoGraph_Digitized_Format)  ) {
                    throw new Error('AXG Version Error');
                }
                return [version, 6];
            } else if ( prefix === exports.kAxoGraphXDocType) {
                // Got an AxoGraph X format file. Check the file version.
                const version = dataview.getInt32(4);
                if ( (version < 3) || (version > exports.kAxoGraph_X_Format) ) {
                    throw new Error('AXG Version Error');
                }
                // update to latest version number
                return [exports.kAxoGraph_X_Format, 8];
            } else {
                throw new Error('AXG Format Error');
            }
        },
        getNumberOfColumns: function getNumberOfColumns(dataview, offset, fileFormat) {
            if ( (fileFormat === exports.kAxoGraph_Digitized_Format) || (fileFormat === exports.kAxoGraph_Graph_Format) ) {
                // Read the number of columns (short integer in AxoGraph 4 files)
                return [dataview.getInt16(offset), offset+2];
            } else if ( fileFormat === exports.kAxoGraph_X_Format ) {
                // Read the number of columns (long integer in AxoGraph X files)
                return [dataview.getInt32(offset), offset+4];
            } else {
                throw new Error('Unknown format');
            }
        },
        readColumn: function readColumn(dataview, offset, fileFormat, columnNumber) {
            const columnData = {
                points: 0,
                title: ''
            };
            switch ( fileFormat ) {
            case exports.kAxoGraph_Graph_Format: {
                // Read the standard column header
                const columnHeader = new exports.ColumnHeader(dataview, offset);
                offset += columnHeader.byteLength;

                // Retrieve the title and number of points in the column
                columnData.type = exports.ColumnType.FloatArrayType;
                columnData.points = columnHeader.points;
                columnData.title = columnHeader.title;

                // create a new pointer to receive the data
                const n = columnHeader.points;
                columnData.floatArray = new Float32Array(n);
                let i = 0;
                for (; i<n; ++i, offset += 4) {
                    columnData.floatArray[i] = dataview.getFloat32(offset);
                }
                break;
            } case exports.kAxoGraph_Digitized_Format: {
                if ( columnNumber === 0 ) {
                    // Read the column header
                    const columnHeader = new exports.DigitizedFirstColumnHeader(dataview, offset);
                    offset += columnHeader.byteLength;

                    // Retrieve the title, number of points in the column, and sample interval
                    columnData.type = exports.ColumnType.SeriesArrayType;
                    columnData.points = columnHeader.points;
                    columnData.title = columnHeader.title;
                    columnData.seriesArray = {
                        firstValue: columnHeader.firstPoint,
                        increment: columnHeader.sampleInterval
                    };
                } else {
                    // Read the column header
                    const columnHeader = new exports.DigitizedColumnHeader(dataview, offset);
                    offset += columnHeader.byteLength;

                    // Retrieve the title and number of points in the column
                    columnData.type = exports.ColumnType.ScaledShortArrayType;
                    columnData.points = columnHeader.points;
                    columnData.title = columnHeader.title;
                    
                    columnData.scaledShortArray = {
                        scale: columnHeader.scalingFactor,
                        offset: 0,
                        // create a new pointer to receive the data
                        shortArray: new Int16Array(columnHeader.points)
                    };

                    // Read in the column's data
                    const n = columnHeader.points;
                    let i = 0;
                    for (; i<n; ++i, offset += 2) {
                        columnData.scaledShortArray.shortArray[i] = dataview.getInt16(offset);
                    }
                }
                break;
            } case exports.kAxoGraph_X_Format: {
                // Read the column header
                const columnHeader = new exports.AxoGraphXColumnHeader(dataview, offset);
                offset += columnHeader.byteLength;
                
                // Retrieve the column type and number of points in the column
                columnData.type = columnHeader.dataType;
                columnData.points = columnHeader.points;
                
                // sanity check on column type
                if ( columnData.type < 0 || columnData.type > 14 ) {
                    throw new Error('Invalid column type: '+columnData.type);
                }
                
                // Read the column title
                columnData.titleLength = columnHeader.titleLength>>1;
                let ic = 0;
                for (; ic<columnData.titleLength; ++ic, offset += 2) {
                    const ccode = dataview.getUint16(offset);
                    if ( ccode ) {
                        columnData.title += String.fromCharCode(ccode);
                    }
                }
                
                const n = columnHeader.points;
                switch ( columnHeader.dataType ) {
                case exports.ColumnType.ShortArrayType: {
                    // create a new pointer to receive the data
                    columnData.shortArray = new Int16Array(n);
                    // Read in the column's data
                    let i = 0;
                    for (; i<n; ++i, offset += 2) {
                        columnData.shortArray[i] = dataview.getInt16(offset);
                    }
                    break;
                } case exports.ColumnType.IntArrayType: {
                    // create a new pointer to receive the data
                    columnData.intArray = new Int32Array(n);
                    // Read in the column's data
                    let i = 0;
                    for (; i<n; ++i, offset += 4) {
                        columnData.intArray[i] = dataview.getInt32(offset);
                    }
                    break;
                } case exports.ColumnType.FloatArrayType: {
                    // create a new pointer to receive the data
                    columnData.floatArray = new Float32Array(n);
                    // Read in the column's data
                    let i = 0;
                    for (; i<n; ++i, offset += 4) {
                        columnData.floatArray[i] = dataview.getFloat32(offset);
                    }
                    break;
                } case exports.ColumnType.DoubleArrayType: {
                    // create a new pointer to receive the data
                    columnData.doubleArray = new Float64Array(n);
                    // Read in the column's data
                    let i = 0;
                    for (; i<n; ++i, offset += 8) {
                        columnData.doubleArray[i] = dataview.getFloat64(offset);
                    }
                    break;
                } case exports.ColumnType.SeriesArrayType: {
                    const seriesParameters = new exports.SeriesArray(dataview, offset);
                    offset += seriesParameters.byteLength;
                    columnData.seriesArray = {
                        firstValue: seriesParameters.firstValue,
                        increment: seriesParameters.increment
                    };
                    break;
                } case exports.ColumnType.ScaledShortArrayType: {
                    columnData.scaledShortArray = {
                        scale: dataview.getFloat64(offset),
                        offset: dataview.getFloat64(offset+8),
                        // create a new pointer to receive the data
                        shortArray: new Int16Array(n)
                    };
                    offset += 16;
                    // Read in the column's data
                    let i = 0;
                    for (; i<n; ++i, offset += 2) {
                        columnData.scaledShortArray.shortArray[i] = dataview.getInt16(offset);
                    }
                    break;
                }
                }
                break;
            } default: {
                return [columnData, offset];
            }
            }
            columnData.nextOffset = offset;
            return [columnData, offset];
        },
        readComment: function readComment(dataview, offset) {
            let comment_size = 0;
            try {
                comment_size = dataview.getInt32(offset)>>1;
            } catch (err) {
                return ['', offset];
            }
            offset += 4;
            let s = '';
            if ( comment_size > 0 ) {
                try {
                    let i = 0;
                    for (; i<comment_size; ++i, offset += 2) {
                        const ccode = dataview.getUint16(offset);
                        if ( ccode ) {
                            s += String.fromCharCode(ccode);
                        }
                    }
                } catch (err) {}
            }
            return [s, offset];
        },
        readNotes: function readNotes(dataview, offset) {
            let notes_size = 0;
            try {
                notes_size = dataview.getInt32(offset)>>1;
            } catch (err) {
                return ['', offset];
            }
            offset += 4;
            let s = 0;
            if ( notes_size ) {
                try {
                    let i = 0;
                    for (; i<notes_size; ++i, offset += 2) {
                        const ccode = dataview.getUint16(offset);
                        if ( ccode ) {
                            s += String.fromCharCode(ccode);
                        }
                    }
                } catch (err) {}
            }
            return [s, offset];
        },
        readTraceHeaders: function readTraceHeaders(dataview, offset) {
            const num_headers = dataview.getInt32(offset),
                  headers = [];
            offset += 4;
            let i = 0;
            for (; i<num_headers; ++i) {
                headers.push(new exports.AxoGraphXTraceHeader(dataview, offset).copy());
            }
            return [headers, offset];
        },
        readFloatColumn: function readFloatColumn(dataview, offset, fileFormat, columnNumber) {
            var columnData;
            [columnData, offset] = exports.readColumn(dataview, offset, fileFormat, columnNumber);

            // If necessary, convert the columnData to FloatArrayType
            switch ( columnData.type ) {
            case exports.ColumnType.ShortArrayType: {
                // Convert in the column data
                columnData.floatArray = new Float32Array(columnData.shortArray.length);
                arrcpy(columnData.floatArray, columnData.shortArray);
                delete columnData.shortArray;
                columnData.type = exports.ColumnType.FloatArrayType;
                return [columnData, offset];
            } case exports.ColumnType.IntArrayType: {
                // Convert in the column data
                columnData.floatArray = new Float32Array(columnData.intArray.length);
                arrcpy(columnData.floatArray, columnData.intArray);
                delete columnData.intArray;
                columnData.type = exports.ColumnType.FloatArrayType;
                return [columnData, offset];
            } case exports.ColumnType.FloatArrayType: {
                // Don't need to convert
                return [columnData, offset];
            } case exports.ColumnType.DoubleArrayType: {
                // Convert in the column data
                columnData.floatArray = new Float32Array(columnData.doubleArray.length);
                arrcpy(columnData.floatArray, columnData.doubleArray);
                delete columnData.doubleArray;
                columnData.type = exports.ColumnType.FloatArrayType;
                return [columnData, offset];
            } case exports.ColumnType.SeriesArrayType: {
                // create a new pointer to receive the converted data
                const firstValue = columnData.seriesArray.firstValue,
                      increment = columnData.seriesArray.increment;
                columnData.floatArray = new Float32Array(columnData.points);
                // Convert in the column data
                let i = 0;
                for (; i<columnData.points; ++i) {
                    columnData.floatArray[i] = firstValue + i*increment;
                }
                columnData.type = exports.ColumnType.FloatArrayType;
                return [columnData, offset];
            } case exports.ColumnType.ScaledShortArrayType: {
                // create a new pointer to receive the converted data
                const scale = columnData.scaledShortArray.scale,
                      off = columnData.scaledShortArray.offset;
                columnData.floatArray = new Float32Array(columnData.points);
                // Convert in the column data
                let i = 0;
                for (; i<columnData.points; ++i) {
                    columnData.floatArray[i] = columnData.scaledShortArray.shortArray[i] * scale + off;
                }
                // free old short array
                delete columnData.scaledShortArray.shortArray;
                columnData.type = exports.ColumnType.FloatArrayType;
                return [columnData, offset];
            } default: {
                return [columnData, offset];
            }
            }
        },

        importAXGFile: function(arraybuffer) {
            const dataview = new DataView(arraybuffer);
            var [fileFormat, offset] = exports.getFileFormat(dataview);
            var numberOfColumns;
            [numberOfColumns, offset] = exports.getNumberOfColumns(dataview, offset, fileFormat);
            if ( numberOfColumns <= 0 ) {
                throw new Error('File format error: number of columns is negative in AxoGraph data file');
            }

            const section_list = [],
                  channel_names = [],
                  channel_units = [],
                  ReturnData = {
                      channels: []
                  };
            let numberOfChannels = 0,
                xscale = 1.0,
                columnNumber = 0;
            for (; columnNumber < numberOfColumns; ++columnNumber) {
                let column;
                [column, offset] = exports.readFloatColumn(dataview, offset, fileFormat, columnNumber);
                if ( column.seriesArray && (columnNumber === 0) ) {
                    xscale = column.seriesArray.increment * 1.0e3;
                } else {
                    section_list.push({
                        points: column.points,
                        title: column.title
                    });
                    const last = section_list.length - 1;
                    
                    if ( column.points < 1 ) {
                        throw new Error("number of points too small");
                    }
                    if ( column.floatArray.length < column.points ) {
                        throw new Error("floatArray too small in importAXGFile()");
                    }
                    if ( section_list[last].points !== column.floatArray.length ) {
                        throw new Error("section too small in importAXGFile()");
                    }
                    section_list[last].data = column.floatArray;
                    // check whether this is a new channel:
                    let isnew = true;
                    // test whether this name has been used before:
                    let n_c = 0;
                    for (; n_c < channel_names.length; ++n_c) {
                        if ( column.title == channel_names[n_c] || column.title.indexOf("Column")===0 ) {
                            isnew = false;
                            break;
                        }
                    }
                    if ( isnew ) {
                        numberOfChannels++;
                        let units = column.title,
                            left = units.lastIndexOf("(") + 1,
                            right = units.lastIndexOf(")"),
                            yunits = units.substring(left, right);
                        channel_units.push( yunits );
                        channel_names.push( column.title );
                    }
                }
            }
            // Distribute Sections to Channels:
            const sectionsPerChannel = ((numberOfColumns-1) / numberOfChannels)|0;
            let c = 0;
            for (; c<numberOfChannels; ++c) {
                const TempChannel = {
                    sections: []
                };
                //Channel TempChannel(sectionsPerChannel);
                let factor = 1.0;
                if (channel_units[c] === "V") {
                    channel_units[c] = "mV";
                    factor = 1.0e3;
                }
                if (channel_units[c] === "A") {
                    channel_units[c] = "pA";
                    factor = 1.0e12;
                }
                let n_s = c;
                for (; n_s < numberOfColumns - 1; ++n_s) {
                    if (factor !== 1.0) {
                        vec_scal_mul(section_list[n_s].data, factor);
                    }
                    TempChannel.sections[((n_s-c)/numberOfChannels)|0] = section_list[n_s];
                }
                TempChannel.channelName = channel_names[c];
                TempChannel.yUnits = channel_units[c];
                ReturnData.channels.push(TempChannel);
            }
            ReturnData.xScale = xscale;
            [ReturnData.comment, offset] = exports.readComment(dataview, offset);
            [ReturnData.notes, offset] = exports.readNotes(dataview, offset);

            return ReturnData;
        },

        AXGReader: class AXGReader {
            // implements qubx_data.Reader interface
            constructor(arraybuffer) {
                this.axg = exports.importAXGFile(arraybuffer);
                this.sampling = this.axg.xScale * 1e-3;
                this.counts = this.axg.channels[0].sections.map(s => s.data.length);
                this.maxN = Math.max(...this.counts);
                this.signals = this.axg.channels.map(ch => ({
                    name: ch.channelName,
                    units: ch.yUnits
                }));
                this.signalNames = this.signals.map(s => s.name);
                this.isSampled = this.signalNames.map(x => true);
                this.constants = {};
                /// anything useful in this.axg.notes?
            }
            readSegments(onseg) {//onseg(this, {buffers, n})
                return this.counts.map((n,i) => this.readSegment(i, onseg));
            }
            readSegment(s, onseg) {
                return onseg(this, {
                    buffers: this.axg.channels.map(ch => ch.sections[s].data),
                    n: this.counts[s]
                }, s);
            }
        }
    };

    function arrcpy(dest, src) {
        const n = Math.min(dest.length, src.length);
        let i = 0;
        for (; i<n; ++i) {
            dest[i] = src[i];
        }
    }

    function vec_scal_mul(arr, factor) {
        const n = arr.length;
        let i = 0;
        for (; i<n; ++i) {
            arr[i] *= factor;
        }
        return arr;
    }

    
    return exports;
});
