/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_data_work", ["qubx", "qubx_workers"], (qubx, qubx_workers) => {
    'use strict';

    const exports = {};

    const DataWorkgroupDefs = {
        count: 3
    };
    exports.DataWorkgroup = class DataWorkgroup {
        constructor(config) {
            this.config = Object.assign({}, DataWorkgroupDefs, config);
            this.transfer = null;
            this._workers = new qubx.PromisedStock(() => new exports.DataWorker(this.transfer), 0, w => w.destroy());
            this.count(this.config.count);
        }
        destroy() {
            this.count(0);
        }
        count(c) {
            if ( typeof(c) === 'undefined' ) {
                return this._workers.instances();
            }
            if ( c === this._workers.instances() ) {
                return this;
            }
            this.config.count = c;
            this._workers.instances(c);
            this.transfer && this.transfer.stock.instances(2*c);
            return this;
        }
        doWork(f) {
            return this._workers.get()
                .then(worker => f(worker)
                      .then(result => {
                          this._workers.put(worker);
                          return result;
                      }));
        }
        reserve(n) {
            if ( (! this.transfer) || (this.transfer.bufferLength < n) ) {
                this.transfer && this.transfer.flush();
                this.transfer = new qubx.PromisedBuffers(1, n, 4, 2*this.count());
            }
            const t = this.transfer;
            this._workers.items.forEach(worker => {
                worker.transfer = t;
            });
        }
        resample(arr, sampling, delta) {
            this.reserve(arr.length);
            return this.doWork(w => w.resample(arr, sampling, delta));
        }
        resampleScan(arr, onprogress) {
            this.reserve(arr.length);
            return this.doWork(w => w.resampleScan(arr, onprogress));
        }
        idealizeMDL(arr, sampling, delta, noSplit3, chunkSize, singlePrecision, onprogress) {
            this.reserve(arr.length);
            return this.doWork(w => w.idealizeMDL(arr, sampling, delta, noSplit3, chunkSize, singlePrecision, onprogress));
        }
        AMP(arr, sampling, model, autoInit, autoInitDown) {
            this.reserve(arr.length);
            return this.doWork(w => w.AMP(arr, sampling, model, autoInit, autoInitDown));
        }
        getFilterOverlap(sampling, freqKHz) {
            return exports.DataWorker.prototype.getFilterOverlap(sampling, freqKHz);
        }
        filter(arr, sampling, freqKHz, dest) {
            this.reserve(arr.length);
            return this.doWork(w => w.filter(arr, sampling, freqKHz, dest));
        }
    };
    
    exports.DataWorker = class DataWorker {
        constructor(transfer) {
            this.worker = new qubx_workers.Worker({
                href: 'js/qubx_data_worker.js'
                // hrefMin: 'js/qubx_data_worker.min.js',
            })
                .method('stop');
            this.transfer = transfer;
        }
        destroy() {
            return this.worker.stop();
        }
        resample(arr, sampling, delta) {
            return this.postData('resample', arr, sampling, {
                delta: delta
            }).then(data => data.resampled);
        }
        resampleScan(arr, onprogress) {
            const n = arr.length;
            return this.postData('resampleScan', arr, 1.0, {}, onprogress).then(data => {
                const params = data.params;
                function getDeltaForFrac(frac) {
                    const L = Math.log(frac),
                          tmp = L/params.a + 1,
                          logdelta = (tmp > 0) ? (params.x0 - Math.log(1/tmp - 1)/params.k) : 0;
                    return Math.exp(logdelta);
                }
                getDeltaForFrac.params = params;
                /// if the scan takes more than (target=.05) ms/pt (decent laptop speed):
                //    report available fraction of target speed to help initialize fast/accurate slider
                const targetMsPerPt = .05;
                getDeltaForFrac.cpuStrength = targetMsPerPt / Math.max(targetMsPerPt, data.runtime/n);
                return getDeltaForFrac;
            });
        }
        idealizeMDL(arr, sampling, delta, noSplit3, chunkSize, singlePrecision, onprogress) {
            return this.postData('idealizeMDL', arr, sampling, {
                combineClose: (typeof(delta) !== 'undefined'),
                combineDelta: delta,
                noSplit3: !! noSplit3,
                chunkSize: chunkSize,
                singlePrecision: singlePrecision
            }, onprogress).then(data => data.idealized);
        }
        AMP(arr, sampling, model, autoInit, autoInitDown) {
            return this.postData('AMP', arr, sampling, {
                model: model,
                autoInit: !!autoInit,
                autoInitUp: !autoInitDown
            }).then(data => data.AMP);
        }
        getFilterOverlap(sampling, freqKHz) {
            return Math.ceil(2/(sampling*1e3*freqKHz))|0;
        }
        filter(arr, sampling, freqKHz, dest) {
            const out = dest || arr,
                  overlap = this.getFilterOverlap(sampling, freqKHz);
            return this.postData('FILTER', arr, sampling, {
                n: arr.length,
                freqKHz: freqKHz,
            }, undefined, buffers => {
                const filtered = new Float32Array(buffers[0], 0, arr.length);
                if ( out.length === (arr.length - 2*overlap) ) {
                    let i = 0;
                    for (; i<out.length; ++i) {
                        out[i] = arr[overlap+i];
                    }
                } else {
                    out.set(filtered);
                }
            }).then(data => out);
        }
        postData(verb, arr, sampling, msgdata, onprogress, onreturn) {
            // pray the workgroup has sized it correctly // this.reserve(arr.length);
            msgdata = msgdata || {}
            msgdata.verb = verb;
            msgdata.sampling = sampling;
            msgdata.n = arr.length;
            return this.transfer.get().then(buffers => {
                const dest = new Float32Array(buffers[0]);
                if ( arr.copyWithin ) {
                    dest.set(arr);
                } else {
                    const n = arr.length;
                    let i = 0;
                    for (; i<n; ++i) {
                        dest[i] = arr[i];
                    }
                }
                msgdata.buffers = buffers;
                msgdata.transfer = buffers;
                return this.worker.request(msgdata, msgdata.transfer, onprogress);
            }).then(data => {
                onreturn && onreturn(data.buffers);
                this.transfer.put(data.buffers);
                data.buffers = data.transfer = null;
                return data;
            }).catch(err => {
                if ( err && err.tkt ) {
                    this.transfer.put(err.buffers);
                    return Promise.reject(err.tktErr || err);
                } else {
                    this.transfer.stock.reportLost();
                    return Promise.reject(err);
                }
            });
        }
    };
    
    return exports;
});

