/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"use strict";

let earlyOnes = [];
const earlyListener = event => earlyOnes.push(event.data);
self.addEventListener('message', earlyListener);

const QUBX_LIB_MEMORY = 4*16*1024*1024;

importScripts('qubx_worker_require.js');
//importScripts('qubx_worker_require.min.js');

qubx_require(["qubx_fit", "qubx_lib", "qubx_worker_util", "qubx_xmeans_math"]).then(libs => {
    const [qubx_fit, lib, qubx_worker_util, qubx_xmeans_math] = libs;
    const Sigmoid = qubx_fit.makeCurve({
        name: 'Sigmoid',
        info: 'y = a*(1/(1+exp(k*(x0-x))) - 1)',
        // y = log(n resampled/ntot) = f(delta)
        // x = x0 - log(1/((y/a) + 1) - 1)/k
        params: [
            {
                name: 'a',
                val: 1.0,
                constant: true,
                // lo: undefined,
                // hi: undefined
            },
            {
                name: 'x0',
                val: 0.0
            },
            {
                name: 'k',
                val: 1.0
            }
        ]
    }, (xx, yy, p) => {
        const a = p.a.val,
              k = p.k.val,
              x0 = p.x0.val,
              n = xx.length;
        let i = 0;
        for (; i<n; ++i) {
            yy[i] = a*(1.0/(1.0+Math.exp(k*(x0-xx[i]))) - 1.0);
        }
    });
    
    function arraymin(arr) {
        const n = arr.length;
        let m = arr.length ? arr[0] : undefined,
            i = 1;
        for (; i<n; ++i) {
            const x = arr[i];
            if ( m > x ) {
                m = x;
            }
        }
        return m;
    }
    function arraymax(arr) {
        const n = arr.length;
        let m = arr.length ? arr[0] : undefined,
            i = 1;
        for (; i<n; ++i) {
            const x = arr[i];
            if ( m < x ) {
                m = x;
            }
        }
        return m;
    }
    
    function mergeChunks(chunks, Nclass) {
        if ( chunks.length === 1 ) {
            return chunks[0];
        }

        // build a table of all clusters' classes' amps and stds
        // and use K-Means to group them into Nclass output classes
        Nclass = Nclass || (chunks.length ? chunks[0].amps.length : 0);/////
        let Ndwell = 0,
            ampCol = [],
            stdCol = [],
            weight = [],
            initColor = [],
            idlSrc = [],
            tableSrc = [],
            ampLo = 1e20,
            ampHi = -1e20;
        chunks.forEach(chunk => chunk.amps.forEach(a => {
            ampLo = Math.min(a, ampLo);
            ampHi = Math.max(a, ampHi);
        }));
        const ampD = ampHi - ampLo;
        chunks.forEach((chunk, iChunk) => {
            Ndwell += chunk.classes.length;
            const occCount = chunk.amps.map(() => 0),
                  n = chunk.durations.length;
            let i = 0;
            for (; i<n; ++i) {
                occCount[chunk.classes[i]] += chunk.durations[i];
            }
            const tableIndex = [];
            for (i=0; i<chunk.amps.length; ++i) {
                tableIndex.push(ampCol.length);
                ampCol.push(chunk.amps[i]);
                stdCol.push(chunk.stds[i]);
                weight.push(occCount[i]);
                initColor.push(((chunk.amps[i] - ampLo) * Nclass / ampD)|0);
                idlSrc.push({
                    iChunk: iChunk,
                    iClass: i
                });
            }
            tableSrc.push(tableIndex);
        });
        const xmeans = new qubx_xmeans_math.Calc();
        xmeans.columns = [ampCol, stdCol];
        xmeans.weight = weight;
        let clusters = xmeans.Iterate(Nclass, initColor, 20*ampCol.length);
        // now we have clusters.color[of each (chunk,cls)]
        // remap colors to have clusters.mean[0][k] increasing in "bigness"
        // (renumbering clusters.color)
        let biggestAbs = 0,
            biggestK = -1;
        const colors = clusters.mean[0].slice(1).map((amp, k) => {
            const absAmp = Math.abs(amp);
            if ( absAmp > biggestAbs ) {
                biggestAbs = absAmp;
                biggestK = k;
            }
            const votes = [];
            let i = 0;
            for (; i<Nclass; ++i) {
                votes[i] = 0;
            }
            let maxVotes = 0;
            let maxVotesC = -1;
            // look through ampstd table [chunk, c] where color=k+1
            // and increment votes[c] (tracking maxVotes, maxVotesC)
            clusters.color.forEach((color, i) => {
                if ( color === (k+1) ) {
                    const c = idlSrc[i].iClass;
                    votes[c] += 1;
                    if ( votes[c] > maxVotes ) {
                        maxVotes = votes[c];
                        maxVotesC = c;
                    }
                }
            });
            return {
                amp: amp,
                std: clusters.mean[1][k+1],
                k: k+1,
                votedK: maxVotesC
            };
        });
        // sort them by their voted class number, try to fulfill if possible
        colors.sort((a, b) => a.votedK - b.votedK);
        const remap = [],
            overflow = [],
              taken = [];
        let lastK = -1;
        colors.forEach((c,i) => {
            if ( (lastK + 1) === Nclass ) {
                overflow.push(c.k);
            } else if ( c.votedK <= lastK ) {
                remap[c.k] = lastK + 1;
            } else {
                remap[c.k] = c.votedK;
            }
            lastK = remap[c.k];
            taken[lastK] = true;
        });
        let nextK = 0;
        while ( overflow.length && (nextK < Nclass) ) {
            if ( ! taken[nextK] ) {
                remap[overflow.splice(0, 1)[0]] = nextK;
            }
            ++nextK;
        }
        clusters.color = clusters.color.map(k => remap[k]);
        
        // fill out amp, std
        const amps = new Float64Array(Nclass),
              stds = new Float64Array(Nclass);
        let k = 0;
        for (; k<Nclass; ++k) {
            amps[k] = colors[k].amp;
            stds[k] = colors[k].std;
        }

        // transcribe durations and classes
        //   replacing (k=classes[i]) with clusters.color[tableSrc[iChunk][k]]
        const durations = new Float32Array(Ndwell),
              classes = new Int32Array(Ndwell);
        let at = 0,
            lastC = -1;
        chunks.forEach((chunk, iChunk) => {
            const n = chunk.durations.length;
            let i = 0;
            for (; i<n; ++i) {
                const c = clusters.color[tableSrc[iChunk][chunk.classes[i]]]; // [Nclass*iChunk + chunk.classes[i]];
                if ( at && (c === lastC) ) {
                    durations[at-1] += chunk.durations[i];
                } else {
                    classes[at] = c;
                    durations[at] = chunk.durations[i];
                    lastC = c;
                    ++at;
                }
            }
        });
        if ( at < Ndwell ) {
            try {
                const condensed = {
                    durations: new Float32Array(at),
                    classes: new Int32Array(at),
                    amps: amps,
                    stds: stds
                };
                condensed.durations.set(durations.slice(0,at));
                condensed.classes.set(classes.slice(0,at));
                return condensed;
            } catch (err) {
                console.log(err);
            }
        }
        return {
            durations: durations,
            classes: classes,
            amps: amps,
            stds: stds
        };
    }

    qubx_worker_util.listen({
        //trace: true,
        messageQueue: earlyOnes,
        stop: data => {
            close();
        },
        resample: data => {
            const rs = new lib.AdaptiveResampler(data.sampling);
            const arr = new Float32Array(data.buffers[0], 0, data.n);
            data.resampled = rs.resample(arr, data.delta);
            rs.destroy();
            // return data; optional/implicit
        },
        resampleScan: (data, onprogress) => {
            const arr = new Float32Array(data.buffers[0], 0, data.n),
                  yrange = (arraymax(arr) - arraymin(arr)) || 1.0,
                  xfit = [];
            let ld = 1;
            for (; ld<11; ++ld) {
                xfit[ld-1] = Math.log(yrange / Math.exp(ld));
            }
            const rs = new lib.AdaptiveResampler(data.sampling),
                  tmPre = self.performance ? performance.now() : new Date(),
                  yfit = xfit.map((logdelta, idelta) => {
                      onprogress(idelta/xfit.length);
                      return Math.log(rs.resample(arr, Math.exp(logdelta)).n / arr.length);
                  }),
                  tmPost = self.performance ? performance.now() : new Date();
            rs.destroy();
            const fitPars = qubx_fit.makeParams(Sigmoid);
            fitPars.a.val = - arraymin(yfit);
            const fitLine = yfit.map(y => 0.0),
                  fitted = qubx_fit.simplexFit(Sigmoid, fitPars, xfit, yfit, fitLine);
            data.params = {
                a: fitPars.a.val,
                k: fitPars.k.val,
                x0: fitPars.x0.val
            };
            data.runtime = tmPost - tmPre;
        },
        idealizeMDL: (data, onprogress) => {
            const arr = new Float32Array(data.buffers[0], 0, data.n),
                  idealizer = new lib.MDLIdealizer(data.chunkSize || arr.length, {
                      combineClose: data.combineClose,
                      combineDelta: data.combineDelta,
                      split3: ! data.noSplit3,
                      singlePrecision: data.singlePrecision
                  });
            if ( data.chunkSize ) {
                const chunks = [];
                let progressChunk = data.chunkSize;
                while ( progressChunk < 2000 ) {
                    progressChunk *= 2;
                }
                let chunkStart = 0;
                for (; chunkStart<arr.length; chunkStart+=data.chunkSize) {
                    if ( ! (chunkStart % progressChunk) ) {
                        onprogress(chunkStart/arr.length);
                    }
                    const chunkN = Math.min(arr.length - chunkStart, data.chunkSize);
                    chunks.push(idealizer.idealize(arr.slice(chunkStart, chunkStart+chunkN), data.sampling));
                }
                data.idealized = mergeChunks(chunks, chunks[0] && Math.min(256, (chunks[0].amps.length * chunks.length)));
            } else {
                data.idealized = idealizer.idealize(arr, data.sampling);
            }
            idealizer.destroy();
        },
        AMP: data => {
            const result = lib.AMP(data.sampling, new Float32Array(data.buffers[0], 0, data.n),
                                 data.model, data.autoInit, data.autoInitUp);
            data.AMP = result;
            data.transfer = ['amp', 'std', 'counts', 'classes'].map(field => result[field].buffer)
                .concat(data.buffers);
        },
        FILTER: data => {
            const n = data.n,
                  arr = new Float32Array(data.buffers[0], 0, n);
            lib.filter(arr, null, data.sampling, 1e3*data.freqKHz);
        }
    });
    earlyOnes = [];
}).catch(err => postMessage({t: 'LOADERR', err}));
