/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_simu", ["qubx", "qubx_ui", "qubx_simulate", "qubx_simu_results", "qubx_data", "qubx_lib", "qubx_model_math"],
    (qubx, qubx_ui, qubx_simulate, qubx_simu_results, qubx_data, qubx_lib, qubx_model_math) => {
        "use strict";
        const exports = {};

        const CONSTDEF = {Ligand: 1.0, Voltage: -50.0, Pressure: 0.0};
        
        const SimModeDefs = {
            constants: {}
        };
        exports.SimMode = class SimMode{
            constructor(main, config) {
                this.main = main;
                let storedPrefs = localStorage.getItem("qubx_simu_prefs");
                storedPrefs = storedPrefs ? JSON.parse(storedPrefs) : {};
                this.config = Object.assign({}, SimModeDefs, config, storedPrefs);
                try {
                    this.stimIdl = JSON.parse(localStorage.getItem("qubx_simu_stimuli"));
                } catch (err) {
                }
                this.stimIdl = this.stimIdl || [];
                this.active = false;
                this.main.panels.then(p => {
                    this.modelView = p.modelView;
                    this.data = p.data;
                    this.controls = p.controls.querySelector("#qubx_demo_data_ctrls");
                    this._makeMenus(p);
                    this._makeDialogs(p);
                    this._makeResults(p);
                }).catch(qubx.reportErr);
            }
            activate() {
                this.main.panels.then(p => {
                    this.active = true;
                    this._modelInsertedRate = p.modelView.models.ev.insertRate.sub(this._onInsertedRate.bind(this));
                    p.modelView.model && p.modelView.model.rates.entries.forEach(this._onInsertedRate.bind(this));
                    this._modelSetRate = p.modelView.models.ev.setRate.sub(this._onSetRate.bind(this));
                    this.sim = new qubx_simulate.Simulator({models: p.modelView.models, constants: this.config.constants});
                    this.sim.ev.run.sub(this._onRun.bind(this));
                    this.sim.ev.clear.sub(this._onClear.bind(this));
                    this.sim.ev.chunk.sub(this._onChunk.bind(this));
                    let simConfig = JSON.parse(localStorage.getItem("qubx_simu_config") || "{}");
                    simConfig.constants = simConfig.constants || {};
                    for (const name of Object.keys(this.config.constants)) {
                        if ( ! (name in simConfig.constants) ) {
                            simConfig.constants[name] = this.config.constants[name];
                        }
                    }
                    this.sim.reconfig(simConfig);
                    this.menus.forEach(node => this.controls.appendChild(node));
                    this.main.vis.add(this.doseResponse);
                    this.main.vis.add(this.matrices);
                    this._dataSetConstant = this.data.data.constants.ev.set.sub(this._onSetConstant.bind(this));
                }).catch(qubx.reportErr);
            }
            deactivate() {
                return Promise.resolve(true).then(() => {
                    this.active = false;
                    this.main.vis.remove(this.matrices);
                    this.main.vis.remove(this.doseResponse);
                    this.data.busy(false);
                    this.menus.forEach(node => this.controls.removeChild(node));
                    this.savePrefs();
                    this.modelView.models.ev.insertRate.unsub(this._modelInsertedRate);
                    this.modelView.models.ev.setRate.unsub(this._modelSetRate);
                    this.data.data.constants.ev.set.unsub(this._dataSetConstant);
                    this.sim.dispose();
                });
            }
            _onInsertedRate(m, row) {
                ["Ligand", "Voltage", "Pressure"].forEach(sName => row[sName] && this.getConstant(row[sName], CONSTDEF[sName]));
            }
            _onSetRate(m, info) {
                if ( info.value && ["Ligand", "Voltage", "Pressure"].indexOf(info.column) !== -1 ) {
                    this.getConstant(info.value, CONSTDEF[info.column]);
                }
            }
            _onSetConstant(info) {
                const c = this.data.data.constants.get(info.row);
                if ( c.Name in this.modelView.modelStim ) {
                    this.config.constants[c.Name] = c.Value;
                    this.sim.reconfig({constants: this.getConstants()});
                }
            }
            _onRun(data) {
                this.active && this.data.busy(true);
            }
            _onClear(data) {
                if ( ! this.active ) { return; }
                this.main.resetModelStim();
                const reshape = () => {
                    this.main.getDeltaForFrac = null;
                    this.data.data.segmentCount(0);
                    this.data.data.signalCount(data.signalNames.length, (sig, iSig) => {
                        sig.name = data.signalNames[iSig];
                        sig.units = '';
                    });
                    this.data.data.segmentCount(data.segmentSize.length);
                    data.segmentSize.forEach((sz, i) => {
                        this.data.data.segmentSize(i, sz);
                    });
                    this.data.data.doneSegmentSize();
                };
                this.data.data.sampling(data.sampling);
                if ( (data.segmentSize.length !== this.data.data.segments.length) || (data.signalNames.length != this.data.data.signals.length) ) {
                    reshape();
                } else {
                    let i = 0;
                    for (; i<data.segmentSize.length; ++i) {
                        if ( data.segmentSize[i] !== this.data.data.segments[i].n ) {
                            reshape();
                            break;
                        }
                    }
                }
                this.data.data.clear();
                [this.matrices, this.doseResponse].forEach(results => {
                    results && results.onClear(data.sampling, this.config.constants, data.classAmp, data.IeqFv, this.config.prefRgns);
                });
                this.data.data.constants.clear();
                for (const cname of Object.keys(this.config.constants)) {
                    this.data.data.constants.append({
                        Name: cname,
                        Value: this.config.constants[cname]
                    });
                }
                this.main.txtTdead.update(this.data.data.tDead());
            }
            _onChunk(data) {
                if ( ! this.active ) { return; }
                data.buffers.forEach((buf, iSig) => {
                    if ( iSig < this.data.data.signalCount() ) {
                        this.data.data.setSampled(data.iSeg, iSig, data.first, new Float32Array(buf, 0, data.n), data.n, 0, true);
                    }
                }); /// clear stale idealizations (onClear?) and fit curves?
                data.classes && data.classes.length && this.data.data.setIdeal(data.iSeg, 0, data.counts, data.classes, data.first, data.amps, data.stds);
                const seg = this.data.data.segments[data.iSeg];
                seg.indVar = data.indVar;
                
                if ( (data.first + data.n) === data.total ) {
                    this.data.busy(false);
                    this.applyStimIdl();
                }
            }
            prepareStimIdl(stimuli) {
                this.stimIdl = stimuli;
                localStorage.setItem("qubx_simu_stimuli", JSON.stringify(stimuli));
            }
            applyStimIdl() {
                this.data.data.signals.forEach((sig, iSig) => {
                    if ( iSig ) {
                        let iStim = 0;
                        for (; iStim<this.stimIdl.length; ++iStim) {
                            if ( this.stimIdl[iStim].name === sig.name ) {
                                this.stimIdl[iStim].segments.forEach((seg, iSeg) => {
                                    if ( iSeg >= this.data.data.segments.length ) {
                                        return;
                                    }
                                    this.data.data.setIdeal(iSeg, iSig, seg.counts, seg.classes, 0, seg.amps, seg.stds);
                                });
                                break;
                            }
                        }
                    }
                });
            }
            savePrefs() {
                try {
                    const prefs = {};
                    for (const k of Object.keys(SimModeDefs)) { // to avoid e.g. config.node
                        prefs[k] = this.config[k];
                    }
                    localStorage.setItem("qubx_simu_prefs", JSON.stringify(prefs));
                    localStorage.setItem("qubx_simu_config", JSON.stringify(this.sim.config));
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            getConstant(name, defval) {
                if ( ! (name in this.config.constants) ) {
                    this.config.constants[name] = defval || 0.0;
                }
                const c = this.config.constants[name];
                if ( this.sim && ! (name in this.sim.config.constants) ) {
                    this.sim.reconfig({constants: this.getConstants()});
                }
                return c;
            }
            getConstants() {
                const stimulus = this.modelView.getStimulus();
                Object.keys(stimulus).forEach(name => {
                    stimulus[name] = this.config.constants[name];
                });
                return stimulus;
            }
            updateConstants(consts) {
                Object.keys(consts).forEach(name => {
                    this.config.constants[name] = consts[name];
                });
            }
            acceptStateIndex(s) {
                // ideally this would be a valid index in all models, but we'll settle for being valid in the first active model
                return this.modelView.active[0].model.acceptStateIndex(s);
            }
            formatStateIndex(i) {
                return this.modelView.active.length && (this.modelView.active[0].model.states.length > i) &&
                    qubx.formatState(this.modelView.active[0].model.states[i]) || (''+i);
            }
            _makeMenus(p) {
                this.mnuSetup = new qubx_ui.Menu({
                    caption: 'Simulation',
                    groups: [[
                        {
                            caption: 'Steady-State...',
                            action: this._onClickSteady.bind(this)
                        },
                        {
                            caption: 'Dose-Response...',
                            action: this._onClickDoseResponse.bind(this)
                        },
                        {
                            caption: 'Agonist-Antagonist...',
                            action: this._onClickAgonistAntagonist.bind(this)
                        },
                        {
                            caption: 'Paired Pulse...',
                            action: this._onClickPairedPulse.bind(this)
                        },
                        {
                            caption: 'Ramp...',
                            action: this._onClickRamp.bind(this)
                        }
                    ]]
                });
                this.mnuSetup.config.node.style.width = '8em';
                
                this.mnuCalc = new qubx_ui.Menu({
                    caption: 'Calculate',
                    groups: [[
                        {
                            caption: 'Mean first passage...',
                            action: this._onClickMFP.bind(this)
                        },
                        {
                            caption: 'Transition state...',
                            action: this._onClickTState.bind(this)
                        }
                    ]]
                });
                this.mnuCalc.config.node.style.width = '8em';

                this.chkAuto = document.createElement('input');
                this.chkAuto.setAttribute('type', 'checkbox');
                this.chkAuto.checked = true;
                this.chkAuto.addEventListener('change', this._onChkAuto.bind(this));
                const lblAuto = document.createElement('label');
                lblAuto.textContent = 'auto';
                lblAuto.appendChild(this.chkAuto);
                this.btnResim = document.createElement('button');
                this.btnResim.textContent = 'Re-simulate';
                this.btnResim.addEventListener('click', this._onClickResim.bind(this));
                
                this.menus = [this.mnuSetup.config.node, this.mnuCalc.config.node, lblAuto, this.btnResim];
            }
            _makeDialogs(p) {
                this.mfpDialog = new qubx_ui.Modal({
                    title: 'Mean first passage',
                    template: '.qubx_dlg_mfp',
                    okLabel: 'Close',
                    canCancel: false
                });
                this.mfpDialog.dom.then(dom => {
                    //// #mfp_index_a|b //// #mfp_ms
                    //// this.mfpDialog.mfp matrix
                    const dotdot = qubx_ui.querydotdot(dom.content);
                    this.mfpDialog.txtA = new qubx_ui.AcceptedInput(dotdot("#mfp_index_a"),
                                                                    0,
                                                                    {accept: this.acceptStateIndex.bind(this),
                                                                     format: this.formatStateIndex.bind(this),
                                                                     widthChars: 5});
                    this.mfpDialog.txtB = new qubx_ui.AcceptedInput(dotdot("#mfp_index_b"),
                                                                    1,
                                                                    {accept: this.acceptStateIndex.bind(this),
                                                                     format: this.formatStateIndex.bind(this),
                                                                     widthChars: 5});
                    this.mfpDialog.txtMS = dotdot("#mfp_ms");
                    this.mfpDialog.update = () => {
                        const a = this.mfpDialog.txtA.value,
                              b = this.mfpDialog.txtB.value,
                              Ns = this.modelView.model.states.size;
                        if ( isNaN(a) || isNaN(b) || (a < 0) || (b < 0) || (a >= Ns) || (b >= Ns) ) {
                            this.mfpDialog.txtMS.value = '';
                        } else {
                            this.mfpDialog.txtMS.value = qubx.formatK0(1e3 * this.mfpDialog.mfp[a][b]);
                        }
                    };
                    [this.mfpDialog.txtA, this.mfpDialog.txtB].forEach((txt, i) => {
                        txt.update(i);
                        txt.ev.change.sub(this.mfpDialog.update);
                    });
                }).catch(qubx.reportErr);

                this.tstateDialog = new qubx_ui.Modal({
                    title: 'Transition state (committor)',
                    template: '#qubx_tstate',
                    canCancel: false,
                    okLabel: 'Close'
                });
                this.tstateDialog.dom.then(dom => {
                    const dotdot = qubx_ui.querydotdot(dom.content);
                    const tmeas = dotdot.sub('#qubx_tstate_tmeas', node => new qubx_ui.AcceptedInput(node, 100, {
                        accept: qubx.acceptFloatGreaterThan(0.0),
                        format: qubx.formatK0,
                        widthChars: 8,
                        localKey: 'qubx_tstate_tmeas'
                    }), 'change', val => this.tstateDialog.update(val));
                    const out = dotdot('#qubx_tstate_msg');
                    this.tstateDialog.update = () => {
                        out.innerHTML = qubx_model_math.findCommitor(this.modelView.model, tmeas.value, this.modelView.dataStim).message;
                    };
                });

                this.steadyDialog = new SetupDialog({
                    title: 'Steady state...',
                    optTemplate: '.qubx_simopts_steady',
                    localKey: 'sim.steady',
                    acceptStateIndex: this.acceptStateIndex.bind(this),
                    modelView: this.modelView,
                    getConstants: this.getConstants.bind(this),
                    updateConstants: this.updateConstants.bind(this)
                });
                
                this.drDialog = new SetupDialog({
                    title: 'Dose-response...',
                    optTemplate: '.qubx_simopts_doseresponse',
                    localKey: 'sim.dr',
                    acceptStateIndex: this.acceptStateIndex.bind(this),
                    modelView: this.modelView,
                    getConstants: this.getConstants.bind(this),
                    updateConstants: this.updateConstants.bind(this),
                    setup: (dlg, dom, optDiv) => {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        this.drDialog.varname = new VarNamePicker(dotdot("#varname"));
                        this.drDialog.preDuration = new qubx_ui.AcceptedInput(dotdot("#preDuration"),
                                                                              this.drDialog.txtDuration.value/10,
                                                                              {accept: qubx.acceptFloatGreaterThan(0),
                                                                               format: qubx.formatK0,
                                                                               localKey: "sim.dr.preDur",
                                                                               widthChars: 6});
                        this.drDialog.pulseDuration = new qubx_ui.AcceptedInput(dotdot("#pulseDuration"),
                                                                                this.drDialog.txtDuration.value/4,
                                                                                {accept: qubx.acceptFloatGreaterThan(0),
                                                                                 format: qubx.formatK0,
                                                                                 localKey: "sim.dr.pulseDur",
                                                                                 widthChars: 6});
                        this.drDialog.levels = new qubx_ui.AcceptedInput(dotdot("#levels"),
                                                                         [-30, -10, 10, 30, 50],
                                                                         {accept: qubx.acceptList(qubx.acceptFloat),
                                                                          format: qubx.formatList(),
                                                                          localKey: "sim.dr.levels",
                                                                          widthChars: 15});
                        this.drDialog.filter = dotdot.listen("#filter", 'change', () => {
                            dom.content.querySelector(".qubx_dlg_dr_filter").style.visibility = this.drDialog.filter.checked ? 'visible' : 'hidden';
                        });
                        this.drDialog.filterkHz = new qubx_ui.AcceptedInput(dotdot("#filter_kHz"),
                                                                            1.0,
                                                                            {accept: qubx.acceptFloatGreaterThan(0),
                                                                             format: qubx.formatK0,
                                                                             localKey: "sim.dr.filter",
                                                                             widthChars: 6});
                    },
                    prerun: stimulus => this.drDialog.varname.update(stimulus),
                    postrun: opts => {
                        opts.varname = this.drDialog.varname.value;
                        opts.resting = opts.consts[opts.varname];
                        opts.preDur = this.drDialog.preDuration.value
                        opts.pulseDur = this.drDialog.pulseDuration.value;
                        opts.postDur = opts.dur - opts.preDur - opts.pulseDur;
                        opts.levels = this.drDialog.levels.value;
                        opts.filterkHz = this.drDialog.filter.checked ? this.drDialog.filterkHz.value : 0;
                    }
                });
                
                this.agonistDialog = new SetupDialog({
                    title: 'Agonist-antagonist...',
                    optTemplate: '.qubx_simopts_agonist',
                    localKey: 'sim.agonist',
                    acceptStateIndex: this.acceptStateIndex.bind(this),
                    modelView: this.modelView,
                    getConstants: this.getConstants.bind(this),
                    updateConstants: this.updateConstants.bind(this),
                    setup: (dlg, dom, optDiv) => {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        this.agonistDialog.agonist = new VarNamePicker(dotdot("#agonist"));
                        this.agonistDialog.antagonist = new VarNamePicker(dotdot("#antagonist"));
                        this.agonistDialog.preDuration = new qubx_ui.AcceptedInput(dotdot("#preDuration"),
                                                                                   this.agonistDialog.txtDuration.value/2,
                                                                                   {accept: qubx.acceptFloatGreaterThan(0),
                                                                                    format: qubx.formatK0,
                                                                                    localKey: "sim.agonist.preDur",
                                                                                    widthChars: 6});
                        this.agonistDialog.agonistDuration = new qubx_ui.AcceptedInput(dotdot("#agonistDuration"),
                                                                                       this.agonistDialog.txtDuration.value/4,
                                                                                       {accept: qubx.acceptFloatGreaterThan(0),
                                                                                        format: qubx.formatK0,
                                                                                        localKey: "sim.agonist.agDur",
                                                                                        widthChars: 6});
                        this.agonistDialog.antiPreDuration = new qubx_ui.AcceptedInput(dotdot("#antiPreDuration"),
                                                                                       this.agonistDialog.txtDuration.value/10,
                                                                                       {accept: qubx.acceptFloatGreaterThan(0),
                                                                                        format: qubx.formatK0,
                                                                                        localKey: "sim.agonist.antiPreDur",
                                                                                        widthChars: 6});
                        this.agonistDialog.antagonistDuration = new qubx_ui.AcceptedInput(dotdot("#antiDuration"),
                                                                                          this.agonistDialog.txtDuration.value/4,
                                                                                          {accept: qubx.acceptFloatGreaterThan(0),
                                                                                           format: qubx.formatK0,
                                                                                           localKey: "sim.agonist.antiDur",
                                                                                           widthChars: 6});
                        this.agonistDialog.agonistLevels = new qubx_ui.AcceptedInput(dotdot("#agonistLevels"),
                                                                                     [1, 1, 1, 1, 1],
                                                                                     {accept: qubx.acceptList(qubx.acceptFloat),
                                                                                      format: qubx.formatList(),
                                                                                      localKey: "sim.agonist.agLevels",
                                                                                      widthChars: 15});
                        this.agonistDialog.antagonistLevels = new qubx_ui.AcceptedInput(dotdot("#antagonistLevels"),
                                                                                        [0.1, 0.5, 1, 5, 10],
                                                                                        {accept: qubx.acceptList(qubx.acceptFloat),
                                                                                         format: qubx.formatList(),
                                                                                         localKey: "sim.agonist.antiLevels",
                                                                                         widthChars: 15});
                        this.agonistDialog.filter = dotdot.listen("#filter", "change", () => {
                            dotdot(".qubx_dlg_agonist_filter").style.visibility = this.agonistDialog.filter.checked ? 'visible' : 'hidden';
                        });
                        this.agonistDialog.filterkHz = new qubx_ui.AcceptedInput(dotdot("#filter_kHz"),
                                                                                 1.0,
                                                                                 {accept: qubx.acceptFloatGreaterThan(0),
                                                                                  format: qubx.formatK0,
                                                                                  localKey: "sim.agonist.filter",
                                                                                  widthChars: 6});
                    },
                    prerun: stimulus => {
                        this.agonistDialog.agonist.update(stimulus);
                        this.agonistDialog.antagonist.update(stimulus);
                    },
                    postrun: opts => {
                        opts.agonist = this.agonistDialog.agonist.value;
                        opts.agonistLevels = this.agonistDialog.agonistLevels.value;
                        opts.antagonist = this.agonistDialog.antagonist.value;
                        opts.antagonistLevels = this.agonistDialog.antagonistLevels.value;
                        opts.preDuration = this.agonistDialog.preDuration.value;
                        opts.agonistDuration = this.agonistDialog.agonistDuration.value;
                        opts.antiPreDuration = this.agonistDialog.antiPreDuration.value;
                        opts.antagonistDuration = this.agonistDialog.antagonistDuration.value;
                        opts.filterkHz = this.agonistDialog.filter.checked ? this.drDialog.filterkHz.value : 0;
                    }
                });
                
                this.ppulseDialog = new SetupDialog({
                    title: 'Paired pulse...',
                    optTemplate: '.qubx_simopts_pairedpulse',
                    localKey: 'sim.ppulse',
                    acceptStateIndex: this.acceptStateIndex.bind(this),
                    modelView: this.modelView,
                    getConstants: this.getConstants.bind(this),
                    updateConstants: this.updateConstants.bind(this),
                    setup: (dlg, dom, optDiv) => {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        dotdot('.qubx_sim_durbox').style.display = 'none';
                        this.ppulseDialog.varname = new VarNamePicker(dotdot("#varname"));
                        this.ppulseDialog.condLevel = new qubx_ui.AcceptedInput(dotdot("#condLevel"),
                                                                                1.0,
                                                                                {accept: qubx.acceptFloat,
                                                                                 format: qubx.formatK0,
                                                                                 localKey: "sim.ppulse.condLevel",
                                                                                 widthChars: 6});
                        this.ppulseDialog.condDuration = new qubx_ui.AcceptedInput(dotdot("#condDuration"),
                                                                                   0.1,
                                                                                   {accept: qubx.acceptFloatGreaterThan(0),
                                                                                    format: qubx.formatK0,
                                                                                    localKey: "sim.ppulse.condDur",
                                                                                    widthChars: 6});
                        this.ppulseDialog.recovery = new qubx_ui.AcceptedInput(dotdot("#recovery"),
                                                                               [0.01, 0.05, 0.1, 0.5, 1],
                                                                               {accept: qubx.acceptList(qubx.acceptFloatGreaterThan(0)),
                                                                                format: qubx.formatList(qubx.formatK0),
                                                                                localKey: "sim.ppulse.recovery",
                                                                                widthChars: 15});
                        this.ppulseDialog.testLevel = new qubx_ui.AcceptedInput(dotdot("#testLevel"),
                                                                                1.0,
                                                                                {accept: qubx.acceptFloat,
                                                                                 format: qubx.formatK0,
                                                                                 localKey: "sim.ppulse.testLevel",
                                                                                 widthChars: 6});
                        this.ppulseDialog.testDuration = new qubx_ui.AcceptedInput(dotdot("#testDuration"),
                                                                                   0.01,
                                                                                   {accept: qubx.acceptFloatGreaterThan(0),
                                                                                    format: qubx.formatK0,
                                                                                    localKey: "sim.ppulse.testDur",
                                                                                    widthChars: 6});
                        this.ppulseDialog.filter = dotdot.listen("#filter", "change", () => {
                            dotdot(".qubx_dlg_ppulse_filter").style.visibility = this.ppulseDialog.filter.checked ? 'visible' : 'hidden';
                        });
                        this.ppulseDialog.filterkHz = new qubx_ui.AcceptedInput(dotdot("#filter_kHz"),
                                                                                1.0,
                                                                                {accept: qubx.acceptFloatGreaterThan(0),
                                                                                 format: qubx.formatK0,
                                                                                 localKey: "sim.ppulse.filter",
                                                                                 widthChars: 6});
                        
                    },
                    prerun: (stimulus => this.ppulseDialog.varname.update(stimulus)),
                    postrun: opts => {
                        opts.varname = this.ppulseDialog.varname.value;
                        opts.resting = opts.consts[opts.varname];
                        opts.condLevel = this.ppulseDialog.condLevel.value;
                        opts.condDur = this.ppulseDialog.condDuration.value;
                        opts.testLevel = this.ppulseDialog.testLevel.value;
                        opts.testDur = this.ppulseDialog.testDuration.value;
                        opts.recovery = this.ppulseDialog.recovery.value;
                        opts.filterkHz = this.ppulseDialog.filter.checked ? this.agonistDialog.filterkHz.value : 0;
                    }
                });
                
                this.rampDialog = new SetupDialog({
                    title: 'Ramp...',
                    optTemplate: '.qubx_simopts_ramp',
                    localKey: 'sim.ramp',
                    acceptStateIndex: this.acceptStateIndex.bind(this),
                    modelView: this.modelView,
                    getConstants: this.getConstants.bind(this),
                    updateConstants: this.updateConstants.bind(this),
                    setup: (dlg, dom, optDiv) => {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        this.rampDialog.varname = new VarNamePicker(dotdot("#varname"));
                        this.rampDialog.preDuration = new qubx_ui.AcceptedInput(dotdot("#preDuration"),
                                                                                this.rampDialog.txtDuration.value/10,
                                                                                {accept: qubx.acceptFloatGreaterThan(0),
                                                                                 format: qubx.formatK0,
                                                                                 localKey: "sim.ramp.preDur",
                                                                                 widthChars: 6});
                        this.rampDialog.rampDuration = new qubx_ui.AcceptedInput(dotdot("#rampDuration"),
                                                                                 this.rampDialog.txtDuration.value/4,
                                                                                 {accept: qubx.acceptFloatGreaterThan(0),
                                                                                  format: qubx.formatK0,
                                                                                  localKey: "sim.ramp.rampDur",
                                                                                  widthChars: 6});
                        this.rampDialog.levelsFrom = new qubx_ui.AcceptedInput(dotdot("#rampFrom"),
                                                                               [0],
                                                                               {accept: qubx.acceptList(qubx.acceptFloat),
                                                                                format: qubx.formatList(),
                                                                                localKey: "sim.ramp.levelsFrom",
                                                                                widthChars: 8});
                        this.rampDialog.levelsTo = new qubx_ui.AcceptedInput(dotdot("#rampTo"),
                                                                             [1,2,3],
                                                                             {accept: qubx.acceptList(qubx.acceptFloat),
                                                                              format: qubx.formatList(),
                                                                              localKey: "sim.ramp.levelsTo",
                                                                              widthChars: 8});
                        this.rampDialog.rampDown = dotdot('#rampDown');
                        this.rampDialog.rampDownDuration = new qubx_ui.AcceptedInput(dotdot('#rampDownDuration'),
                                                                                     this.rampDialog.txtDuration.value/4,
                                                                                     {accept: qubx.acceptFloatGreaterThan(0),
                                                                                      format: qubx.formatK0,
                                                                                      localKey: 'sim.ramp.rampDur',
                                                                                      widthChars: 6});
                    },
                    prerun: (stimulus => (this.rampDialog.varname.update(stimulus))),
                    postrun: opts => {
                        opts.varname = this.rampDialog.varname.value;
                        opts.resting = opts.consts[opts.varname];
                        opts.preDur = this.rampDialog.preDuration.value
                        opts.rampDur = this.rampDialog.rampDuration.value;
                        opts.rampDown = this.rampDialog.rampDown.checked && this.rampDialog.rampDownDuration.value || 0.0;
                        opts.postDur = opts.dur - opts.preDur - opts.rampDur - opts.rampDown;
                        opts.levelsFrom = this.rampDialog.levelsFrom.value;
                        opts.levelsTo = this.rampDialog.levelsTo.value;
                    }
                });
            }
            _makeResults(panels) {
                this.matrices = new qubx_simu_results.SimMatrices({
                    models: panels.modelView,
                    choose: panels.data.plot.chosen.bind(panels.data.plot)
                });
                this.matrices.onClear(1.0, this.config.constants);
                this.matrices.tabTitle = 'Matrices';
                this.matrices.tabIcon = '<b>Q</b>';
                
                this.doseResponse = new qubx_simu_results.DoseResponse({
                    models: panels.modelView,
                    choose: panels.data.plot.chosen.bind(panels.data.plot)
                });
                this.doseResponse.onClear(1.0, this.config.constants);
                this.doseResponse.tabTitle = 'Dose-Response';
                this.doseResponse.tabIcon = '<img src="images/appbar.graph.line.up.svg">';
                this.doseResponse.tabOffscreen = true;
                
                const onChoose = chosen => {
                    this.matrices.onChoose(chosen);
                    this.doseResponse.onChoose(chosen);
                };
                panels.data.plot.ev.choose.sub(onChoose);
                onChoose(panels.data.plot.chosen());
                
                const onSelect = (offset, count, segments) => {
                    [this.matrices, this.doseResponse].forEach(results => {
                        results.onSelect(offset, count, segments,
                                         this.data.data.signals.map(qubx.m('name')),
                                         this.sim.config.sampling, this.config.constants);
                    });
                };
                panels.data.plot.ev.select.sub(onSelect);
            }
            _onClickMFP(event) {
                qubx_lib.then(lib => {
                    this.mfpDialog.mfp = lib.meanFirstPassage(this.modelView.info.Q, this.sim.config.sampling);
                    this.mfpDialog.update();
                    return this.mfpDialog.run()
                        .catch(qubx.reportErr);
                }).catch(qubx.reportErr);
            }
            _onClickTState(event) {
                this.tstateDialog.run()
                    .catch(qubx.reportErr);
                this.tstateDialog.update();
           }
            _onClickSteady(event) {
                this.steadyDialog.run()
                    .then(opts => this._setupSteady(opts))
                    .catch(qubx.reportErr);
            }
            _setupSteady(opts) {
                const n = isFinite(opts.dur) ? Math.ceil(opts.dur / opts.samplingSec) : 1,
                      consts = this.getConstants(),
                      constStr = Object.keys(consts).map(name => name+"="+qubx.formatK0(consts[name]));
                this.prepareStimIdl([]);
                this.sim.reconfig({
                    sampling: opts.samplingSec,
                    usePeq: opts.usePeq,
                    constants: opts.consts,
                    segmentSize: [n],
                    segmentIndVar: [{}],
                    stimulus: {},
                    about: 'Steady-state for '+qubx.formatK0(opts.dur)+' seconds at '+opts.constStr.join(", ")
                });
                if ( ! opts.usePeq ) {
                    const states = this.modelView.model.states,
                          n = states.size;
                    for (; i<n; ++i) {
                        states.set(i, "Pr", (i === opts.startState) ? 1 : 0);
                    }
                }
                this.config.prefRgns = null;
            }
            _onClickDoseResponse(event) {
                this.drDialog.run()
                    .then(opts => this._setupDoseResponse(opts))
                    .catch(qubx.reportErr);
            }
            _setupDoseResponse(opts) {
                // opts: samplingSec, dur, consts, constStr, usePeq || startState;
                // varname, resting, preDur, pulseDur, postDur, levels
                const n = isFinite(opts.dur) ? Math.ceil(opts.dur / opts.samplingSec) : 1,
                      stimulus = {};
                stimulus[opts.varname] = {
                    stimType: 'steps',
                    filter_kHz: opts.filterkHz,
                    traces: opts.levels.map(level => ({
                        levels: [opts.resting, level, opts.resting],
                        durations: [opts.preDur, opts.pulseDur, opts.postDur]
                    }))
                };
                const nPre = Math.round(opts.preDur / opts.samplingSec),
                      iPost = Math.round((opts.preDur + opts.pulseDur) / opts.samplingSec),
                      nPulse = iPost - nPre,
                      nPost = n - iPost;
                this.prepareStimIdl([// per stimulus
                    {
                        name: opts.varname,
                        segments: opts.levels.map(level => ({
                            classes: [0,1,0],
                            counts: [nPre, nPulse, nPost],
                            amps: [opts.resting, level],
                            stds: [0, 0]
                        }))
                    }
                ]);
                this.sim.reconfig({
                    sampling: opts.samplingSec,
                    usePeq: opts.usePeq,
                    constants: this.getConstants(),
                    segmentSize: opts.levels.map(() => n),
                    segmentIndVar: opts.levels.map(level => {
                        const v = {};
                        v[opts.varname] = level;
                        return v;
                    }),
                    stimulus: stimulus,
                    about: 'Dose-response with '+opts.varname+' ladder at '+opts.levels.map(qubx.formatK0).join(", ")
                });
                if ( ! opts.usePeq ) {
                    const states = this.modelView.model.states,
                          n = states.size;
                    let i = 0;
                    for (; i<n; ++i) {
                        states.set(i, "Pr", (i === opts.startState) ? 1 : 0);
                    }
                }
                this.config.prefRgns = null;
            }
            _onClickAgonistAntagonist(event) {
                this.agonistDialog.run()
                    .then(opts => this._setupAgonistAntagonist(opts))
                    .catch(qubx.reportErr);
            }
            _setupAgonistAntagonist(opts) {
                const n = isFinite(opts.dur) ? Math.ceil(opts.dur / opts.samplingSec) : 1,
                      stimulus = {},
                      defStim = (varname, pre, dur, post, resting, levels) => {
                          stimulus[varname] = {
                              stimType: 'steps',
                              filter_kHz: opts.filterkHz,
                              traces: levels.map(level => ({
                                  levels: [resting, level, resting],
                                  durations: [pre, dur, post]
                              }))
                          };
                      };
                defStim(opts.agonist, opts.preDuration, opts.agonistDuration, opts.samplingSec,
                        opts.consts[opts.agonist], opts.agonistLevels);
                defStim(opts.antagonist, opts.antiPreDuration, opts.antagonistDuration, opts.samplingSec,
                        opts.consts[opts.antagonist], opts.antagonistLevels);
                const segmentSize = ((opts.agonistLevels.length > opts.antagonistLevels.length) ? opts.agonistLevels : opts.antagonistLevels).map(() => n);
                this.prepareStimIdl([
                    {
                        name: opts.agonist,
                        pre: opts.preDuration,
                        dur: opts.agonistDuration,
                        levels: opts.agonistLevels
                    },
                    {
                        name: opts.antagonist,
                        pre: opts.antiPreDuration,
                        dur: opts.antagonistDuration,
                        levels: opts.agonistLevels
                    }
                ].map((stim, iStim) => {
                    const nPre = Math.round(stim.pre / opts.samplingSec),
                          iPost = Math.round((stim.pre + stim.dur) / opts.samplingSec),
                          nPulse = iPost - nPre,
                          nPost = n - iPost,
                          resting = opts.consts[stim.name];
                    return {
                        name: stim.name,
                        segments: segmentSize.map((n, iSeg) => ({
                            classes: [0,1,0],
                            counts: [nPre, nPulse, nPost],
                            amps: [resting, +stim.levels[iSeg], resting],
                            stds: [0,0]
                        }))
                    };
                }));
                this.sim.reconfig({
                    sampling: opts.samplingSec,
                    usePeq: opts.usePeq,
                    constants: this.getConstants(),
                    segmentSize: segmentSize,
                    segmentIndVar: opts.agonistLevels.map((agonistLevel, iL) => {
                        const v = {};
                        v[agonist] = agonistLevel;
                        v[antagonist] = opts.antagonistLevels[iL];
                        return v;
                    }),
                    stimulus: stimulus,
                    about: 'Agonist-antagonist ('+opts.agonist+'-'+opts.antagonist+')'
                });
                if ( ! opts.usePeq ) {
                    const states = this.modelView.model.states,
                          n = states.size;
                    let i = 0;
                    for (; i<n; ++i) {
                        states.set(i, "Pr", (i === opts.startState) ? 1 : 0);
                    }
                }
                this.config.prefRgns = null;
            }
            _onClickPairedPulse(event) {
                this.ppulseDialog.run()
                    .then(opts => this._setupPairedPulse(opts))
                    .catch(qubx.reportErr);
            }
            _setupPairedPulse(opts) {
                const nn = opts.recovery.map(r => (1 + Math.ceil((opts.condDur+r+opts.testDur+opts.testDur)/opts.samplingSec))|0),
                      stimulus = {};
                stimulus[opts.varname] = {
                    stimType: 'steps',
                    filter_kHz: opts.filterkHz,
                    traces: opts.recovery.map(r => ({
                        levels: [opts.resting, opts.condLevel, opts.resting, opts.testLevel, opts.resting],
                        durations: [opts.samplingSec, opts.condDur, r, opts.testDur, opts.testDur]
                    }))
                };
                const nPre = 1,
                      iRec = Math.round((opts.samplingSec + opts.condDur) / opts.samplingSec);
                this.prepareStimIdl([
                    {
                        name: opts.varname,
                        segments: opts.recovery.map((r, iSeg) => {
                            const iTest = Math.round((opts.samplingSec + opts.condDur + r) / opts.samplingSec),
                                  iPost = Math.round((opts.samplingSec + opts.condDur + r + opts.testDur) / opts.samplingSec),
                                  nCond = iRec - nPre,
                                  nRec = iTest - iRec,
                                  nTest = iPost - iTest,
                                  nPost  = nn[iSeg] - iPost;
                            return {
                                classes: [0,1,0,2,0],
                                counts: [nPre, nCond, nRec, nTest, nPost],
                                amps: [opts.resting, opts.condLevel, opts.testLevel],
                                stds: [0, 0, 0]
                            };
                        })
                    }
                ]);
                this.sim.reconfig({
                    sampling: opts.samplingSec,
                    usePeq: opts.usePeq,
                    constants: this.getConstants(),
                    segmentSize: nn,
                    segmentIndVar: opts.recovery.map(dur => ({t_recovery: opts.dur})),
                    stimulus: stimulus,
                    about: 'Paired pulses with varying recovery times: '+opts.recovery.map(qubx.formatK0).join(", ")
                });
                if ( ! opts.usePeq ) {
                    const states = this.modelView.model.states,
                          n = states.size;
                    let i = 0;
                    for (; i<n; ++i) {
                        states.set(i, "Pr", (i === opts.startState) ? 1 : 0);
                    }
                }
                this.config.prefRgns = opts.recovery.map(r => ({
                    f: (1 + Math.ceil((opts.condDur+r)/opts.samplingSec)) |0,
                    n: Math.ceil((opts.testDur+opts.testDur)/opts.samplingSec) |0
                }));
            }
            _onClickRamp(event) {
                this.rampDialog.run()
                    .then(opts => this._setupRamp(opts))
                    .catch(qubx.reportErr);
            }
            _setupRamp(opts) {
                const n = isFinite(opts.dur) ? Math.ceil(opts.dur / opts.samplingSec) : 1,
                      levelsFrom = opts.levelsFrom,
                      levelsTo = opts.levelsTo;
                while ( levelsFrom.length < levelsTo.length ) {
                    levelsFrom.push(levelsFrom.length ? levelsFrom[levelsFrom.length-1] : 0);
                }
                while ( levelsTo.length < levelsFrom.length ) {
                    levelsTo.push(levelsTo.length ? levelsTo[levelsTo.length-1] : 0);
                }
                const stimulus = {};
                stimulus[opts.varname] = {
                    stimType: 'ramps',
                    traces: levelsFrom.map((levelFrom, iLevel) => {
                        const levelTo = levelsTo[iLevel],
                              levels = [levelFrom, levelFrom, levelTo],
                              durations = [0, opts.preDur, opts.rampDur];
                        if ( opts.rampDown ) {
                            levels.push(levelFrom);        levels.push(levelFrom);
                            durations.push(opts.rampDown); durations.push(opts.postDur);
                        } else {
                            levels.push(levelTo);
                            durations.push(opts.postDur);
                        }
                        return {levels, durations};
                    })
                };
                const nPre = Math.round(opts.preDur / opts.samplingSec),
                      iPost = Math.round((opts.preDur + opts.rampDur) / opts.samplingSec),
                      iDown = Math.round((opts.preDur + opts.rampDur + opts.rampDown) / opts.samplingSec),
                      nRampUp = iPost - nPre,
                      nRampDown = iDown - iPost,
                      nPost = n - iDown;
                this.prepareStimIdl([// per stimulus
                    {
                        name: opts.varname,
                        segments: levelsFrom.map((levelFrom, iLevel) => {
                            const levelTo = levelsTo[iLevel],
                                  events = {
                                      classes: [0],
                                      counts: [nPre],
                                      amps: [levelFrom],
                                      stds: [0]
                                  },
                                  addEvent = (n,a,s) => {
                                      events.classes.push(events.amps.length);
                                      events.counts.push(n);
                                      events.amps.push(a);
                                      events.stds.push(s||0);
                                  },
                                  addRamp = (yFrom, yTo, count) => {
                                      const nStep = 100,
                                            dx = count / nStep,
                                            m = (yTo - yFrom) / count;
                                      let x = 0,
                                          iStep = 1;
                                      for (; iStep<=nStep; ++iStep) {
                                          const nextX = Math.round(dx*iStep);
                                          if ( nextX > x ) {
                                              addEvent(nextX-x, yFrom + .5*(m*x + m*nextX));
                                              x = nextX;
                                          }
                                      }
                                  };
                            addRamp(levelFrom, levelTo, nRampUp);
                            if ( opts.rampDown ) {
                                addRamp(levelTo, levelFrom, nRampDown);
                                addEvent(nPost, levelFrom);
                            } else {
                                addEvent(nPost, levelTo);
                            }
                            return events;
                        })
                    }
                ]);
                this.sim.reconfig({
                    sampling: opts.samplingSec,
                    usePeq: opts.usePeq,
                    constants: this.getConstants(),
                    segmentSize: levelsFrom.map(() => n),
                    segmentIndVar: levelsTo.map(level => {
                        const v = {};
                        v[opts.varname] = level;
                        return v;
                    }),
                    stimulus: stimulus,
                    about: 'Ramps of '+opts.varname+' to '+levelsTo.map(qubx.formatK0).join(", ")
                });
                if ( ! opts.usePeq ) {
                    const states = this.modelView.model.states,
                          n = states.size;
                    let i = 0;
                    for (; i<n; ++i) {
                        states.set(i, "Pr", (i === opts.startState) ? 1 : 0);
                    }
                }
                this.config.prefRgns = levelsTo.map(r => ({
                    f: ((opts.preDur + .8*opts.rampDur)/opts.samplingSec) |0,
                    n: Math.ceil((opts.preDur + opts.rampDur)/opts.samplingSec) |0
                }));
            }
            _onChkAuto() {
                this.sim.paused = ! this.chkAuto.checked;
                this.chkAuto.checked && this.sim.update(true);
            }
            _onClickResim() {
                this.sim.update(true);
            }
        };
        
        class StimConstPanel {
            constructor(node) {
                this.node = node;
                this.template = node.querySelector(".qubx_sim_stimconst");
                node.removeChild(this.template);
                this.stims = [];
            }
            read(stimulus) {
                const names = Object.keys(stimulus);
                names.sort((a,b) => a.toLowerCase().localeCompare(b.toLowerCase()));
                this.stims.forEach(rec => this.node.removeChild(rec.stim));
                this.stims = names.map(name => {
                    const stim = this.template.cloneNode(true);
                    this.node.appendChild(stim);
                    stim.querySelector(".qubx_opt_name").textContent = name+':';
                    const input = new qubx_ui.AcceptedInput(stim.querySelector("input"),
                                                            stimulus[name],
                                                            {
                                                                accept: qubx.acceptFloat,
                                                                format: qubx.formatK0,
                                                                widthChars: 6
                                                            });
                    return {name, stim, input};
                });
            }
            write(constants) {
                this.stims.forEach(rec => {
                    constants[rec.name] = rec.input.value;
                });
            }
        }
        
        class VarNamePicker {
            constructor(node) {
                this.node = node;
                this.node.addEventListener('change', () => {
                    this.value = this.node.value;
                });
                this.value = "";
            }
            update(stimulus) {
                const names = Object.keys(stimulus);
                names.sort((a,b) => a.toLowerCase().localeCompare(b.toLowerCase()));
                const options = d3.select(this.node)
                    .selectAll('option')
                    .data(names);
                options.enter().insert('option');
                options.exit().remove();
                options.text(qubx.i);
                options.property('value', qubx.i);
                if ( names.length ) {
                    this.node.selectedIndex = Math.max(0, names.indexOf(this.value));
                    this.value = this.node.value;
                }
/// auto-sel e.g. dose-response varname
            }
        }

        class SetupDialog {
            constructor(config) {
                this.config = config;
                this.dlg = new qubx_ui.Modal({
                    title: config.title,
                    template: '.qubx_dlg_sim_setup'
                });
                this.dlg.dom.then(dom => {
                    const dotdot = qubx_ui.querydotdot(dom.content);
                    qubx_ui.loadTemplate(config.optTemplate).then(opts => {
                        dotdot('.qubx_sim_setup_opts').appendChild(opts);
                        opts.classList.add('qubx_fullsize');
                        this.stims = new StimConstPanel(dotdot(".qubx_sim_stimconsts"));
                        this.stimsLbl = dotdot("#restLvlLbl");
                        this.txtSampling = new qubx_ui.AcceptedInput(dotdot("#sampling"),
                                                                     10.0,
                                                                     {
                                                                         accept: qubx.acceptFloatGreaterThan(0),
                                                                         format: qubx.formatK0,
                                                                         localKey: this.config.localKey+".sampling",
                                                                         widthChars: 6
                                                                     });
                        this.txtDuration = new qubx_ui.AcceptedInput(dotdot("#duration"),
                                                                     1.0,
                                                                     {
                                                                         accept: qubx.acceptFloatGreaterThan(0),
                                                                         format: qubx.formatK0,
                                                                         localKey: this.config.localKey+".dur",
                                                                         widthChars: 6
                                                                     });
                        this.chkStartEq = dotdot("#startEq");
                        this.txtStartState = new qubx_ui.AcceptedInput(dotdot("#startState"),
                                                                       0,
                                                                       {
                                                                           accept: this.config.acceptStateIndex,
                                                                           format: qubx.formatStr,
                                                                           localKey: this.config.localKey+".start",
                                                                           widthChars: 3
                                                                       });
                        this.txtStartState.node.disabled = this.chkStartEq.checked;
                        this.startLabel = {
                            node: dotdot(".startStateLabel"),
                            update: () => {
                                const i = this.txtStartState.value,
                                      lbl = this.config.modelView && this.config.modelView.active[0].model.states.get(i, 'Label');
                                this.startLabel.node.textContent = lbl ? ('('+lbl+')') : '';
                            }
                        };
                        this.txtStartState.ev.change.sub(this.startLabel.update);
                        this.chkStartEq.addEventListener("change", () => {
                            this.txtStartState.node.disabled = this.chkStartEq.checked;
                        });
                        config.setup && config.setup(this, dom, opts);
                    });
                }).catch(qubx.reportErr);
            }
            run() {
                return Promise.resolve(true).then(() => {
                    const stimulus = this.config.getConstants();
                    this.stims.read(stimulus);
                    this.stimsLbl.style.display = Object.keys(stimulus).length ? '' : 'none';
                    this.config.prerun && this.config.prerun(stimulus);
                    this.startLabel.update();
                    return this.dlg.run();
                }).then(() => {
                    let opts = {
                        samplingSec: 1e-3/this.txtSampling.value,
                        usePeq: this.chkStartEq.checked,
                        dur: this.txtDuration.value,
                        consts: this.config.getConstants()
                    };
                    this.stims.write(opts.consts);
                    this.config.updateConstants && this.config.updateConstants(opts.consts);
                    if ( ! opts.usePeq ) {
                        opts.startState = this.txtStartState.value;
                    }
                    opts.constStr = Object.keys(opts.consts).map(name => name+"="+qubx.formatK0(opts.consts[name]));
                    opts = this.config.postrun && this.config.postrun(opts) || opts;
                    return opts;
                });
            }
        }
        
        return exports;
    });
