/* Copyright 2016 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_expo_landowne", ['qubx_fit'], function(qubx_fit) {
    'use strict';

    /* This module implements a method to find all significant exponential components of a duration histogram, published in 

Exponential Sum-Fitting of Dwell-Time Distributions without Specifying Starting Parameters
David Landowne, Bin Yuan, and Karl L. Bagleby
Biophysical Journal Volume 104 June 2013 2383-2391
    */

    var exports = {
        maxN: 20,
        areaCutoff: 1e-5,
        nearThreshold: 0.02,
        fitRepeats: 5,
        fitIterations: 30
    };

    function llrOfSignificance(sig) {
        sig = +sig;
        return Math.log((1-.0005)/sig);
    }
    function llOfSSR(ssr, n) {
        ssr = +ssr;
        n = +n;
        var s2 = ssr/n;
        return -(.5*n)*(Math.log(2*Math.PI)+Math.log(s2)) - ssr/(2*s2);
    }

    // 1. Log bin the data -- xx contains log10(left bound of each bin)
    
    exports.AllExponentials = function() {
        var params = [{
            name: 'N',
            val: exports.maxN,
            constant: true
        }];
        for (var i=0; i<exports.maxN; ++i ) {
            params.push({
                name: 't'+i,
                val: 0.0
            });
            params.push({
                name: 'a'+i,
                val: 0.0
            });
        }
        this.curve = qubx_fit.makeCurve({
            name: 'All Exponentials',
            info: 'see Landowne et al, 2013',
            params: params,
            init: function(p, xx, yy) {
                // p.N.val = ...
            }
        }, function(xx, yy, p) {
            var r = calcR(xx);
            var xp = xx.map(function(x) {
                return Math.pow(10.0, x);
            });
            for (var i=0, nd=xx.length; i<nd; ++i) {
                yy[i] = 0.0;
            }
            var ne = p.N.val | 0;
            for (var ie=0; ie<ne; ++ie) {
                var tau = p['t'+ie].val;
                if ( tau > 0.0 ) {
                    var area = p['a'+ie].val;
                    addExpo(xp, yy, tau, area/tau, r);
                }
            }
        });
    };
    // 2. Generate the time constants for all possible exponentials
    exports.AllExponentials.prototype.getInitialParams = function(xx, x0, x1) {
        // Returns initial param values spaced evenly between min(xx) and max(xx) on log axis, taus constant
        var N = exports.maxN;
        var area = 1/N;
        var pvals = {
            N: { val: N }
        };
        if ( typeof(x0) === 'undefined' ) {
            x0 = xx[0];
            x1 = xx[1];
        } else {
            x0 = Math.log10(x0);
            x1 = Math.log10(x1);
        }
        var t0 = x0;
        var m = (x1 - x0) / (N - 1);
        for (var i=0; i<N; ++i) {
            pvals['t'+i] = {
                val: t0 + i*m,
                constant: true
            };
            pvals['a'+i] = {
                val: area
            };
        }
        return qubx.fit.makeParams(this.curve, pvals);
    };
    // 3. Fit all areas with fixed time constants (done using a Fitter)
    // 4. Delete the end exponentials with insignificant area, generate new evenly spaced taus, fit their areas
    exports.AllExponentials.prototype.getParamsMinusEnds = function(params) {
        // Returns updated param values spaced evenly between smallest and largest legit expo in input params, taus constant
        var N = exports.maxN;
        var i0 = 0;
        while ( (i0 < N) && (params['a'+i0].val < exports.areaCutoff) ) {
            ++i0;
        }
        i1 = N - 1;
        while ( (i0 < i1) && (params['a'+i1].val < exports.areaCutoff) ) {
            --i1;
        }
        if ( (i0 !== 0) || (i1 !== (N-1)) ) {
            return this.getInitialParams(null, params['t'+i0].val, params['t'+i1].val);
        }
        return params;
    };
    // 5. Delete all exponentials with negligible area
    exports.AllExponentials.prototype.getParamsMinusNegligible = function(params) {
        // Returns updated param values with the smallest expos removed, none constant
        var N = exports.maxN;
        var paramsOut = {};
        var iOut = 0;
        for (var i=0; i<N; ++i) {
            var area = params['a'+i].val;
            if ( area >= exports.areaCutoff ) {
                paramsOut['t'+iOut] = {val: params['t'+i]};
                paramsOut['a'+iOut] = {val: area};
                ++iOut;
            }
        }
        paramsOut.N = {val: iOut, constant: true};
        return paramsOut;
    };
    // 6. Fit remaining areas and taus
    // 7. Combine exponentials with very close time constants (in order of decreasing time const, re-fitting after each)
    exports.AllExponentials.prototype.getParamsCombined = function(params, fit, xx, yy, ssr) {
        // Returns {ssr: final, params: updated, fitted param values with near-identical time constants combined into one expo}
        var N = params.N.val|0;
        var paramsOut = {
            N: {val: N, constant: true}
        };
        // re-order exponentials by tau
        var expos = [];
        for (var i=0; i<n; ++i) {
            expos[i] = {
                tau: params['t'+i].val,
                amp: params['a'+i].val,
            };
        }
        expos.sort(function(a,b) {return a.tau - b.tau;});
        for (var i=0; i<N; ++i) {
            paramsOut['t'+i] = {val: expos[i].tau};
            paramsOut['a'+i] = {val: expos[i].area}
        }
        function combine(i) {
            var t1 = paramsOut['t'+i].val,
                a1 = paramsOut['a'+i].val,
                t2 = paramsOut['t'+(i+1)].val,
                a2 = paramsOut['a'+(i+1)].val;
            paramsOut['t'+i].val = t1*a1/(a1+a2) + t2*a2/(a1+a2);
            paramsOut['a'+i].val = a1+a2;
            for (var j=i+1; j<(N-1); ++j) {
                paramsOut['t'+j].val = paramsOut['t'+(j+1)].val;
                paramsOut['a'+j].val = paramsOut['a'+(j+1)].val;
            }
            --N;
            paramsOut.N.val = N;
        }
        var ff = new Float32Array(xx.length);
        for (var i=N-2; i>=0; --i) {
            var t0 = paramsOut['t'+i].val;
            if ( ((paramsOut['t'+(i+1)].val - t0) / t0) < exports.nearThreshold ) {
                combine(i);
                var fitted = fit(curve, paramsOut, xx, yy, ff, function(pp, it, rss) {
                    //console.log('iteration',it,pp,rss);
                }); /*min, params, iterations, fevals*/
                ssr = fitted.ssr;
            }
        }
        return {
            ssr: ssr,
            params: paramsOut
        };
    };
    // Determine the best fit with one fewer exponential
    

        
        
    function calcR(xx, logbase) {
        logbase = logbase || 10.0;
        if ( xx.length < 2 ) {
            return 1.0;
        }
        var min_dx = xx[1] - xx[0];
        for (var i=2, n=xx.length; i<n; ++i) {
            var dx = xx[i] - xx[i-1];
            if ( dx < min_dx ) {
                min_dx = dx;
            }
        }
        return Math.exp(min_dx * Math.log(logbase));
    }
            // area under amp*exp(t/-tau)
            // amp*tau*(exp(-10**x / -tau) - exp(-10**x) * r / -tau))
            // where r = exp( log(10) * min(delta between two log10 x bins)
    function addExpo(xp, yy, tau, area, r) {
        tau = +tau;
        area = +area;
        r = +r;
        for (var i=0, n=xp.length; i<n; ++i) {
            var power = xp[i] / (- tau);
            yy[i] = area * (Math.exp(power) - Math.exp(r*power));
        }
    }
        
    
    


    return exports;
});
