/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_fit", [], () => {
    "use strict";
    const exports = {};
    
    const simplexDefaults = {
        MAX_IT: 1000, // maximum number of iterations
        ALPHA: 1.0, // reflection coefficient
        BETA: 0.5, // contraction coefficient
        GAMMA: 2.0, // expansion coefficient
        EPSILON: 1e-7, // stops iteration when sqrt(rss) < EPSILON
        scale: 1.0
        // onIteration: function(params, itr, rss)
    };
    exports.simplex = function simplex(start, config, func) {
        config = Object.assign({}, simplexDefaults, config);
        const n = start.length,
              ALPHA = config.ALPHA,
              BETA = config.BETA,
              GAMMA = config.GAMMA;

        var vs;         /* vertex with smallest value */
        var vh;         /* vertex with next smallest value */
        var vg;         /* vertex with largest value */
        
        var i,j,m,row;
        var k;           /* track the number of function evaluations */
        var itr;        /* track the number of iterations */
        
        var v;     /* holds vertices of simplex */
        var pn,qn;   /* values used to create initial simplex */
        var f;      /* value of function at each vertex */
        var fr;      /* value of function at reflection point */
        var fe;      /* value of function at expansion point */
        var fc;      /* value of function at contraction point */
        var vr;     /* reflection - coordinates */
        var ve;     /* expansion - coordinates */
        var vc;     /* contraction - coordinates */
        var vm;     /* centroid - coordinates */
        var min;
        
        var fsum,favg,s,cent;
        
        /* dynamically allocate arrays */
        v = [];
        f = [];
        vr = [];
        ve = [];
        vc = [];
        vm = [];
        for (i=0;i<=n;i++) {
            v[i] = [];
        }
        
        /* create the initial simplex */
        /* assume one of the vertices is 0,0 */
        
        pn = config.scale*(Math.sqrt(n+1)-1+n)/(n*Math.sqrt(2.0));
        qn = config.scale*(Math.sqrt(n+1)-1)/(n*Math.sqrt(2.0));
        
        for (i=0;i<n;i++) {
            v[0][i] = start[i];
        }
        
        for (i=1;i<=n;i++) {
            for (j=0;j<n;j++) {
                if (i-1 == j) {
                    v[i][j] = pn + start[j];
                }
                else {
                    v[i][j] = qn + start[j];
                }
            }
        }
        
        /* find the initial function values */
        for (j=0;j<=n;j++) {
            f[j] = func(v[j]);
        }
        
        k = n+1;
        
        /* print out the initial values */
        /*
        console.log("Initial Values");
        for (j=0;j<=n;j++) {
            console.log(v[j][0], v[j][1], f[j]);
        }
        */
        
        /* begin the main loop of the minimization */
        for (itr=1;itr<=config.MAX_IT;itr++) {
            /* find the index of the largest value */
            vg=0;
            for (j=0;j<=n;j++) {
                if (f[j] > f[vg]) {
                    vg = j;
                }
            }
            
            /* find the index of the smallest value */
            vs=0;
            for (j=0;j<=n;j++) {
                if (f[j] < f[vs]) {
                    vs = j;
                }
            }
            
            /* find the index of the second largest value */
            vh=vs;
            for (j=0;j<=n;j++) {
                if (f[j] > f[vh] && f[j] < f[vg]) {
                    vh = j;
                }
            }
            
            /* calculate the centroid */
            for (j=0;j<=n-1;j++) {
                cent=0.0;
                for (m=0;m<=n;m++) {
                    if (m!=vg) {
                        cent += v[m][j];
                    }
                }
                vm[j] = cent/n;
            }
            
            /* reflect vg to new vertex vr */
            for (j=0;j<=n-1;j++) {
                /*vr[j] = (1+ALPHA)*vm[j] - ALPHA*v[vg][j];*/
                vr[j] = vm[j]+ALPHA*(vm[j]-v[vg][j]);
            }
            fr = func(vr);
            k++;
            
            if (fr < f[vh] && fr >= f[vs]) {
                for (j=0;j<=n-1;j++) {
                    v[vg][j] = vr[j];
                }
                f[vg] = fr;
            }
            
            /* investigate a step further in this direction */
            if ( fr <  f[vs]) {
                for (j=0;j<=n-1;j++) {
                    /*ve[j] = GAMMA*vr[j] + (1-GAMMA)*vm[j];*/
                    ve[j] = vm[j]+GAMMA*(vr[j]-vm[j]);
                }
                fe = func(ve);
                k++;
                
                /* by making fe < fr as opposed to fe < f[vs],
                   Rosenbrocks function takes 63 iterations as opposed
                   to 64 when using double variables. */
                
                if (fe < fr) {
                    for (j=0;j<=n-1;j++) {
                        v[vg][j] = ve[j];
                    }
                    f[vg] = fe;
                }
                else {
                    for (j=0;j<=n-1;j++) {
                        v[vg][j] = vr[j];
                    }
                    f[vg] = fr;
                }
            }
            
            /* check to see if a contraction is necessary */
            if (fr >= f[vh]) {
                if (fr < f[vg] && fr >= f[vh]) {
                    /* perform outside contraction */
                    for (j=0;j<=n-1;j++) {
                        /*vc[j] = BETA*v[vg][j] + (1-BETA)*vm[j];*/
                        vc[j] = vm[j]+BETA*(vr[j]-vm[j]);
                    }
                    fc = func(vc);
                    k++;
                }
                else {
                    /* perform inside contraction */
                    for (j=0;j<=n-1;j++) {
                        /*vc[j] = BETA*v[vg][j] + (1-BETA)*vm[j];*/
                        vc[j] = vm[j]-BETA*(vm[j]-v[vg][j]);
                    }
                    fc = func(vc);
                    k++;
                }
                
                
                if (fc < f[vg]) {
                    for (j=0;j<=n-1;j++) {
                        v[vg][j] = vc[j];
                    }
                    f[vg] = fc;
                }
                /* at this point the contraction is not successful,
                   we must halve the distance from vs to all the
                   vertices of the simplex and then continue.
                   10/31/97 - modified to account for ALL vertices.
                */
                else {
                    for (row=0;row<=n;row++) {
                        if (row != vs) {
                            for (j=0;j<=n-1;j++) {
                                v[row][j] = v[vs][j]+(v[row][j]-v[vs][j])/2.0;
                            }
                        }
                    }
                    f[vg] = func(v[vg]);
                    k++;
                    f[vh] = func(v[vh]);
                    k++;
                }
            }
            
            /* test for convergence */
            fsum = 0.0;
            for (j=0;j<=n;j++) {
                fsum += f[j];
            }
            favg = fsum/(n+1);
            s = 0.0;
            for (j=0;j<=n;j++) {
                s += Math.pow((f[j]-favg),2.0)/(n);
            }
            s = Math.sqrt(s);

            /* print out the value at each iteration */
            config.onIteration && config.onIteration(v[vs], itr, s);

            if (s < config.EPSILON) break;
        }
        /* end main loop of the minimization */
        
        /* find the index of the smallest value */
        vs=0;
        for (j=0;j<=n;j++) {
            if (f[j] < f[vs]) {
                vs = j;
            }
        }
        
        //console.log("The minimum was found at");
        for (j=0;j<n;j++) {
            //console.log(v[vs][j]);
            //start[j] = v[vs][j];
        }
        min=func(v[vs]);
        k++;
        //console.log(k,"Function Evaluations");
        //console.log(itr,"Iterations through program");
        
        return {
            min: min,
            param: v[vs],
            iterations: itr,
            fevals: k
        };
    };

    exports.makeCurve = function makeCurve(opts, f) {
        f.curveName = opts.name || f.name;
        f.curveInfo = opts.info || "";
        f.curveInit = opts.init || function(pars, xx, yy) {};
        f.curveParams = opts.params || [];
        f.curveParams.forEach(p => {
            p.val = p.val || 0.0;
            p.constant = p.constant || false;
            p.info = p.info || "";
        });
        return f;
    };

    exports.makeParams = function makeParams(curve, override) {
        const params = {};
        curve.curveParams.forEach(function(par) {
            let p = par;
            if ( override && (par.name in override) ) {
                p = Object.assign({}, p, override[par.name]);
            }
            params[par.name] = {
                val: p.val,
                lo: p.lo,
                hi: p.hi,
                constant: p.constant
            };
        });
        return params;
    };

    exports.line = exports.makeCurve({
        name: 'Line',
        info: 'y = slope * (x - x0) + icept',
        params: [
            {
                name: 'slope',
                val: 1.0,
                //info: ''
                // constant: false,
                // lo: undefined,
                // hi: undefined
            },
            {
                name: 'icept',
                val: 0.0,
            },
            {
                name: 'x0',
                val: 0.0,
                constant: true,
                info: 'x or time offset'
            }
        ],
        init: (pars, xx, yy) => {
            pars.icept.val = yy[0];
            pars.slope.val = (yy[yy.length-1] - yy[0]) / (xx[xx.length-1] - xx[0]);
        }
    }, (xx, yy, p) => {
        const m = p.slope.val,
              x0 = p.x0.val,
              b = p.icept.val,
              n = xx.length;
        let i = 0;
        for (; i<n; ++i) {
            yy[i] = m*(xx[i] - x0) + b;
        }
    });

    exports.hill = exports.makeCurve({
        name: 'Hill',
        info: 'y = x^H / (x^H + EC50^H)',
        params: [
            {
                name: 'H',
                val: 1.0,
                info: ''
                // constant: false,
                // lo: undefined,
                // hi: undefined
            },
            {
                name: 'EC50',
                val: 1.0,
                lo: 0.0
            }
        ]
    }, (xx, yy, p) => {
        const H = p.H.val,
              EC50 = p.EC50.val,
              EH = Math.pow(EC50, H),
              n = xx.length;
        let i = 0;
        for (; i<n; ++i) {
            const xH = Math.pow(xx[i], H);
            yy[i] = xH / (xH + EH);
        }
    });

    exports.SS_FAIL = 1e24;
    exports.simplexFit = function simplexFit(curve, params, xx, yy, ff, onIteration) {
        const pVals = [],
              pOrder = [];
        curve.curveParams.forEach(p => {
            var q = params[p.name];
            if ( ! q.constant ) {
                pVals.push(q.val);
                pOrder.push(q);
            }
        });
        const Npar = pVals.length;
        const result = exports.simplex(pVals, {onIteration: onIteration}, pars => {
            let i = 0;
            for (; i<Npar; ++i) {
                const v = pars[i],
                      p = pOrder[i];
                if ( (typeof(p.lo) !== 'undefined') && (v < p.lo) ) {
                    return exports.SS_FAIL;
                }
                if ( (typeof(p.hi) !== 'undefined') && (v > p.hi) ) {
                    return exports.SS_FAIL;
                }
                pOrder[i].val = pars[i];
            }
            curve(xx, ff, params);
            let ssd = 0.0;
            const n = xx.length;
            for (i=0; i<n; ++i) {
                const d = yy[i] - ff[i];
                ssd += d*d;
            }
            return ssd;
        });
        let j = 0;
        for (; j<Npar; ++j) {
            pOrder[j].val = result.param[j];
        }
        result.params = params;
        return result;
    };
    
    exports.test = function test(xx, yy) {
        xx = xx || new Float32Array([0, 1, 2, 3, 4, 5]);
        yy = yy || new Float32Array([.1, .3, .3, .4, .6, .7]);
        const ff = xx.map(() => 0.0);

        const curve = exports.line;
        const pars = exports.makeParams(curve, {
            slope: {
                val: 2.0
            }
        });
        console.log(JSON.stringify(pars));
        
        curve.curveInit(pars, xx, yy);
        console.log(JSON.stringify(pars));
        
        const fitted = exports.simplexFit(curve, pars, xx, yy, ff, (pp, it, rss) => {
            console.log('iteration',it,pp,rss);
        });
        console.log(fitted);
        console.log(ff);
        console.log(JSON.stringify(pars));
    };
    
    return exports;
});
