/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_hist", ['qubx', 'qubx_ui', 'qubx_lib', 'qubx_chart', 'd3', 'fitness_ui'], (qubx, qubx_ui, qubx_lib, qubx_chart, d3, fitness_ui) => {
    'use strict';

    const exports = {};

    const AmpHistDefs = {
        Nbin: 64
    };
    exports.AmpHist = class AmpHist {
        constructor(config) {
            this.config = Object.assign({}, AmpHistDefs, config);
            this.dataView = config.dataView;
            this.modelView = config.modelView;
            this.node = config.node || document.createElement('div');
            this.node.classList.add('qubx_hist_amp');
            this.box = d3.select(this.node).append('div');
            this.box.classed('qubx_fullsize', true);
            this.dlgOptions = new qubx_ui.Modal({
                title: 'Amplitude histogram options',
                canCancel: false,
                okLabel: 'Close',
                template: '#qubx_hist_amp_options'
            });
            this.dlgOptions.dom.then(dom => {
                this.txtBins = new qubx_ui.AcceptedInput(dom.content.querySelector('.qubx_hist_amp_bins'), this.config.Nbin, {
                    accept: qubx.acceptIntAtLeast(2),
                    localKey: 'qubx_hist_amp_bins',
                    widthChars: 5
                });
                this.config.Nbin = this.txtBins.value;
                this.schedule(true);
                this.txtBins.ev.change.sub(this._onChangeBins.bind(this));
            }).catch(qubx.reportErr);
            this.chart = new qubx_chart.Chart({
                node: this.node,
                legend: {
                    show: true
                }
            });
            this.chart.laidOut.then(() => {
                this.btnFit = this.chart.buttons.append('button')
                    .attr('title', 'Curve fitting...')
                    .on('click', () => fitness_ui.runFitControls({
                        category: 'AmpHist',
                        expr: fitness_ui.presets[1].expr, // Gaussian
                        xName: 'I',
                        yName: 'Frequency',
                        columns: {
                            I: this.bins,
                            Frequency: this.bars
                        },
                        oncurve: (curve, info) => this.setFit(curve)
                    }));
                this.btnFit.append('img').attr({
                    src: 'images/curvefit.svg',
                    alt: '[Curve Fitting...]'
                }).style('width', '2em');
                this.btnOptions = this.chart.buttons.append('button')
                    .attr('title', 'Options...')
                    .on('click', () => this.dlgOptions.run());
                this.btnOptions.append('img').attr({
                    src: 'images/appbar.settings.svg',
                    alt: '[Settings]'
                }).style('width', '2em');
            });
            this.hOnChoose = this.dataView.ev.choose.sub(this._onChoose.bind(this));
            this.hOnChangeData = this.dataView.ev.changeData.sub(this._onChangeData.bind(this));
            this.hOnPeq = this.modelView.ev.Peq.sub(this._onPeq.bind(this));
            this.hOnAmps = this.modelView.ev.amps.sub(this._onAmps.bind(this));
            this.hOnResize = qubx_ui.onResize(this.node, this._onResize.bind(this));
            this.updateSoon = qubx.batchDelayed(this.update.bind(this), 200);
            this.data = null;
            this.showing = false;
            this.spinner = new qubx_ui.Spinner(this.node);
            
            qubx_lib.then(lib => {
                this.maker = new lib.AmpHistMaker(this.config.Nbin);
                this._onChangeData(this.dataView.data());
            }).catch(qubx.reportErr);
        }
        destroy() {
            this.node.innerHTML = '';
            this.node.classList.remove('qubx_hist_amp');
            this.maker && this.maker.destroy();
            this.maker = null;
            this._onChangeData(null);
            this.dataView.ev.choose.unsub(this.hOnChoose);
            this.dataView.ev.changeData.unsub(this.hOnChangeData);
            this.modelView.ev.Peq.unsub(this.hOnPeq);
            this.modelView.ev.amps.unsub(this.hOnAmps);
            qubx_ui.clearResize(this.hOnResize);
        }
        onShow(showing) {
            this.showing = showing;
            this.schedule(true);
        }
        _onChangeBins(Nbin) {
            this.config.Nbin = Nbin;
            this.schedule(true);
        }
        _onResize(info) {
            this.chart.resize();
        }
        _onChoose(i) {
            this.schedule(true);
        }
        _onChangeData(data) {
            if ( this.data ) {
                this.data.ev.changeSamples.unsub(this.hOnChangeSamples);
            }
            this.data = data;
            if ( this.data ) {
                this.hOnChangeSamples = this.data.ev.changeSamples.sub(this._onChangeSamples.bind(this));
            }
            this.schedule(true);
        }
        _onChangeSamples() {
            this.schedule(true);
        }
        _onPeq() {
            this.schedule();
        }
        _onAmps() {
            this.schedule();
        }
        schedule(binsToo) {
            this.requested = true;
            this.requestBins = this.requestBins || binsToo;
            this.showing && (! this.scheduled) && this.updateSoon();
        }
        update() {
            if ( this.scheduled ) {
                return;
            }
            if ( ! (this.maker && this.data) ) {
                this.scheduled = false;
                return;
            }
            const wantBins = this.requestBins;
            this.scheduled = true;
            this.requested = false;
            this.requestBins = false;
            this.spinner.show();
            
            let step1 = Promise.resolve(false);
            if ( wantBins ) {
                if ( (! this.bins) || (this.config.Nbin !== this.maker.Nbin) ) {
                    this.maker.setNbin(this.config.Nbin);
                    this.bins = new Float32Array(this.config.Nbin);
                    this.bars = new Float32Array(this.config.Nbin);
                    this.pdf = new Float32Array(this.config.Nbin);
                } else {
                    this.maker.reset();
                }
                step1 = this.data.eachChunk(this.dataView.chosen(), 0, 10240, chunk => this.maker && this.maker.scan(chunk))
                    .then(a => this.maker && this.maker.bin(this.bins))
                    .then(b => this.data.eachChunk(this.dataView.chosen(), 0, 10240, chunk => this.maker && this.maker.add(chunk)))
                    .then(c => {
                        if ( this.maker ) {
                            this.maker.hist(this.bars);
                        }
                        return this.requested ? this.update() : true;
                    });
            }
            return step1.then(newBins => {
                // do the pdf
                let factorCount = 0;
                this.modelView.active.forEach(m => {
                    factorCount += m.model.channelCount();
                });
                if ( factorCount > 10 ) {
                    this.setHist(this.bins, this.bars);
                } else {
                    var factors = []; // if channel count === 2, list it twice
                    this.modelView.active.forEach(m => {
                        let i = 0;
                        for (; i<m.model.channelCount(); ++i) {
                            factors.push(m);
                        }
                    });
                    const pdf = amppdfAll(factors, this.bins, this.pdf);
                    if ( newBins ) {
                        this.setHist(this.bins, this.bars, this.pdf);
                    } else if ( this.chart.trends.length ) {
                        this.chart.trends[0].setData(Array.from(this.pdf));
                    }
                }
                this.scheduled = false;
                this.spinner.hide();
                return this.requested ? this.update() : true;
            })
                .catch(qubx.reportErr)
                .catch(err => {
                    this.scheduled = false;
                    //? this.spinner.hide();
                    return Promise.reject(err);
                });
        }
        setHist(bins, bars, pdf) {
            if ( bins[0] === bins[1] ) {
                return;
            }
            this.chart.setHist({
                xvar: 'pA',
                yvar: 'Amplitude',
                xlabel: '',
                ylabel: 'count/total',
                series: {
                    'pA': Array.from(bins),
                    'Amplitude': Array.from(bars)
                }
            });
            this._iFit = undefined;
            if ( pdf ) {
                this.chart.trendCount(1, i => 'PDF', i => '#f00');
                this.chart.trends[0].setData(Array.from(pdf));
            } else {
                this.chart.trendCount(0);
            }
        }
        setFit(fit) {
            let iFit = this._iFit;
            if ( (typeof(iFit) === 'undefined') || ! this.chart.trends[iFit] ) {
                iFit = this._iFit = this.chart.trends.length;
                this.chart.trendCount(iFit+1, i => 'Fit', i => '#f06');
            }
            this.chart.trends[iFit].setData(Array.from(fit));
        }
    };

    const EPS = 1e-5;
    function gaussian(mean, xms, x) {
        mean = +mean;
        xms = +xms;
        x = +x;
        if ( xms === 0.0 ) {
            if ( Math.abs(x-mean) < EPS ) {
                return 1e6; // warning this is a sub for infinity
            } else {
                return 0.0;
            }
        }
        let tmp = (x-mean)/xms;
        tmp = Math.exp(-tmp*tmp/2.0);  // this is where it would underflow
        tmp/= 2.5066283*xms;
        return (tmp<1.0e-8) ? 1.0e-8 : tmp;
    }

    function amppdf(amp, xms, pe, bins, pdf) {
        const nclass = pe.length,
              nbin = bins.length;
        let sum = 0.0;
        
        // Sum per bin and overall
        let iBin = 0;
        for (; iBin<nbin; iBin++) {
            let binsum = 0.0,
                iClass = 0;
            for (; iClass<nclass; iClass++) { // sum all classes for bin
                const y = pe[iClass]*gaussian(amp[iClass],xms[iClass],bins[iBin]);   // pdf[iClass][iBin]
                pdf[iBin] = y;
                binsum += y;
            }
            pdf[iBin] = binsum;
            sum += binsum;
        }
        
        // Normalize
        if( sum !== 0.0 ) {
            for(iBin=0; iBin<nbin; iBin++) {
                pdf[iBin] /= sum;
            }
        }
    }

    function amppdfAll(factors, bins, pdf) {
        const Nclass = [],
              prep = [];
        let Nclass_i = [];
        factors.forEach(m => {
            const amp = [],
                  vrc = [];
            m.classCalc.entries.forEach((c, i) => {
                amp[i] = c.Amp;
                vrc[i] = c.Std * c.Std;
            });
            const peq = amp.map(() => 0.0),
                  sts = m.model.states;
            m.stateCalc.entries.forEach((s, i) => {
                peq[sts.get(i,'Class')] += s.Peq;
            });
            let iOut = 0,
                i = 0;
            for (; i<amp.length; ++i) {
                if ( peq[i] > 1e-8 ) {
                    amp[iOut] = amp[i];
                    vrc[iOut] = vrc[i];
                    peq[iOut] = peq[i];
                    ++iOut;
                }
            }
            if ( iOut ) {
                Nclass.push(iOut);
                Nclass_i.push(0);
                prep.push({
                    amp: amp.slice(0, iOut),
                    vrc: vrc.slice(0, iOut),
                    peq: peq.slice(0, iOut)
                });
            }
        });
        function next() {
            /// iterates over cartesian product of model classes, as indexed in Nclass_i
            /// returns false on rollover
            let iModel = 0,
                carry = 1;
            while ( carry && (iModel < prep.length) ) {
                const d = Nclass_i[iModel] + carry;
                Nclass_i[iModel] = d % Nclass[iModel];
                carry = (d / Nclass[iModel])|0;
                ++iModel;
            }
            return ! carry;
        }
        const amp = [],
            std = [],
            peq = [];
        while ( Nclass_i.length ) {
            let a = 0.0,
                v = 0.0,
                p = 1.0;
            prep.forEach((m, iModel) => {
                const c = Nclass_i[iModel];
                a += m.amp[c];
                v += m.vrc[c];
                p *= m.peq[c];
            });
            amp.push(a);
            std.push(Math.sqrt(v));
            peq.push(p);
            if ( ! next() ) {
                Nclass_i = [];
            }
        }
        amppdf(amp, std, peq, bins, pdf);
        return pdf;
    }
    
    const DurHistDefs = {
        Nbin: 32,
        maxNbin: 16384,
        dwellBatch: 16384,
        smooth: false,
        sqrt: true
    };
    exports.DurHists = class DurHists {
        constructor(config) {
            this.config = Object.assign({}, DurHistDefs, config);
            this.dataView = config.dataView;
            this.modelView = config.modelView;
            this.grid = new qubx_ui.AspectGrid3({
                node: this.config.node
            });
            this.node = this.grid.node;
            this.entries = [];
            this.dlgOptions = new qubx_ui.Modal({
                title: 'Duration histogram options',
                canCancel: false,
                okLabel: 'Close',
                template: '#qubx_hist_dur_options'
            });
            this.dlgOptions.dom.then(dom => {
                const dotdot = qubx_ui.querydotdot(dom.content);
                this.txtBins = new qubx_ui.AcceptedInput(dotdot('.qubx_hist_dur_bins'), this.config.Nbin, {
                    accept: qubx.acceptIntAtLeast(2),
                    localKey: 'qubx_hist_dur_bins',
                    widthChars: 5
                });
                this.txtBins.ev.change.sub(this._onChangeBins.bind(this));
                this.chkSmooth = dotdot.listen('.qubx_hist_dur_smooth', 'change', () => {
                    this.config.smooth = this.chkSmooth.checked;
                    this.schedule(true);
                });
                this.chkSmooth.checked = this.config.smooth;
                this.chkSqrt = dotdot.listen('.qubx_hist_dur_sqrt', 'change', () => {
                    this.config.sqrt = this.chkSqrt.checked;
                    this.schedule(true);
                });
                this.chkSqrt.checked = this.config.sqrt;
                this.config.Nbin = this.txtBins.value;
                this.schedule(true);
            }).catch(qubx.reportErr);
            this.hOnChoose = this.dataView.ev.choose.sub(this._onChoose.bind(this));
            this.hOnChangeData = this.dataView.ev.changeData.sub(this._onChangeData.bind(this));
            this.hOnPeq = this.modelView.ev.Peq.sub(this._onPeq.bind(this));
            this.hOnActive = this.modelView.ev.active.sub(this._onActive.bind(this));
            this.hOnSelectModel = this.modelView.context.ev.selectModel.sub(this._onSelectModel.bind(this));
            this.updateSoon = qubx.batchDelayed(this.update.bind(this), 1000);
            this.data = null;
            this.showing = false; // (otherwise it computes in bg for no reason)
            this.spinner = new qubx_ui.Spinner(this.node);
            this.waiting = [];
            
            qubx_lib.then(lib => {
                this.maker = new lib.DurHistMaker(this.config.dwellBatch, this.config.maxNbin);
                this._onChangeData(this.dataView.data());
            }).catch(qubx.reportErr);
        }
        destroy() {
            this.maker && this.maker.destroy();
            this.maker = null;
            this._onChangeData(null);
            this.dataView.ev.choose.unsub(this.hOnChoose);
            this.dataView.ev.changeData.unsub(this.hOnChangeData);
            this.modelView.ev.Peq.unsub(this.hOnPeq);
            this.modelView.ev.active.unsub(this.hOnActive);
            this.modelView.context.ev.selectModel.unsub(this.hOnSelectModel);
        }
        _setClassCount(nc) {
            const entries = [];
            let i = 0;
            for (; i<nc; ++i) {
                let entry = this.entries[i];
                if ( ! entry ) {
                    entry = this.entries[i] = {c:i};
                }
                entries.push(entry);
            }
            if ( nc < this.entries.length ) {
                this.entries.splice(nc, this.entries.length-nc);
            }
            const self = this;
            this.grid.data(entries, qubx.m('c'),
                           function(d, i) { // init
                               d.node = this;
                               d.node.classList.add('qubx_hist_dur');
                               d.chart = new qubx_chart.Chart({
                                   node: d.node,
                                   legend: {
                                       show: true
                                   }
                               });
                               d.chart.laidOut.then(() => {
                                   d.btnFit = d.chart.buttons.append('button')
                                       .attr('title', 'Curve fitting...')
                                       .on('click', () => fitness_ui.runFitControls({
                                           category: 'DurHist'+i,
                                           expr: fitness_ui.presets[8].expr, // Expo log x
                                           xName: 'log10ms',
                                           yName: 'Frequency',
                                           columns: {
                                               log10ms: d.chart.columnData[0],
                                               Frequency: self.config.sqrt ? d.chart.columnData[1].map(x => x*x) : d.chart.columnData[1]
                                           },
                                           oncurve: (curve, info) => self.setFit(i, self.config.sqrt ? curve.map(Math.sqrt) : curve)
                                       }));
                                   d.btnFit.append('img').attr({
                                       src: 'images/curvefit.svg',
                                       alt: '[Curve Fitting...]'
                                   }).style('width', '2em');
                                   if ( (i === 0) && ! d.btnOptions ) {
                                       d.btnOptions = d.chart.buttons.append('button')
                                           .attr('title', 'Options...')
                                           .on('click', () => self.dlgOptions.run());
                                       d.btnOptions.append('img').attr({
                                           src: 'images/appbar.settings.svg',
                                           alt: '[Settings]'
                                       }).style('width', '2em');
                                   }
                               });
                           },
                           function(d, i) { // layout
                               d.chart.resize();
                           },
                           function(d, i) { // teardown
                               d.chart = d.box = d.node = null;
                           });
        }
        onShow(showing) {
            this.showing = showing;
            this.schedule(true);
        }
        _onChangeBins(Nbin) {
            this.config.Nbin = Nbin;
            this.schedule(true);
        }
        _onChoose(i) {
            this.schedule(true);
        }
        _onChangeData(data) {
            if ( this.data ) {
                this.data.ev.changeIdeal.unsub(this.hOnChangeIdeal);
                this.data.ev.changeTDead.unsub(this.hOnChangeTDead);
            }
            this.data = data;
            if ( this.data ) {
                this.hOnChangeIdeal = this.data.ev.changeIdeal.sub(this._onChangeIdeal.bind(this));
                this.hOnChangeTDead = this.data.ev.changeTDead.sub(this._onChangeTDead.bind(this));
            }
            this.schedule(true);
        }
        _onChangeIdeal() {
            this.schedule(true);
        }
        _onChangeTDead() {
            this.schedule(true);
        }
        _onPeq() {
            this.schedule();
        }
        _onActive() {
            // this.schedule();
        }
        _onSelectModel() {
            this.schedule();
        }
        _onAmps() {
            this.schedule();
        }
        getSVG(c) {
            if ( ! (this.requested || this.scheduled) ) {
                return Promise.resolve(this._getSVG(c));
            }
            this.scheduled || this.updateSoon();
            return new Promise((resolve, reject) => this.waiting.push({
                resolve: () => resolve(this._getSVG(c)),
                reject: reject
            }));
        }
        _getSVG(c) {
            if ( typeof(c) === 'undefined' ) {
                return this.entries.map(item => item.chart.asSVG());
            }
            return this.entries[c].chart.asSVG();
        }
        schedule(binsToo) {
            this.requested = true;
            this.requestBins = this.requestBins || binsToo;
            this.showing && (! this.scheduled) && this.updateSoon();
        }
        update() {
            if ( this.scheduled ) {
                return;
            }
            if ( ! (this.maker && this.data) ) {
                return;
            }
            const wantBins = this.requestBins;
            this.scheduled = true;
            this.requested = false;
            this.requestBins = false;
            this.spinner.show();
            
            const tdMS = this.data.tDead();
            let step1 = Promise.resolve(false);
            if ( wantBins ) {
                // assemble events, find tmin, tmax via qubx.extrema
                const sampling = 1e3 * this.data.sampling(),
                      chunks = [];
                // events are in msec, with tdMS subtracted before finding tmin and tmax
                let tmin = 1e20,
                    tmax = -1e20,
                    maxClass = -1;
                step1 = this.data.eachChunk(this.dataView.chosen(), 0, 0, chunk => {
                    const ff = chunk.firsts,
                          ll = chunk.lasts,
                          cc = chunk.classes,
                          nd = ff.length,
                          dd = new Float32Array(nd);
                    chunk.durations = dd;
                    let di = 0;
                    for (; di<nd; ++di) {
                        const dur = (ll[di] - ff[di] + 1) * sampling;
                        if ( dur < tmin ) {
                            tmin = dur;
                        }
                        if ( dur > tmax ) {
                            tmax = dur;
                        }
                        dd[di] = dur;
                        maxClass = Math.max(maxClass, cc[di]);
                    }
                    chunk.count = qubx.applyTdead(chunk.classes, dd, tdMS, {subtract: true});
                    chunks.push(chunk);
                }, true).then(() => {
                    if ( tmin < tmax ) {
                        this._setClassCount(maxClass+1);
                        this.bins = this.maker.makeBins(this.config.Nbin, tmin, tmax, tdMS, maxClass+1);
                        chunks.forEach(chunk => this.maker.add(-(maxClass+1), chunk.classes, chunk.durations, chunk.count, tdMS, sampling, this.config.smooth)); ///?
                        this.bars = [];
                        let i = 0;
                        for (; i<=maxClass; ++i) {
                            this.bars[i] = this.maker.read(i);
                            this.config.sqrt && sqrtize(this.bars[i]);
                        }
                    } else {
                        this.bins = this.bars = null;
                    }
                    return this.requested ? this.update() : true;
                });
            }
            return step1.then(newBins => qubx_lib.then(lib => {
                if ( this.bins ) {
                    if ( newBins ) {
                        this.setHist(this.bins, this.bars);
                    }
                    // do the pdf
                    const model = this.modelView.model;
                    if ( model && model.active && (model.channelCount() === 1) ) { /// we need the auto-multimodel for small nchannel
                        const clazz = model.states.entries.map(qubx.m('Class')),
                              Qe = lib.QtoQe(this.modelView.info.Q, clazz, tdMS*1e-3);
                        this.bars.forEach((bars, c) => {
                            const d = this.entries[c],
                                  pdf = this.maker.pdf(model.states.size, Qe, clazz, c, tdMS);
                            d.chart.trendCount(pdf.components.length+1,
                                               i => i ? ('c'+i) : 'PDF',
                                               i => i ? 'rgba(0,0,0,0.5)' : qubx_chart.trendColors[c]) ///?
                            this.config.sqrt && sqrtize(pdf.pdf);
                            d.chart.trends[0].setData(Array.from(pdf.pdf));
                            pdf.components.forEach((comp,i) => {
                                this.config.sqrt && sqrtize(comp);
                                d.chart.trends[i+1].setData(Array.from(comp));
                            });
                            d.chart.tbltop.html(Array.from(pdf.tau).map((tau,i) => {
                                const area = pdf.amp[i] * tau;
                                return 'c'+(i+1)+': '+qubx.formatK0(1e3*tau)+' ms, area '+area.toFixed(3);
                            }).join('<br>'));
                        });
                    } else {
                        this.bars.forEach((bars, c) => {
                            this.entries[c].chart.trendCount(0);
                        });
                    }
                }
                this.scheduled = false;
                this.spinner.hide();
                return this.requested ? this.update() : true;
            })).then(() => {
                setTimeout(() => {
                    this.waiting.forEach(w => w.resolve());
                    this.waiting = [];
                }, 800);
            }).catch(qubx.reportErr)
                .catch(err => {
                    this.scheduled = false;
                    //? this.spinner.hide();
                    this.waiting.forEach(w => w.reject());
                    this.waiting = [];
                    return Promise.reject(err);
                });
        }
        setHist(bins, bars) {
            if ( bins[0] === bins[1] ) {
                return;
            }
            const series = {'log10 ms': Array.from(bins)};
            this.entries.forEach((d,i) => {
                if ( d.chart ) {
                    const barLabel = 'Class '+i,
                          thisSeries = Object.assign({}, series);
                    thisSeries[barLabel] = Array.from(bars[i]);
                    d.chart.setHist({
                        xvar: 'log10 ms',
                        yvar: 'Class '+i,
                        xlabel: '',
                        ylabel: this.config.sqrt ? 'sqrt(n/total)' : 'count/total',
                        series: thisSeries
                    });
                    d.chart.histFit = undefined;
                }
            });
        }
        setFit(iChart, fit) {
            const chart = this.entries[iChart].chart;
            let iFit = chart.histFit;
            if ( (typeof(iFit) === 'undefined') || ! chart.trends[iFit] ) {
                iFit = chart.histFit = chart.trends.length;
                chart.trendCount(iFit+1, i => 'Fit', i => '#f06');
            }
            chart.trends[iFit].setData(Array.from(fit));
        }
    };
    
    function sqrtize(a) {
        const n = a.length;
        let i = 0;
        for (; i<n; ++i) {
            const y = a[i];
            a[i] = (y >= 0.0) ? Math.sqrt(y) : 0.0;
        }
    }

    return exports;
});

