/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_mac", ["qubx", "qubx_optimize", "lmfit", "qubx_workers"], (qubx, qubx_optimize, lmfit, qubx_workers) => {
    "use strict";
    
    const exports = {};

    const MacWorkgroupDefs = {
        count: 3
    };
    exports.MacWorkgroup = class MacWorkgroup {
        constructor(config) {
            this._workers = [];
            this.layout = null;
            this.models = [];
            this.transfer = new exports.MacTransfer();
            this.config = {
                count: config && config.count || MacWorkgroupDefs.count
            };
            this.count(this.config.count);
        }
        destroy() {
            this.count(0);
        }
        count(c) {
            if ( typeof(c) === 'undefined' ) {
                return this._workers.length;
            }
            if ( c === this._workers.length ) {
                return this;
            }
            while ( c > this._workers.length ) {
                this._workers.push(new exports.MacWorker(this.transfer));
                if ( this.model ) {
                    this._workers[this._workers.length-1].setModel(this.model);
                }
            }
            while ( c < this._workers.length ) {
                this._workers.pop().destroy();
            }
            if ( this.layout ) {
                this._distributeLayout();
            }
            return this;
        }
        setModels(models) {
            this.models = models;
            this._workers.forEach(w => w.setModels(models));
            return this;
        }
        setModel(iModel, model) {
            this.models[iModel] = model;
            this._workers.forEach(w => w.setModel(iModel, model));
            return this;
        }
        setLayout(layout, getChunk) {
            this.layout = layout;
            this.getChunk = getChunk;
            let Nchunk = 0,
                maxChunk = 0;
            layout.segments.forEach(seg => {
                Nchunk += seg.chunks.length|0;
                seg.chunks.forEach(chunk => {
                    maxChunk = Math.max(maxChunk, chunk.n);
                });
            });
            this.transfer.setup(layout.signalNames.length, Math.ceil(Nchunk/this._workers.length), maxChunk);
            return this._distributeLayout();
        }
        _distributeLayout() {
            const l = this.layout;
            this.batch = Math.ceil(l.segments.length / this._workers.length)|0;
            let s0 = 0;
            const nextBatch = () => {
                s0 += this.batch;
                return {
                    sampling: l.sampling,
                    signalNames: l.signalNames,
                    delays: l.delays,
                    defaults: l.defaults,
                    resampled: l.resampled,
                    segments: l.segments.slice(s0-this.batch, Math.min(l.segments.length, s0))
                }
            };
            return Promise.all(this._workers.map((w, iw) => w.setLayout(nextBatch(), (arrays, iSeg, iChunk) => {
                return this.getChunk(arrays, this.batch*iw + iSeg, iChunk);
            })));
        }
        setRate(iModel, index, field, value) {
            if ( this.models[iModel] ) {
                this.models[iModel].rates[index][field] = value;
            }
            this._workers.forEach(w => w.setRate(iModel, index, field, value));
            return this;
        }
        setChannelCount(iModel, cc) {
            if ( this.models[iModel] ) {
                this.models[iModel].channelCount = cc;
            }
            this._workers.forEach(w => w.setChannelCount(iModel, cc));
            return this;
        }
        setDelay(name, value) {
            if ( this.layout ) {
                this.layout.delays[name] = value;
            }
            this._workers.items.forEach(w => w.setDelay(name, value));
            return this;
        }
        setConstant(name, value) {
            if ( this.layout ) {
                this.layout.defaults[name] = value;
            }
            this._workers.forEach(w => w.setConstant(name, value));
            return this;
        }
        optPars(modelOpts, optNchannel) {
            return Promise.all(this._workers.map(w => w.optPars(modelOpts, optNchannel)))
                .then(results => results[0]);
        }
        getLL(params) {
            if ( ! (this.layout && this.layout.segments.length) ) {
                return Promise.resolve({
                    ll: 0.0,
                    resultCode: 0
                });
            }
            let ll = 0.0,
                n = 0;
            const pp = params && params.slice();
            return Promise.all(this._workers.map((w, iw) => {
                if ( iw*this.batch < this.layout.segments.length ) {
                    return w.getLL(pp);
                } else {
                    return Promise.resolve({ll: 0.0, ndata: 0});
                }
            })).then(results => {
                let result = {};
                results.forEach(r => {
                    ll += r.ll;
                    n += r.ndata;
                });
                if ( results.length ) {
                    result = results[0];
                    result.ll = ll;
                    result.ndata = n;
                    if ( pp && pp.length ) {
                        result.params = pp;
                        result.aic = 2*pp.length - 2*ll;
                        result.bic = pp.length*Math.log(n) - 2*ll;
                    }
                }
                return result;
            });
        }
        getOutput(onChunk) {
            if ( ! (this.layout && this.layout.segments.length) ) {
                return Promise.resolve(true);
            }
            return Promise.all(this._workers.map((w, iw) => {
                if ( this.batch*iw < this.layout.segments.length ) {
                    return w.getOutput((iSeg, iChunk, arr, chunk) => onChunk(this.batch*iw + iSeg, iChunk, arr, chunk));
                } else {
                    return Promise.resolve(true);
                }
            }));
        }
        getOptStd(invHessian) {
            if ( ! (this.layout && this.layout.segments.length) ) {
                return Promise.resolve({});
            }
            return this._workers[0].getOptStd(invHessian);
        }
        optimize(modelOpts, opts) {
            /* modelOpts: {constraints, balanceLoops, opt: {rates, cc, ...}} */
            /* opts: {
               lm, dfp, simplex: boolean; or:
               OptClass: e.g. qubx_optimize.Simplex or qubx_optimize.DFP
               onIteration: function(model, params, ll, iteration[, grad])
               ...any constructor config options for OptClass, in particular:
               maxIter
               epsilon
               }*/
            let lastModels = [];
            const score = params => this.getLL(params)
                  .then(data => {
                      lastModels = data.models;
                      return data.resultCode ? -1e20 : -data.ll;
                  });
            const iter = (params, score, iteration, grad) => {
                if ( opts.onIteration ) {
                    return opts.onIteration(lastModels, params, -score, iteration, grad);
                } else {
                    return true;
                }
            };
            const run = (OptClass, params, randomize) => {
                const pp = params.slice();
                if ( randomize ) {
                    console.log(params);
                    let i = 0;
                    for (; i<pp.length; ++i) {
                        pp[i] *= 1+randomize*qubx.gaussRandom();
                    }
                    console.log(pp);
                }
                return (new OptClass(pp, opts)).run({
                    score: score,
                    grad: qubx_optimize.numGrad2Side(score),
                    iter: iter,
                    dx: opts.dx
                }).then(result => {
                    result.ll = - result.score;
                    if ( result.hessian ) {
                        return this.getOptStd([].concat.apply([], result.hessian)) // from 2d arrays to 1d
                            .then(models => {
                                result.models = models;
                                return result;
                            });
                    } else {
                        result.models = lastModels;
                        return result;
                    }
                });
            };
            const runLM = params => {//////
                const lmParams = params.slice();
                let ndata = 0;
                const segOffset = this.layout.segments.map(seg => {
                    let segOff = ndata;
                    ndata += seg.n;
                    return segOff;
                });
                const evalLM = (pp, fvec, iteration) => {
                    return score(pp).then(negLL => this.getOutput((iSeg, iChunk, arr, chunk) => {
                        /// pretty sure there's only one chunk per segment, otherwise this is wrong
                        const n = this.layout.segments[iSeg].n;
                        let iOut = segOffset[iSeg],
                            iIn = 0;
                        for (; iIn < n; ++iIn, ++iOut) {
                            fvec[iOut] = arr[4*iIn + 3]; // to minimize error scaled by std
                        }
                    }).then(() => iter(pp, negLL, iteration) ? 0 : -1));
                };
                const control = {
                    maxcall: Math.ceil(opts.maxIter/lmParams.length)|0
                };
                return lmfit.minimizeSoon(lmParams, evalLM, ndata, control)
                    .then(() => score(lmParams)).then(negLL => ({
                        score: negLL,
                        ll: -negLL,
                        grad: lmParams.map(p => 0.0),
                        iteration: control.iteration,
                        params: lmParams,
                        models: lastModels
                    }));
            };
            return this.optPars(modelOpts, opts.optNchannel).then(params => {
                let optResult = Promise.resolve({
                    params: params,
                    score: 0,
                    models: []
                });
                const chain = (OptClass, randomize) => {
                    optResult = optResult.then(result => run(OptClass, result.params, randomize));
                };
                if ( opts.lm || opts.simplex || opts.dfp ) {
                    opts.simplex && chain(qubx_optimize.Simplex);
                    if ( opts.lm ) {
                        optResult = optResult.then(result => runLM(result.params));
                    }
                    opts.dfp && chain(qubx_optimize.DFP, opts.simplex && .02);// dfp needs distance from solution to develop hessian error estimates
                } else {
                    chain(opts.OptClass || qubx_optimize.Simplex);
                }
                return optResult.then(optResult => this.getLL(optResult.params)
                                      .then(llResult => {
                                          optResult.ndata = llResult.ndata;
                                          optResult.aic = llResult.aic;
                                          optResult.bic = llResult.bic;
                                          return optResult;
                                      }));
            });
        }
    };

    exports.MacTransfer = class MacTransfer {
        constructor(outCapMB, instances) {
            this.data = null;
            this.out = null;
            this.outCapMB = outCapMB || 24;
            this.instances = instances || 2;
        }
        setup(Nsig, Nchunk, chunkCap) {
            if ( (! this.data)/* || (this.data.bufferCount != Nsig)*/ || (this.data.bufferLength != chunkCap) ) {
                this.data && this.data.flush();
                this.data = new qubx.PromisedBuffers(2/*Nsig*/, chunkCap, 4, this.instances);
                const outCapChunks = this.outCapMB * 1024 * 1024 / (4*4*this.instances*chunkCap),
                      Nbatch = Math.max(1, Math.min(Nchunk, Math.floor(outCapChunks)))|0;
                this.out && this.out.flush();
                this.out = new qubx.PromisedBuffers(Nbatch, 4*chunkCap, 4, this.instances);
            }
        }
    };

    exports.MacWorker = class MacWorker {
        constructor(transfer) {
            this.worker = new qubx_workers.Worker({
                href: 'js/qubx_mac_worker.js'
                // hrefMin: 'js/qubx_mac_worker.min.js'
            });
            this.sched = new qubx.PromiseTickets();
            this.llWait = new qubx.CircularStock(4);
            this.transfer = transfer || new exports.MacTransfer();
            this.ownTransfer = (typeof(transfer) === 'undefined');
            this.layoutNum = -1;
            this.models = [];
        }
        destroy() {
            this.worker.destroy();
        }
        setModels(models) {
            this.models = models;
            this.worker.post({
                t: 'MODELS',
                models: models
            });
            return this;
        }
        setModel(iModel, model) {
            this.models[iModel] = model;
            this.worker.post({
                t: 'MODEL',
                iModel: iModel,
                model: model
            });
            return this;
        }
        setLayout(layout, getChunk) {
            // sampling (sec), signalNames[], delays[name], defaults[name], segments = [{chunks: [{dur, t0, n, stimuli: {name: {classes: [], durations: [], amps: []}}}]}]
            this.layingOut = true;
            this.getChunk = getChunk;
            ++this.layoutNum;
            layout.layoutNum = this.layoutNum;
            this.layout = layout;
            let Nchunk = 0,
                maxChunk = 0;
            layout.segments.forEach(seg => {
                Nchunk += seg.chunks.length|0;
                this.ownTransfer && seg.chunks.forEach(chunk => {
                    maxChunk = Math.max(maxChunk, chunk.n);
                });
            });
            this.Nchunk = Nchunk;
            this.worker.post({
                t: 'LAYOUT',
                layout: layout
            });
            this.ownTransfer && this.transfer.setup(layout.signalNames.length, Nchunk, maxChunk);
            let result = Promise.resolve(true),
                expired = false;
            layout.segments.forEach((seg, iSeg) => seg.chunks.forEach((chunk, iChunk) => {
                if ( expired ) { return; }
                result = result.then(() => {
                    return expired ? null : this.transfer.data.get();
                }).then(buffers => {
                    if ( expired ) { return {}; }
                    this.getChunk(buffers.map(buf => new Float32Array(buf)),
                                  iSeg, iChunk);
                    return this._sendChunk(iSeg, iChunk, layout.layoutNum, buffers);
                }).then(data => data.buffers && this.transfer.data.put(data.buffers))
                    .catch(qubx.reportErr)
                    .catch(data => {
                        data.buffers && this.transfer.data.put(data.buffers);
                        expired = true;
                    });
            }));
            
            if ( expired ) {
                return Promise.resolve({});
            }
            
            return result.then(() => {
                this.layingOut = false;
                while ( this.llWait.size ) {
                    const w = this.llWait.pop();
                    this.sched.resolve(w.tkt, this.getLL(w.params));
                }
            });
        }
        _sendChunk(iSeg, iChunk, layoutNum, buffers) {
            return this.worker.request({
                t: 'CHUNK',
                iSeg: iSeg,
                iChunk: iChunk,
                layoutNum: layoutNum,
                buffers: buffers
            }, buffers);
        }
        setRate(iModel, index, field, value) {
            this.worker.post({
                t: 'RATE',
                iModel: iModel,
                index: index,
                field: field,
                value: value
            });
            return this;
        }
        setChannelCount(iModel, cc) {
            this.worker.post({
                t: 'CC',
                iModel: iModel,
                channelCount: cc
            });
            return this;
        }
        setDelay(name, value) {
            this.worker.post({
                t: 'DELAY',
                name: name,
                value: value
            });
            return this;
        }
        setConstant(name, value) {
            this.worker.post({
                t: 'CONSTANT',
                name: name,
                value: value
            });
            return this;
        }
        optPars(modelOpts, optNchannel) {
            return this.worker.request({
                t: 'OPTPARS',
                modelOpts: modelOpts,
                optNchannel: optNchannel
            }).then(data => data.params);
        }
        getLL(params) {
            const pp = params ? params.slice() : params;
            if ( this.layingOut ) {
                return new Promise((resolve, reject) => this.llWait.push({
                    tkt: this.sched.register(resolve, reject),
                    params: pp
                }));
            }
            const msg = {t: 'LL'};
            if ( typeof(pp) !== 'undefined' ) {
                msg.params = pp;
            }
            return this.worker.request(msg);
        }
        _getOutputChunks(buffers, iSegs, iChunks, layoutNum) {
            return this.worker.request({
                t: 'OUTPUT',
                buffers: buffers,
                iSegs: iSegs,
                iChunks: iChunks,
                layoutNum: layoutNum
            }, buffers);
        }
        getOutput(onChunk) {
            const layoutNum = this.layoutNum,
                  workItems = [];
            let iSegs = [],
                iChunks = [];
            const send = () => {
                if ( iSegs.length === 0 ) {
                    return;
                }
                const iS = iSegs.slice(),
                      iC = iChunks.slice();
                workItems.push(
                    this.transfer.out.get()
                        .then(buffers => this._getOutputChunks(buffers, iS, iC, layoutNum))
                        .then(data => {
                            if ( data.layoutNum === this.layoutNum ) {
                                let i = 0;
                                for (; i<iS.length; ++i) {
                                    onChunk(iS[i], iC[i], new Float32Array(data.buffers[i]),
                                            this.layout.segments[iS[i]].chunks[iC[i]]);
                                }
                            }
                            this.transfer.out.put(data.buffers);
                        })
                );
                iSegs = [];
                iChunks = [];
            };
            const push = (iSeg, iChunk) => {
                iSegs.push(iSeg);
                iChunks.push(iChunk);
                if ( iSegs.length === this.transfer.out.bufferCount ) {
                    send();
                }
            };
            
            this.layout.segments.forEach((seg, iSeg) => seg.chunks.forEach((chunk, iChunk) => push(iSeg, iChunk)));
            send();
            
            return Promise.all(workItems);
        }
        getOptStd(invHessian) {
            return this.worker.request({
                t: 'OPTSTD',
                invHessian: invHessian
            }).then(data => data.models);
        }
    };
    
    return exports;
});

