/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"use strict";

const QUBX_LIB_MEMORY = 64*16*1024*1024;

importScripts('qubx_worker_require.js');
//importScripts('qubx_worker_require.min.js');

qubx_require(['qubx_lib', 'qubx_worker_util']).then(libs => {
    const [lib, util] = libs;

    function buildChunkIdl(chunk, signals) {
        const idlSignals = [];
        let iSig = 1;
        while ( iSig < signals.delays.length ) {
            const nm = signals.names[iSig];
            idlSignals.push(chunk.stimuli[nm]);
            ++iSig;
        }
        while ( iSig < signals.defaults.length ) {
            const nm = signals.names[iSig];
            idlSignals.push({
                classes: [0],
                durations: [1e3*chunk.dur],
                amps: [signals.defaults[iSig]]
            });
            ++iSig;
        }
        return idlSignals;
    }

    var worker = {
        sampled: {
            signals: new util.SignalMap(),
            macData: null,
            segData: [],
            layout: null,
            stimuli: [], // [segment as {chunks: {name: {classes, durations, amps}}}]
            totalN: 0,
            sampledT: null,
            
            reset: function() {
                if ( this.macData ) {
                    this.macData.destroy();
                    this.macData = null;
                    this.segData.forEach(seg => seg.chunks.forEach(chunk => chunk.idl.destroy()));
                    this.segData = [];
                }
            },
            setLayout: function(layout, models) {
                this.layout = layout;
                this.signals.update(models, layout);
                this.totalN = 0;
                const Nsig = this.signals.names.length;
                if ( Nsig ) {
                    let dataCap = 0,
                        chunkCap = 0;
                    layout.segments.forEach(seg => {
                        dataCap += seg.n;
                        seg.chunks.forEach(chunk => {
                            chunkCap = Math.max(chunkCap, chunk.n);
                        });
                    });
                    if ( this.macData ) {
                        this.macData.reset(Nsig-1, dataCap, chunkCap);
                    } else {
                        this.macData = new lib.MacData(Nsig-1, dataCap, chunkCap);
                    }
                    let sampledT = this.sampledT;
                    if ( (! layout.resampled) && ((! sampledT) || (sampledT.length < chunkCap)) ) {
                        sampledT = new Float32Array(chunkCap);
                        this.sampledT = sampledT;
                    }
                    const dt = layout.sampling;
                    let i = 0;
                    for (; i<chunkCap; ++i) {
                        sampledT[i] = i*dt;
                    }
                }
                this.segData.forEach(seg => seg.chunks.forEach(chunk => chunk.idl.destroy()));
                this.segData = [];
            },
            nextChunk: function(data) {
                const chunk = this.layout.segments[data.iSeg].chunks[data.iChunk];
                if ( data.iChunk === 0 ) {
                    this.macData.appendSegment();
                    this.segData.push({chunks: []});
                }
                const idl = new lib.IdlArrays(buildChunkIdl(chunk, this.signals));
                this.signals.delays.slice(1).forEach((d,i) => {
                    idl.signals[i].delay(d);
                });
                const offset = this.macData.appendChunk(data.iSeg, chunk.dur, chunk.t0, chunk.n,
                                                        this.layout.resampled ? new Float32Array(data.buffers[1]) : this.sampledT,
                                                        new Float32Array(data.buffers[0]), idl);
                this.segData[data.iSeg].chunks[data.iChunk] = {
                    dur: chunk.dur,
                    t0: chunk.t0,
                    n: chunk.n,
                    idl: idl,
                    offset: offset
                };
                this.totalN += chunk.n;
            },
            reload: function() {
                if ( this.macData ) {
                    this.macData.reset(this.signals.names.length-1, this.macData.dataCap, this.macData.chunkCap);
                    this.segData.forEach((seg, iSeg) => {
                        this.macData.appendSegment();
                        seg.chunks.forEach((chunk, iChunk) => {
                            chunk.idl.set(buildChunkIdl(this.layout.segments[iSeg].chunks[iChunk], this.signals));
                            this.signals.delays.slice(1).forEach((d,i) => {
                                chunk.idl.signals[i].delay(d);
                            });
                            this.macData.appendChunk(iSeg, chunk.dur, chunk.t0, chunk.n, null, null, chunk.idl);
                        });
                    });
                    return true;
                }
                return false;
            },
            readOutput: function(layoutNum, iSegs, iChunks, buffers) {
                var nthSeg = iSegs[iSegs.length-1];
                if ( (layoutNum === this.layout.layoutNum) &&
                     (nthSeg < this.segData.length) &&
                     (iChunks[iChunks.length-1] < this.segData[nthSeg].chunks.length) ) {
                    let i = 0;
                    for (; i<iSegs.length; ++i) {
                        const chunk = this.segData[iSegs[i]].chunks[iChunks[i]];
                        this.macData.readOutput(new Float32Array(buffers[i]), chunk.offset, chunk.n);
                    }
                } else {
                    /// do them the favor of zeroing the buffers?
                }
            }
        },
        
        calc: {
            macModels: [],
            macCalc: null,
            ampms: [],
            ratesms: [],
            staleRates: [],
            reset: function(preserveModel) {
                if ( this.macCalc ) {
                    this.macCalc.destroy(preserveModel);
                    this.macCalc = null;
                    this.ampms = [];
                    this.ratesms = [];
                    if ( ! preserveModel ) {
                        this.macModels = [];
                    }
                    this.staleRates = [];
                }
            },
            setup: function(models, macData, signals, modelOpts, optNchannel, dataOnly) {
                this.reset(dataOnly);
                macData.getStimclasses();
                var Nmodel = 0;
                models.forEach(m => {
                    if ( m ) {
                        ++Nmodel;
                    }
                });
                this.macCalc = new lib.MacLLCalc(macData, Nmodel);
                models.forEach((model, i) => {
                    if ( ! model ) {
                        return;
                    }
                    const opt = modelOpts[i] ? modelOpts[i].opt : {};
                    if ( ! this.macModels[i] ) {
                        this.macModels[i] = lib.modelFromJson(model, signals.indexMap);
                    }
                    this.macModels[i].iVoltage(signals.names.indexOf('Voltage'));
                    this.ampms[i] = new lib.StimAmps(this.macModels[i], macData.Nstimclass, signals.names.length, macData.stimclasses, macData.sc_frac);
                    const Nclass = this.macModels[i].Nclass();
                    if ( ! opt.cond ) {
                        if ( model.IeqFv ) {
                            this.ampms[i].fixAmp0();
                        }
                        let c = 0;
                        for (; c<Nclass; ++c) {
                            this.ampms[i].fixAmp(c);
                        }
                    }
                    if ( ! opt.variance ) {
                        if ( model.IeqFv ) {
                            this.ampms[i].fixVar0();
                        }
                        let c = 0;
                        for (; c<Nclass; ++c) {
                            this.ampms[i].fixVar(c);
                        }
                    }
                    this.ratesms[i] = new lib.StimRates(this.macModels[i], macData.Nstimclass, signals.names.length, macData.stimclasses, macData.sc_frac, 0.0);
                    if ( ! opt.rates ) {
                        this.ratesms[i].fixAll();
                    } else {
                        model.constraints.forEach(cns => {
                            const Nstate = model.states.length;
                            let i = 0;
                            for (; i<cns.States.length; ++i) {
                                if ( cns.States[i] >= Nstate ) {
                                    return; // skip invalid cns
                                }
                            }
                            if ( cns.Type === 'FixRate' ) {
                                this.ratesms[i].fixRate(cns.States[0], cns.States[1]);
                            } else if ( cns.Type === 'FixExp' ) {
                                this.ratesms[i].fixExp(cns.States[0], cns.States[1]);
                            } else if ( cns.Type === 'FixPress' ) {
                                this.ratesms[i].fixPress(cns.States[0], cns.States[1]);
                            } else if ( cns.Type === 'ScaleRate' ) {
                                this.ratesms[i].scaleRate(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                            } else if ( cns.Type === 'ScaleExp' ) {
                                this.ratesms[i].scaleExp(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                            } else if ( cns.Type === 'ScalePress' ) {
                                this.ratesms[i].scalePress(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                            }
                        });
                        if ( model.balanceLoops ) {
                            this.ratesms[i].balanceLoops();
                        }
                    }
                    this.macCalc.addModel(this.macModels[i], this.ampms[i], this.ratesms[i]);
                    this.staleRates[i] = false;
                });
                this.macCalc.optNchannel(optNchannel);
                this.macCalc.setupConstraints();
                const params = [],
                      np = this.macCalc.NfPar();
                let p = 0;
                for (; p<np; ++p) {
                    params.push(this.macCalc.fPars(p));
                }
                return params;
            },
            setRate: function(iModel, index, field, value) {
                if ( this.macModels[iModel] ) {
                    const r = worker.models[iModel].rates[index],
                          n = this.macModels[iModel].Nstate(),
                          ij = n * r.From + r.To;
                    if ( field === 'k0' ) {
                        this.macModels[iModel].K0(ij, value);
                    } else if ( field === 'k1' ) {
                        this.macModels[iModel].K1(ij, value);
                    } else if ( field === 'k2' ) {
                        this.macModels[iModel].K2(ij, value);
                    }
                    this.staleRates[iModel] = true;
                }
            },
            freshenRates: function() {
                let any = false;
                this.ratesms.forEach((ratesm, i) => {
                    if ( this.staleRates[i] ) {
                        ratesm.setupQQ();
                        ratesm.setupAA();
                        this.staleRates[i] = false;
                        any = true;
                    }
                });

            },
            setChannelCount: function(iModel, n) {
                if ( this.macModels[iModel] ) {
                    this.macModels[iModel].Nchannel(n);
                    this.macCalc.setupNchannel();
                }
            },
            do_fParsToModel: function(params) {
                const NfPar = this.macCalc ? this.macCalc.NfPar() : 0;
                params.forEach((p,i) => {
                    if ( i < NfPar ) {
                        this.macCalc.fPars(i, p);
                    } else {
                        console.log('Extra parameter',i,p);
                    }
                });
                this.macCalc.do_fParsToModel();
            },
            updateQMJ: function(models, modelOpts, optNchannel) {
                models.forEach((model, iModel) => {
                    if ( ! this.macModels[iModel] ) {
                        return;
                    }
                    const opt = modelOpts[iModel] ? modelOpts[iModel].opt : {};
                    if ( opt.rates ) {
                        const Nstate = model.states.length;
                        model.rates.forEach(r => {
                            r.k0 = this.macModels[iModel].K0(r.From*Nstate + r.To);
                            r.k1 = this.macModels[iModel].K1(r.From*Nstate + r.To);
                            r.k2 = this.macModels[iModel].K2(r.From*Nstate + r.To);
                        });
                    }
                    if ( opt.cond ) {
                        const ampName = model.IeqFv ? "Cond" : "Amp";
                        model.classes[0].Amp = this.macModels[iModel].amp0();
                        model.classes.forEach((c, i) => {
                            c[ampName] = this.macModels[iModel].amp(i);
                        });
                    }
                    if ( opt.variance ) {
                        const stdName = model.IeqFv ? "CondStd" : "Std";
                        model.classes[0].Std = Math.sqrt(this.macModels[iModel].var0());
                        model.classes.forEach((c, i) => {
                            c[stdName] = Math.sqrt(this.macModels[iModel].var(i));
                        });
                    }
                    if ( optNchannel ) {
                        model.channelCount = this.macModels[iModel].Nchannel();
                    }
                });
            },
            getStd: function(models, invHessian, modelOpts, optNchannel) {
                this.macCalc && this.macCalc.calcStd(invHessian);
                models.forEach((model, iModel) => {
                    const optimizing = modelOpts[iModel] ? modelOpts[iModel].opt : {};
                    if ( optNchannel ) {
                        model.channelCount_std = this.macModels[iModel].d_Nchannel();
                    }
                    const ampName = model.IeqFv ? "Cond" : "Amp",
                          stdName = model.IeqFv ? "CondStd" : "Std";
                    if ( optimizing.cond ) {
                        model.amp0_std = this.macModels[iModel].d_amp0();
                        model.classes.forEach((cls, c) => {
                            cls[ampName+"_std"] = this.macModels[iModel].d_amp[c];
                        });
                    }
                    if ( optimizing.variance ) {
                        model.var0_std = this.macModels[iModel].d_var0();
                        model.classes.forEach((cls, c) => {
                            cls[stdName+"_std"] = this.macModels[iModel].d_var[c]; /// sqrt?
                        });
                    }
                    if ( optimizing.rates ) {
                        const Nstate = this.macModels[iModel].Nstate();
                        model.rates.forEach((r,i) => {
                            var j = r.From*Nstate + r.To;
                            r.k0_std = this.macModels[iModel].d_K0(j);
                            r.k1_std = this.macModels[iModel].d_K1(j);
                            r.k2_std = this.macModels[iModel].d_K2(j);
                        });
                    }
                });
            }
        },

        models: [],
        layout: null,
        modelOpts: [],
        optNchannel: false,
        
        setupModels: function(dataOnly) { // dataOnly=true when model and signals are unchanged
            if ( ! this.sampled.macData ) {
                return [];
            }
            return this.calc.setup(this.models, this.sampled.macData, this.sampled.signals, this.modelOpts, this.optNchannel, dataOnly);
        },
        setModels: function(mm) {
            this.models = mm;
            mm.forEach(model => {
                if ( model ) {
                    model.constraints = [];
                }
            });
            if ( ! this.updateSignals() ) {
                this.setupModels();
            }
        },
        setModel: function(i, m) {
            this.models[i] = m;
            m.constraints = [];
            if ( ! this.updateSignals() ) {
                this.setupModels();
            }
        },
        setConstraints: function(modelOpts, optNchannel) {
            if ( ! this.sampled.macData ) {
                return [];
            }
            this.models.forEach(function(model, iModel) {
                if ( ! model ) {
                    return;
                }
                const opts = modelOpts[iModel] || {};
                model.constraints = opts.constraints || [];
                model.balanceLoops = !! opts.balanceLoops;
            });
            this.modelOpts = modelOpts;
            this.optNchannel = optNchannel;
            return this.setupModels(true);
        },
        cleanupModels: function(andData) {
            this.calc.reset();
            andData && this.sampled.reset();
        },
        setLayout: function(l) {
            this.layout = l;
            this.cleanupModels();
            this.sampled.setLayout(l, this.models);
        },
        reloadData: function(dataOnly) {
            this.sampled.reload() && this.setupModels(dataOnly);
        },
        updateSignals: function() {
            var updated = this.sampled.signals.update(this.models, this.layout);
            if ( updated ) {
                this.cleanupModels();
                this.reloadData();
                return true;
            }
            return false;
        },


        onMessage: function(event) {
            var model;
            console.log('w',event.data.t,event.data);
            /* try {
               console.log(JSON.stringify(event.data));
               } catch(e) {} */
            try {
                if ( event.data.t === 'STOP' ) {
                    this.cleanupModels(true);
                    postMessage(event.data);
                    close();
                } else if ( event.data.t === 'MODELS' ) {
                    this.setModels(event.data.models);
                } else if ( event.data.t === 'MODEL' ) {
                    this.setModel(event.data.iModel, event.data.model);
                } else if ( event.data.t === 'LAYOUT' ) {
                    this.setLayout(event.data.layout);
                } else if ( event.data.t === 'CHUNK' ) {
                    if ( (event.data.layoutNum !== this.layout.layoutNum) || ! this.sampled.macData ) {
                        event.data.tktErr = 'Layout expired';
                        postMessage(event.data, event.data.buffers);
                    } else {
                        this.sampled.nextChunk(event.data);
                        postMessage(event.data, event.data.buffers);
                        if ( (event.data.iSeg == (this.layout.segments.length-1)) && (event.data.iChunk == (this.layout.segments[event.data.iSeg].chunks.length-1)) ) {
                            this.setupModels();
                        }
                    }
                } else if ( event.data.t === 'DELAY' ) {
                    if ( this.layout ) {
                        this.layout.delay[event.data.name] = event.data.value;
                    }
                    const iSig = this.sampled.signals.indexOf(event.data.name);
                    if ( (iSig > 0) && (iSig < this.sampled.signals.delays.length) ) {
                        this.sampled.signals.delays[iSig] = event.data.value;
                    }
                    this.reloadData(true);
                } else if ( event.data.t === 'CONSTANT' ) {
                    if ( this.layout ) {
                        this.layout.defaults[event.data.name] = event.data.value;
                    }
                    const iSig = this.sampled.signals.indexOf(event.data.name);
                    if ( iSig > 0 ) {
                        const oldVal = this.sampled.signals.defaults[iSig];
                        if ( oldVal !== event.data.value ) {
                            this.sampled.signals.defaults[iSig] = event.data.value;
                            if ( iSig >= this.sampled.signals.delays.length ) {
                                this.reloadData(true);
                            }
                        }
                    }
                } else if ( event.data.t === 'RATE' && this.models[event.data.iModel] ) {
                    // index, field, value
                    model = this.models[event.data.iModel];
                    model.rates[event.data.index][event.data.field] = event.data.value;
                    if ( 0 <= ["Ligand", "Voltage", "Pressure"].indexOf(event.data.field) ) {
                        if ( ! this.updateSignals() ) {
                            this.setupModels();
                        }
                    } else {
                        this.calc.setRate(event.data.iModel, event.data.index, event.data.field, event.data.value);
                    }
                } else if ( event.data.t === 'CC' && this.models[event.data.iModel] ) {
                    model = this.models[event.data.iModel];
                    model.channelCount = event.data.channelCount;
                    this.calc.setChannelCount(event.data.iModel, event.data.channelCount);
                } else if ( event.data.t === 'OPTPARS' ) {
                    postMessage({
                        tkt: event.data.tkt,
                        params: this.setConstraints(event.data.modelOpts, event.data.optNchannel)
                    });
                } else if ( event.data.t === 'LL' ) {
                    if ( ! this.calc.macCalc ) {
                        postMessage({
                            tkt: event.data.tkt,
                            tktErr: 'Not ready'
                        });
                    } else {
                        const response = {
                            t: 'LL',
                            tkt: event.data.tkt
                        };
                        this.calc.freshenRates();
                        if ( event.data.params ) {
                            this.calc.do_fParsToModel(event.data.params);
                            this.calc.updateQMJ(this.models, this.modelOpts, this.optNchannel);
                            //console.log('get LL of params',event.data.params, 'rates', model.rates.map(function(r) { return r.k0; }));
                            response.models = this.models;
                        }
                        response.resultCode = this.calc.macCalc.run();
                        response.ll = this.calc.macCalc.ll();
                        response.ndata = this.sampled.totalN;
                        postMessage(response);
                    }
                } else if ( event.data.t === 'OUTPUT' ) {
                    this.sampled.readOutput(event.data.layoutNum, event.data.iSegs, event.data.iChunks, event.data.buffers);
                    postMessage(event.data, event.data.buffers);
                } else if ( event.data.t === 'OPTSTD' ) {
                    var mdlOut = this.models.map(model => model && JSON.parse(JSON.stringify(model)));
                    this.calc.getStd(mdlOut, event.data.invHessian, this.modelOpts, this.optNchannel);
                    postMessage({
                        tkt: event.data.tkt,
                        models: mdlOut
                    });
                }
            } catch (err) {
                if ( event.data.tkt ) {
                    event.data.tktErr = {
                        message: err.message,
                        stack: err.stack
                    };
                    console.log(err.stack);
                    postMessage(event.data);
                }
            }
        },
        setup: function() {
            self.addEventListener('message', this.onMessage.bind(this));
            postMessage({t: 'READY'});
            return this;
        }
    }.setup();
}).catch(err => postMessage({t: 'LOADERR', err}));

/*
MODELS models
MODEL iModel, model
RATE iModel, index, field, value
CC iModel, channelCount

LAYOUT layout: {layoutNum, sampling, segments: [{chunks: {n, t0, dur}}], signalNames: [name], defaults: {name: value}, delay: {name: value}}
 - signalNames: ["Current", stim0, stim1, ...]; implying that chunk.stimuli[stim*] is defined for all chunks
CHUNK layoutNum, iSeg, dur, buffers, t0, stimuli: {name: {classes, durations, amps}}
 -> CHUNK
DELAY name, value // msec? sec?
CONSTANT name, val

LL tkt [, params]
 -> LL resultCode, ll, tkt [, tktErr]; [, model] if params

OPTPARS tkt, modelOpts: {constraints, balanceLoops, opt}, optNchannel
 -> OPTPARS params

OPTSTD tkt, invHessian[flat Nstate*Nstate array]
 -> OPTSTD models

OUTPUT layoutNum, buffers, iSegs, iChunks
 -> OUTPUT
 */
