/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_model", ["qubx"], qubx => {
    'use strict';
    
    const exports = {};

    let seqNumber = 0;

    exports.stateFields = {
        Class: {name: 'Class', accept: qubx.acceptIntAtLeast(0), format: qubx.formatStr, widthChars: 3},
        Pr: {name: 'Pr', accept: qubx.acceptFloatBetween(0, 1), format: qubx.formatStr, widthChars: 7},
        x: {name: 'x', accept: qubx.acceptFloatBetween(0, 100), format: qubx.formatStr, widthChars: 7},
        y: {name: 'y', accept: qubx.acceptFloatBetween(0, 100), format: qubx.formatStr, widthChars: 7},
        Label: {name: 'Label', defval: "", accept: qubx.acceptStr, format: qubx.formatStr, widthChars: 7}
    };
    exports.stateFieldOrder = ['Class', 'Pr', 'x', 'y', 'Label'];
    
    exports.classFields = {
        Amp: {name: 'Amp', defval: 0.0, accept: qubx.acceptFloat, widthChars: 7},
        Std: {name: 'Std', defval: 0.1, accept: qubx.acceptFloatAtLeast(0), widthChars: 7},
        Cond: {name: 'Cond', defval: 0.0, accept: qubx.acceptFloat, widthChars: 7},
        CondStd: {name: 'CondStd', defval: 1.0, accept: qubx.acceptFloatAtLeast(0), widthChars: 7}
    };
    exports.classFieldOrder = ['Amp', 'Std', 'Cond', 'CondStd'];
    
    exports.rateFields = {
        From: {name: 'From', accept: qubx.acceptNone, format: qubx.formatStr, widthChars: 3},
        To: {name: 'To', accept: qubx.acceptNone, format: qubx.formatStr, widthChars: 3},
        k0: {name: 'k0', defval: 1.0, accept: qubx.acceptFloatGreaterThan(0.0), format: qubx.formatK0, widthChars: 7},
        k1: {name: 'k1', defval: 0.0, accept: qubx.acceptFloat, format: qubx.formatK0, widthChars: 7},
        k2: {name: 'k2', defval: 0.0, accept: qubx.acceptFloat, format: qubx.formatK0, widthChars: 7},
        Ligand: {name: 'Ligand', defval: "", accept: qubx.acceptStr, format: qubx.formatStr, widthChars: 10},
        Voltage: {name: 'Voltage', defval: "", accept: qubx.acceptStr, format: qubx.formatStr, widthChars: 10},
        Pressure: {name: 'Pressure', defval: "", accept: qubx.acceptStr, format: qubx.formatStr, widthChars: 10}
    };
    exports.rateFieldOrder = ['From', 'To', 'k0', 'k1', 'k2', 'Ligand', 'Voltage', 'Pressure'];
    
    exports.constraintFields = {
        Type: {name: 'Type', defval: 'FixRate', choices: ['FixRate', 'ScaleRate', 'FixExp', 'ScaleExp', 'FixPress', 'ScalePress']},
        States: {name: 'States', defval: [0, 1], accept: qubx.acceptList(qubx.acceptIntAtLeast(0)), format: qubx.formatList()}
    };
    exports.constraintFieldOrder = ['Type', 'States'];
    
    exports.Model = class Model {
        constructor(config) {
            this.undo = new qubx.Undo();
            this.undo.ev.change.sub(() => { this.changed = true; });
            this.states = new qubx.Table({label: 'States'});
            exports.stateFieldOrder.forEach(fieldName => {
                this.states.addField(exports.stateFields[fieldName]);
            });
            this.states.ev.inserted.sub(this._onInsertedState.bind(this));
            this.states.ev.removing.sub(this._onRemovingState.bind(this));
            this.states.ev.set.sub(this._onSetState.bind(this));
            this.states_undo = new qubx.TableUndo(this.states, this.undo);
            this.rates = new qubx.Table({label: 'Rates'});
            exports.rateFieldOrder.forEach(fieldName => {
                this.rates.addField(exports.rateFields[fieldName]);
            });
            this.rates.ev.inserted.sub(this._onInsertedRate.bind(this));
            this.rates.ev.removing.sub(this._onRemovingRate.bind(this));
            this.rates.ev.removed.sub(this._onRemovedRate.bind(this));
            this.rates.ev.set.sub(this._onSetRate.bind(this));
            this.rates_undo = new qubx.TableUndo(this.rates, this.undo);
            this.classes = new qubx.Table({label: 'Classes'});
            exports.classFieldOrder.forEach(fieldName => {
                this.classes.addField(exports.classFields[fieldName]);
            });
            this.classes_undo = new qubx.TableUndo(this.classes, this.undo);
            this.constraints = new qubx.Table({label: 'Constraints'});
            exports.constraintFieldOrder.forEach(fieldName => {
                this.constraints.addField(exports.constraintFields[fieldName]);
            });
            this.changed = false;
            this._channelCount = 1;
            this._IeqFv = false;
            this._balanceLoops = true;
            this._vRev = 0.0;
            this._variables = [];
            this._varcount = {};
            this._rate_conn = {};
            this.active = true;
            this.ev = qubx.makeEvents(['setChannelCount', 'setIeqFv', 'setVRev', 'setBalanceLoops', 'changeVariables', 'active']);
        }
        clear() {
            this.rates.clear();
            this.states.clear();
            this.classes.clear();
            return this;
        }
        setActive(b) {
            b = !!b;
            if ( this.active === b ) {
                return;
            }
            this.active = b;
            this.ev.active.trigger(b);
        }
        channelCount(n, noSeal, undoing) {
            if ( n ) {
                if ( n === this._channelCount ) {
                    return this;
                }
                const prev = this._channelCount;
                this._channelCount = n;
                if ( ! undoing ) {
                    this.undo.pushUndo(() => this.channelCount(prev, false, true),
                                       () => this.channelCount(n, false, true));
                    if ( ! noSeal ) {
                        this.undo.sealUndo("Set channel count");
                    }
                }
                this.ev.setChannelCount.trigger(n);
                return this;
            } else {
                return this._channelCount;
            }
        }
        IeqFv(x, undoing) {
            if ( typeof(x) !== 'undefined' ) {
                x = !!x;
                if ( x === this._IeqFv ) {
                    return this;
                }
                this._IeqFv = x;
                if ( ! undoing ) {
                    this.undo.pushUndo(() => this.IeqFv(! x, true),
                                       () => this.IeqFv(x, true));
                    this.undo.sealUndo("Set I=f(v)");
                }
                this.ev.setIeqFv.trigger(x);
                this._onSetRate({}, {row: -1, column: 'Voltage', value: x?'Voltage':'', prev:x?'':'Voltage'});
                return this;
            } else {
                return this._IeqFv;
            }
        }
        vRev(x, undoing) {
            if ( typeof(x) !== 'undefined' ) {
                if ( x === this._vrev ) {
                    return this;
                }
                const prev = this._vRev;
                this._vRev = x;
                if ( ! undoing ) {
                    this.undo.pushUndo(() => this.vRev(prev, true),
                                       () => this.vRev(x, true));
                    this.undo.sealUndo("Set Vrev");
                }
                this.ev.setVRev.trigger(x);
                return this;
            } else {
                return this._vRev;
            }
        }
        balanceLoops(x, undoing) {
            if ( typeof(x) !== 'undefined' ) {
                x = !!x;
                if ( x === this._balanceLoops ) {
                    return this;
                }
                this._balanceLoops = x;
                if ( ! undoing ) {
                    this.undo.pushUndo(() => this.balanceLoops(! x, true),
                                       () => this.balanceLoops(x, true));
                    this.undo.sealUndo("Set Balance Loops");
                }
                this.ev.setBalanceLoops.trigger(x);
                return this;
            } else {
                return this._balanceLoops;
            }
        }
        formatStateAsLbl(i) {
            return this.states.get(i, 'Label') || (''+i);
        }
        _gen_rate_conn() {
            this._rate_conn = {};
            for (const row of this.rates) {
                this._rate_conn[`${row.From}_${row.To}`] = row.Index;
            }
        }
        rate_connecting(a, b) {
            const i = this._rate_conn[`${a}_${b}`];
            if ( typeof(i) === 'undefined' ) {
                return -1;
            }
            return i;
        }
        getStimulus() {
            const stimulus = {};
            var name;
            for (const r of this.rates) {
                ['Ligand', 'Voltage', 'Pressure'].forEach((field, notLigand) => {
                    const name = r[field];
                    if ( name ) {
                        stimulus[name] = notLigand ? 0.0 : 1.0
                    }
                });
            }
            if ( this._IeqFv ) {
                stimulus['Voltage'] = 0.0;
            }
            return stimulus;
        }
        _rem_var(name) {
            const ix = this._varcount[name][0];
            delete this._varcount[name];
            this._variables.splice(ix, 1);
            for (const nm of Object.keys(this._varcount)) {
                const ix2 = this._varcount[nm][0];
                if ( ix2 > ix ) {
                    this._varcount[nm][0] = ix2 - 1;
                }
            }
        }
        _onSetRate(info, undoing) {
            if ( ['Ligand', 'Voltage', 'Pressure'].indexOf(info.column) !== -1 ) {
                if ( info.value == info.prev ) {
                    return;
                }
                if ( info.prev ) {
                    const ct = this._varcount[info.prev][1];
                    if ( ct > 1 ) {
                        this._varcount[info.prev][1] = ct - 1;
                    } else {
                        this._rem_var(info.prev);
                    }
                }
                if ( info.value ) {
                    if ( info.value in this._varcount ) {
                        const ct = this._varcount[info.value][1];
                        this._varcount[info.value][1] = ct + 1;
                    } else {
                        const ix = this._variables.length;
                        this._variables.push(info.value);
                        this._varcount[info.value] = [ix, 1];
                    }
                }
                this.ev.changeVariables.trigger(this._variables);
            }
        }
        _onInsertedRate(row, undoing) {
            this._gen_rate_conn();
            if ( 0 < /* if (number of variables added by the following function):*/
                 [row.Ligand, row.Voltage, row.Pressure].map(name => {
                     if ( name ) {
                         if ( name in this._varcount ) {
                             this._varcount[name][1] += 1;
                             return 0;
                         } else {
                             const ix = this._variables.length;
                             this._variables.push(name);
                             this._varcount[name] = [ix, 1];
                             return 1;
                         }
                     }
                 }).reduce((sum, one) => sum+one, 0) ) {
                this.ev.changeVariables.trigger(this._variables);
            }
        }
        _onRemovingRate(row, undoing) {
            if ( 0 < [row.Ligand, row.Voltage, row.Pressure].map(name => {
                if ( name ) {
                    const ct = this._varcount[name][1];
                    if ( ct > 1 ) {
                        this._varcount[name][1] = ct - 1;
                        return 0;
                    } else {
                        this._rem_var(name);
                        return 1;
                    }
                }
            }).reduce((sum, one) => sum+one, 0) ) {
                this.ev.changeVariables.trigger(this._variables);
            }
        }
        _onRemovedRate(row, undoing) {
            this._gen_rate_conn();
        }
        _onInsertedState(row, undoing) {
            this.ensureClass(row.Class);
        }
        _onRemovingState(row, undoing) {
            var f, t, j;
            for (j=this.rates.size-1; j>=0; --j) {
                f = this.rates.get(j, 'From');
                t = this.rates.get(j, 'To');
                if ( (row.Index === f) || (row.Index === t) ) {
                    if ( (j%2) === 0 ) {
                        this.rates.remove(j+1);
                        this.rates.remove(j);
                    }
                } else {
                    if ( f > row.Index ) {
                        this.rates.set(j, 'From', f-1);
                    }
                    if ( t > row.Index ) {
                        this.rates.set(j, 'To', t-1);
                    }
                }
            }
        }
        _onSetState(info, undoing) {
            if ( info.column === 'Class' ) {
                this.ensureClass(info.value);
            }
        }
        ensureClass(c) {
            while ( c >= this.classes.size ) {
                this.classes.append({Amp: this.classes.size,
                                     Std: this.classes.size ? this.classes.get(this.classes.size-1, 'Std') : 0.1,
                                     Cond: 10*this.classes.size,
                                     CondStd: this.classes.size ? this.classes.get(this.classes.size-1, 'CondStd') : 0.1});
            }
        }
        toJson() {
            return {
                channelCount: this._channelCount,
                IeqFv: this._IeqFv,
                balanceLoops: this._balanceLoops,
                vRev: this._vRev,
                states: this.states.toJson(),
                classes: this.classes.toJson(),
                rates: this.rates.toJson(),
                constraints: this.constraints.toJson(),
                modelName: this.modelName || '',
                modelAbout: this.modelAbout || ''
            };
        }
        fromJson(obj) {
            if ( 'channelCount' in obj ) {
                this.channelCount(obj.channelCount);
            }
            if ( 'IeqFv' in obj ) {
                this.IeqFv(obj.IeqFv);
            }
            if ( 'balanceLoops' in obj ) {
                this._balanceLoops = obj.balanceLoops; // this.balanceLoops(obj.balanceLoops);
            }
            if ( 'vRev' in obj ) {
                this.vRev(obj.vRev);
            }
            if ( 'classes' in obj ) {
                this.classes.fromJson(obj.classes);
            } else {
                this.classes.clear();
            }
            if ( 'states' in obj ) {
                this.states.fromJson(obj.states);
            } else {
                this.states.clear();
            }
            if ( 'rates' in obj ) {
                this.rates.fromJson(obj.rates);
                const n = this.rates.length;
                let i = 0;
                for (; i<n; ++i) {
                    const r = this.rates.get(i);
                    if ( r.Voltage ) {
                        this.rates.set(i, "Voltage", "Voltage");
                    }
                    if ( r.Pressure ) {
                        this.rates.set(i, "Pressure", "Pressure");
                    }
                }
            }
            if ( 'constraints' in obj ) {
                this.constraints.fromJson(obj.constraints);
            } else {
                this.constraints.clear();
            }
            this.modelName = obj.modelName || '';
            this.modelAbout = obj.modelAbout || '';
            return this;
        }
        makeQ(stimulus) {
            stimulus = stimulus || {};
            const Q = this.states.entries.map(state => this.states.entries.map(state => 0)),
                  n = this.states.size;
            for (const rate of this.rates) {
                let k = rate.k0;
                if ( rate.Ligand && (rate.Ligand in stimulus) ) {
                    k *= stimulus[rate.Ligand];
                }
                let kExp = 0.0;
                if ( rate.Voltage && (rate.Voltage in stimulus) ) {
                    kExp += rate.k1 * stimulus[rate.Voltage];
                }
                if ( rate.Pressure && (rate.Pressure in stimulus) ) {
                    kExp += rate.k2 * stimulus[rate.Pressure];
                }
                k *= Math.exp(kExp);
                Q[rate.From][rate.To] = k;
            }
            let i = 0;
            for (; i<n; ++i) {
                let sum = 0.0;
                let j = 0;
                for (; j<n; ++j) {
                    sum += Q[i][j];
                }
                Q[i][i] = - sum;
            }
            return Q;
        }
        acceptStateIndex(s) {
            const n = this.states.size;
            let i = 0;
            for (; i<n; ++i) {
                if ( this.states.get(i, 'Label') === s ) {
                    break;
                }
            }
            if ( i === n ) {
                i = qubx.acceptInt(s);
            }
            if ( i < 0 || i >= n ) {
                throw new TypeError('Expected a number between '+0+' and '+(n-1));
            }
            return i;
        }
    };

    exports.QMFtoJson = function QMFtoJson(qmf) {
        const mj = {
            channelCount: qmf.ch.ChannelCount.data[0],
            IeqFv: qmf.ch.Properties && qmf.ch.Properties.ch.IeqFv && qmf.ch.Properties.ch.IeqFv.data[0],
            vRev: qmf.ch.Vrev && qmf.ch.VRev.data[0],
            balanceLoops: qmf.ch.BalanceLoops && qmf.ch.BalanceLoops.data[0],
            states: qmf.ch.States.children.map(s => ({
                Class: s.ch.Class.data[0],
                Label: s.ch.Label && s.ch.Label.data || "",
                Pr: s.ch.Pr.data[0],
                x: s.ch.x.data[0],
                y: s.ch.y.data[0]
            })),
            classes: [],
            rates: [],
            constraints: qmf.ch.Constraints.children.map(c => ({
                Type: c.ch.Type && c.ch.Type.data || "",
                States: c.ch.States && c.ch.States.data || []
            }))
        };
        for (const r of qmf.ch.Rates.children) {
            mj.rates.push({
                From: r.ch.States.data[0],
                To: r.ch.States.data[1],
                k0: r.ch.k0.data[0],
                k1: r.ch.k1.data[0],
                k2: r.ch.k2 ? r.ch.k2.data[0] : 0.0,
                Ligand: r.ch.P.data[0] && r.ch.PNames.children[0].data || "",
                Voltage: r.ch.Q.data[0] && r.ch.QNames.children[0].data || "",
                Pressure: r.ch.Pressure && r.ch.Pressure.data[0] && r.ch.PressureNames.children[0].data || ""
            });
            mj.rates.push({
                From: r.ch.States.data[1],
                To: r.ch.States.data[0],
                k0: r.ch.k0.data[1],
                k1: r.ch.k1.data[1],
                k2: r.ch.k2 ? r.k2.data[1] : 0.0,
                Ligand: r.ch.P.data[1] && r.ch.PNames.children[1].data || "",
                Voltage: r.ch.Q.data[1] && r.ch.QNames.children[1].data || "",
                Pressure: r.ch.Pressure && r.ch.Pressure.data[1] && r.ch.PressureNames.children[1].data || ""
            });
        }
        const aa = qmf.ch.Amps,
              ss = qmf.ch.Stds,
              cc = qmf.ch.Conds,
              cs = qmf.ch.CondStds;
        let i = 0;
        for (; i<aa.length; ++i) {
            mj.classes.push({
                Amp: aa[i],
                Std: ss[i],
                Cond: cc[i],
                CondStd: cs[i]
            });
        }
        return mj;
    };

    exports.Models = class Models {
        constructor() {
            this.models = [];
            this.ev = qubx.makeEvents(['insert', 'remove', 'activate', 'setChannelCount', 'setIeqFv', 'setVRev',
                                       'insertState', 'removeState', 'setState', 'setClass', 'insertRate', 'removeRate', 'setRate']);
        }
        insert(i, model) {
            i = (typeof(i) !== 'undefined') ? (i|0) : this.models.length;
            const rec = {
                model: model || new exports.Model(),
                active: true,
                index: i,
                seq: ++seqNumber
            };
            const m = rec.model;
            rec.__modelsOnSetChannelCount = m.ev.setChannelCount.sub(n => m.active && this.ev.setChannelCount.trigger(rec, n));
            rec.__modelsOnSetIeqFv = m.ev.setIeqFv.sub(b => m.active && this.ev.setIeqFv.trigger(rec, b));
            rec.__modelsOnSetVRev = m.ev.setVRev.sub(v => m.active && this.ev.setVRev.trigger(rec, v));
            rec.__modelsOnInsertState = m.states.ev.inserted.sub(row => m.active && this.ev.insertState.trigger(rec, row));
            rec.__modelsOnRemoveState = m.states.ev.removed.sub(row => m.active && this.ev.removeState.trigger(rec, row));
            rec.__modelsOnSetState = m.states.ev.set.sub(info => m.active && this.ev.setState.trigger(rec, info));
            rec.__modelsOnSetClass = m.classes.ev.set.sub(info => m.active && this.ev.setClass.trigger(rec, info));
            rec.__modelsOnInsertRate = m.rates.ev.inserted.sub(row => m.active && this.ev.insertRate.trigger(rec, row));
            rec.__modelsOnRemoveRate = m.rates.ev.removed.sub(row => m.active && this.ev.removeRate.trigger(rec, row));
            rec.__modelsOnSetRate = m.rates.ev.set.sub(info => m.active && this.ev.setRate.trigger(rec, info));
            this.models.splice(i, 0, rec);
            let j = i + 1;
            for (; j<this.models.length; ++j) {
                this.models[j].index = j;
            }
            this.ev.insert.trigger(rec, i);
            m.rates.entries.forEach(row => this.ev.insertRate.trigger(rec, row));
            return rec;
        }
        remove(i) {
            i = i|0;
            if ( i >= 0 ) {
                const rec = this.models[i];
                const m = rec.model;
                m.setActive(false);
                m.states.ev.inserted.unsub(rec.__modelsOnInsertState);
                m.states.ev.removed.unsub(rec.__modelsOnRemoveState);
                m.states.ev.set.unsub(rec.__modelsOnSetState);
                rec.__modelsOnInsertState = rec.__modelsOnRemoveState = rec.__modelsOnSetState = null;
                m.classes.ev.set.unsub(rec.__modelsOnSetClass);
                rec.__modelsOnSetClass = null;
                m.rates.ev.inserted.unsub(rec.__modelsOnInsertRate);
                m.rates.ev.removed.unsub(rec.__modelsOnRemoveRate);
                m.rates.ev.set.unsub(rec.__modelsOnSetRate);
                rec.__modelsOnInsertRate = rec.__modelsOnRemoveRate = rec.__modelsOnSetRate = null;
                m.ev.setChannelCount.unsub(rec.__modelsOnSetChannelCount);
                m.ev.setIeqFv.unsub(rec.__modelsOnSetIeqFv);
                m.ev.setVRev.unsub(rec.__modelsOnSetVRev);
                rec.__modelsOnSetChannelCount = rec.__modelsOnSetIeqFv = rec.__modelsOnSetVRev = null;
                this.ev.remove.trigger(rec, i);
                this.models.splice(i, 1);
                let j = i;
                for (; j<this.models.length; ++j) {
                    this.models[j].index = j;
                }
            }
            return this;
        }
        activate(i, active) {
            i = i|0;
            active = !!active;
            if ( active !== this.models[i].model.active ) {
                this.models[i].model.setActive(active);
                this.ev.activate.trigger(this.models[i], i, active);
                if ( active ) {
                    this.models[i].model.rates.entries.forEach(row => this.ev.insertState.trigger(this.models[i], row, true));
                    this.models[i].model.rates.entries.forEach(row => this.ev.insertRate.trigger(this.models[i], row, true));
                }
            }
        }
        _onSomeChange() {
            this.ev.change.trigger();
        }
    };

    exports.modelToLandscape = function modelToLandscape(model, opts) {
        if ( ! model ) {
            return {
                zeroState: opts.zeroState || 0,
                states: [],
                rates: []
            };
        }
        const Q = opts.Q || model.makeQ(),
              landscape = opts.landscape || {},
              kRef = opts.kRef || 1000000,
              logKref = Math.log(Math.max(1, kRef));
        let zeroState = opts.zeroState;
        if ( typeof(zeroState) === 'undefined' ) {
            const statesSort = model.states.entries.slice();
            statesSort.sort((a,b) => a.x - b.x);
            zeroState = statesSort[0].Index;
        }
        const states = landscape.states || model.states.entries.map(st => Object.assign({}, st));
        const rates = landscape.rates || model.rates.entries.map(rate => Object.assign({}, rate));
        states.forEach(st => {
            st.lsLevel = undefined;
        });
        rates.forEach(rate => {
            lsLevel: undefined;
        });
        (function visitState(st, level) {
            level = level || 0.0;
            states[st].lsLevel = level;
            rates.forEach(rate => {
                if ( (rate.From === st) && (typeof(rate.lsLevel) === 'undefined') ) {
                    rate.lsLevel = logKref + level - Math.log(Q[rate.From][rate.To]);
                    visitState(rate.To, level + Math.log(Q[rate.To][rate.From] / Q[rate.From][rate.To]));
                }
            });
        })(zeroState);
        return {
            zeroState,
            states,
            rates
        };
    };
    
    return exports;
});
