/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define('qubx_model_math', ['qubx', 'qubx_fit', 'qubx_model', 'qubx_lib'], (qubx, qubx_fit, qubx_model, qubx_lib) => {
    'use strict';

    const exports = {};
    var lib;

    qubx_lib.then(loaded => {
        lib = loaded;
    });

    exports.findCommitor = function findCommitor(mdl, tmeas_ms, stimulus) {
        if ( ! lib ) {
            return {
                stateOrder: [],
                commitor: 0,
                Pt: [],
                residual: 0,
                iterations: 0,
                message: 'Waiting for the matrix exponential function to load...'
            };
        }
        let output = "";
        const states = exports.findLinearStates(mdl);
        // warn about out-of-order states. we can't just translate to the true state numbers because the result could be midway between two states.
        let i = 0;
        for (; i<states.length; ++i) {
            if ( i !== states[i] ) {
                output += "<div>States are out of order. The answer will be a zero-based index into this array:<br>";
                output += JSON.stringify(states);
                output += "</div>";
                break;
            }
        }
        
        // get model matrices: Q (rate) and A (transition probability after tmeas)
        const Q = mdl.makeQ(stimulus),
              A = lib.expm(Q, tmeas_ms*1e-3);
        
        // we use a 1d numerical search to find the optimal commitor, so functions are defined in terms of par = [fractionalCommitorIndex]
        // the optimal commitor has Pt[first state] == Pt[last state]
        function calcPt(par) { // Pt = P0(commitor) * A
            // separate commitor param into integer and fraction parts
            const i = Math.floor(par[0]),
                  frac = par[0] - i,
                  // use i,frac to initialize state probability vector
                  P0 = Q.map(row => 0.0);
            P0[states[i]] = 1 - frac;
            if ( frac ) {
                P0[states[i+1]] = frac;
            }
            // advance probabilities by tmeas
            return vxm(P0, A);
        }
        function testCommitor(par) {
            // quantity to minimize: squared difference between start- and end-state probability
            const Pt = calcPt(par),
                  d = Pt[states[0]] - Pt[states[states.length-1]];
            return d*d;
        }
        // search for best commitor using simplex
        const startParam = [0.5*(states.length-1)],
              fit = qubx_fit.simplex(startParam, {}, bounded(testCommitor, 0, states.length-1)),
              Pt = calcPt(fit.param);
        output += '<div>Commitor: '+fit.param[0].toPrecision(2)+'</div>';
        output += '<div>P('+states[0]+')='+Pt[states[0]].toPrecision(5)+'; ';
        output += 'P('+states[states.length-1]+')='+Pt[states[states.length-1]].toPrecision(5)+'</div>';
        output += '<div>Residual: '+fit.min.toPrecision(3)+' in '+fit.iterations+' iterations</div>';
        return {
            stateOrder: states,
            commitor: fit.param[0],
            Pt: Pt,
            residual: fit.min,
            iterations: fit.iterations,
            message: output
        };
    };
    
    function vxm(vec, mat) {
        if ( vec.length === 0 ) {
            return vec;
        }
        const out = [],
              m = mat.length,
              n = mat[0] && mat[0].length || 0;
        let i = 0;
        for (; i<n; ++i) {
            let s = 0.0,
                j = 0;
            for (; j<m; ++j) {
                s += vec[j] * mat[j][i];
            }
            out.push(s);
        }
        return out;
    }
    
    function bounded(func, lo, hi) {
        return par => ((par[0] < lo) || ( (typeof(hi) !== 'undefined') && (par[0] > hi) )) ? 1e10 : func(par);
    }
    
    function adjacencyList(model) {
        const connTo = model.states.entries.map(() => []);
        let i = 0;
        for (; i<model.rates.size; ++i) {
            connTo[model.rates.get(i, 'From')].push(model.rates.get(i, 'To'));
        }
        return connTo;
    }
    
    exports.findLinearStates = function findLinearStates(model) {
        // returns a list of state indices which are linearly connected, in order
        const Ns = model.states.size,
              connTo = adjacencyList(model);
        let s = 0;
        // find first terminal state: only one neighbor
        while ( (s < Ns) && (connTo[s].length != 1) ) {
            s += 1;
        }
        const linear = [s];
        while ( connTo[s].length ) {
            // visit the neighbor, if any, and remove the reverse connection
            const s2 = connTo[s][0];
            linear.push(s2);
            const iS = connTo[s2].indexOf(s);
            connTo[s2].splice(iS, 1);
            s = s2;
        }
        return linear;
    };
    
    exports.buildLinearModel = function buildLinearModel(model, n, ffwd, fbwd, sfwd, sbwd) {
        model.clear();
        const dx = 100.0 / (n+1);
        let i = 0;
        for (; i<n; ++i) {
            model.states.append({
                Class: (i == (n-1)) ? 1 : 0,
                x: (i+1)*dx,
                y: 50
            });
            if ( i ) {
                model.rates.append({
                    From: i-1,
                    To: i,
                    k0: ffwd
                });
                model.rates.append({
                    From: i,
                    To: i-1,
                    k0: fbwd
                });
            }
        }
        model.rates.set(0, 'k0', sfwd);
        model.rates.set(model.rates.size-1, 'k0', sbwd);
        model.undo.sealUndo("Build model");
    };

    const pathYminimaDefs = {
        divs: 128,
        refine: 3,
        sign: -1 // smallest graphed mathematically has highest gfx y coord
    };
    exports.pathYminima = function pathYminima(path, opts) { // path: svg <path> element
        opts = Object.assign({}, pathYminimaDefs, opts);
        const pathLen = path.getTotalLength();
        let dp = pathLen / opts.divs;
        let minima = [];
        let lastY = opts.sign*1e20;
        let dec = true;
        let i = 0, p = 0.0;
        for (; i<opts.divs; ++i, p += dp) {
            let y = path.getPointAtLength(p).y;
            if ( (opts.sign*y >= opts.sign*lastY) && (dec || (i === 1)) ) {
                minima.push({p: p-dp, y: lastY});
            }
            dec = opts.sign*y < opts.sign*lastY;
            lastY = y;
        }
        if ( dec ) {
            minima.push({p: p-dp, y: lastY});
        }
        for (i=0; i<opts.refine; ++i) {
            dp *= 0.5;
            minima = minima.map(m => refineYmin(path, dp, m, opts.sign))
        }
        return minima;
    }
    function refineYmin(path, dp, m, sign) {
        const r = Object.assign({}, m);
        let y = path.getPointAtLength(m.p-dp).y;
        if ( sign*y < sign*r.y ) {
            r.p = m.p-dp;
            r.y = y;
        }
        y = path.getPointAtLength(m.p+dp).y;
        if ( sign*y < sign*r.y ) {
            r.p = m.p+dp;
            r.y = y;
        }
        return r;
    }

    return exports;
});
