/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_optimize", [], () => {
    "use strict";

    const exports = {};

    function sqr(x) { return x*x; }
    
    function sumColumns(mat) {
        const n = mat.length,
              m = n ? mat[0].length : 0,
              sums = [];
        let j = 0;
        for (; j<m; ++j) {
            let sum = 0.0,
                i = 0;
            for (; i<n; ++i) {
                sum += mat[i][j];
            }
            sums[j] = sum;
        }
        return sums;
    }
    
    exports.Simplex = class Simplex {
        constructor(StartPars, config) {
            let Steps = config && config.steps || 0.1;
            const epsilon = config && config.epsilon || 1e-8;
            this.maxIter = config && config.maxIter || 200;
            /*
              % The structure elements are the following:
              %   t.n  number of elements in the parameter vector
              %   t.prow row vector of parameters presently being tested
              %   t.simp the simplex matrix, (n+1) row vectors
              %   t.vals the (n+1) element column vector of function values
              %   t.high index of the highest t.vals element
              %   t.low index of the lowest t.vals element
              %   t.centr centroid row vector of the best n vertices
              %   t.index counter for loops
              %   t.state the state variable of the machine
              %   t.epsilon is the largest score that is considered converged
            */
            const t = this;
            t.n=StartPars.length;
            t.prow=StartPars.slice();
            
            // Handle defaults for the step size.
            if ( typeof(Steps.length) === 'undefined' ) { // Only a scalar given.
                const zerostep = 1e-3;
                Steps = t.prow.map(p => Steps*p || zerostep);
            }
            
            // The simplex is (n+1) row vectors.
            // vals is a column vector of function values
            t.simp = [];
            t.vals = [];
            let i = 0;
            for (; i<=t.n; ++i) {
                t.simp[i] = t.prow.slice();
                if ( i ) {
                    t.simp[i][i-1] += Steps[i-1];
                    t.vals[i-1] = 0;
                }
            }
            
            // Initialize the other variables
            t.index=0;
            t.state=1;
            t.high=0;
            t.low=0;
            t.centr=t.prow;
            t.epsilon = epsilon;
        }
        centroid() {
            const t = this;
            return sumColumns(t.simp).map(function(s) { return s/(t.n+1); });
        }
        converged(epsilon) {
            epsilon = epsilon || t.epsilon;
            const t = this,
                  err = Math.max(...t.vals) - Math.min(...t.vals);
            return (t.state === 3) && (err < epsilon);
        }
        next(y) {
            const t = this;
            var i;
            if ( t.state === 1 ) { // Start-up
                t.vals[t.index]=y;  // pick up the function value from last time.
                t.index++;
                if ( t.index <= t.n ) { // % continue to fill up the simplex
                    t.prow = t.simp[t.index];
                } else {// Simplex is full, make the first move
                    t.state=3;
                }
            } else if ( t.state === 3 ) { // Test a new vertex
                i = t.high;
                if ( y < t.vals[i] ) { // The new vertex is better than some.
                    t.simp[i] = t.prow.slice(); // replace the worst one.
                    t.vals[i]=y;
                    if ( y < t.vals[t.low] ) { // The new vertex is better than the best,
                        t.prow = t.simp[i].map((s,is) =>s + 1.1*(s-t.centr[is])); // so, expand in the new direction.
                        t.prevy=y;
                        t.state=4;
                    } else {
                        t.state=3;
                    }
                } else { // the new vertex is worse than the worst: contract.
                    t.prow=t.simp[t.high].map((s,is) => 0.5*(s+t.centr[is]));
                    t.state=5;
                }
            } else if ( t.state === 4 ) { // Test an expansion
                if ( y < t.prevy ) { // Accept the expansion
                    t.simp[t.high] = t.prow.slice();
                    t.vals[t.high] = y;
                }
                t.state=3;
            } else if ( t.state === 5 ) { // Test a contraction
                if ( y < t.vals[t.high] ) { // Accept the contraction
                    t.simp[t.high] = t.prow.slice();
                    t.vals[t.high]=y;
                    t.state=3;
                } else { // contract the whole simplex toward the best vertex.
                    t.index=0;
                    t.simp[0] = t.simp[0].map((s, is) =>0.5*(s + t.simp[t.low][is]));
                    t.prow=t.simp[0];
                    t.state=6;
                }
            } else if ( t.state === 6 ) {
                t.vals[t.index] = y; // pick up the function value.
                t.index++;
                i=t.index;
                if ( i <= t.n ) {
                    t.simp[i] = t.simp[i].map((s,is) => 0.5*(s + t.simp[t.low][is]));
                    t.prow = t.simp[i].slice();
                    t.state=6; // continue evaluating the vertices
                } else {
                    t.state=3;
                }
            } // end switch
            
            if ( t.state === 3 ) { // Normal exit mode: sort the vertices and try a reflection.
                // assign min and max
                let lo = t.vals[0],
                    iLo = 0,
                    hi = t.vals[0],
                    iHi = 0;
                t.vals.forEach((v,i) => {
                    if ( v < lo ) {
                        lo = v;
                        iLo = i;
                    }
                    if ( v > hi ) {
                        hi = v;
                        iHi = i;
                    }
                });
                t.low = iLo;
                t.high = iHi;
                // find the excluded centroid
                const sum = sumColumns(t.simp);
                t.centr = t.simp[t.high].map((s,is) =>(sum[is] - s) / t.n);
                // reflect about the centroid from the highest vertex
                t.prow = t.simp[t.high].map((s,is) =>2*t.centr[is] - s);
            }
            
            return t.prow;
        }
        /* Promise-based run() method */
        run(cb) {
            // cb: {iter(params, score, iteration) => bool shouldContinue,
            //      score(params)=>Promise(score)}
            // returns: Promise({params, score, iteration, centroid})
            const shouldContinue = cb.iter || (() => true),
                  t = this;
            t.iter = 0;
            function finished(params, score) {
                // Returns true on max iterations, !shouldContinue(), or convergence
                return ( t.iter && ((t.iter === t.maxIter) ||
                                    (! shouldContinue(params, score, t.iter)) ||
                                    t.converged(t.epsilon)) );
            }
            // The promised result is an object with params, score, iteration, and centroid; available when finished()
            return new Promise((resolve, reject) => {
                // Each iteration begins with the asynchronous resolution of the previous params' score
                function nextIteration(params, score) {
                    if ( finished(params, score) ) {
                        const centroid = t.centroid();
                        cb.score(centroid).then(centroidScore => resolve({
                            score: centroidScore,
                            iteration: t.iter,
                            params: centroid,
                            centroid: centroid
                        }));
                        return;
                    } else if ( t.iter === 0 ) { // first iteration's params are from Simplex constructor
                        params = t.prow;
                    } else { // for subsequent iterations they are computed given the previous params' score
                        params = t.next(score);
                    }
                    t.iter++;
                    cb.score(params).then(score => nextIteration(params, score));
                }
                // Start the first iteration immediately
                nextIteration();
            });
        }
        test(params, fun) {
            // if not provided, we test the optimizer's ability to find params=[2,-5]
            fun = fun || (params => sqr(params[0]-2) + sqr(params[1]-(-5)));
            params = params || [1, 1];
            const simplex = new exports.Simplex(params);
            let iter = 0;
            for (; iter<200; ++iter) {
                const y = fun(params);
                params = simplex.next(y);
                if ( ! (i%10) ) { // every 10 iterations print out the fitted value
                    console.log('value',y,'at params',params,'iteration',iter);
                }
                if ( simplex.converged() ) {
                    break;
                }
            }
            params = simplex.centroid();
            console.log('final params',params);
            return params;
        }
    };
    
    exports.Simplex.testRun = function(params, scoreFunc, maxIterations, shouldContinue) {
        scoreFunc = scoreFunc || (params => Promise.resolve(sqr(params[0]-2) + sqr(params[1]-(-5))));
        shouldContinue = (params, score, iter) => {
            console.log('iteration',iter,'score',score,'params',params);
            return true;
        };
        params = params || [1,1];
        
        new exports.Simplex(params, {
            maxIter: maxIterations
        })
            .run({
                score: scoreFunc,
                iter: shouldContinue
            })
            .then(result => console.log('finished after iteration',result.iteration,'score',result.score,'centroid',result.centroid));
    };

    /* ============================================================================================== */
    
    const DFPDefaults = {
        maxIter: 50,
        convLL: 1e-8,
        convGrad: 1e-8,
        maxStep: 1.0,
        maxRestarts: 2,
        alf: 1e-4,
        tolx: 1e-7,
        epsilon: 3e-8
    };
    exports.DFP = class DFP {
        constructor(params, config) {
    
/*
Adaptation of Davidon-Fletcher-Powell minimization for cooperative control flow

// basic interface -- methods return status indicating what to compute and which method to call next

var dfp = new DFP();
var params = [1, 1, 1];
var status = dfp.start(params);
while ( status.need ) {
  if ( status.need === 'score' ) {
    status = dfp.score( compute_score(status.params) )
  } else if ( status.need === 'grad' ) {
    var result = compute_grads(status.params);
    status = dfp.grad(result.score, result.grad);
  } else if ( status.need === 'iter' ) {
    // if you've had enough at this point, just discard dfp, otherwise...
    status = dfp.iter()
  }
}
if ( status.err ) {
  console.log('Error', status.err, status.errMsg);
} else {
  console.log('Success', status.score, status.params, status.grad, status.hessian);
}

// DFP.run(): alt interface based on Promise -- you provide score, grad callbacks which return promises, as does DFP.run()

var dfp = new DFP();
var params = [1, 1, 1];
dfp.run(params, {
    iter: function(params, score, iteration, grad) {
        // return false to interrupt
        return true;
    },
    score: function(params) {
        return Promise.resolve(compute_score(params));
    },
    // either:
    grad: function(params) {
        return Promise.resolve({
            score: compute_score(params),
            grad: compute_grad(params)
        });
    },
    // or:
    dx: 1e-6 // the default (forward) numerical differentiator step
    // or:
    grad: numGrad2Side(score, dx) // two-sided numerical gradient, twice as slow, score and dx as above
})
    .then(function(status) {
        // look at status.iteration, status.score, ..., status.hessian
    })
    .catch(function(status) {
        // look at status.err, status.errMsg
    });
*/

/*
function nodetest(name, deps, func) {
    func().test();
}
if (typeof module !== 'undefined' && module.exports) { // detect node.js
    define = nodetest;
}
*/
            this.params = params;
            this.config = {
                maxIter: config && config.maxIter || DFPDefaults.maxIter,
                convLL: config && config.convLL || DFPDefaults.convLL,
                convGrad: config && config.convGrad || DFPDefaults.convGrad,
                maxStep: config && config.maxStep || DFPDefaults.maxStep,
                maxRestarts: config && config.maxRestarts || DFPDefaults.maxRestarts,
                alf: config && config.alf || DFPDefaults.alf,
                tolx: config && config.tolx || DFPDefaults.tolx,
                epsilon: config && config.epsilon || DFPDefaults.epsilon
            };
        }
        start(params, restarts) {
            const n = params.length;
            this.dfpstate = {
                first: true,
                dg: exports.newVector(n),
                hdg: exports.newVector(n),
                xi: exports.newVector(n),
                z: params.slice(),
                stpmax: 0,
                scorePre: 0
            };
            this.status = {
                score: 0,
                iteration: 1,
                err: 0,
                restarts: restarts|0,
                params: params.slice(),
                grad: exports.newVector(n),
                hessian: exports.newDiagMatrix(n),
                need: 'grad'
            };
            return this.status;
        }
        grad(score, grad) {
            const hessian = this.status.hessian,
                  d = this.dfpstate,
                  xi = d.xi,
                  dg = d.dg,
                  hdg = d.hdg;
            
            if ( d.first ) {
                d.first = false;
                this.status.score = score;
                this.status.grad = grad;
                const ng = grad.length;
                let sum = 0.0,
                    i = 0;
                for (; i<ng; ++i) {
                    xi[i] = - grad[i];
                    sum += d.z[i] * d.z[i];
                }
                d.stpmax = this.config.maxStep * Math.max(Math.sqrt(sum), ng);
                this.status.need = 'iter'; // -> lnsrch -> grad
                return this.status;
            }
            if ( score > 1e10 ) {
                if ( this.status.restarts < this.config.maxRestarts ) {
                    return this.start(this.status.params, this.status.restarts+1);
                } else {
                    this.status.params = d.z;
                    this.status.score = d.scorePre;
                    this.status.err = -4;
                    this.status.errMsg = "too many restarts";
                    this.status.need = "";
                    return this.status;
                }
            }
            const tolf = d.scorePre ? Math.abs((score - d.scorePre) / d.scorePre) : 1.0;
            d.scorePre = score;
            const n = grad.length;
            let i = 0;
            for (; i<n; ++i) {
                xi[i] = this.status.params[i] - d.z[i];
                d.z[i] = this.status.params[i];
                dg[i] = this.status.grad[i];
                this.status.grad[i] = grad[i];
            }
            let tolg = 0.0;
            for (i=0; i<n; ++i) {
                tolg = Math.max(tolg, Math.abs(grad[i]));
            }
            if ( (tolf < this.config.convLL) && (tolg < this.config.convGrad) && (score < 1e10) ) {
                this.status.need = ""; // success
                this.status.score = score;
                this.status.params = d.z;
                return this.status;
            }
            
            for (i=0; i<n; ++i) {
                dg[i] = grad[i] - dg[i];
            }
            for (i=0; i<n; ++i) {
                let sum = 0.0,
                    j = 0;
                for (; j<n; ++j) {
                    sum += hessian[i][j] * dg[j];
                }
                hdg[i] = sum;
            }
            let fac = 0.0,
                fae = 0.0,
                sumdg = 0.0,
                sumxi = 0.0;
            for (i=0; i<n; ++i) {
                fac += dg[i] * xi[i];
                fae += dg[i] * hdg[i];
                sumdg += dg[i] * dg[i];
                sumxi += xi[i] * xi[i];
            }
            if ( (fac*fac) > (this.config.epsilon*sumdg*sumxi) ) {
                fac = 1.0 / fac;
                const fad = 1.0 / fae;
                for (i=0; i<n; ++i) {
                    dg[i] = fac*xi[i] - fad*hdg[i];
                }
                for (i=0; i<n; ++i) {
                    let j = 0;
                    for (; j<n; ++j) {
                        hessian[i][j] += fac*xi[i]*xi[j] - fad*hdg[i]*hdg[j] + fae*dg[i]*dg[j];
                    }
                }
            }
            for (i=0; i<n; ++i) {
                let sum = 0.0,
                    j = 0;
                for (; j<n; ++j) {
                    sum += hessian[i][j]*grad[j];
                }
                xi[i] = - sum;
            }
            
            if ( this.status.iteration === this.config.maxIter ) {
                this.status.err = 0; // not exceptional
                this.status.errMsg = "too many iterations";
                this.status.need = "";
                return this.status;
            }
            ++ this.status.iteration;
            this.status.need = 'iter'; // -> lnsrch -> grad
            return this.status;
        }
        iter() {
            this.dfpstate.scorePre = this.status.score;
            return this.lnsrch1();
        }
        // lnsrch exits with need: "grad"
        // lnsrch loops on need: "score"
        lnsrch1() {
            const d = this.dfpstate,
                  n = d.z.length,
                  l = {
                      check: 0,
                      flagRund: 0,
                      f2: 0,
                      fold2: 0,
                      alam2: 0
                  };
            this.lnstate = l;
            let sum = 0.0,
                i = 0;
            for (; i<n; ++i) {
                sum += d.xi[i] * d.xi[i];
            }
            sum = Math.sqrt(sum);
            if ( sum > d.stpmax ) {
                for (i=0; i<n; ++i) {
                    d.xi[i] *= d.stpmax/sum;
                }
            }
            sum = 0.0;
            for (i=0; i<n; ++i) {
                sum += this.status.grad[i] * d.xi[i];
            }
            l.slope = sum;
            let test = 1e-12;
            for (i=0; i<n; ++i) {
                let tmp = Math.abs(d.xi[i]) / Math.max(Math.abs(d.z[i]), 1);
                if ( (tmp > test) && isFinite(tmp) ) {
                    test = tmp;
                }
            }
            l.alamin = this.config.tolx / test;
            l.alam = 1.0;
            if ( l.alam < l.alamin ) {
                this.status.need = 'grad'; // exit
                return this.status;
            }
            
            const znew = this.status.params;
            for (i=0; i<n; ++i) {
                znew[i] = d.z[i] + l.alam*d.xi[i];
            }
            this.status.need = 'score'; // repeat
            return this.status;
        }
        score(score) {
            this.status.score = score;
            const d = this.dfpstate,
                  l = this.lnstate,
                  alam = l.alam,
                  alam2 = l.alam2;
            if ( alam < l.alamin ) {
                this.status.params = d.z.slice();
                l.check = true;
                this.status.need = 'grad'; // exit
                return this.status;
            }
            if ( score <= (d.scorePre + this.config.alf*alam*l.slope) ) {
                this.status.need = 'grad'; // exit
                return this.status;
            }
            var tmplam;
            if ( alam === 1.0 ) {
                tmplam = - l.slope / (2.0 * (score - d.scorePre - l.slope));
            } else {
                const rhs1 = score - d.scorePre - alam*l.slope,
                      rhs2 = l.f2 - l.fold2 - alam2*l.slope,
                      a = (rhs1/(alam*alam) - rhs2/(alam2*alam2)) / (alam - alam2),
                      b = (-alam2*rhs1/(alam*alam) + alam*rhs2/(alam2*alam2)) / (alam - alam2);
                if ( a === 0 ) {
                    tmplam = - l.slope / (2*b);
                } else {
                    const disc = b*b - 3*a*l.slope;
                    if ( disc < 0 ) {
                        l.flagRund = 1;
                        this.status.need = 'grad'; // exit
                        return this.status;
                    }
                    tmplam = (-b + Math.sqrt(disc))/(3*a);
                }
                if ( tmplam > (0.5*alam) ) {
                    tmplam = 0.5*alam;
                }
            }
            l.alam2 = alam;
            l.f2 = score;
            l.fold2 = d.scorePre;
            l.alam = Math.max(tmplam, 0.1*alam);
            
            const znew = this.status.params,
                  zn = znew.length;
            let i = 0;
            for (; i<zn; ++i) {
                znew[i] = d.z[i] + l.alam*d.xi[i];
            }
            this.status.need = 'score'; // repeat
            return this.status;
        }
        run(cb) {
            // cb: {iter(params, score, iteration, grad),
            //      score(params)=>Promise(score),
            //      grad(params)=>Promise({score,grad})
            //      or dx: 1e-6 for numerical gradient}
            // returns Promise(status)
            return new Promise((resolve, reject) => {
                const next = status => {
                    if ( status.need === 'iter' ) {
                        if ( (! cb.iter) || cb.iter(status.params, status.score, status.iteration, status.grad) ) {
                            return next(this.iter());
                        } else {
                            status.err = -10;
                            status.errMsg = 'Stopped'
                        }
                    }
                    if ( status.need === 'score' ) {
                        return cb.score && cb.score(status.params).then(score => next(this.score(score)));
                    }
                    if ( status.need === 'grad' ) {
                        return (cb.grad && cb.grad || exports.numGradFwd(cb.score, cb.dx||1e-7))(status.params)
                            .then(result => next(this.grad(result.score, result.grad)));
                    }
                    (status.err ? reject : resolve)(status);
                };
                next(this.start(this.params));
            });
        }
    };

    exports.numGradFwd = function numGradFwd(getScore, dx) {
        dx = dx || 1e-6;
        return params => {
            const paramGrid = [params].concat(params.map((p,i) => {
                const xpars = params.slice();
                xpars[i] += dx;
                return xpars;
            }));
            return Promise.all(paramGrid.map(pp => getScore(pp)))
                .then(scores => {
                    const score = scores[0];
                    return {
                        score: score,
                        grad: params.map((p,i) => (scores[i+1] - score) / dx)
                    };
                });
        };
    };

    exports.numGrad2Side = function numGrad2Side(getScore, dx) {
        dx = dx || 1e-6;
        return params => {
            const paramGrid = [params]
                  .concat(params.map((p,i) => {
                      const xpars = params.slice();
                      xpars[i] += dx;
                      return xpars;
                  }))
                  .concat(params.map((p,i) => {
                      const xpars = params.slice();
                      xpars[i] -= dx;
                      return xpars;
                  }));
            return Promise.all(paramGrid.map(pp => getScore(pp)))
                .then(scores => {
                    const n = params.length;
                    return {
                        score: scores[0],
                        grad: params.map((p, i) => (scores[i+1] - scores[n+i+1]) / (2*dx))
                    };
                });
        };
    };

    exports.newVector = function newVector(n) {
        const v = [];
        let i = 0;
        for (; i<n; ++i) {
            v[i] = 0.0;
        }
        return v;
    }
    exports.newMatrix = function newMatrix(n, m) {
        const mat = [];
        let i = 0;
        for (; i<n; ++i) {
            const row = [];
            let j = 0;
            for (; j<m; ++j) {
                row[j] = 0.0;
            }
            mat[i] = row;
        }
        return mat;
    }
    exports.newDiagMatrix = function newDiagMatrix(n) {
        const mat = exports.newMatrix(n, n);
        let i = 0;
        for (; i<n; ++i) {
            mat[i][i] = 1.0;
        }
        return mat;
    }

    exports.test = function test() {
        const data = [1, 2, 3.3],
              params = [2, 1]; // [m, b]; data[i] = m*i + b
        const score = params => {
            /// resolves to sum square difference between m*i+b and data
            const m = params[0],
                  b = params[1];
            return Promise.resolve(data.reduce((sum, one, i) => {
                const d = m*i+b-one;
                return sum + d*d;
            }, 0));
        };
        return (new exports.Simplex(params))
            .run({
                score: score,
                iter: (params, score, iteration) => {
                    console.log('simplex iter',iteration,params,score);
                    return true;
                }
            })
            .then(simplexResult => (new exports.DFP(simplexResult.params)).run({
                iter: (params, score, iteration, grad) => {
                    console.log('dfp iter', params, score, iteration, grad);
                    return true; // ( ! stopFlag );
                },
                score: score,
                //dx: 1e-6
                grad: exports.numGrad2Side(score)
            })).then(dfpResult => console.log('done', dfpResult.params, dfpResult))
            .catch(status => console.log('err', status.err, status.errMsg));
    };
    
    return exports;
});
        

