/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_simu_results", ["d3", "qubx", "qubx_lib", "qubx_ui", "qubx_chart", "qubx_fit"],
   (d3, qubx, qubx_lib, qubx_ui, qubx_chart, qubx_fit) => {
       "use strict";
       const exports = {};

       const templates = qubx_ui.loadTemplates([".qubx_simResultsBlob", ".qubx_simResultsInstructions"]).then(items => ({
           blob: items[0],
           instructions: items[1]
       })).catch(qubx.reportErr);
       
       exports.SimMatrices = class SimMatrices {
           constructor(config) {
               this.models = config.models;
               this.models.context.ev.selectModel.sub(this._onSelectModel.bind(this));
               this.choose = config.choose || function(iSeg) {};
               this.node = document.createElement('div');
               this.node.className = "qubx_simResults";
               this.node.style.overflowY = 'scroll';
               this.showing = false;
               this.stale = true;
               templates.then(templates => {
                   const ssItem = title => {
                       const item = {node: templates.blob.cloneNode(true)};
                       item.node.style.float = 'left';
                       item.data = item.node.querySelector(".qubx_simResultsBlobData");
                       item.title = item.node.querySelector(".qubx_simResultsBlobTitle");
                       d3.select(item.title).text(title);
                       this.node.appendChild(item.node);
                       return item;
                   };
                   this.Peq = ssItem("Peq: Equilibrium state occupancy");
                   this.Keq = ssItem("Keq: Equilibrium constants");
                   this.delG = ssItem("delG: Delta free energy");
                   this.Qflux = ssItem("Qflux: ");
                   this.MFP = ssItem("Mean First Passage: avg. msec from state A to B");
                   this.Q = ssItem("Q: rate constants");
                   this.A = ssItem("A: sampled transition probabilities");
               });
           }
           update() {
               templates.then(templates => {
                   this.stale = false;
                   const levels = this.indVar.map(indVar => Object.assign({}, this.constants, indVar));
                   levels.push(this.constants);
                   const multi = (levels.length > 1);
                   qubx_lib.then(lib => {
                       const out = {Peq: "", Q: "", A: "", Keq: "", delG: "", Qflux: "", MFP: ""};
                       levels.forEach((constants, i) => {
                           if ( i ) {
                               Object.keys(out).forEach(name => {
                                   out[name] += "<br>";
                               });
                           }
                           if ( multi ) {
                               const lbl = Object.keys(this.indVar[0]).map(name => ''+name+': '+qubx.formatK0(constants[name]))
                                     .join(', ');
                               Object.keys(out).forEach(name => {
                                   out[name] += lbl+'<br>';
                               });
                           }
                           const Q = this.models.model.makeQ(constants);
                           if ( ! Q.length ) { return; }
                           const Peq = lib.QtoPe(Q),
                                 A = lib.expm(Q, this.sampling),
                                 Keq = lib.QtoKeq(Q);
                           out.Peq += qubx.pformat(Peq);
                           out.Q += qubx.pformat(Q);
                           out.A += qubx.pformat(A);
                           out.Keq += qubx.pformat(Keq);
                           out.delG += qubx.pformat(lib.KeqToDelG(Keq));
                           out.Qflux += qubx.pformat(lib.makeQflux(Q));
                           out.MFP += qubx.pformat(lib.meanFirstPassage(Q, this.sampling));
                       });
                       this.Peq.data.innerHTML = out.Peq;
                       this.Q.data.innerHTML = out.Q;
                       this.A.data.innerHTML = out.A;
                       this.Keq.data.innerHTML = out.Keq;
                       this.delG.data.innerHTML = out.delG;
                       this.Qflux.data.innerHTML = out.Qflux;
                       this.MFP.data.innerHTML = out.MFP;
                   }).catch(qubx.reportErr);
               });
           }
           onShow(showing) {
               this.showing = showing;
               if ( showing && this.stale ) {
                   this.update();
               }
           }
           tryUpdate() {
               if ( this.showing ) {
                   this.update();
               } else {
                   this.stale = true;
               }
           }
           onClear(sampling, constants) {
               this.sampling = sampling;
               this.constants = constants;
               this.indVar = [];
               this.tryUpdate();
           }
           onSelect(offset, count, segments, signalNames, sampling, constants) {
               if ( segments.length && segments[0].indVar && ! this.indVar.length ) {
                   this.indVar = segments.map(qubx.m('indVar'));
               }
               if ( this.showing ) {
                   this.update();
               } else {
                   this.stale = true;
               }
           }
           onChoose(iSeg) {
           }
           _onSelectModel() {
               this.tryUpdate();
           }
       };
       
       const assignToSegments = function assignToSegments(dose, meas, normvar, peakDir, normMax) {
           const n = dose.length;
           let dy = 1e8,
               ylast = dose[0],
               i = 1;
           for (; i<n; ++i) {
               const y = dose[i];
               dy = Math.min(dy, Math.abs(y - ylast));
               ylast = y;
           }
           dy /= 2; // tolerance for matching doses
           const result = {
               norm: []
           };
           const fields = ['mean', 'std', 'response'];
           fields.forEach(ff => {
               if ( ff in meas ) {
                   result[ff] = [];
               }
           });
           dose.forEach((d, i) => {
               if ( d !== null ) {
                   const nMeas = meas.dose.length;
                   let j = 0;
                   for (; j<nMeas; ++j) {
                       if ( Math.abs(meas.dose[j] - d) < dy ) {
                           fields.forEach(ff => {
                               if ( ff in meas ) {
                                   result[ff][i] = meas[ff][j];
                               }
                           });
                           break;
                       }
                   }
               }
           });
           const minmax = (peakDir > 0) ? qubx.arrayMax : qubx.arrayMin;
           if ( typeof(normMax) === 'undefined' ) {
               normMax = minmax(result[normvar]) || 0.0;
           }
           result.normMax = normMax;
           result.norm = result[normvar].map(x => x/normMax);
           return result;
       };
       const intersection_or_first = function intersection_or_first(f0, n0, f1, n1) {
           const f = Math.max(f0, f1),
                 l = Math.min(f0+n0-1, f1+n1-1);
           if ( f <= l ) {
               return {f: f, l: l, n: l-f+1};
           } else {
               return {f: f0, l: f0+n0-1, n: n0};
           }
       };

       exports.DoseResponse = class DoseResponse {
           constructor(config) {
               this.models = config.models;
               this.choose = config.choose || (iseg => {});
               this.node = document.createElement('div');
               this.node.className = "qubx_simResults";
               this.grid = new qubx_ui.AspectGrid({
                   node: this.node
               });
               this.charts = [];
               this.showing = false;
               this.stale = true;
               this.chosen = -1;
               templates.then(templates => {
                   this.instructions = templates.instructions;
                   this.charts.length || this.grid.insert(this.instructions);
               });
           }
           chartCount(n) {
               if ( n === this.charts.length ) {
                   return Promise.resolve(true);
               }
               return templates.then(templates => {
                   this.charts.length || this.grid.remove(this.instructions);
                   const ssItem = () => {
                       const item = {
                           node: templates.blob.cloneNode(true)
                       };
                       item.hdr = item.node.querySelector(".qubx_simResultsBlobHdr");
                       item.title = item.hdr.querySelector(".qubx_simResultsBlobTitle");
                       item.data = item.node.querySelector(".qubx_simResultsBlobData");
                       item.node.classList.add('qubx_fullsize');
                       item.chart = new qubx_chart.Chart({
                           node: item.data,
                           onselected: this.choose
                       });
                       this.grid.insert(item.node, undefined, item.chart.resize.bind(item.chart));
                       return item;
                   };
                   while ( this.charts.length < n ) {
                       this.charts.push(ssItem());
                   }
                   while ( this.charts.length > n ) {
                       const extra = this.charts.pop().node;
                       this.grid.remove(extra);
                   }
                   if ( n === 0 ) {
                       this.grid.insert(this.instructions);
                   }
               });
           }
           update() {
               if ( this.charts && this.indVar.length && this.indVar[this.indVar.length-1] ) {
                   qubx_lib.then(lib => {
                       let maxN = 8;
                       this.segments.forEach(seg => {
                           maxN = Math.max(maxN, seg.selTo - seg.selFrom + 1);
                       });
                       // Identify the quantities of interest: they are listed in indVar and take different values in different segments
                       const signalVals = [],
                             nonSignal = [];
                       let signalVaryCount = 0,
                           incomplete = false;
                       Object.keys(this.indVar[0]).forEach(name => {
                           let vary = false;
                           const firstVal = this.indVar[0][name],
                                 vals = this.indVar.map(iv => {
                                     if ( iv ) {
                                         const val = iv[name];
                                         vary = vary || (val != firstVal);
                                         return val;
                                     } else {
                                         incomplete = true;
                                         return 0;
                                     }
                                 });
                           if ( vary ) {
                               const iSig = this.signalNames.indexOf(name);
                               if ( iSig >= 0 ) {
                                   signalVals[iSig] = vals;
                                   ++signalVaryCount;
                               } else {
                                   nonSignal.push({
                                       name: name,
                                       vals: vals
                                   });
                               }
                           }
                       });
                       if ( incomplete ) {
                           return;
                       }
                       // Measure and Plot:
                       this.stale = false;
                       return this.chartCount(2*signalVaryCount + nonSignal.length).then(() => {
                           let iChart = 0,
                               dy = (this.classAmp.length > 1) ? (this.classAmp[1] - this.classAmp[0]): 1;
                           if ( this.IeqFv ) {
                               if ( 'Voltage' in this.indVar[0] ) {
                                   dy *= this.indVar.Voltage;
                               } else if ( 'Voltage' in this.constants ) {
                                   dy *= this.constants.Voltage;
                               }
                           }
                           const peakDir = (dy >= 0) ? 1 : -1;
                           // First, peak and equilibrium v. dose, when we have a signal to show us where to measure
                           signalVaryCount && signalVals.forEach((dose, iSig) => {
                               if ( dose ) {
                                   let indHi = dose[0],
                                       indLo = dose[0];
                                   dose.forEach(d => {
                                       if ( d < indLo ) { indLo = d; }
                                       if ( d > indHi ) { indHi = d; }
                                   });
                                   const measPeak = new lib.DoseResponsePeak(this.sampling, 0/*filter*/, peakDir, maxN, 3*this.segments.length),
                                         measEquil = new lib.DoseResponseEquil(Math.round(maxN/8)|0, (indHi-indLo)/1000.0, maxN, 3*this.segments.length);
                                   this.segments.map(seg => {
                                       measPeak.add(seg.sources[iSig], seg.sources[0], seg.selFrom, seg.selTo-seg.selFrom+1);
                                       measEquil.add(seg.sources[iSig], seg.sources[0], seg.selFrom, seg.selTo-seg.selFrom+1);
                                   });
                                   const mPeak = measPeak.getResult(),
                                         mEquil = measEquil.getResult();
                                   measEquil.destroy();
                                   measPeak.destroy();
                                   
                                   const sortPeak = assignToSegments(dose, mPeak, 'response', peakDir),
                                         sortEquil = assignToSegments(dose, mEquil, 'mean', peakDir, sortPeak.normMax);
                                   
                                   let item = this.charts[iChart];
                                   item.data.style.maxHeight = ""; // allows it to resize next time generated
                                   /*///item.title.html(`normalized I peak v. ${this.signalNames[iSig]}`);*/
                                   const xvar = this.signalNames[iSig];
                                   let yvar = "norm. I peak",
                                       chartData = {
                                           xvar: xvar,
                                           yvar: yvar,
                                           columns: [xvar, yvar, "I peak"],
                                           series: {},
                                           fitCurve: qubx_fit.hill,
                                           fitMarks: {x: ["EC50"]}
                                       };
                                   chartData.series[xvar] = dose;
                                   chartData.series[yvar] = sortPeak.norm;
                                   chartData.series["I peak"] = sortPeak.response;
                                   item.chart.setScatter(chartData);
                                   
                                   item = this.charts[iChart+1];
                                   item.data.style.maxHeight = ""; // allows it to resize next time generated
                                   /*///item.title.html(`normalized I equil. v. ${this.signalNames[iSig]}`);*/
                                   yvar = "norm. I eq.";
                                   chartData = {
                                       xvar: xvar,
                                       yvar: yvar,
                                       columns: [xvar, yvar, "I eq", "I eq std."],
                                       series: {},
                                       fitCurve: qubx_fit.hill,
                                       fitMarks: {x: ["EC50"]}
                                   };
                                   chartData.series[xvar] = dose;
                                   chartData.series[yvar] = sortEquil.norm;
                                   chartData.series["I eq"] = sortEquil.mean;
                                   chartData.series["I eq std."] = sortEquil.std;
                                   item.chart.setScatter(chartData);
                                   
                                   iChart += 2;
                               }
                           });
                           
                           // Then, peak current for non-signal variables (e.g. tRecovery for paired pulse simulation)
                           // If the setup has set .prefRgn = {f: iFirst, n: count}, we will
                           // prefer to measure the intersection with onscreen selection,
                           // but if the intersection is empty, we will use the whole onscreen selection
                           nonSignal.forEach(ns => {
                               const dose = this.indVar.map(qubx.m('ns.name'));
                               // check for prefRgn, find intersection or fall back on main selection
                               // measure peak per segment
                               const peak = this.segments.map((seg, iSeg) => {
                                   const pref = this.prefRgns ? this.prefRgns[iSeg] : null,
                                         rgn = intersection_or_first(seg.selFrom, seg.selTo-seg.selFrom+1,
                                                                     pref ? pref.f : -1, pref ? pref.n : -1),
                                         arr = seg.sources[0];
                                   return (peakDir > 0) ? qubx.arrayMax(arr, rgn.f, rgn.l+1) : qubx.arrayMin(arr, rgn.f, rgn.l+1);
                               });
                               const minmax = (peakDir > 0) ? qubx.arrayMax : qubx.arrayMin,
                                     maxPeak = minmax(peak) || 0.0,
                                     norm = peak.map(p => p/maxPeak),
                                     item = this.charts[iChart];
                               item.data.style.maxHeight = ""; // allows it to resize next time generated
                               /*///item.title.html(`normalized I peak v. ${ns.name}`);*/
                               const yvar = "norm. I peak",
                                     chartData = {
                                         xvar: ns.name,
                                         yvar: yvar,
                                         columns: [ns.name, yvar, "I peak"],
                                         series: {}
                                     };
                               chartData.series[ns.name] = dose;
                               chartData.series[yvar] = norm;
                               chartData.series["I peak"] = peak;
                               item.chart.setScatter(chartData);
                               
                               ++iChart;
                           });
                           this.onChoose(this.chosen);
                       });
                   }).catch(qubx.reportErr);
               }
           }
           onShow(showing) {
               this.showing = showing;
               if ( showing && this.stale ) {
                   this.update();
               } else {
                   this.onChoose(this.chosen);
               }
           }
           onClear(sampling, constants, classAmp, IeqFv, prefRgns) {
               this.sampling = sampling;
               this.constants = constants;
               this.indVar = [];
               this.segments = [];
               this.signalNames = [];
               this.prefRgns = prefRgns;
               this.classAmp = classAmp;
               this.IeqFv = IeqFv;
               if ( this.showing ) {
                   this.update();
               } else {
                   this.stale = true;
               }
           }
           onSelect(offset, count, segments, signalNames, sampling, constants) {
               this.segments = segments;
               this.signalNames = signalNames;
               if ( segments.length && segments[segments.length-1].indVar && ! this.indVar.length ) {
                   this.indVar = segments.map(qubx.m('indVar'));
               }
               if ( this.showing ) {
                   this.update();
               } else {
                   this.stale = true;
               }
           }
           onChoose(iSeg) {
               this.chosen = iSeg;
               this.charts.forEach(item => item.chart.select(iSeg));
           }
       };
       
       return exports;
   });
