/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define ("qubx_simulate", ["qubx", "qubx_model"], (qubx, qubx_model) => {
    'use strict';

    const exports = {};
    
    const SimulatorDefaults = {
        usePeq: true,
        segmentSize: [10000],
        segmentIndVar: [{}], // e.g. {Ligand: 1.5} per segment
        sampling: 1e-4,
        chunk: 2048,
        auto: true,
        constants: {},
        stimulus: {
            /* e.g. 'Voltage' : {
                        stimType: 'steps',
                        filter_kHz: 0,
                        traces: [
                            {
                                levels: [-50, 50, -50],
                                durations: [.3, .3, .3], // in seconds; final value is held to end of segment
                            } //, additional segments; final segment will be repeated as needed for config.segmentSize
                        ]
                     }
            */
        },
        about: "from QUB Online"
    };
    
    exports.Simulator = class Simulator {
        constructor(config) {
            this.config = Object.assign({}, SimulatorDefaults, config);
            const path = 'js/qubx_simulate_worker.js';
            this.worker = new Worker(path);
            this.worker.addEventListener('message', this._onMessage.bind(this), false);
            this.worker.addEventListener('error', this._onError.bind(this), false);
            this.models = this.config.models || new qubx_model.Models();
            delete this.config.models; // posted separately in json form
            this._modelInsertModel = this.models.ev.insert.sub(this._onInsertModel.bind(this));
            this._modelRemoveModel = this.models.ev.remove.sub(this._onRemoveModel.bind(this));
            this._modelActivateModel = this.models.ev.activate.sub(this._onActivateModel.bind(this));
            this._modelSetChannelCount = this.models.ev.setChannelCount.sub(this._onSetChannelCount.bind(this));
            this._modelSetIeqFv = this.models.ev.setIeqFv.sub(this._onSetIeqFv.bind(this));
            this._modelSetVRev = this.models.ev.setVRev.sub(this._onSetVRev.bind(this));
            this._modelInsertedState = this.models.ev.insertState.sub(this._onInsertedState.bind(this));
            this._modelRemovedState = this.models.ev.removeState.sub(this._onRemovedState.bind(this));
            this._modelSetState = this.models.ev.setState.sub(this._onSetState.bind(this));
            this._modelSetClass = this.models.ev.setClass.sub(this._onSetClass.bind(this));
            this._modelInsertedRate = this.models.ev.insertRate.sub(this._onInsertedRate.bind(this));
            this._modelRemovedRate = this.models.ev.removeRate.sub(this._onRemovedRate.bind(this));
            this._modelSetRate = this.models.ev.setRate.sub(this._onSetRate.bind(this));
            
            this._q = [[new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk)],
                       [new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk)]
                      ];
            this._running = 0;
            this._serial = 0;
            this.paused = false;
            this.ev = qubx.makeEvents(['clear', 'chunk', 'run']);
            this.updateConfig();
            this.update();
        }
        dispose() {
            this.worker.postMessage({t: 'STOP'});
            this.models.ev.setChannelCount.unsub(this._modelSetChannelCount);
            this.models.ev.setIeqFv.unsub(this._modelSetIeqFv);
            this.models.ev.setVRev.unsub(this._modelSetVRev);
            this.models.ev.insertState.unsub(this._modelInsertedState);
            this.models.ev.removeState.unsub(this._modelRemovedState);
            this.models.ev.setState.unsub(this._modelSetState);
            this.models.ev.setClass.unsub(this._modelSetClass);
            this.models.ev.insertRate.unsub(this._modelInsertedRate);
            this.models.ev.removeRate.unsub(this._modelRemovedRate);
            this.models.ev.setRate.unsub(this._modelSetRate);
            this.models.ev.insert.unsub(this._modelInsertModel);
            this.models.ev.remove.unsub(this._modelRemoveModel);
            this.models.ev.activate.unsub(this._modelActivateModel);
        }
        _onMessage(event) {
            if ( event.data.serial === this._serial ) {
                if ( (event.data.first === 0) && (event.data.iSeg === 0) ) {
                    this.ev.clear.trigger(event.data);
                }
                if ( event.data.n ) {
                    this.ev.chunk.trigger(event.data);
                }
            }
            this._running -= 1;
            this._q.push(event.data.buffers);
            this._postFill();
        }
        _onError(e) {
            console.log(e);
            qubx.errOut.send(e);
        }
        _onInsertModel(m, i) {
            this.update();
        }
        _onRemoveModel(m, i) {
            this.update();
        }
        _onActivateModel(m, i, active) {
            this.update();
        }
        _onSetChannelCount(m, n) {
            this.updateChannelCount(m.index, n);
        }
        _onSetIeqFv(m, n) {
            this.updateOne(m.index);
        }
        _onSetVRev(m, n) {
            this.updateOne(m.index);
        }
        _onInsertedState(m, row, justActivating) {
            if ( ! justActivating) {
                this.updateOne(m.index);
            }
        }
        _onRemovedState(m, row) {
            this.updateOne(m.index);
        }
        _onSetState(m, info) {
            if ( (info.column === 'Class') || ((info.column === 'Pr') && (! this.config.usePeq)) ) {
                this.updateOne(m.index);
            }
        }
        _onSetClass(m, info) {
            this.updateOne(m.index);
        }
        _onInsertedRate(m, row, justActivating) {
            if ( (! justActivating) && (row.Index % 2) ) {
                this.updateOne(m.index);
            }
        }
        _onRemovedRate(m, row) {
            if ( (row.Index % 2) === 0 ) {
                this.updateOne(m.index);
            }
        }
        _onSetRate(m, info) {
            if ( ['k0', 'k1', 'k2'].indexOf(info.column) !== -1 ) {
                this.updateRate(m.index, info);
            }
            else {
                this.update();
            }
        }
        reconfig(cfg) {
            this.config = Object.assign(this.config, cfg);
            this.updateConfig();
        }
        updateConfig() {
            if ( this.paused ) {
                return;
            }
            //qubx.log({t: 'CONFIG', config: this.config});
            this.worker.postMessage({t: 'CONFIG', config: this.config});
            this.schedule();
        }
        update(anyway) {
            if ( this.paused && ! anyway ) {
                return;
            }
            //qubx.log({t: 'MODEL', model: this.model.toJson()});
            this.worker.postMessage({
                t: 'MODELS',
                models: this.models.models.map(function(m) {
                    return m.model.active ? m.model.toJson() : '';
                })
            });
            this.schedule();
        }
        updateOne(index) {
            if ( this.paused ) {
                return;
            }
            this.worker.postMessage({
                t: 'MODEL',
                i: index,
                model: this.models.models[index].model.toJson()
            });
            this.schedule();
        }
        updateRate(index, info) {
            if ( this.paused ) {
                return;
            }
            //qubx.log({t: 'RATE', info: info});
            this.worker.postMessage({t: 'RATE', i: index, info: info});
            this.schedule();
        }
        updateChannelCount(index, n) {
            if ( this.paused ) {
                return;
            }
            //qubx.log({t: 'CC', n: n});
            this.worker.postMessage({t: 'CC', i: index, n: n});
            this.schedule();
        }
        schedule() {
            if ( this._scheduled ) {
                clearTimeout(this._scheduled);
            }
            if ( this.config.auto ) {
                this._scheduled = setTimeout(this.run.bind(this), 100);
            }
        }
        run() {
            this._segmentSize = this.config.segmentSize;
            this._segmentIndVar = this.config.segmentIndVar;
            this._serial = (this._serial + 1) % 128;
            this._sampling = this.config.sampling;
            this._iSeg = 0;
            this._first = 0; // offset in segment
            this.ev.run.trigger();
            if ( this._running ) {
                return; // will postFill on receipt
            }
            while ( this._postFill() ) {
            }
        }
        _postFill() {
            if ( this._iSeg >= this._segmentSize.length ) {
                return false;
            }
            while ( this._first === this._segmentSize[this._iSeg] ) {
                ++this._iSeg;
                this._first = 0;
                if ( this._iSeg === this._segmentSize.length ) {
                    return false;
                }
            }
            const iSeg = this._iSeg,
                  first = this._first,
                  total = this._segmentSize[iSeg],
                  n = Math.min(total-first, this.config.chunk);
            if ( n && this._q.length ) {
                let buffers = this._q.pop();
                if ( buffers[0].byteLength !== (4*this.config.chunk) ) {
                    buffers = [new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk), new ArrayBuffer(4*this.config.chunk)]; ///
                }
                this._running += 1;
                const msg = {
                    t: 'FILL',
                    buffers: buffers,
                    segmentSize: this._segmentSize,
                    indVar: this._segmentIndVar[iSeg],
                    iSeg: iSeg,
                    first: first,
                    n: n,
                    total: total,
                    sampling: this._sampling,
                    serial: this._serial
                };
                this.worker.postMessage(msg, buffers);
                this._first += n;
                return true;
            }
            return false;
        }
    };

    return exports;
});
