/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_single", ["qubx", "qubx_optimize", "qubx_workers", "qubx_xmeans_math"], (qubx, qubx_optimize, qubx_workers, qubx_xmeans_math) => {
    "use strict";
    
    const exports = {};

    const SingleWorkgroupDefs = {
        count: 3,
        skmMaxChunk: 1024*1024
    };
    exports.SingleWorkgroup = class SingleWorkgroup {
        constructor(config) {
            this.skmModel = this.skmSampling = null;
            this.config = Object.assign({}, SingleWorkgroupDefs, config);
            this.layout = null;
            this._workers = new qubx.PromisedStock(() => {
                const w = new qubx_workers.Worker({
                    href: 'js/qubx_single_worker.js'
                    //hrefMin: 'js/qubx_single_worker.min.js'
                })
                      .method('stop')
                      .method('setupSKM', true)
                      .method('SKM')
                      .method('setLayout', true)
                      .method('setDelay', true)
                      .method('setConstant', true)
                      .method('getOptPars')
                      .method('getLL')
                      .method('getOptStd');
                this.skmModel && w.setupSKM({
                    model: this.skmModel,
                    sampling: this.skmSampling,
                    maxN: this.config.skmMaxChunk
                });
                this.layout && w.setLayout({layout: this.layout});
                return w;
            }, 0, w => w.stop());
            /// this._workers.items: all workers, even if out of stock
            this.count(this.config.count);
        }
        destroy() {
            this.count(0);
        }
        count(c) {
            if ( typeof(c) === 'undefined' ) {
                return this._workers.instances();
            }
            if ( c === this._workers.instances() ) {
                return this;
            }
            this.config.count = c;
            this._workers.instances(c);
            return this;
        }
        doWork(f) {
            return this._workers.get()
                .then(worker => f(worker).then(result => {
                    this._workers.put(worker);
                    return result;
                }));
        }
        SKM(model, sampling, segments, getSamples, maxIter, fixA, fixAmp, fixStd) {
            this.skmModel = model;
            this.skmSampling = sampling;
            this._workers.items.forEach(worker => {
                worker.setupSKM({
                    model: model,
                    sampling: sampling,
                    maxN: this.config.skmMaxChunk
                });
            });
            const maxChunk = this.config.skmMaxChunk,
                  chunks = [];
            segments.forEach(seg => {
                let offset = 0;
                for (; offset<seg.n; offset += maxChunk) {
                    chunks.push({
                        seg: seg,
                        offset: offset,
                        n: Math.min(maxChunk, seg.n-offset)
                    });
                }
            });
            const idlsegs = segments.map(() => ({chunks: []}));
            return Promise.all(chunks.map(chunk =>this.doWork(worker => {
                const data = getSamples(chunk.seg, chunk.offset, chunk.n);
                return worker.SKM({
                    buffers: [data.buffer],
                    iSeg: chunk.seg.index,
                    offset: chunk.offset,
                    maxIter: maxIter,
                    fixA: fixA,
                    fixAmp: fixAmp,
                    fixStd: fixStd
                }, [data.buffer]);
            }))).then(idlchunks => {
                idlchunks.forEach(chunk => idlsegs[chunk.iSeg].chunks.push(chunk));
                return idlsegs.map((seg, iSeg) => {
                    const segOut = mergeChunks(seg.chunks);
                    segOut.iSeg = iSeg;
                    segOut.offset = 0;
                    return segOut;
                });
            });
        }
        setLayout(layout) {
            this.layout = layout;
            this._workers.items.forEach(worker => worker.setLayout({layout: layout}));
        }
        setDelay(name, value) {
            if ( this.layout ) {
                this.layout.delays[name] = value;
            }
            this._workers.items.forEach(w => w.setDelay({
                name: name,
                value: value
            }));
        }
        setConstant(name, value) {
            if ( this.layout ) {
                this.layout.defaults[name] = value;
            }
            this._workers.items.forEach(w => w.setConstant({
                name: name,
                value: value
            }));
        }
        setModel(model, tdead, tcrit, hjc, chs, msl) {
            return Promise.all(this._workers.items.map(w => w.getOptPars({
                model: model,
                tdead: tdead,
                tcrit: tcrit,
                hjc: hjc,
                chs: chs,
                msl: msl
            }))).then(paramses => paramses[0]);
        }
        getLL(params, tcrit) {
            if ( ! (this.layout && this.layout.segments.length) ) {
                return Promise.resolve({
                    ll: 0.0,
                    resultCode: 0
                });
            }
            return this.doWork(worker => worker.getLL({
                params: params,
                tcrit: tcrit
            })).then(result => {
                result.params = params;
                result.aic = 2*params.length - 2*result.ll;
                result.bic = params.length * Math.log(result.ndwell) - 2*result.ll;
                return result;
            });
        }
        getOptStd(params, invHessian) {
            if ( ! (this.layout && this.layout.segments.length) ) {
                return Promise.resolve({});
            }
            return this.doWork(worker => worker.getOptStd({
                params: params,
                invHessian: invHessian
            }));
        }
        optimize(model, tdead, tcrit, opts) {
            /* opts: {
               dfp, simplex: boolean; or:
               OptClass: e.g. qubx_optimize.Simplex or qubx_optimize.DFP
               onIteration: function(model, params, ll, iteration[, grad])
               ...any constructor config options for OptClass, in particular:
               maxIter
               epsilon
               }*/
            let lastModel = {};
            const score = params => this.getLL(params.slice(), tcrit)
                  .then(data => {
                      lastModel = data.model;
                      return data.resultCode ? 1e20 : -data.ll;
                  });
            const iter = (params, score, iteration, grad) => {
                if ( opts.onIteration ) {
                    return opts.onIteration(lastModel, params, -score, iteration, grad);
                } else {
                    return true;
                }
            };
            const run = (OptClass, params, randomize) => {
                const pp = params.slice();
                if ( randomize ) {
                    console.log(params);
                    let i = 0;
                    for (; i<pp.length; ++i) {
                        pp[i] *= 1+randomize*qubx.gaussRandom();
                    }
                    console.log(pp);
                }
                return (new OptClass(pp, opts)).run({
                    score: score,
                    grad: qubx_optimize.numGrad2Side(score),
                    iter: iter,
                    dx: opts.dx
                }).then(result => {
                    result.ll = - result.score;
                    if ( result.hessian ) {
                        return this.getOptStd(result.params, [].concat.apply([], result.hessian)) // from 2d arrays to 1d
                            .then(modelResult => {
                                result.model = modelResult.model;
                                return result;
                            });
                    } else {
                        result.model = lastModel;
                        return result;
                    }
                });
            };
            return this.setModel(model, tdead, tcrit, opts.hjc, opts.chs, opts.msl).then(ready => {
                let optResult = Promise.resolve({
                    params: ready.params,
                    hjc: ready.hjc,
                    score: 0,
                    model: model
                });
                const chain = (OptClass, randomize) => {
                    optResult = optResult.then(result => run(OptClass, result.params, randomize));
                };
                if ( opts.simplex || opts.dfp ) {
                    opts.simplex && chain(qubx_optimize.Simplex);
                    opts.dfp && chain(qubx_optimize.DFP, opts.simplex && .02);// dfp needs distance from solution to develop hessian error estimates
                } else {
                    chain(opts.OptClass || qubx_optimize.Simplex);
                }
                return optResult.then(optResult => this.getLL(optResult.params, tcrit)
                                      .then(llResult => {
                                          optResult.ndata = optResult.ndwell = llResult.ndwell;
                                          optResult.aic = llResult.aic;
                                          optResult.bic = llResult.bic;
                                          optResult.hjc = ready.hjc;
                                          optResult.msl = ready.msl;
                                          return optResult;
                                      }));
            });
        }
    };
    
    function mergeChunks(chunks) {
        if ( chunks.length === 1 ) {
            return chunks[0];
        }
        chunks.sort((a, b) => a.offset - b.offset);

        // build a table of all clusters' classes' amps and stds
        // and use K-Means to group them into Nclass output classes
        let Ndwell = 0;
        const Nclass = chunks.length ? chunks[0].amp.length : 0,
            ampCol = [],
            stdCol = [],
            weight = [],
            initColor = [];
        chunks.forEach(chunk => {
            Ndwell += chunk.classes.length;
            const occCount = chunk.amp.map(() => 0);
            for (const cc of chunk.counts) {
                occCount[chunk.classes[i]] += cc;
            }
            for (i=0; i<Nclass; ++i) {
                ampCol.push(chunk.amp[i]);
                stdCol.push(chunk.std[i]);
                weight.push(occCount[i]);
                initColor.push(i+1);
            }
        });
        const xmeans = new qubx_xmeans_math.Calc();
        xmeans.columns = [ampCol, stdCol];
        xmeans.weight = weight;
        const clusters = xmeans.Iterate(Nclass, initColor, 500);

        // now we have clusters.color[of each (chunk,cls)]
        // remap colors to have clusters.mean[0][k] increasing in "bigness"
        // (renumbering clusters.color)
        let biggestAbs = 0,
            biggestK = -1;
        const colors = clusters.mean[0].slice(1).map((amp, k) => {
            const absAmp = Math.abs(amp);
            if ( absAmp > biggestAbs ) {
                biggestAbs = absAmp;
                biggestK = k;
            }
            const votes = [];
            let i = 0;
            for (; i<Nclass; ++i) {
                votes[i] = 0;
            }
            let maxVotes = 0,
                maxVotesC = -1;
            // look through ampstd table [chunk, c] where color=k+1
            // and increment votes[c] (tracking maxVotes, maxVotesC)
            clusters.color.forEach((color, i) => {
                if ( color === (k+1) ) {
                    const c = i%Nclass;
                    votes[c] += 1;
                    if ( votes[c] > maxVotes ) {
                        maxVotes = votes[c];
                        maxVotesC = c;
                    }
                }
            });
            return {
                amp: amp,
                std: clusters.mean[1][k+1],
                k: k+1,
                votedK: maxVotesC
            };
        });
        // sort them by their voted class number, try to fulfill if possible
        colors.sort((a, b) => a.votedK - b.votedK);
        const remap = [],
              overflow = [],
              taken = [];
        let lastK = -1;
        colors.forEach((c,i) => {
            if ( (lastK + 1) === Nclass ) {
                overflow.push(c.k);
            } else if ( c.votedK <= lastK ) {
                remap[c.k] = lastK + 1;
            } else {
                remap[c.k] = c.votedK;
            }
            lastK = remap[c.k];
            taken[lastK] = true;
        });
        let nextK = 0;
        while ( overflow.length && (nextK < Nclass) ) {
            if ( ! taken[nextK] ) {
                remap[overflow.splice(0, 1)[0]] = nextK;
            }
            ++nextK;
        }
        clusters.color = clusters.color.map(k => remap[k]);
        
        // fill out amp, std
        const amp = new Float64Array(Nclass),
              std = new Float64Array(Nclass);
        let k = 0;
        for (; k<Nclass; ++k) {
            amp[k] = colors[k].amp;
            std[k] = colors[k].std;
        }

        // transcribe counts and classes
        //   replacing (k=classes[i]) with clusters.color[Nclass*iChunk + k]
        const counts = new Int32Array(Ndwell),
              classes = new Int32Array(Ndwell);
        let at = 0;
        chunks.forEach((chunk, iChunk) => {
            const n = chunk.counts.length;
            let i = 0;
            for (; i<n; ++i, ++at) {
                counts[at] = chunk.counts[i];
                classes[at] = clusters.color[Nclass*iChunk + chunk.classes[i]];
            }
        });
        
        return {
            counts: counts,
            classes: classes,
            amp: amp,
            std: std
        };
    }
    
    return exports;
});
