/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"use strict";

importScripts('qubx_worker_require.js');
//importScripts('qubx_worker_require.min.js');

const QUBX_LIB_MEMORY = 2*16*1024*1024;

qubx_require(['qubx_lib', 'qubx_worker_util', 'hjcfit']).then(libs => {
    const [lib, util, hjcfit] = libs;

    const skm = {
        sameModel: function(ma, mb) {
            if ( ! (ma && mb) ) {
                return false;
            }
            if ( ma.channelCount !== mb.channelCount ) {
                return false;
            }
            if ( ma.states.length !== mb.states.length ) {
                return false;
            }
            if ( ma.rates.length !== mb.rates.length ) {
                return false;
            }
            const ns = ma.states.length;
            let i = 0;
            for (; i<ns; ++i) {
                const c = ma.states[i].Class;
                if ( c !== mb.states[i].Class ) {
                    return false;
                }
                if ( ma.classes[c].Amp !== mb.classes[c].Amp ) {
                    return false;
                }
                if ( ma.classes[c].Std !== mb.classes[c].Std ) {
                    return false;
                }
            }
            const nr = ma.rates.length;
            for (i=0; i<nr; ++i) {
                const ra = ma.rates[i],
                      rb = mb.rates[i];
                if ( (ra.From !== rb.From) || (ra.To !== rb.To) || (ra.k0 !== rb.k0) ) {
                    return false;
                }
            }
            return true;
        },
        destroy: function() {
            this.skm && this.skm.destroy();
            this.skmodel && this.skmodel.destroy();
        },
        setup: function(model, sampling, maxN) {
            model.channelCount = Math.min(5, model.channelCount);
            if ( this.skm ) {
                if ( this.sameModel(model, this.modelQmj) ) {
                    this.skm.reset(1e3*sampling);
                } else {
                    this.skm.destroy();
                    this.skmodel.destroy();
                    this.skm = this.skmodel = null;
                }
            }
            if ( ! this.skm ) {
                this.modelQmj = model;
                this.maxN = maxN;
                this.skmodel = lib.modelFromJson(model);
                this.skm = new lib.SKM(1e3*sampling, maxN, this.skmodel);
            }
        },
        run: function(data, maxIter, fixA, fixAmp, fixStd) {
            const out = {};
            if ( maxIter ) {
                const result = this.skm.iterate(data, null, maxIter, fixA, fixAmp, fixStd);
                out.ll = result.ll;
                out.iterations = result.iterations;
            } else {
                out.ll = this.skm.viterbi(data);
                out.iterations = 0;
            }
            const events = this.skm.getEvents();
            ['amp', 'std', 'classes', 'counts'].forEach(field => {
                out[field] = events[field];
            });
            return out;
        },
        setupMsg: function(data) {
            this.setup(data.model, data.sampling, data.maxN);
        },
        runMsg: function(data) {
            // buffers[0], n[, maxIter[, fixA, fixAmp, fixStd]]
            // => { ll, classes, counts, amp, std[, iterations] }
            const buffers = data.buffers;
            data.buffers = [];
            if ( ! this.skm ) {
                throw new Error('missing SKM_SETUP');
            }
            const result = this.run(new Float32Array(buffers[0], 0, data.n),
                                  data.maxIter, data.fixA, data.fixAmp, data.fixStd);
            result.transfer = [];
            Object.keys(result).forEach(field => {
                const x = result[field];
                data[field] = x;
                if ( x && x.buffer ) {
                    result.transfer.push(x.buffer);
                }
            });
        }
    };

    const opt = {
        ctx: new lib.MSL(0),
        idlData: new lib.IdlArraySegments(),
        signals: new util.SignalMap(),
        destroy: function() {
            this.reset();
            this.layout = null;
            this.idlData.destroy();
            this.ctx.destroy();
        },
        reset: function() {
            if ( this.libModel ) {
                this.ratesm.destroy();
                this.libModel.destroy();
            }
            if ( this.multiModel ) {
                this.multiModel.destroy();
                this.optModel.destroy();
                this.optData.destroy();
            }
            if ( this.hjcData ) {
                this.hjcData.destroy();
            }
            this.hjcData = this.optData = this.libModel = this.multiModel = this.optModel = this.ratesm = this.modelQmj = null;
        },
        setLayout: function(data) {
            const layout = data.layout;
            this.reset();
            this.layout = layout;
            // deduplicate stimulus class indices across segments; they are multiplexed by amp anyway, but this helps estimate max multiplexed classes
            const stimAmp = {};
            let empty = false;
            layout.segments.forEach((seg, iSeg) => {
                Object.keys(seg.stimuli).forEach(name => {
                    if ( empty ) {
                        return;
                    }
                    const stim = seg.stimuli[name],
                          amp = stimAmp[name] || [];
                    stimAmp[name] = amp;
                    const nextC = amp.length,
                          na = stim.amps.length;
                    let i = 0;
                    for (; i<na; ++i) {
                        amp.push(stim.amps[i]);
                    }
                    const nd = stim.classes.length;
                    for (i=0; i<nd; ++i) {
                        if ( stim.classes[i] < 0 ) {
                            empty = true;
                            return;
                        }
                        stim.classes[i] += nextC;
                    }
                });
            });
            if ( empty ) {
                layout.segments = [];
            }
            layout.segments.forEach((seg, iSeg) => {
                Object.keys(seg.stimuli).forEach(name => {
                    seg.stimuli[name].amps = stimAmp[name];
                });
            });
        },
        setDelay: function(data) {
            if ( this.layout ) {
                const name = data.name,
                      value = data.value;
                this.layout.delays[name] = value;
            }
            this.reset();
        },
        setConstant: function(data) {
            const name = data.name,
                  value = data.value;
            if ( this.layout && this.layout.defaults[name] === value ) {
                return;
            }
            if ( this.layout ) {
                this.layout.defaults[name] = value;
            }
            this.reset();
        },
        copyMatchingModel: function(model) {
            if ( ! this.modelQmj ) {
                return false;
            }
            if ( this.modelQmj.channelCount !== model.channelCount ) {
                return false;
            }
            if ( this.modelQmj.states.length !== model.states.length ) {
                return false;
            }
            const ns = model.states.length;
            let i = 0;
            for (; i<ns; ++i ) {
                if ( this.modelQmj.states[i].Class !== model.states[i].Class ) {
                    return false;
                }
            }
            if ( this.modelQmj.rates.length !== model.rates.length ) {
                return false;
            }
            const nr = model.rates.length;
            for (i=0; i<nr; ++i ) {
                const a = this.modelQmj.rates[i],
                      b = model.rates[i];
                if ( (a.From !== b.From) ||
                     (a.To !== b.To) ||
                     (a.Ligand !== b.Ligand) ||
                     (a.Voltage !== b.Voltage) ||
                     (a.Pressure !== b.Pressure) ) {
                    return false;
                }
                a.k0 = b.k0;
                a.k1 = b.k1;
                a.k2 = b.k2;
                const ij = a.From*model.states.length+a.To
                this.libModel.K0(ij, b.k0);
                this.libModel.K1(ij, b.k1);
                this.libModel.K2(ij, b.k2);
            }
            this.multiModel && this.multiModel.update();
            return true;
        },
        shouldHJC: function(model, wantHJC, wantMSL) {
            if ( wantMSL || ! wantHJC ) {
                return false;
            }
            if ( this.signals.delays.length > 1 ) {
                return false;
            }
            if ( model.channelCount > 1 ) {
                return false;
            }
            for (const s in model.states) {
                if ( s.Class > 1 ) {
                    return false;
                }
            }
            return true;
        },
        updateModel: function(model, tdead, hjc, msl) {
            this.signals.update(model, this.layout);
            hjc = this.shouldHJC(model, hjc, msl);
            this.msl = msl;
            this.mslChosen = msl || (this.signals.defaults.length > 1);
            if ( (this.hjc !== hjc) || (! this.copyMatchingModel(model)) ) {
                this.hjc = hjc;
                this.reset();
                this.modelQmj = model;
                this.libModel = lib.modelFromJson(model);
                if ( hjc ) {
                    this.setupHJC(model, tdead);
                } else {
                    this.setupMXL(model, tdead);
                }
            }
        },
        setupHJC: function(model, tdead) {
            let ndwell = 0;
            this.hjcData = new hjcfit.Bursts();
            this.layout.segments.forEach(seg => {
                const durs = hjcTrimAndApplyDeadTime(seg.idl.classes, seg.idl.durations, 1e3*tdead);
                if ( durs.length ) {
                    this.hjcData.append(durs);
                    ndwell += durs.length;
                }
            });
            this.ndwell = ndwell;
        },
        setupMXL: function(model, tdead) {
            this.multiModel = new lib.MultiModel(this.libModel);
            const segments = this.layout.segments.map(seg => {
                const signals = [seg.idl];
                let iSig = 1;
                for (; iSig<this.signals.delays.length; ++iSig) {
                    signals.push(seg.stimuli[this.signals.names[iSig]]);
                }
                for (iSig=this.signals.delays.length; iSig<this.signals.defaults.length; ++iSig) {
                    signals.push({
                        classes: [0],
                        durations: [1e3*seg.dur],
                        amps: [this.signals.defaults[iSig]]
                    });
                }
                return {
                    signals: signals
                };
            });
            this.idlData.set(segments);
            this.optData = this.ctx.newData(this.idlData, tdead);
            let dim = 4;
            while ( dim < model.states.length ) {
                dim *= 2;
            }
            this.optModel = this.ctx.newModels(dim, 1, this.optData.Nstim);
        },
        getOptPars: function(data) {
            const model = data.model,
                  tdead = data.tdead,
                  tcrit = data.tcrit,
                  hjc = data.hjc,
                  chs = data.chs,
                  msl = data.msl;
            if ( model.channelCount > 5 ) {
                throw new Error('Channel count is too high for MIL');
            }
            this.updateModel(model, tdead, hjc, msl);
            this.tdead = tdead;
            this.tcrit = tcrit;
            this.chs = chs;
            if ( this.ratesm ) {
                this.ratesm.destroy();
            }
            const scFrac = lib.dvec().size(1).set(0, 1); /// why 1 stimulus level? with hjc we find out if it works with Nstim=0
            this.ratesm = new lib.StimRates(this.libModel, this.optData|0, this.signals.names.length, this.optData && this.optData.stimclasses.p, scFrac.p, 0.0);
            scFrac.destroy();
            model.constraints.forEach(cns => {
                const Nstate = model.states.length,
                      ncs = cns.States.length;
                let i = 0;
                for (; i<ncs; ++i) {
                    if ( cns.States[i] >= Nstate ) {
                        return; // skip invalid cns
                    }
                }
                if ( cns.Type === 'FixRate' ) {
                    this.ratesm.fixRate(cns.States[0], cns.States[1]);
                } else if ( cns.Type === 'FixExp' ) {
                    this.ratesm.fixExp(cns.States[0], cns.States[1]);
                } else if ( cns.Type === 'FixPress' ) {
                    this.ratesm.fixPres(cns.States[0], cns.States[1]);
                } else if ( cns.Type === 'ScaleRate' ) {
                    this.ratesm.scaleRate(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                } else if ( cns.Type === 'ScaleExp' ) {
                    this.ratesm.scaleExp(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                } else if ( cns.Type === 'ScalePress' ) {
                    this.ratesm.scalePres(cns.States[0], cns.States[1], cns.States[2], cns.States[3]);
                }
            });
            if ( model.balanceLoops ) {
                this.ratesm.balanceLoops();
            }
            this.ratesm.setupConstraints();
            const params = [],
                  np = this.ratesm.cns.NfPar();
            let ip = 0;
            for (; ip<np; ++ip) {
                params.push(this.ratesm.cns.fPars(ip));
            }
            return {
                params: params,
                hjc: this.hjc,
                msl: this.mslChosen
            };
        },
        setParams: function(params) {
            const NfPar = this.ratesm ? this.ratesm.cns.NfPar() : 0;
            params.forEach((p,i) => {
                if ( i < NfPar ) {
                    this.ratesm.cns.fPars(i, p);
                } else {
                    console.log('Extra parameter',i,p);
                }
            });
            this.ratesm.cns.freeToPars();
            this.ratesm.parsToModel();
            this.updateQmj();
            if ( this.hjc ) {
                const stimulus = {};
                let i = this.signals.delays.length;
                for (; i<this.signals.names.length; ++i) {
                    stimulus[this.signals.names[i]] = this.signals.defaults[i];
                }
                const info = hjcMakeQ(this.modelQmj, stimulus);
                this.numOpen = info.numOpen;
                this.Q = info.Q;
            } else {
                this.multiModel.update();
                this.optModel.reset();
                this.optModel.setupModel((this.modelQmj.channelCount > 1) ? this.multiModel.multi : this.libModel, this.optData);
            }
        },
        updateQmj: function() {
            const model = this.modelQmj,
                  n = model.rates.length;
            let i = 0;
            for (; i<n; ++i ) {
                const r = this.modelQmj.rates[i],
                      ij = r.From*model.states.length + r.To;
                r.k0 = this.libModel.K0(ij);
                r.k1 = this.libModel.K1(ij);
                r.k2 = this.libModel.K2(ij);
            }
        },
        getLL: function(data) {
            const params = data.params;
            this.setParams(params);
            if ( this.hjc ) {
                let ll = -1e20;
                try {
                    ll = Math.log(10) * hjcfit.likelihood(this.hjcData, this.Q, this.numOpen, 1e3*this.tdead, 1e3*this.tcrit, this.chs);
                } catch (err)
                {}
                if ( ! isFinite(ll) ) {
                    ll = -1e21;;
                }
                return {
                    ll: ll,
                    ndwell: this.ndwell,
                    code: 0,
                    model: this.modelQmj
                };
            } else {
                const outCode = this.mslChosen ?
                    this.ctx.msl(this.optData, this.optModel, 0) :
                    this.ctx.mil(this.optData, this.optModel, 0);
                return {
                    ll: this.optModel.getLL()[0],
                    ndwell: this.optData.Ndwell,
                    code: outCode,
                    model: this.modelQmj
                };
            }
        },
        getOptStd: function(data) {
            const params = data.params,
                  invHessian = data.invHessian;
            this.setParams(params);
            const pHess = lib.dvec().set(invHessian);
            this.ratesm.calcStd(pHess.p);
            pHess.destroy();
            const Nstate = this.libModel.Nstate(),
                  model = JSON.parse(JSON.stringify(this.modelQmj));
            model.rates.forEach((r,i) => {
                const j = r.From*Nstate + r.To;
                r.k0_std = this.libModel.d_K0(j);
                r.k1_std = this.libModel.d_K1(j);
                r.k2_std = this.libModel.d_K2(j);
            });
            return {
                model: model
            };
        }
    };

    function hjcMakeQ(model, stimulus) {
        const Nstate = model.states.length,
              Q = [];
        let i = 0;
        for (; i<Nstate; ++i) {
            const row = [];
            let j = 0;
            for (; j<Nstate; ++j) {
                row.push(0.0);
            }
            Q.push(row);
        }
        for (const rate of model.rates) {
            let k = rate.k0;
            if ( rate.Ligand && (rate.Ligand in stimulus) ) {
                k *= stimulus[rate.Ligand];
            }
            let kExp = 0.0;
            if ( rate.Voltage && (rate.Voltage in stimulus) ) {
                kExp += rate.k1 * stimulus[rate.Voltage];
            }
            if ( rate.Pressure && (rate.Pressure in stimulus) ) {
                kExp += rate.k2 * stimulus[rate.Pressure];
            }
            k *= Math.exp(kExp);
            Q[rate.From][rate.To] = k;
        }
        for (i=0; i<Nstate; ++i) {
            let sum = 0.0,
                j = 0;
            for (; j<Nstate; ++j) {
                sum += Q[i][j];
            }
            Q[i][i] = - sum;
        }

        // reordered with open states first
        const clazz = model.states.map(s => s.Class),
              sourceState = [];
        let numOpen = 0;
        clazz.forEach((c,i) => {
            if ( c ) {
                ++numOpen;
                sourceState.push(i);
            }
        });
        clazz.forEach((c,i) => c || sourceState.push(i));
        const Q_ordered = Q.map(row => row.slice());
        for (i=0; i<Nstate; ++i) {
            let j = 0;
            for (; j<Nstate; ++j) {
                Q_ordered[i][j] = Q[sourceState[i]][sourceState[j]];
            }
        }
        return {
            numOpen: numOpen,
            Q: Q_ordered
        };
    }
    
    function hjcTrimAndApplyDeadTime(classes, durations, tdead) {
        // trims events if they are (a) initial cls=0 (b) final cls=0 (c) initial dur<td
        // joins subsequent events with dur<td to their neighbors
        // returns new durations array
        let L = durations.length;
        if ( ! L ) {
            return durations;
        }
        let firstAlive = 0;
        while ( firstAlive < L ) {
            if ( (durations[firstAlive] < tdead) || (classes[firstAlive] === 0) ) {
                ++firstAlive;
            } else {
                break;
            }
        }
        const dursOut = [];
        let lastCls = 0,
            iRd = firstAlive;
        --L; // handling the final dwell separately in case it's cls0
        for (; iRd<L; ++iRd) {
            if ( (classes[iRd] === lastCls) || (durations[iRd] < tdead) ) {
                dursOut[dursOut.length-1] += durations[iRd];
            } else {
                dursOut.push(durations[iRd]);
            }
            lastCls = classes[iRd];
        }
        if ( durations[L] < tdead ) {
            dursOut[dursOut.length-1] += durations[L];
        } else if ( classes[L] ) {
            dursOut.push(durations[L]);
        }
        if ( dursOut.length && ! (dursOut.length % 2) ) { // prevent ending on cls0
            dursOut.pop();
        }
        return dursOut;
    }
    
    util.listen({
        // trace: true,
        stop: data => {
            try {
                skm.destroy();
                opt.destroy();
            } catch(err) {
                console.log(err);
            }
            close();
        },
        
        setupSKM: skm.setupMsg.bind(skm),
        SKM: skm.runMsg.bind(skm),

        setLayout: opt.setLayout.bind(opt),
        setDelay: opt.setDelay.bind(opt),
        setConstant: opt.setConstant.bind(opt),
        getOptPars: opt.getOptPars.bind(opt),
        getLL: opt.getLL.bind(opt),
        getOptStd: opt.getOptStd.bind(opt)
    });
}).catch(err => postMessage({t: 'LOADERR', err}));

//// Idealization:
// SKM_SETUP {model, maxN, sampling}
// SKM {buffers: [Float32Array], n[, maxIter[, fixA, fixAmp, fixStd]]}
//     => {ll, classes, counts, amp, std[, iterations]}

//// Optimization:
// LAYOUT {layout:
//   sampling: (sec),
//   signalNames: ['Current', 'Voltage', ...]
//   delays: [0, 0, ...],
//   defaults: [0, 0, ...],
//   segments: {
//       idl: {classes, durations},
//       stimuli: {
//           Voltage: {classes, durations, amp},
//           ...
//       },
//       dur: (sec)
//   }
// }}
// DELAY {name, value}
// CONSTANT {name, value)
// OPTPARS {model, tdead, hjc, chs, msl} => {params, hjc}
// LL {[params], tcrit} => {ndwell, ll}
// OPTSTD {params, invHessian} => {model}
