/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

// vibration support
navigator.vibrate = navigator.vibrate || navigator.webkitVibrate || navigator.mozVibrate || navigator.msVibrate;

define ("qubx_ui", ["qubx", "d3", "element-resize-detector"], (qubx, d3, elementResize) => {
    'use strict';

    const exports = {};
    
    const tmpl = qubx.get("qubx_templates.html").then(data => {
        const frag = document.createDocumentFragment();
        const templates = document.createElement('div');
        frag.appendChild(templates);
        templates.innerHTML = data;
        return templates;
    });
    const loadTemplate = function loadTemplate(query) {
        return tmpl.then(templates => templates.querySelector(query).cloneNode(true));
    };
    const loadTemplates = function loadTemplates(queries) {
        return tmpl.then(templates => queries.map(query => {
            const t = templates.querySelector(query);
            return t && t.cloneNode(true);
        }));
    };
    exports.loadTemplate = loadTemplate;
    exports.loadTemplates = loadTemplates;

    exports.eventOffset = function eventOffset(e, pageX, pageY) {
        const target = e.target || e.srcElement,
            rect = target.getBoundingClientRect(),
            x = (typeof(pageX) === 'undefined') ? e.clientX : (pageX - (document.body.scrollLeft + document.documentElement.scrollLeft)),
            y = (typeof(pageY) === 'undefined') ? e.clientY : (pageY - (document.body.scrollTop + document.documentElement.scrollTop));
        return {
            x: x - rect.left,
            y: y - rect.top
        };
    };

    exports.escapeHTML = function escapeHTML(html) {
        return document.createElement('div').appendChild(document.createTextNode(html)).parentNode.innerHTML;
    };

    const resizeChecker = elementResize({
        strategy: "scroll"
    });
    exports.onResize = function onResize(node, callback) {
        const info = {node: node, callback: callback, width: 0, height: 0, justResized: false};
        function handler() {
            info.width = node.clientWidth;
            info.height = node.clientHeight;
            try {
                callback(info);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        resizeChecker.listenTo(node, handler);
        return info;
    };
    exports.clearResize = function clearResize(info) {
        resizeChecker.removeAllListeners(info.node);
    };

    exports.dragEnter = node => event => {
        qubx.stopAndPrevent(event);
        node.classList.add('qubx_dragtarget');
    };
    exports.dragLeave = node => event => {
        qubx.stopAndPrevent(event);
        node.classList.remove('qubx_dragtarget');
    };

    exports.querydotdot = function querydotdot(container) {
        // inspired by dart's '..' operator, querydotdot returns a function which
        // queries container and calls action(result)
        // so you can bind the child node to a local name while initializing.
        // By default it returns the query result node, but it will pass along action's return value if truthy
        // (e.g. so you can wrap the node and keep the wrapped reference).
        const dotdot = (selector, action) => {
            const child = container.querySelector(selector);
            return action && action(child) || child;
        };
        // shorthand to add a listener, e.g. dotdot.listen('#myBtn', 'click', function(ev) {})
        dotdot.listen = (selector, evType, handler, action) => dotdot(selector, node => {
            node.addEventListener(evType, handler);
            return action && action(node);
        });
        // shorthand to add a qubx_ui.makeEvents-style listener
        dotdot.sub = (selector, action, evType, handler) => dotdot(selector, node => {
            const wrapped = action(node) || node;
            evType && wrapped.ev[evType].sub(handler);
            return wrapped;
        });
        return dotdot;
    }
    
    exports.Palette = class Palette {
        constructor(node, config) {
            this.node = node;
            const onclick = config ? config.onClick : i => {};
            node.classList.add("qubx_noselect");
            for (let i=0; i<9; ++i) {
                const box = document.createElement('div');
                box.className = 'qubx_palette_box qubx_class'+i;
                const lbl = document.createElement('div');
                lbl.className = 'qubx_palette_lbl';
                lbl.innerHTML = ''+i;
                box.appendChild(lbl);
                node.appendChild(box);
                box.addEventListener('click', event => onclick(i));
            }
        }
    };

    let acceptedValues = localStorage.getItem("qubx_panels.AcceptedInput.values");
    acceptedValues = acceptedValues ? JSON.parse(acceptedValues) : {};
    exports.SaveInputs = function SaveInputs() {
        localStorage.setItem("qubx_panels.AcceptedInput.values", JSON.stringify(acceptedValues));
    };
    setInterval(exports.SaveInputs, 60000);
    
    exports.AcceptedInput = class AcceptedInput {
        constructor(node, value, config) {
            this.node = node;
            this.accept = config && config.accept || qubx.acceptStr;
            this.format = config && config.format || qubx.formatStr;
            this.localKey = config && config.localKey;
            node.style.width = `${config && config.widthChars || 10}em`;
            node.addEventListener('input', this._onInput.bind(this));
            node.addEventListener('change', this._onChange.bind(this));
            this.ev = qubx.makeEvents(['change']);
            if ( this.localKey && (this.localKey in acceptedValues) ) {
                try {
                    value = this.accept(acceptedValues[this.localKey]);
                } catch (e) {}
            }
            this.update(value);
        }
        update(v, step) {
            if ( typeof(v) !== 'undefined' ) {
                this.value = v;
            }
            if ( typeof(step) !== 'undefined' ) {
                this.node.step = step;
            }
            try {
                const fmt = this.format(this.value);
                this.node.value = fmt;
                if ( (typeof(v) !== 'undefined') && this.localKey ) {
                    acceptedValues[this.localKey] = fmt;
                }
            } catch (err) {
                console.log(err);
            }
        }
        _onChange(event) {
            const s = this.node.value;
            try {
                const v = this.accept(s);
                if ( v !== this.value ) {
                    this.value = v;
                    if ( this.localKey ) {
                        acceptedValues[this.localKey] = s;
                    }
                    this.ev.change.trigger(v);
                }
            }
            catch (e) {
                if ( ! s ) { return; }
                const l = s.length, final = s[l];
                if ( (final in {'-':1, '.':1}) || ((l == 1) && (final === '0')) ) { return; }
                this.node.value = this.value;
            }
        }
        _onInput(event) {
            const s = this.node.value;
            if ( (! s) || (s in {'-': true, '.': true}) ) { return; }
            try {
                const v = this.accept(s);
                if ( Math.abs(v - this.value) in {1: true, 10: true} ) { // spinner steps, not keystrokes
                    this.value = v;
                    if ( this.localKey ) {
                        acceptedValues[this.localKey] = s;
                    }
                    this.ev.change.trigger(v);
                }
            }
            catch (e) {
            }
        }
        focus() {
            this.node.focus();
        }
        select() {
            this.node.select();
        }
    };

    exports.setSelectOptions = function setSelectOptions(select, options) {
        const opts = d3.select(select)
            .selectAll("option")
            .data(options)
        opts.enter()
            .append("option")
        opts.exit().remove()
        opts.text(qubx.i);
    };

    exports.Menu = class Menu {
        constructor(config) {
            this.config = {
                node: config.node || document.createElement("select"),
                caption: config.caption || 'Menu...',
                groups: [],
                onfocus: config.onfocus
            };
            this.optOrder = [];
            d3.select(this.config.node)
                .property('size', 1)
                .on('change', () => {
                    const i = this.config.node.selectedIndex;
                    this.config.node.selectedIndex = 0;
                    const d = this.optOrder[i];
                    d.action && d.action();
                }).on('focus', () => {
                    this.config.onfocus && this.optOrder.map(this.config.onfocus);
                });
            config.groups && this.configure({groups: config.groups});
        }
        configure(config) {
            if ( 'node' in config ) {
                this.config.node = config.node;
            }
            if ( 'caption' in config ) {
                this.config.caption = config.caption;
            }
            if ( 'groups' in config ) {
                this.config.groups = [[{caption: this.config.caption}]]
                    .concat(config.groups);
            }
            if ( 'onfocus' in config ) {
                this.config.onfocus = config.onfocus;
            }
            
            const mnu = d3.select(this.config.node);
            const groupSel = mnu.selectAll("optgroup")
                .data(this.config.groups);
            groupSel.enter()
                .append("optgroup")
                .selectAll("option")
                .data(qubx.i)
                .enter().insert("option");
            groupSel.exit().remove();
            const itemSel = groupSel.selectAll("option")
                .data(qubx.i);
            itemSel.enter().insert("option");
            itemSel.exit().remove();
            
            this.optOrder = [];
            mnu.selectAll("optgroup")
                .selectAll("option") // re-select to get both new and existing?
                .text(d => {
                    this.optOrder.push(d);
                    return d.caption;
                });
            
            this.config.node.selectedIndex = 0;
        }
    };

    const PopupDefs = {
        getOptions: () => [],
        onChoose: (label, i) => {},
        width: '1.5em'
    };
    exports.Popup = class Popup {
        constructor(config) {
            this.config = Object.assign({}, PopupDefs, config);
            this.node = config.node || document.createElement('select');
            this.node.style.width = this.config.width;
            const sel = d3.select(this.node);
            sel.append('option');
            this.node.selectedIndex = 0;
            sel.on('focus', this._onFocus.bind(this));
            sel.on('change', this._onChange.bind(this));
        }
        _onFocus() {
            this.labels = this.config.getOptions();
            const options = d3.select(this.node)
                  .selectAll('option')
                  .data([''].concat(this.labels));
            options.enter().insert('option');
            options.exit().remove();
            options.property('value', qubx.i);
            options.text(qubx.i);
        }
        _onChange() {
            const ix = this.node.selectedIndex - 1;
            if ( ix >= 0 ) {
                this.node.selectedIndex = 0;
                this.config.onChoose(this.labels[ix], ix);
            }
        }
    };

    const SelectDefs = {
        options: [],
        width: ''
        // onChange: (label, i) => {},
    };
    exports.Select = class Select {
        constructor(config) {
            this.config = Object.assign({}, SelectDefs, config);
            this.node = config.node || document.createElement('select');
            this.node.style.width = this.config.width;
            this._selected = undefined;
            this._selectedIndex = -1;
            this.options(this.config.options);
            d3.select(this.node)
                .on('change', this._onChange.bind(this));
        }
        options(opts) {
            if ( typeof(opts) === 'undefined' ) {
                return this.config.options;
            }
            this.config.options = opts;
            const options = d3.select(this.node)
                  .selectAll('option')
                  .data(opts.map(name => ({lbl: name})), qubx.m('lbl'));
            options.enter().insert('option')
                .property('value', qubx.m('lbl'))
                .text(qubx.m('lbl'));
            options.exit().remove();
            // does "it" stay selected by d3 key?
            const ix = this.node.selectedIndex,
                  lbl = (ix >= 0) ? opts[ix] : undefined,
                  changed = (lbl !== this._selected);
            this._selectedIndex = ix;
            this._selected = lbl;
            if ( changed ) {
                this.config.onChange(this._selected, this._selectedIndex);
            }
            return this;
        }
        selected(s) {
            if ( typeof(s) === 'undefined' ) {
                return this._selected;
            }
            return this.selectedIndex(this.config.options.indexOf(s));
        }
        selectedIndex(i) {
            if ( typeof(i) === 'undefined' ) {
                return this._selectedIndex;
            }
            this._selected = (i >= 0) ? this.config.options[i] : undefined;
            this.node.selectedIndex = i;
            this._selectedIndex = i;
            return this;
        }
        _onChange() {
            this._selectedIndex = this.node.selectedIndex;
            this._selected = (this._selectedIndex >= 0) ? this.config.options[this._selectedIndex] : undefined;
            this.config.onChange(this._selected, this._selectedIndex);
        }
    };

    exports.ColumnTable = class ColumnTable {
        constructor(config) {
            this.config = {
                columns: (config && config.columns) || [],
                node: config.node
            };
            if ( typeof(config.autohide) !== 'undefined' ) {
                this.config.autohide = d3.select(config.autohide.tagName ? config.autohide : config.node);
            }
            this.tbl = d3.select(config.node);
        }
        render(rowData, rowKey) {
            if ( this.tbl.select("thead").empty() ) {
                this.theadRow = this.tbl.append("thead").append("tr");
                this.tbody = this.tbl.append("tbody");
            }
            if ( this.config.autohide ) {
                this.config.autohide.style({display: rowData.length ? '' : 'none'});
            }
            this.rowData = rowData;
            this.rowData.forEach((item, i) => {
                item._ctIndex = i;
            });
            const hdrs = this.theadRow.selectAll("td")
                  .data(this.config.columns);//key-func is trouble with duplicate names//, function(d, i) { return d.label; });
            hdrs.enter()
                .append("td");
            let showHdr = false;
            hdrs.each(function(d,i) {
                if ( d.label ) {
                    showHdr = true;
                }
                d.cells = [];
                if ( d.render ) {
                    d.render.bind(this)(d,i);
                } else {
                    d3.select(this).text(d.label||"");
                }
            });
            hdrs.exit().remove();
            this.theadRow.style('visibility', showHdr ? 'visible' : 'hidden');

            const cellData = row => this.config.columns.map(col => ({row, col}));
            const rows = this.tbody.selectAll("tr")
                .data(rowData, (typeof(rowKey)!=='undefined') ? d => d[rowKey] : undefined);
            rows.exit().remove();
            rows.selectAll("td").data(cellData)
                .each(function(d,i,j) {
                    d.col.update(d3.select(this), d, j); // d.row._ctIndex);
                });
            const cells = rows.enter()
                .append("tr")
                .selectAll("td")
                .data(cellData)
            cells.enter()
                .append("td")
                .each(function(d,i,j) { // unfortunately j seems to be the row's position in the enter selection rather than global index
                    d.col.init(d3.select(this), d, d.row._ctIndex);
                    d.col.cells.push(this);
                });
            cells.exit().remove();
        }
    };
    exports.ColumnTable.checkable = function(col, checked, labelTag, onCheck) {
        col.render = function(d,i) {
            if ( typeof(d.checked) === 'undefined' ) {
                d.checked = !!checked;
            }
            const lbl = d3.select(this)
                .html("")
                .append(labelTag || "label");
            const chk = lbl.append("input")
                .property({
                    type: "checkbox",
                    checked: d.checked
                })
                .on('change', () => {
                    d.checked = chk.node().checked;
                    onCheck && onCheck(d,i);
                });
            lbl.append("span").text(col.label);
        };
        return col;
    };
    exports.ColumnTable.htmlCol = function(label, field) {
        const render = (cell, d) => cell.html(d.row[field]);
        return {
            label: label,
            init: render,
            update: render
        };
    };
    exports.ColumnTable.txtCol = function(label, field, format) {
        function render(cell, d) {
            let v = d.row[field];
            if ( format ) {
                v = format(v);
            }
            cell.text(v);
        }
        return {
            label: label,
            init: render,
            update: render
        };
    };
    exports.ColumnTable.inputCol = function(label, field, inputType, accept, format, onChange, min, max, step) {
        accept = accept || qubx.acceptStr;
        format = format || qubx.formatStr;
        return {
            label: label,
            init: (cell, d, i) => {
                const input = cell.append("input")
                    .property('type', inputType||'text')
                    .property('value', format(d.row[field]))
                    .on('input', function() {
                        try {
                            d.row[field] = accept(input.property('value'));
                            onChange && onChange(d.row, d.col, d.row[field]);
                        } catch (e) {}
                    });
                if ( typeof(min) !== 'undefined' ) {
                    input.property('min', min);
                }
                if ( typeof(max) !== 'undefined' ) {
                    input.property('max', max);
                }
                step && input.property('step', step);
            },
            update: (cell, d, i) => {
                cell.select("input")
                    .property('value', format(d.row[field]))
                    .on('input', function() {
                        try {
                            d.row[field] = accept(this.value);
                            onChange && onChange(d.row, d.col, d.row[field]);
                        } catch (e) {}
                    });
            }
        };
    };
    exports.ColumnTable.numCol = function(label, field, accept, format, onChange) {
        return exports.ColumnTable.inputCol(label, field, 'number', accept||qubx.acceptFloat, format||qubx.formatK0, onChange);
    };
    exports.ColumnTable.chkCol = function(label, field, onChange) {
        return {
            label: label,
            init: (cell, d, i) => {
                const input = cell.append("input")
                    .property('type', 'checkbox')
                    .property('checked', d.row[field])
                    .on('change', function() {
                        d.row[field] = !! input.property('checked');
                        onChange && onChange(d.row, d.col, d.row[field]);
                    });
            },
            update: (cell, d, i) => {
                cell.select("input")
                    .property('checked', !! d.row[field])
                    .on('change', function() {
                        d.row[field] = !! this.checked;
                        onChange && onChange(d.row, d.col, d.row[field]);
                    });
            }
        };
    };
    exports.ColumnTable.selectCol = function(label, field, choices, onChange) {
        return {
            label: label,
            init: (cell, d, i) => {
                const mnu = cell.append("select");
                exports.setSelectOptions(mnu.node(), choices);
                mnu.node().selectedIndex = choices.indexOf(d.row[field]);
                mnu.on('change', function() {
                    const i = mnu.node().selectedIndex;
                    d.row[field] = choices[i];
                    onChange && onChange(d.row, d.col, d.row[field]);
                });
            },
            update: (cell, d, i) => {
                cell.select("select").node().selectedIndex = choices.indexOf(d.row[field]);
            }
        };
    };
    exports.ColumnTable.btnCol = function(label, caption, onClick, title) {
        return {
            label: label,
            init: (cell, d, i) => {
                cell.append("button")
                    .html(caption)
                    .property('title', title||'')
                    .on('click', () => onClick(d.row, d.row._ctIndex));
            },
            update: (cell, d, i) => {
                cell.select("button")
                    .on('click', () => onClick(d.row, d.row._ctIndex));
            }
        };
    };
    exports.ColumnTable.fileCol = function(label, onchange) {
        return {
            label: label,
            init: (cell, d, i) => {
                const input = cell.append("input");
                input.attr({type: 'file'})
                    .on('change', () => onchange && onchange(input.node(), d, i));
            },
            update: (cell, d, i) => {
                const input = cell.select("input");
                input.on('change', () => onchange && onchange(input.node(), d, i));
            }
        };
    };

    exports.makeTableViewColumn = function makeTableViewColumn(table, field) {
        return exports.ColumnTable.txtCol(field, field, table.fields[table.fieldIndex[field]].format);
    };
    exports.makeTableEditColumn = function makeTableEditColumn(table, field, onChange) {
        const info = table.fields[table.fieldIndex[field]];
        if ( info.accept === qubx.acceptNone ) {
            return exports.makeTableViewColumn(table, field);
        } else if ( info.choices && info.choices.length ) {
            return exports.ColumnTable.selectCol(field, field, info.choices, onChange);
        } else {
            return exports.ColumnTable.inputCol(field, field, isFinite(info.defval) ? 'number' : 'text', info.accept, info.format, info.onChange);
        }
    };
    exports.TableView = class TableView {
        constructor(config) {
            config.columns = config.columns || config.table.fields.map(qubx.m('name'));
            config.columns.forEach((c, i) => {
                if ( typeof(c) === 'string' ) {
                    if ( c === 'delete' ) {
                        config.columns[i] = exports.ColumnTable.btnCol('delete', 'X', (d,di) => this.table.remove(di));
                    } else {
                        config.columns[i] = exports.makeTableEditColumn(config.table, c, this._onChange.bind(this));
                    }
                }
            });
            this.view = new exports.ColumnTable(config);
            this.render(config.table);
        }
        render(table) {
            if ( this.table && table && (this.table !== table) ) {
                this.table.ev.inserted.unsub(this.tableOnIns);
                this.table.ev.removed.unsub(this.tableOnRem);
                this.table.ev.set.unsub(this.tableOnSet);
                this.table = this.tableOnIns = this.tableOnRem = this.tableOnSet = null;
            }
            if ( table && ! this.table ) {
                this.table = table;
                this.tableOnIns = this.table.ev.inserted.sub(this.reRender.bind(this));
                this.tableOnRem = this.table.ev.removed.sub(this.reRender.bind(this));
                this.tableOnSet = this.table.ev.set.sub(this.reRender.bind(this));
            }
            if ( ! this._userEdit ) {
                this.view.render(this.table.entries, 'Index');
            }
        }
        reRender() {
            this.render();
        }
        _onChange(row, col, val) {
            this._userEdit = true;
            this.table.set(row.Index, col.label, val);
            this._userEdit = false;
        }
    };

    exports.Opener = class Opener {
        constructor(config) {
            this.config = config || {}; // {noComplaints: false, onprogress: function(x in 0..1)}
            this.res = null;
            this.rej = null;
            this.defer = this.config.defer;
            this.dlg = new exports.Modal({
                title: "Choose a file",
                template: ".qubx_opener"
            });
            this.dlg.dom.then(dom => {
                const dotdot = exports.querydotdot(dom.content);
                this.node = dom.content;
                this.txtURL = dotdot("#qubx_opener_url");
                this.input = dotdot.listen("#qubx_opener_file", "change", this._onChange.bind(this), input => {
                    if ( 'multiple' in this.config ) {
                        input.multiple = !!this.config.multiple;
                    }
                });
            }).catch(qubx.reportErr);
        }
        run(acceptType, instructionNode, defer) {
            if ( typeof(defer) === 'undefined' ) {
                defer = !! this.config.defer;
            }
            this.defer = !! defer;
            return new Promise((resolve, reject) => {
                this.res = resolve;
                this.rej = reject;
                this.input.setAttribute("accept", acceptType || this.config.acceptType || "*/*");
                setTimeout(() => this.txtURL.focus(), 10);
                instructionNode && this.node.appendChild(instructionNode);
                this.dlg.run().then(() => {
                    instructionNode && this.node.removeChild(instructionNode);
                    const res = this.res,
                          rej = this.rej;
                    if ( res ) {
                        this.res = null;
                        this.rej = null;
                        let url = this.txtURL.value;
                        if ( url ) {
                            if ( url.indexOf('://') < 0 ) {
                                url = 'https://'+url;
                            }
                            const get = onprogress => {
                                return qubx.get(url, 'blob', onprogress || this.config.onprogress).then(blob => {
                                    blob.name = url;
                                    return blob;
                                }).catch(err => {
                                    let complaint = err.message || '';
                                    if ( complaint.indexOf('Network Error') >= 0 ) {
                                        complaint = 'The file could not be read.  If the URL is not in error, perhaps the host is not configured to allow cross-origin requests.';
                                    }
                                    this.config.noComplaints || setTimeout(() => alert(complaint), 10)
                                    return Promise.reject(err);
                                });
                            };
                            if ( defer ) {
                                res([{url, get}]);
                            } else {
                                res(get().then(blob => [blob]));
                            }
                        } else {
                            rej('Canceled');
                        }
                    }
                });
                // hopefully the fake click will work, and the dialog will close instantly
                //this.input.click();
            });
        }
        reset() {
            this.input.value = "";
            this.txtURL.value = "";
        }
        _onChange() {
            const res = this.res;
            if ( res ) {
                this.res = null;
                this.rej = null;
                this.dlg.close();
                res(this.defer ? Array.prototype.map.call(this.input.files, file => ({
                    name: file.name,
                    get: () => Promise.resolve(file)
                })) : this.input.files);
            }
        }
    };

    const SaverDefaults = {
        mimeType: 'text/plain'
    };
    exports.Saver = class Saver {
        constructor(config) {
            this.config = Object.assign({}, SaverDefaults, config);
            this.link = document.createElement("a");
            this.link.setAttribute("style", "visibility: hidden; position: absolute; bottom: 0");
            this.link.innerHTML = 'download';
            document.querySelector("body").appendChild(this.link);
            this.promptDlg = new exports.Modal({
                title: 'Save as...',
                template: '.qubx_saver_prompt',
                okLabel: 'Save'
            });
            this.promptDlg.dom.then(dom => {
                const dotdot = exports.querydotdot(dom.content);
                this.txtName = dotdot('.qubx_saver_name');
                this.mnuExt = dotdot('.qubx_saver_ext');
                this.divInfo = dotdot('.qubx_saver_info');
            });
        }
        release() {
            if ( this.lastObject ) {
                URL.revokeObjectURL(this.lastObject);
                this.lastObject = null;
            }
        }
        prompt(opts) {
            opts = opts || {};
            const name = opts.name || 'Untitled',
                  iDot = name.lastIndexOf('.');
            let ext = opts.ext,
                basename = name;
            if ( iDot >= 0 ) {
                basename = name.substring(0, iDot);
                ext = ext || name.substring(iDot);
            }
            if ( ! ext ) {
                ext = opts.exts && opts.exts.length && opts.exts[0] || '.txt';
            }
            let exts = opts.exts || [];
            if ( ! exts.length ) {
                exts = [ext];
            }
            this.txtName.value = basename;
            exports.setSelectOptions(this.mnuExt, exts);
            this.mnuExt.selectedIndex = exts.indexOf(ext);
            this.divInfo.innerHTML = opts.info || '';
            return this.promptDlg.run().then(() => {
                const name = this.txtName.value;
                if ( ! name ) {
                    return Promise.reject('Canceled');
                }
                const ext = this.mnuExt.value || '';
                return {
                    name,
                    ext,
                    fullname: name+ext
                };
            });
        }
        save(name, blob) {
            this.release();
            this.lastObject = URL.createObjectURL(blob);
            this.link.download = name;
            this.link.href = this.lastObject;
            this.link.click();
        }
        saveText(name, s, typ) {
            const t = typ || this.config.mimeType;
            this.save(name, new Blob([s], {type: t}));
        }
        saveJSON(name, obj) {
            this.saveText(name, JSON.stringify(obj), 'application/json');
        }
    };
    
    const ModalConfig = {
        title: "",
        template: '#qubx_modal_blank',
        canCancel: true,
        cancelLabel: 'Cancel',
        canOK: true,
        okLabel: 'OK',
        focus: '.qubx_modal_ok', // selector within dom.content or dom.buttons
        buttons: [] // in addition to OK, Cancel if enabled
        //             list of [label, f()]
        // template, onCancel, onOK, (onOK return True to keep it open)
        //
    };
    exports.Modal = class Modal {
        constructor(config) {
            this.config = Object.assign({}, ModalConfig, config);
            this.running = false;
            this.dom = exports.loadTemplates([".qubx_modal", this.config.template]).then(items => {
                const t = items[0],
                      content = items[1],
                      dotdot = exports.querydotdot(t),
                      dom = {
                          root: t,
                          title: dotdot(".qubx_modal_title"),
                          buttons: dotdot('.qubx_modal_buttons'),
                          content
                      };
                dotdot('.qubx_modal_content').appendChild(content);
                const btns = dom.buttons,
                      cancel = btns.querySelector('.qubx_modal_cancel');
                cancel.textContent = this.config.cancelLabel;
                if ( this.config.canCancel ) {
                    cancel.addEventListener('click', this._onClickCancel.bind(this));
                } else {
                    cancel.style.display = 'none';
                }
                const ok = btns.querySelector('.qubx_modal_ok');
                ok.textContent = this.config.okLabel;
                if ( this.config.canOK ) {
                    ok.addEventListener('click', this._onClickOK.bind(this));
                } else {
                    ok.style.display = 'none';
                }
                this.config.buttons.forEach(spec => {
                    const lbl = spec[0],
                          onclick = spec[1],
                          btn = document.createElement('button');
                    btn.className = 'qubx_modal_btn';
                    btn.textContent = lbl;
                    btns.appendChild(btn);
                    onclick && btn.addEventListener('click', onclick);
                });
                
                t.style.display = 'none';
                document.body.appendChild(t);
                if ( this.config.title ) {
                    dom.title.textContent = this.config.title;
                } else {
                    dom.title.style.display = 'none';
                }
                return dom;
            }).catch(qubx.reportErr);
        }
        close(err) {
            const sess = this.sess;
            this.sess = null;
            if ( err ) {
                this.config.onCancel && this.config.onCancel();
                this.dom.then(dom => { dom.root.style.display = 'none'; });
                this.running = false;
                sess.reject(err);
            } else {
                if ( (! this.config.onOK) || (! this.config.onOK()) ) {
                    this.dom.then(dom => { dom.root.style.display = 'none'; });
                    this.running = false;
                    sess.resolve(true);
                } else {
                    this.sess = sess;
                }
            }
        }
        _onClickCancel(event) {
            this.close('Canceled');
        }
        _onClickOK(event) {
            this.close();
        }
        run() {
            return new Promise((resolve, reject) => {
                this.sess = {resolve, reject};
                this.dom.then(dom => {
                    dom.root.style.display = '';
                    this.running = true;
                    (dom.content.querySelector(this.config.focus) ||
                     dom.buttons.querySelector(this.config.focus)).focus();
                });
            });
        }
        destroy() {
            this.dom.then(dom => {
                document.body.removeChild(dom.root);
            }).catch(qubx.reportErr);
        }
    };

    const ModalProgressConfig = {
        canCancel: true,
        status: "",
        progress: 0.0
    };
    exports.ModalProgress = class ModalProgress {
        constructor(config) {
            this.config = Object.assign({}, ModalProgressConfig, config);
            this.dom = exports.loadTemplate(".qubx_modal_progress").then(t => {
                const dotdot = exports.querydotdot(t),
                      dom = {
                          root: t,
                          progress: dotdot('.qubx_modal_prog_progress'),
                          status: dotdot('.qubx_modal_prog_status')
                      };
                dotdot('.qubx_modal_cancel').addEventListener('click', () => {
                    if ( ! this.canceled ) {
                        this.canceled = true;
                        this.config.oncancel && this.config.oncancel();
                    }
                });
                if ( ! this.config.canCancel ) {
                    dotdot('.qubx_modal_cancel').style.display = 'none';
                }
                t.style.display = 'none';
                document.body.appendChild(t);
                return dom;
            }).catch(qubx.reportErr);
        }
        destroy() {
            this.dom.then(dom => document.body.removeChild(dom.root));
        }
        show(status, progress) {
            this.status(status||"")
                .progress(progress||0.0);
            this.canceled = false;
            this.dom.then(dom => {
                dom.root.style.display = '';
            });
        }
        hide() {
            this.dom.then(dom => {
                dom.root.style.display = 'none';
            });
        }
        status(s) {
            if ( typeof(s) === 'undefined' ) {
                return this.config.status;
            } else {
                this.config.status = s;
                this.dom.then(dom => {
                    dom.status.textContent = s;
                });
            }
            return this;
        }
        progress(p) {
            if ( typeof(p) === 'undefined' ) {
                return this.config.progress;
            } else {
                this.config.progress = Math.max(0.0, Math.min(1.0, p));
                this.dom.then(dom => {
                    dom.progress.style.width = `${100*p}%`;
                });
            }
            return this;
        }
    };
    
    var confirm, confirmReady;
    exports.confirm = function(cfg) {
        cfg = cfg || {};
        if ( ! confirm ) {
            confirm = new exports.Modal({
                template: '#qubx_confirm'
            });
            confirmReady = confirm.dom.then(dom => {
                confirm.hdr = dom.content.querySelector('#qubx_confirm_head');
                confirm.q = dom.content.querySelector('#qubx_confirm_q');
            });
        }
        return confirmReady.then(() => {
            confirm.hdr.innerHTML = cfg.hdr || '';
            confirm.q.innerHTML = cfg.q || '';
            return confirm.run();
        });
    };
    
    exports.SimplePicker = class SimplePicker {
        constructor(title, okname) {
            this.dlg = new exports.Modal({
                title: title || "Browse",
                template: '.qubx_simplepick',
                onOK: this._onClickOK.bind(this),
                onCancel: this._onClickCancel.bind(this)
            });
            this.dlg.dom.then(dom => {
                const dotdot = exports.querydotdot(dom.root);
                if ( okname ) {
                    dotdot('.qubx_modal_ok').textContent = okname;
                }
                this.desc = dotdot('.qubx_simpledesc');
                this.list = dotdot.listen('.qubx_simplelist', 'change', this._onItemChange.bind(this));
            }).catch(qubx.reportErr);
        }
        _onItemChange() {
            this.index = this.list.node().selectedIndex;
            this.item = this.items[this.index];
            this.desc.textContent = this.item[this.descKey];
        }
        bindItems(items, nameKey, descKey) {
            this.items = items;
            this.nameKey = nameKey || "name";
            this.descKey = descKey || "desc";
            
            this.list.property('size', Math.max(3, items.length));
            const opts = this.list.selectAll("option")
                .data(items);
            const getName = qubx.m(this.nameKey);
            opts.text(getName);
            opts.enter()
                .append("option")
                .text(getName);
            opts.exit()
                .remove();
            if ( items.length ) {
                this.item = items[0];
                this.index = 0;
                this.desc.textContent = this.item[this.descKey];
            }
        }
        unbind() {
            this.items = null;
        }
        run(items, nameKey, descKey) {
            this.bindItems(items, nameKey, descKey);
            this.dlg.run();
            return new Promise((resolve, reject) => {
                this.doneRun = resolve;
                this.cancelRun = reject;
            });
        }
        _onClickOK() {
            this.doneRun({
                item: this.item,
                index: this.index
            });
            this.doneRun = this.cancelRun = null;
            this.unbind();
        }
        _onClickCancel() {
            this.cancelRun(Error("Canceled"));
            this.doneRun = this.cancelRun = null;
            this.unbind();
        }
    };

    var showCopyableDlg;
    setTimeout(() => {
        // want this on top of everything, so wait to create
        showCopyableDlg = new exports.Modal({
            title: 'QUB - Output',
            template: '.qubx_showCopyable',
            canCancel: false,
            okLabel: 'Close'
        });
        showCopyableDlg.dom.then(dom => {
            showCopyableDlg.lbl = dom.content.querySelector(".qubx_showCopyableLbl");
            showCopyableDlg.txt = dom.content.querySelector(".qubx_showCopyableTxt");
        });
    }, 5000);
    exports.ShowCopyable = function ShowCopyable(label, txt) {
        showCopyableDlg.lbl.innerHTML = label;
        showCopyableDlg.txt.value = txt;
        setTimeout(() => {
            showCopyableDlg.txt.focus();
            showCopyableDlg.txt.select();
        }, 250);
        return showCopyableDlg.run();
    };

    const SelectListDefs = {
    };
    // events: onchange: function(selectedIndex)
    exports.SelectList = class SelectList {
        constructor(config, events) {
            this.config = Object.assign({}, SelectListDefs, config);
            this.events = events || {};
            this._selectedIndex = -1;
            this._data = [];
        }
        data(data, asHTML) {
            if ( typeof(data) === 'undefined' ) {
                return this._data;
            }
            this._data = data;
            if ( this._data.length && (this._selectedIndex > (this._data.length-1)) ) {
                this._selectedIndex = -1;
            }
            this.config.node.classList.add('qubx_selectlist');
            const dd = d3.select(this.config.node)
                .selectAll("div")
                .data(data);
            dd.enter().insert("div")
                .on('click', (d, i) => {
                    this.selectedIndex(i);
                    this.events.onchange && this.events.onchange(i);
                });
            dd.exit().remove();
            (asHTML ? dd.html : dd.text).bind(dd)(qubx.i);
            dd.classed("qubx_selectlist_item", true);
            dd.classed("qubx_selectlist_sel", (d, i) => (i === this._selectedIndex));
            const divs = [];
            dd.each(function(d, i) { divs.push(this); });
            this.divs = divs;
        }
        selectedIndex(selIx) {
            if ( typeof(selIx) === 'undefined' ) {
                return this._selectedIndex;
            }
            if ( selIx >= this._data.length ) {
                selIx = -1;
            }
            if ( (this._selectedIndex >= 0) && (this._selectedIndex < this._data.length) ) {
                this.divs[this._selectedIndex].classList.remove("qubx_selectlist_sel");
            }
            this._selectedIndex = selIx;
            if ( selIx >= 0 ) {
                this.divs[selIx].classList.add("qubx_selectlist_sel");
            }
        }
    };
        
    exports.resultDlg = new exports.Modal({
        title: 'Browse Results',
        template: '.qubx_resultDocs',
        canCancel: false,
        okLabel: 'Close'
    });
    exports.resultDlg.results = localStorage.getItem('qubx_resultDocs_results');
    exports.resultDlg.results = exports.resultDlg.results ? JSON.parse(exports.resultDlg.results) : [];
    exports.resultDlg.results.forEach(entry => {
        entry.date = new Date(entry.date);
    });
    exports.resultDlg.savePrefs = function() {
        const results = exports.resultDlg.results;
        if ( results.length && ! results[results.length-1].pinned ) {
            results.pop();
        }
        localStorage.setItem('qubx_resultDocs_results', JSON.stringify(results));
    };
    exports.resultDlg.add = function(srcdoc, from, date) {
        const results = exports.resultDlg.results;
        if ( results.length && ! results[results.length-1].pinned ) {
            results.pop();
        }
        results.push({
            srcdoc,
            from: from || "",
            date: date || new Date(),
            comment: "",
            pinned: false
        });
    };
    exports.resultDlg.dom.then(dom => {
        const dotdot = exports.querydotdot(dom.content),
              ui = {
                  date: dotdot('.qubx_resultDocs_date'),
                  from: dotdot('.qubx_resultDocs_from'),
                  comment: dotdot('.qubx_resultDocs_comment'),
                  pin: dotdot('.qubx_resultDocs_pin'),
                  download: dotdot('.qubx_resultDocs_download'),
                  pinlist: new exports.SelectList({node: dotdot('.qubx_resultDocs_pinlist')},
                                                  {onchange: selectResult}),
                  content: dotdot('.qubx_resultDocs_content')
              };
        let selection = {};
        exports.resultDlg.resUI = ui;
        function showResult(entry) {
            entry = entry || {};
            ui.date.textContent = ""+entry.date;
            ui.from.textContent = entry.from;
            ui.comment.value = entry.comment;
            ui.pin.checked = entry.pinned;
            ui.download.href = 'data:text/html;charset=utf-8,' + encodeURI(entry.srcdoc);
            ui.download.download = `QUB Results ${ui.from ? ("("+ui.from+")") : ""}.html`;
            ui.content.srcdoc = entry.srcdoc;
            selection = entry;
        }
        function saveComment() {
            selection.comment = ui.comment.value;
        }
        function listPinned() {
            const names = exports.resultDlg.results.map(entry => {
                return `${entry.comment ? (entry.comment+" ") : ""}${qubx.formatDate(entry.date)} (${entry.from$})`;
            });
            names.reverse();
            ui.pinlist.data(names);
        }
        function getResult(i) {
            return (i >= 0) ? exports.resultDlg.results[exports.resultDlg.results.length-(i+1)] : {};
        }
        function selectResult(i) {
            saveComment();
            showResult(getResult(i));
        };
        ui.pin.addEventListener('change', () => {
            selection.pinned = ui.pin.checked;
        });

        const _run = exports.resultDlg.run.bind(exports.resultDlg);
        exports.resultDlg.run = () => {
            listPinned();
            ui.pinlist.selectedIndex(0);
            selectResult(0);
            return _run().then(result => {
                saveComment();
                ui.content.srcdoc = "";
                ui.download.href = "";
                let i = exports.resultDlg.results.length-2;
                for (; i>=0; --i) {
                    if ( ! exports.resultDlg.results[i].pinned ) {
                        exports.resultDlg.results.splice(i, 1);
                    }
                }
                return result;
            });
        };
    }).catch(qubx.reportErr);

    exports.Spinner = class Spinner {
        constructor(parent) {
            this.node = document.createElement('div');
            this.node.className = 'spinBox';
            this.node.style.display = 'none';
            this.spinner = document.createElement('div');
            this.spinner.className = 'spinner';
            this.node.appendChild(this.spinner);
            parent && parent.appendChild(this.node);
        }
        show() {
            this.node.style.display = '';
        }
        hide() {
            this.node.style.display = 'none';
        }
    };

    exports.OutlineRender = {
        Text: (field, container, containerClass) => {
            field = field || 'label';
            if ( container ) {
                return function(d,i) {
                    const c = d3.select(this)
                          .append(container)
                          .text(d[field]);
                    containerClass && c.classed(containerClass, true);
                    return c;
                };
            }
            return function(d,i) {
                d3.select(this).text(d[field]);
            };
        },
        Check: (labelField, checkedField, boxOnly) => {
            labelField = labelField || 'label';
            checkedField = checkedField || 'checked';
            return function(d,i) {
                const lbl = d3.select(this).append(boxOnly ? 'span' : 'label'),
                      chk = lbl.append('input').property({
                    type: 'checkbox',
                    checked: !! d[checkedField]
                }).on('change', () => {
                    d[checkedField] = chk.property('checked');
                });
                lbl.append('span').text(d[labelField]);
            };
        }
    };
    
    const OutlineDefs = {
        renderHead: exports.OutlineRender.Text(),
        renderLeaf: exports.OutlineRender.Text(),
        selectHead: false,
        selectLeaf: false,
        onSelect: item => {}
    };
    exports.Outline = class Outline {
        constructor(config) {
            this.config = Object.assign({}, OutlineDefs, config);
            this.node = config.node || document.createElement('div');
            this.node.classList.add('qubx_outline');
            this.selectedItem = undefined;
            this.selectedNode = undefined;
        }
        // tree = node = {expand: false, children: [], label: ''}
        // absence of children: is leaf; tree root shouldn't be a leaf
        data(tree) {
            const root = d3.select(this.node).append('div');
            this._renderNode(root, tree);
            this.selectedItem = undefined;
            this.selectedNode = undefined;
        }
        _renderNode(div, data) {
            const self = this; // d3 abuses this, can't bind()
            div.html('');
            const head = div.append('div');
            this.config.renderHead.bind(head.node())(data, 0);
            head.classed('qubx_outline_head', true);
            let canSelect = data.canSelect;
            if ( typeof(canSelect) === 'undefined' ) {
                canSelect = self.config.selectHead;
            }
            head.on('click', function() {
                const selChange = canSelect && (self.selectedItem !== data);
                if ( (! data.expand) || (! selChange) ) {
                    data.expand = ! data.expand;
                    self._renderChildren(div, data);
                }
                if ( selChange ) {
                    self.selectedNode && self.selectedNode.classList.remove('qubx_outline_sel');
                    self.selectedItem = data;
                    self.selectedNode = this;
                    self.selectedNode && self.selectedNode.classList.add('qubx_outline_sel');
                    self.config.onSelect(data);
                }
                d3.event.stopPropagation();
            });
            this._renderChildren(div, data);
        }
        _renderChildren(div, data) {
            const self = this, // d3 abuses this, can't bind()
                  head = div.select('.qubx_outline_head');
            head.classed('qubx_outline_show', ! data.expand);
            head.classed('qubx_outline_hide', data.expand);
            let canSelect = data.canSelect;
            if ( typeof(canSelect) === 'undefined' ) {
                canSelect = self.config.selectLeaf;
            }
            if ( data.expand ) {
                const children = div.append('div');
                children.classed('qubx_outline_children', true);
                const child = children.selectAll('div')
                    .data(data.children);
                child.enter().insert('div');
                child.exit().remove();
                if ( canSelect ) {
                    child.on('click', function(d,i) {
                        if ( ! d.children ) {
                            if ( self.selectedItem !== d ) {
                                self.selectedNode && self.selectedNode.classList.remove('qubx_outline_sel');
                                self.selectedItem = d;
                                self.selectedNode = this;
                                self.selectedNode && self.selectedNode.classList.add('qubx_outline_sel');
                                self.config.onSelect(d);
                            }
                            d3.event.stopPropagation();
                        }
                    });
                }
                child.each(function(d,i) {
                    const childNode = d3.select(this);
                    if ( d.children ) {
                        self._renderNode(childNode, d);
                    } else {
                        childNode.classed('qubx_outline_leaf', true);
                        self.config.renderLeaf.bind(this)(d,i);
                    }
                });
            } else {
                div.selectAll('.qubx_outline_children').remove();
            }
        }
    };

    function aspectDimensions(w, h, n, aspect) {
        n = n || 1;
        aspect = aspect || 1.618;
        let r = 1,
            c = 1,
            wOne = w,
            hOne = h;
        while ( (r*c) < n ) {
            if ( aspect < (wOne/hOne) ) {
                ++c;
                wOne = Math.max(1, w/c);
            } else {
                ++r;
                hOne = Math.max(1, h/r)
            }
        }
        while ( (r > 1) && ((r-1)*c >= n) ) {
            --r;
        }
        return {
            rows: r,
            columns: c
        };
    }
    
    const AspectDefaults = {
        aspectRatio: 1.618
    };
    exports.AspectGrid = class AspectGrid {
        constructor(config) {
            this.config = Object.assign({}, AspectDefaults, config);
            this.node = config && config.node || document.createElement('div');
            this.node.classList.add('qubx_aspect_grid');
            this.resizeHandle = exports.onResize(this.node, info => this.__layout());
            this.layoutSoon = qubx.batchDelayed(this.__layout.bind(this), 50);
            this.items = [];
        }
        destroy() {
            this.node.innerHTML = '';
            exports.clearResize(this.resizeHandle);
        }
        insert(child, index, onLayout) {
            child = child || document.createElement('div');
            const item = {
                box: document.createElement('div'),
                child,
                onLayout
            };
            item.box.appendChild(child);
            item.box.classList.add('qubx_aspect_box');
            child.classList.add('qubx_aspect_child');
            if ( typeof(index) === 'undefined' ) {
                index = this.items.length;
            }
            if ( index < this.items.length ) {
                this.node.insertBefore(item.box, this.items[index].box);
            } else {
                this.node.appendChild(item.box);
            }
            this.items.splice(index, 0, item);
            this.layoutSoon();
            return child;
        }
        remove(childOrIndex) {
            const child = childOrIndex;
            let index = childOrIndex | 0;
            if ( index !== childOrIndex ) {
                index = -1;
                const n = this.items.length;
                let i = 0;
                for (; i<n; ++i) {
                    if ( this.items[i].child === child ) {
                        index = i;
                        break;
                    }
                }
            }
            if ( 0 <= index && index < this.items.length ) {
                this.items[index].child.classList.remove('qubx_aspect_child');
                this.node.removeChild(this.items[index].box);
                this.items.splice(index, 1);
                this.layoutSoon();
            }
        }
        clear() {
            this.items.forEach(item => item.child.classList.remove('qubx_aspect_child'));
            this.node.innerHTML = '';
            this.items = [];
            // this.layoutSoon();
        }
        __layout() {
            const dim = aspectDimensions(this.node.clientWidth, this.node.clientHeight, this.items.length, this.config.aspectRatio),
                  itemWidth = 0.95 * this.node.clientWidth / dim.columns,
                  itemHeight = 0.95 * this.node.clientHeight / dim.rows;
            this.items.forEach(item => {
                item.box.style.width = itemWidth + 'px';
                item.box.style.height = itemHeight + 'px';
            });
            this.items.forEach(item => item.onLayout && item.onLayout());
        }
    };
    
    exports.AspectGrid3 = class AspectGrid3 {
        constructor(config) {
            this.config = Object.assign({}, AspectDefaults, config);
            this.node = config && config.node || document.createElement('div');
            this.node.classList.add('qubx_aspect_grid');
            this.resizeHandle = exports.onResize(this.node, this.__layout.bind(this));
            this.__data = [];
        }
        destroy() {
            this.node.innerHTML = '';
            exports.clearResize(this.resizeHandle);
        }
        data(arr, keyfn, setup, onLayout, teardown) {
            this.__data = arr;
            this.onLayout = onLayout;
            const bound = d3.select(this.node)
                  .selectAll('.qubx_aspect_box')
                  .data(arr, keyfn);
            bound.enter().insert('div').each(function(d,i) {
                d.aspectBox = d3.select(this);
                d.aspectBox.classed('qubx_aspect_box', true);
                d.aspectChild = d.aspectBox.append('div');
                d.aspectChild.classed('qubx_aspect_child', true);
                d.aspectChild.datum(() => null); // prevent d3 data binding propagation to children
                setup && setup.bind(d.aspectChild.node())(d, i);
            });
            bound.exit()
                .each((d,i) => {
                    teardown && teardown.bind(d.aspectChild.node())(d, i);
                    d.aspectBox = null;
                    d.aspectChild = null;
                }).remove();
            this.__layout();
            return bound; // .selectAll('.qubx_aspect_child'); /// return bound;
        }
        __layout() {
            const dim = aspectDimensions(this.node.clientWidth, this.node.clientHeight, this.__data.length, this.config.aspectRatio),
                  itemWidth = 0.95 * this.node.clientWidth / dim.columns,
                  itemHeight = 0.95 * this.node.clientHeight / dim.rows;
            this.__data.forEach(item => {
                item.aspectBox.style({
                    width: itemWidth + 'px',
                    height: itemHeight + 'px'
                });
            });
            this.__data.forEach((item, i) => this.onLayout && this.onLayout(item, i));
        }
    };
    
    const TwoPaneDefs = {
        vertical: false,
        skews: [.3, .5, .7],
        skewType: 1,
        snap: 0.05,
        // node, first, second
        // firstLabel, secondLabel if you want descriptive "Show thisthat..." buttons when minimized
    };
    exports.TwoPane = class TwoPane {
        constructor(config) {
            config = config || {};
            this.config = Object.assign({}, TwoPaneDefs, config);
            this.node = config.node || document.createElement('div');
            this.node.classList.add('qubx_twopane');
            this.node.classList.add(this.config.vertical ? 'qubx_twopane_v' : 'qubx_twopane_h');
            this.first = config.first || document.createElement('div');
            this.first.classList.add('qubx_twopane_item');
            this.first.classList.add('qubx_twopane_first');
            if ( this.config.firstLabel ) {
                this.showFirstBox = document.createElement('div');
                this.showFirstBox.className = 'qubx_twopane_showbox qubx_twopane_hid';
                this.showFirst = document.createElement('button');
                this.showFirst.textContent = this.config.firstLabel;
                this.showFirst.addEventListener('click', this.moreA.bind(this));
                this.showFirstBox.appendChild(this.showFirst);
            } else {
                this.showFirstBox = document.createElement('div');
                this.showFirstBox.className = 'qubx_twopane_spacer';
            }
            this.second = config.second || document.createElement('div');
            this.second.classList.add('qubx_twopane_item');
            this.second.classList.add('qubx_twopane_second');
            if ( this.config.secondLabel ) {
                this.showSecondBox = document.createElement('div');
                this.showSecondBox.className = 'qubx_twopane_showbox qubx_twopane_hid';
                this.showSecond = document.createElement('button');
                this.showSecond.textContent = this.config.secondLabel;
                this.showSecond.addEventListener('click', this.moreB.bind(this));
                this.showSecondBox.appendChild(this.showSecond);
            } else {
                this.showSecondBox = document.createElement('div');
                this.showSecondBox.className = 'qubx_twopane_spacer';
            }
            this.sep = document.createElement('div');
            this.sep.className = 'qubx_twopane_sep qubx_noselect';
            this.sepHandle = document.createElement('div');
            this.sepHandle.className = 'qubx_twoPane_sepHandle';
            this.sep.appendChild(this.sepHandle);
            this.btnMoreB = document.createElement('button');
            this.btnMoreB.className = 'qubx_twopane_moreA';
            this.btnMoreB.addEventListener('mousedown', event => {
                event.preventDefault();
                event.stopPropagation();
            });
            this.btnMoreB.addEventListener('mouseup', event => this.moreB());
            this.btnMoreB.textContent = '\u279C';
            this.btnMoreB.style.transform = `rotate(${this.config.vertical ? -90 : 180}deg)`;
            this.sepHandle.appendChild(this.btnMoreB);
            this.btnMoreA = document.createElement('button');
            this.btnMoreA.className = 'qubx_twopane_moreA';
            this.btnMoreA.addEventListener('mousedown', event => {
                event.preventDefault();
                event.stopPropagation();
            });
            this.btnMoreA.addEventListener('mouseup', event => this.moreA());
            this.btnMoreA.textContent = '\u279C';
            this.btnMoreA.style.transform = `rotate(${this.config.vertical ? 90 : 0}deg)`;
            if ( this.config.vertical ) {
                this.sepHandle.insertBefore(this.btnMoreA, this.btnMoreB);
            } else {
                this.sepHandle.appendChild(this.btnMoreA);
            }
            this.sep.addEventListener('mousedown', this.__onMouseDown.bind(this));
            this.node.appendChild(this.showFirstBox);
            this.node.appendChild(this.first);
            this.node.appendChild(this.sep);
            this.node.appendChild(this.second);
            this.node.appendChild(this.showSecondBox);
            setTimeout(this.onResize.bind(this), 10);
            this.resizeHandle = exports.onResize(this.node, info => this.onResize());
        }
        destroy() {
            exports.clearResize(this.resizeHandle);
        }
        addControl(node) {
            node.classList.add('qubx_twopane_sepControl');
            this.sepHandle.appendChild(node);
            return this;
        }
        onResize(setOnly, pos) {
            const limit = this.config.vertical ? this.node.clientHeight : this.node.clientWidth;
            if ( ! setOnly ) {
                this.skewsPix = this.config.skews.map(skew => (skew * limit) | 0);
            }
            if ( typeof(pos) !== 'undefined' ) {
                pos = pos + 'px';
            } else {
                if ( this.config.skewType < 0 ) {
                    pos = ((this.config.snap*limit)|0) + 'px';
                } else if ( this.config.skewType < this.skewsPix.length ) {
                    pos = this.skewsPix[this.config.skewType] + 'px';
                } else {
                    pos = (((limit - 2*(this.config.vertical ? this.showSecondBox.clientHeight : this.showSecondBox.clientWidth))|0)) + 'px';
                }
                d3.select(this.showFirstBox).classed('qubx_twopane_hid', this.config.skewType >= 0);
                d3.select(this.showSecondBox).classed('qubx_twopane_hid', this.config.skewType !== this.skewsPix.length);
                if ( this.config.skewType < 0 ) {
                    this.first.classList.add('qubx_twopane_hid');
                } else {
                    this.first.classList.remove('qubx_twopane_hid');
                }
                if ( this.config.skewType === this.skewsPix.length ) {
                    this.second.classList.add('qubx_twopane_hid');
                } else {
                    this.second.classList.remove('qubx_twopane_hid');
                }
            }
            if ( this.config.vertical ) {
                if ( this.config.skewType >= 0 ) {
                    this.first.style.height = pos;
                }
            } else {
                if ( this.config.skewType >= 0 ) {
                    this.first.style.width = pos;
                }
            }
            if ( navigator.userAgent.indexOf('CPU iPhone OS') > 0 ) {
                // Apple over-optimized: they don't re-flow the container when an item's size changes.
                // We will force a re-flow:
                this.node.style.display='none';
                this.__whatever = this.node.offsetHeight; // no need to store this anywhere, the reference is enough (store it for the minifier's sake)
                this.node.style.display='';
            }
            this.btnMoreB.style.display = (this.config.skewType < 0) ? 'none' : '';
            this.btnMoreA.style.display = (this.config.skewType === this.skewsPix.length) ? 'none' : '';
        }
        moreA() {
            if ( this.config.skewType < this.skewsPix.length ) {
                ++this.config.skewType;
                this.onResize(true);
            }
        }
        moreB() {
            if ( this.config.skewType >= 0 ) {
                --this.config.skewType;
                this.onResize(true);
            }
        }
        /// todo: hammertime
        __onMouseDown(event) {
            const limit = this.config.vertical ? this.node.clientHeight : this.node.clientWidth;
            let p0 = this.config.vertical ? event.clientY : event.clientX,
                skewPix = 0;
            if ( 0 <= this.config.skewType ) {
                if ( this.config.skewType < this.skewsPix.length ) {
                    skewPix = this.skewsPix[this.config.skewType];
                } else {
                    skewPix = limit;
                }
            }
            const onmousemove = evt => {
                const p = (this.config.vertical ? evt.clientY : evt.clientX)
                skewPix = Math.max(0, Math.min(limit, skewPix + (p - p0)));
                p0 = p;
                this.onResize(true, skewPix);
            };
            const onmouseup = evt => {
                onmousemove(evt);
                const Nskew = this.skewsPix.length,
                      skew = skewPix / limit;
                if ( skew < this.config.snap ) {
                    this.config.skewType = -1;
                } else if ( skew > (1 - this.config.snap) ) {
                    this.config.skewType = Nskew;
                } else {
                    const skewFrac = 1/Nskew;
                    let s = 0;
                    for (; s<Nskew; ++s) {
                        if ( skew < (s+1)*skewFrac ) {
                            this.config.skewType = s;
                            this.config.skews[s] = skew;
                            this.skewsPix[s] = skewPix;
                            break;
                        }
                    }
                }
                this.onResize(true);
                this.node.removeEventListener('mousemove', onmousemove);
                document.body.removeEventListener('mouseup', onmouseup);
            };
            this.node.addEventListener('mousemove', onmousemove);
            document.body.addEventListener('mouseup', onmouseup);
        }
    };

    const TabsDefs = {
        position: 'top'
    };
    exports.Tabs = class Tabs {
        constructor(config) {
            this.config = Object.assign({}, TabsDefs, config);
            this.node = this.config.node || document.createElement('div');
            this.node.classList.add('qubx_tabs');
            this.node.classList.add('qubx_tabs_'+this.config.position, true);
            this.bar = d3.select(this.node).append('div');
            this.bar.classed('qubx_tabs_bar', true);
            this.content = d3.select(this.node).append('div').node();
            this.content.classList.add('qubx_tabs_content');
            this.bg = this.config.defaultContent;
            if ( this.bg ) {
                this.bg.classList.add('qubx_fullsize');
                this.content.appendChild(this.bg);
            }
            this.items = [];
            this.item = null;
            this.ev = qubx.makeEvents(['pick']);
            this.updateBar = qubx.batchDelayed(this.renderBar.bind(this), 100);
        }
        add(item) {
            this.insert(this.items.length, item);
        }
        insert(index, item) {
            item.node.classList.add('qubx_fullsize');
            if ( item.tabOffscreen ) {
                if ( item.tabOffscreen === true ) {
                    item.tabOffAuto = true;
                    item.tabOffscreen = document.createElement('div');
                    item.tabOffscreen.className = 'qubx_tabs_offscreen';
                    document.body.appendChild(item.tabOffscreen);
                }
                item.tabOffscreen.appendChild(item.node);
            }
            this.items.splice(index, 0, item);
            this.updateBar();
        }
        remove(item) {
            if ( item === this.item ) {
                this.pick(null);
            }
            let i = this.items.indexOf(item);
            if ( i < 0 ) {
                return;
            }
            if ( item.tabOffscreen ) {
                item.tabOffscreen.removeChild(item.node);
                if ( item.tabOffAuto ) {
                    document.body.removeChild(item.tabOffscreen);
                    item.tabOffscreen = true;
                    delete item.tabOffAuto;
                }
            }
            this.items.splice(i, 1);
            this.updateBar();
        }
        pick(item) {
            if ( item && item.length ) { // by title
                item = this.items.filter(it => it.tabTitle === item)[0];
            }
            if ( item === this.item ) {
                return Promise.resolve(true);
            }
            return (this.item && this.item.onShow && this.item.onShow(false) || Promise.resolve(true)).then(() => {
                if ( this.item ) {
                    this.content.removeChild(this.item.node);
                    this.item.tabOffscreen && this.item.tabOffscreen.appendChild(this.item.node);
                } else if ( this.bg ) {
                    this.content.removeChild(this.bg);
                }
                this.item = item;
                if ( item ) {
                    this.item.tabOffscreen && this.item.tabOffscreen.removeChild(this.item.node);
                    this.content.appendChild(item.node);
                    item.onShow && item.onShow(true)
                } else if ( this.bg ) {
                    this.content.appendChild(this.bg);
                }
                this.updateBar();
                this.ev.pick.trigger(item);
            });
        }
        renderBar() {
            const buttons = this.bar.selectAll('.qubx_tab_btn')
                  .data(this.items);
            buttons.enter().append('div')
                .html(it => it.tabIcon || '\u274d')
                .attr('title', it => it.tabTitle || '')
                .on('click', it => this.pick(it))
                .classed('qubx_tab_btn', true);
            buttons.exit().remove();
            buttons.classed('qubx_tab_active', it => it === this.item);
            buttons.classed('qubx_tab_inactive', it => it !== this.item);
        }
    };
    
    exports.SnapSliderTransform = maxScale => (xInitial, sliderOffset) => {
        // -1 <= sliderOffset <= 1, transform(x,0)===x
        if ( ! sliderOffset ) {
            return xInitial;
        }
        if ( ! xInitial ) {
            xInitial = (sliderOffset > 0) ? 1.0 : -1.0;
        }
        let scale = Math.pow(maxScale, Math.abs(sliderOffset));
        if ( (xInitial > 0) !== (sliderOffset > 0) ) {
            scale = 1.0 / scale;
        }
        return xInitial * scale;
    };
    const SnapSliderDefs = {
        transform: exports.SnapSliderTransform(6.283e2)
    };
    exports.SnapSlider = class SnapSlider {
        constructor(config) {
            this.config = Object.assign({}, SnapSliderDefs, config);
            this.node = this.config.node || document.createElement('input');
            this.node.type = 'range';
            this.node.min = "-1.0";
            this.node.max = "1.0";
            this.node.step = ".001";
            this.node.value = "0.0";
            this.__value = 1.0;
            this.node.addEventListener('input', this.__onInput.bind(this)); // immediate
            this.node.addEventListener('change', this.__onChange.bind(this)); // mouseup
            this.ev = qubx.makeEvents(['change', 'recenter']);
        }
        value(v, unlessStretched) {
            const sliderPos = +this.node.value;
            if ( unlessStretched && sliderPos ) {
                return;
            }
            if ( typeof(v) === 'undefined' ) {
                return this.config.transform(this.__value, sliderPos);
            }
            this.__value = v;
            return this;
        }
        focus() {
            this.node.focus();
        }
        __moved() {
            this.ev.change.trigger(this.value());
        }
        __onInput() {
            this.__moved();
            this.__snapBack(true);
        }
        __onChange() {
            this.__moved();
            this.__snapBack();
        }
        __snapBack(cancel) {
            if ( cancel ) {
                if ( this.__timeout ) {
                    cancelTimeout(this.__timeout);
                    this.__timeout = null;
                }
                return;
            }
            if ( ! this.__timeout ) {
                this.__timeout = setTimeout(this.__postChange.bind(this), 10);
            }
        }
        __postChange() {
            this.__timeout = null;
            this.value(this.value());
            this.node.value = "0.0";
            this.ev.recenter.trigger();
        }
    };

    return exports;
});
