/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_ui_data", ["qubx", "qubx_ui", "qubtrace", "qubx_lib", "qubx_data_work", "qubx_idl", "d3", "hammer", "cloud"], (qubx, qubx_ui, qubtrace, qubx_lib, qubx_data_work, qubx_idl, d3, Hammer, cloud) => {
    'use strict';
    
    const exports = {};

    const PlotConfig = {
        backStyle: "rgb(255,255,255)",
        dataStyle: "rgb(100,100,100)",
        dataChosenStyle: "rgb(10,100,200)",
        normStyle: "rgb(0,0,0)",
        normChosenStyle: "rgb(55,155,255)",
        stimulusStyle: ["rgb(0,128,0)", "rgb(128,40,0)"],
        stimulusChosenStyle: ["rgb(0,255,0)", "rgb(255,120,0)"],
        lineStyle: "rgb(0,0,0)",
        lineChosenStyle: "rgb(255,0,40)",
        selStyle: "rgba(100,150,250, 0.2)",
        fitStyle: "rgba(255, 80, 120, 0.7)",
        llStyle: "rgba(100,100,10, 0.8)",
        nodeStyle: "rgba(128, 128, 255, 0.9)",
        resampledStyle: "rgba(155, 0, 111, 0.4)",
        resampledRad: 2,
        idlStyle: "rgba(255,80,120,0.5)",
        idlChosenStyle: "rgba(155, 155, 255,0.5)",
        sampling: 1,
        minSppRaster: 5.0,
        lineHeight: 1/50, // as fraction of trace height
        nodeHeight: .1,
        scalebarHeight: .15,
        scalebarFont: "sans-serif",
        scalebarLores: false,
        scalebarLoresPx: .99,
        scalebarLoresPy: .1,
        scalebarHires: true,
        scalebarHiresPx: 1,
        scalebarHiresPy: 0,
        scalebarBack: "rgba(254,254,254,.5)",
        scalebarStyle: "rgb(0,0,0)"
        // also expected: node: from the DOM
        // optional: hires: DOM node for detail view
    };
    const wheelDeltaGrow = [0, 1.1, 1.5]; // for wheelEvent.deltaMode in [pixel, line, page]
    const FILTER_CHUNK = 4096;

    //// baseline nodes:
    /// data-entry mode includes baseline correction phase
    /// lower pane showing hires subtracted preview and baselineSel controls
    ///  (sets view.baselining = true)
    ///  view.ev.pickNode.sub(function(baselineSel) {
    ///    edit text of baselineSel.t, baselineSel.ix, baselineSel.level
    ///    button delete
    ///    (if baselineSel !== null)
    ///  });
    
    exports.Plot = class Plot {
        constructor(config) {
            this.ev = qubx.makeEvents(["select", "hilite", "choose", 'changeData', 'changeFilter', 'pickNode']);
            this.node = config.node;
            this.hires = config.hires;
            this.config = Object.assign({}, PlotConfig, config);
            this.w = this.node.clientWidth;
            this.h = this.node.clientHeight;
            this.dpRatio = 1.0;
            this.baselining = false;
            this._rawStim = false;
            this.workers = null; // needs to be set from outside for filtering; qubx_data_work.DataWorkgroup
            const plotLayer = d3.select(this.node).append('div');
            plotLayer.classed('plotLayer', true);
            this.plotLayer = plotLayer.node();
            this.canvas = plotLayer.append('canvas').property({width: this.w, height: this.h}).node();
            const selLayer = d3.select(this.node).append('div');
            selLayer.classed('plotLayer', true);
            this.selLayer = selLayer.node();
            this.selCanvas = selLayer.append('canvas').property({width: this.w, height: this.h}).attr({oncontextmenu: "return false;"}).node();
            if ( this.config.scalebarLores ) {
                this.scaleCanvas = document.createElement('canvas');
                this.scaleCanvas.className = 'scaleCanvas qubx_noselect';
                this.scaleCanvas.setAttribute("oncontextmenu", "return false;");
                this.node.appendChild(this.scaleCanvas);
            }
            if ( this.hires ) {
                const hiresLayer = d3.select(this.hires).append('div');
                hiresLayer.classed('plotLayer', true);
                this.hiresLayer = hiresLayer.node();
                this.hiresCanvas = hiresLayer.append('canvas').property({
                    width: this.hires.clientWidth,
                    height: this.hires.clientHeight
                }).node();
                const fitLayer = d3.select(this.hires).append('div');
                fitLayer.classed('plotLayer', true);
                this.fitLayer = fitLayer.node();
                this.fitCanvas = fitLayer.append('canvas').property({
                    width: this.hires.clientWidth,
                    height: this.hires.clientHeight
                }).node();
                const hiSelLayer = d3.select(this.hires).append('div');
                hiSelLayer.classed('plotLayer', true);
                this.hiSelLayer = hiSelLayer.node();
                this.hiSelCanvas = hiSelLayer.append('canvas')
                    .property({
                        width: this.hires.clientWidth,
                        height: this.hires.clientHeight
                    }).attr({
                        oncontextmenu: "return false;"
                    }).node();
                if ( this.config.scalebarHires ) {
                    this.hiresScaleCanvas = document.createElement('canvas');
                    this.hiresScaleCanvas.className = 'scaleCanvas qubx_noselect';
                    this.hiresScaleCanvas.setAttribute("oncontextmenu", "return false;");
                    this.hires.appendChild(this.hiresScaleCanvas);
                }
            }
            [this.scaleCanvas, this.hiresScaleCanvas].forEach((scale, hires) => {
                const propPx = hires ? 'scalebarHiresPx' : 'scalebarLoresPx',
                      propPy = hires ? 'scalebarHiresPy' : 'scalebarLoresPy';
                if ( scale ) {
                    const parent = scale.parentNode;
                    let down = null;
                    const onMouseUp = evt => {
                        try {
                            document.body.removeEventListener('mouseup', onMouseUp);
                            parent.removeEventListener('mousemove', onMouseMove);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    };
                    const onMouseMove = evt => {
                        try {
                            const rel = qubx.relMouseCoords(parent, evt),
                                  x = Math.max(0, Math.min(parent.clientWidth-5, parseInt(scale.style.left) + rel.x - down.x)),
                                  y = Math.max(0, Math.min(parent.clientHeight-5, parseInt(scale.style.top) + rel.y - down.y));
                            scale.style.left = ''+x+'px';
                            scale.style.top = ''+y+'px';
                            this.config[propPx] = (x < (parent.clientWidth/2)) ? (x/parent.clientWidth) : (x/(parent.clientWidth - scale.clientWidth));
                            this.config[propPy] = (y < (parent.clientHeight/2)) ? (y/parent.clientHeight) : (y/(parent.clientHeight - scale.clientHeight));
                            down = rel;
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    };
                    scale.addEventListener('mousedown', evt => {
                        try {
                            evt.stopPropagation();
                            evt.preventDefault();
                            down = qubx.relMouseCoords(parent, evt);
                            document.body.addEventListener('mouseup', onMouseUp);
                            parent.addEventListener('mousemove', onMouseMove);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                }
            });
            
            const hovering = d3.select(this.node).append('div');
            hovering.classed('qubx_ui_data_hovering', true);
            hovering.append('div').classed('qubx_ui_data_hovering_bg', true);
            const fg = hovering.append('div');
            fg.classed('qubx_ui_data_hovering_fg', true);
            const choCon = fg.append('div');
            this.chosenControl = choCon.node();
            choCon.classed('qubx_ui_data_ctrl', true);
            choCon.append('span').text('Feat. trace:')
                .classed('qubx_ui_data_ctrl_caption', true);
            choCon.append('button').text('-')
                .on('click', this._onClickChosenMinus.bind(this))
                .classed('qubx_ui_data_chosen_minus', true);
            choCon.append('input').property({
                type: 'text',
                value: 'All'
            })
                .classed('qubx_ui_data_chosen', true);
            choCon.append('button').text('+')
                .on('click', this._onClickChosenPlus.bind(this))
                .classed('qubx_ui_data_chosen_plus', true);
            this._chosen = -1;
            this.txtChosen = new qubx_ui.AcceptedInput(this.chosenControl.querySelector(".qubx_ui_data_chosen"), this._chosen+1, {
                accept: this.acceptSegIndex1.bind(this),
                format: this.formatSegIndex1.bind(this),
                defval: 1,
                widthChars: 2.5
            });
            this.txtChosen.ev.change.sub(this._onChangeChosen.bind(this));
            
            this.selFrom = this.selTo = 0; // as fractions 0..1 of total
            this.hiliteFrom = this.hiliteTo = 0;
            this._data = null;
            this.maxNpoint = 0;
            this.bounds = [{lo: 0, hi: 0, dy: 0}];
            this.drawSoon = qubx.batchDelayed(this.draw.bind(this), 50);
            this.drawHiresSoon = qubx.batchDelayed(this.drawHires.bind(this), 50);
            this.drawFitSoon = qubx.batchDelayed(this.drawFit.bind(this), 100, true);
            this.reSelectSoon = qubx.batchDelayed(this.reSelect.bind(this), 50);
            this.reHiSelSoon = qubx.batchDelayed(this.reHiSel.bind(this), 10);
            this.sendHiSelSoon = qubx.batchDelayed(this.sendHiSel.bind(this), 100);
            this.reshape();
            this.mouseX0 = null;
            this.canvasTouch = new Hammer(this.selCanvas);
            this.canvasTouch.on("pan", this._onPanSel.bind(this));
            this.canvasTouch.on("doubletap", this._onDoubleTap.bind(this));
            if ( this.hires ) {
                this.hiresCanvasTouch = new Hammer(this.hiSelCanvas);
                this.hiresCanvasTouch.get('pinch').set({ enable: true });
                this.hiresCanvasTouch.on("pan", this._onPanHires.bind(this));
                this.hiresCanvasTouch.on("panend", this._onPanendHires.bind(this));
                this.hiresCanvasTouch.on("pinch", this._onPinchHires.bind(this));
                this.hiresCanvasTouch.on("pinchend", this._onPinchendHires.bind(this));
                this.hiresCanvasTouch.on("doubletap", this._onDoubleTapHires.bind(this));
                this.hiSelCanvas.addEventListener('wheel', this._onWheelHires.bind(this));
                this.hiSelCanvas.addEventListener('mousemove', this._onMouseMoveHires.bind(this));
                this.hiSelCanvas.addEventListener('mousedown', this._onMouseDownHires.bind(this));
                this.hiSelCanvas.addEventListener('touchstart', this._onTouchStartHires.bind(this));
                this.zoom(0,1);
                this.hiMoving = 0;
                this.filterSoon = qubx.batchDelayed(this.__startFilter.bind(this), 100);
                this.__filtering = 0; // 1: busy, -1: interrupt&restart
                this.__filterKHz = 0;
            }
            this.raster_samples = qubtrace.raster_samples;
            qubx_lib.then(lib => {
                // prefer asm.js version
                this.raster_samples = lib.raster_samples;
            }).catch(qubx.reportErr);
        }
        data(d) {
            if ( typeof(d) === 'undefined' ) {
                return this._data;
            }
            if ( d === this._data ) {
                return this;
            }
            if ( this._data ) {
                this._data.ev.changeSampling.unsub(this._dataOnChangeSampling);
                this._data.ev.changeSegCount.unsub(this._dataOnChangeSegCount);
                this._data.ev.changeSegSize.unsub(this._dataOnChangeSegSize);
                this._data.ev.doneSegSize.unsub(this._dataOnDoneSegSize);
                this._data.ev.changeSamples.unsub(this._dataOnChangeSamples);
                this._data.ev.changeBounds.unsub(this._dataOnChangeBounds);
                this._data.ev.changeRaw.unsub(this._dataOnChangeRaw);
                this._data.ev.changeIdeal.unsub(this._dataOnChangeIdeal);
                this._data.ev.changeFit.unsub(this._dataOnChangeFit);
                this._data.ev.changePointLL.unsub(this._dataOnChangePointLL);
                this._data.ev.changeBaseline.unsub(this._dataOnChangeBaseline);
                this._data.ev.clear.unsub(this._dataOnClear);
            }
            this._data = d;
            if ( this._data ) {
                this._dataOnChangeSampling = this._data.ev.changeSampling.sub(this._onChangeSampling.bind(this));
                this._dataOnChangeSegCount = this._data.ev.changeSegCount.sub(this._onChangeSegCount.bind(this));
                this._dataOnChangeSegSize = this._data.ev.changeSegSize.sub(this._onChangeSegSize.bind(this));
                this._dataOnDoneSegSize = this._data.ev.doneSegSize.sub(this._onDoneSegSize.bind(this));
                this._dataOnChangeSamples = this._data.ev.changeSamples.sub(this._onChangeSamples.bind(this));
                this._dataOnChangeBounds = this._data.ev.changeBounds.sub(this._onChangeBounds.bind(this));
                this._dataOnChangeRaw = this._data.ev.changeRaw.sub(this._onChangeRaw.bind(this));
                this._dataOnChangeIdeal = this._data.ev.changeIdeal.sub(this._onChangeIdeal.bind(this));
                this._dataOnChangeFit = this._data.ev.changeFit.sub(this._onChangeFit.bind(this));
                this._dataOnChangePointLL = this._data.ev.changePointLL.sub(this._onChangePointLL.bind(this));
                this._dataOnChangeBaseline = this._data.ev.changeBaseline.sub(this._onChangeBaseline.bind(this));
                this._dataOnClear = this._data.ev.clear.sub(this._onClear.bind(this));
            }
            this._onChangeSegCount(this._data && this._data.segments.length || 0);
            this._onDoneSegSize();
            this.filterSoon();
            
            this.ev.changeData.trigger(d);
            return this;
        }
        filterKHz(freq) {
            if ( typeof(freq) === 'undefined' ) {
                return this.__filterKHz;
            }
            if ( freq !== this.__filterKHz ) {
                this.__filterKHz = freq;
                this.ev.changeFilter.trigger(freq);
                if ( freq ) {
                    this.filterSoon();
                } else {
                    this.__unfilter();
                }
            }
            return this;
        }
        rawStim(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._rawStim;
            }
            x = !!x;
            if ( this._rawStim !== x ) {
                this._rawStim = x;
                this.drawSoon();
                if ( this.hires ) {
                    this.drawHiresSoon();
                }
            }
            return this;
        }
        _onChangeSampling(sampling) {
            this.reshape();
        }
        _onChangeSegCount(count) {
            this.chosen(Math.min(this._chosen, count-1));
        }
        _onChangeSegSize(iSeg, n) {
            
        }
        _onDoneSegSize() {
            this.erase();
            this.reshape();
        }
        _onClear(iSeg) {
            const oneSeg = (typeof(iSeg)!=='undefined');
            if ( ! oneSeg ) {
                this.bounds = this._data.signals.map(sig => ({
                    lo: 1e20,
                    hi: -1e20,
                    dy: null
                }));
                this.erase();
            } else {
                this.drawSoon();
            }
        }
        _onChangeSamples(iSeg, iSignal, offset, count, immediate) {
            const seg = this._data.segments[iSeg];
            if ( (iSignal === 0) || (! seg.ideal[iSignal]) || (seg.ideal[iSignal].classAt(0) < 0) ) {
                let ex = qubx.extrema(seg.sampled[iSignal], offset, count);
                if ( (! (ex.lo || ex.hi)) && seg.ideal[iSignal] && seg.ideal[iSignal].amp && seg.ideal[iSignal].amp.length  ) {
                    ex = qubx.extrema(seg.ideal[iSignal].amp)
                }
                const bd = this.bounds[iSignal] = this.bounds[iSignal] || {};
                bd.lo = (typeof(bd.lo) !== 'undefined') ? bd.lo : 1e20;
                bd.hi = (typeof(bd.hi) !== 'undefined') ? bd.hi : -1e20;
                ex.lo = Math.min(bd.lo, ex.lo);
                ex.hi = Math.max(bd.hi, ex.hi);
                if ( (ex.lo !== bd.lo) || (ex.hi !== bd.hi) ) {
                    this.setBounds(iSignal, ex.lo, ex.hi);
                    if ( immediate && ! this.drawSoon.active )  {
                        this.draw(offset, count);
                    } else {
                        this.drawSoon();
                    }
                } else {
                    this.drawSoon();
                }
            }
            this._drawHiresOverlap(iSeg, offset, count);
            this.reSelectSoon();
        }
        _onChangeBounds(iSignal) {
            let lo = 1e20,
                hi = -1e20;
            for (const seg of this._data.segments) {
                const ex = qubx.extrema(seg.sampled[iSignal]);
                lo = Math.min(lo, ex.lo);
                hi = Math.max(hi, ex.hi);
            }
            if ( lo !== 1e20 ) {
                this.setBounds(iSignal, lo, hi);
                this.drawSoon();
                this.hires && this.drawHiresSoon();
            }
        }
        _onChangeRaw(iSeg, offset, count, immediate) {
            this._onChangeSamples(iSeg, 0, offset, count, immediate);
            const seg = this._data.segments[iSeg];
            if ( this.__filterKHz && (iSeg === (this._data.segments.length-1)) && ((offset+count) === seg.n) ) {
                this.filterSoon();
            } else {
                this._data.ev.changeSamples.trigger(iSeg, 0, offset, count, immediate);
            }
        }
        __startFilter() {
            if ( this.__filtering ) {
                this.__filtering = -1;
                return;
            }
            if ( ! (this.__filterKHz && this._data.segments.length && this._data.segments[0].sampled[0]) ) {
                return;
            }
            const sampling = this._data.sampling(),
                  overlap = this.workers.getFilterOverlap(sampling, this.__filterKHz),
                  Nfull = FILTER_CHUNK + 2*overlap;
            this.workers.reserve(Nfull);
            Promise.all(this._data.segments.map((seg, iSeg) => Promise.all(qubx.range(0, seg.n, FILTER_CHUNK).map(at => this.workers.doWork(worker => {
                if ( this.__filtering < 0 ) {
                    return;
                }
                const chunk = new Float32Array(Nfull),
                      limit = at + Nfull - overlap,
                      y0 = seg.raw[at],
                      yn = seg.raw[Math.min(seg.n, at+FILTER_CHUNK)-1];
                let i = 0,
                    j = at - overlap;
                for (; j<limit; ++i, ++j) {
                    if ( j < 0 ) {
                        chunk[i] = y0;
                    } else if ( j >= seg.n ) {
                        chunk[i] = yn;
                    } else {
                        chunk[i] = seg.raw[j];
                    }
                }
                return worker.filter(chunk, sampling, this.__filterKHz).then(() => {
                    const subChunk = chunk.subarray(overlap, overlap+Math.min(seg.n-at, FILTER_CHUNK));
                    this._data.setSampled(iSeg, 0, at, subChunk, subChunk.length, 0, false, true);
                });
            }))))).catch(qubx.reportErr)
                .catch(err => {})
                .then(() => {
                    if ( this.__filtering < 0 ) {
                        this.filterSoon();
                    }
                    this.__filtering = 0;
                });
        }
        __unfilter() {
            this._data.segments.forEach((seg, iSeg) => {
                seg.sampled[0].set(seg.raw);
                this._data.ev.changeSamples.trigger(iSeg, 0, 0, seg.n);
            });
        }
        _onChangeIdeal(iSeg, iSignal) {
            const seg = this._data.segments[iSeg],
                  bd = this.bounds[iSignal],
                  lo = Math.min(bd ? bd.lo : 1e100, qubx.arrayMin(seg.ideal[iSignal].amp)),
                  hi = Math.max(bd ? bd.hi : -1e100, qubx.arrayMax(seg.ideal[iSignal].amp));
            if ( (! bd) || (lo !== bd.lo) || (hi !== bd.hi) ) {
                this.setBounds(iSignal, lo, hi);
            } else if ( ! this.drawSoon.active ) {
                this.draw();
            }
            this.hires && this.drawHiresSoon();
        }
        _onChangeFit(iSeg) {
            this.drawFitSoon();
        }
        _onChangePointLL(iSeg) {
            this._pointLLex = null;
            this.drawFitSoon();
        }
        _onChangeBaseline(iSeg) {
            this.drawFitSoon();
        }
        chosen(i) {
            if ( typeof(i) === 'undefined' ) {
                return this._chosen;
            }
            if ( (this._chosen !== i) && (i >= -1) && this._data && (i < this._data.segments.length) ) {
                this._chosen = i;
                this.txtChosen.update(i+1);
                this._pointLLex = null;
                this.drawSoon();
                if ( this.hires ) {
                    this.drawHiresSoon();
                    this.reHiSelSoon();
                }
                this.ev.choose.trigger(i);
            }
            return this;
        }
        _onClickChosenMinus() {
            try {
                this.chosen(this._chosen-1);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onClickChosenPlus() {
            try {
                this.chosen(this._chosen+1);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onChangeChosen(chosen) {
            this.chosen(chosen);
        }
        acceptSegIndex1(s) {
            if ( s === 'All' ) {
                return 0;
            }
            return qubx.acceptIntBetween(0, this_data && this._data.segmentCount()||0)(s);
        }
        formatSegIndex1(x) {
            if ( x === 0 ) {
                return 'All';
            }
            return ""+x;
        }
        setBounds(iSignal, lo, hi) {
            const dy = hi - lo;
            this.bounds[iSignal] = {
                lo: lo,
                hi: hi,
                dy: dy
            };
            this.drawSoon();
            this.hires && this.drawHiresSoon();
        }
        zoom(from, to, immediate) {
            if ( typeof(to) === 'undefined' ) {
                const z = {
                    from: this.selFrom,
                    to: this.selTo,
                    offset: Math.round(this.selFrom*this.maxNpoint)
                };
                z.count = Math.round(z.to*this.maxNpoint) - z.offset + 1;
                return z;
            }
            from = Math.max(0, from);
            to = Math.min(1, to);
            if ( (from !== this.selFrom) || (to !== this.selTo) ) {
                this.selFrom = from;
                this.selTo = to;
                this.sppHi = this.zoom().count / this.w;
                const sampling = this._data ? this._data.sampling() : 1.0;
                this.xppHi = sampling * this.sppHi;
                this.drawSel();
                if ( this.hires ) {
                    if ( immediate ) {
                        this.drawHires();
                        this.reSelect();
                    } else {
                        this.drawHiresSoon();
                        this.reSelectSoon();
                    }
                } else {
                    this.reSelect();
                }
            }
            return this;
        }
        reSelect(onSelect) {
            onSelect = onSelect || this.ev.select.trigger.bind(this.ev.select);
            if ( ! this._data ) {
                onSelect(-1, -1, []);
                return;
            }
            const iFrom = Math.round(this.selFrom*this.maxNpoint),
                  iTo = this.selTo ? Math.round(this.selTo*(this.maxNpoint-1)) : (this.maxNpoint-1),
                  segments = this._data.segments.map((seg, iSeg) => {
                      const segOut = {
                          index: iSeg,
                          n: seg.n,
                          selFrom: Math.min(seg.n-1, iFrom),
                          selTo: Math.min(seg.n-1, iTo),
                          indVar: seg.indVar,
                          sources: seg.sampled,
                          raw: seg.raw,
                          signals: []
                      };
                /*segOut.sources.forEach(function(src, i) {
                  segOut.signals[i] = src.slice && src.slice(iFrom, Math.min(iTo, src.length));
                  }); wait for TypedArray.slice to become mainstream */
                
                      return segOut;
                  });
            onSelect(iFrom, iTo-iFrom+1, segments);
        }
        hilite(from, to, immediate) {
            if ( typeof(to) === 'undefined' ) {
                const z = {
                    from: this.hiliteFrom,
                    to: this.hiliteTo,
                    offset: Math.round(this.hiliteFrom*this.maxNpoint)
                };
                z.count = Math.round(z.to * this.maxNpoint) - z.offset + 1;
                return z;
            }
            from = Math.max(0, from);
            to = Math.min(1, to);
            if ( (from !== this.hiliteFrom) || (to !== this.hiliteTo) ) {
                this.hiliteFrom = from;
                this.hiliteTo = to;
            }
            if ( immediate ) {
                this.reHiSel();
            } else {
                this.reHiSelSoon();
            }
        }
        reHiSel(onSelect) {
            this.drawHiSel();
            this.sendHiSelSoon();
        }
        sendHiSel() {
            if ( ! this._data ) {
                this.ev.hilite.trigger(-1, -1, []);
                return;
            }
            const hilite = this.hilite(),
                  src = (this._chosen < 0) ? this._data.segments : [this._data.segments[this._chosen]];
            if ( hilite.to ) {
                const segments = src.map((seg, i) => {
                    const segOut = {
                        index: (this._chosen < 0) ? i : this._chosen,
                        from: Math.min(seg.n-1, hilite.offset),
                        to: Math.min(seg.n-1, hilite.offset+hilite.count-1)
                    };
                    segOut.count = segOut.to - segOut.from + 1;
                    return segOut;
                });
                this.ev.hilite.trigger(hilite.offset, hilite.count, segments);
            } else {
                const iFrom = Math.round(this.selFrom*this.maxNpoint),
                      iTo = this.selTo ? Math.round(this.selTo*(this.maxNpoint-1)) : (this.maxNpoint-1),
                      segments = src.map((seg, i) => {
                          const segOut = {
                              index: (this._chosen < 0) ? i : this._chosen,
                              from: Math.min(seg.n-1, iFrom),
                              to: Math.min(seg.n-1, iTo)
                          };
                          segOut.count = segOut.to - segOut.from + 1;
                          return segOut;
                      });
                this.ev.hilite.trigger(iFrom, iTo-iFrom+1, segments);
            }
        }
        drawHiSel() {
            const hilite = this.hilite(),
                  ctx = this.hiSelCanvas.getContext('2d');
            ctx.clearRect(0, 0, this.wHi, this.hHi);
            if ( ! hilite.count ) {
                return;
            }
            const zoom = this.zoom();
            if ( ! zoom.count ) {
                return;
            }
            const selW = zoom.to - zoom.from,
                  x0 = (hilite.from - zoom.from) / selW,
                  dx = (hilite.to - hilite.from) / selW;
            ctx.fillStyle = this.config.selStyle;
            ctx.fillRect(x0*this.wHi, 0, dx*this.wHi, this.hHi);
        }
        _drawHiresOverlap(iSeg, offset, count) {
            if ( this.hires && (this._chosen < 0) || (iSeg === this._chosen) ) {
                const zoom = this.zoom();
                if ( Math.max(offset, zoom.offset) <= Math.min(offset+count-1, zoom.offset+zoom.count-1) ) {
                    this.drawHiresSoon();
                }
            }
        }
        reshape() {
            this.dpRatio = window.devicePixelRatio || 1.0;
            this.w = (this.dpRatio * this.plotLayer.clientWidth) || 100;
            this.h = (this.dpRatio * this.plotLayer.clientHeight) || 100;
            this.canvas.width = this.w;
            this.canvas.height = this.h;
            this.selCanvas.width = this.w;
            this.selCanvas.height = this.h;
            let maxNpoint = 1;
            this._data && this._data.segments.map(seg => {
                if ( seg.n > maxNpoint ) {
                    maxNpoint = seg.n;
                }
            });
            const sampling = this._data ? this._data.sampling() : 1;
            this.maxNpoint = maxNpoint;
            this.spp = maxNpoint / this.w;
            this.xpp = sampling * this.spp;
            if ( this.hires ) {
                this.wHi = this.hiresLayer.clientWidth * this.dpRatio;
                this.hHi = this.hiresLayer.clientHeight * this.dpRatio;
                this.hiresCanvas.width = this.wHi;
                this.hiresCanvas.height = this.hHi;
                this.fitCanvas.width = this.wHi;
                this.fitCanvas.height = this.hHi;
                this.hiSelCanvas.width = this.wHi;
                this.hiSelCanvas.height = this.hHi;
                this.sppHi = this.zoom().count / this.wHi;
                this.xppHi = sampling * this.sppHi;
            }
            this.drawSoon();
            this.drawSel();
            this.hires && this.drawHiresSoon();
        }
        drawSel() {
            const ctx = this.selCanvas.getContext('2d');
            ctx.clearRect(0, 0, this.w, this.h);
            ctx.fillStyle = this.config.selStyle;
            ctx.fillRect(this.selFrom*this.w, 0, (this.selTo - this.selFrom)*this.w, this.h);
        }
        erase(loresOnly) {
            const toErase = [this.canvas];
            if ( this.hires && ! loresOnly ) {
                toErase.push(this.hiresCanvas);
                toErase.push(this.fitCanvas);
            }
            toErase.forEach(canvas => {
                const ctx = canvas.getContext('2d');
                ctx.fillStyle = this.config.backStyle;
                ctx.fillRect(0, 0, canvas.width, canvas.height);
            });
        }
        draw(offset, count) {
            try {
                this.emsize = (this.dpRatio*.8*this.chosenControl.clientHeight)|0; // ctx.measureText('M').width;
                if ( typeof(offset) === 'undefined' ) {
                    offset = 0;
                    count = this.maxNpoint;
                    this.erase(true);
                } else if ( typeof(count) === 'undefined' ) {
                    count = this.maxNpoint - offset;
                }
                if ( ! this._data ) {
                    return;
                }
                this.layoutLo = qubtrace.layoutSignals(this.w, this.h, this.emsize, this._data.signalCount()-1);
                const ctx = this.canvas.getContext('2d'),
                      spp = this.spp,
                      i0 = Math.max(0, Math.min(this.w-1, Math.floor(offset/spp))),
                      i1 = Math.max(0, Math.min(this.w-1, Math.ceil((offset+count)/spp))),
                      n = this._data.segments.length;
                let i = 0;
                for (; i<n; ++i) {
                    if ( i !== this._chosen ) {
                        this._draw(ctx, i0, (i1-i0+1), this.h, this.layoutLo, i, offset, count,
                                   this.config.dataStyle, this.config.normStyle,
                                   this.config.lineStyle, this.config.stimulusStyle,
                                   null, this.config.idlStyle);
                    }
                }
                if ( (0 <= this._chosen) && (this._chosen < this._data.segments.length) ) {
                    this._draw(ctx, i0, (i1-i0+1), this.h, this.layoutLo, this._chosen, offset, count,
                               this.config.dataChosenStyle, this.config.normChosenStyle,
                               this.config.lineChosenStyle, this.config.stimulusChosenStyle,
                               null, this.config.idlChosenStyle);
                }
                if ( this.config.scalebarLores ) {
                    this._drawScalebar(this.scaleCanvas, this.w, this.h, this.xpp,
                                       this.bounds[0].dy/this.layoutLo.current.height);
                    this.scaleCanvas.style.width = (this.scaleCanvas.width / this.dpRatio)+'px';
                    this.scaleCanvas.style.height = (this.scaleCanvas.height / this.dpRatio)+'px';
                    this.scaleCanvas.style.left = ((this.config.scalebarLoresPx < 0.5) ? ((this.config.scalebarLoresPx * this.canvas.width / this.dpRatio)|0)
                                                   : ((this.config.scalebarLoresPx * (this.canvas.width - this.scaleCanvas.width) / this.dpRatio)|0)) + 'px';
                    this.scaleCanvas.style.top = ((this.config.scalebarLoresPy < 0.5) ? ((this.config.scalebarLoresPy * this.canvas.height / this.dpRatio)|0)
                                                  : ((this.config.scalebarLoresPy * (this.canvas.height - this.scaleCanvas.height) / this.dpRatio)|0)) + 'px';
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        drawHires() {
            try {
                const ctx = this.hiresCanvas.getContext('2d'),
                      zoom = this.zoom();
                ctx.fillStyle = this.config.backStyle;
                ctx.fillRect(0, 0, this.wHi, this.hHi);
                if ( ! this._data ) {
                    return;
                }
                this.layoutHi = qubtrace.layoutSignals(this.wHi, this.hHi, this.emsize, this._data.signalCount()-1);
                if ( (0 <= this._chosen) && (this._chosen < this._data.segments.length) ) {
                    this._draw(ctx, 0, this.wHi, this.hHi, this.layoutHi, this._chosen, zoom.offset, zoom.count,
                               this.config.dataStyle, this.config.normStyle, this.config.lineStyle, this.config.stimulusStyle, this.config.resampledStyle, this.config.idlStyle);
                } else {
                    const n = this._data.segments.length;
                    let i = 0;
                    for (; i<n; ++i ) {
                        this._draw(ctx, 0, this.wHi, this.hHi, this.layoutHi, i, zoom.offset, zoom.count,
                                   this.config.dataStyle, this.config.normStyle, this.config.lineStyle, this.config.stimulusStyle, this.config.resampledStyle, this.config.idlStyle);
                    }
                }
                this.drawFitSoon();
                this.drawHiSel();
                if ( this.config.scalebarHires ) {
                    this._drawScalebar(this.hiresScaleCanvas, this.wHi, this.hHi, this.xppHi, this.layoutHi);
                    this.hiresScaleCanvas.style.width = (this.hiresScaleCanvas.width / this.dpRatio)+'px';
                    this.hiresScaleCanvas.style.height = (this.hiresScaleCanvas.height / this.dpRatio)+'px';
                    this.hiresScaleCanvas.style.left = ((this.config.scalebarHiresPx < 0.5) ? ((this.config.scalebarHiresPx * this.hiresCanvas.width / this.dpRatio)|0)
                                                        : ((this.config.scalebarHiresPx * (this.hiresCanvas.width - this.hiresScaleCanvas.width) / this.dpRatio)|0)) + 'px';
                    this.hiresScaleCanvas.style.top = ((this.config.scalebarHiresPy < 0.5) ? ((this.config.scalebarHiresPy * this.hiresCanvas.height / this.dpRatio)|0)
                                                       : ((this.config.scalebarHiresPy * (this.hiresCanvas.height - this.hiresScaleCanvas.height) / this.dpRatio)|0)) + 'px';
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _drawScalebar(canvas, plotW, plotH, xpp, layout) {
            const signals = this._data.signals.map(sig => ({units: sig.units}));
            if ( ! signals.length ) {
                return;
            }
            signals[0].ypp = this.bounds[0].dy / (layout.current.height||1);
            signals[0].color = this.config.scalebarStyle;
            layout.stimulus.forEach((stim, i) => {
                signals[i+1].ypp = this.bounds[i+1].dy / stim.height;
                signals[i+1].color = this.config.stimulusStyle[i];
            });
            qubtrace.drawScalebar(canvas, this.emsize, plotW, plotH, xpp, signals,
                                  this.config.scalebarHeight, this.config.scalebarFont, this.config.scalebarBack, this.config.scalebarStyle);
        }
        drawFit() {
            const ctx = this.fitCanvas.getContext('2d'),
                  zoom = this.zoom();
            ctx.clearRect(0, 0, this.wHi, this.hHi);
            if ( ! this._data ) {
                return;
            }
            this.pointLLex = this.pointLLex || (this._data.segments.length && this._data.segments[0].pointLL && qubx.extremas(
                this._data.segments.map(qubx.m('pointLL'))
            ));
            const pointLLlyt = {
                height: Math.min(3*this.emsize, this.hHi)
            };
            pointLLlyt.top = (this.hHi - pointLLlyt.height)/2;
            ctx.save();
            ctx.translate(0, this.layoutHi.current.top);
            if ( (0 <= this._chosen) && (this._chosen < this._data.segments.length) ) {
                this._drawNodes(ctx, 0, this.wHi, this.layoutHi.current.height, this._chosen, zoom.offset, zoom.count, this.config.nodeStyle);
                this._drawFit(ctx, 0, this.wHi, this.layoutHi.current.height, this._chosen, zoom.offset, zoom.count, this.config.fitStyle);
                ctx.translate(0, pointLLlyt.top - this.layoutHi.current.top);
                ctx.lineWidth = 0.5*Math.max(1, Math.round(pointLLlyt.height*this.config.lineHeight)|0);
                this.pointLLex && qubtrace.drawLines(ctx, 0, this.wHi, pointLLlyt.height, this._data.segments[this._chosen].pointLL,
                                                     zoom.offset, zoom.count, this.config.llStyle, zoom.count/this.wHi,
                                                     this.pointLLex.lo, (this.pointLLex.hi - this.pointLLex.lo)/pointLLlyt.height);
            } else {
                const n = this._data.segments.length;
                let i = 0;
                for (; i<n; ++i ) {
                    this._drawFit(ctx, 0, this.wHi, this.layoutHi.current.height, i, zoom.offset, zoom.count, this.config.fitStyle);
                }
            }
            ctx.restore();
        }
        _drawFit(ctx, px0, pxw, pxh, iSeg, offset, count, fitStyle) {
            if ( ! this._data ) {
                return;
            }
            let n = count;
            const seg = this._data.segments[iSeg];
            if ( (seg.seen - offset) < count ) {
                n = seg.seen - offset;
                pxw = Math.round(pxw * n/count);
            }
            if ( n <= 1 ) { return; }
            const spp = n / pxw,
                  extrema = seg.extrema || this.bounds[0];
            if ( spp >= this.config.minSppRaster ) {
                this._drawRasterStd(ctx, px0, pxw, pxh, seg.fit, offset, n, fitStyle, spp, extrema.lo, extrema.dy/pxh);
            } else {
                qubtrace.drawLinesStd(ctx, px0, pxw, pxh, seg.fit, offset, n, fitStyle, spp, extrema.lo, extrema.dy/pxh);
            }
        }
        _drawNodes(ctx, px0, pxw, pxh, iSeg, offset, count, nodeStyle) {
            if ( ! this._data ) {
                return;
            }
            let n = count;
            const seg = this._data.segments[iSeg];
            if ( ! seg.baseline ) {
                return;
            }
            if ( (seg.seen - offset) < count ) {
                n = seg.seen - offset;
                pxw = Math.round(pxw * n/count);
            }
            if ( n <= 1 ) { return; }
            const spp = n / pxw,
                  extrema = seg.extrema || this.bounds[0],
                  samples = seg.baseline.toSampled(offset, offset+n-1);
            if ( spp >= this.config.minSppRaster ) {
                this._drawRaster(ctx, px0, pxw, pxh, samples, 0, n, nodeStyle, null, spp, extrema.lo, extrema.dy/pxh);
            } else {
                ctx.lineWidth = Math.max(1, Math.round(pxh*this.config.lineHeight)|0);
                qubtrace.drawLines(ctx, px0, pxw, pxh, samples, 0, n, nodeStyle, spp, extrema.lo, extrema.dy/pxh);
            }
            const nodes = seg.baseline.getNodes(offset, offset+n-1),
                  rad = 0.5*Math.max(1, Math.round(pxh*this.config.nodeHeight)|0);
            qubtrace.drawDots(ctx, px0, pxw, pxh, nodes.indices, nodes.levels, offset, n, nodeStyle, rad, spp, extrema.lo, extrema.dy/pxh,
                              1.5*rad, this.baselineSel && this.baselineSel.ix);
        }
        _draw(ctx, px0, pxw, pxh, layout, iSeg, offset, count, dataStyle, normStyle, lineStyle, stimulusStyle, resampledStyle, idlStyle) {
            let n = count;
            if ( ! this._data ) {
                return;
            }
            const seg = this._data.segments[iSeg];
            if ( (seg.seen - offset) < count ) {
                n = seg.seen - offset;
                pxw = Math.round(pxw * n/count);
            }
            if ( n <= 1 ) { return; }
            const spp = n / pxw;
            if ( spp >= this.config.minSppRaster ) {
                ctx.lineWidth = 1.0;
                seg.sampled.forEach((points, iStim) => {
                    if ( iStim ) {
                        const lyt = layout.stimulus[iStim-1],
                              bd = this.bounds[iStim];
                        ctx.save();
                        ctx.translate(0, lyt.top);
                        if ( (! this._rawStim) && seg.ideal[iStim] && (seg.ideal[iStim].classAt(0) >= 0) ) {
                            this._drawIdl(ctx, px0, pxw, lyt.height, seg.ideal[iStim], offset, n, stimulusStyle[iStim-1], spp, bd.lo,
                                          bd.dy/(lyt.height||1));
                        } else if ( points ) {
                            this._drawRaster(ctx, px0, pxw, lyt.height, points, offset, n, stimulusStyle[iStim-1], null, spp, bd.lo,
                                             bd.dy/(lyt.height||1));
                        }
                        ctx.restore();
                    }
                });
                ctx.save();
                ctx.translate(0, layout.current.top);
                if ( seg.sampled[0] ) {
                    this._drawRaster(ctx, px0, pxw, layout.current.height, seg.sampled[0], offset, n, dataStyle, normStyle, spp, this.bounds[0].lo,
                                     this.bounds[0].dy/layout.current.height);
                }
            } else {
                ctx.lineWidth = 0.5*Math.max(1, Math.round(pxh*this.config.lineHeight)|0);
                seg.sampled.forEach((points, iStim) => {
                    if ( iStim ) {
                        const lyt = layout.stimulus[iStim-1],
                              bd = this.bounds[iStim];
                        ctx.save();
                        ctx.translate(0, lyt.top);
                        if ( seg.ideal[iStim] && seg.ideal[iStim].classAt(0) >= 0 ) {
                            this._drawIdl(ctx, px0, pxw, lyt.height, seg.ideal[iStim], offset, n, stimulusStyle[iStim-1], spp, bd.lo,
                                          bd.dy/(lyt.height||1));
                        } else if ( points ) {
                            qubtrace.drawLines(ctx, px0, pxw, lyt.height, points, offset, n, stimulusStyle[iStim-1], spp, bd.lo,
                                               bd.dy/(lyt.height||1));
                        }
                        ctx.restore();
                    }
                });
                ctx.save();
                ctx.translate(0, layout.current.top);
                if ( seg.sampled[0] ) {
                    qubtrace.drawLines(ctx, px0, pxw, layout.current.height, seg.sampled[0], offset, n, lineStyle, spp, this.bounds[0].lo,
                                       this.bounds[0].dy/layout.current.height);
                }
            }
            if ( resampledStyle && seg.resampled ) {
                qubtrace.drawDots(ctx, px0, pxw, layout.current.height, seg.resampled.closests, seg.resampled.means, offset, n, resampledStyle, this.config.resampledRad, spp, this.bounds[0].lo, this.bounds[0].dy/layout.current.height);
            }
            if ( seg.ideal[0] && idlStyle ) {
                this._drawIdl(ctx, px0, pxw, layout.current.height, seg.ideal[0], offset, n, idlStyle, spp, this.bounds[0].lo, this.bounds[0].dy/layout.current.height);
            }
            ctx.restore();
        }
        _drawIdl(ctx, px0, pxw, pxh, idl, offset, n, idlStyle, spp, yLo, ypp) {
            const lineWidth = 0.5*Math.max(1, Math.round(pxh*this.config.lineHeight)|0),
                  events = idl.getDwellsFLC(offset, offset+n-1);
            if ( events.classes.length > pxw ) {
                const sampled = qubx_idl.sampleDwells(events, idl.amp, idl.std, offset, offset+n-1, lineWidth*ypp, pxw, null, qubx.UNSET);
                exports.drawSampledIdl(ctx, px0, pxw, pxh, sampled.lo, sampled.hi, idlStyle, yLo, ypp);
            } else {
                exports.drawIdl(ctx, px0, pxw, pxh, events, offset, idl.amp, idl.std, idlStyle, lineWidth, spp, yLo, ypp);
            }
        }
        _drawRaster(ctx, px0, pxw, pxh, points, offset, n, dataStyle, normStyle, spp, yLo, ypp) {
            if ( n > (points.length - offset) ) {
                const coverage = (points.length - offset) / n;
                n = points.length - offset;
                pxw = Math.round(pxw * coverage)|0;
            }
            const lineWidth = Math.max(1, Math.round(pxh*this.config.lineHeight)|0),
                  raster = this.raster_samples(points.buffer, offset, n, pxw, spp, 2*lineWidth*ypp);
            qubtrace.drawRastered(ctx, px0, pxw, pxh, raster, dataStyle, normStyle, spp, yLo, ypp);
        }
        _drawRasterStd(ctx, px0, pxw, pxh, points, offset, n, dataStyle, spp, yLo, ypp) {
            const raster = qubtrace.raster_samples_std(points.buffer, offset, n, pxw, spp, ypp);
            qubtrace.drawRastered(ctx, px0, pxw, pxh, raster, dataStyle, null, spp, yLo, ypp);
        }
        _onPanSel(event) {
            try {
                const offset = qubx_ui.eventOffset(event, event.center.x, event.center.y);
                var x0, x1;
                if ( event.deltaX < 0 ) {
                    x0 = offset.x;
                    x1 = x0 - event.deltaX;
                } else {
                    x1 = offset.x;
                    x0 = x1 - event.deltaX;
                }
                this.zoom(this.dpRatio*x0/this.w, this.dpRatio*x1/this.w);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onDoubleTap(event) {
            try {
                this.zoom(0, 1);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onDoubleTapHires(event) {
            try {
                const h = this.hilite();
                if ( h.count ) {
                    this.zoom(h.from, h.to);
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onPanHires(event) {
            try {
                if ( this._hiPinch ) {
                    return;
                }
                if ( ! this._hiPan ) {
                    this._hiPan = {
                        selFrom: this.selFrom,
                        selTo: this.selTo,
                        pw: this.selTo - this.selFrom
                    };
                }
                if ( this.baselining && this.baselineSel ) {
                    // could try moving the node up and down?
                } else if ( event.srcEvent.shiftKey ) {
                    const rq = - this._hiPan.pw * event.deltaX * this.dpRatio / this.wHi,
                          d = Math.max(-this._hiPan.selFrom, Math.min(1-this._hiPan.selTo, rq));
                    this.zoom(this._hiPan.selFrom+d, this._hiPan.selTo+d, true);
                } else {
                    const offset = qubx_ui.eventOffset(event, event.center.x, event.center.y),
                          pw = this.selTo - this.selFrom;
                    var x0, x1;
                    if ( event.deltaX < 0 ) {
                        x0 = offset.x;
                        x1 = x0 - event.deltaX;
                    } else {
                        x1 = offset.x;
                        x0 = x1 - event.deltaX;
                    }
                    this.hilite(this.selFrom + pw * this.dpRatio * x0/this.w,
                                this.selFrom + pw * this.dpRatio * x1/this.w);
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onPanendHires(event) {
            try {
                this._hiPan = null;
                this._onMouseMoveHires(event.srcEvent, 1); // update cursor
                if ( this.baselining && (this._chosen >= 0) ) {
                    const hilite = this.hilite(),
                          center = hilite.offset + (Math.round(hilite.count/2)|0),
                          seg = this._data.segments[this._chosen],
                          mean = qubx.meanstd(seg.sampled[0], hilite.offset, hilite.count).mean;
                    this._data.setBaseline(this._chosen, b => {
                        b.setNodes([center], [mean]);
                        this.baselineSel = {
                            ix: center,
                            level: mean,
                            t: center * this._data.sampling()
                        };
                        this.ev.pickNode.trigger(this.baselineSel);
                    });
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onPinchHires(event) {
            try {
                const touches = event.srcEvent.touches;
                if ( touches && touches.length !== 2 ) {
                    return;
                }
                const offset = qubx_ui.eventOffset(event, event.center.x, event.center.y);
                if ( ! this._hiPinch ) {
                    this._hiPinch = {
                        x: offset.x * this.dpRatio,
                        scale: 1.0,
                    };
                } else {
                    const rq = (this.selTo - this.selFrom) * (this._hiPinch.x - offset.x*this.dpRatio) / this.wHi,
                          d = Math.max(-this.selFrom, Math.min(1-this.selTo, rq));
                    this.zoom(this.selFrom+d, this.selTo+d, true);
                    this._hiPinch.x = offset.x*this.dpRatio;
                    
                    const center = this.selFrom + (this.selTo - this.selFrom)*offset.x*this.dpRatio/this.wHi,
                          scaled = qubtrace.pinchInterval(this.selFrom, this.selTo, center, event.scale/this._hiPinch.scale);
                    this.zoom(scaled.from, scaled.to, true);
                    this._hiPinch.scale = event.scale;
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onPinchendHires(event) {
            setTimeout(() => { // avoid misinterpreting tail end of pinch as a pan
                this._hiPinch = null;
            }, 100);
        }
        _onWheelHires(event) {
            try {
                const offset = qubx_ui.eventOffset(event),
                      scale = Math.pow(wheelDeltaGrow[event.deltaMode] || (1 + .3/this.hiresLayer.clientHeight), - event.deltaY),
                      center = this.selFrom + (this.selTo - this.selFrom)*offset.x*this.dpRatio/this.wHi,
                      scaled = qubtrace.pinchInterval(this.selFrom, this.selTo, center, scale);
                this.zoom(scaled.from, scaled.to, true);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onMouseDownHires(event) {
            this._onMouseMoveHires(event, 2);
            this.baselining && (this._chosen >= 0) && this._pickNode(qubx_ui.eventOffset(event));
        }
        _onTouchStartHires(event) {
            this._onMouseMoveHires(event, 2);
            this.baselining && (this._chosen >= 0) && this._pickNode(qubx_ui.eventOffset(event, event.center.x, event.center.y));
        }
        _pickNode(pos) {
            // find node in [pos.x-emsize, pos.x+emsize] which minimizes distance(pos, node); distance <= emsize
            // this.baselineSel = {ix, t, level} or null
            this._data.setBaseline(this._chosen); // ensure existence
            const emSamples = (this.sppHi||1) * this.emsize,
                  bl = this._data.segments[this._chosen].baseline,
                  pCenter = this.selFrom + (this.selTo - this.selFrom)*this.dpRatio*pos.x/this.wHi,
                  iCenter = Math.round(this.maxNpoint*pCenter),
                  tCenter = iCenter * this._data.sampling(),
                  layout = this.layoutHi.current,
                  vCenter = this.bounds[0].lo + (this.bounds[0].dy / layout.height)*(layout.height - (this.dpRatio*pos.y - layout.top)),
                  nodes = bl.getNodes(Math.max(0, iCenter-emSamples), Math.min(this.maxNpoint-1, iCenter+emSamples));
            let iBest, dBest = 1e20;
            nodes.indices.forEach((ix, i) => {
                const dPixH = (ix - iCenter) / this.sppHi,
                      dPixV = (nodes.levels[i] - vCenter) * (layout.height / this.bounds[0].dy),
                      d2 = dPixH*dPixH + dPixV*dPixV;
                if ( d2 < dBest ) {
                    iBest = i;
                    dBest = d2;
                }
            });
            if ( dBest <= (this.emsize*this.emsize) ) {
                this.pickNode({
                    ix: nodes.indices[iBest],
                    level: nodes.levels[iBest],
                    t: nodes.indices[iBest] * this._data.sampling()
                });
            } else {
                this.pickNode(null);
            }
        }
        pickNode(node) {
            if ( (node || (node === 0)) && (typeof(node.level) === 'undefined') ) {
                const bl = this._data.segments[this._chosen].baseline;
                node = {
                    ix: node,
                    level: bl.levels[bl.idl.classAt(node)],
                    t: node*this._data.sampling()
                };
            }
            this.baselineSel = node;
            this.ev.pickNode.trigger(this.baselineSel);
            this.drawHiresSoon();
            return this.baselineSel;
        }
        _onMouseMoveHires(event, moving) {
            try {
                if ( typeof(moving) === 'undefined' ) {
                    moving = event.shiftKey ? (this.hiMoving || 1) : 0;
                }
                if ( moving !== this.hiMoving ) {
                    this.hiMoving = moving;
                    this.hiSelCanvas.style.cursor = '';
                    if ( moving ) {
                        const ctype = (moving === 1) ? 'grab' : 'grabbing';
                        this.hiSelCanvas.style.cursor = ctype;
                        if ( ! this.hiSelCanvas.style.cursor ) {
                            this.hiSelCanvas.style.cursor = '-webkit-'+ctype;
                        }
                        if ( ! this.hiSelCanvas.style.cursor ) {
                            this.hiSelCanvas.style.cursor = '-moz-'+ctype;
                        }
                    }
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        getComposite() {
            const w = this.canvas.width,
                  hLo = this.canvas.height,
                  hHi = this.hiresCanvas.height,
                  h = hLo + hHi;
            
            if ( ! this._composite ) {
                this._composite = document.createElement('canvas');
                this._composite.width = w;
                this._composite.height = h;
                this._composite.style.position = 'fixed';
                this._composite.style.visibility = 'hidden';
                document.body.appendChild(this._composite);
            }
            if ( (this._composite.width != w) || (this._composite.height != h) ) {
                this._composite.width = w;
                this._composite.height = h;
            }
            
            const ctx = this._composite.getContext('2d');
            ctx.drawImage(this.canvas, 0, 0);
            if ( this.config.scalebarLores && this.scaleCanvas ) {
                ctx.drawImage(this.scaleCanvas, (this.dpRatio * parseInt(this.scaleCanvas.style.left))|0, (this.dpRatio * parseInt(this.scaleCanvas.style.top))|0);
            }
            ctx.drawImage(this.hiresCanvas, 0, hLo);
            ctx.drawImage(this.fitCanvas, 0, hLo);
            if ( this.config.scalebarHires && this.hiresScaleCanvas ) {
                ctx.drawImage(this.hiresScaleCanvas, (this.dpRatio * parseInt(this.hiresScaleCanvas.style.left))|0, (hLo+this.dpRatio*parseInt(this.hiresScaleCanvas.style.top))|0);
            }
            return this._composite;
        }
    };

    exports.PlotFrame = class PlotFrame {
        constructor(config) {
            this.node = config && config.node || document.createElement('div');
            this.view = document.createElement('div');
            this.view.className = 'qubx_ui_data_view';
            this.node.appendChild(this.view);
            const lores = document.createElement('div');
            lores.className = 'qubx_ui_data_lo';
            this.view.appendChild(lores);
            const hires = document.createElement('div');
            hires.className = 'qubx_ui_data_hi';
            this.view.appendChild(hires);
            this.plot = new exports.Plot({
                node: lores,
                hires: hires
            });
            this.data = null;
            this.plot.ev.changeData.sub(data => {
                this.data = data;
            });
            this.spinner = new qubx_ui.Spinner(this.node);
            this._resizeHandle = qubx_ui.onResize(this.node, info => this.plot.reshape());
        }
        dispose() {
            this.plot.data(null);
            qubx_ui.clearResize(this._resizeHandle);
        }
        setData(data) {
            this.plot.data(data);
        }
        busy(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._busy;
            } else if ( this._busy != x ) {
                if ( x ) {
                    this.spinner.show();
                } else {
                    this.spinner.hide();
                }
            }
            return this;
        }
    };

    exports.ConstantsTable = class ConstantsTable {
        constructor(config) {
            this.node = config.node || document.createElement('table');
            this.node.classList.add('qubx_data_consts');
            this.plot = config.plot;
            this.stimulus = {};
            this.rows = [];
            this.tbl = new qubx_ui.ColumnTable({
                node: this.node,
                columns: [
                    qubx_ui.ColumnTable.txtCol('Constants:', 'Name'),
                    qubx_ui.ColumnTable.inputCol('', 'Value', 'number', qubx.acceptFloat, qubx.formatK0, this._onChangeConst.bind(this))
                ]
            });
            this.plot.ev.changeData.sub(this._onChangeData.bind(this));
            this._boundOnSet = this._onSetConst.bind(this);
            this._boundOnInserted = this._onInsertedConst.bind(this);
            this._onChangeData(this.plot.data());
        }
        _onChangeData(data) {
            if ( this._data ) {
                this._data.constants.ev.set.unsub(this._boundOnSet);
                this._data.constants.ev.inserted.unsub(this._boundOnInserted);
            }
            this._data = data;
            if ( data ) {
                data.constants.ev.set.sub(this._boundOnSet);
                data.constants.ev.inserted.sub(this._boundOnInserted);
                this.layout();
            }
        }
        _onChangeConst(row, col, val) {
            this._data.constants.setByName(row.Name, val);
        }
        _onSetConst(info) {
            if ( info.column === 'Value' ) {
                this.layout();
            }
        }
        _onInsertedConst(c) {
            this.layout();
        }
        layout(stimulus) {
            this.stimulus = stimulus = stimulus || this.stimulus;
            const modelConst = [],
                  signalNames = this._data.signals.map(qubx.m('name'));
            for (const name of Object.keys(stimulus)) {
                if ( signalNames.indexOf(name) < 0 ) {
                    modelConst.push(name);
                }
            }
            const constants = this._data.constants;
            this.rows = modelConst.map(name => ({
                Name: name,
                Value: constants.getByName(name) || 0.0
            }));
            this.tbl.render(this.rows, 'Name');
        }
    };
    
    const CursorInfoDefs = {
        format: qubx.formatK0
    };
    exports.CursorInfo = class CursorInfo {
        constructor(config) {
            this.config = Object.assign({}, CursorInfoDefs, config);
            this.node = config.node || document.createElement('div');
            this.node.classList.add('qubx_cursor_info');
            this.plot = config.plot;
            this.nodeSel = d3.select(this.node).append('div');
            this.nodeMouse = d3.select(this.node).append('div');
            this.nodeT = d3.select(this.node).append('div');
            this.nodeGeneral = d3.select(this.node).append('div');
            this.nodeGeneral.style('width', '100%');
            d3.select(this.node).selectAll('div').each(function(d,i) {
                this.className = 'qubx_cursinfo';
            });
            let line = this.nodeSel.append('div');
            this.sel = {};
            this.sel.t0 = line.append('span');
            line.append('span').html('&ndash;');
            this.sel.t1 = line.append('span');
            line.append('span').text(' s (pts ');
            this.sel.p0 = line.append('span');
            line.append('span').html('&ndash;');
            this.sel.p1 = line.append('span');
            line.append('span').text(')');
            this.sel.nodeStim = this.nodeSel.append('div');
            line = this.nodeSel.append('div');
            line.append('span').text('Dur: ');
            this.sel.dur = line.append('span');
            line.append('span').text(' s, ');
            this.sel.durPts = line.append('span');
            line.append('span').text(' pts');
            line = this.nodeSel.append('div');
            line.append('span').text('I: ');
            this.sel.Imean = line.append('span');
            line.append('span').html(' &#xb1; ');
            this.sel.Istd = line.append('span');
            ['min', 'max'].forEach(stat => {
                line = this.nodeSel.append('div');
                line.append('span').text('I '+stat+': ');
                this.sel['I'+stat] = line.append('span');
                line.append('span').text(' at ');
                this.sel['T'+stat] = line.append('span');
                line.append('span').text(' sec');
            });
            this.updateSoon = qubx.batchDelayed(this._update.bind(this), 10);
            this.updateSelSoon = qubx.batchDelayed(this._updateSel.bind(this), 250);
            this.updateGeneralSoon = qubx.batchDelayed(this._updateGeneral.bind(this), 500);
            this.plot.ev.select.sub(this.updateSelSoon);
            this.plot.ev.hilite.sub(this.updateSelSoon);
            this.plot.ev.choose.sub(this.updateSelSoon);
            this.plot.selCanvas.addEventListener('mousemove', this._onMouse.bind(this), true);
            this.plot.selCanvas.addEventListener('mouseleave', this._onMouseLeave.bind(this), true);
            this.plot.hiSelCanvas.addEventListener('mousemove', this._onMouseHi.bind(this), true);
            this.plot.hiSelCanvas.addEventListener('mouseleave', this._onMouseLeaveHi.bind(this), true);
            this.plot.ev.changeData.sub(this._onChangeData.bind(this));
            this._onChangeData(this.plot.data());
        }
        _onChangeData(data) {
            if ( this._data ) {
                this._data.ev.changeSampling.unsub(this.updateGeneralSoon);
                this._data.ev.doneSegSize.unsub(this.updateGeneralSoon);
                this._data.ev.changeSamples.unsub(this.updateSoon);
            }
            this._data = data;
            if ( data ) {
                this._data.ev.changeSampling.sub(this.updateGeneralSoon);
                this._data.ev.doneSegSize.sub(this.updateGeneralSoon);
                this._data.ev.changeSamples.sub(this.updateSoon);
                this.updateSoon();
                this.updateGeneralSoon();
            }
        }
        _onMouse(event) {
            this.mouse = qubx_ui.eventOffset(event);
            this.updateSoon();
        }
        _onMouseLeave() {
            this.mouse = null;
            this.updateSoon();
        }
        _onMouseHi(event) {
            this.mouse = qubx_ui.eventOffset(event);
            this.mouse.hires = true;
            this.updateSoon();
        }
        _onMouseLeaveHi() {
            this.mouse = null;
            this.updateSoon();
        }
        _update() {
            try {
                const data = this._data;
                if ( ! (this.mouse && data) ) {
                    [this.nodeMouse, this.nodeT].forEach(node => node.html('').style('display', 'none'));
                    this.nodeMouse.html('');
                    this.nodeT.html('');
                    return;
                }
                var proportion;
                if ( this.mouse.hires ) {
                    proportion = this.plot.selFrom + (this.plot.selTo - this.plot.selFrom)*this.plot.dpRatio*this.mouse.x/this.plot.wHi;
                } else {
                    proportion = this.plot.dpRatio*this.mouse.x/this.plot.w;
                }
                const p = Math.round(this.plot.maxNpoint*proportion)|0,
                      t = p*data.sampling();
                
                const iSeg = Math.max(this.plot.chosen(), 0),
                      seg = data.segments[iSeg],
                      signals = data.signals.map((sig, iSig) => {
                          const layouts = this.mouse.hires ? this.plot.layoutHi : this.plot.layoutLo,
                                layout = iSig ? layouts.stimulus[iSig-1] : layouts.current;
                          if ( ! layout ) {
                              return {
                                  lbl: '',
                                  v: [0, 0]
                              };
                          }
                          const h = layout.height,
                                vMouse = this.plot.bounds[iSig].lo + this.plot.bounds[iSig].dy*(h - (this.plot.dpRatio*this.mouse.y - layout.top))/h;
                          let vT = seg.sampled[iSig][p];
                          if ( iSig && seg.ideal[iSig] ) {
                              const stimCls = seg.ideal[iSig].classAt(p);
                              if ( stimCls >= 0 ) {
                                  vT = seg.ideal[iSig].amp[stimCls];
                              }
                          }
                          const name = iSig ? sig.name : 'I';
                          return {
                              lbl: [name+'(<span style="font-size: 150%">&#x2316;</span>):', name+'(t):'],
                              v: [vMouse, vT]
                          };
                      });
                signals.splice(0, 0, {
                    lbl: ['t [sec]:', 't [sec]:'],
                    v: [t, t]
                });
                
                const format = this.config.format;
                [this.nodeMouse, this.nodeT].forEach((node, iNode) => {
                    let tbl = node.select('table');
                    if ( tbl.empty() ) {
                        tbl = node.append('table');
                    }
                    const rows = tbl.selectAll('tr').data(signals, function(d,i) { return d.lbl[iNode]; }),
                          row = rows.enter().append('tr');
                    row.append('td').html(function(d,i) {
                        return d.lbl[iNode];
                    });
                    row.append('td').classed('qubx_cursinfo_val', true);
                    rows.exit().remove();
                    rows.select('.qubx_cursinfo_val').text(function(d,i) {
                        return format(d.v[iNode]);
                    });
                    node.style('display', '');
                });
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _updateSel() { // of highlighted data
            try {
                const hilite = this.plot.hilite();
                this.plot.reSelect((selOffset, selCount, segments) => {
                    const chosen = this.plot.chosen();
                    if ( chosen >= 0 ) {
                        segments = [segments[chosen]];
                    }
                    if ( hilite.count && hilite.offset ) {
                        selOffset = hilite.offset;
                        selCount = hilite.count;
                        segments.forEach(seg => {
                            seg.selFrom = Math.max(seg.selFrom, hilite.offset);
                            seg.selTo = Math.min(seg.selTo, hilite.offset + hilite.count - 1);
                        });
                    } else {
                        this.sel.nodeStim.html('');
                        this.nodeSel.style('display', 'none');
                        return;
                    }
                    const data = this._data,
                          dt = data.sampling(),
                          format = this.config.format;
                    this.sel.t0.text(format(dt*selOffset));
                    this.sel.t1.text(format(dt*(selOffset + selCount - 1)));
                    this.sel.p0.text(selOffset);
                    this.sel.p1.text(selOffset + selCount - 1);
                    this.sel.dur.text(format(dt*selCount));
                    this.sel.durPts.text(selCount);
                    const segmentsI = segments.map(seg => seg.sources[0]),
                          stats = qubx.meanstd(segmentsI, selOffset, selCount);
                    this.sel.Imean.text(format(stats.mean));
                    this.sel.Istd.text(format(stats.std));
                    const ex = qubx.extremas(segmentsI, selOffset, selCount);
                    this.sel.Imin.text(format(ex.lo));
                    this.sel.Tmin.text(format(dt * ex.iLo)); /// 'in segment (?)'
                    this.sel.Imax.text(format(ex.hi));
                    this.sel.Tmax.text(format(dt * ex.iHi)); /// 'in segment (?)'
                    
                    /// number of segments (chosen)?
                    
                    const signals = data.signals.slice(1).map((sig, iSig) => {
                        /// should prefer ideal[iSig+1] sampled/rastered to 1 point
                        const segmentsS = segments.map(seg => seg.sources[iSig+1]),
                              ex = qubx.extremas(segmentsS, selOffset, selCount),
                              info = (ex.lo < ex.hi) ? (format(ex.lo)+'&ndash;'+format(ex.hi)) : format(ex.lo);
                        return {
                            lbl: (sig.units ? (sig.name + ' ['+sig.units+']') : sig.name)+':',
                            info: info
                        };
                    });
                    let tbl = this.sel.nodeStim.select('table');
                    if ( tbl.empty() ) {
                        tbl = this.sel.nodeStim.append('table');
                    }
                    const rows = tbl.selectAll('tr').data(signals, qubx.m('lbl')),
                          row = rows.enter().append('tr');
                    row.append('td').html(qubx.m('lbl'));
                    row.append('td').classed('qubx_cursinfo_val', true);
                    rows.exit().remove();
                    rows.select('.qubx_cursinfo_val').html(qubx.m('info'));
                    this.nodeSel.style('display', '');
                });
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _updateGeneral() {
            try {
                this.nodeGeneral.html('');
                if ( ! this._data ) {
                    return;
                }
                const data = this._data,
                      format = this.config.format;
                let Ntotal = 0;
                data.segments.forEach(seg => {
                    Ntotal += seg.n;
                });
                const Nseg = data.segments.length;
                this.nodeGeneral.append('div')
                    .html(Nseg+' segment'+((Nseg!==1)?'s':'')+' at '+format(1e-3/data.sampling())+ ' kHz');
                this.nodeGeneral.append('div')
                    .html('Samples: '+Ntotal);
                this.nodeGeneral.append('div')
                    .html('Time: '+format(data.sampling()*Ntotal)+' sec');
            } catch (err) {
                qubx.reportErr(err);
            }
        }
    };

    exports.drawIdl = function drawIdl(ctx, px0, pxw, pxh, events, first, amp, std, idlStyle, lineWidth, spp, yLo, ypp) {
        ctx.save();
        ctx.translate(0, pxh);
        ctx.scale(1, -1/ypp);
        ctx.translate(0, -yLo);
        ctx.strokeStyle = idlStyle;
        const pps = 1/spp,
            n = events.classes.length;
        let l = -2,
            a_l = 0.0,
            s_l = 0.0,
            c_l = -2,
            i = 0;
        for (; i<n; ++i) {
            const f = events.firsts[i],
                  c = events.classes[i],
                  x0 = pps * (f - first),
                  a = amp[c],
                  s = Math.max(2*std[c], ypp*lineWidth);
            if ( (c !== c_l) && (f === (l + 1)) ) {
                ctx.lineWidth = lineWidth;
                ctx.beginPath();
                ctx.moveTo(x0, (a_l < a) ? (a_l + s_l/2) : (a_l - s_l/2));
                ctx.lineTo(x0, (a_l < a) ? (a - s/2) : (a + s/2));
                ctx.stroke();
            }
            l = events.lasts[i];
            a_l = a;
            s_l = s;
            c_l = c;
            ctx.lineWidth = s;
            ctx.beginPath();
            ctx.moveTo(x0-lineWidth/2, a);
            ctx.lineTo(pps * (l - first + 1) + lineWidth/2, a);
            ctx.stroke();
        }
        ctx.restore();
    };

    exports.drawSampledIdl = function drawSampledIdl(ctx, px0, pxw, pxh, lows, highs, idlStyle, yLo, ypp) {
        const UNSET = qubx.UNSET;
        let exists = false,
            i = 0;
        for (; i<pxw; ++i) {
            if ( lows[i] !== UNSET ) {
                exists = true;
                break;
            }
        }
        if ( ! exists ) {
            return;
        }
        ctx.save();
        ctx.translate(0.5, pxh);
        ctx.scale(1.0, -1/ypp);
        ctx.translate(0.0, -yLo);
        ctx.strokeStyle = idlStyle;
        ctx.lineWidth = 1;
        ctx.beginPath();
        for (i=0; i<pxw; ++i) {
            const lo = lows[i];
            if ( lo !== UNSET ) {
                ctx.moveTo(px0+i, lo);
                ctx.lineTo(px0+i, highs[i]);
            }
        }
        ctx.stroke();
        ctx.restore();
    };

    exports.Converter = class Converter {
        constructor(config) {
            this.config = Object.assign({}, config);
            this.progress = new qubx_ui.ModalProgress({
                oncancel: this._onCancel.bind(this)
            });
        }
        receiveFile(file) {
            const srcName = file.name || file.url || '',
                  lwr = srcName.toLowerCase(); 
            if ( ! lwr.length ) {
                return Promise.reject('Empty');
            }
            let ext = '',
                iDot = lwr.lastIndexOf('.');
            if ( iDot > 0 ) {
                ext = lwr.substring(iDot);
            }
            if ( ['.qdf', '.qsf', '.txt', '.csv', '.atf', '.axgx', '.axgr'].indexOf(ext) >= 0 ) {
                return file.get();
            }
            const statusHead = 'Using a web service to read '+(file.name || file.url)+'...\n';
            this.progress.show(statusHead+'Uploading...', 0.0);
            var putUrl;
            const cleanup = () => {
                this.progress.hide();
                putUrl && cloud.releaseFile(putUrl);
                this._request = null;
            };
            const step1 = file.url ? Promise.resolve(true) : file.get().then(blob => {
                return cloud.canPutFile(file.name, blob.size).then(upload => new Promise((resolve, reject) => {
                    const request = new XMLHttpRequest();
                    this._request = request;
                    request.open('POST', '/online/data/put', true);
                    request.responseType = 'json';
                    request.onload = () => resolve(request.response);
                    request.onerror = reject;
                    const form = new FormData();
                    form.append('id', upload.id);
                    form.append('file', blob);
                    request.send(form);
                }));
            }).then(reply => {
                file.url = reply.url;
                putUrl = reply.url;
            });
            return step1.then(() => {
                this._request = null;
                this.progress.status(statusHead+'Converting...');
                this.progress.progress(.25);
                return cloud.convertFile(file.url);
            }).then(reply => {
                this.progress.status(statusHead+'Downloading...');
                return qubx.get(reply.url, 'blob', p => {
                    this.progress.progress(.5+.5*p);
                }).catch (err => {
                        var complaint = err.message || '';
                        if ( complaint.indexOf('Network Error') >= 0 ) {
                            complaint = 'The file could not be read.  If the URL is not in error, perhaps the host is not configured to allow cross-origin requests.';
                        }
                        setTimeout(() => alert(complaint), 10);
                        return Promise.reject(err);
                    });
            }).then(blob => {
                blob.name = srcName.substring(Math.max(0, srcName.lastIndexOf('/')), srcName.length-ext.length)+'.qdf';
                cleanup();
                return blob;
            }).catch(err => {
                cleanup();
                return Promise.reject(err);
            });
        }
        _onCancel() {
            this._request && this._request.abort();
            cloud.cancelConvertFile();
        }
    };
    
    return exports;
});

