/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_ui_model", ["qubx", "qubx_lib", "qubx_model", "qubx_model_math", "qubx_ui", "snap.svg", "d3"], (qubx, qubx_lib, qubx_model, qubx_model_math, qubx_ui, snap, d3) => {
    'use strict';

    const exports = {};

    const modelStyle = qubx.get("css/qubx_model.css");
    
    const STATE_WIDTH2 = 5; // half-width
    const LONG_PRESS_MS = 700;
    const SNAP_TO = 2.5;
    
    function distance(x0, y0, x1, y1) {
        const dx = x1 - x0;
        const dy = y1 - y0;
        return Math.sqrt(dx*dx + dy*dy);
    }

    function snap_to_grid(x, d) {
        return d * Math.round(x/d);
    }

    const ContextMethods = {
        init: function(view, node) {
            this.view = view;
            this.node = node || document.createElement('div');
            this.node.classList.add('qubx_model_context');
            this.ev = qubx.makeEvents(['unselect', 'selectModel', 'selectRate', 'selectState']);

            this.stateProp = qubx_ui.loadTemplate(".qubx_state_props").then(t => {
                const dotdot = qubx_ui.querydotdot(t);
                const stateProp = {
                    node: t,
                    Class: dotdot.sub("#Class", input => new qubx_ui.AcceptedInput(input, 0, qubx_model.stateFields.Class),
                                      'change', this._enteredClass.bind(this)),
                    Label: dotdot.sub("#Label", input => new qubx_ui.AcceptedInput(input, "", qubx_model.stateFields.Label),
                                      'change', this._enteredLabel.bind(this)),
                    units: dotdot(".qubx_class_amp_units"),
                    amp: dotdot.sub("#amp", input => new qubx_ui.AcceptedInput(input, 0.0, qubx_model.classFields.Amp),
                                    'change', this._enteredAmp.bind(this)),
                    std: dotdot.sub("#std", input => new qubx_ui.AcceptedInput(input, 0.0, qubx_model.classFields.Std),
                                    'change', this._enteredStd.bind(this)),
                    Peq: dotdot(".qubx_state_Peq"),
                    TimeConstant: dotdot(".qubx_state_TimeConstant"),
                    calcAmp: dotdot(".qubx_class_calcAmp"),
                    calcStd: dotdot(".qubx_class_calcStd"),
                    grab: dotdot("#qubx_class_grab"),
                    grabAll: dotdot("#qubx_class_grab_all"),
                    del: dotdot(".qubx_state_del")
                };
                this.node.appendChild(t);
                t.style.display = 'none';
                stateProp.grab.addEventListener('click', () => {
                    try {
                        stateProp.grab.disabled = true;
                        const c = this._selRow.Class;
                        this.view.config.grabAmp()
                            .then(stats => {
                                this.model.classes.set(c, 'Amp', stats.mean);
                                this.model.classes.set(c, 'Std', stats.std);
                                stateProp.amp.update(stats.mean);
                                stateProp.std.update(stats.std);
                                stateProp.grab.disabled = false;
                            })
                            .catch(qubx.reportErr)
                            .catch(err => {
                                stateProp.grab.disabled = false;
                                return Promise.reject(err);
                            });
                    } catch (err) {
                        qubx.reportErr(err);
                    }
                });
                stateProp.grabAll.addEventListener('click', () => {
                    try {
                        stateProp.grabAll.disabled = true;
                        const c = this._selRow.Class;
                        this.view.config.grabAmps(this.model).then(result => {
                            const n = result.amp.length;
                            let i = 0;
                            for (; i<n; ++i) {
                                this.model.classes.set(i, 'Amp', result.amp[i]);
                                this.model.classes.set(i, 'Std', result.std[i]);
                            }
                            stateProp.amp.update(result.amp[c]);
                            stateProp.std.update(result.std[c]);
                            stateProp.grabAll.disabled = false;
                        }).catch(qubx.reportErr)
                            .catch(err => {
                                stateProp.grabAll.disabled = false;
                                return Promise.reject(err);
                            });
                    } catch (err) {
                        qubx.reportErr(err);
                    }
                });
                stateProp.del.addEventListener("click", event => {
                    try {
                        this.modelRec.deleteState(this._selRow.Index);
                    } catch (err) {
                        qubx.reportErr(err);
                    }
                });
                return stateProp;
            }).catch(qubx.reportErr);
            
            this.rateProp = qubx_ui.loadTemplate(".qubx_rate_props").then(t => {
                const dotdot = qubx_ui.querydotdot(t);
                const rateProp = {
                    node: t,
                    From: dotdot(".qubx_rate_from"),
                    To: dotdot(".qubx_rate_to"),
                    K: dotdot(".qubx_rate_k"),
                    k0: dotdot.sub("#k0", input => new qubx_ui.AcceptedInput(input, 1.0, qubx_model.rateFields.k0),
                                   'change', this._enteredK0.bind(this)),
                    k1: dotdot.sub("#k1", input => new qubx_ui.AcceptedInput(input, 1.0, qubx_model.rateFields.k1),
                                   'change', this._enteredK1.bind(this)),
                    k2: dotdot.sub("#k2", input => new qubx_ui.AcceptedInput(input, 1.0, qubx_model.rateFields.k2),
                                   'change', this._enteredK2.bind(this)),
                    Ligand: dotdot.sub("#ligand", input => new qubx_ui.AcceptedInput(input, "", qubx_model.rateFields.Ligand),
                                       'change', this._enteredLigand.bind(this)),
                    noLigand: dotdot.listen('#noLigand', 'click', () => {
                        rateProp.Ligand.update('');
                        this._enteredLigand('');
                    }),
                    LigandSuggest: new qubx_ui.Popup({
                        node: dotdot("#ligand_suggest"),
                        getOptions: () => {
                            const notLigand = ['Voltage', 'Pressure'];
                            return this.view.config.listStimuli()
                                .filter(name => notLigand.indexOf(name) < 0);
                        },
                        onChoose: name => {
                            rateProp.Ligand.update(name);
                            this._enteredLigand(name);
                        }
                    }),
                    Voltage: dotdot.listen("#voltage", 'change', this._enteredVoltage.bind(this)),
                    Pressure: dotdot.listen("#pressure", 'change', this._enteredPressure.bind(this)),
                    del: dotdot.listen('.qubx_rate_del', 'click', event => {
                        try {
                            this.modelRec.deleteRate(this._selRow.Index);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    }),
                };
                this.node.appendChild(t);
                t.style.display = 'none';
                return rateProp;
            }).catch(qubx.reportErr);
            
            this.modelProp = qubx_ui.loadTemplate(".qubx_model_props").then(t => {
                const dotdot = qubx_ui.querydotdot(t);
                const modelProp = {
                    node: t,
                    name: dotdot("#modelName"),
                    active: dotdot.listen("#modelActive", 'change', this._enteredActive.bind(this)),
                    channelCount: dotdot.sub('#channelCount', input => new qubx_ui.AcceptedInput(input, 1, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        defval: 1,
                        widthChars: 5
                    }), 'change', this._enteredChannelCount.bind(this)),
                    IeqFv: dotdot.listen("#IeqFv", 'change', this._enteredIeqFv.bind(this)),
                    Vrev: dotdot.sub("#Vrev", input => new qubx_ui.AcceptedInput(input, 0.0, {
                        accept: qubx.acceptFloat,
                        format: qubx.formatStr,
                        defval: 0.0,
                        widthChars: 4
                    }), 'change', this._enteredVrev.bind(this))
                };
                this.node.appendChild(t);
                t.style.display = 'none';
                return modelProp;
            }).catch(qubx.reportErr);
            this._boundSetChannelCount = this._onSetChannelCount.bind(this);
            this._boundSetIeqFv = this._onSetIeqFv.bind(this);
            this._boundSetVRev = this._onSetVRev.bind(this);
            this._boundInsertedState = this._onInsertedState.bind(this);
            this._boundRemovedState = this._onRemovedState.bind(this);
            this._boundSetState = this._onSetState.bind(this);
            this._boundInsertedRate = this._onInsertedRate.bind(this);
            this._boundRemovingRate = this._onRemovingRate.bind(this);
            this._boundSetRate = this._onSetRate.bind(this);
            this._boundQ = this._onQ.bind(this);
            this._boundPeq = this._onPeq.bind(this);
            this._boundAmps = this._onAmps.bind(this);

            return this;
        },
        unselect: function(nextSel) {
            if ( ! this.model ) {
                return;
            }
            if ( this._selRow ) {
                if ( 'Class' in this._selRow ) {
                    this.stateProp.then(stateProp => { stateProp.node.style.display = 'none'; });
                } else {
                    this.rateProp.then(rateProp => { rateProp.node.style.display = 'none'; });
                }
            }
            this._selRow = null;
            if ( this.modelRec !== nextSel ) {
                this.model.ev.setChannelCount.unsub(this._boundSetChannelCount);
                this.model.ev.setIeqFv.unsub(this._boundSetIeqFv);
                this.model.ev.setVRev.unsub(this._boundSetVRev);
                this.model.states.ev.inserted.unsub(this._boundInsertedState);
                this.model.states.ev.removed.unsub(this._boundRemovedState);
                this.model.states.ev.set.unsub(this._boundSetState);
                this.model.rates.ev.inserted.unsub(this._boundInsertedRate);
                this.model.rates.ev.removing.unsub(this._boundRemovingRate);
                this.model.rates.ev.set.unsub(this._boundSetRate);
                this.modelRec.evCalc.Q.unsub(this._boundQ);
                this.modelRec.evCalc.Peq.unsub(this._boundPeq);
                this.modelRec.evCalc.amps.unsub(this._boundAmps);
                this.modelProp.then(modelProp => {
                    modelProp.node.style.display = 'none';
                });
                this.ev.unselect.trigger();
                this.modelRec = null;
                this.model = null;
            }
        },
        select: function(modelRec) {
            if ( typeof(modelRec) === 'undefined' ) {
                modelRec = this.modelRec;
            }
            this.unselect(modelRec);
            if ( this.modelRec !== modelRec ) {
                this.modelRec = modelRec;
                this.model = modelRec.model;
                this.model.ev.setChannelCount.sub(this._boundSetChannelCount);
                this.model.ev.setIeqFv.sub(this._boundSetIeqFv);
                this.model.ev.setVRev.sub(this._boundSetVRev);
                this.model.states.ev.inserted.sub(this._boundInsertedState);
                this.model.states.ev.removed.sub(this._boundRemovedState);
                this.model.states.ev.set.sub(this._boundSetState);
                this.model.rates.ev.inserted.sub(this._boundInsertedRate);
                this.model.rates.ev.removing.sub(this._boundRemovingRate);
                this.model.rates.ev.set.sub(this._boundSetRate);
                modelRec.evCalc.Q.sub(this._boundQ);
                modelRec.evCalc.Peq.sub(this._boundPeq);
                modelRec.evCalc.amps.sub(this._boundAmps);
                this.modelProp.then(modelProp => {
                    modelProp.name.textContent = this.model.modelName||'Untitled';
                    modelProp.active.checked = this.model.active;
                    modelProp.channelCount.update(this.model.channelCount());
                    modelProp.IeqFv.checked = this.model.IeqFv();
                    modelProp.Vrev.update(this.model.vRev());
                    modelProp.node.style.display = '';
                });
            }
            modelRec && this.ev.selectModel.trigger(modelRec);
        },
        onActivate: function(rec, i, active) {
            this.modelProp.then(modelProp => {
                modelProp.active.checked = active;
            });
        },
        updateState: function() {
            this._selRow && this.stateProp.then(stateProp => {
                if ( (! this._selRow) || ! ('Class' in this._selRow) ) {
                    return;
                }
                const state = this._selRow;
                const i = this._selRow.Index;
                const cls = this.model.classes.get(state.Class);
                stateProp.Class.update(cls.Index);
                stateProp.Class.node.className = "qubx_opt_class qubx_classx qubx_class"+cls.Index;
                stateProp.Label.update(state.Label);
                stateProp.amp.update(this.model.IeqFv() ? cls.Cond : cls.Amp);
                stateProp.std.update(this.model.IeqFv() ? cls.CondStd : cls.Std);
                stateProp.Peq.textContent = qubx.formatK0(this.modelRec.stateCalc.get(i, 'Peq'));
                stateProp.TimeConstant.textContent = qubx.formatK0(this.modelRec.stateCalc.get(i, 'TimeConstant'));
                stateProp.calcAmp.textContent = qubx.formatK0(this.modelRec.classCalc.get(state.Class, 'Amp'));
                stateProp.calcStd.textContent = qubx.formatK0(this.modelRec.classCalc.get(state.Class, 'Std'));
            }).catch(qubx.reportErr);
        },
        selectState: function(modelRec, i, field) {
            this.select(modelRec);
            this._selRow = this.model.states.get(i);
            this.stateProp.then(stateProp => {
                if ( (! this._selRow) || ! ('Class' in this._selRow) ) {
                    return;
                }
                stateProp.node.style.display = '';
            }).catch(qubx.reportErr);
            this.updateState();
            this.ev.selectState.trigger(modelRec, i, field);
        },
        updateRate: function() {
            this._selRow && this.rateProp.then(rateProp => {
                if ( (! this._selRow) || ! ('k0' in this._selRow) ) {
                    return;
                }
                const rate = this._selRow;
                rateProp.From.textContent = qubx.formatState(this.model.states.get(rate.From));
                rateProp.To.textContent = qubx.formatState(this.model.states.get(rate.To));
                rateProp.K.textContent = qubx.formatK0u(this.modelRec.rateCalc.get(rate.Index, 'K'));
                rateProp.k0.update(rate.k0);
                rateProp.k1.update(rate.k1);
                rateProp.k2.update(rate.k2);
                rateProp.Ligand.update(rate.Ligand);
                rateProp.Voltage.checked = !!rate.Voltage;
                rateProp.Pressure.checked = !!rate.Pressure;
            });
        },
        selectRate: function(modelRec, i, field) {
            this.select(modelRec);
            this._selRow = this.model.rates.get(i);
            this.rateProp.then(rateProp => {
                rateProp.node.style.display = '';
                const input = rateProp[field];
                input && setTimeout(() => {
                    input.focus();
                    input.select();
                }, 10);
            }).catch(qubx.reportErr);
            this.updateRate();
            this.ev.selectRate.trigger(modelRec, i, field);
        },
        _onSetChannelCount: function(n) {
            this.modelProp.then(modelProp => {
                modelProp.channelCount.update(n);
            }).catch(qubx.reportErr);
        },
        _onSetIeqFv: function(IeqFv) {
            this.modelProp.then(modelProp => {
                modelProp.IeqFv.checked = IeqFv;
            }).catch(qubx.reportErr);
            this.stateProp.then(stateProp => {
                stateProp.units.textContent = IeqFv ? "pS" : "pA";
            }).catch(qubx.reportErr);
            if ( this._selRow && ('Class' in this._selRow) ) {
                this.updateState();
            }
        },
        _onSetVRev: function(Vrev) {
            this.modelProp.then(modelProp => {
                modelProp.Vrev.update(Vrev);
            }).catch(qubx.reportErr);
        },
        _onInsertedState: function(row) {
            this.selectState(this.modelRec, row.Index);
        },
        _onRemovedState: function(row) {
            if ( this._selRow === row ) {
                this.select(this.modelRec);
            }
        },
        _onSetState: function(info) {
            if ( this._selRow === this.model.states.get(info.row) ) {
                this.updateState();
            }
        },
        _onInsertedRate: function(row) {
            if ( (row.Index % 2) ) {
                this.selectRate(this.modelRec, row.Index-1, 'k0');
            }
        },
        _onRemovingRate: function(row) {
            if ( this._selRow === row ) {
                this.select(this.modelRec);
            }
        },
        _onSetRate: function(info) {
            const rate = this.model.rates.get(info.row);
            const selected = this._selRow === rate;
            if ( selected ) {
                this.updateRate();
            }
        },
        _onQ: function() {
            if ( this._selRow && this._selRow.k0 ) {
                this.rateProp.then(rateProp => {
                    rateProp.K.textContent = qubx.formatK0(this.modelRec.rateCalc.get(this._selRow.Index, 'K'));
                }).catch(qubx.reportErr);
            }
        },
        _onPeq: function() {
            if ( this._selRow && ('Class' in this._selRow) ) {
                this.stateProp.then(stateProp => {
                    stateProp.Peq.textContent = qubx.formatK0(this.modelRec.stateCalc.get(this._selRow.Index, 'Peq'));
                    stateProp.TimeConstant.textContent = qubx.formatK0(this.modelRec.stateCalc.get(this._selRow.Index, 'TimeConstant'));
                }).catch(qubx.reportErr);
            }
        },
        _onAmps: function() {
            if ( this._selRow && ('Class' in this._selRow) ) {
                this.stateProp.then(stateProp => {
                    stateProp.calcAmp.textContent = qubx.formatK0(this.modelRec.classCalc.get(this._selRow.Class, 'Amp'));
                    stateProp.calcStd.textContent = qubx.formatK0(this.modelRec.classCalc.get(this._selRow.Class, 'Std'));
                }).catch(qubx.reportErr);
            }
        },
        _enteredActive: function(event) {
            this.modelProp.then(modelProp => {
                this.view.models.activate(this.modelRec.index, modelProp.active.checked);
            });
        },
        _enteredChannelCount: function(n) {
            this.model.channelCount(n);
        },
        _enteredIeqFv: function(event) {
            this.modelProp.then(modelProp => {
                this.model.IeqFv(modelProp.IeqFv.checked);
            }).catch(qubx.reportErr);
        },
        _enteredVrev: function(Vrev) {
            this.model.vRev(Vrev);
        },
        _enteredClass: function(cls) {
            if ( this._selRow ) {
                this.model.states.set(this._selRow.Index, 'Class', cls);
                this.model.undo.sealUndo("Set Class");
            }
        },
        _enteredLabel: function(lbl) {
            if ( this._selRow ) {
                this.model.states.set(this._selRow.Index, 'Label', lbl);
                this.model.undo.sealUndo("Set Label");
            }
        },
        _enteredAmp: function(x) {
            if ( this._selRow ) {
                this.model.classes.set(this._selRow.Class, this.model.IeqFv() ? 'Cond' : 'Amp', x);
                this.model.undo.sealUndo("Set Amp");
            }
        },
        _enteredStd: function(x) {
            if ( this._selRow ) {
                this.model.classes.set(this._selRow.Class, this.model.IeqFv() ? 'CondStd' : 'Std', x);
                this.model.undo.sealUndo("Set Std");
            }
        },
        _enteredK0: function(k0) {
            if ( this._selRow ) {
                this.model.rates.set(this._selRow.Index, 'k0', k0);
                this.model.undo.sealUndo("Set k0");
            }
        },
        _enteredK1: function(k1) {
            if ( this._selRow ) {
                this.model.rates.set(this._selRow.Index, 'k1', k1);
                this.model.undo.sealUndo("Set k1");
            }
        },
        _enteredK2: function(k2) {
            if ( this._selRow ) {
                this.model.rates.set(this._selRow.Index, 'k2', k2);
                this.model.undo.sealUndo("Set k2");
            }
        },
        _enteredLigand: function(ligand) {
            if ( this._selRow ) {
                this.model.rates.set(this._selRow.Index, 'Ligand', ligand);
                this.model.undo.sealUndo("Set Ligand name");
            }
            this.rateProp.then(rateProp => {
                rateProp.noLigand.disabled = ! ligand;
            });
        },
        _enteredVoltage: function(event) {
            try {
                if ( this._selRow ) {
                    this.rateProp.then(rateProp => {
                        const yes = rateProp.Voltage.checked;
                        this.model.rates.set(this._selRow.Index, 'Voltage', yes ? 'Voltage' : '');
                        this.model.undo.sealUndo("Set Voltage-sensitive");
                    }).catch(qubx.reportErr);
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        },
        _enteredPressure: function(event) {
            try {
                if ( this._selRow ) {
                    this.rateProp.then(rateProp => {
                        const yes = rateProp.Pressure.checked;
                        this.model.rates.set(this._selRow.Index, 'Pressure', yes ? 'Pressure' : '');
                        this.model.undo.sealUndo("Set Pressure-sensitive");
                    }).catch(qubx.reportErr);
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        }
    };
    
    // mixins for each qubx_model.Models.models[i]
    const EditMethods = {
        clear: function() {
            this.model.clear();
            this.model.undo.sealUndo("Delete all states");
        },
        deleteState: function(i) {
            this.view.context.select(this);
            this.model.states.remove(i);
            this.model.undo.sealUndo("Delete state");
        },
        deleteRate: function(i) {
            this.view.context.select(this);
            const iMain = Math.floor(i/2)*2 + 1;
            this.model.rates.remove(iMain);
            this.model.rates.remove(iMain-1);
            this.model.undo.sealUndo("Delete rate");
        },
        close: function() {
            const view = this.view,
                  model = this.model,
                  divUndo = document.createElement('div');
            divUndo.className = 'qubx_model_undoCloseBox';
            divUndo.textContent = `Closed ${model.modelName||'Untitled'}.`;
            const btnUndo = document.createElement('button');
            btnUndo.textContent = 'Undo';
            divUndo.appendChild(btnUndo);
            view.context.node.insertBefore(divUndo, view.context.node.firstChild);
            let timeout = setTimeout(() => {
                timeout = null;
                view.context.node.removeChild(divUndo);
            }, 10000);
            btnUndo.addEventListener('click', () => {
                timeout && clearTimeout(timeout);
                view.models.insert(undefined, model);
                view.context.node.removeChild(divUndo);
            });
            this.view.models.remove(this.index);
        }
    };
    const CalcMethods = {
        initCalc: function(view) {
            this.view = view;
            this._modelStim = {};
            this.evCalc = qubx.makeEvents(['Q', 'Peq', 'amps', 'stimulus']);
            this.stateCalc = new qubx.Table({label: 'States'});
            this.stateCalc.addField({name: 'Peq', defval: 0, format: qubx.formatK0});
            this.stateCalc.addField({name: 'TimeConstant', units: 'ms', defval: 1, format: qubx.formatK0});
            this.classCalc = new qubx.Table({label: 'Classes'});
            this.classCalc.addField({name: 'Amp', units: 'pA', defval: 0.0, format: qubx.formatK0});
            this.classCalc.addField({name: 'Std', defval: 0.0, format: qubx.formatK0});
            this.rateCalc = new qubx.Table({label: 'Rates'});
            this.rateCalc.addField({name: 'K', defval: 1, format: qubx.formatK0});
            this.calculateStateStatsSoon = qubx.batchDelayed(this._calculateStateStats.bind(this), 100);
            this.updateQSoon = qubx.batchDelayed(this._updateQ.bind(this), 100);
            this.sendAmpsSoon = qubx.batchDelayed(this._sendAmps.bind(this), 50);
            this._calcOnSetIVRev = this._onSetIVRevC.bind(this);
            this.model.ev.setIeqFv.sub(this._calcOnSetIVRev);
            this.model.ev.setVRev.sub(this._calcOnSetIVRev);
            this._calcOnInsertedState = this._onInsertedStateC.bind(this);
            this._calcOnRemovedState = this._onRemovedStateC.bind(this);
            this.model.states.ev.inserted.sub(this._calcOnInsertedState);
            this.model.states.ev.removed.sub(this._calcOnRemovedState);
            this._calcOnInsertedClass = this._onInsertedClassC.bind(this);
            this._calcOnRemovingClass = this._onRemovingClassC.bind(this);
            this._calcOnSetClass = this._onSetClassC.bind(this);
            this.model.classes.ev.inserted.sub(this._calcOnInsertedClass);
            this.model.classes.ev.removing.sub(this._calcOnRemovingClass);
            this.model.classes.ev.set.sub(this._calcOnSetClass);
            this._calcOnInsertedRate = this._onInsertedRateC.bind(this);
            this._calcOnRemovingRate = this._onRemovingRateC.bind(this);
            this._calcOnSetRate = this._onSetRateC.bind(this);
            this.model.rates.ev.inserted.sub(this._calcOnInsertedRate);
            this.model.rates.ev.removing.sub(this._calcOnRemovingRate);
            this.model.rates.ev.set.sub(this._calcOnSetRate);
            view.ev.tDead.sub(this.updateQSoon);
            this._updateQ();
            return this;
        },
        disposeCalc: function() {
            this.model.ev.setIeqFv.unsub(this._calcOnSetIVRev);
            this.model.ev.setVRev.unsub(this._calcOnSetIVRev);
            this.model.states.ev.inserted.unsub(this._calcOnInsertedState);
            this.model.states.ev.removed.unsub(this._calcOnRemovedState);
            this.model.classes.ev.inserted.unsub(this._calcOnInsertedClass);
            this.model.classes.ev.removing.unsub(this._calcOnRemovingClass);
            this.model.classes.ev.set.unsub(this._calcOnSetClass);
            this.model.rates.ev.inserted.unsub(this._calcOnInsertedRate);
            this.model.rates.ev.removing.unsub(this._calcOnRemovingRate);
            this.model.rates.ev.set.unsub(this._calcOnSetRate);
            this.view.ev.tDead.unsub(this.updateQSoon);
            this.view = this.evCalc = this.stateCalc  = this.classCalc = this.rateCalc = null;
        },
        setStimulus: function(dataStim, getMissing) {
            this._dataStim = dataStim || this._dataStim;
            this.stimulus = Object.assign({}, this._modelStim, this._dataStim);
            this.updateQSoon();
        },
        _updateQ: function() {
            qubx_lib.then(lib => {
                if ( this.view ) {
                    this.Q = this.model.makeQ(this.stimulus);
                    const clazz = this.model.states.entries.map(qubx.m('Class'));
                    this.Qe = lib.QtoQe(this.Q, clazz, 1e-3*this.view.tDead());
                    this._calculateStateStats();
                    this._calculateRateStats();
                    this.evCalc.Q.trigger(this.Q, this.Qe);
                }
            }).catch(qubx.reportErr);
        },
        _calculateStateStats: function() {
            qubx_lib.then(lib => {
                if ( this.view && this.Qe && (this.Qe.length === this.model.states.size) ) {
                    const Peq = lib.QtoPe(this.Qe);
                    const ksum = Peq.map(() => 0.0);
                    this.model.rates.entries.forEach(rate => {
                        ksum[rate.From] += this.Qe[rate.From][rate.To];
                    });
                    const tc = ksum.map(s => s && (1e3 / s));
                    Peq.forEach((p,i) => {
                        this.stateCalc.set(i, 'Peq', p);
                        this.stateCalc.set(i, 'TimeConstant', tc[i]);
                    });
                    this.evCalc.Peq.trigger(Peq, tc);
                }
            }).catch(qubx.reportErr);
        },
        _calculateRateStats: function() {
            this.model.rates.entries.forEach((rate, i) => {
                this.rateCalc.set(i, 'K', this.Q[rate.From][rate.To]);
            });
            const index = this.model.rates.entries.indexOf(this._selRow);
        },
        _calculateClassAmp: function(iRow) {
            if ( this.model.IeqFv() ) {
                const dV = 1e-3 * ((this.stimulus.Voltage||0.0) - this.model.vRev());
                return dV * this.model.classes.get(iRow, 'Cond');
            } else {
                return this.model.classes.get(iRow, 'Amp');
            }
        },
        _calculateClassStd: function(iRow) {
            if ( this.model.IeqFv() ) {
                const dV = 1e-3 * ((this.stimulus.Voltage||0.0) - this.model.vRev());
                return dV * this.model.classes.get(iRow, 'CondStd');
            } else {
                return this.model.classes.get(iRow, 'Std');
            }
        },
        _onSetIVRevC: function() {
            this._modelStim = this.model.getStimulus();
            this.setStimulus();
            this.evCalc && this.evCalc.stimulus.trigger();
            this.model.classes.entries.forEach((state, i) => {
                this.classCalc.set(i, 'Amp', this._calculateClassAmp(i));
                this.classCalc.set(i, 'Std', this._calculateClassStd(i));
            });
            this.sendAmpsSoon();
        },
        _onInsertedStateC: function(row) {
            this.stateCalc.insert(row.Index, {});
            this.calculateStateStatsSoon();
        },
        _onRemovedStateC: function(row) {
            this.stateCalc.remove(row.Index);
            this.calculateStateStatsSoon();
        },
        _onInsertedClassC: function(row) {
            this.classCalc.insert(row.Index, {
                Amp: this._calculateClassAmp(row.Index),
                Std: this._calculateClassStd(row.Index)
            });
            this.sendAmpsSoon();
        },
        _onRemovingClassC: function(row) {
            this.classCalc.remove(row.Index);
            this.sendAmpsSoon();
        },
        _onSetClassC: function(info) {
            let newCalc = false;
            if ( (info.column === 'Cond') || (info.column === 'CondStd') ) {
                if ( this.model.IeqFv() ) {
                    if ( info.column === 'Cond' ) {
                        this.classCalc.set(info.row, info.column, this._calculateClassAmp(info.row));
                    } else {
                        this.classCalc.set(info.row, info.column, this._calculateClassStd(info.row));
                    }
                    newCalc = true;
                }
            } else if ( (info.column === 'Amp') || (info.column === 'Std') ) {
                if ( ! this.model.IeqFv() ) {
                    this.classCalc.set(info.row, info.column, info.value);
                    newCalc = true;
                }
            }
            if ( newCalc && this._selRow && (this._selRow.Class === info.row) ) {
                this.sendAmpsSoon();
            }
        },
        _sendAmps: function() {
            this.evCalc && this.evCalc.amps.trigger();
        },
        _onInsertedRateC: function(row) {
            this.rateCalc.insert(row.Index, {});
            if ( (row.Index % 2) ) {
                this._modelStim = this.model.getStimulus();
                this.setStimulus();
                this.evCalc && this.evCalc.stimulus.trigger();
            }
        },
        _onRemovingRateC: function(row) {
            this.rateCalc.remove(row.Index);
            if ( row.Index % 2 ) {
                this._modelStim = this.model.getStimulus();
                this.setStimulus();
                this.evCalc && this.evCalc.stimulus.trigger();
            }
        },
        _onSetRateC: function(info) {
            this._modelStim = this.model.getStimulus();
            this.setStimulus();
            this.evCalc && this.evCalc.stimulus.trigger();
        },
    };
    const formatRate = r => r.sens ? qubx.formatK1u(r.row[r.field], r.row[r.sens], r.lbl) :
          (qubx.formatK0u(r.row[r.field]) + (r.row.Ligand ? ' L' : ''));
    const DisplayMethods = {
        initDisplay: function(view, node) {
            this.view = view;
            this.displayNode = node || document.createElement('div');
            this.displayNode.classList.add('qubx_model_view');
            this.displayNode.classList.add('qubx_noselect');
            this.displayNode.setAttribute("tabindex", 0);
            
            this.svg = Snap(100, 100);
            this.svg.select('defs').append(Snap.parse('<filter id="inactive" x="0" y="0"><feGaussianBlur in="SourceGraphic" stdDeviation=".4" /><feComponentTransfer><feFuncR type="linear" slope="0.5" intercept="0"></feFuncR><feFuncG type="linear" slope="0.5" intercept="0"></feFuncG><feFuncB type="linear" slope="0.5" intercept="0"></feFuncB></filter>'))
            this._svgPt = this.svg.node.createSVGPoint();
            this.svgMain = this.svg.group();
            this.svgMain.attr('id', 'modelMain');
            this._border = this.svgMain.rect(0, 0, 100, 100).addClass("qubx_model_bg");
            this._edge = this.svgMain.rect(0, 0, 100, 100).addClass("qubx_model_edge").attr({stroke: '#69f', opacity: 0});
            this._border.click((e,x,y) => {
                const local = this.cursorPoint(e);
                if ( this.rootX || (this === this.view.context.modelRec) ) { // root menu already showing, or this model already in front
                    this.toggleRootmenu(local.x/this.xScale, local.y/this.yScale);
                }
                this.view.context.select(this);
            });
            this.displayNode.appendChild(this.svg.node);
            this._newRate = this.svgMain.line(0, 0, 1, 1).addClass("qubx_model_newRate").attr({visibility: 'hidden'});
            
            this.btnClose = document.createElement('button');
            this.btnClose.textContent = 'X';
            this.btnClose.className = 'qubx_model_close';
            this.displayNode.appendChild(this.btnClose);
            this.btnClose.addEventListener('click', () => {
                this.close();
            });
            
            if ( view.config.hasButtons ) {
                this.modelBtns = qubx_ui.loadTemplate(".qubx_model_btns").then(t => {
                    const btns = {
                        root: t,
                        undo: t.querySelector(".qubx_btn_undo"),
                        redo: t.querySelector(".qubx_btn_redo")
                    };
                    btns.undo.addEventListener("click", this._onClickUndo.bind(this));
                    btns.redo.addEventListener("click", this._onClickRedo.bind(this));
                    [btns.undo, btns.redo].forEach(b => {
                        const mouseup = e => {
                            b.style.top = '0';
                            b.style.left = '0';
                            document.body.removeEventListener("mouseup", mouseup);
                        };
                        b.addEventListener("mousedown", e => {
                            if ( ! b.classList.contains("qubx_btn_do_disabled") ) {
                                b.style.top = '.2em';
                                b.style.left = '.2em';
                                document.body.addEventListener("mouseup", mouseup);
                            }
                        });
                    });
                    this.displayNode.appendChild(t);
                    return btns;
                }).catch(qubx.reportErr);
            } else {
                this.modelBtns = null;
            }
            
            this.rootMenu = qubx_ui.loadTemplate(".qubx_model_rootmenu").then(t => {
                const rootMenu = {
                    node: t,
                    newStates: new qubx_ui.Palette(t.querySelector(".qubx_palette"), {
                        onClick: this._onClickNewClass.bind(this)
                    })
                };
                t.style.top = '0';
                t.style.left = '0';
                t.style.visibility = 'hidden';
                this.displayNode.appendChild(t);
                return rootMenu;
            }).catch(qubx.reportErr);
            
            this.displayNode.addEventListener('keyup', this._onKeyUp.bind(this));
            
            this.states = [];
            this.rates = [];
            this._sel = null;

            this._boundUndoChange = this.model.undo.ev.change.sub(this._onUndoChangeD.bind(this));
            this._boundInsertedState = this.model.states.ev.inserted.sub(this._onInsertedStateD.bind(this));
            this._boundRemovedState = this.model.states.ev.removed.sub(this._onRemovedStateD.bind(this));
            this._boundSetState = this.model.states.ev.set.sub(this._onSetStateD.bind(this));
            this._boundInsertedRate = this.model.rates.ev.inserted.sub(this._onInsertedRateD.bind(this));
            this._boundRemovingRate = this.model.rates.ev.removing.sub(this._onRemovingRateD.bind(this));
            this._boundSetRate = this.model.rates.ev.set.sub(this._onSetRateD.bind(this));
            this._boundSetActive = this.model.ev.active.sub(this._onSetActiveD.bind(this));

            this.model.states.entries.forEach(this._onInsertedStateD.bind(this));
            this.model.rates.entries.forEach(this._onInsertedRateD.bind(this));
            
            this.w = this.displayNode.clientWidth;
            this.h = this.displayNode.clientHeight;
            
            return this;
        },
        onResize: function() {
            const new_w = this.displayNode.clientWidth;
            const new_h = this.displayNode.clientHeight
            const aspect = new_h ? (new_w / new_h) : 1;
            if ( aspect > 1 ) {
                this.h = 100;
                this.w = aspect * this.h;
            } else {
                this.w = 100;
                this.h = this.w / aspect;
            }
            this.svg.attr({viewBox: `0 0 ${this.w} ${this.h}`, width: new_w, height: new_h});
            this.layout();
        },
        disposeDisplay: function() {
            this.displayNode.removeChild(this.svg.node);
            this.model.undo.ev.change.unsub(this._boundUndoChange);
            this.model.states.ev.inserted.unsub(this._boundInsertedState);
            this.model.states.ev.removed.unsub(this._boundRemovedState);
            this.model.states.ev.set.unsub(this._boundSetState);
            this.model.rates.ev.inserted.unsub(this._boundInsertedRate);
            this.model.rates.ev.removing.unsub(this._boundRemovingRate);
            this.model.rates.ev.set.unsub(this._boundSetRate);
            this.model.ev.active.unsub(this._boundSetActive);
            this.modelBtns.then(modelBtns => {
                this.displayNode.removeChild(modelBtns.root);
            }).then(() => this.rootMenu.then(rootMenu => {
                this.displayNode.removeChild(rootMenu.node);
                this.displayNode = this.modelBtns = this.rootMenu = null;
            }));
            this.svg = this._svgPt = this._border = this._edge = this._newRate = this.states = this.rates = null;
        },
        getSVG: function() {
            return modelStyle.then(styleText => {
                const svg = this.svg.node.cloneNode(true);
                [".qubx_model_edge", ".qubx_model_hilite", ".qubx_model_newRate"].forEach(cls => {
                    const node = svg.querySelector(cls);
                    node.parentNode.removeChild(node);
                });
                svg.setAttribute("xmlns", "http://www.w3.org/2000/svg");
                svg.setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
                svg.setAttribute("version", "1.1");
                const style = document.createElement("style");
                style.innerHTML = styleText;
                svg.insertBefore(style, svg.firstChild);
                return svg;
            }).catch(qubx.reportErr);
        },
        unselected: function(model) {
            if ( this._sel ) {
                this._sel.attr({visibility: "hidden"});
                this._sel = null;
                this._selRow = null;
            }
            this._edge.attr({opacity: (model === this.model) ? '0.6' : '0'});
        },
        selectedState: function(i, field) {
            this.unselected(this.model);
            this._sel = this.states[i].hilite;
            this._sel.attr({visibility: "visible"});
            this._selRow = this.model.states.get(i);
        },
        selectedRate: function(i, field) {
            this.unselected(this.model);
            const iMain = Math.floor(i/2)*2 + 1;
            this._sel = this.rates[iMain].hilite;
            this._sel.attr({visibility: "visible"});
            this._selRow = this.model.rates.get(i);
        },
        _onKeyUp: function(event) {
            try {
                if ( (event.keyCode == 8) || (event.keyCode == 46) ) {
                    if ( this._selRow ) {
                        if ( 'Class' in this._selRow ) {
                            this.deleteState(this._selRow.Index);
                        } else {
                            this.deleteRate(this._selRow.Index);
                        }
                    }
                }
            } catch (err) {
                qubx.reportErr(err);
            }
        },
        _onClickNewClass: function(j) {
            this.model.states.append({
                Class: j,
                x: snap_to_grid(this.rootX*100/this.displayNode.clientWidth, SNAP_TO),
                y: snap_to_grid(this.rootY*100/this.displayNode.clientHeight, SNAP_TO)
            });
            this.model.undo.sealUndo("Add state");
            this.hideRootmenu();
        },
        _onClickUndo: function() {
            try {
                this.model.undo.canUndo() && this.model.undo.undo();
            } catch (err) {
                qubx.reportErr(err);
            }
        },
        _onClickRedo: function(event) {
            try {
                this.model.undo.canRedo() && this.model.undo.redo();
            } catch (err) {
                qubx.reportErr(err);
            }
        },
        _onUndoChangeD: function() {
            this.modelBtns && this.modelBtns.then(btns => {
                if ( this.model.undo.canUndo() ) {
                    btns.undo.classList.remove("qubx_btn_do_disabled");
                } else {
                    btns.undo.classList.add("qubx_btn_do_disabled");
                }
                if ( this.model.undo.canRedo() ) {
                    btns.redo.classList.remove("qubx_btn_do_disabled");
                } else {
                    btns.redo.classList.add("qubx_btn_do_disabled");
                }
            }).catch(qubx.reportErr);
        },
        _onInsertedStateD: function(row) {
            const state = {g: this.svgMain.group()};
            state.frame = state.g.rect(0, 0, 2*STATE_WIDTH2, 2*STATE_WIDTH2).addClass("qubx_model_state").addClass("qubx_model_cx").addClass("qubx_model_c"+row.Class);
            state.lbl = state.g.text(STATE_WIDTH2, STATE_WIDTH2+2.2, qubx.formatState(row)).addClass("qubx_model_state_lbl");
            state.hilite = state.g.rect(-1, -1, 2*STATE_WIDTH2+2, 2*STATE_WIDTH2+2).addClass("qubx_model_hilite").attr({visibility: "hidden"});
            this.states.splice(row.Index, 0, state);
            this.layoutState(row.Index);
            state.g.mousedown((e,x,y) => this._onMousedownState(this.states.indexOf(state), e, x, y));
        },
        _onRemovedStateD: function(row) {
            this.states[row.Index].g.remove();
            this.states.splice(row.Index, 1);
            const n = this.states.length;
            let i = row.Index;
            for (; i<n; ++i) {
                this.states[i].lbl.node.textContent = qubx.formatState(this.model.states.get(i));
            }
        },
        _onSetStateD: function(info) {
            const state = this.states[info.row];
            if ( info.column === 'Class' ) {
                state.frame.removeClass("qubx_model_c"+info.prev);
                state.frame.addClass("qubx_model_c"+info.value);
            } else if ( (! this._setXY) && (['x', 'y'].indexOf(info.column) != -1) ) {
                this.layoutState(info.row, true);
                const nr = this.rates.length;
                let r = 1;
                for (; r<nr; r += 2) {
                    if ([this.model.rates.get(r, 'From'), this.model.rates.get(r, 'To')].indexOf(info.row) != -1) {
                        this.layoutRate(r);
                    }
                }
            } else if ( ['Index', 'Label'].indexOf(info.column) != -1 ) {
                state.lbl.node.textContent = qubx.formatState(this.model.states.get(info.row));
                this.layoutState(info.row);
            }
        },
        cursorPoint: function(event) {
            this._svgPt.x = event.clientX; this._svgPt.y = event.clientY;
            return this._svgPt.matrixTransform(this.svg.node.getScreenCTM().inverse());
        },
        _onMousedownState: function(i, event, x, y) {
            const view = this.states[i];
            if ( view.dragHandle ) { // doubled mousedown maybe from touch emulation; ignore
                return;
            }
            view.x0 = view.x;
            view.y0 = view.y;
            this._newRateHit = null;

            this.hideRootmenu();
            this.view.context.selectState(this, i);

            this.svg.mousemove((e,mx,my) => this._onMousemoveState(i, e, mx, my));
            this.svg.mouseup((e,mx,my) => this._onMouseupState(i, e, mx, my));
            view.dragHandle = setTimeout(() => {
                if ( navigator.vibrate ) {
                    navigator.vibrate(100);
                }
                const x0 = this.model.states.get(i, 'x'),
                      y0 = this.model.states.get(i, 'y');
                let toX = x0,
                    toY = y0;
                view.frame.animate({x: -STATE_WIDTH2, y: -STATE_WIDTH2, width: 4*STATE_WIDTH2, height: 4*STATE_WIDTH2}, 70, mina.easein, () => {
                    view.frame.animate({x: 0, y: 0, width: 2*STATE_WIDTH2, height: 2*STATE_WIDTH2}, 70, mina.easein);
                });
                this.svg.unmousemove().unmouseup();
                view.dragHandle = null;
                view.dragging = true; // unmousemove doesn't seem to work in mobile chrome
                view.g.drag(event, x, y, (dx, dy) => {
                    toX = snap_to_grid(Math.max(0, Math.min(100, x0 + dx*100/this.w)), SNAP_TO);
                    toY = snap_to_grid(Math.max(0, Math.min(100, y0 + dy*100/this.h)), SNAP_TO);
                    if ( this.view.config.removeOnEdges ) {
                        if ( (toX === 0) || (toX === 100) || (toY === 0) || (toY === 100) ) {
                            this._edge.attr({opacity: 0.9});
                        } else {
                            this._edge.attr({opacity: 0.4});
                        }
                    }
                    this._setXY = toY !== y0;
                    this.model.states.set(i, 'x', toX);
                    this._setXY = false;
                    this.model.states.set(i, 'y', toY);
                }, () => { // onstart
                    this.view.config.removeOnEdges && this._edge.attr({stroke: 'red', opacity: 0.4});
                }, () => { // onend
                    view.g.undrag();
                    const removing = this.view.config.removeOnEdges && ((toX === 0) || (toX === 100) || (toY === 0) || (toY === 100));
                    this.view.config.removeOnEdges && this._edge.attr({stroke: '#69f', opacity: 0});
                    if ( removing ) {
                        this.deleteState(i);
                    } else {
                        this.model.undo.sealUndo('Move state');
                    }
                    view.dragging = false;
                });
            }, LONG_PRESS_MS);
        },
        _onMouseupState: function(i, event, x, y) {
            const view = this.states[i];
            if ( view.dragHandle ) {
                clearTimeout(view.dragHandle);
                view.dragHandle = null;
            } else {
                this._newRate.attr({visibility: 'hidden'});
                if ( (this._newRateHit !== null) && (-1 === this.model.rate_connecting(i, this._newRateHit)) ) {
                    this.model.rates.append({From: i, To: this._newRateHit});
                    this.model.rates.append({From: this._newRateHit, To: i});
                    this.model.undo.sealUndo("Add rate");
                }
            }
            this.svg.unmousemove().unmouseup();
        },
        _onMousemoveState: function(i, event, x, y) {
            const view = this.states[i],
                  local = this.cursorPoint(event);
            if ( view.dragHandle && ! this.ptInState(local.x, local.y, i) ) {
                clearTimeout(view.dragHandle);
                view.dragHandle = null;
                this._newRate.attr({visibility: 'visible', x1: this.states[i].x, y1: this.states[i].y})
                    .removeClass("qubx_model_newRateHit");
            }
            if ( ! view.dragging ) {
                this._newRate.attr({x2: local.x, y2: local.y});
                //this._newRate.attr({x2: x*this.xScale, y2: y*this.yScale});
                const s = this.stateAtLocal(local.x, local.y, [i]);
                if ( s !== this._newRateHit ) {
                    if ( this._newRateHit !== null ) {
                        this._newRate.removeClass("qubx_model_newRateHit");
                    }
                    if ( s !== null ) {
                        this._newRate.addClass("qubx_model_newRateHit");
                    }
                    this._newRateHit = s;
                }
            }
        },
        layoutState: function(i) {
            const state = this.model.states.get(i),
                  view = this.states[i],
                  lblW = (1.1 * view.lbl.node.getBoundingClientRect().width * (this.w/(this.displayNode.clientWidth||1)))|0,
                  w2 = Math.max(STATE_WIDTH2, lblW>>1);
            view.x = state.x*this.w/100;
            view.y = state.y*this.h/100;
            view.l = view.x - w2;
            view.r = view.x + w2;
            view.t = view.y - STATE_WIDTH2;
            view.b = view.y + STATE_WIDTH2;
            view.g.transform(`translate(${view.l},${view.t})`);
            view.frame.attr({width: 2*w2});
            view.lbl.attr({x: w2});
            view.hilite.attr({width: 2*w2+2});
        },
        ptInState: function(sx, sy, i) {
            const s = this.states[i];
            return ( (sx >= s.l) && (sx <= s.r) && (sy >= s.t) && (sy <= s.b) );
        },
        stateAtLocal: function(x, y, ignoreStates) {
            const ignore = ignoreStates || [],
                  n = this.states.length;
            let i = 0;
            for (; i<n; ++i) {
                if ( this.ptInState(x, y, i) && (-1 === ignoreStates.indexOf(i)) ) {
                    return i;
                }
            }
            return null;
        },
        _onInsertedRateD: function(row) {
            if ( (row.Index % 2) === 0 ) { // operate on symmetric pair of rates
                this.rates.splice(row.Index, 0, {});
            } else {
                const rate = {g: this.svgMain.group()},
                      rev = this.rates[row.Index-1],
                      rowRev = this.model.rates.get(row.Index-1);
                rate.line = rate.g.line(1.5*STATE_WIDTH2, 0, 0, 0).addClass("qubx_model_rate");
                rate.arrowF = rate.g.line(-STATE_WIDTH2, 0, 0, 0).addClass("qubx_model_rate");
                rate.arrowB = rate.g.line(0, 0, STATE_WIDTH2, 0).addClass("qubx_model_rate");
                rate.arrowB.transform(`translate(${1.5*STATE_WIDTH2},0) rotate(45)`);
                rate.k = rate.g.text(50, 5, ['', '', ''])
                    .addClass("qubx_model_rate_lbl");
                rev.k = rate.g.text(50, -2, ['', '', ''])
                    .addClass("qubx_model_rate_lbl");
                function tspanInfo(rate, row, tspanSel) {
                    return Array.prototype.map.call(tspanSel, (span, i) => ({
                        rate: rate,
                        row: row,
                        field: 'k'+i,
                        sens: ['', 'Voltage', 'Pressure'][i],
                        lbl: ['', ' V ', ' P '][i],
                        span: span
                    }));
                }
                const spans = tspanInfo(rate, row, rate.k.selectAll('tspan'));
                d3.select(rate.k.node)
                    .selectAll('tspan')
                    .data(spans)
                    .text(formatRate);
                const spansRev = tspanInfo(rev, rowRev, rev.k.selectAll('tspan'));
                d3.select(rev.k.node)
                    .selectAll('tspan')
                    .data(spansRev)
                    .text(formatRate);
                spans.concat(spansRev).forEach(d => {
                    d.span.mousedown((e,x,y) => this._onMousedownRate(this.rates.indexOf(d.rate), d.field, e, x, y));
                    d.span.wheel((e,x,y) => this._onWheelRate(this.rates.indexOf(d.rate), d.field, e, x, y));
                });
                rate.hilite = rate.g.rect(1.5*STATE_WIDTH2-1, -6.5, 0, 12.5).addClass("qubx_model_hilite").attr({visibility: "hidden"});
                this.rates.splice(row.Index, 0, rate);
                this.layoutRate(row.Index);
                rate.g.mousedown((e,x,y) => this._onMousedownRate(this.rates.indexOf(rate), null, e, x, y));
            }
        },
        _onRemovingRateD: function(row) {
            if ( row.Index % 2 ) {
                this.rates[row.Index].g.remove();
            }
            this.rates.splice(row.Index, 1);
        },
        _onSetRateD: function(info) {
            d3.select(this.rates[info.row].k.node)
                .selectAll('tspan').text(formatRate);
        },
        _onSetActiveD: function(active) {
            this.svgMain.attr('filter', active ? '' : 'url(#inactive)');
        },
        _onMousedownRate: function(i, field, event, x, y) {
            const rate = this.model.rates.get(i);
            if ( (field === 'k1') && ! rate.Voltage ) {
                field = rate.Pressure ? 'k2' : 'Ligand';
            }
            this.view.context.selectRate(this, i, field);
            this.hideRootmenu();
            event.stopPropagation();
        },
        _onWheelRate: function(i, field, event, x, y) {
            /// focus the text input?
            if ( event.deltaY ) {
                const dy = event.deltaMode ? event.deltaY : ((event.deltaY >> 3)||((event.deltaY > 0) ? 1 : -1));
                let k = this.model.rates.get(i, field);
                k = qubx.scrolledDigits(k, dy, event.shiftKey, event.deltaMode>1, (field==='k0')?1e-10:undefined);
                this.model.rates.set(i, field, k);
            }
            event.stopPropagation();
        },
        layoutRate: function(i) {
            const rate = this.model.rates.get(i);
            let from = this.states[rate.From],
                to = this.states[rate.To];
            const view = this.rates[i],
                  viewR = this.rates[i-1],
                  up = from.x <= to.x;
            var theta, rad, lA, lB;
            if ( up ) {
                view.k.attr({y:-2});
                viewR.k.attr({y:5});
            } else {
                view.k.attr({y:5});
                viewR.k.attr({y:-2});
                from = to;
                to = this.states[rate.From];
            }
            theta = Math.atan2(to.y-from.y, to.x-from.x);
            rad = distance(from.x, from.y, to.x, to.y);
            lA = 1.5*STATE_WIDTH2;
            lB = rad - 1.5*STATE_WIDTH2;
            // set up rotated, translated context along positive x axis
            view.g.transform(`translate(${from.x},${from.y}) rotate(${180*theta/Math.PI})`);
            view.arrowF.transform(`translate(${lB},0) rotate(45)`);
            view.line.attr({x2: lB});
            view.hilite.attr({width: Math.max(0, lB-lA+2)});
            lA += (lB - lA)/2;
            [view.k, viewR.k].forEach(t => t.attr({x: lA}));
        },
        layout: function() {
            var i, n;
            this._border.attr({width: this.w, height: this.h});
            this._edge.attr({width: this.w, height: this.h});
            this.xScale = this.w / this.displayNode.clientWidth;
            this.yScale = this.h / this.displayNode.clientHeight;
            for (i=0, n=this.states.length; i<n; ++i ) {
                this.layoutState(i);
            }
            for (i=1, n=this.rates.length; i<n; i += 2 ) {
                this.layoutRate(i);
            }
        },
        toggleRootmenu: function(x, y) {
            if ( this.rootX ) {
                this.hideRootmenu();
            } else {
                this.rootMenu.then(rootMenu => {
                    this.rootX = x;
                    this.rootY = y;
                    x = Math.min(x, this.displayNode.clientWidth - rootMenu.node.clientWidth);
                    y = Math.min(y, this.displayNode.clientHeight - rootMenu.node.clientHeight);
                    rootMenu.node.style.left = x;
                    rootMenu.node.style.top = y;
                    rootMenu.node.style.visibility = 'visible';
                }).catch(qubx.reportErr);
            }
        },
        hideRootmenu: function() {
            if ( this.rootX ) {
                this.rootMenu.then(rootMenu => {
                    rootMenu.node.style.visibility = 'hidden';
                    this.rootX = 0;
                    this.rootY = 0;
                }).catch(qubx.reportErr);
            }
        }
    };

    const ModelViewConfig = {
        hasButtons: true,
        hasContext: true,
        removeOnEdges: false, // drag states to edge to delete
        listStimuli: () => [],
        grabAmp: () => Promise.reject('Unimplemented'),
            // Promise.resolve({mean, std})
        grabAmps: model => Promise.reject('Unimplemented')
            // Promise.resolve({amp: [cls0, cls1, ...], std: [cls0, cls1, ...]})
    };
    exports.ModelView = class ModelView {
        constructor(config) {
            this.config = Object.assign({}, ModelViewConfig, config);
            this.models = config.models || new qubx_model.Models();
            this.count = this.models.models.length;
            this.modelStim = {};
            this.node = config.node;
            if ( ! this.node ) {
                this.node = document.createElement("div");
                this.node.style.width = "100%";
                this.node.style.height = "100%";
            }
            this.node.classList.add("qubx_model");
            this.node.style.overflow = 'hidden';
            this.gridbox = document.createElement('div');
            this.gridbox.classList.add('qubx_models_view');
            this.node.insertBefore(this.gridbox, this.node.firstChild);
            this.grid = new qubx_ui.AspectGrid();
            this.grid.node.classList.add('qubx_fullsize');
            this.gridbox.appendChild(this.grid.node);
            
            this.ev = qubx.makeEvents(['tDead', 'amps', 'Peq', 'Q', 'count', 'active']);
            this._passThruAmps = this.ev.amps.trigger.bind(this.ev.amps);
            this._passThruPeq = this.ev.amps.trigger.bind(this.ev.Peq);
            this._passThruQ = this.ev.amps.trigger.bind(this.ev.Q);
            this._hOnStimulus = this._onStimulus.bind(this);
            this._tDead = 0;
            this.active = [];
            
            this.node.addEventListener('dragenter', qubx_ui.dragEnter(this.node));
            this.node.addEventListener('dragleave', qubx_ui.dragLeave(this.node));
            this.node.addEventListener('dragover', qubx.stopAndPrevent);
            this.node.addEventListener('drop', this._onDrop.bind(this));
            
            this.context = Object.create(ContextMethods)
                .init(this);
            this.node.appendChild(this.context.node);
            if ( ! this.config.hasContext ) {
                this.context.node.style.display = 'none';
            }
            this.context.ev.unselect.sub(this._onUnselect.bind(this));
            this.context.ev.selectModel.sub(this._onSelectModel.bind(this));
            this.context.ev.selectRate.sub(this._onSelectRate.bind(this));
            this.context.ev.selectState.sub(this._onSelectState.bind(this));
            
            this.dataStim = {};
            
            this.models.ev.insert.sub(this._onInsert.bind(this));
            this.models.ev.remove.sub(this._onRemove.bind(this));
            this.models.ev.activate.sub(this._onActivate.bind(this));
            this.models.models.forEach((rec, i) => this._onInsert(rec, i));
            
            const resizeFlowHandle = qubx_ui.onResize(this.node, this.onResize.bind(this));
            
            this.dispose = () => {
                qubx_ui.clearResize(resizeFlowHandle);
                let i = this.models.models.length - 1;
                for (; i>=0; --i) {
                    this._onRemove(this.models.models[i], i);
                }
            };
        }
        onResize() {
            const aspect = this.node.clientWidth / this.node.clientHeight;
            if ( aspect > 1 ) {
                this.node.style.flexFlow = 'row nowrap';
            } else {
                this.node.style.flexFlow = 'column nowrap';
            }
        }
        toJson() {
            return this.models.models.map(rec => rec.model.toJson());
        }
        fromJson(models) {
            while ( this.models.models.length ) {
                this.models.remove(this.models.models.length-1);
            }
            if ( models.channelCount ) {
                this.models.insert().model.fromJson(models);
            } else {
                models.forEach(qmj => {
                    const model = this.models.insert().model;
                    model.fromJson(qmj);
                    model.undo.clear();
                });
            }
        }
        setStimulus(dataStim, getMissing) {
            this.dataStim = dataStim || this._dataStim;
            for (const name of Object.keys(this.modelStim)) {
                if ( ! (name in this.dataStim) ) {
                    this.dataStim[name] = getMissing ? getMissing(name) : this.modelStim[name];
                }
            }
            this.models.models.forEach(rec => rec.setStimulus(this.dataStim));
            return this;
        }
        getStimulus() {
            return Object.assign.apply(null, [{}].concat(this.active.map(m => m.model.getStimulus())));
        }
        _onStimulus() {
            this.modelStim = Object.assign({}, ...this.active.map(m => m.model.getStimulus()));
        }
        tDead(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._tDead;
            }
            if ( this._tDead !== x ) {
                this._tDead = x;
                this.ev.tDead.trigger(x);
            }
            return this;
        }
        _onInsert(rec, i) {
            Object.assign(rec, EditMethods, CalcMethods, DisplayMethods);
            rec.initCalc(this);
            rec.initDisplay(this);
            rec.evCalc.amps.sub(this._passThruAmps);
            rec.evCalc.Peq.sub(this._passThruPeq);
            rec.evCalc.Q.sub(this._passThruQ);
            rec.evCalc.stimulus.sub(this._hOnStimulus);
            this.grid.insert(rec.displayNode, i, rec.onResize.bind(rec));
            this._updateCount();
            this.context.select(rec);
            this._showhideClose();
        }
        _onRemove(rec, i) {
            if ( this.context.modelRec === rec ) {
                this.context.unselect();
            }
            this.grid.remove(i);
            rec.evCalc.amps.unsub(this._passThruAmps);
            rec.evCalc.Peq.unsub(this._passThruPeq);
            rec.evCalc.Q.unsub(this._passThruQ);
            rec.evCalc.stimulus.unsub(this._hOnStimulus);
            this._updateCount();
            rec.disposeDisplay();
            rec.disposeCalc();
            this._showhideClose();
        }
        _showhideClose() {
            const closeDisplay = (this.grid.items.length > 1) ? '' : 'none';
            this.models.models.forEach(rec => {
                rec.btnClose.style.display = closeDisplay;
            });
        }
        _updateCount() {
            this.count = this.grid.items.length;
            this.ev.count.trigger(this.count);
            this._genActive();
            [this.ev.amps, this.ev.Peq, this.ev.Q].forEach(e => e.trigger());
        }
        _genActive() {
            this.active = this.models.models.filter(rec => rec.model.active);
            this.ev.active.trigger(this.active);
        }
        _onActivate(rec, i, active) {
            this.context.onActivate(rec, i, active);
            this._genActive();
        }
        _onUnselect() {
            this.context.modelRec && this.context.modelRec.unselected();
            this.model = null;
            this.info = null;
        }
        _onSelectModel(rec) {
            this.model = this.context.model;
            this.info = rec;
            this.context.modelRec.unselected(this.model);
        }
        _onSelectRate(rec, i, field) {
            this.model = this.context.model;
            this.context.modelRec.selectedRate(i, field);
        }
        _onSelectState(rec, i, field) {
            this.model = this.context.model;
            this.context.modelRec.selectedState(i, field);
        }
        _onDrop(event) {
            event.stopPropagation();
            event.preventDefault();
            this.openFiles(event.dataTransfer.files);
            this.node.classList.remove('qubx_dragtarget');
        }
        openFiles(files) {
            for (const f of files) {
                const ext = (f.name.length > 4) ? f.name.slice(f.name.length-4).toLowerCase() : '.qmj';
                if ( ext === '.qmj' ) {
                    const reader = new FileReader();
                    reader.onload = e => this._onLoadFile(f, JSON.parse(e.target.result));
                    reader.readAsText(f);
                } else if ( ext === '.qmf' ) {
                    const reader = new FileReader();
                    reader.onload = e => this._onLoadFile(f, qubx_model.QMFtoJson(qubx.readTree(e.target.result)));
                    reader.readAsArrayBuffer(f);
                } else {
                    const reader = new FileReader();
                    reader.onload = e => {
                        qubx.errOut.send('unknown model type: '+f.name+': '+((e.target.result.length > 65536) ? e.target.result.substring(0, 65536) : e.target.result));
                    };
                    reader.readAsText(f);
                    console.log('unknown model type', f);
                }
            }
        }
        _onLoadFile(f, qmj) {
            const model = this.models.insert().model;
            model.fromJson(qmj);
            model.modelName = f.name.replace(/.qmf$/i, '');
            model.modelAbout = '';
            model.modelId = null;
            model.changed = false;
        }
        getSVG() {
            const modelRec = this.context.modelRec || this.models.models[0];
            return modelRec ? modelRec.getSVG() : Promise.reject('No model');
        }
    };

    exports.ModelPicker = class ModelPicker {
        constructor() {
            this.dlg = new qubx_ui.Modal({
                title: 'Browse models',
                template: '.qubx_mbrowse',
                onOK: this._onClickOK.bind(this),
                onCancel: this._onClickCancel.bind(this)
            });
            this.dlg.dom.then(dom => {
                const dotdot = qubx_ui.querydotdot(dom.content);
                this.modelView = new exports.ModelView({hasButtons: false, hasContext: false, node: dotdot('.qubx_mbrowse_model')});
                this.modelView.models.insert();
                this.modelView.model.undo.disable();
                this.list = new exports.ModelPickerList({
                    node: dotdot('.qubx_mbrowse_list'),
                    modelView: this.modelView,
                    info: {
                        modelName: dotdot('.qubx_mbrowse_mName'),
                        modelAbout: dotdot('.qubx_mbrowse_mAbout'),
                        modelAuthors: dotdot('.qubx_mbrowse_mAuthors'),
                        modelWhenPub: dotdot('.qubx_mbrowse_mWhenPub'),
                        modelDOI: dotdot('.qubx_mbrowse_mDOI'),
                        personName: dotdot('.qubx_mbrowse_pName'),
                        personAbout: dotdot('.qubx_mbrowse_pAbout'),
                        modelWhenAdded: dotdot('.qubx_mbrowse_mWhenAdded'),
                        link: dotdot('.qubx_mbrowse_link')
                    }
                });
                dotdot.listen('.qubx_mbrowse_search', 'click', event => {
                    try {
                        this.list.populate(this.qNode.value);
                    } catch (err) {
                        qubx.reportErr(err);
                    }
                });
                this.qNode = dotdot.listen('.qubx_mbrowse_q', keypress, event => {
                    try {
                        if ( event.which === 13 ) {
                            this.list.populate(qNode.value);
                            qNode.select();
                        }
                    } catch (err) {
                        qubx.reportErr(err);
                    }
                });
                dom.root.querySelector('.qubx_modal_ok').textContent = 'Load';
            }).catch(qubx.reportErr);
        }
        run(storage) {
            this.list.storage = storage;
            this.list.populate();
            this.qNode.value = "";
            this.dlg.run();
            return new Promise((resolve, reject) => {
                this.doneRun = resolve;
            });
        }
        _onClickOK() {
            this.doneRun({
                modelId: this.list.modelIDs[this.list.index],
                modelQmj: this.modelView.model.toJson(),
                modelName: this.modelView.model.modelName,
                modelAbout: this.modelView.model.modelAbout
            });
            this.doneRun = null;
            this.list.storage = null;
        }
        _onClickCancel() {
            this.list.storage = null;
            this.doneRun({});
            this.doneRun = null;
        }
    };

    const ModelPickerListConfig = {
        // node, modelView, info{modelName...personAbout}
    };
    exports.ModelPickerList = class ModelPickerList {
        constructor(config) {
            this.config = Object.assign({}, ModelPickerListConfig, config);
            this.select = d3.select(config.node)
                .append('select');
            this.select.property('size', 10);
            this.select.on('change', this._onChange.bind(this));
            this.index = -1;
        }
        selected(i, qmj) {
            this.index = i;
            if ( i < 0 ) {
                this.config.modelView.model.clear();
                const info = this.config.info;
                info.modelName.textContent = "";
                info.modelAbout.textContent = "";
                info.modelAuthors.textContent = "";
                info.modelWhenPub.textContent = "";
                info.modelDOI.textContent = "";
                info.personName.textContent = "";
                info.personAbout.textContent = "";
                info.modelWhenAdded.textContent = "";
                info.link.innerHTML = '';
                return;
            }
            if ( typeof(qmj) !== 'undefined' ) {
                this.qmjCache[i] = qmj;
            }
            if ( typeof(this.qmjCache[i]) !== 'undefined' ) {
                this.config.modelView.model.fromJson(this.qmjCache[i]);
                const info = this.config.info;
                info.modelName.textContent = this.models[i].modelName;
                info.modelAbout.textContent = this.models[i].modelAbout;
                info.modelAuthors.textContent = this.models[i].modelAuthors||"";
                info.modelWhenPub.textContent = this.models[i].modelWhenPub||"";
                info.modelDOI.textContent = this.models[i].modelDOI||"";
                info.personName.textContent = this.models[i].personName||"";
                info.personAbout.textContent = (this.models[i].personAbout||"").replace(/\n/g, " | ");
                info.modelWhenAdded.textContent = this.models[i].modelWhenAdded||"";
                info.link.innerHTML = `link to this model: <a href="${this.models[i].url}">${this.models[i].url}</a>`;
                this.config.modelView.model.modelName = this.models[i].modelName;
                this.config.modelView.model.modelAbout = this.models[i].modelAbout;
                setTimeout(() => {
                    this.select.node().focus();
                }, 200);
            } else {
                this.storage.getModelQmj(i).then(result => {
                    this.selected(i, result);
                }).catch(qubx.reportErr);
            }
        }
        populate(queryString) {
            this.storage.listModels(queryString).then(models => {
                this.models = models;
                this.modelIDs = models.map(qubx.m('modelId'));
                this.qmjCache = [];
                const bound = this.select.selectAll('option').data(models);
                bound.enter().insert('option');
                bound.exit().remove();
                bound.property('value', qubx.m('linkId'));
                bound.text(qubx.m('modelName'));
                if ( models.length ) {
                    this.select.node().selectedIndex = 0;
                    this.selected(0);
                } else {
                    this.selected(-1);
                }
            }).catch(qubx.reportErr);
        }
        _onChange() {
            this.selected(this.select.node().selectedIndex);
        }
    };

    function expand(extrema, factor) {
        factor = factor || 1.15;
        const d = factor * (extrema.hi - extrema.lo);
        const c = 0.5*(extrema.hi + extrema.lo);
        return {
            d: d,
            lo: c-0.5*d,
            hi: c+0.5*d
        };
    }
    function formatPoint(x, y) {
        return `${x.toFixed(3)},${y.toFixed(3)}`;
    }
    const EnergyLandscape2DDefs = {
    };
    exports.EnergyLandscape2D = class EnergyLandscape2D {
        constructor(config) {
            this.config = Object.assign({}, EnergyLandscape2DDefs, config);
            this.node = this.config.node || document.createElement('div');
            this.node.classList.add('qubx_landscape');
            this.height = this.width = 100;
            this.svg = Snap(this.width, this.height);
            this.svg.select('defs').append(Snap.parse('<marker orient="auto" refY="0.0" refX="0.0" id="TriangleInM" style="overflow:visible"><path id="path4424" d="M 5.77,0.0 L -2.88,5.0 L -2.88,-5.0 L 5.77,0.0 z " fill="#000000" stroke="#000000" style="fill-rule:evenodd;stroke-width:1pt;stroke-opacity:1;fill-opacity:1" transform="scale(-0.4)" /></marker>'));
            this.svg.select('defs').append(Snap.parse('<marker orient="auto" refY="0.0" refX="0.0" id="TriangleOutM" style="overflow:visible"> <path id="path4433" d="M 5.77,0.0 L -2.88,5.0 L -2.88,-5.0 L 5.77,0.0 z " style="fill-rule:evenodd;stroke-width:1pt;stroke-opacity:1;fill-opacity:1" transform="scale(0.4)" /></marker>'));
            this.node.appendChild(this.svg.node);
            this.svgG = this.svg.group();
            d3.select(this.svgG.node)
                .append('circle').attr('id', 'mmmLandscapeBall');
            d3.select(this.svgG.node)
                .append('circle').attr('id', 'mmmTstate');
            this.paths = [];
        }
        render(model, opts) {
            opts.landscape = opts;
            this.ls = qubx_model.modelToLandscape(model, opts);
            this.ls.states.forEach(st => {
                st.lsRates = [];
            });
            this.ls.rates.forEach(rate => {
                this.ls.states[rate.From].lsRates[rate.To] = rate;
            });
            let exX = expand(qubx.extrema(this.ls.states.map(qubx.m('x'))), 1.25);
            let exY = expand(qubx.extrema(this.ls.states
                                          .concat(this.ls.rates)
                                          .map(qubx.m('lsLevel'))),
                             1.25);
            let anyLabel = false;
            this.ls.states.forEach(st => {
                if ( st.Label ) {
                    anyLabel = true;
                }
            });
            const fontSize = anyLabel ? (opts.fontSize || 10) : 0,
                  h = this.height - 2.2*fontSize,
                  my = h / exY.d,
                  sharp = opts.sharp || 0.6;
            let xLo = 200, iLo;
            if ( ! opts.xFromModel ) {
                exX = expand({lo: 0, hi: this.ls.states.length-1}, 1.25);
            }
            const mx = this.width / exX.d;
            this.ls.states.forEach((st, s) => {
                st.lsX = mx * ((opts.xFromModel ? st.x : s) - exX.lo);
                st.lsY = my * (exY.hi - st.lsLevel);
                if ( st.x < xLo ) {
                    xLo = st.x;
                    iLo = st.Index;
                }
            });
            this.ls.rates.forEach(rate => {
                rate.lsY = my * (exY.hi - rate.lsLevel);
            });
            this.paths = [];
            depthFirstChains(iLo, i => this.ls.rates.filter(rate => (rate.From === i)).map(qubx.m('To')), chain => {
                let state = this.ls.states[chain[0]],
                    rate = state.lsRates[chain[1]],
                    nextState = this.ls.states[chain[1]],
                    dx = nextState.lsX - state.lsX,
                    sharpA = 0.5*dx*sharp;
                state.lsX += 0.25*sharpA;
                let pathStr = `M ${formatPoint(state.lsX, state.lsY)} C ${formatPoint(state.lsX, state.lsY+0.5*(rate.lsY - state.lsY))}`,
                    i = 1;
                for (; i<chain.length; ++i) {
                    rate = state.lsRates[chain[i]];
                    nextState = this.ls.states[chain[i]];
                    dx = nextState.lsX - state.lsX;
                    sharpA = 0.5*dx*sharp;
                    const sharpB = 0.5*dx*(1-sharp);
                    if ( i > 1 ) {
                        pathStr += ' '+formatPoint(state.lsX+sharpA, state.lsY);
                    }
                    pathStr += ' '+formatPoint(state.lsX+0.5*dx-sharpB, rate.lsY);
                    pathStr += ' '+formatPoint(state.lsX+0.5*dx, rate.lsY);
                    pathStr += ' '+formatPoint(state.lsX+0.5*dx+sharpB, rate.lsY);
                    if ( i < (chain.length-1) ) {
                        pathStr += ' '+formatPoint(nextState.lsX-sharpA, nextState.lsY);
                        pathStr += ' '+formatPoint(nextState.lsX, nextState.lsY);
                    } else {
                        nextState.lsX -= 0.25 * sharpB;
                        pathStr += ' '+formatPoint(nextState.lsX, nextState.lsY+0.5*(rate.lsY - nextState.lsY));
                        pathStr += ' '+formatPoint(nextState.lsX, nextState.lsY);
                    }
                    state = nextState;
                }
                var tstate;
                if ( opts.commitorRadius ) {
                    tstate = qubx_model_math.findCommitor(model, opts.tMeas, opts.stimulus);
                    let matching = tstate.stateOrder.length === chain.length;
                    matching = matching && (tstate.residual < .1);
                    matching = matching && tstate.stateOrder.map((s,i) => s === chain[i]).filter(qubx.i).length === chain.length;
                    if ( ! matching ) {
                        tstate = undefined;
                    }
                }
                this.paths.push({
                    p: pathStr,
                    t: tstate && tstate.commitor
                });
            });

            const lineColor = opts.stroke || 'black';
            const pathSel = d3.select(this.svgG.node)
                .selectAll('path')
                .data(this.paths);
            pathSel.enter().append('path')
                .attr({
                    id: 'mmmLandscape',
                    fill: 'none',
                    'marker-start': 'url(#TriangleInM)',
                    'marker-end': 'url(#TriangleOutM)'
                });
            pathSel.exit().remove();
            pathSel.attr('d', qubx.m('p'));
            pathSel.attr({
                stroke: lineColor,
                'stroke-width': opts.strokeWidth || 2.2
            });
            d3.select(this.svg.node)
                .selectAll('marker path')
                .attr({
                    fill: lineColor,
                    stroke: lineColor
                });

            let commitorCoord = undefined;
            pathSel.each(function(d,i) {
                if ( typeof(d.t) !== 'undefined' ) {
                    const minima = qubx_model_math.pathYminima(this).map(m => m.p);
                    if ( d.t < minima.length ) {
                        const sFrom = d.t|0,
                              pFrom = minima[sFrom],
                              pTo = minima[sFrom+1];
                        commitorCoord = this.getPointAtLength((d.t - sFrom) ? (pFrom + (d.t - sFrom)*(pTo-pFrom)) : pFrom);
                    }
                }
            });
            d3.select(this.svgG.node)
                .select('#mmmTstate')
                .attr({
                    cx: commitorCoord ? commitorCoord.x : 0,
                    cy: commitorCoord ? commitorCoord.y : 0,
                    r: opts.commitorRadius || 4,
                    fill: opts.lineColor,
                    opacity: commitorCoord ? 1 : 0
                });

            d3.select(this.svgG.node)
                .select('#mmmLandscapeBall')
                .attr({
                    cx: this.ls.states[0].lsX,
                    cy: this.ls.states[0].lsY + 1.2*fontSize,
                    r: opts.ballRadius || 2.2,
                    stroke: opts.ballStroke || '#990000',
                    'stroke-width': opts.ballStrokeWidth || 1,
                    fill: opts.ballFill || '#777700'
                });

            const stateY = Math.max(...this.ls.states.map((st, j) => st.lsY + fontSize*((j && (j < (this.ls.states.length-1))) ? 1.5 : 2.5)));
            const txtSel = d3.select(this.svgG.node)
                .selectAll('text')
                .data(anyLabel ? this.ls.states.map((st,j) => ({
                    x: st.lsX,
                    y: stateY,
                    lbl: st.Label,
                    id: 'stateLabel'+st.Index
                })) : []);
            const txtEnter = txtSel.enter().append('text');
            txtEnter.style({
                'font-weight': 900,
                'text-anchor': 'middle',
            });
            txtEnter.append('tspan');
            txtSel.select('tspan').text(qubx.m('lbl'));
            txtSel.attr('id', qubx.m('id'))
                .attr('x', qubx.m('x'))
                .attr('y', qubx.m('y'))
                .style({
                    'font-family': opts.textFont || 'sans-serif',
                    'font-size': `${fontSize|0}px`,
                    fill: opts.textFill
                });
            txtSel.exit().remove();
        }
    };
    const EnergyLandscapeVisDefs = {
        kRef: 1000000,
        fontSize: 10,
        sharp: 0.6,
        stroke: '#000000',
        strokeWidth: 2.2,
        ballRadius: 2.2,
        ballStroke: '#990000',
        ballFill: '#777700',
        ballStrokeWidth: 1,
        textFill: '#000000',
        xFromModel: true,
        textFont: 'Arial, sans-serif',
        tMeas: 100,
        commitorRadius: 4
    };
    exports.EnergyLandscapeVis = class EnergyLandscapeVis {
        constructor(config) {
            this.landscape = new exports.EnergyLandscape2D(config);
            this.node = this.landscape.node;
            this.saver = config && config.saver || new qubx_ui.Saver();
            const floater = d3.select(this.node).append('div');
            floater.classed('qubx_landscape_float', true);
            floater.append('button')
                .attr('title', 'Save')
                .on('click', this.onClickDownload.bind(this))
                .append('img').attr({
                    src: 'images/dl.svg',
                    alt: '[Save]'
                });
            floater.append('button')
                .attr('title', 'About')
                .text('?')
                .on('click', () => this.about.run());
            floater.append('button')
                .attr('title', 'Settings')
                .on('click', () => this.settings.run()
                    .catch(() => {} )
                    .then(() => {
                        this.renderSoon();
                        localStorage.setItem('qubx_landscape_opts', JSON.stringify(this.opts));
                    }))
                .append('img').attr({
                    src: 'images/appbar.settings.svg',
                    alt: '[Settings]'
                });
            this.about = new qubx_ui.Modal({
                title: 'Energy Landscape',
                template: '#qubx_landscape_about',
                canCancel: false,
                okLabel: 'Close'
            });
            this.settings = new qubx_ui.Modal({
                title: 'Energy Landscape Settings',
                template: "#qubx_landscape_settings",
                canCancel: false,
                okLabel: 'Close'
            });
            this.opts = {};
            try {
                this.opts = JSON.parse(localStorage.getItem('qubx_landscape_opts'));
            } catch (err) {}
            this.opts = Object.assign({}, EnergyLandscapeVisDefs, this.opts);
            this.settings.dom.then(dom => {
                const setInOpts = (row, col, val) => {
                    this.opts[row.field] = val;
                    this.renderSoon();
                };
                this.settings.numTbl = new qubx_ui.ColumnTable({
                    node: dom.content.querySelector('#qubx_landscape_numOpts'),
                    columns: [
                        qubx_ui.ColumnTable.htmlCol('', 'label'),
                        qubx_ui.ColumnTable.inputCol('', 'value', 'number', qubx.acceptFloat, qubx.formatK0, setInOpts)
                    ]
                });
                this.settings.numTbl.render([{
                    field: 'kRef',
                    label: 'k<sub>ref</sub>:'
                }, {
                    field: 'fontSize',
                    label: 'Font size:'
                }, {
                    field: 'sharp',
                    label: 'Peak sharpness:'
                }, {
                    field: 'strokeWidth',
                    label: 'Line width:'
                }, {
                    field: 'ballRadius',
                    label: 'Ball radius:'
                }, {
                    field: 'ballStrokeWidth',
                    label: 'Ball line width:'
                }, {
                    field: 'commitorRadius',
                    label: 'Committor radius:'
                }, {
                    field: 'tMeas',
                    label: 'Committor t<sub>meas</sub>:'
                }].map(opt => {
                    opt.value = this.opts[opt.field];
                    return opt;
                }));
                this.settings.txtTbl = new qubx_ui.ColumnTable({
                    node: dom.content.querySelector('#qubx_landscape_txtOpts'),
                    columns: [
                        qubx_ui.ColumnTable.htmlCol('', 'label'),
                        qubx_ui.ColumnTable.inputCol('', 'value', 'text', qubx.acceptStr, qubx.formatStr, setInOpts)
                    ]
                });
                this.settings.txtTbl.render([{
                    field: 'stroke',
                    label: 'Line color:'
                }, {
                    field: 'textFill',
                    label: 'Text color:'
                }, {
                    field: 'textFont',
                    label: 'Font:'
                }, {
                    field: 'ballStroke',
                    label: 'Ball outline:'
                }, {
                    field: 'ballFill',
                    label: 'Ball fill:'
                }].map(opt => {
                    opt.value = this.opts[opt.field];
                    return opt;
                }));
                this.settings.chkTbl = new qubx_ui.ColumnTable({
                    node: dom.content.querySelector('#qubx_landscape_chkOpts'),
                    columns: [
                        qubx_ui.ColumnTable.htmlCol('', 'label'),
                        qubx_ui.ColumnTable.chkCol('', 'value', setInOpts)
                    ]
                });
                this.settings.chkTbl.render([{
                    field: 'xFromModel',
                    label: "use model's x coordinates",
                    value: this.opts.xFromModel
                }]);
            });
            qubx_ui.onResize(this.landscape.node, () => {
                const newW = this.landscape.node.clientWidth,
                      newH = this.landscape.node.clientHeight,
                      aspect = newH ? (newW / newH) : 1.0;
                this.landscape.width = aspect*this.landscape.height;
                this.landscape.svg.attr({
                    viewBox: `0 0 ${this.landscape.width} ${this.landscape.height}`,
                    width: newW,
                    height: newH
                });
                this.onChangeSomething();
            });
            const mv = config.modelView;
            this.renderSoon = qubx.batchDelayed(() => {
                this.opts.stimulus = mv.dataStim;
                this.landscape.render(mv.model, Object.assign({Q: (mv.info || mv.active[0]).Q}, this.opts));
            }, 120, true);
            ['setState', 'insertRate', 'removeRate', 'setRate'].forEach(ename => {
                mv.models.ev[ename].sub(this.onChangeSomething.bind(this));
            });
            this.renderSoon();
        }
        onClickDownload() {
            try {
                this.saver.prompt({name: "Landscape.svg"})
                    .then(fi => this.saver.saveText(fi.fullname, this.asSVG().outerHTML, "image/svg+xml"))
                    .catch(qubx.reportErr);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        asSVG() {
            const svg = this.landscape.svg.node.cloneNode(true);
            svg.setAttribute("xmlns", "http://www.w3.org/2000/svg");
            svg.setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
            svg.setAttribute("version", "1.1");
            return svg;
        }
        onShow(showing) {
            this.showing = showing;
            showing && this.renderSoon();
        }
        onChangeSomething() {
            this.showing && this.renderSoon();
        }
    };

    function depthFirstChains(head, getChildren, onChain, visited, chained) {
        chained = chained || [];
        chained.push(head);
        visited = visited || [];
        visited.push(head);
        const children = getChildren(head).filter(ch => visited.indexOf(ch) < 0);
        if ( ! children.length ) {
            onChain(chained);
        } else {
            depthFirstChains(children[0], getChildren, onChain, visited, chained);
            let i = 1;
            for (; i<children.length; ++i) {
                depthFirstChains(children[i], getChildren, onChain, visited, [head]);
            }
        }
    }
    
    return exports;
});
