/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_worker_util", [], () => {
    'use strict';

    const exports = {};

    let messageConfig = {},
        messageListening = false;
    const handleMessage = data => {
        const f = messageConfig[data.verb];
        if ( f ) {
            if ( messageConfig.trace ) {
                console.log(data.verb,data.tkt,data,'in',location.pathname);
            }
            var result;
            try {
                result = f(data, progress => postMessage({
                    t: 'PROGRESS',
                    tkt: data.tkt,
                    progress: progress
                }));
                result = (result && result.then) ? result : Promise.resolve(result || data);
            } catch (err) {
                result = Promise.reject(err);
            }
            result.then(result => {
                if ( messageConfig.trace ) {
                    console.log(data.verb,data.tkt,'->',result);
                }
                if ( data.tkt ) {
                    result.tkt = data.tkt;
                    postMessage(result, result.transfer||[]);
                }
            }).catch(err => {
                if ( data.tkt ) {
                    data.tktErr = {
                        message: err.message,
                        stack: err.stack
                    };
                    postMessage(data, data.transfer||[]);
                } else {
                    console.log(err);
                }
            });
        } else {
            console.log('unregistered verb:',data.verb,data);
        }
    };
    const messageListener = event => handleMessage(event.data);
    
    exports.listen = function listen(config) {
        messageConfig = config || {};
        if ( messageListening ) {
            if ( messageConfig.removeListener ) {
                self.removeEventListener('message', messageListener);
                messageListening = false;
                return;
            }
        } else {
            self.addEventListener('message', messageListener);
            messageListening = true;
            if ( config.ready || (typeof(config.ready) === 'undefined') ) {
                postMessage({t: 'READY'});
            }
        }
        if ( config.trace && config.messageQueue && config.messageQueue.length ) {
            console.log(config.messageQueue.length + ' early messages for '+location.pathname);
        }
        config.messageQueue && config.messageQueue.forEach(handleMessage);
    };

    exports.getStimulus = function getStimulus(models) {
        const stimulus = {};
        models.forEach(model => {
            if ( model ) {
                model.rates.forEach(r => {
                    if ( r.Ligand ) {
                        stimulus[r.Ligand] = 1.0;
                    }
                    if ( r.Voltage ) {
                        stimulus[r.Voltage] = 0.0;
                    }
                    if ( r.Pressure ) {
                        stimulus[r.Pressure] = 0.0;
                    }
                });
                if ( model.IeqFv ) {
                    stimulus['Voltage'] = 0.0;
                }
            }
        });
        return stimulus;
    };
    
    exports.SignalMap = class SignalMap {
        constructor() {
            this.names = [];
            this.src = [];
            this.indexMap = {};
            this.defaults = [];
            this.delays = [];
        }
        indexOf(name) {
            const i = this.indexMap[name];
            if ( typeof(i) === 'undefined' ) {
                return 0;
            }
            return i;
        }
        update(models, layout) {
            if ( models && (typeof(models.length) === 'undefined') ) {
                models = [models];
            }
            let lastSrc = this.src,
                lastDefs = this.defaults,
                lastDelays = this.delays;
            this.names = [];
            this.src = [];
            this.indexMap = {};
            this.defaults = [];
            this.delays = [];
            
            //console.log('u');
            if ( models.length && layout && layout.signalNames.length ) {
                this.names[0] = layout.signalNames[0];
                this.src[0] = 0;
                this.defaults[0] = 0.0;
                this.delays[0] = 0.0;
                this.indexMap[this.names[0]] = 0;
                const modelStim = exports.getStimulus(models);
                layout.signalNames.slice(1).forEach((name,i) => {
                    if ( name in modelStim ) {
                        this.names.push(name);
                        this.indexMap[name] = this.src.length;
                        this.src.push(i+1);
                        this.defaults.push(0.0);
                        this.delays.push(0.0);
                        delete modelStim[name];
                    }
                });
                for (const name of Object.keys(modelStim)) {
                    this.names.push(name);
                    this.defaults.push(layout.defaults && layout.defaults[name] || 0.0);
                }
            }
            
            if ( (lastSrc.length !== this.src.length) || (lastDefs.length !== this.defaults.length) ) {
                return true;
            }
            let i = 0;
            for (; i<lastSrc.length; ++i) {
                if ( (lastSrc[i] !== this.src[i]) || (lastDelays[i] !== this.delays[i]) ) {
                    return true;
                }
            }
            for (i=0; i<lastDefs.length; ++i) {
                if ( lastDefs[i] !== this.defaults[i] ) {
                    return true;
                }
            }
            return false; // all signals the same as before
        }
    };

    return exports;
});
