/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_xmeans_math", ["random"], Random => {
    'use strict';

    const engine = Random.engines.mt19937().autoSeed();
    
    class Calc{
        constructor() {
            this.columns = [];
            this.weight = [];
            this.colorWeights = [];
            this.colorWeightK = 0;
        }
        ColorWeight(color, K) {
            if ( K !== this.colorWeightK ) {
                this.colorWeightK = K;
                this.colorWeights = [];
            }
            let c = 0;
            for (; c<=K; c++) {
                this.colorWeights[c] = 0.0;
            }
            const n = color.length;
            let i = 0;
            for (; i<n; ++i) {
                c = color[i];
                if ( c ) {
                    this.colorWeights[c] += this.weight[i];
                }
            }
            return this.colorWeights;
        }
        WeightedSumInColor(column, color, k) {
            const N = column.length;
            let sum = 0.0,
                i = 0;
            for (; i<N; ++i) {
                if ( color[i] === k ) {
                    sum += column[i] * this.weight[i];
                }
            }
            return sum;
        }
        WeightedSSDInColor(column, color, k, mean) {
            const N = column.length;
            let sum = 0.0,
                i = 0;
            for (; i<N; ++i) {
                if ( color[i] === k ) {
                    const dev = column[i] - mean;
                    sum += dev*dev * this.weight[i];
                }
            }
            return sum;
        }
        Means(color, colorWeight, K) {
            return this.columns.map(data => colorWeight.map((cw, k) => {
                return cw ? (this.WeightedSumInColor(data, color, k) / cw) : 0.0;
            }));
        }
        Stds(color, colorWeight, K, means) {
            return this.columns.map((data, j) => colorWeight.map((cw, k) => {
                return cw ? Math.sqrt(this.WeightedSSDInColor(data, color, k, means[j][k]) / cw) : 0.0;
            }));
        }
        Measure(means, measures, K) {
            const M = this.columns.length,
                  n = measures.length;
            let i = 0;
            for (; i<n; i++) {
                let k = 1;
                for (; k<=K; k++) {
                    let sum = 0.0,
                        j = 0;
                    for (; j<M; j++) {
                        const dev = (this.columns[j][i] - means[j][k]);
                        sum += dev * dev;
                    }
                    measures[i][k] = sum;
                }
            }
        }
        Score(measures, color) {
            const n = measures.length;
            let score = 0.0,
                i = 0;
            for (; i<n; ++i) {
                const c = color[i];
                if ( c ) {
                    score += measures[i][c] * this.weight[i];
                }
            }
            return score;
        }
        PickColors(meas) {
            const n = meas.length,
                  nk = n && meas[0].length,
                  colors = new Int32Array(n);
            let i = 0;
            for (; i<n; ++i) {
                const meas_of_i = meas[i];
                let lo = meas_of_i[1],
                    lo_k = 1,
                    k = 2;
                for (; k<nk; k++) {
                    const m = meas_of_i[k];
                    if ( m < lo ) {
                        lo = m;
                        lo_k = k;
                    }
                }
                colors[i] = lo_k;
            }
            return colors;
        }
        RandomColors(k, n) {
            const color = [],
                  dist = Random.integer(1, k);
            let i = 0;
            for (; i<n; ++i) {
                color.push(dist(engine));
            }
            return color;
        }
        CanonicalColors(input) {
            let next_i = 1;
            const renum = input.map(() => -1);
            input.forEach(x => {
                if ( x && (renum[x] < 0) ) {
                    renum[x] = next_i;
                    ++next_i;
                }
            });
            return input.map(x => x ? renum[x] : 0);
        }
        Iterate(k, color, iterations) {
            iterations = iterations|0;
            const m = this.columns.length;
            let n = 0;
            if ( (k > 0) && m && this.columns[0].length ) {
                n = this.columns[0].length;
                const measures = [];
                let i = 0;
                for (; i<n; ++i) {
                    measures[i] = [];
                }
                let colorWeight = this.ColorWeight(color, k),
                      means = this.Means(color, colorWeight, k);
                this.Measure(means, measures, k);
                let lastSSD = this.Score(measures, color),
                    lastColor = color,
                    t = 0;
                for (; t<iterations; ++t) {
                    color = this.PickColors(measures);
                    const ssd = this.Score(measures, color);
                    if ( t && (ssd >= lastSSD) ) {
                        break;
                    }
                    lastColor = color;
                    lastSSD = ssd;
                    colorWeight = this.ColorWeight(lastColor, k);
                    means = this.Means(lastColor, colorWeight, k);
                    this.Measure(means, measures, k);
                }
                
                if ( iterations === 0 ) {
                    color.forEach(c => {
                        if ( c === 0 ) {
                            --n;
                        }
                    });
                }
                
                const nf = 3*k*m;
                var aic, aicc;
                if ( lastSSD )
                    aic = (n * Math.log(lastSSD / n) + 2*nf);
                else
                    aic = 0.0;
                if (n-nf-1)
                    aicc = 2*nf*(nf+1)*(1.0/(n-nf-1)) + aic;
                else
                    aicc = (n * Math.log(lastSSD / n) + 2*nf);
                
                lastColor = this.CanonicalColors(lastColor);
                colorWeight = this.ColorWeight(lastColor, k);
                return {
                    k: k,
                    color: lastColor,
                    ssd: lastSSD,
                    aicc: aicc,
                    mean: this.Means(lastColor, colorWeight, k),
                    std: this.Stds(lastColor, colorWeight, k, means)
                };
            }
            return null;
        }
        TryK(k, trials, iterations) {
            let best = {
                k: k,
                ssd: 1e20,
                aicc: -1e20,
                color: [],
                mean: [],
                std: []
            };
            if ( (k > 0) && this.columns.length && this.columns[0].length ) {
                const n = this.columns[0].length;
                let trial = 0;
                for (; trial<trials; ++trial) {
                    const last = this.Iterate(k, this.RandomColors(k, n), iterations);
                    if ( last.ssd < best.ssd ) {
                        best = last;
                    }
                }
            }
            return best;
        }
    };
    
    return {
        Calc: Calc
    };
});
